//
//  graph_utilities.c
//  
//
//  Created by Ivaylo Dobrikov on 06/02/14.
//
//

#include "graph_utilities.h"

#include <stdio.h>
#include <assert.h>

static int open_vertex_mallocs = 0;
static int open_edge_mallocs = 0;

void init_mallocs_graph_counter() {
    open_vertex_mallocs = 0;
    open_edge_mallocs = 0;
}

/* debugging functions for printing the vertices and the edges of a graph */
#ifdef FAIRNESSDEBUG
static void print_vertices(vertex_list vertices) {
    vertex_list currrent_node = vertices;
    GraphVertex v;
    printf("Vertices list: ");
    while (currrent_node != NULL) {
        v = currrent_node->vertex;
        printf("Vertex ( {%d,%d} ), ", v.atomid, v.graphid);
        currrent_node = currrent_node->next;
    }
    printf("\n");
}

static void print_edges(edge_list edges) {
    GraphEdge current_edge;
    edge_list currrent_node = edges;
    printf("Edges list: ");
    while (currrent_node != NULL) {
        current_edge = currrent_node->edge;
        printf("( {%d,%d} --> {%d,%d} ), ", current_edge.src.atomid, current_edge.src.graphid, current_edge.dst.atomid, current_edge.dst.graphid);
        currrent_node = currrent_node->next;
    }
    printf("\n");
}

#endif

/* functions for freeing the allocated memory for the Graph type */
static void free_vertex_list(vertex_list head) {
    vertex_list node = head;
    vertex_list temp;
    while (node != NULL) {
        temp = node;
        node = node->next;
        open_vertex_mallocs--;
        ltl_free(temp);
    }
}

static void free_edge_list(edge_list head) {
    edge_list node = head;
    edge_list temp;
    while (node != NULL) {
        temp = node;
        node = node->next;
        open_edge_mallocs--;
        ltl_free(temp);
    }
}

void delete_graph_data(Graph* g) {
    free_vertex_list(g->vertices);
    free_edge_list(g->edges);
}

void delete_scc_components_data(Graph* scc_components, int num) {
    int i;
    for (i=0; i<num; i++) {
        delete_graph_data(&scc_components[i]);
    }
    ltl_free(scc_components);
}

void assert_open_mallocs_graph_are_zero() {
    assert(open_vertex_mallocs == 0);
    assert(open_edge_mallocs == 0);
}

/* further utility functions */
static int is_not_in_list(int atomid, int* atomids, int len) {
    int i;
    for (i=0; i<len; i++) {
        if (atomids[i] == atomid) {
            return 0;
        }
    }
    return 1;
}

static int vertex_in_list(int atomid, vertex_list head) {
    GraphVertex v;
    vertex_list currrent_node = head;
    
    while (currrent_node != NULL) {
        v = currrent_node->vertex;
        if (v.atomid == atomid) {
            return v.graphid;
        }
        currrent_node = currrent_node->next;
    }
    return -1;
}

// inserting a node uniquely
vertex_list insert_vertex(vertex_list head, GraphVertex v) {
    
    // insert a new vertex node to the node's list as the new head of the list
    vertex_list new_node = (vertex_list) ltl_malloc(sizeof(struct vertex_node));
    new_node->vertex = v;
    new_node->next = head;
    open_vertex_mallocs++;
    
    return new_node;
}

// inserting a node uniquely
static edge_list insert_edge(edge_list head, GraphEdge e) {
    
    // insert a new vertex node to the node's list as the new head of the list
    edge_list new_node = (edge_list) ltl_malloc(sizeof(struct edge_node));
    new_node->edge = e;
    new_node->next = head;
    open_edge_mallocs++;
    
    return new_node;
}

static GraphEdge create_edge(GraphVertex a, GraphVertex b) {
    GraphEdge e;
    e.src = a; e.dst = b;
    return e;
}

static GraphVertex create_vertex(int atomid, int graphid) {
    GraphVertex v;
    v.atomid = atomid; v.graphid = graphid;
    return v;
}

int empty_graph(Graph g) {
    if (g.vertices == NULL) {
        return 1;
    }
    return 0;
}

/*
 Read the SCC from the saved tarjan data t and save it into a variable of type Graph.
 All vertices with an atom id from the "atomids" (the impartial vertices) array will
 not be saved to the expected graph. The deleting of the imparial vertices happens thus
 in the process of initializing the graph of the given SCC.
 */
Graph get_scc_graph(tarjandata *t, int *atomids, int *len, int scc_size) {
    
    ltl_formula *f = t->f;
    
    int i;
    int atomid,succid;
    int enumerate = 0;
    int vertex_index;
    
    edge_list edges = NULL;
    vertex_list vertices = NULL;
    GraphVertex v_atom, v_succ;
    
    for (i=scc_size; i>0; i--) {
        
        ltlatom *atom = peek_stack(t,i-1);
        atomid = find_atom_id(f,atom);
        
        // adding only impartial nodes to the SCC graph
        if (is_not_in_list(atomid, atomids, *len)) {
            
            ltlatom *succ;
            unsigned int first_tid = atom->first_transition_id;
            int nbr_transitions = atom->number_of_transitions;
            
            // adding uniquely a vertex to the vertices list
            vertex_index = vertex_in_list(atomid,vertices);
            if (vertex_index == -1 ) {
                v_atom = create_vertex(atomid,enumerate);
                vertices = insert_vertex(vertices,v_atom);
                enumerate++;
            } else {
                v_atom = create_vertex(atomid,vertex_index);
            }
            
            int j;
            for (j=0; j<nbr_transitions; j++) {
                
                succ = get_atom_trans(f,first_tid+j);
                succid = find_atom_id(f,succ);
                
                // adding only edges to impartial nodes
                if (is_not_in_list(succid, atomids, *len)) {
                    
                    vertex_index = vertex_in_list(succid,vertices);
                    if (vertex_index == -1) {
                        v_succ = create_vertex(succid,enumerate);
                        vertices = insert_vertex(vertices,v_succ);
                        enumerate++;
                    } else {
                        v_succ = create_vertex(succid,vertex_index);
                    }
                    
                    GraphEdge e = create_edge(v_atom, v_succ);
                    edges = insert_edge(edges,e);
                }
            }
        }
    }
    IFFAIRNESSDEBUG(print_vertices(vertices);print_edges(edges));
    
    // initialize the SCC graph
    Graph g;
    g.vertices = vertices;
    g.edges = edges;
    
    return g;
}
