
#include <stdio.h>
#include <stdint.h>
#include "ltlarray.h"
#include "ltlmem.h"

#ifndef DEBUG
#define DEBUG 0
#endif

#ifndef LTL_MEMALIGNMENT
#define LTL_MEMALIGNMENT 8
#endif

static const int reg_size_step = 64;

void init_array(ltlarray *st, size_t unitlength, initfunction init, const char *name) {
  if (DEBUG) printf("ltlarray: init_array called for %s.\n", name);
  /* round the size of an unit up to the memory alignment */
  if (unitlength==0) {
    fprintf(stderr, "fatal error: unitlength for arrays must be >0, stopping.\n");
    exit(1);
  }
  st->unit = ((unitlength-1) / LTL_MEMALIGNMENT + 1) * LTL_MEMALIGNMENT;
  st->size = 0;
  st->allocated = 0;
  st->regsize = 0;
  st->reg = 0;
  st->init = init;
  st->name = name;
}

void delete_array(ltlarray *st) {
  unsigned int i,allocated;
  void **reg;

  if (DEBUG>0) printf("ltlarray: delete_array called for %s\n", st->name);

  allocated = st->allocated;
  reg = st->reg;
  for (i=0; i<allocated; i++) {
    ltl_free(reg[i]);
  }
  ltl_free(reg);

  st->size = 0;
  st->allocated = 0;
  st->regsize = 0;
  st->reg = 0;
}

void *get_array_element(ltlarray *st, unsigned int id) {
  unsigned int pageno,index;
  void *result;
  if (DEBUG>2) printf("ltlarray: get_array_element(%s,%u) called.\n", st->name, id);
  pageno = id >> array_pagesize_log2;
  if (pageno < st->allocated) {
    if (DEBUG>2) printf("ltlarray:   pageno %u found.\n", pageno);
    index = id & (array_pagesize_abs - 1);
    result = (void *) ( ((uintptr_t) (st->reg)[pageno]) + (index * st->unit) );
  } else {
    fprintf(stderr,"error: array (%s) range exceeded.\n", st->name);
    result = 0;
  }
  return result;
}

int assure_arraysize(ltlarray *st, unsigned int size) {
  int ok = 1;
  unsigned int i,j,dif,newreg;
  uintptr_t elem;
  void *page;
  if (DEBUG>2) printf("ltlarray: assure_arraysize(%s,%u) called.\n", st->name, size); fflush(stdout);
  if (size > st->size) {
    dif = ((size - st->size - 1) >> array_pagesize_log2) + 1;
    if (DEBUG>0) printf("ltlarray:   current size (%u, %u pages) for %s exceeded, allocate %i new pages.\n",
                        st->size, st->allocated, st->name, dif);
    newreg = (st->allocated + dif) - st->regsize;
    if (newreg > 0) {
      newreg += reg_size_step;
      if (DEBUG>0) printf("ltlarray:   current regsize (%u) exceeded, allocate %u new entries.\n", 
                          st->regsize, newreg);
      st->reg = (void **) ltl_realloc(st->reg, sizeof(void *) * (st->allocated+newreg));
      st->regsize += newreg;
      if (!st->reg) {
        fprintf(stderr, "ltlarray: realloc(reg) returns NULL.\n");
        ok = 0;
      }
    }
    for (i=0; i<dif && ok; i++) {
      page = ltl_malloc(st->unit * array_pagesize_abs);
      (st->reg)[st->allocated] = page;
      if (page != 0) {
        st->allocated++;
        st->size = st->allocated * array_pagesize_abs;
        if (st->init != 0) {
          if (DEBUG>1) printf("ltlarray:   a_as: calling init function.\n");
          elem = (uintptr_t) page;
          for (j=0; j<array_pagesize_abs; j++) {
            (*st->init)((void *) elem);
            /* jump to the next element */
            elem += st->unit;
          }
        }
      } else {
        fprintf(stderr, "malloc() returned NULL in assure_arraysize().\n");
      }
      ok = st->reg != 0;
    }
  } else {
    if (DEBUG>2) printf("ltlarray:   current size (%u) ok.\n", st->size);
    ok = 1;
  }
  return ok;
}

size_t ltl_array_memory_usage(ltlarray *st) {
  size_t mem = sizeof(ltlarray);
  mem += st->unit * array_pagesize_abs * st->allocated;
  mem += st->regsize * sizeof(void*);
  return mem;
}
