#include <stdio.h>
#include <stdint.h>
#include <sicstus/sicstus.h>

#include "ltlc.h"
#include "time.h"

static SP_term_ref maketerm(ltlnode *node);
static int process_outgoing(ltl_formula *formula, ltlnode *node, SP_term_ref list, int *number_of_transevals);
static int create_atoms(ltl_formula *formula, ltlnode *node, SP_term_ref atomlist, int is_start_node, int nr_diff_trans);
static void save_nodetrans(ltl_formula *f, ltlnode *target, SP_term_ref op, bitfield **eval, unsigned int *transid);
static void ltlevaluate(ltl_formula *formula, bitfield *evaluation);
static int atoms_match(ltl_formula *formula, bitfield *a, bitfield *b);
static void eval_props(int size, bitfield *evaluation, SP_term_ref atomlist);
static void init_untils(ltl_formula *formula);

void init_formula(ltl_formula *f) {
  int i;

  f->unexplored_nodes = 0;

  /* initialise timer to check when to update the prolog side */
  f->last_update_call = clock();
  f->canceled = 0;
  f->update_countdown = update_countdown_init;

  bitoperations_init();

  init_nodestore(&f->nstore);

  f->tmp_evaluation = (bitfield *) ltl_malloc(get_bitfield_memsize(f->closure_size));
  f->tmp_evaluation2 = (bitfield *) ltl_malloc(get_bitfield_memsize(f->closure_size));

  init_array(&f->nodetrans, sizeof(ltlnodetrans) + get_bitfield_memsize(f->number_of_tps), 0, "nodetrans");
  f->nodetrans_number = 0;

  IFLTLDEBUG(ltlmsg("closure_size=%i\n", f->closure_size););

  init_array(&f->astore, sizeof(ltlatom) + get_bitfield_memsize(f->closure_size), 0, "atoms");
  f->number_of_atoms = 0;

  init_array(&f->atomtransstore, sizeof(ltlatom*), 0, "atomtrans");
  f->number_of_atomtrans = 0;

  f->number_of_equal_evaluations = 2*f->number_of_nexts + 2*f->number_of_yesterdays;
  f->equal_evaluations = (equal_evaluation_positions*) ltl_malloc(sizeof(equal_evaluation_positions)*
                                                                  f->number_of_equal_evaluations);

  IFLTLDEBUG(ltlmsg("init equal_evaluations\n");
             ltldeb_inc();)
    
  for (i=0; i<f->number_of_equal_evaluations; i++) {
    int pos = f->index_nexts + i;
    switch (f->closure[pos].op) {
    case ltlnext:
      f->equal_evaluations[i].a = pos;
      f->equal_evaluations[i].b = f->closure[pos].args.case_next;
      f->equal_evaluations[i].signum = 1;
      break;
    case ltlnextnot:
      f->equal_evaluations[i].a = pos;
      f->equal_evaluations[i].b = f->closure[pos].args.case_nextnot;
      f->equal_evaluations[i].signum = 0;
      break;
    case ltlyesterday:
      f->equal_evaluations[i].a = f->closure[pos].args.case_yesterday;
      f->equal_evaluations[i].b = pos;
      f->equal_evaluations[i].signum = 1;
      break;
    case ltlyesterdaynot:
      f->equal_evaluations[i].a = f->closure[pos].args.case_yesterdaynot;
      f->equal_evaluations[i].b = pos;
      f->equal_evaluations[i].signum = 0;
      break;
    default:
      fprintf(stderr,"fatal: next or nextnot expected in initialisation of equal_evaluations");
      exit(1);
    }

    IFLTLDEBUG(ltlmsg("%i equals %s%i\n",
                      f->equal_evaluations[i].a,
                      f->equal_evaluations[i].signum ? "" : "-",
                      f->equal_evaluations[i].b); );
  }

  init_array(&f->tmp_transeval_store, sizeof(bitfield*), 0, "tmp_transeval");
  if (f->number_of_tps > 0) {
    f->zero_transeval = (bitfield*) ltl_malloc(get_bitfield_memsize(f->number_of_tps));
    clear_bitfield(f->zero_transeval, f->number_of_tps);
  } else {
    f->zero_transeval = 0;
  }

  /* initialise number_of_untils and untils */
  init_untils(f);

  IFLTLDEBUG( ltldeb_dec(); )
}

static void init_untils(ltl_formula *f) {
  int i,j;
  f->number_of_untils = 0;
  for (i=f->index_generics; i<f->closure_size; i++) {
    if (f->closure[i].op == ltluntil) {
      f->number_of_untils++;
    }
  }
  if (f->number_of_untils > 0) {
    f->untils = (unsigned int*) ltl_malloc(sizeof(unsigned int) * f->number_of_untils);
    if (f->untils == NULL) {
      fprintf(stderr, "malloc() returned NULL for f->untils (size %zu)\n", sizeof(unsigned int) * f->number_of_untils);
      exit(1);
    }
    j=0;
    for (i=f->index_generics; i<f->closure_size; i++) {
      if (f->closure[i].op == ltluntil) {
        f->untils[j] = i;
        j++;
      }
    }
  } else {
    f->untils = 0;
  }
}

void delete_formula(ltl_formula *f)
{
  delete_nodestore(&f->nstore);
  delete_array(&f->nodetrans);
  f->nodetrans_number = 0;
  delete_array(&f->astore);
  f->number_of_atoms = 0;
  delete_array(&f->atomtransstore);
  f->number_of_atomtrans = 0;
  f->number_of_equal_evaluations = 0;
  ltl_free(f->equal_evaluations);
  f->equal_evaluations=0;
  ltl_free(f->tmp_evaluation);
  ltl_free(f->tmp_evaluation2);
  delete_array(&f->tmp_transeval_store);
  f->number_of_untils = 0;
  if (f->untils) ltl_free(f->untils);
  if (f->zero_transeval) ltl_free(f->zero_transeval);
  if (f->closure) ltl_free(f->closure);

  bitoperations_delete();
}

static void emergency_break(const char *msg) {
  fprintf(stderr,"%s\n",msg);
  exit(1);
}

void ltl_reraise_exception() {
  SP_term_ref exception = SP_new_term_ref();
  SP_exception_term(exception);
  SP_raise_exception(exception);
}

int explore_node(ltl_formula *formula, ltlnode *node, int is_start_node) {
  IFLTLDEBUG(ltlmsg("explore node ");ltl_printnode(node);printf("\n");
             ltldeb_inc(););

  if (node->number_of_atoms == node_not_opened) {
    SP_term_ref ref = maketerm(node);
    SP_term_ref aps_term = SP_new_term_ref(); SP_put_variable(aps_term);
    SP_term_ref outgoing_term = SP_new_term_ref(); SP_put_variable(outgoing_term);
    int nr_diff_trans,status;
    clock_t now;

    IFLTLTIMING ( clock_t timer_start = clock(); );
    status = SP_query(formula->callback_predicate, 
                      ref,
                      aps_term, outgoing_term,
                      formula->ap_list, formula->tp_list);
    now = clock();
    IFLTLTIMING( formula->callback_time += (now - timer_start);
                 formula->callback_calls += 1; );
    if (status == SP_SUCCESS) {
      /* IFLTLDEBUG(ltlmsg("call succeeded, process outgoing\n");); */
      process_outgoing(formula, node, outgoing_term, &nr_diff_trans);
      create_atoms(formula, node, aps_term, is_start_node, nr_diff_trans);
    } else if (status == SP_FAILURE) {
      IFLTLDEBUG( ltlmsg("callback failed, node is not explorable.\n"); );
      node->number_of_atoms = node_unexplorable;
      formula->unexplored_nodes += 1;
    } else {
      IFLTLDEBUG( ltlmsg("callback raised exception!\n");
                  ltldeb_dec(); );
      ltl_reraise_exception();
      return 0;
    }

    if (now - formula->last_update_call >= ltl_update_interval) {
      /* force update */
      formula->update_countdown = -1;
    }
    
  }
  IFLTLDEBUG( ltldeb_dec(); );
  return 1;
}

void get_nodetransop(ltl_formula *f, unsigned int id, ltlnode **dest, bitfield **evaluation, SP_term_ref *op) {
  void *mempos = get_array_element(&f->nodetrans, id);
  if (dest!=NULL) {
    *dest = ((ltlnodetrans*) mempos)->dest;
  }
  if (evaluation!=NULL) {
    *evaluation = (bitfield*) (((uintptr_t) mempos) + sizeof(ltlnodetrans));
  }
  if (op!=NULL) {
    *op = ((ltlnodetrans*) mempos)->operation;
  }
}

static int transition_match(ltl_formula *f, bitfield *atomevaluation, bitfield *transevaluation) {
  int i;
  for (i=0; i<f->number_of_tps; i++) {
    if (getbit(transevaluation, i) != getbit(atomevaluation, f->index_tps + i))
      return 0;
  }
  return 1;
}

/*
 * A helper function to compare two SP_integers. Needed for qsort().
 */
int sp_int_cmp(const void *va, const void *vb) {
  const SP_integer a = *((const SP_integer *) va);
  const SP_integer b = *((const SP_integer *) vb);
  return a<b ? -1 : (a>b ? 1 : 0);
}

int transition_in_array(SP_term_ref op, SP_integer *matches, int nrmatches) {
  SP_integer transid;
  if (SP_is_integer(op)) {
    SP_get_integer(op,&transid);
    if (bsearch(&transid, matches, nrmatches, sizeof(SP_integer), sp_int_cmp)!=0) {
      return 1;
    }
  }
  return 0;
}

/*
 * Search for a transition between two atoms a and b by looking at the
 * transitions between their nodes and testing each transition if it
 *  a) matches all transition propositions in a (by using transition_match())
 *  b) (if nrmatches>0) the transition is in the sorted array "matches"
 * If a transition is found, 1 is returned and *found set.
 * If not, 0 is returned, found keeps unmodified.
 */
int find_transition_operation_that_matches(ltl_formula *f, ltlatom *a, ltlatom *b, SP_integer *matches, int nrmatches, SP_term_ref *found) {
  if (a==0 || b==0) {
    SP_term_ref result = SP_new_term_ref();
    SP_atom name = SP_atom_from_string("none");
    SP_cons_functor(result, name, 0);
    if (found != NULL) {
      *found = result;
    }
    return 1;
  } else {
    ltlnode *na = a->node;
    ltlnode *nb = b->node;
    ltlnode *dest;
    bitfield *transeval,*atomeval;
    SP_term_ref op;
    int i;

    atomeval = get_atom_evaluation(a);
    
    for (i=0; i<na->number_of_transitions; i++) {
      get_nodetransop(f, na->first_transid + i, &dest, &transeval, &op);
      if (dest==nb && transition_match(f, atomeval, transeval)) {
        if (nrmatches==0 || transition_in_array(op,matches,nrmatches)) {
          if (found != NULL) {
            *found = op;
          }
          return 1;
        } 
      }
    }
    return 0;
  }
}

SP_term_ref find_transition_operation(ltl_formula *f, ltlatom *a, ltlatom *b) {
  SP_term_ref op;
  int result = find_transition_operation_that_matches(f,a,b,NULL,0,&op);
  if (!result) {
    IFLTLDEBUG(ltlmsg("transition not matched in the state space: ");
               ltl_printnode(a->node);printf(" --> ");ltl_printnode(b ->node);printf("\n"););
    fprintf(stderr, "Warning: no matching operation found between atoms.\n");
    op = SP_new_term_ref();
  }
  return op;
}

/*
  returns the transition ids of the transitions between two atoms a and b
  as a Prolog list
 */
SP_term_ref find_transition_operations_diflist(ltl_formula *f, ltlatom *a, ltlatom *b, SP_term_ref list) {
  ltlnode *na = a->node;
  ltlnode *nb = b->node;
  ltlnode *dest;
  bitfield *transeval,*atomeval;
  SP_term_ref op;
  int i;

  atomeval = get_atom_evaluation(a);
 
  SP_term_ref tail;
  for (i=0; i<na->number_of_transitions; i++) {
    get_nodetransop(f, na->first_transid + i, &dest, &transeval, &op);
    if (dest==nb && transition_match(f, atomeval, transeval)) {
      tail = list;
      list = SP_new_term_ref();
      SP_cons_list(list,op,tail);
    }
  }
  return list;
}

SP_term_ref find_transition_operations(ltl_formula *f, ltlatom *a, ltlatom *b) {
  SP_term_ref list = SP_new_term_ref();
  return find_transition_operations_diflist(f,a,b,list);
}

int explore_atom(ltl_formula *formula, ltlatom *atom) {
  int nr_atom_trans = atom->number_of_transitions;
  if (nr_atom_trans < 0) {
    ltlnode *node = atom->node;
    int nr_node_trans = node->number_of_transitions;
    unsigned int transid = node->first_transid;
    unsigned int nextatomid;
    ltlnode *refnode;
    int i,j,nr_next_atoms;
    ltlatom *nextatom;
    bitfield *evaluation,*nextevaluation,*transevaluation;

    IFLTLDEBUG(ltlmsg("explore atom %u, node has %i transitions\n",
                      find_atom_id(formula, atom), nr_node_trans);
               ltldeb_inc(); );
    
    evaluation = get_atom_evaluation(atom);

    /* iterate over all subsequent nodes */
    nr_atom_trans = 0;
    for (i=0; i<nr_node_trans; i++) {
      get_nodetransop(formula, transid, &refnode, &transevaluation, NULL);
      if (refnode) {
        IFLTLDEBUG(ltlmsg("checking transitions to refnode ");
                   ltl_printnode(refnode); printf("\n");
                   ltldeb_inc(); );

        if (transition_match(formula, evaluation, transevaluation)) {
          if (explore_node(formula, refnode, 0)) {
            nr_next_atoms = refnode->number_of_atoms;
            nextatomid = refnode->first_atom_id;

            /* iterate over all atoms of the next node */
            for (j=0; j<nr_next_atoms; j++) {
              get_atom(&formula->astore, nextatomid+j, &nextatom, &nextevaluation);
                
              IFLTLDEBUG(ltlmsg("checking transition to atom %u\n", nextatomid+j););
                
              if (atoms_match(formula, evaluation, nextevaluation)) {
                /* IFLTLDEBUG(ltlmsg("- transition found.\n");); */
                assure_arraysize(&formula->atomtransstore, formula->number_of_atomtrans + 1);
                if (nr_atom_trans==0) {
                  atom->first_transition_id = formula->number_of_atomtrans;
                }
                set_atom_trans(formula, formula->number_of_atomtrans, nextatom);
                  
                nr_atom_trans++;
                formula->number_of_atomtrans += 1;
              }
            }
          } else {
            IFLTLDEBUG( ltlmsg("explore_node() failed.\n");
                        ltldeb_dec(); );
            return 0;
          }
        }
        
        IFLTLDEBUG( ltldeb_dec(); );
      } else {
        emergency_break("invalid pointer to next node encountered!");
      }
      transid++;
    }
    atom->number_of_transitions = nr_atom_trans;
    
    IFLTLDEBUG(ltldeb_dec(););
  }
  return 1;
}

int atoms_match(ltl_formula *formula, bitfield *a, bitfield *b) {
  int i,equal,max = formula->number_of_equal_evaluations;
  equal_evaluation_positions *eep = formula->equal_evaluations;
  int va,vb;
  for (i=0; i<max; i++) {
    va = eep[i].a<0 ? 1 : getbit(a,eep[i].a);
    vb = eep[i].b<0 ? 1 : getbit(b,eep[i].b);
    equal = va ? vb : !vb;
    if (eep[i].signum ? !equal : equal) return 0;
  }
  return 1;
}

static int transeval_present(ltl_formula *f, bitfield *evaluation, int size) {
  int i;
  for (i=0; i<size; i++) {
    if (equal_bitfields(evaluation,
                        *((bitfield**) (get_array_element(&f->tmp_transeval_store, i))),
                        f->number_of_tps))
      return 1;
  }
  return 0;
}

int process_outgoing(ltl_formula *formula, ltlnode *node, SP_term_ref list, int *number_of_transevals) {
  SP_term_ref tail,elem,nodeterm,opterm,evallist=0;
  nodestore *nstore = &formula->nstore;
  ltlnode *refnode;
  unsigned int counter,transid;
  int tps_present = formula->number_of_tps>0;
  int ok = 1;
  bitfield *evaluation;

  *number_of_transevals = 0;

  tail = SP_new_term_ref();
  elem = SP_new_term_ref();
  nodeterm = SP_new_term_ref();
  if (tps_present) evallist = SP_new_term_ref();
  
  SP_put_term(tail,list);
  counter=0;
  while(SP_get_list(tail, elem, tail)) {
    opterm = SP_new_term_ref();
    ok = SP_get_arg(1, elem, nodeterm);
    ok = ok && SP_get_arg(2, elem, opterm);
    if (tps_present) ok = ok && SP_get_arg(3, elem, evallist);

    if (ok) {
      get_node_by_term(nstore, nodeterm, &refnode);
      save_nodetrans(formula, refnode, opterm, &evaluation, &transid);
      if (counter==0) {
        node->first_transid = transid;
      }
      if (tps_present) {
        eval_props(formula->number_of_tps, evaluation, evallist);
        if (!transeval_present(formula, evaluation, *number_of_transevals)) {
          assure_arraysize(&formula->tmp_transeval_store, *number_of_transevals+1);
          *((bitfield **) get_array_element(&formula->tmp_transeval_store, *number_of_transevals)) = evaluation;
          (*number_of_transevals)++;
        }
      }
      counter++;
    } else {
      fprintf(stderr, "wrong format of node transition!\n");
    }
  }
  node->number_of_transitions = counter;
  return counter;
}

static void eval_props(int size, bitfield *evaluation, SP_term_ref atomlist) {
  SP_term_ref tail = SP_new_term_ref();
  SP_term_ref elem = SP_new_term_ref();
  long l;
  int i;
  
  SP_put_term(tail,atomlist);
  i = 0;
  while(SP_get_list(tail, elem, tail)) {
    if(SP_get_integer(elem, &l)) {
      setbit(evaluation, i, l != 0L);
    } else {
      emergency_break("list of propositions contains a non-number element.\n");
    }
    i++;
  }
  if (i != size) {
    fprintf(stderr, "list of propositions has wrong size (%i instead of %i)\n", i, size);
  }
}

static void setbits(ltl_formula *f, bitfield *evaluation, 
                    unsigned long nextindex, 
                    unsigned long yestindex,
                    int deadlock,
                    int start) {
  if (deadlock) {
    /* deadlock, X(...) == FALSE and X(not(...)) == FALSE */
    setbit_by_long(evaluation, f->index_nextnots, 0L, f->number_of_nexts * 2);
  } else {
    /* setting the next formulas X(a) where a!=not(...) */
    setbit_by_long(evaluation, f->index_nexts, nextindex, f->number_of_nexts);
    /* no deadlock, X(not(a)) == not(X(a)) */
    setbit_by_long(evaluation, f->index_nextnots, ~nextindex, f->number_of_nexts);
  }

  if (start) {
    /* Y(...) == FALSE and Y(not(...)) == FALSE */
    setbit_by_long(evaluation, f->index_yesterdays, 0L, f->number_of_yesterdays * 2);
  } else {
    setbit_by_long(evaluation, f->index_yesterdays, yestindex, f->number_of_yesterdays);
    setbit_by_long(evaluation, f->index_yesterdaynots, ~yestindex, f->number_of_yesterdays);
  }
}

int create_atoms(ltl_formula *formula, ltlnode *node, SP_term_ref atomlist, int is_start_node, int trans_combinations) {
  int deadlock = node->number_of_transitions == 0;
  ltlarray *astore = &formula->astore;
  unsigned int atomid =formula->number_of_atoms;
  bitfield *evaluation = formula->tmp_evaluation;
  bitfield *transeval = 0;
  ltlatom *atom;
  bitfield *current;
  unsigned long nextaslong, yestaslong;
  int is_startatom;
  int tps_present = formula->number_of_tps>0;
  int transiter;

  unsigned int next_combinations = deadlock ? 1 : (1 << formula->number_of_nexts);
  unsigned int yest_combinations;
  if (formula->number_of_yesterdays > 0)
    yest_combinations = (1 << formula->number_of_yesterdays) + (is_start_node ? 1 : 0);
  else
    yest_combinations = 1;
  if (tps_present)  {
    if (deadlock) trans_combinations = 1;
  }
  else
    trans_combinations = 1;

  // the number of atoms corresponding to a state in the original state space
  unsigned int number_of_atoms = next_combinations * yest_combinations * trans_combinations;

  /*  IFLTLDEBUG(ltlmsg("create atoms for node ");ltl_printnode(node);printf("\n");
      ltldeb_inc();); */

  node->first_atom_id = atomid;
  node->number_of_atoms = number_of_atoms;
  formula->number_of_atoms += number_of_atoms;
  assure_arraysize(astore, formula->number_of_atoms);

  /* IFLTLDEBUG(ltlmsg("%u new atoms, total %u\n", number_of_atoms, formula->number_of_atoms);); */

  eval_props(formula->number_of_aps, evaluation, atomlist);

  for (transiter=0; transiter < trans_combinations; transiter++) {
    if (tps_present) {
      if (deadlock)
        transeval = formula->zero_transeval;
      else
        transeval = *((bitfield**) get_array_element(&formula->tmp_transeval_store, transiter));
    }
    for (nextaslong=0; nextaslong < next_combinations; nextaslong++) {
      for (yestaslong=0; yestaslong < yest_combinations; yestaslong++) {
        get_atom(astore, atomid, &atom, &current);

        /* init atom's data */
        atom->node = node;
        atom->number_of_transitions = -1;
        atom->tarjan_number = -1;
        atom->tarjan_on_stack = 0;

        /* setting atomic propositions */
        copy_bitfield(evaluation, current, 0, formula->number_of_aps);

        /* setting transition propositions */
        if (tps_present)
          insert_bitfield(transeval, current, formula->index_tps, formula->number_of_tps);

        /* setting next and yesterday bits */
        is_startatom = is_start_node && yestaslong+1 == yest_combinations;
        setbits(formula, current, nextaslong, yestaslong, deadlock, is_startatom);

        /* evaluate the rest of the formula */
        ltlevaluate(formula, current);

        /*IFLTLDEBUG(ltlmsg("atom %u of node %i created, evaluation = ", atomid, 
          node->type==int_node ? node->id.id : -1);
          show_bitfield(current, formula->closure_size);
          printf("\n");); */
        atomid++;
      }
    }
  }

  /* IFLTLDEBUG(ltldeb_dec();) */

  return number_of_atoms;
}

void save_nodetrans(ltl_formula *f, ltlnode *target, SP_term_ref op, bitfield **eval, unsigned int *transid) {
  *transid = f->nodetrans_number;
  f->nodetrans_number++;
  assure_arraysize(&f->nodetrans, f->nodetrans_number);
  void *mempos = get_array_element(&f->nodetrans, *transid);
  ((ltlnodetrans*) mempos)->dest = target;
  ((ltlnodetrans*) mempos)->operation = op;
  *eval = (bitfield*) (((uintptr_t) mempos) + sizeof(ltlnodetrans));
}

SP_term_ref maketerm(ltlnode *node) {
  SP_term_ref ref = SP_new_term_ref();
  if (node->type == int_node) {
    SP_put_integer(ref, (long) node->id.id);
  } else {
    SP_put_term(ref, node->id.ref);
  }
  return ref;
}

int eval(bitfield *evaluation, ltl_pform_ref *ref) {
  int value = (ref->ref < 0) ? 1 : getbit(evaluation, ref->ref);
  return ref->signum ? value : !value;
}

void ltlevaluate(ltl_formula *formula, bitfield *evaluation) {
  int i;
  int value = 0;
  ltl_part_formula *f;

  for (i=formula->index_generics; i<formula->closure_size; i++) {
    f = formula->closure + i;
    switch (f->op) {
    case ltlap: emergency_break("can not evaluate atomic proposition.\n"); break;
    case ltltp: emergency_break("can not evaluate transition proposition.\n"); break;
    case ltlnext: emergency_break("can not evaluate next.\n"); break;
    case ltlnextnot: emergency_break("can not evaluate nextnot.\n"); break;
    case ltlyesterday: emergency_break("can not evaluate yesterday.\n"); break;
    case ltlyesterdaynot: emergency_break("can not evaluate yesterdaynot.\n"); break;
    case ltlor: value = eval(evaluation,&f->args.case_or.a) || eval(evaluation,&f->args.case_or.b); break;
    case ltluntil: value = eval(evaluation,&f->args.case_until.goal) || 
        (eval(evaluation,&f->args.case_until.inv) &&  getbit(evaluation,f->args.case_until.next)); break;
    case ltlsince: value = eval(evaluation,&f->args.case_since.goal) || 
        (eval(evaluation,&f->args.case_since.inv) &&  getbit(evaluation,f->args.case_since.yest)); break;
    }
    setbit(evaluation,i,value);
  }
}

void get_atom(ltlarray *astore, unsigned int atomid, ltlatom **atom, bitfield **evaluation) {
  void *mempos = get_array_element(astore, atomid);
  if (mempos) {
    *atom = (ltlatom *) mempos;
    *evaluation = (bitfield *) (((uintptr_t) mempos) + sizeof(ltlatom));
  } else {
    emergency_break("get_atom() called for unknown ID");
  }
}

bitfield *get_atom_evaluation(ltlatom *atom) {
  void *mempos = (void*) atom;
  return (bitfield*) (((uintptr_t) mempos) + sizeof(ltlatom));
}

ltlatom *get_atom_trans(ltl_formula *f, unsigned int id) {
  return *((ltlatom **) get_array_element(&(f->atomtransstore), id));
}

void set_atom_trans(ltl_formula *f, unsigned int id, ltlatom *dest) {
  IFLTLDEBUG( ltlmsg("set atom trans %i to %p\n", id, dest); )
  *((ltlatom **) get_array_element(&(f->atomtransstore), id)) = dest;
}

size_t ltl_formula_memory_usage(ltl_formula *f) {
  size_t mem = sizeof(ltl_formula);
  mem += f->closure_size * sizeof(ltl_part_formula);
  mem += f-> number_of_equal_evaluations * sizeof(equal_evaluation_positions);
  mem += ltl_nodestore_memory_usage(&f->nstore);
  mem += ltl_array_memory_usage(&f->nodetrans);
  mem += ltl_array_memory_usage(&f->astore);
  mem += ltl_array_memory_usage(&f->atomtransstore);
  return mem;
}
