% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(probhash, [add_term/3, reset_probhash/0, hash_term/2, hash_to_atom/2, raw_sha_hash/2]).
:- use_module(library(fastrw),[ fast_buf_read/2,
                                fast_buf_write/3 ]).

:- use_module('../../src/module_information.pl').

:- module_info(group,infrastructure).
:- module_info(description,'This is the interface to C code for generating good hashes (SHA-1).').

foreign_resource(probhash,[sha1]).
foreign(sha1,sha1(+address(char),+integer,[-term])).

:- dynamic is_initialised/0.

init_probhash :- is_initialised,!.
init_probhash :- 
    load_foreign_resource(library(probhash)),
    assert(is_initialised).

% input: term
% output: hash value as a biginteger (the byte order is reversed)
hash_term(Term,Hash) :- raw_sha_hash(Term,H), hash_to_int(H,Hash).

:- dynamic hash_id/1.
:- dynamic hash_h1h2/2.
:- dynamic hash_h2h3/2.
:- dynamic hash_h3h4/2.
:- dynamic hash_h4h5/2.
:- dynamic hash_h5id/2.

reset_probhash :-
    retractall(hash_id(_)), assert(hash_id(0)),
    retractall(hash_h1h2(_,_)), retractall(hash_h2h3(_,_)),
    retractall(hash_h3h4(_,_)), retractall(hash_h4h5(_,_)),
    retractall(hash_h5id(_,_)).

hash_id(0).

gen_new_id(X) :- hash_id(N), retract(hash_id(N)), X is N + 1, assert(hash_id(X)).

lookup(H1,H2,H3,H4,H5,X) :-
  hash_h1h2(H1,H2), hash_h2h3(H2,H3), hash_h3h4(H3,H4), hash_h4h5(H4,H5), hash_h5id(H5,X).

store_hash(H1,H2,H3,H4,H5,X) :-
   assert_hash_h1h2(H1,H2), assert_hash_h2h3(H2,H3), assert_hash_h3h4(H3,H4),
   assert_hash_h4h5(H4,H5), assert(hash_h5id(H5,X)).

assert_hash_h1h2(X,Y) :- hash_h1h2(X,Y) -> true ; assert(hash_h1h2(X,Y)).
assert_hash_h2h3(X,Y) :- hash_h2h3(X,Y) -> true ; assert(hash_h2h3(X,Y)).
assert_hash_h3h4(X,Y) :- hash_h3h4(X,Y) -> true ; assert(hash_h3h4(X,Y)).
assert_hash_h4h5(X,Y) :- hash_h4h5(X,Y) -> true ; assert(hash_h4h5(X,Y)).

add_term(Term, Id, New) :- raw_sha_hash(Term,TH), hash_to_32bit(TH,[H1,H2,H3,H4,H5]),
    ( lookup(H1,H2,H3,H4,H5,X) ->
        New=false, Id=X 
    ; otherwise ->
        New = true, gen_new_id(Id), store_hash(H1,H2,H3,H4,H5,Id)).


% input: raw hash 
% output: hash value as a biginteger (the byte order is reversed)
hash_to_int(X,Y) :- hash_to_int(X,0,Y).
hash_to_int([],A,A).
hash_to_int([H|T],A,R) :- A2 is A * 256 + H, hash_to_int(T,A2,R).

% raw sha hashing, input is a term, output a raw hash, i.e., a list of 20 bytes
raw_sha_hash(Term,Digest) :-
    init_probhash,
    fast_buf_write(Term,Len,Addr),
    sha1(Addr,Len,Digest).

% input: raw hash value
% output: hash value as an atom 
hash_to_atom(H,ResultAtom) :-
        raw_sha_hash(H,H2), hash_to_int(H2,X),
        number_codes(X, C),atom_codes(ResultAtom,C).

% input: raw hash value
% output: hash value as 5 32-bit chunks 
hash_to_32bit([V1,V2,V3,V4,V5,V6,V7,V8,V9,V10,V11,V12,V13,V14,V15,V16,V17,V18,V19,V20],[H1,H2,H3,H4,H5]) :-
  hash_to_int([V1,V2,V3,V4],H1),
  hash_to_int([V5,V6,V7,V8],H2),
  hash_to_int([V9,V10,V11,V12],H3),
  hash_to_int([V13,V14,V15,V16],H4),
  hash_to_int([V17,V18,V19,V20],H5).
