/*
    Title: A Profiler for SICStus Prolog
    Author: Philip Höfges
    Heinrich-Heine-University Düsseldorf
    Lehrstuhl für Softwaretechnik und Programmiersprachen
    Prof. Dr. Michael Leuschel, Sebastian Krings
*/

% declare the programm as a module
:- module(profiler, [reset_dynamics/0,call_for_profiling/1,
        enable_profiling_naming/2,enable_profiling_naming/3,
        enable_profiling/1,enable_profiling/2,
        enable_profiling_all/1,
        coroutines/1,call_for_profiling/1,
        create_new_terms/2,file/0,pen/0,memory/0,clear_times/0]).

% define dynamic predicates to remember results
%    profiling_term: profiles all predicates in form of Name/Arity
%    profiling_term_with_if: profiles if arguments of Name(Arg1,...) match
%    profiling_condition: profiles in form of (Name/Arity,Cond) where Cond is a predicate which need to be true
%    predicate: remembers every predicate that has already a startterm
%    blocked: remembers all predicate which a co-routines in form of :- block ...
%    corouting: decide if coroutines should be profiled
%    prof_all: alle predicates of this module should be profiled
:- dynamic profiling_term/2,profiling_term_with_if/3,profiling_condition/3,predicate/3,blocked/1,corouting/1,prof_all/1.

% set corouting to false
:- assert(corouting(false)).

% declare cpp resource
foreign_resource('profiler', [file,built_list,clear_timers,handle_entry,handle_success,handle_redo,handle_fail,pen,stop_timers,restart_timers]).
foreign(handle_entry, c, handle_entry(+atom)).
foreign(handle_success, c, handle_success(+atom)).
foreign(handle_redo, c, handle_redo(+atom)).
foreign(handle_fail, c, handle_fail(+atom)).
foreign(stop_timers, c, stop_timers).
foreign(restart_timers, c, restart_timers).
foreign(pen, c, pen).
foreign(clear_timers, c, clear_timers).
foreign(built_list, c, built_list([-term])).
foreign(file, c, file).
:- load_foreign_resource(profiler).

%--------------------------------------------------------------------------------------
% profiling functions
%--------------------------------------------------------------------------------------

% startpoint for time counting: use backtracking to differ between cases
prof_on(Name) :- handle_entry(Name).
prof_on(Name) :- handle_fail(Name),fail.

% end of counting time: use backtracking to differ between cases
prof_off(Name) :- handle_success(Name).
prof_off(Name) :- handle_redo(Name),fail.

% startpoint for corouting: use backtracking to differ between enter and leave
enter_coro :- stop_timers.
enter_coro :- restart_timers, fail.

% end of corouting: use backtracking to differ between enter and leave
leave_coro :- restart_timers.
leave_coro :- stop_timers, fail.

%-------

% enable which functions should be profiled (naming added)
enable_profiling_all(Module) :-
    assert(prof_all(Module)).

:- meta_predicate enable_profiling_naming(-,:).
enable_profiling_naming(Module:Name/Arity,NP) :- !,
    assert(profiling_term(Module:Name/Arity,NP)).
enable_profiling_naming(Name/Arity,NP) :- !,
    assert(profiling_term(user:Name/Arity,NP)).

enable_profiling_naming(Module:Term,NP) :- !,
    functor(Term,Name,Arity),
    assert(profiling_term_with_if(Module:Term,Name/Arity,NP)).
enable_profiling_naming(Term,NP) :- !,
    functor(Term,Name,Arity),
    assert(profiling_term_with_if(user:Term,Name/Arity,NP)).

:- meta_predicate enable_profiling_naming(-,:,:).
enable_profiling_naming(_:Name/Arity,Condition,NP) :- !,
    assert(profiling_condition(Condition,Name/Arity,NP)).
enable_profiling_naming(Name/Arity,Condition,NP) :- !,
    assert(profiling_condition(Condition,Name/Arity,NP)).

%-------

% enable which functions should be profiled (no naming added)
:- meta_predicate enable_profiling(:).
enable_profiling(Module:Name/Arity) :- !,
    assert(profiling_term(Module:Name/Arity,profiler:default_name)).
enable_profiling(Name/Arity) :-
    assert(profiling_term(user:Name/Arity),profiler:default_name).

enable_profiling(Module:Term) :- !,
    functor(Term,Name,Arity),
    assert(profiling_term_with_if(Module:Term,Name/Arity,profiler:default_name)).
enable_profiling(Term) :-
    functor(Term,Name,Arity),
    assert(profiling_term_with_if(user:Term,Name/Arity,profiler:default_name)).

% enable which functions should be profiled with special conditions
:- meta_predicate enable_profiling(-,:).
enable_profiling(_:Name/Arity,Condition) :- !,
    assert(profiling_condition(Condition,Name/Arity,profiler:default_name)).
enable_profiling(Name/Arity,Condition) :-
    assert(profiling_condition(Condition,Name/Arity,profiler:default_name)).

% default for naming predicate
default_name(N,_,N).

% reset all dynamic predicates
reset_dynamics :-
   retractall(profiling_term(_,_)),
   retractall(profiling_term_with_if(_,_,_)),
   retractall(profiling_condition(_,_,_)),
   retractall(predicate(_,_,_)),
   retractall(blocked(_)),
   retractall(tcl(_)),
   retractall(rect(_,_)),
   retractall(corouting(_)).

% add a function to be profiled at runtime
call_for_profiling([]) :- pen.
call_for_profiling([Pred|T]) :- !,
    functor(Pred,Name,_),
    prof_on(Name),
    call(user:Pred),
    prof_off(Name),
    call_for_profiling(T).
call_for_profiling(_) :-
    print('This predicate is not available. Error! Try again!'), nl,
    fail.

% switch to decide about behavior of co-routines
coroutines(X) :-
    retract(corouting(_)),
    assert(corouting(X)).
coroutines(X) :-
    assert(corouting(X)).

%--------------------------------------------------------------------------------------
% output functions
%--------------------------------------------------------------------------------------

% starts to print memory usage
memory :-
    statistics.

%--------------------------------------------------------------------------------------
% term_expansion
%--------------------------------------------------------------------------------------

% check if the desired function should be expanded
to_be_expanded(Name/Arity,fully) :-
    (profiling_term_with_if(_,Name/Arity,_);
    profiling_condition(_,Name/Arity,_)).
to_be_expanded(Name/Arity,partly) :-
    prolog_load_context(module,Module),
    profiling_term(Module:Name/Arity,_).
to_be_expanded(Name/Arity,partly) :-
    prof_all(Module),
    profiling_term(Module:Name/Arity,_).

% add the predicate that should blocked
to_be_blocked(List) :- !,
    List = (H,T),
    functor(H,Name,Arity),
    assert(blocked(Name/Arity)),
    to_be_blocked(T).
to_be_blocked(Last) :-
    functor(Last,Name,Arity),
    assert(blocked(Name/Arity)).

% assert if the whole module should be profiled
assert_if_needed((H :- _)) :- !,
    assert_if_needed(H).
assert_if_needed(Head) :-
    functor(Head,Name,Arity),
    prolog_load_context(module,Module),
    prof_all(Module),
    (\+ do_not_assert(Name,Arity)),
    (\+ profiling_term(Module:Name/Arity,profiler:default_name)), !,
    assert(profiling_term(Module:Name/Arity,profiler:default_name)).
assert_if_needed(_).

% the "prediacates" that should not be asserted
do_not_assert(':-',1).
do_not_assert(end_of_file,0).

% create new terms on top level
% backtracks into the last case (no expansion) if to_be_expanded fails always
% first case for co-routines with the block predicate
create_new_terms(A,B) :-
    assert_if_needed(A),
    create_new_terms2(A,B).
create_new_terms2((:- block List),(:- block List)) :-
    to_be_blocked(List),
    !.
create_new_terms2((:- if(X)),(:- if(X))) :- !.
create_new_terms2((:- elif(X)),(:- elif(X))) :- !.
create_new_terms2((:- endif),(:- endif)) :- !.
create_new_terms2((:- else),(:- else)) :- !.
create_new_terms2((Head :- Body),[Term1,Term2]) :-
    functor(Head,Name,Arity),
    prolog_load_context(module,Module),
    predicate(Module,Name,Arity), !,
    to_be_expanded(Name/Arity,Sort),
    create_new_term((Head :- Body),'_p',Term1),
    (Sort = fully -> create_new_term((Head :- Body),'_np',Term2) ; true).
create_new_terms2((Head :- Body),[Term1,Term2,Term3]) :-
    functor(Head,Name,Arity),
    to_be_expanded(Name/Arity,Sort), !,
    create_new_term((Head :- Body),'_p',Term1),
    (Sort = fully -> create_new_term((Head :- Body),'_np',Term2) ; true),
    create_new_startterm((Head :- Body),Term3).

% dcgs
create_new_terms2((Head --> Body),[Term1,Term2]) :-
    functor(Head,Name,Arity),
    prolog_load_context(module,Module),
    predicate(Module,Name,Arity), !,
    to_be_expanded(Name/Arity,Sort),
    create_new_term((Head --> Body),'_p',Term1),
    (Sort = fully -> create_new_term((Head --> Body),'_np',Term2) ; true).
create_new_terms2((Head --> Body),[Term1,Term2,Term3]) :-
    functor(Head,Name,Arity),
    to_be_expanded(Name/Arity,Sort), !,
    create_new_term((Head :- Body),'_p',Term1),
    (Sort = fully -> create_new_term((Head --> Body),'_np',Term2) ; true),
    create_new_startterm((Head --> Body),Term3).

create_new_terms2(Term,[Term1,Term2]) :-
    functor(Term,Name,Arity),
    prolog_load_context(module,Module),
    predicate(Module,Name,Arity), !,
    to_be_expanded(Name/Arity,Sort),
    create_new_term(Term,'_p',Term1),
    (Sort = fully -> create_new_term(Term,'_np',Term2) ; true).
create_new_terms2(Term,[Term1,Term2,Term3]) :-
    functor(Term,Name,Arity),
    to_be_expanded(Name/Arity,Sort), !,
    create_new_term(Term,'_p',Term1),
    (Sort = fully -> create_new_term(Term,'_np',Term2) ; true),
    create_new_startterm(Term,Term3).
create_new_terms2(Term,[Term]).

% create new body with new names but the same rules
% backtracks into the last case (Func) if there is no other match
create_new_body(when(Cond,Goal),_,when(Cond,New_Goal)) :-
    corouting(true),
    to_be_expanded(Goal,_),
    New_Goal = (profiler:enter_coro,Goal,profiler:leave_coro), !.
create_new_body(Fact,_,Fact) :- var(Fact).
create_new_body(\+(X),Ex,(\+(Body))) :-
    create_new_body(X,Ex,Body).
create_new_body((A->B),Ex,(Body_A->Body_B)) :-
    create_new_body(A,Ex,Body_A),
    create_new_body(B,Ex,Body_B).
create_new_body((A->B;C),Ex,(Body_A->Body_B;Body_C)) :-
    create_new_body(A,Ex,Body_A),
    create_new_body(B,Ex,Body_B),
    create_new_body(C,Ex,Body_C).
create_new_body(if(A,B,C),Ex,New_Body) :-
    create_new_body((A->B;C),Ex,New_Body).
create_new_body((A,B),Ex,(Body_A,Body_B)) :-
    create_new_body(A,Ex,Body_A),
    create_new_body(B,Ex,Body_B).
create_new_body((A;B),Ex,(Body_A;Body_B)) :-
    create_new_body(A,Ex,Body_A),
    create_new_body(B,Ex,Body_B).
create_new_body(Func,_,Func) :-
    functor(Func,N,Arity),
    blocked(N/Arity),
    corouting(true),
    to_be_expanded(N/Arity,_).
create_new_body(Func,Ex,New_Body) :-
    functor(Func,N,Arity),
    to_be_expanded(N/Arity,_),
    Func =.. [Name|T],
    atom_concat(Name,Ex,New_Name),
    New_Body =.. [New_Name|T].
create_new_body(Func,_,Func).

% create a new term with new name (_p and _np) with the same body
create_new_term((Head :- Body),Ex,(New_Head :- New_Body)) :-
    Head =.. [Name|T],
    atom_concat(Name,Ex,New_Name),
    New_Head =.. [New_Name|T],
    create_new_body(Body,Ex,New_Body).

% dcgs
create_new_term((Head --> Body),Ex,(New_Head --> New_Body)) :-
    Head =.. [Name|T],
    atom_concat(Name,Ex,New_Name),
    New_Head =.. [New_Name|T],
    create_new_body(Body,Ex,New_Body).

create_new_term(Fact,Ex,New_Fact) :-
    Fact =.. [Name|T],
    atom_concat(Name,Ex,New_Name),
    New_Fact =.. [New_Name|T].

% do preperations for creating a new startterm
do_the_common_stuff((Head :- _),Name,Len,T2,Caller,Caller2,New_Head) :- !,
    do_the_common_stuff_aux(Head,Name,Len,T2,Caller,Caller2,New_Head).
do_the_common_stuff(Term,Name,Len,T2,Caller,Caller2,New_Head) :-
    do_the_common_stuff_aux(Term,Name,Len,T2,Caller,Caller2,New_Head).
do_the_common_stuff_aux(Head,Name,Len,T2,Caller,Caller2,New_Head) :-
    Head =.. [Name|T],
    length(T,Len),
    length(T2,Len),
    atom_concat(Name,'_p',New_Name),
    atom_concat(Name,'_np',New_Name2),
    New_Head =.. [Name|T2],
    Caller =.. [New_Name|T2],
    Caller2 =.. [New_Name2|T2].

% create a new function with the same name but different body
% and differs between coroutines and normal predicates
% if->then;else is necessary here
create_new_startterm(In,Out) :-
    functor(In,Func,_),
    do_the_common_stuff(In,Name,Len,Args,Caller,Caller2,New_Head), !,
    ( (blocked(Name/Len),corouting(true))
    -> create_new_startterm_blocked(Name,Len,Args,Caller,Caller2,New_Head,Func,Out)
    ; create_new_startterm_aux(Name,Len,Args,Caller,Caller2,New_Head,Func,Out)
    ),
    prolog_load_context(module,Module),
    assert(predicate(Module,Name,Len)).

% create a new startterm as coroutine with profiling on
create_new_startterm_blocked(Name,Len,Args,Caller,Caller2,New_Head,Func,Out) :-
    profiling_term_with_if(_:Term,Name/Len,NP),
    Out =.. [Func|(New_Head, (New_Head=Term->call(NP,Name,Args,A),profiler:enter_coro,profiler:prof_on(A),Caller,profiler:prof_off(A),profiler:leave_coro;Caller2))].
create_new_startterm_blocked(Name,Len,Args,Caller,Caller2,New_Head,Func,Out) :-
    profiling_condition(Modul:Cond,Name/Len,NP),
    Out =.. [Func|(New_Head,(call(Modul:New_Cond)->call(NP,Name,Args,A),profiler:enter_coro,profiler:prof_on(A),Caller,profiler:prof_off(A),profiler:leave_coro;Caller2))],
    New_Head =.. [_|T],
    New_Cond =.. [Cond|T].
create_new_startterm_blocked(Name,Len,Args,Caller,_,New_Head,Out) :-
    profiling_term(_:Name/Len,NP),
    Out =.. [Func|(New_Head, (call(NP,Name,Args,A),profiler:enter_coro,profiler:prof_on(A),Caller,profiler:prof_off(A),profiler:leave_coro))].

% create a new startterm as normal predicate with profiling on
create_new_startterm_aux(Name,Len,Args,Caller,Caller2,New_Head,Func,Out) :-
    profiling_term_with_if(_:Term,Name/Len,NP),
    Out =.. [Func|(New_Head, (New_Head=Term->call(NP,Name,Args,A),profiler:prof_on(A),Caller,profiler:prof_off(A);Caller2))].
create_new_startterm_aux(Name,Len,Args,Caller,Caller2,New_Head,Out) :-
    profiling_condition(Modul:Cond,Name/Len,NP),
    Out =.. [Func|(New_Head, (call(Modul:New_Cond)->call(NP,Name,Args,A),profiler:prof_on(A),Caller,profiler:prof_off(A);Caller2))],
    New_Head =.. [_|T],
    New_Cond =.. [Cond|T].
create_new_startterm_aux(Name,Len,Args,Caller,_,New_Head,Func,Out) :-
    profiling_term(_:Name/Len,NP),
    Out =.. [Func|(New_Head, (call(NP,Name,Args,A),profiler:prof_on(A),Caller,profiler:prof_off(A)))].
