#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "queue.h"
#include "ring.h"
#include "sha.h"

#define MAP_THRESH 50


static void enqueue_fifo(wQueue *q, char *term, int len, char digest[20]);
static void enqueue_lifo(wQueue *q, char *term, int len, char digest[20]);

wQueue *init_queue(int strategy) {
    ring_env *env = NULL;
    if (strategy == 0) {
        env = ring_init(1048576);
    }
    wQueue *new = malloc(sizeof(wQueue));
    new->first = NULL;
    new->last = NULL;
    new->c = 0;
    new->strategy = strategy;
    new->env = env;
    return new;
}

void print_term(tCell *t) {
	int i;
	for (i=0;i<t->term_length;i++) {
		printf("%c",(t->term)[i]);
	}
	printf("\n");
}

// TODO: use this in enqueue_fifo
void enqueue_cell(wQueue *q, tCell *t) {
    q->c++;
    if (q->first == NULL) {
        q->first = t; 
        q->last = t;
        q->first->next = NULL;
    } else { 
        tCell *temp = q->last; // for multithreading
        q->last = t;
        temp->next = t;
    }
}

void enqueue(wQueue *q, char *term, int len, char digest[20]) {
    int r = rand() % 100;
    if (r < q->strategy) {
        assert(0);
        enqueue_lifo(q, term, len, digest);
    }
    else {
        enqueue_fifo(q, term, len, digest);
    }
}

#ifdef DEBUG
void enqueue_debug(wQueue *q, char *term, int len, char digest[20]) {
    q->c++;
    // ignore if in Hashmap 
    // else add to working queue
    tCell *new = malloc(sizeof(tCell));
    
    new->term = malloc(len);
    memcpy(new->term, term, len);
    new->term_length = len;
    new->digest = malloc(20);
    memcpy(new->digest, digest, 20);
    new->next = NULL;
    new->pos = (size_t) -1;
    
    //  printf("queued: %s %s\n",term,digest);
    
    
    
    if (q->first == NULL) {
        q->first = new; 
        q->last = new;
        q->first->next = NULL;
    }
    else { 
        tCell *temp = q->last; // for multithreading
        q->last = new;
        temp->next = new;
    }
        
}

void clear_debug(wQueue *q) {
    while (q->first != NULL) {
        tCell *ele = dequeue(q);
        free(ele->term);
        free(ele->digest);
    }
    free(q);
}
#endif

static void enqueue_fifo(wQueue *q, char *term, int len, char digest[20]) {
    q->c++;
    // ignore if in Hashmap 
    // else add to working queue
    tCell *new = malloc(sizeof(tCell));
//    printf("strat: %d\n", q->strategy);

    size_t pos = ring_put(q->env, term, len, digest);
    free(term);
    new->term = NULL;
    new->pos = pos;
    new->term_length = len;
    new->digest = digest;
    new->next = NULL;

    
    //	printf("queued: %s %s\n",term,digest);
    
    
    
    if (q->first == NULL) {
        q->first = new; 
        q->last = new;
        q->first->next = NULL;
    }
    else { 
        tCell *temp = q->last; // for multithreading
        q->last = new;
        temp->next = new;
    }
    
}

static void enqueue_lifo(wQueue *q, char *term, int len, char digest[20]) {
    q->c++;
    tCell *new = malloc(sizeof(tCell));
    new->term = term;
    new->term_length = len;
    new->digest = digest;
    new->next = q->first;
    q->first = new;
    if (q->last == NULL)
        q->last = new;
}

static void load_terms(tCell *cell, wQueue *q) {
    int i = 0;
    while (i++ < MAP_THRESH && cell != NULL && cell->term == NULL) {
        assert(cell);
        cell->term = ring_retrieve(q->env, cell->pos, cell->term_length, cell->digest);
        assert(cell->term);
        //ring_free(cell->pos, cell->term_length);
// TODO: remove assertions and additional sha1 call after I'm confident it works
        char buf[20];
        sha1(cell->term, cell->term_length, buf);
        assert(0 == memcmp(buf, cell->digest, 20));
        cell = cell->next;
    }
}

tCell *dequeue(wQueue *q) {
    if (q->c == 0) {
        return NULL;
    }
    q->c--;
    tCell *res = q->first;
    if (q->strategy == 0 && res->term == NULL) {
        load_terms(res, q);
    }
    assert(res->term);
    q->first = q->first->next;
    if (q->first == NULL) { q->last = NULL; }
    res->next = NULL;
    return res;
}

int is_empty(wQueue *q) {
    return (q->first == NULL);
}

int q_size(wQueue *q) {
    return q->c;
}


void print_queue(wQueue *q) {
    printf("Queue: [");
    tCell *p = q->first;
    //	if (first != NULL) printf("%s %i | ", first->term, c);
    while (p != NULL) {
        printf("%s ",p->term);
        //		print_key(p->digest);
        //		printf("\n");
        p = p->next;
    }
    printf("]\n");
}

void teardown_queue(wQueue **q) {
    while((*q)->c > 0) {
        tCell *cell = dequeue(*q);
        if (cell->term != NULL) {
            free(cell->term);
        }
        free(cell->digest);
        free(cell);
    }
    ring_teardown((*q)->env);
    free(*q);
    *q = NULL;
}

/* returns 0 if two different terms share the same digest */
#ifdef DEBUG
int check_debugq(wQueue *q, char digest[20], char *term, int len) {
    tCell *cell = q->first;
    while (cell != NULL) {
        if (memcmp(digest, cell->digest, 20) == 0 && (cell->term_length != len || memcmp(term, cell->term, len)))
            return 0;
        cell = cell->next;
    }
    return 1;
}
#endif
