#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <errno.h>

#include "mapper.h"
#include "ring.h"
#include "queue.h"

char *ring_retrieve(ring_env *env, size_t pos, size_t len, char *digest) {
    env->retrieve_count++;
    char *addr = get_in_mmap(env->mmap_env, pos, len + 21 + sizeof(size_t), 0);
// TODO: remove assertions after I'm confident it works
    assert(addr[0] == 'X');
    assert(memcmp(addr+1, &len, sizeof(size_t)) == 0);
    assert(memcmp(addr+1+sizeof(size_t), digest, 20) == 0);
    char *res = (char*) malloc(len);
    assert(res);

    memcpy(res, addr+21+sizeof(size_t), len);
    return res;
}

void ring_free(ring_env *env, tCell *cell) {
    assert(cell->pos != (size_t) -1);
    env->free_count++;
    size_t len = cell->term_length + 21 + sizeof(size_t);
    char *addr = get_in_mmap(env->mmap_env, cell->pos, len, 0);
    memset(addr, 0, len);
    env->pos_read = cell->pos + len;
    if (env->pos_read == env->extending) {
        env->extending = 0;
    }
    cell->pos = (size_t) -1;
}

static void write_term(char *addr, char *term, size_t term_len, char *digest) {
    addr[0] = 'X';
    memcpy(addr + 1,                  &term_len, sizeof(size_t));
    memcpy(addr + 1 + sizeof(size_t),  digest,   20);
    memcpy(addr + 21 + sizeof(size_t), term,     term_len);
}

size_t ring_put(ring_env *env, char *term, size_t term_len, char *digest) {
    env->put_count++;
    size_t len = sizeof(size_t) + 20 + term_len + 1;
    if (env->pos_write >= env->pos_read) {
        if (env->extending) {
            // append at end of file
            int rc = ftruncate(env->mmap_env->fd, env->pos_write + len);
            assert(rc == 0);
            env->bufsize = env->pos_write + len;
            char *addr = get_in_mmap(env->mmap_env, env->pos_write, len, 1);
//            memcpy(addr, term, len);
            write_term(addr, term, term_len, digest);
            size_t res = env->pos_write;
            env->pos_write += len;
            return res;
        } else if (env->pos_write + len <= env->bufsize) {
            // fits in buffer, write it
            char *addr = get_in_mmap(env->mmap_env, env->pos_write, len, 0);
            //memcpy(addr, term, len);
            write_term(addr, term, term_len, digest);
            size_t res = env->pos_write;
            env->pos_write += len;
            return res;
        } else if (len + 1 <= env->pos_read) {
            // buffer circular, write at beginning
            char *addr = get_in_mmap(env->mmap_env, 0, len, 0);
            //memcpy(addr, term, len);
            write_term(addr, term, term_len, digest);
            env->pos_write = len;
            return 0;
        } else {
            int rc = ftruncate(env->mmap_env->fd, env->pos_write + len);
            assert(rc == 0);
            env->bufsize = env->pos_write + len;
            char *addr = get_in_mmap(env->mmap_env, env->pos_write, len, 1);
            //memcpy(addr, term, len);
            write_term(addr, term, term_len, digest);
            size_t res = env->pos_write;
            env->pos_write += len;
            return res;
        }
    } else {
        // write in mmap because buffer has space
        // note the < instead of <=
        // this ensures that we have an unsed byte that distinguishes
        // a full buffer from an empty one
        if (env->pos_write + len < env->pos_read) {
            char *addr = get_in_mmap(env->mmap_env, env->pos_write, len, 0);
            //memcpy(addr, term, len);
            write_term(addr, term, term_len, digest);
            size_t res = env->pos_write;
            env->pos_write += len;
            return res;
        } else {
            // append at end of file
            env->extending = env->pos_write;
            int rc = ftruncate(env->mmap_env->fd, env->bufsize + len);
            assert(rc == 0);
            size_t res = env->bufsize;
            char *addr = get_in_mmap(env->mmap_env, env->bufsize, len, 1);
            //memcpy(addr, term, len);
            write_term(addr, term, term_len, digest);
            env->bufsize = env->bufsize + len;
            env->pos_write = env->bufsize;
            return res;
        }
    }
}


ring_env *ring_init(size_t bfsze) {
#ifndef TMPFOLDER
#define TMPFOLDER "/var/tmp/"
#endif
    char *filename = tempnam(TMPFOLDER, "zmq-");
    puts(filename);
    mmap_env *menv = mapper_init(filename);

    ring_env *env = malloc(sizeof(ring_env));
    env->mmap_env = menv;
    env->bufsize = bfsze;
    env->pos_read = 0;
    env->pos_write = 0;
    env->extending = 0;
    env->retrieve_count = 0;
    env->put_count = 0;
    env->free_count = 0;

    int rc = ftruncate(env->mmap_env->fd, env->bufsize);
    assert(rc == 0);
    return env;
}

void ring_teardown(ring_env *env) {
    int rc = close(env->mmap_env->fd);
    assert(rc == 0);
    remove(env->mmap_env->filename);
    free(env->mmap_env->filename);
    free(env->mmap_env);
    printf("bufsize: %lu\n", env->bufsize);
    printf("putcount: %d\n", env->put_count);
    printf("retcount: %d\n", env->retrieve_count);
    printf("frecount: %d\n", env->free_count);
    free(env);
}

