#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include "../common/queue.h"
#include "../common/sha.h"
#include "../common/ringcovery.h"
#include "CuTest.h"


tCell *gen_cell(char *str) {
    tCell *c = (tCell*) malloc(sizeof(tCell));
    char *digest = malloc(20);
    sha1(str, strlen(str) + 1, digest);
    c->term = str;
    c->digest = digest;
    c->term_length = strlen(str) + 1;
    c->pos = 0;
    c->next = 0;

    return c;
}


void write_cell(int fd, tCell *c) {
    write(fd, "X", 1);
    size_t len = c->term_length;
    write(fd, &len, sizeof(size_t));
    write(fd, c->digest, 20);
    write(fd, c->term, c->term_length);
}

void test_one_wp_is_all(CuTest *tc) {
    tCell *c = gen_cell("foobar123");
    int fd = open("testfile.zmq", O_CREAT | O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
    write_cell(fd, c);
    close(fd);
    
    wQueue *q = recover_file("testfile.zmq");

    CuAssertPtrNotNull(tc, q);
    CuAssertIntEquals(tc, q_size(q), 1);

    tCell *rec = dequeue(q);
    CuAssertStrEquals(tc, rec->term, c->term);
    CuAssertMemEquals(tc, rec->digest, c->digest, 20);
    CuAssertFalse(tc, rec->next);

    remove("testfile.zmq");
}

void test_two_wps_block(CuTest *tc) {
    int fd = open("testfile.zmq", O_CREAT | O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

    tCell *c1 = gen_cell("foobar123");
    write_cell(fd, c1);

    tCell *c2 = gen_cell("this_is_yet_another_term");
    write_cell(fd, c2);

    close(fd);
    
    wQueue *q = recover_file("testfile.zmq");

    CuAssertPtrNotNull(tc, q);
    CuAssertIntEquals(tc, q_size(q), 2);

    tCell *rec = dequeue(q);
    CuAssertPtrNotNull(tc, rec->term);
    CuAssertStrEquals(tc, rec->term, c1->term);
    CuAssertMemEquals(tc, rec->digest, c1->digest, 20);
    free(rec->term);
    free(rec->digest);
    free(rec);


    rec = dequeue(q);
    CuAssertPtrNotNull(tc, rec->term);
    CuAssertStrEquals(tc, rec->term, c2->term);
    CuAssertMemEquals(tc, rec->digest, c2->digest, 20);
    free(rec->term);
    free(rec->digest);
    free(rec);

    remove("testfile.zmq");
}

void test_leading_zeros(CuTest *tc) {
    int fd = open("testfile.zmq", O_CREAT | O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

    char *block = malloc(1000);
    memset(block, 0, 1000);
    write(fd, block, 1000);
    free(block);

    tCell *c1 = gen_cell("foobar123");
    write_cell(fd, c1);

    tCell *c2 = gen_cell("this_is_yet_another_term");
    write_cell(fd, c2);

    close(fd);
    
    wQueue *q = recover_file("testfile.zmq");

    CuAssertPtrNotNull(tc, q);
    CuAssertIntEquals(tc, q_size(q), 2);

    tCell *rec = dequeue(q);
    CuAssertPtrNotNull(tc, rec->term);
    CuAssertStrEquals(tc, rec->term, c1->term);
    CuAssertMemEquals(tc, rec->digest, c1->digest, 20);
    free(rec->term);
    free(rec->digest);
    free(rec);


    rec = dequeue(q);
    CuAssertPtrNotNull(tc, rec->term);
    CuAssertStrEquals(tc, rec->term, c2->term);
    CuAssertMemEquals(tc, rec->digest, c2->digest, 20);
    free(rec->term);
    free(rec->digest);
    free(rec);

    remove("testfile.zmq");
}


void test_trailing_zeros(CuTest *tc) {
    int fd = open("testfile.zmq", O_CREAT | O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

    tCell *c1 = gen_cell("foobar123");
    write_cell(fd, c1);

    tCell *c2 = gen_cell("this_is_yet_another_term");
    write_cell(fd, c2);

    char *block = malloc(1000);
    memset(block, 0, 1000);
    write(fd, block, 1000);
    free(block);
    close(fd);
    
    wQueue *q = recover_file("testfile.zmq");

    CuAssertPtrNotNull(tc, q);
    CuAssertIntEquals(tc, q_size(q), 2);

    tCell *rec = dequeue(q);
    CuAssertPtrNotNull(tc, rec->term);
    CuAssertStrEquals(tc, rec->term, c1->term);
    CuAssertMemEquals(tc, rec->digest, c1->digest, 20);
    free(rec->term);
    free(rec->digest);
    free(rec);


    rec = dequeue(q);
    CuAssertPtrNotNull(tc, rec->term);
    CuAssertStrEquals(tc, rec->term, c2->term);
    CuAssertMemEquals(tc, rec->digest, c2->digest, 20);
    free(rec->term);
    free(rec->digest);
    free(rec);

    remove("testfile.zmq");
}


void test_many_zeros(CuTest *tc) {
    int fd = open("testfile.zmq", O_CREAT | O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

    char *block = malloc(1000);
    memset(block, 0, 1000);
    write(fd, block, 1000);

    tCell *c1 = gen_cell("foobar123");
    write_cell(fd, c1);

    write(fd, block, 1000);

    tCell *c2 = gen_cell("this_is_yet_another_term");
    write_cell(fd, c2);

    write(fd, block, 1000);
    free(block);
    close(fd);
    
    wQueue *q = recover_file("testfile.zmq");

    CuAssertPtrNotNull(tc, q);
    CuAssertIntEquals(tc, q_size(q), 2);

    tCell *rec = dequeue(q);
    CuAssertPtrNotNull(tc, rec->term);
    CuAssertStrEquals(tc, rec->term, c1->term);
    CuAssertMemEquals(tc, rec->digest, c1->digest, 20);
    free(rec->term);
    free(rec->digest);
    free(rec);


    rec = dequeue(q);
    CuAssertPtrNotNull(tc, rec->term);
    CuAssertStrEquals(tc, rec->term, c2->term);
    CuAssertMemEquals(tc, rec->digest, c2->digest, 20);
    free(rec->term);
    free(rec->digest);
    free(rec);

    remove("testfile.zmq");
}


void RunAllTests(void) 
{
    CuString *output = CuStringNew();
    CuSuite* suite = CuSuiteNew();


    SUITE_ADD_TEST(suite, test_one_wp_is_all);
    SUITE_ADD_TEST(suite, test_two_wps_block);
    SUITE_ADD_TEST(suite, test_leading_zeros);
    SUITE_ADD_TEST(suite, test_trailing_zeros);
    SUITE_ADD_TEST(suite, test_many_zeros);

    CuSuiteRun(suite);
    CuSuiteSummary(suite, output);
    CuSuiteDetails(suite, output);
    printf("%s\n", output->buffer);
    CuStringDelete(output);
    CuSuiteDelete(suite);
}

int main(void)
{
    RunAllTests();
    return 0;
}

