#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "../../../include/zmq.h"
#include "../../../include/czmq.h"

#include "../../../common/queue.h"
#include "../../../common/hashtrie.h"
#include "../../../common/ring.h"

#include "wtok.h"
#include "wwork.h"
#include "whash_recv.h"
#include "wdebug.h"

static void enqueue_packages(zmsg_t *msg);
static void append_wp(zmsg_t *wpmsg);
static void wp_sendmore(zmsg_t *msg, char *term, int len, char *digest);

extern wQueue *local_queue;
extern char *worker_id;

static char *filename = NULL;

static void *recv_work = NULL;
static int fd = -1;

static void send_work(int max_packages);

int h_work_recv (zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    DEBUG_PRINT(DEBUG_VERBOSE,,,":enter :h_work_recv");
    zmsg_t *answer = zmsg_recv(recv_work);
    if (zmsg_size(answer) == 1) {
        assert(filename);
        assert(fd >= 0);
        char *s = zmsg_popstr(answer);
        assert(0 == strcmp(s, "ACK"));
        free(s);
        int rc = close(fd);
        assert(rc == 0);
        remove(filename);
        free(filename);

        filename = NULL;
        fd = -1;
    } else if (zmsg_size(answer) == 2 && zframe_size(zmsg_last(answer)) <= 20) {
        // TODO: make this one frame
        assert(filename == NULL);
        char *s = zmsg_popstr(answer);
        assert(0 == strcmp(s, "give work"));
        free(s);
        s = zmsg_popstr(answer);
        int max_packages = atoi(s);
        free(s);
        zmsg_destroy(&answer);

        // enforce FIFO freeing in ring buffer
        clear_durable_queue();
        send_work(max_packages);
    } else {
        enqueue_packages(answer);
        zmsg_destroy(&answer);
        tick();
    }
    DEBUG_PRINT(DEBUG_VERBOSE,,,":exit :h_work_recv");
    return 0;
}

static void enqueue_packages(zmsg_t *msg) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :enqueue_packages");
    int i;
    char *from = zmsg_popstr(msg);

    int size = (int) zmsg_size(msg);
    for (i = 0; i < size; i++) {
        zframe_t *frame = zmsg_pop(msg);
        char *str = zframe_strdup(frame);
        char *digest = (char*) malloc(20);
        int wplen = zframe_size(frame) - 20;
        assert(wplen > 0);
        char *term = (char*) malloc(wplen);
        
        memcpy(term, str + 20, wplen);
        memcpy(digest, str, 20);   
        enqueue(local_queue, term, wplen, digest);
        free(str);
        zframe_destroy(&frame);
    }
    zstr_send(recv_work, "ACK %s", from);
    free(from);
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :enqueue_packages");
}


static void send_work(int max_packages) {
    if (q_size(local_queue) < 2) {
       zstr_send(recv_work, "done.");
       return;
    }  

    zmsg_t *wpmsg = zmsg_new();


    int len = local_queue->c;
    int i;

#ifndef TMPFOLDER
#define TMPFOLDER "/var/tmp/"
#endif
    filename = tempnam(TMPFOLDER, "zmqw-");
    assert(filename);
    fd = open(filename, O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
    assert(fd >= 0);
    for(i = 0; i < len ; i++) {
        if (i % 2 == 0 && i / 2 <= max_packages) {
            append_wp(wpmsg);
        } else {
            tCell *item = dequeue(local_queue);
            enqueue(local_queue, item->term, item->term_length, item->digest);
            ring_free(local_queue->env, item);
            free(item);
        }
    }

    zmsg_pushstr(wpmsg, worker_id);
    zmsg_send(&wpmsg, recv_work);
    zmsg_destroy(&wpmsg);
    
}

static void append_wp(zmsg_t *wpmsg) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :append_wp");
    tCell *item = NULL;
    if (!is_empty(local_queue)) {
        item = dequeue(local_queue); 
    }
    assert(item);

    if (!trie_contains_processed(item->digest)) {
        wp_sendmore(wpmsg, item->term, item->term_length, item->digest);
    }
    int rc = write(fd, "X", 1);
    assert(rc == 1);
    size_t tmp = (size_t) item->term_length;
    rc = write(fd, &tmp, sizeof(size_t));
    assert(rc == sizeof(size_t));
    rc = write(fd, item->digest, 20);
    assert(rc == 20);
    rc = write(fd, item->term, item->term_length);
    assert(rc == item->term_length);

    ring_free(local_queue->env, item);
    free(item->term);
    free(item->digest);
    free(item);
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :append_wp");
}

static void wp_sendmore(zmsg_t *msg, char *term, int len, char *digest) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,
                char d[41];print_readable_digest((unsigned char*) digest, d);,,
                ":enter :wp_sendmore :digest :%s", d);
    char *content = (char*) malloc(20 + len);
    memcpy(content, digest, 20);
    memcpy(content+20, term, len);
    zmsg_addmem(msg, content, 20 + len);
    free(content);
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :wp_sendmore");
}

// this needs some love
// socket: send work request / poller -> recv_work
// socket: send work
void init_work_handler(zctx_t *ctx, zmq_pollitem_t *poller_get,
                       char *masteraddr, char *id, int port) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :init_work_handler");

    assert(recv_work == NULL);

    recv_work = zsocket_new (ctx, ZMQ_DEALER);

    zsocket_set_identity(recv_work, id);

    int rc = zsocket_connect(recv_work,"tcp://%s:%i", masteraddr, port+2);
    assert(rc == 0);

    *poller_get = (zmq_pollitem_t) { recv_work, 0, ZMQ_POLLIN };
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :init_work_handler");
}

void teardown_work_handler(zctx_t *ctx) {
    assert(recv_work != NULL);
    zsocket_destroy(ctx, recv_work);
    recv_work = NULL;
}

void signal_ready(char *id) {
    int acknowledged = 0;
    while (!acknowledged) {
        zstr_send(recv_work, "OHAI %s", worker_id);
        if (zsocket_poll(recv_work, 1000) == true) {
            char *rep = zstr_recv(recv_work);
            assert(0 == strcmp(rep, "OHAI-OK"));
            free(rep);
            acknowledged = 1;
        }
    }
}
