% (c) 2012-2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(ample_sets, [
      compute_dependendency_relation_of_all_events_in_the_model/3,
	    compute_ample_set2/5,
      enable_graph/1,
      dependent_act/2, dependent_actions/3,
      clear_dynamic_predicates_for_POR/0,
      reset_runtime_dynamic_predicates_POR/0,
      stutter_action/1, visible_action/1 % exported for ltl.pl for asserting and retracting stutter/visible actions
      ]).

/* SICSTUS Libraries */
:- use_module(library(ordsets)).
:- use_module(library(random)).
:- use_module(library(lists)).
:- use_module(library(ugraphs)).
:- use_module(library(timeout)).

/* B module sources */
:- use_module(probsrc(bmachine),[b_get_invariant_from_machine/1,b_operation_preserves_full_invariant/1,b_top_level_operation/1]).
:- use_module(probsrc(bsyntaxtree),[conjunction_to_list/2,occurs_in_expr/2]).
:- use_module(probsrc(specfile),[prepare_state_for_specfile_trans/2,specfile_possible_trans_name/2,specfile_trans/6]).
:- use_module(probsrc(state_space),[visited_expression/2,use_no_timeout/1,assert_time_out_for_node/3]).

/* The Enabling Analysis' module */
:- use_module(probsrc(enabling_analysis),[dependent_actions/4,b_get_read_write/3,enable_analysis/6]).
:- use_module(probsrc(enable_graph)).

:- use_module(probsrc(specfile),[state_corresponds_to_initialised_b_machine/2]).
:- use_module(probsrc(state_space),[visited_expression/2]).
:- use_module(probsrc(store), [empty_state/1]).

/* Other modules from ProB */
:- use_module(probsrc(debug),[debug_println/2]).
:- use_module(probsrc(preferences), [get_preference/2]).
:- use_module(probsrc(error_manager), [time_out_with_enum_warning_for_findall/3]).
:- use_module(probsrc(tools),[cputime/1]).

/* Modules and Infos for the code coverage analysis */
:- use_module(module_information).
:- module_info(group,model_checker).
:- module_info(description,'This module contains predictes for partial order reduction using the ample sets\' theory.').

% Dynamic predicates for keeping track of visible and stutter actions, 
% as well as of all dependent reations between the actions of the B model
:- dynamic stutter_action/1,visible_action/1,dependent_act/2,enables/2,expanded_state/1.

:- dynamic enable_graph/1.

clear_dynamic_predicates_for_POR :-
    reset_static_dynamic_predicates_POR,
    reset_runtime_dynamic_predicates_POR.

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(clear_specification,clear_dynamic_predicates_for_POR,
                           'Reset Ample Sets.').
:- register_event_listener(reset_specification,reset_static_dynamic_predicates_POR,
                           'Reset Ample Sets.').

% predicates that will be asserted once on every loading of the machine
reset_static_dynamic_predicates_POR :- 
   retractall(dependent_act(_,_)),
   retractall(enable_graph(_)).

% predicates that will be asserted at runtime
reset_runtime_dynamic_predicates_POR :-
    retractall(stutter_action(_)),
    retractall(visible_action(_)),
    retractall(expanded_state(_)),
    retractall(enables(_,_)). % used to not track in the enable graph

/* This implementation of the ample sets approach sastisfies all four conditions of the ample sets theory (not proved yet).
   The conditions are taken from the book "Principles of Model Checking" written by C.Baier and J.P.Katoen.

For every state s in the state space, where Act(s) stands for all enabled actions in state s, the following conditions 
must be satisfied:

1. Nonemptiness condition (A1) and terminal states
   (A 1.1) ample(s) <: Act(s) (set equality is possible)
   (A 1.2) ample(s) = {} <=> Act(s) = {}

2. Dependency condition
   (A 2) Let s -a_{1}> s_{1} -a_{2}> ... -a_{n}> s_{n} -a> s' be a finite execution
         fragment in the original state space. If a depends on ample(s), then a_{i}:ample(s) for some 0<i<=n.
   
   This condition can be satisfied by the following conditions (Local Criteria):
      (A 2.1) Any action b: Act(s)\ample(s) is independent of any a:ample(s) (Here we use the dependency graph generated from the Enabling Analysis)
      (A 2.2) Any disabled action b:Act\en(s) which is dependent on ample(s) cannot become enabled by a possible sequence of actions starting with an action
   	 		  from en(s)\ample(s). (We check this condition by means of using the enable graph generated from the Enabling Analysis)

      Note that determining the local criteria's conditions above is dependent on the formalism on which we are applying the partial order reduction method.
      The conditions (A 2.1) and (A 2.2) are adjusted to the semantics of the B method/Event-B. For other formalisms such as CSP we will need to define
      the conditions differently. From one hand because we cannot use the Constraint Based Checker of ProB (it is not implemented for CSP yet), from other hand
      because CSP differs semantically from B/Event-B (the process structure). An idea to apply Partial Order Reduction on CSP is to make the reduction
      on the CSP operator-level ('|||', [|X|] and so on). For more detailed description of how to apply POR on CSP read the following papers:
            - H. Wehrheim. Partial order reductions for failure refinement.
            - J. Sung, Y. Liu and J.S. Dong. Model Checking CSP Revisited: Introducing a Process Analysis Toolkit.

3. Stutter condition
   (A 3) If ample(s) /= Act(s) then any action t:ample(s) is a stutter action.

4. Cycle condition
   (A 4) For any cycle s_{0}s_{1}...s_{n} in the reduced state space and actions t:Act(s_{i}), for some 
         0 < i <= n, there exists j:{1..n} such that t:ample(s_{j}). (s_{0},...,s_{n} are states)

 */

% compute_ample_set2(+CurID,+CurState,+FindInvViolations,+EnableGraph,+DisableGraph,+WithPGE)
%
% The main predicate for computing ample sets by respecting the four ample set conditions specified above.
% A subset of all possible enabled outgoing transitions in CurState will be computed and added to the state space. The predicate is similar
% to the compute_all_transitions/2 predicate implemented in tcltk_interface.pl.

% Arguments:
% CurID - the current ID of the state.
% CurState - the current evaluation of all global variables in the machine.
% FindInvVionlations - possible values 1 (option setted) and 0 (option not setted). If FindInvViolations equals to 0
%                      the stutter condition (A 3) will be not checked.
% DisableGraph - the disable graph for the PGE optimisations. If option WithPGE is on then a combined optimisation using POR and PGE will be started.
% WithPGE - the option for using the PGE optimisation. Possible values are with_pge and without_pge.
compute_ample_set2(CurID,CurState,FindInvViolations,DisableGraph,POROptions) :-
   user: reset_max_reached,
   prepare_state_for_specfile_trans(CurState,PreparedCurState),
   % getting all enabled actions at CurID
   get_all_possible_action_names_from_current_state(CurID,PreparedCurState,DisableGraph,WithPGE,Actions),
   % only top level action names will be considered
   get_action_names_only(Actions,ActionNames),
   % mark state as visited (Cycle Condition (A 4))
   % we want only unique action names for analyzing and computing of ample sets
   list_to_ord_set(ActionNames,AllEnabledActions),
   % compute a sound ample set from all actions enabled in state
   compute_ample_actions(CurID,AllEnabledActions,FindInvViolations,POROptions,AmpleSet),
   debug_println(20,sets(CurID,allActions(AllEnabledActions),ampleSet(AmpleSet))),
   assert(expanded_state(CurID)),
   ( (AmpleSet \= []) ->
	    % compute just the actions from the ample set (case of state space reduction)
        % AmpleSet is a non-empty subset of enabled(CurID), AmpleSet can be equal to AllEnabledActions
	    add_transitions(AmpleSet,CurID,Actions,FindInvViolations,WithPGE)
   ; otherwise ->
	    true /* DEADLOCK */
   ),
   (user: max_reached ->
	    user: assert_max_reached_for_node(CurID),
	    MaxReached=true
   ;    MaxReached=false
   ),
   (CurID==root ->
	    user: save_constants(MaxReached)
   ;    user: add_new_transitions_to_cache(CurID)
   ),!. % once we have computed the ample set in CurState we don't want to backtrack

% computing a sound ample set of actions.
compute_ample_actions(CurID,AllEnabledActions,FindInvViolations,POROptions,AmpleSet) :- 
   % get_ample_actions/5 call only reasonable to be executed when more than one actions are enabled in CurID
   ((length(AllEnabledActions,L),L>1) -> 
	    get_ample_actions(CurID,AllEnabledActions,FindInvViolations,POROptions,AmpleSet)
   ; otherwise ->
        AmpleSet = AllEnabledActions
   ).

/************************** ADD TRANSITIONS (BEGIN) ******************************/

% Adding transitions from the particular ample set to the state space.
add_transitions([],_CurID,_Actions,_FindInvViolations,_WithPGE).
add_transitions([ActionName|ActNames],CurID,Actions,FindInvViolations,WithPGE) :-
   add_action_name_transitions(Actions,CurID,ActionName,FindInvViolations,WithPGE,[],RemActions),
   add_transitions(ActNames,CurID,RemActions,FindInvViolations,WithPGE).

% An event can have parameters, hence there can be multiple actions with the same ActionName.
% e.g. act1(1),act1(2),act1(3),...
% Therefore, we must add every single action template with same action name to the state space.
add_action_name_transitions([],_CurID,_ActionName,_FindInvViolations,_WithPGE,RemActions,RemActions).
add_action_name_transitions([(ActionName1,Act,NewExpression,TransInfo,Residue)|R],CurID,ActionName,FindInvViolations,WithPGE,IgnoredActions_so_far,RemActions) :-
   (ActionName == ActionName1 ->
        user: add_trans_id(CurID,Act,NewExpression,Residue,NewId,TransInfo,_),
	    (WithPGE==with_pge -> assert(user: back_transition(NewId,ActionName,CurID)) ; true),
        % if successor state already visited we have to explore all actions
        (cycle_condition_check_positive(FindInvViolations,CurID,NewId) ->
            % Case in which (A 4) can be violited,
            % therefore we must explore all possible actions in the current state.
            ord_union(IgnoredActions_so_far,R,AllRemainedActions),
            debug_println(9,computing_all_other_enabled_actions(back_trans(ActionName),AllRemainedActions)),
            compute_all_other_enabled_actions(AllRemainedActions,CurID,WithPGE),
            %compute_all_enabled_actions(CurID,CurState,DisableGraph,WithPGE),
     	    debug_println(9,'*********** CYCLE DETECTED *************'),
            RemActions=[]
        ; otherwise ->
            % add only the actions from the computed ample set
     	    add_action_name_transitions(R,CurID,ActionName,FindInvViolations,WithPGE,IgnoredActions_so_far,RemActions)
        )
   ; otherwise ->
        % keeping track of all ignored actions so far, we don't want to go through the same order of actions every time we
        % want to add a transition from AmpleSet to the state space
        ord_add_element(IgnoredActions_so_far, (ActionName1,Act,NewExpression,TransInfo,Residue), IgnoredActions_so_far1),
	    add_action_name_transitions(R,CurID,ActionName,FindInvViolations,WithPGE,IgnoredActions_so_far1,RemActions)
   ).

cycle_condition_check_positive(FindInvViolations,CurID,NewId) :-
    FindInvViolations==1,
    \+var(NewId),
    %visited_expression(NewId,_State),
    (CurID >= NewId),
    expanded_state(NewId).
    %\+not_all_transitions_added(NewId).

compute_all_other_enabled_actions([],_CurID,_WithPGE).
compute_all_other_enabled_actions([(ActionName,Act,NewExpression,TransInfo,Residue)|R],CurID,WithPGE) :-
    user: add_trans_id(CurID,Act,NewExpression,Residue,NewId,TransInfo,_),
    (WithPGE==with_pge -> assert(user: back_transition(NewId,ActionName,CurID)) ; true),
    compute_all_other_enabled_actions(R,CurID,WithPGE).

/************************** ADD TRANSITIONS (END) ******************************/

/* Auxiliary predicates for getting action names enabled at the current state */ 

% getting action names from the action tuples
get_action_names_only(Actions,ActionNames) :-
    maplist(get_action_name,Actions,ActionNames).

get_action_name((ActionName,_A,_NewE,_TrInf,_Residue),ActionName).

% Getting all enabled actions in the current state (check performance)
get_all_possible_action_names_from_current_state(CurID,CurState,DisableGraph,WithPGE,AllActions) :- 
   findall((ActionName,Act,NewExpression,TransInf,Residue),
	(specfile_possible_trans_name(CurState,ActionName), % getting possible action names from the loaded B/Event-B model
	 spec_file_trans_if_enabled(CurID,CurState,ActionName,Act,NewExpression,TransInf,Residue,WithPGE,DisableGraph)),
		AllActions).

spec_file_trans_if_enabled(CurID,CurState,ActionName,Act,NewExpression,TransInf,Residue,WithPGE,DisableGraph) :-
  ((WithPGE==with_pge) ->
         specfile_trans_pge(CurID,DisableGraph,CurState,ActionName,Act,NewExpression,TransInf,Residue)
  ; otherwise ->
         % How should we proceed with the ample computations when a time out occurs?
         specfile_trans_timeout(CurID,CurState,ActionName,Act,NewExpression,TransInf,Residue)
  ).

specfile_trans_pge(CurID,DisableGraph,CurState,ActionName,Act,NewExpression,TransInf,Residue) :-
    user: get_action_neighbors(ActionName, DisableGraph, Neighbors),
    action_not_disabled(ActionName,CurID,Neighbors),
    specfile_trans_timeout(CurID,CurState,ActionName,Act,NewExpression,TransInf,Residue).

action_not_disabled(ActionName,CurID,Neighbors) :-
    user: action_disabled(ActionName,CurID,Neighbors),
    debug_println(9,action_disabled(CurID,ActionName)),
    !,fail.
action_not_disabled(_ActionName,_CurID,_Neighbors).

specfile_trans_timeout(CurID,CurState,ActionName,Act,NewExpression,TransInf,Residue) :-
    (use_no_timeout(CurID) ->
        specfile_trans(CurState,ActionName,Act,NewExpression,TransInf,Residue)
    ;   preferences:preference(time_out,CurTO),
        time_out_with_enum_warning_for_findall(specfile_trans(CurState,ActionName,Act,NewExpression,TransInf,Residue),CurTO,TimeOutRes),
        user: check_timeout_result(CurID,ActionName,CurTO,TimeOutRes),
        Residue==[]
    ).
/* Auxiliary predicates for getting action names enabled at the curent state */ 

/************************** DEPENDENCY RELATION (BEGIN) ******************************/
    
:- use_module(error_manager,[call_in_fresh_error_scope_for_one_solution/1]).

/* predicate which determines the dependency relations between all actions in the model,
   it should be called only once (prior to the model checking) */

compute_dependendency_relation_of_all_events_in_the_model(FindInvViolations,por(_TYPE,UseEnableGraph,Depth,_PGE),EnableGraph) :-
   (enable_graph(EnableGraph) ->
        % In case the dependency relations and the enable graph are already computed.
        % Case possible if the user stopped the model checking for a while (and meanwhile didn't reloaded the model)
        % and in case that he wants to continue the verification of the model the enable graph don't have
        % to be computed again.
        true
   ; otherwise ->
	      findall(Action, b_top_level_operation(Action),Actions),
        sort(Actions,AllActions),
   	    retractall(enable_graph(_)),
   	    print('********** DETERMINE ACTION DEPENDENCIES AND ENABLE GRAPH *************'),nl,
        get_preference(timeout_cbc_analysis,Timeout),
   	    cputime(T1),
        (UseEnableGraph = true ->
            call_in_fresh_error_scope_for_one_solution(determine_enabling_relations(AllActions,Timeout,FindInvViolations,Depth,EnableGraph))
   	    ; otherwise ->
            call_in_fresh_error_scope_for_one_solution(determine_dependent_enable_action_relations(AllActions,Timeout,FindInvViolations,EnableGraph))
        ),
        cputime(T2),
   	    print('********** FINISHED ANALYSIS *************'),nl,
   	    print('Analysis Checking Time: '), D is T2-T1, print(D),print(' ms.'),nl,
   	    assert(enable_graph(EnableGraph))
   ).

create_enable_graph(FindInvViolations,EnableGraph) :-
    findall(Action, b_top_level_operation(Action),Actions),
    sort(Actions,AllActions),
    get_preference(timeout_cbc_analysis,Timeout),
    determine_enabling_relations(AllActions,Timeout,FindInvViolations,1,EnableGraph),
    print(enable_graph(EnableGraph)),nl.

determine_enabling_relations(AllActions,Timeout,FindInvViolations,Depth,EnableGraph) :- 
    determine_enabling_relations(AllActions,AllActions,Timeout,FindInvViolations,Depth,Edges),
    enable_graph: vertices_edges_predicates_to_egraph(AllActions,Edges,EnableGraph).

determine_enabling_relations([],_AllActions,_Timeout,_FindInvViolations,_Depth,[]).
determine_enabling_relations([Act|Actions],AllActions,Timeout,FindInvViolations,Depth,Edges) :-
    determine_enabling_relations1(AllActions,Act,Timeout,FindInvViolations,Depth,Edges1),
    append(Edges1,Rest,Edges),
    determine_enabling_relations(Actions,AllActions,Timeout,FindInvViolations,Depth,Rest).

determine_enabling_relations1([],_Act,_Timeout,_FindInvViolations,_Depth,[]).
determine_enabling_relations1([Act2|Acts],Act1,Timeout,FindInvViolations,Depth,Result) :-
    ( (Act1==Act2,\+dependent_act(Act1,Act1)) ->
       assert(dependent_act(Act1,Act1))
    ; dependent_actions(Act1,Act2,Timeout) ->
       assert(dependent_act(Act1,Act2)), % the dependency relation is symmetric
       assert(dependent_act(Act2,Act1))
    ; otherwise ->
       true
    ),
    (may_enable(Act1,Act2,FindInvViolations,Timeout,Depth,Predicate) ->
       Result = [Act1-Predicate-Act2|Res1]
    ; otherwise ->
       Result = Res1
    ),
    determine_enabling_relations1(Acts,Act1,Timeout,FindInvViolations,Depth,Res1).

may_enable(Act1,Act2,FindInvViolations,Timeout,Depth,Predicate) :- 
    enable_analysis(Act1,Act2,FindInvViolations,true,Timeout,Enable),
    (memberchk(Enable,[guaranteed,possible,possible_enable]) ->
        Predicate = b(truth,pred,[])
    ; Enable = pred(Expr) ->
        (Depth=0 -> Predicate=b(truth,pred,[]); Predicate = Expr)
    ; otherwise ->
        fail
    ).

determine_dependent_enable_action_relations(AllActions,Timeout,FindInvViolations,EnGraph) :- 
   vertices_edges_to_ugraph([],[],Graph),
   determine_dependent_relations_and_enable_graph(AllActions,AllActions,Timeout,FindInvViolations,Graph,EnGraph).

/* auxiliary predicates (determine_dependent_relations_graph/4 and determine_enable_graph_and_dependency_wrt_one_action/4) 
   for computing the enable graph */
determine_dependent_relations_and_enable_graph([],_AllActions,_Timeout,_FindInvViolations,Graph,Graph).
determine_dependent_relations_and_enable_graph([Action|Actions],AllActions,Timeout,FindInvViolations,Graph,EnGraph) :-
   determine_enable_graph_and_dependency_wrt_one_action(AllActions,Action,Timeout,FindInvViolations,Graph,ResGraph),
   determine_dependent_relations_and_enable_graph(Actions,Actions,Timeout,FindInvViolations,ResGraph,EnGraph).

determine_enable_graph_and_dependency_wrt_one_action([],_Act,_Timeout,_FindInvViolations,Graph,Graph).
determine_enable_graph_and_dependency_wrt_one_action([Act2|Actions],Act1,Timeout,FindInvViolations,Graph,ResGraph) :-
   % dependent realtions should be asserted only once
   ( (Act1==Act2,\+dependent_act(Act1,Act1)) ->
       assert(dependent_act(Act1,Act1))
   ; dependent_actions(Act1,Act2,Timeout) ->
	   assert(dependent_act(Act1,Act2)), % the dependency relation is symmetric
	   assert(dependent_act(Act2,Act1))
   ; otherwise ->
	    true
   ),
   add_enable_edge(Act1,Act2,FindInvViolations,Timeout,Graph,Graph1),
   add_enable_edge(Act2,Act1,FindInvViolations,Timeout,Graph1,Graph2),
   determine_enable_graph_and_dependency_wrt_one_action(Actions,Act1,Timeout,FindInvViolations,Graph2,ResGraph).

add_enable_edge(Src,Dest,FindInvViolations,Timeout,Graph,ResGraph) :-
    ( may_enable1(Src,Dest,FindInvViolations,Timeout,Graph) ->
 	    add_vertices(Graph,[Src,Dest],Graph1), % we don't need this predicate call
 	    add_edges(Graph1,[Src-Dest],ResGraph)
    ; otherwise ->
 	    ResGraph=Graph
    ).

may_enable1(Act1,Act2,_FindInvViolations,_Timeout,Graph) :-
    neighbors(Act1, Graph, Neighbors),
    memberchk(Act1-Act2,Neighbors),!,fail. % enabling relation is already determined
may_enable1(Act1,Act2,FindInvViolations,Timeout,_Graph) :-
    enable_analysis(Act1,Act2,FindInvViolations,false,Timeout,Enable),
    ( (Enable == guaranteed)
    ; (Enable == possible)
    ; (Enable == possible_enable)
    ; (get_preference(use_optimistic_por,false),Enable == timeout)  % pessimistic approach
    ).

/* 
   dependent_actions/2 determine statically or dynamically if two actions are dependent
   two events/actions are syntactically dependent in the following three cases:
	1. If both events modify at least one common variable.
	2. Act1 modifies a variable in the guard of Act2. (Act1 can enable or disable Act2)
	3. Act2 modifies a variable in the guard of Act1. (Act2 can enable or disable Act1)
*/

%%% Act1 and Act2 are enabled in current state State,
%%% we don't have to check if both actions are enabled in 
%%% some state

dependent_actions(Act1,Act2,_Timeout) :-
	dependent_act(Act1,Act2),!,fail. % dependent relation is already computed
dependent_actions(Act1,Act2,Timeout) :-
	dependent_actions(Act1,Act2,Timeout,Res),
	((Res == '-') -> 
		dependent_actions(Act2,Act1,Timeout)
	; (Res == syntactic_independent) -> 
		fail
	; (Res == independent) ->
		fail
	; otherwise ->
		true
	).

/************************** DEPENDENCY RELATION (END) ******************************/

%---------------------------------------------------------------------------------------------------------------------------------
/************************** DETERMINING POSSIBLE AMPLE SETS (BEGIN) ******************************/

/************************** PREDICATES COMPUTING AMPLE SETS WITH RESPECT TO CONDITIONS (A 2) and (A 3) *************************/

/* 

   implementation of the ample-sets approach satisfying the dependency condition (A 2)
   by using enabling and dependence analysis (see module enabling analysis)

*/

get_ample_actions(CurID,AllEnabledActions,FindInvViolations,POROptions,AmpleSet) :-
    get_possible_ample_actions(AllEnabledActions,CurID,AllEnabledActions,FindInvViolations,POROptions,AmpleSet).

get_possible_ample_actions([],_CurID,AllEnabledActions,_FindInvViolations,_POROptions,AllEnabledActions).
get_possible_ample_actions(Actions,CurID,AllEnabledActions,FindInvViolations,POROptions,AmpleSet) :-
    % choosing of action for computing a possible ample set for CurID should be random
    random_select(Act,Actions,RemainedActions),
    get_dependent_actions(CurID,POROptions,Act,AllEnabledActions,DepActs), % finding set (DepActs) satisfying (A 2.1) and (A 2.2)
    (check_stutter_condition(DepActs,FindInvViolations,AllEnabledActions) -> % checking (A 3)
        AmpleSet = DepActs
    ; get_possible_ample_actions(RemainedActions,CurID,AllEnabledActions,FindInvViolations,POROptions,AmpleSet)
    ).

check_stutter_condition(AmpleSet,FindInvViolations,AllEnabledActions) :-
    % in case that we don't check for invariant violations, we don't have to prove the stuttering condition for AmpleSet
    ((FindInvViolations==0;ord_seteq(AmpleSet,AllEnabledActions)) ->
        true
    ; otherwise ->
        stutter_actions(AmpleSet)
    ).

% get_dependent_actions(+CurID,+POROptions,+Act,+AllEnabledActions,+EnGraph,-DepActs)
% CurID - the ID of the current state (used here only for debugging purposes)
% POROptions - partial order reduction options (e.g. UseEnableGraph, Depth)
% Act - the action with which the current dependency class will be computed
% AllEnabledActions - the set of all in CurID enabled actions and it is an ordinary set
% DepActs - the computed candidate for ample set
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% EnableGraph - the enable graph for all possible enable actions' relations in the current machine. The enable graph will
%               be computed only single time (before starting with the verification of the state space).
get_dependent_actions(CurID,POROptions,Act,AllEnabledActions,DepActs) :-
    % finding all dependent actions to Act (A 2.1)
    % DepActions_so_far is an ordinary set because AllEnabledActions is an ordinary set and every 'dependent_act(Act1,Act)' is uniquely asserted
    %% (CurID == 4 -> trace ; true),
    findall(Act1,(member(Act1,AllEnabledActions), dependent_act(Act,Act1)),DepActions_so_far),
    ord_subtract(AllEnabledActions,DepActions_so_far,IndActions),
    (IndActions == [] ->
        DepActs=DepActions_so_far
    ; otherwise ->
        enable_graph(EnableGraph),
        vertices(EnableGraph,Vertices), % getting all possible actions that can be enabled
        % DisabledActions = Vertices\AllEnabledActions
        ord_subtract(Vertices,AllEnabledActions,DisabledActions),
        % finding all independent actions to Act that could enable a currently disabled action that is dependent to Act (A 2.2)
        findall(Act2,(member(Act2,IndActions),may_enable_dep_action(Act2,DepActions_so_far,DisabledActions,EnableGraph,CurID,POROptions)),DepEnActions),
        ord_union(DepActions_so_far,DepEnActions,DepActs)
        ,debug_println(9,ord_union(CurID,DepActions_so_far,DepEnActions,DepActs))
    ).

/* searching for enabling path beginning from an action which is not from AmpleSet and 
	which enables action dependent on AmpleSet */
may_enable_dep_action(Act,AmpleSet,DisabledActions,EnableGraph,CurID,POROptions) :-
   member(DisAct,DisabledActions),
   % we check here only currently disabled actions dependent on AmpleSet that can be enabled later
   depends_on_ample_set(DisAct,AmpleSet,_DepAct),
   % is there a path that can enable Act1, which is dependent on Act
   enables_action(CurID,Act,DisAct,AmpleSet,EnableGraph,POROptions),
   !.

enables_action(CurID,Act,DisAct,AmpleSet,EnableGraph,por(_TYPE,UseEnableGraph,Depth,_PGE)) :-
    ( enables(Act,DisAct) ->
        % should improve performance a little bit (not applicable when using enable graphs)
        true
    ; check_if_enabled(UseEnableGraph,Depth,CurID,Act,DisAct,AmpleSet,EnableGraph) ->
        true
    ; otherwise ->
        fail
    ).

check_if_enabled(true,Depth,CurID,Act,DisAct,AmpleSet,EnableGraph) :- !,
    enables_pred_action(Act,DisAct,AmpleSet,CurID,EnableGraph,Depth).
check_if_enabled(false,_Depth,CurID,Act,DisAct,AmpleSet,EnableGraph) :-
    min_path(Act,DisAct,EnableGraph,P,L),
    sets_disjoint(AmpleSet,P),
    debug_print_path(CurID,P,L),
    assert(enables(Act,DisAct)).

enables_pred_action(Act,DisAct,AmpleSet,CurID,EnableGraph,Depth) :-
    % visited_expression/2 and state_corresponds_to_initialised_b_machine/2 should be called once for each state,
    % avoid multiple calls per state
    (Depth==0 ->
       empty_state(ST) % empty state for 
    ; otherwise ->
       visited_expression(CurID,RawState),
       state_corresponds_to_initialised_b_machine(RawState,ST)
    ),
    Pred =.. [evaluate_weakest_precondition_for_por, ST, AmpleSet],
    enable_graph: path(EnableGraph,Act,DisAct,enable_graph,Pred,Depth,Path,Length),
    debug_print_path(CurID,Path,Length).

% guaranteeing that the dependent action to AmpleSet has been enabled from an action off AmpleSet
sets_disjoint(AmpleSet,Path) :-
    list_to_ord_set(Path,OrdSet),
    \+ord_intersection(AmpleSet,OrdSet).

depends_on_ample_set(Act,AmpleSet,Act1) :-
   member(Act1,AmpleSet),
   dependent_act(Act,Act1).

debug_print_path(CurID,Path,Length) :-
   debug_println(20,'***** Enable Path *****'),
   debug_println(20,path(CurID,Path,Length)),
   debug_println(20,'***********************').

/************************** DETERMINING POSSIBLE AMPLE SETS (END) ******************************/

/************************** CHECKING IF ALL ACTIONS FROM AMPLE SET ARE STUTTER ACTIONS (BEGIN) WITH REGARD TO THE INVARIANT ******************************/
/************************** PREDICATES WHICH SATISFY CONDITION (A3) ******************************/

stutter_actions(Actions) :-
    maplist(is_stutter_action,Actions).

% visible actions are non stutter actions
is_stutter_action(Act) :- 
   ( stutter_action(Act) -> 
	     debug_println(9,stutter_action(Act))
   ; visible_action(Act) -> 
	     debug_println(9,visible_action(Act)),
	     fail
   ; otherwise -> 
	     check_if_action_is_stutter_wrt_inv(Act)
   ),!.

% A stutter action cannot change the current invariant value.
% That's why we must check if the particular action preserves
% the full invariant in order to determine if the action is
% stutter under the following LTL formula:
% G Phi, where Phi is the invariant.
check_if_action_is_stutter_wrt_inv(Act) :- 
    b_operation_preserves_full_invariant(Act),!,
    assert(stutter_action(Act)),
    debug_println(9,stutter_according_to_proof_info(Act)).
check_if_action_is_stutter_wrt_inv(Act) :- 
   b_get_invariant_from_machine(Invariant),
   conjunction_to_list(Invariant,Conjunctions), % getting all conjunctions from the invariant
   b_get_read_write(Act,_Reads,Writes),
   ( action_does_not_modify_important_variables_in_the_inv(Writes,Conjunctions) 
	-> assert(stutter_action(Act)), 
	   debug_println(9,stutter_according_to_read_write_info(Act))
   ; otherwise 
	-> assert(visible_action(Act)),fail
   ).

action_does_not_modify_important_variables_in_the_inv([],_Conjuncions).
action_does_not_modify_important_variables_in_the_inv([Var|WrVars],Conjunctions) :- 
   var_not_changing_visible_condition_in_the_inv(Var,Conjunctions),
   action_does_not_modify_important_variables_in_the_inv(WrVars,Conjunctions).

var_not_changing_visible_condition_in_the_inv(_Var,[]).
var_not_changing_visible_condition_in_the_inv(Var,[Conj|Conjs]) :- 
   ( occurs_in_expr(Var,Conj) 
	-> (\+ trivial_predicate(Var,Conj) -> fail; true)
   ; otherwise -> true
   ),
   var_not_changing_visible_condition_in_the_inv(Var,Conjs).

trivial_predicate(Var,Conj) :- 
   \+var(Conj),
   trivial_member_predicate(Var,Conj).

/* 
   
   TODO: Define formal when an action in Event-B/B can be seen as a stutter action.
   Maybe we should use here the predicate b_specialized_invariant_for_op/2 from bmachine.pl.

*/

/*trivial_member_predicate(Var,Conj) :-
   Conj=b(member(b(identifier(Var),integer,_),b(integer_set('NATURAL'),set(integer),_)),pred,_).
trivial_member_predicate(Var,Conj) :-
   Conj=b(member(b(identifier(Var),integer,_),b(integer_set('NAT'),set(integer),_)),pred,_).*/
trivial_member_predicate(Var,Conj) :-
   Conj=b(member(b(identifier(Var),integer,_),b(integer_set('INT'),set(integer),_)),pred,_). 
trivial_member_predicate(Var,Conj) :-
   Conj=b(member(b(identifier(Var),integer,_),b(integer_set('INTEGER'),set(integer),_)),pred,_).

/************************** CHECKING IF ALL ACTIONS FROM AMPLE SET ARE STUTTER ACTIONS (END) ******************************/
