% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html

:- module(b_ast_cleanup, [clean_up/3, clean_up_pred/3, clean_up_pred_or_expr/3, 
                          clean_up_l_wo_optimizations/4,
                          clean_up_l_with_optimizations/4, 
                          check_used_ids_info/4, recompute_used_ids_info/2,
                          definitely_not_empty_and_finite/1, % TO DO: move to another module
                          predicate_level_optimizations/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module implements transformations/simplifications on the AST.').

:- use_module(tools, [safe_atom_chars/3,exact_member/2,foldl/4,filter/4]).
:- use_module(error_manager).
:- use_module(self_check).
:- use_module(bsyntaxtree).
:- use_module(translate,[print_bexpr/1]).
:- use_module(btypechecker, [unify_types_strict/2]).
:- use_module(preferences,[get_preference/2]).

:- use_module(library(lists)).
:- use_module(library(ordsets)).

% entry point for cleaning up predicates; ensures that global, predicate-level optimizations also applied
clean_up_pred(Expr,NonGroundExceptions,CleanedUpExpr) :-
   clean_up(Expr,NonGroundExceptions,CExpr),
   (get_texpr_type(CExpr,pred)
    -> predicate_level_optimizations(CExpr,CleanedUpExpr)
    ;  add_internal_error('Not predicate: ',clean_up_pred(Expr,NonGroundExceptions,CleanedUpExpr)),
       CleanedUpExpr = CExpr).

% Warning: arguments swapped with clean_up for maplist !
clean_up_pred_or_expr(NonGroundExceptions,Expr,CleanedUpExpr) :-
   clean_up_pred_or_expr_with_path(NonGroundExceptions,Expr,CleanedUpExpr,[]).
clean_up_pred_or_expr_with_path(NonGroundExceptions,Expr,CleanedUpExpr,Path) :-
%print(clean_up_pred_or_expr_with_path(NonGroundExceptions,Expr,CleanedUpExpr,Path)),nl,
   clean_up(Expr,NonGroundExceptions,CExpr,Path),
   (get_texpr_type(CExpr,pred)
    -> predicate_level_optimizations(CExpr,CleanedUpExpr,Path)
    ;  CleanedUpExpr = CExpr).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% clean up some code afterwards
clean_up(Expr,NonGroundExceptions,CExpr) :-
    clean_up(Expr,NonGroundExceptions,CExpr,[]).
clean_up(Expr,NonGroundExceptions,CExpr,Path) :-
    get_texpr_type(Expr,Type1),
    ground_type(Type1,NonGroundExceptions),
    cleanups(pre,Expr,[],TPExpr,Path),
    remove_bt(TPExpr,PExpr,LExpr,TLExpr),
    syntaxtransformation(PExpr,Subs,_,NewSubs,LExpr),
    functor(PExpr,F,N),
    % recursively clean up sub-expressions
    clean_up_l(Subs,NonGroundExceptions,NewSubs,F/N,1,Path),
    cleanups(post,TLExpr,[],CExpr,Path).

% apply the clean-up rules to an expression until all
% applicable rules are processed
% cleanups(Phase,Expr,AppliedRules,Result,Path):
%   Phase: pre or post
%   Expr:  the expression to clean up
%   AppliedRules: a list of clean up rules that have been already applied
%                 and must only be apply once ("single" mode)
%   Result: the cleaned-up expression
%   Path: list of outer functors leading to this expression; can be used to decide about applicability of rules
cleanups(Phase,Expr,AppliedRules,Result,Path) :-
    % set up co-routines that ensure that "Rule" is not applied if
    % is in the list AppliedRules
    assure_single_rules(AppliedRules,Rule),
    ( cleanup_phase(Phase,Expr,NExpr,Mode/Rule,Path) ->       % try to apply a rule (matching the current phase)
        ( Mode==single -> AppRules = [Rule|AppliedRules] % if the rule is marked as "single", we add to the list of already applied rules
        ; Mode==multi  -> AppRules = AppliedRules        % if "multi", we do not add it to the list, the rule might be applied more than once
        ; otherwise    -> add_error_fail(b_ast_cleanup,'Unexpected rule mode ',Mode)),
        cleanups(Phase,NExpr,AppRules,Result,Path)            % continue recursively with the new expression
    ; otherwise ->              % if no rule matches anymore,
        Result = Expr,          %   we leave the expression unmodified
        Rule = none).           %   and unblock the co-routine (see assure_single_rules/2)

assure_single_rules([],_Rule) :- !.
assure_single_rules(AppliedRules,Rule) :-
        assure_single_rules2(AppliedRules,Rule).
:- block assure_single_rules2(?,-).
assure_single_rules2(_AppliedRules,none) :- !.
assure_single_rules2(AppliedRules,Rule) :-
        \+ member(Rule, AppliedRules).

cleanup_phase(Phase,OTExpr,NTExpr,Mode/Rule,Path) :-
        create_texpr(OExpr,OType,OInfo,OTExpr),
        create_texpr(NExpr,NType,NInfo,NTExpr),
        cleanup_phase2(Phase,OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule,Path).
cleanup_phase2(pre,OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule,_Path) :-
        cleanup_pre(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode_Rule),
        (functor(Mode_Rule,'/',2) -> Mode_Rule = Mode/Rule
         ; add_internal_error('Illegal cleanup_pre rule, missing mode: ',Mode_Rule),fail).
cleanup_phase2(post,OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule,Path) :-
        cleanup_post_with_path(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode_Rule,Path),
        (functor(Mode_Rule,'/',2)
         -> Mode_Rule = Mode/Rule %,format(' cleanup_post Rule: ~w/~w~n',[Mode,Rule])
         ; add_internal_error('Illegal cleanup_post rule, missing mode: ',Mode_Rule),fail).

clean_up_l_wo_optimizations(Rest,NonGroundExceptions,CRest,SectionName) :-
   clean_up_l(Rest,NonGroundExceptions,CRest,top_level(SectionName),1,[]).
clean_up_l([],_,[],_Functor,_Nr,_Path).
clean_up_l([Expr|Rest],NonGroundExceptions,[CExpr|CRest],Functor,ArgNr,Path) :-
    clean_up(Expr,NonGroundExceptions,CExpr,[arg(Functor,ArgNr)|Path]),
    A1 is ArgNr+1,
    clean_up_l(Rest,NonGroundExceptions,CRest,Functor,A1,Path).

% same as clean_up_l but also applies predicate_level_optimizations
% Context is just the name of the section/context in which the optimizations are run
clean_up_l_with_optimizations(Rest,NonGroundExceptions,CRest,Context) :-
   %clean_up_l(Rest,NonGroundExceptions,CRest,top_level,1,[]).
   clean_up_l_with_opt(Rest,NonGroundExceptions,CRest,top_level(Context),1,[]).
clean_up_l_with_opt([],_,[],_Functor,_Nr,_Path).
clean_up_l_with_opt([Expr|Rest],NonGroundExceptions,[CExpr|CRest],Functor,ArgNr,Path) :-
    clean_up_pred_or_expr_with_path(NonGroundExceptions,Expr,CExpr,[arg(Functor,ArgNr)|Path]),
    A1 is ArgNr+1,
    clean_up_l_with_opt(Rest,NonGroundExceptions,CRest,Functor,A1,Path).

:- use_module(typing_tools,[is_infinite_type/1]).
:- use_module(specfile,[animation_mode/1, animation_minor_mode/1]).
% cleanup_pre(OldExpr,OldType,OldInfo,NewExpr,NewType,NewInfo,Mode/Rule)

cleanup_pre(block(TS),subst,_,Subst,subst,Info,multi/remove_block) :-
    !,get_texpr_expr(TS,Subst),get_texpr_info(TS,Info).
% replace finite by truth
cleanup_pre(finite(S),pred,I,truth,pred,[was(finite(S))|I],multi/remove_finite) :-
    preferences:get_preference(disprover_mode,false), % keep finite in disprover goals 
    get_texpr_type(S,Type), 
    \+ is_infinite_type(Type),!.
    %!,(is_infinite_type(Type) -> print(keep_finite(Type)),nl,fail ; print(removed_finite(Type)),nl).
%% implement fnc by rewriting using fnc(r) == %x.(x:dom(r)|r[{x}])
cleanup_pre(trans_function(Rel),Type,Info,NewExpr,Type,[was(trans_function)|Info],multi/trans_function) :-
    unify_types_strict(Type,set(couple(T1,set(T2)))),
    %print(trans_function(T1,T2,Rel)),nl,
    NewExpr = lambda([IDx],MemberPred,ImageExpr),
    safe_create_texpr(domain(Rel),set(T1),RangeSet),
    create_texpr(identifier('_zzzz_unary'),T1,[generated],IDx),
    safe_create_texpr(member(IDx,RangeSet),pred,MemberPred),
    safe_create_texpr(set_extension([IDx]),set(T1),SingletonX), % {x}
    safe_create_texpr(image(Rel,SingletonX),set(T2),ImageExpr). % r[{x}]
    %print('Rewriting fnc: '),translate:print_bexpr(NewExpr),nl.
%% implement rel by rewriting using rel(r) == {x,y| x:dom(r) & y: union(r[{x}])}
cleanup_pre(trans_relation(Rel),Type,Info,NewExpr,Type,[was(trans_relation)|Info],multi/trans_relation) :-
    unify_types_strict(Type,set(couple(T1,T2))),
    %print(trans_relation(T1,T2,Rel)),nl,
    NewExpr = comprehension_set([IDx,IDy],MemberPred),
    create_texpr(conjunct(MemberPredX,MemberPredY),pred,[generated],MemberPred),
    create_texpr(identifier('_zzzz_unary'),T1,[],IDx), % was [generated]
    create_texpr(identifier('_zzzz_binary'),T2,[],IDy), % was [generated]
    safe_create_texpr(domain(Rel),set(T1),RangeSet),
    safe_create_texpr(member(IDx,RangeSet),pred,MemberPredX), % x:dom(r)
    safe_create_texpr(set_extension([IDx]),set(T1),SingletonX), % {x}
    safe_create_texpr(image(Rel,SingletonX),set(set(T2)),ImageExpr), % r[{x}]
    safe_create_texpr(general_union(ImageExpr),set(T2),UnionSet), % union(r[{x}])
    safe_create_texpr(member(IDy,UnionSet),pred,MemberPredY). % y: union(r[{x}])
    %print('Rewriting rel: '),translate:print_bexpr(NewExpr),nl.
/*
 cleanup_pre(domain(SETC),Type,I,
             comprehension_set(DomainIds,NewPred),Type,I,multi/dom_let_pred) :-
    get_texpr_expr(SETC,comprehension_set(CompIds,CompPred)),
    get_domain_range_ids(CompIds,DomainIds,[RangeId]),
    get_texpr_ids(CompIds,UnsortedIds), sort(UnsortedIds,Blacklist),
    conjunction_to_list(CompPred,Preds),
    select_equality(TId,Preds,Blacklist,Expr,Rest,_UsedIds,check_well_definedness),
    same_id(TId,RangeId,_),
    !,
    LetIds = [RangeId], Exprs = [Expr],
    conjunct_predicates(Rest,RestPred),
    NewPred = b(let_predicate(LetIds,Exprs,RestPred),pred,[generated(domain)]),
     print('Translated dom({...,x|x=E&...}) into: '),  
    translate:print_bexpr(b(comprehension_set(DomainIds,NewPred),Type,[])),nl.
*/
/* strangely enough: this does not seem to buy anything:
 cleanup_pre(domain(SETC),Type,I,
             let_expression(LetIds,Exprs,NewExpr),Type,[generated(domain)|I],multi/dom_let_expr) :-
    get_texpr_expr(SETC,comprehension_set(CompIds,CompPred)),
    get_domain_range_ids(CompIds,DomainIds,[RangeId]),
    get_texpr_ids(CompIds,UnsortedIds), sort(UnsortedIds,Blacklist),
    conjunction_to_list(CompPred,Preds),
    select_equality(TId,Preds,Blacklist,Expr,Rest,_UsedIds,check_well_definedness),
    same_id(TId,RangeId,_),
    !,
    LetIds = [RangeId], Exprs = [Expr],
    conjunct_predicates(Rest,RestPred),
    get_texpr_info(SETC,CInfo),
    NewExpr = b(comprehension_set(DomainIds,RestPred),Type,CInfo),
     print('Translated dom({...,x|x=E&...}) into: '), translate:print_bexpr(b(let_expression(LetIds,Exprs,NewExpr),Type,[])),nl.
*/
% exchange quantified union by generalized union
cleanup_pre(QUANT,Type,I,
            let_expression(LetIds,Exprs,NewExpr),Type,[generated(QuantOP)|I],multi/quant_union_inter_let) :-
    quantified_set_operator(QUANT,QuantOP,AllIds,Pred,Expr),
    conjunction_to_list(Pred,Preds),
    % The ids are needed to build a "black list"
    get_texpr_ids(AllIds,UnsortedIds), sort(UnsortedIds,Ids),
    find_one_point_rules(AllIds,Preds,Ids,LetIds,Exprs,RestIds,NewPreds,check_well_definedness),
    %print(find_one_point_rules(AllIds,Preds,Ids,LetIds,Exprs,RestIds,NewPreds)),nl,
    % only succeed if we found at least one id which can be rewritten as let
    LetIds = [_ID1|_],
    !,
    ((RestIds=[],NewPreds=[]) -> NewExpr=Expr
      ; RestIds = [] -> conjunct_predicates(NewPreds,NewPred),
        NewExpr = b(if_then_else(NewPred,Expr,b(empty_set,Type,[])),Type,I)
      ; conjunct_predicates(NewPreds,NewPred),
        quantified_set_operator(NewQUANT,QuantOP,RestIds,NewPred,Expr),
        NewExpr = b(NewQUANT,Type,I)),
     (debug_mode(off) -> true
       ; print('Translated UNION/INTER into: '), translate:print_bexpr(b(let_expression(LetIds,Exprs,NewExpr),Type,[])),nl).
cleanup_pre(quantified_union(Ids,Pred,Expr),Type,I,general_union(Set),Type,I,multi/quant_union) :-
   %print(ids(Ids)),nl, print(pred(Pred)),nl, print(expr(Expr)),nl,
    !,quantified_set_op(Ids,Pred,Expr,quantified_union,Type,Set).
% exchange quantified intersection by generalized intersection
cleanup_pre(quantified_intersection(Ids,Pred,Expr),Type,I,general_intersection(Set),Type,I,multi/quant_inter) :-
    !,quantified_set_op(Ids,Pred,Expr,quantified_intersection,Type,Set).
% In reply to PROB-240: Check if arguments of Prj1/2 are types only using is_just_type:
cleanup_pre(function(TProjection,Argument),Type,I,Result,Type,I,multi/projection_call) :-
    get_texpr_expr(TProjection,Projection),
    cleanup_function_projection(Projection,Argument,I,Result),
    %print(prj_fun(Result,I)),nl,
    !.
cleanup_pre(function(Lambda,Argument),Type,I,assertion_expression(Cond,ErrMsg,Expr),Type,I,multi/lambda_guard1) :-
    get_texpr_expr(Lambda,LambdaExpr),
    is_lambda(LambdaExpr,TId,TPre,TVal),
    get_texpr_id(TId,ID),
    \+ occurs_in_expr(ID,Argument), % otherwise name clash
    !, % TO DO: support for functions of multiple arguments
    get_texpr_id(TId,Id),
    ErrMsg = 'function called outside of domain: ',
    ~~mnf( replace_id_by_expr(TPre,Id,Argument,Cond) ),
    ~~mnf( replace_id_by_expr(TVal,Id,Argument,Expr) ).
    %print(assertion(expression)),nl,translate:print_bexpr(Cond),nl, translate:print_bexpr(Expr),nl.
cleanup_pre(first_projection(A,B),Type,I,Set,Type,[was(prj1),prob_annotation('SYMBOLIC')|I],multi/first_projection) :-
    !,create_projection_set(A,B,first,Set).
cleanup_pre(second_projection(A,B),Type,I,Set,Type,[was(prj2),prob_annotation('SYMBOLIC')|I],multi/second_projection) :-
    !,create_projection_set(A,B,second,Set).
cleanup_pre(event_b_first_projection(Rel),Type,I,Set,Type,[prob_annotation('SYMBOLIC')|I],multi/ev_first_projection) :-
    !,create_event_b_projection_set(Rel,first,Set).
cleanup_pre(event_b_second_projection(Rel),Type,I,Set,Type,[prob_annotation('SYMBOLIC')|I],multi/ev_second_projection) :-
    !,create_event_b_projection_set(Rel,second,Set).
cleanup_pre(event_b_first_projection_v2,Type,I,Set,Type,[prob_annotation('SYMBOLIC')|I],multi/ev2_first_projection) :-
    !,create_event_b_projection_set_v2(Type,first,Set). % TO DO: Daniel, does this work ???
cleanup_pre(event_b_second_projection_v2,Type,I,Set,Type,[prob_annotation('SYMBOLIC')|I],multi/ev2_second_projection) :-
    !,create_event_b_projection_set_v2(Type,second,Set). % TO DO: Daniel, does this work ???
cleanup_pre(function(Fun,Arg),integer,I,ArithOp,integer,I,multi/succ_pred_optimisation) :-
    get_texpr_expr(Fun,PS),
    ( PS=predecessor -> Op=minus
    ; PS=successor   -> Op=add),
    ArithOp =.. [Op,Arg,Integer],
    create_texpr(integer(1),integer,[],Integer).
/* cleanup_pre(event_b_comprehension_set([ID],ID,Pred),T,I,comprehension_set([Result],NewPred),T,
        [was(event_b_comprehension_set)|I],multi/ev_compset_single_id) :-
         % Event_B_Comprehension with a single ID which is also the expression
         % TO DO: expand for multiple IDs
         !,
         Result = ID, NewPred=Pred. */
cleanup_pre(event_b_comprehension_set(Ids,Expr,Pred),T,I,NewExpression,T,
        [was(event_b_comprehension_set)|I],multi/ev_compset) :-
    rewrite_event_b_comprehension_set(Ids,Expr,Pred, T, NewExpression).
cleanup_pre(domain_restriction(A,B),T,I,identity(A),T,I,multi/event_b_to_normal_identity1) :-
  /* translate S <| id into id(S) */
   is_event_b_identity(B).
cleanup_pre(range_restriction(B,A),T,I,identity(A),T,I,multi/event_b_to_normal_identity2) :-
  /* translate id |> S into id(S) */
   is_event_b_identity(B).
% what about translating id(TOTAL TYPE) into event_b_identity ??
cleanup_pre(ring(A,B),T,I,composition(B,A),T,I,multi/ring_composition).
cleanup_pre(minus_or_set_subtract(A,B),integer,I,minus(A,B),integer,I,multi/remove_ambiguous_minus_int).
cleanup_pre(minus_or_set_subtract(A,B),set(T),I,set_subtraction(A,B),set(T),I,multi/remove_abiguous_minus_set).
cleanup_pre(minus_or_set_subtract(A,B),seq(T),I,set_subtraction(A,B),seq(T),I,multi/remove_abiguous_minus_seq).
cleanup_pre(mult_or_cart(A,B),integer,I,multiplication(A,B),integer,I,multi/remove_ambiguous_times_int).
cleanup_pre(mult_or_cart(A,B),set(T),I,cartesian_product(A,B),set(T),I,multi/remove_ambiguous_times_set).
cleanup_pre(mult_or_cart(A,B),seq(T),I,cartesian_product(A,B),seq(T),I,multi/remove_ambiguous_times_seq).
cleanup_pre(E,T,Iin,E,T,Iout,single/remove_rodinpos) :-
    selectchk(nodeid(rodinpos(_,[],_)),Iin,Iout). % reomve rodinpos information with Name=[]
cleanup_pre(partition(X,[Set]),pred,I,equal(X,Set),pred,I,multi/remove_partition_one_element) :-
      !,
      %  partition(X,Set) <=> X=Set
     print('Introducing equality for partition: '),translate:print_bexpr(X), print(' = '), translate:print_bexpr(Set),nl.
cleanup_pre(exists(AllIds,P),pred,I,NewP1,pred,I,single/factor_out) :-
    conjunction_to_list(P,Preds),
    % transform, e.g., #(x).(y>2 & x=y) -->  y>2 & #(x).(x=y)
    bsyntaxtree:create_exists_opt(AllIds,Preds,b(NewP1,pred,_I),Modified), 
    %print(exists(AllIds)),nl,print(NewP1),nl, % move things which do not depend on AllIds outside
     % the rule will fire again on the newly generated sub predicate ! -> fix ?
     Modified=true.% check if anything modified; otherwise don't fire rule
cleanup_pre(exists(AllIds,P),pred,I,let_predicate(LetIds,Exprs,NewP),pred,INew,multi/exists_to_let) :-
    % rewrite predicates of the form #x.(x=E & P(x)) into (LET x==E IN P(x))
    % side condition for #(ids).(id=E & P(ids)): no identifiers of ids occur in E
    conjunction_to_list(P,Preds),
    % The ids are needed to build a "black list"
    get_texpr_ids(AllIds,UnsortedIds), sort(UnsortedIds,Ids),
    find_one_point_rules(AllIds,Preds,Ids,LetIds,Exprs,RestIds,NewPreds,no_check), % TO DO: fully check that this is ok in the context of existential quantification:  #x.(1:dom(f) & x=f(1) & P) -->  LET x=f(1) IN 1:dom(f) & P END
    % only succeed if we found at least one id which can be rewritten as let
    LetIds = [ID1|_],!,
    %print(generate_let(LetIds,I)),translate:print_bexpr(P),nl,
    (atomic(ID1) -> add_internal_error(cleanup_pre,unwrapped_let_identifier(ID1)), INew=I
     ; get_texpr_id(ID1,AtomicID), remove_used_id_from_info(I,AtomicID,INew)), % probably not necessary ?!
    create_exists_opt(RestIds,NewPreds,NewP).
% now the same LET extraction but for universal qunatification:
cleanup_pre(forall(AllIds,P,Rhs),pred,I,let_predicate(LetIds,Exprs,NewP),pred,I,multi/forall_to_let) :-
    conjunction_to_list(P,Preds),
    % The ids are needed to build a "black list"
    get_texpr_ids(AllIds,UnsortedIds), sort(UnsortedIds,Ids),
    find_one_point_rules(AllIds,Preds,Ids,LetIds,Exprs,RestIds,NewPreds,check_well_definedness),
    % only succeed if we found at least one id which can be rewritten as let
    LetIds = [ID1|_],!,
    (atomic(ID1) -> add_internal_error(cleanup_pre,unwrapped_let_identifier(ID1)) ; true),
    conjunct_predicates(NewPreds,NewLhs),
    create_implication(NewLhs,Rhs,NewForallBody),
    create_forall(RestIds,NewForallBody,NewP),
    (debug_mode(on) ->  print('Introduced let in forall: '), print(LetIds),nl ; true).
    % translate:print_bexpr(b(let_predicate(LetIds,Exprs,NewP),pred,I)),nl. % warning: used_identifier information not yet computed; translate may generate warnings
cleanup_pre(exists(AllIds,P),pred,I,disjunct(NewP1,NewP2),pred,I,multi/partition_exists) :-
    is_a_disjunct(P,D1,D2),
    %print('### Warning: body of exists is a disjunction: '), translate:print_bexpr(P),nl,
    (animation_minor_mode(eventb) -> true
      ; translate:translate_bexpression(P,PS),
        add_warning(b_ast_cleanup,'Warning: body of existential quantifier is a disjunction (not allowed by Atelier-B): ',PS,I)),
    /* note that even if D2 is only well-defined in case D1 is false; it is ok to seperate this out
       into two existential quantifiers:  #x.(x=0 or 1/x=10) is ok to transform into #x.(x=0) or #x.(1/x=10) */
    create_exists_opt(AllIds,[D1],NewP1), % print('P1: '),translate:print_bexpr(NewP1),nl,
    create_exists_opt(AllIds,[D2],NewP2). %, print('P2: '),translate:print_bexpr(NewP1),nl.
cleanup_pre(exists(AllIds,P),pred,I,exists(AllIds,P),pred,[warning_generated|I],single/exists_implication_warning) :-
    (is_an_implication(P,Q,_)
     ; % we run as cleanup_pre: the other simplifications which remove typing have not run yet
       % such typing conjuncts typicially come from Rodin translations
       is_a_conjunct(P,Typing,Imp),
       is_an_implication(Imp,Q,_), 
       is_typing_predicate(Typing)
    ),
    \+ member(warning_generated,I),
    \+ is_typing_predicate(Q), % TRUTH => RHS  is ok
    % we could also catch something like x:TYPING & (Q=>R)
    (animation_minor_mode(eventb) -> true ;
    translate:translate_bexpression(P,PS),
    add_warning(b_ast_cleanup,'Warning: body of existential quantifier is an implication (did you mean a conjunction ?): ',PS,I)).
cleanup_pre(forall(AllIds,P,Rhs),pred,I,NewPred,pred,I,multi/forall_to_post_let) :-
% translate something like !(x,y).(y:1..100 & x=y*y => x<=y) into !(y).(y : 1 .. 100 => (#(x).( (x)=(y * y) & x <= y)))
    post_let_forall(AllIds,P,Rhs,NewPred,modification),
    !,
    (debug_mode(on) -> print('POST LET INTRODUCTION: '),translate:print_bexpr(b(NewPred,pred,[])),nl ; true).
cleanup_pre(set_extension(List),Type,I, value(AVL),Type,I, single/eval_set_extension) :- 
   List \= [],
   List = [_|ListT], ListT \= [], % do not do this for singleton sets so as not to prevent triggering of other rules
   preferences:get_preference(try_kodkod_on_load,false),
   maplist(eval_set_extension,List,EList),
   custom_explicit_sets:convert_to_avl(EList,AVL),
   % evaluate simple explicit set extensions: avoid storing & traversing position info & AST
   !,
   (debug_mode(on) -> print('EVAL SET EXTENSION: '), translate:print_bvalue(AVL),nl ; true).
% TO DO: same for sequence extension
cleanup_pre(set_extension(List),Type,I, set_extension(NList),Type,I, single/remove_pos) :- 
   remove_position_info_from_list(List,I,NList),!.
cleanup_pre(sequence_extension(List),Type,I, sequence_extension(NList),Type,I, single/remove_pos) :- 
    remove_position_info_from_list(List,I,NList),!.
cleanup_pre(typeset,SType,I,Expr,SType,I,multi/remove_typeset) :- !,
    ( ground(SType) ->
        (SType=set(Type),create_type_expression2(Type,Expr) -> true
        ; otherwise ->
            add_error_and_fail(b_ast_cleanup,'Creating type expression for type set failed: ',SType))
    ; otherwise    -> add_error_and_fail(b_ast_cleanup,'Non-ground type for typeset expression: ',SType)).

% TODO: Some types (namely records) are still missing
create_type_expression(Type,TExpr) :-
    create_texpr(Expr,set(Type),[],TExpr),
    create_type_expression2(Type,Expr).
create_type_expression2(integer,integer_set('INTEGER')).
create_type_expression2(boolean,bool_set).
create_type_expression2(global(G),identifier(G)).
create_type_expression2(string,string_set).
create_type_expression2(set(T),pow_subset(Sub)) :-
    create_type_expression(T,Sub).
create_type_expression2(couple(TA,TB),cartesian_product(A,B)) :-
    create_type_expression(TA,A),create_type_expression(TB,B).
create_type_expression2(freetype(FT),freetype_set(FT)).

% the case below happens frequently in data validation:
remove_position_info_from_list(List,I,NList) :-
   member(nodeid(pos(C,FilePos,Line,From,Line,To)),I),
   To-From > 1000, % the entire set/sequence extension is on one large line
   length(List,Len), Len>100, % it has many elements
   % we replace all position infos by the same top-level position info (enabling sharing)
   maplist(remove_position_info(nodeid(pos(C,FilePos,Line,From,Line,To))),List,NList),
   (debug_mode(off) -> true
     ; format('SIMPLIFY POSITION INFO IN SET/SEQUENCE EXTENSION: line # ~w, length ~w~n',[Line,Len])).
remove_position_info(NI,b(Expr,Type,Infos),b(NExpr,Type,NewInfos)) :-
    syntaxtransformation(Expr,Subs,_Names,NSubs,NExpr),
    (select(nodeid(pos(_,_FilePos,_,_,_,_)),Infos,NT) -> NewInfos=[NI|NT] ; NewInfos=Infos),
    maplist(remove_position_info(NI),Subs,NSubs).

% rules for function application of various projection functions
cleanup_function_projection(first_projection(A,B),Argument,I,Result) :-
    gen_assertion_expression(A,B,Argument,first_of_pair(Argument),first,I,Result).
cleanup_function_projection(second_projection(A,B),Argument,I,Result) :-
    gen_assertion_expression(A,B,Argument,second_of_pair(Argument),second,I,Result).
cleanup_function_projection(event_b_second_projection(A),Argument,_I,Result) :- % old style Rodin projection
    check_is_just_type(A),Result = first_of_pair(Argument).
cleanup_function_projection(event_b_second_projection(A),Argument,_I,Result) :- % old style Rodin projection
    check_is_just_type(A),Result = second_of_pair(Argument).
cleanup_function_projection(event_b_first_projection_v2,Argument,_I,Result) :- Result = first_of_pair(Argument).
cleanup_function_projection(event_b_second_projection_v2,Argument,_I,Result) :- Result = second_of_pair(Argument).

check_is_just_type(_A) :- preferences:get_preference(ignore_prj_types,true),!.
check_is_just_type(A) :- (is_just_type(A) -> true ; debug_println(9,not_type_for_prj(A)),fail).

gen_assertion_expression(A,B,_Argument,ProjExpr,_ProjType,_I,Result) :-
    check_is_just_type(A),check_is_just_type(B),
    !,
    Result = ProjExpr.
% TO DO: add simplification rule for couple(x,y) : A*B  with A or B being just types
gen_assertion_expression(A,B,Argument,ProjExpr,ProjType,Info,Result) :- 
    create_cartesian_product(A,B,CartAB),
    safe_create_texpr(member(Argument,CartAB),pred,MemCheck),
    ErrMsg = 'projection function called outside of domain: ', % TO DO: provide better user message with Argument result
    (ProjType == first -> get_texpr_type(A,set(TT)) ; get_texpr_type(B,set(TT))),
    %get_texpr_pos_infos(Argument,Info), % add position infos
    extract_pos_infos(Info,PosInfo),
    safe_create_texpr(ProjExpr,TT,PosInfo,TProjExpr),
    Result = assertion_expression(MemCheck,ErrMsg,TProjExpr). 
:- use_module(bsyntaxtree,[is_set_type/2]).
create_cartesian_product(A,B,CartAB) :- 
    get_texpr_types([A,B],[STA,STB]),
    is_set_type(STA,TypeA), is_set_type(STB,TypeB),
    safe_create_texpr(cartesian_product(A,B),set(couple(TypeA,TypeB)),CartAB).


% TO DO: support for functions of multiple arguments
is_lambda(lambda([TId],TPred,TValue), TId, TPred,TValue).
is_lambda(event_b_comprehension_set([TId],Expr,TPred), TId, TPred, TValue) :-
% rewrite_event_b_comprehension_set does not seem to get called before the function/lambda rule is applied
     % {ID.ID|->Val | PRed}
     Expr = b(couple(LHS,RHS),_,_),
     same_texpr(LHS,TId),
     TValue=RHS.

% rewriting Event-B comprehension sets into classical B style ones
rewrite_event_b_comprehension_set(IDs,CoupleExpr,Pred, _T, NewExpression) :-
        % detect lambda expressions in classical B style
        nested_couple_to_list(CoupleExpr,List),
        check_ids(IDs,List,Expr),!,
        %print(rewriting_event_b1(IDs,Pred,Expr)),nl,
        NewExpression = lambda(IDs,Pred,Expr).
rewrite_event_b_comprehension_set(IDList,CoupleExpr,Pred, _T, NewExpression) :-
         % Event_B_Comprehension with a several IDs which are also used as the couple expression
         nested_couple_to_list(CoupleExpr,List), %print(List),nl,
         List = IDList,
         !,
         NewExpression = comprehension_set(IDList,Pred).
rewrite_event_b_comprehension_set(Ids,Expr,Pred, T, NewExpression) :-
    NewExpression = comprehension_set([Result],NewPred),
    unify_types_strict(T,set(Type)),
    % print(event_b_comprehension_set(Ids,Expr,Pred)),nl,
    (select(Expr,Ids,RestIds)
      -> % the Expr is an identifier which is part of Ids: we can avoid complicated translation below
       % example {f,n•n:INT & f:1..n-->Digit|f} --translated-> {f|#(n).(n:INT & f:1..n-->Digit)}
      % print(remove(Expr,RestIds)),nl,
         ExPred=Pred, Result=Expr
      ; get_unique_id_inside('_comp_result_',Pred,Expr,ResultId),
        create_texpr(identifier(ResultId),Type,[],Result),
        safe_create_texpr(equal(Result,Expr),pred,Equal),
        conjunct_predicates([Equal,Pred],ExPred),
        RestIds=Ids
    ),
    create_exists(RestIds,ExPred,NewPred). %, print(done),nl, translate:print_bexpr(NewPred),nl.

check_ids([],[CoupleExpr],CoupleExpr). % we terminate with a single expression
check_ids([ID|T],[CoupleExprID|CT],Rest) :-
  same_id(ID,CoupleExprID,_),
  check_ids(T,CT,Rest).


eval_set_extension(b(E,_,_),EE) :- eval_set_extension_aux(E,EE).
eval_set_extension_aux(boolean_false,pred_false).
eval_set_extension_aux(boolean_true,pred_true).
eval_set_extension_aux(couple(A,B),(EA,EB)) :- eval_set_extension(A,EA), eval_set_extension(B,EB).
eval_set_extension_aux(integer(I),int(I)).
eval_set_extension_aux(string(S),string(S)).
eval_set_extension_aux(unary_minus(b(integer(I),_,_)),int(MI)) :- MI is -I.


post_let_forall(AllIds,P,Rhs,NewPred,modification) :-
    %print('try forall: '), translate:print_bexpr(P),nl,
    conjunction_to_list(P,Preds), reverse(Preds,RPreds),
    select_equality(TId,RPreds,[],Expr,RRest,UsedIds,no_check),
    select(TId,AllIds,RestIds),
    get_texpr_id(TId,Id),
    \+ member(Id,UsedIds), % not a recursive equality
    reverse(RRest,Rest),
    conjunct_predicates(Rest,RestPred),
    \+ occurs_in_expr(Id,RestPred),
    !,
    NewRhs = b(let_predicate([TId],[Expr],Rhs),pred,[]),
    post_let_forall(RestIds,RestPred,NewRhs,NewPred,_).
post_let_forall(AllIds,P,Rhs,NewPred, no_modification) :-
    create_implication(P,Rhs,NewForallBody),
    create_forall(AllIds,NewForallBody,NewP), 
    get_texpr_expr(NewP,NewPred).


is_interval(b(interval(A,B),_,_),A,B).
is_empty_set(b(empty_set,_,_)).
is_empty_set(b(empty_sequence,_,_)).
% create a lambda expression for a projection
create_projection_set(A,B,_Switch,Res) :- 
    (is_empty_set(A) ; is_empty_set(B)),!,
    Res = empty_set.
create_projection_set(A,B,Switch,lambda(Ids,SPred,Expr)) :- %  generate lambda to be able to use function(lambda) rule
    Ids = [TArg1,TArg2],
    ( Switch==first  -> Expr = TArg1
    ; Switch==second -> Expr = TArg2),
    get_texpr_type(A,TA1), unify_types_strict(TA1,set(Type1)), 
    get_texpr_type(B,TB2), unify_types_strict(TB2,set(Type2)),
    (contains_no_ids(A,B) -> Arg1 = '_zzzz_unary', Arg2 = '_zzzz_binary' % avoid generating fresh ids; relevant for test 1313 and ticket PROB-346
    % TO DO: check whether _zzzz_unary/binary are actually used; we should avoid generating fresh ids whenever possible (otherwise syntactically identical formulas become different)
     ; get_unique_id_inside('_prj_arg1__',A,B,Arg1),
       get_unique_id_inside('_prj_arg2__',A,B,Arg2)),
    create_texpr(identifier(Arg1),Type1,[generated(Switch)],TArg1),
    create_texpr(identifier(Arg2),Type2,[generated(Switch)],TArg2),
    safe_create_texpr(member(TArg1,A),pred,MembA),
    safe_create_texpr(member(TArg2,B),pred,MembB),
    conjunct_predicates([MembA,MembB],Pred), SPred=Pred.
%    bsyntaxtree:mark_bexpr_as_symbolic(Pred,SPred). % TO DO: put mark code into another module; maybe only mark as symbolic if types large enough ??

contains_no_ids(A,B) :- contains_no_ids(A), contains_no_ids(B).
contains_no_ids(b(E,_,_)) :- contains_no_ids_aux(E).
contains_no_ids_aux(bool_set).
contains_no_ids_aux(integer_set(_)).
contains_no_ids_aux(mult_or_cart(A,B)) :- contains_no_ids(A),contains_no_ids(B).
contains_no_ids_aux(relations(A,B)) :- contains_no_ids(A),contains_no_ids(B).
contains_no_ids_aux(pow_subset(A)) :- contains_no_ids(A).
contains_no_ids_aux(string_set).
% TO DO: add more

create_event_b_projection_set(Rel,Switch,lambda(Ids,SPred,Expr)) :-
    Ids = [TArg1,TArg2],
    ( Switch==first  -> Expr = TArg1
    ; Switch==second -> Expr = TArg2),
    get_texpr_type(Rel,RT),unify_types_strict(RT,set(couple(Type1,Type2))),
    get_unique_id_inside('_prj_arg1__',Rel,Arg1),
    get_unique_id_inside('_prj_arg2__',Rel,Arg2),
    create_texpr(identifier(Arg1),Type1,[generated(Switch)],TArg1),
    create_texpr(identifier(Arg2),Type2,[generated(Switch)],TArg2),
    create_texpr(couple(TArg1,TArg2),couple(Type1,Type2),[],Couple),
    safe_create_texpr(member(Couple,Rel),pred,Member),
    SPred=Member.
    %bsyntaxtree:mark_bexpr_as_symbolic(Pred,SPred).

create_event_b_projection_set_v2(RelType,Switch,comprehension_set(Ids,SPred)) :-
    % we are generating {p1,p2,lambda | lambda=p1/p2}
    Ids = [TArg1,TArg2,TArg3],
    ( Switch==first  -> ResultExpr = TArg1, Type1 = Type3
    ; Switch==second -> ResultExpr = TArg2, Type2 = Type3),
    unify_types_strict(RelType,set(couple(couple(Type1,Type2),T3))),
    (T3==Type3 -> true ; add_error(create_event_b_projection_set,'Unexpected return type: ',T3)),
    %print(types(Type1,Type2,Type3)),nl,
    Arg1 = '_zzzz_unary',
    Arg2 = '_zzzz_binary',
    Arg3 = '_lambda_result_', % the comprehension set contains no other expressions: no clash possible
    create_texpr(identifier(Arg1),Type1,[generated(Switch)],TArg1),
    create_texpr(identifier(Arg2),Type2,[generated(Switch)],TArg2),
    create_texpr(identifier(Arg3),Type3,[lambda_result,generated(Switch)],TArg3),
    safe_create_texpr(equal(TArg3,ResultExpr),pred,[lambda_result],Equal),
    conjunct_predicates([Equal],Pred), 
    SPred=Pred.
%    bsyntaxtree:mark_bexpr_as_symbolic(Pred,SPred).
    %,print(Pred),nl.

:- use_module(btypechecker,[couplise_list/2,prime_identifiers/2]).
% create a comprehension set for quantified union or intersection
quantified_set_op(Ids,Pred,Expr,Loc,set(Type),Set) :-
    Info = [generated(Loc)],
    create_texpr(range(Lambda),set(set(Type)),Info,Set),
    create_texpr(lambda(Ids,Pred,Expr),set(couple(ArgType,set(Type))),Info,Lambda),
    get_texpr_types(Ids,Types),couplise_list(Types,ArgType). % , print(quantified),nl,translate:print_bexpr(Set),nl.

quantified_set_operator(quantified_union(AllIds,Pred,Expr),quantified_union,AllIds,Pred,Expr).
quantified_set_operator(quantified_intersection(AllIds,Pred,Expr),quantified_intersection,AllIds,Pred,Expr).

:- use_module(debug).
:- use_module(kernel_records,[normalise_record_type/2]).
:- use_module(library(lists),[last/2]).

% first the rules that require the path:
cleanup_post_with_path(assign([b(identifier(ID),TYPE,INFO)],[EXPR]),subst,I,
             assign_single_id(b(identifier(ID),TYPE,INFO),EXPR),subst,I,single/assign_single_id,Path) :- 
    \+ animation_minor_mode(eventb), % there is no support in the Event-B interpreter for assign_single_id yet
    (simple_expression(EXPR) % the assign_single_id is not guarded by a waitflag; EXPR should not be too expensive too calculate
      -> true
       ; % if we are in an unguarded context; then we do not need to guard EXPR by waitflag anyway
         maplist(unguarded,Path)
    ),
    !,
    (debug_mode(on) -> print('Single Assignment: '),
        translate:print_subst(b(assign([b(identifier(ID),TYPE,INFO)],[EXPR]),subst,I)),nl 
     ;  true).
cleanup_post_with_path(any(Ids,Pred,Subst),subst,I,Result,subst,[generated|I],single/transform_any_into_let,Path) :-
    (last(Path,arg(top_level(_),_)) /* do not remove top-level ANY if show_eventb_any_arguments is true; see, e.g., test 1271 */
     -> preferences:preference(show_eventb_any_arguments,false) ; true),
    find_lets(Ids,Pred,LetIDs,LetDefs,RestIds,RestPred),
    LetIDs \= [],
      % print(found_lets(LetIDs,RestIds,RestPred)),nl,print(Path),nl,
    conjunct_predicates(LetDefs,LetDefPred),
    ((RestIds = [], is_truth(RestPred)) % complete ANY can be translated to LET
       -> Result = let(LetIDs,LetDefPred,Body), Body = Subst
      ;  split_predicate(RestPred,Ids,RestUsingIds,RestNotUsingIds),
         % print('USING: '),translate:print_bexpr(RestUsingIds),nl, print('NOT USING: '),translate:print_bexpr(RestNotUsingIds),nl,
         (RestIds = []
           -> (is_truth(RestUsingIds)
               % RestPred does not use the LET identifiers; move outside of the LET !
                -> Result = select([b(select_when(RestPred,SelectBody),subst,[generated|I])]),
                   SelectBody = b(let(LetIDs,LetDefPred,Subst),subst,[generated|I])
                ; is_truth(RestNotUsingIds)
               % RestPred uses LET identifiers in all conjuncts; move inside LET
                -> Result = let(LetIDs,LetDefPred,LetBody),
                   LetBody = b(select([b(select_when(RestPred,Subst),subst,[])]),subst,[generated|I])
                ; otherwise
                % we would need to generate an outer and inner select; transformation probably not worth it
                 -> fail
              )
            ; is_truth(RestUsingIds)
               % RestPred does not use LET identifiers move outside of LET
              -> Result = any(RestIds,RestPred,SelectBody),
                 SelectBody = b(let(LetIDs,LetDefPred,Subst),subst,[generated|I])
            ; is_truth(RestNotUsingIds)
               % RestPred uses LET identifiers in all conjuncts; move inside LET
              -> Result = let(LetIDs,LetDefPred,LetBody),
                 LetBody = b(any(RestIds,RestPred,Subst),subst,I)
            ; otherwise
                % we would need to generate an outer and inner any; transformation probably not worth it
                 -> fail
            )
     ),
     !. %,translate:print_subst(b(Result,subst,[])),nl.
cleanup_post_with_path(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule,_Path) :-
    cleanup_post(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule).

%unguarded(arg(sequence/1,1)). % first argument of sequence is not guarded
unguarded(arg(X,_)) :- unguarded_aux(X).
unguarded_aux(top_level(_)).
unguarded_aux(operation/4).
unguarded_aux(parallel/1).
unguarded_aux(init_parallel/1).
unguarded_aux(var/2).
unguarded_aux(let/3).
% what about choice/2 ??

:- load_files(library(system), [when(compile_time), imports([environ/2])]).

% now the rules which do not need the Path
cleanup_post(Expr,Type,I,Expr,Type,I2,single/remove_erroneous_info_field) :- 
    \+ ground(I),!,
    functor(Expr,F,_N),
    add_internal_error('Information field not ground: ',I:F),
    I2=[].
:- if(environ(prob_safe_mode,true)).
cleanup_post(comprehension_set(Ids,E),Type,I,comprehension_set(Ids,E),Type,I,single/sanity_check) :- 
    get_texpr_ids(Ids,UnsortedIds),sort(UnsortedIds,SIds),
    \+ same_length(UnsortedIds,SIds),
    add_error(cleanup_post,'Identifier clash in comprehension set: ',UnsortedIds),
    print(E),nl,
    fail.
:- endif.
%cleanup_post(block(TS),subst,_,Subst,subst,Info,multi/remove_block) :-
%    !,get_texpr_expr(TS,Subst),get_texpr_info(TS,Info).
cleanup_post(Expr,Type,I,Expr,NType,I,single/normalise_record_type) :-
    nonvar(Type),Type=record(Fields),
    normalise_record_type(record(Fields),NType),
    NType \== Type,
    !.
cleanup_post(conjunct(b(truth,pred,_),b(B,pred,I)),pred,_,B,pred,I,multi/remove_truth_conj1) :- !.
cleanup_post(conjunct(b(A,pred,I),b(truth,pred,_)),pred,_,A,pred,I,multi/remove_truth_conj2) :- !.
cleanup_post(conjunct(b(falsity,pred,_),_),pred,I,falsity,pred,I,multi/simplify_falsity_conj1) :- !.
% cleanup_post(conjunct(_,b(falsity,pred,_)),pred,_,falsity,pred,I,multi/simplify_falsity_conj2) :- !.  % WD Implications ???
cleanup_post(disjunct(b(truth,pred,_),_),pred,I,truth,pred,I,multi/simplify_truth_disj1) :- !.
%cleanup_post(disjunct(_,b(truth,pred,_)),pred,_,truth,pred,I,multi/simplify_truth_disj2) :- !. % WD Implications ???
cleanup_post(disjunct(b(A,pred,I),b(falsity,pred,_)),pred,_,A,pred,I,multi/remove_falsity_disj1) :- !.
cleanup_post(disjunct(b(falsity,pred,_),b(B,pred,I)),pred,_,B,pred,I,multi/remove_falsity_disj2) :- !.
cleanup_post(implication(b(truth,pred,_),b(B,pred,I)),pred,_,B,pred,I,multi/remove_truth_impl1) :- !.
cleanup_post(implication(b(falsity,pred,_),_),pred,I,truth,pred,I,multi/simplify_falsity_impl1) :- !.
cleanup_post(implication(_,b(truth,pred,_)),pred,I,truth,pred,I,multi/simplify_truth_impl2) :- !.
cleanup_post(equivalence(b(truth,pred,_),b(B,pred,I)),pred,_,B,pred,I,multi/remove_truth_equiv1) :- !.
cleanup_post(equivalence(b(A,pred,I),b(truth,pred,_)),pred,_,A,pred,I,multi/remove_truth_equiv2) :- !.
% TO DO: more rules for implication/equivalence to introduce negations  (A <=> FALSITY  --->  not(A)) ?
% detect certain tautologies/inconsistencies
cleanup_post(member(X,B),pred,I,truth,pred,[was(member(X,B))|I],multi/remove_type_member) :-
      is_just_type(B),!.
cleanup_post(not_member(X,B),pred,I,falsity,pred,[was(not_member(X,B))|I],multi/remove_type_not_member) :- is_just_type(B),!.
cleanup_post(member(X,b(Set,_,_)),pred,I,equal(X,One),pred,I,multi/remove_member_one_element_set) :- 
      singleton_set_extension(Set,One),
      !,
      %  X:{One} <=> X=One
     true. %print('Introducing equality: '),translate:print_bexpr(X), print(' = '), translate:print_bexpr(One),nl.
cleanup_post(member(X,b(Set,_,_)),pred,I,not_equal(X,One),pred,I,multi/remove_member_setdiff) :-
     Set = set_subtraction(MaximalSet,SONE),
     SONE = b(Set2,_,_), singleton_set_extension(Set2,One),
     definitely_maximal_set(MaximalSet),
     !,  % x : INTEGER-{One}  <=>  x/=One
     (debug_mode(off) -> true
      ; print('Replacing member of set_subtraction: '), translate:print_bexpr(MaximalSet), print(' - '), translate:print_bexpr(SONE),nl).
cleanup_post(not_member(X,b(Set,_,_)),pred,I,not_equal(X,One),pred,I,multi/remove_member_one_element_set) :-  
      singleton_set_extension(Set,One),
      !,
      %  X/:{One} <=> X/=One
      true. %print('Introducing disequality: '),translate:print_bexpr(X), print(' /= '), translate:print_bexpr(One),nl.
cleanup_post(member(E,b(fin_subset(E2),_,_)),pred,I,finite(E),pred,I,multi/introduce_finite) :- 
    (same_texpr(E,E2); is_just_type(E2)),!. % print(introduce(finite(E))),nl.  
cleanup_post(member(E,b(image(Rel,SONE),TypeImg,InfoImg)),pred,I,member(Couple,Rel),pred,I,multi/replace_image_by_member) :- 
     SONE = b(Set,_,_), singleton_set_extension(Set,One),
      !,
     create_couple(One,E,Couple),
     (debug_mode(off) -> true
      ;  print('Member of Image: '),translate:print_bexpr(b(member(E,b(image(Rel,SONE),TypeImg,InfoImg)),pred,I)),nl,
         print('    replaced by: '),translate:print_bexpr(b(member(Couple,Rel),pred,I)),nl
     ).
/* do we want need this rule ?:
clenaup_post(member(b(couple(A,B),couple(TA,TB),IC),ID),pred,I,equal(A,B),pred,I,multi/replace_member_id) :-
     is_is_event_b_identity(ID),
     !,
     print('Replace member of id by equality: '),
     tranlsate:print_bexpr(b(member(b(couple(A,B),couple(TA,TB),IC),ID),pred,I)),nl.
*/
cleanup_post(member(b(couple(A,B),couple(TA,TB),IC),b(reverse(Rel),_,_)),pred,I,member(ICouple,Rel),pred,I,multi/remove_reverse) :- !,
     (debug_mode(off) -> true ;  print('Removed inverse (~): '),translate:print_bexpr(Rel),nl),
     ICouple = b(couple(B,A),couple(TB,TA),IC).
cleanup_post(subset(A,B),pred,I,truth,pred,[was(subset(A,B))|I],multi/remove_type_subset) :- is_just_type(B),!.
cleanup_post(subset(A,B),pred,I,NewPred,pred,[generated_conjunct|I],multi/replace_subset_by_element) :- 
     get_texpr_expr(A,set_extension(List)),!, % for sequence extension we don't need this as the interpreter knows exactly the cardinality of a sequence_extension ?
     % applying rule {x1,x2,...} <: B  <=>  x1:B & x2:B & ...
     maplist(gen_member_predicates(B),List,Conjuncts),
     conjunct_predicates(Conjuncts,TNewPred),
      % print('detected subset-member rule: '),translate:print_bexpr(TNewPred),nl,
     get_texpr_expr(TNewPred,NewPred).
cleanup_post(subset(A,B),pred,I,NewPred,pred,[generated_conjunct|I],multi/replace_union_subset) :- 
     % mark conjunct as generated: used e.g. by flatten_conjunct in predicate_evaluator
     get_texpr_expr(A,union(_,_)),!,
     % applying rule A1 \/ A2 <: B  <=>  A1 <: B & A2 <: B
     % could be detrimental if checking that something is an element of B is expensive
     extract_unions(A,As),
     findall(Subi,(member(Ai,As),safe_create_texpr(subset(Ai,B),pred,I,Subi)),Conj),  % we could try and re-run clean-up ? safe_create_texpr will ensure WD info set
     conjunct_predicates(Conj,TNewPred),
     %translate:print_bexpr(TNewPred),nl,
     get_texpr_expr(TNewPred,NewPred).
cleanup_post(Comp,pred,I,SComp,pred,I,multi/simplify_cse_comparison) :-
  % simplify comparison operations; can result in improved constraint propagation
  % e.g., ia + CSE1 * 2 > ia + fa <=> CSE1 * 2 > fa for Setlog/prob-ttf/qsee-TransmitMemoryDumpOk21_SP_3.prob
     comparison(Comp,A,B,SComp,SA,SB),
     %print(try_simplify(A,B)),nl,
     simplify_comparison_terms(A,B,SA,SB),!,
     (debug_mode(off) -> true
      ;  print('Simplified: '),translate:print_bexpr(b(Comp,pred,I)),
         print(' <=> '),translate:print_bexpr(b(SComp,pred,I)),nl).
cleanup_post(equal(A,B),pred,I,truth,pred,[was(equal(A,B))|I],multi/remove_equality) :- 
   same_texpr(A,B),always_well_defined_or_disprover_mode(A),!. % ,print(removed_equal(A,B)),nl. 
cleanup_post(equal(A,B),pred,I,falsity,pred,[was(equal(A,B))|I],multi/remove_equality_false) :- 
   different_texpr_values(A,B),!. %,print(removed_equal_false(A,B)),nl. 
cleanup_post(equal(A,B),pred,I,greater(Low,Up),pred,I,multi/remove_equality) :- 
   % Low..Up = {}  <=>  Low>Up   % is also handled by constraint solver; but other simplifications can apply here
   (is_empty_set(B), is_interval(A,Low,Up) ;
    is_empty_set(A), is_interval(B,Low,Up)),!,
    (debug_mode(off) -> true
      ;  print('Simplified: '), translate:print_bexpr(b(equal(A,B),pred,I)), print(' <=> '),
         translate:print_bexpr(b(greater(Low,Up),pred,I)),nl).
cleanup_post(equal(A,B),pred,I,equal(A,RLet),pred,I,single/detect_recursion) :-
    % "A" should be an identifier
    get_texpr_id(A,ID),
    % check if some side conditions are fulfilled where the recursion detection can be enabled
    recursion_detection_enabled(A,B,I),
    % A must be recursively used in B:
    find_recursive_usage(B,ID),
    % TO DO: also find mutual recursion !
    debug_println(9,recursion_detected(ID)),
    !, % create an recursive_let where the body is annotated to be symbolic
    get_texpr_type(B,Type),  add_texpr_infos(B,[prob_annotation('SYMBOLIC')],B2),
    mark_recursion(B2,ID,B3),
    print(marked_recursion(ID)),nl,
    safe_create_texpr(recursive_let(A,B3),Type,RLet).
cleanup_post(equal(A,B),pred,_I,ResultExpr,pred,Info2,multi/simplify_bool_true_false) :- 
   % simplify bool(X)=TRUE -> X and bool(X)=FALSE -> not(X)
   ( get_texpr_expr(A,convert_bool(X)), get_texpr_boolean(B,BOOLVAL)
   ;
     get_texpr_boolean(A,BOOLVAL),get_texpr_expr(B,convert_bool(X))
   ),
   !, debug_println(9,simplify_bool_equal_TRUE(X)),
   get_texpr_info(X,Info2),
   (BOOLVAL = boolean_true -> get_texpr_expr(X,ResultExpr) ; ResultExpr = negation(X)).
cleanup_post(not_equal(A,B),pred,I,falsity,pred,[was(not_equal(A,B))|I],multi/remove_disequality) :- 
   same_texpr(A,B),always_well_defined_or_disprover_mode(A),!. % ,print(removed_not_equal(A,B)),nl. 
   % sometimes one uses & TRUE=TRUE to finish off guards, invariants, ...
   % exchange lambda expressions by a comprehension set
% TO DO: also add rule for bool(X)=FALSE -> not(X)
cleanup_post(not_equal(A,B),pred,I,truth,pred,[was(not_equal(A,B))|I],multi/remove_disequality_false) :- 
   different_texpr_values(A,B),!. %  ,print(removed_not_equal_false(A,B)),nl.
cleanup_post(not_equal(A,B),pred,I,NewP,pred,[was(not_equal(A,B))|I],multi/not_disjoint_disequality) :- 
  /* Set1 /\ Set2 /= {}    <===>   #(zz).(zz:Set1 & zz:Set2) */
   preferences:preference(use_smt_mode,true), /* currently this rewriting makes test 1112 fail; TO DO: investigate */
   is_empty_set(B),
   get_texpr_expr(A,intersection(Set1,Set2)),!, 
   get_texpr_type(Set1,Set1Type), unify_types_strict(Set1Type,set(T)),
   ID = b(identifier('_zzzz_unary'),T,[generated]),
   ESet1 = b(member(ID,Set1),pred,[]),
   ESet2 = b(member(ID,Set2),pred,[]),
   create_exists_opt([ID],[ESet1,ESet2],NewPredicate),
   (debug_mode(off) -> true
      ;  print('Transformed not disjoint disequality: '),translate:print_bexpr(NewPredicate),nl),
   get_texpr_expr(NewPredicate,NewP).
cleanup_post(negation(A),pred,I,falsity,pred,[was(negation(A))|I],multi/remove_negation_truth) :- 
   is_truth(A),!. % ,print(negation(A)),nl. 
cleanup_post(negation(A),pred,I,truth,pred,[was(negation(A))|I],multi/remove_negation_falsity) :- 
   is_falsity(A),!. % ,print(negation(A)),nl. 
cleanup_post(convert_bool(A),boolean,I,boolean_true,boolean,I,multi/remove_convert_bool) :- 
   is_truth(A),!. %  ,print(convert_bool(A)),nl. 
cleanup_post(convert_bool(A),boolean,I,boolean_false,boolean,I,multi/remove_convert_bool_false) :- 
   is_falsity(A),!. %  ,print(convert_bool_false(A)),nl. 
cleanup_post(assertion_expression(Cond,_ErrMsg,Expr),_T,_I,BE,TE,IE,multi/remove_assertion_expression) :- 
   is_truth(Cond),!,  %print(remove_assertion_expression(Cond)),nl,
   Expr = b(BE,TE,IE).
cleanup_post(union(A,B),T,I,Res,T,[add_element_to_set|I],single/add_element_to_set) :- A=b(AE,_,_),B=b(BE,_,_),
    \+ member(add_element_to_set,I),
    (   singleton_set_extension(BE,_El) -> Res = union(A,B)
      ; singleton_set_extension(AE,_El) -> Res = union(B,A)),
    !. %,print(detected_add_singleton_element(_El)),nl.
cleanup_post(general_union(SetExt),Type,_I,Res,Type,Info,multi/general_union_set_extension) :- % union_generalized
    SetExt = b(set_extension(LIST),_,Info), 
    construct_union_from_list(LIST,Type,Info,TRes),
    !, (debug:debug_mode(on) -> print('translated_general_union: '), translate:print_bexpr(TRes),nl ; true),
    get_texpr_expr(TRes,Res).
cleanup_post(function(Override,X),Type,_I,Res,Type,Info,multi/function_override) :-
    preferences:get_preference(disprover_mode,true), % only applied in Disprover mode as it can remove WD problem; TO DO make it also applicable in normal mode
    get_texpr_expr(Override,overwrite(_F,SEXt)),
    SEXt = b(set_extension(LIST),_,_),
    member(b(couple(From,To),_,_),LIST),
    same_texpr(From,X),
    % ( F <+ { ... X|->To ...}) (X)   ==>  To
    %translate:print_bexpr(b(function(Override,X),Type,_I)), print(' ==> '), translate:print_bexpr(To),nl,
    !,
    get_texpr_expr(To,Res),
    get_texpr_info(To,Info).
cleanup_post(function(SEXt,ARG),Type,_I,Res,Type,Info,multi/function_set_extension) :-
    SEXt = b(set_extension(LIST),_,_),  % TO DO: also support b(value(avl_set(A)),_,_)
    eval_set_extension(ARG,Value), %nl,print(arg(Value)),nl,
    select(b(couple(LHS,RHS),_,_),LIST,REST),
    eval_set_extension(LHS,Value), %nl,print(found(RHS)),nl,
    % WE NEED TO Check that all LHS can be compared against ARG
    \+ member((b(couple(LHS2,_),_,_),REST),eval_set_extension(LHS2,Value)), % no other potential match
    \+ member((b(couple(LHS3,_),_,_),LIST), \+ eval_set_extension(LHS3,_)), % all left-hand-sides can be evaluated
    !, get_texpr_expr(RHS,Res), get_texpr_info(RHS,Info).
% Detect if_then_else in format as printed by pp_expr2(if_then_else( ....)...) or as generated by B2TLA:
cleanup_post(function(IFT,DUMMYARG),Type,Info,if_then_else(IFPRED,THEN,ELSE),Type,Info,multi/function_if_then_else) :-
   is_if_then_else(IFT,DUMMYARG,IFPRED,THEN,ELSE),
   (debug:debug_mode(off) -> true
    ; print('% Recognised if-then-else expression: IF '), translate:print_bexpr(IFPRED),
      print('  THEN '),translate:print_bexpr(THEN), print(' ELSE '),translate:print_bexpr(ELSE),nl
   ).
/*
cleanup_post(function(Lambda,Argument),Type,I,assertion_expression(Cond,ErrMsg,Expr),Type,I,multi/lambda_guard2) :-
    get_texpr_info(Lambda,IL), member(was(lambda),IL),
    print(function(Lambda,Argument)),nl,
    get_texpr_expr(Lambda,comprehension_set([TId1,TId2,LAMBDARES],TPre,TVal)),
    Argument=b(couple(Argument1,Argument2),_,_),
    !, % TO DO: support for functions of multiple arguments
    get_texpr_id(TId1,Id1),
    get_texpr_id(TId2,Id2),
    ErrMsg = 'function called outside of domain: ',
    ~~mnf( replace_id_by_expr(TPre,Id1,Argument1,Cond1) ),
    ~~mnf( replace_id_by_expr(TVal,Id1,Argument1,Expr1) ),
    ~~mnf( replace_id_by_expr(Cond1,Id2,Argument2,Cond) ),
    ~~mnf( replace_id_by_expr(Expr1,Id2,Argument2,Expr) ).  */
cleanup_post(lambda(Ids,Pred,Expr),Type,I, comprehension_set(CompIds,CompPred),Type,NewInfo,multi/remove_lambda) :- !,
    NewInfo = I, % was NewInfo = [was(lambda)|I],
    unify_types_strict(Type,set(couple(_ArgType,ResType))),
    %ResultId = '_lambda_result_',
    get_unique_id_inside('_lambda_result_',Pred,Expr,ResultId),
    % get_unique_id('_lambda_result_',ResultId),
    def_get_texpr_id(Result,ResultId),
    get_texpr_type(Result,ResType),
    get_texpr_info(Result,[lambda_result]),
    append(Ids,[Result],CompIds),
    get_texpr_expr(Equal,equal(Result,Expr)),
    get_texpr_type(Equal,pred),
    extract_important_info_from_subexpression(Expr,EqInfo), % mark equality with wd condition if Expr has wd condition
    get_texpr_info(Equal,[lambda_result|EqInfo]),
    conjunct_predicates([Pred,Equal],CompPred0),
    add_texpr_infos(CompPred0,[prob_annotation('LAMBDA')],CompPred).
cleanup_post(range(SETC),Type,I, comprehension_set(RangeIds,NewCompPred),Type,I,single/range_setcompr) :-
   % translate ran({x1,...xn|P})  into {xn| #(x1,...).(P)} ; particularly interesting if x1... contains large datavalues (e.g., C_02_001.mch from test 1131)
    get_texpr_expr(SETC,comprehension_set(CompIds,CompPred)),
    get_domain_range_ids(CompIds,DomainIds,RangeIds), % print(range(CompIds,DomainIds,RangeIds)),nl,
    \+ (member(ID,RangeIds),get_texpr_id(ID,'_lambda_result_')), % TO DO: rename in this case; if we don't rename then _lambda_result_ will not be enumerated !
    !, % TO DO: use create_exists_opt + also detect patterns such as dom(dom( or ran(ran( ...
    create_outer_exists(DomainIds,CompPred,NewCompPred),
    (debug_mode(off) -> true
      ;  print('Encode range as existential quantification: '), translate:print_bexpr(NewCompPred),nl).
/*  fails test 306 : */
/*
cleanup_post(domain(SETC),Type,I, comprehension_set(DomainIds,NewCompPred),Type,I,single/domain_setcompr) :-
   % translate dom({x1,...xn|P})  into {x1,..| #(xn).(P)} ; particularly interesting if xn contains large datavalues
    get_texpr_expr(SETC,comprehension_set(CompIds,CompPred)),
    get_domain_range_ids(CompIds,DomainIds,RangeIds), % print(domain(CompIds,DomainIds,RangeIds)),nl,
    \+ (member(ID,CompIds),get_texpr_id(ID,'_lambda_result_')),
    !,
    % TO DO: use create_optimized exists; also treat inner existential quantification and merge
    create_outer_exists(RangeIds,CompPred,NewCompPred), print('Encode domain as existential quantification: '),translate:print_bexpr(NewCompPred),nl.
    % WE HAVE TO BE CAREFUL if xn = LAMBDA_RESULT ?
*/
cleanup_post(domain(SETC),Type,I, comprehension_set(DomainIds,RestPred),Type,I,single/domain_setcompr) :-
   % translate dom({x1,...xn|P & xn=E})  into {x1,..| P} ; particularly interesting if xn contains large datavalues
    get_texpr_expr(SETC,comprehension_set(CompIds,CompPred)),
    get_domain_range_ids(CompIds,DomainIds,RangeIds), % print(domain(CompIds,DomainIds,RangeIds)),nl,
    \+ (member(ID,CompIds),get_texpr_id(ID,'_lambda_result_')),
    conjunction_to_list(CompPred,Preds),
    RangeIds = [TId],
    get_sorted_ids(RangeIds,Blacklist),
    select_equality(TId,Preds,Blacklist,_Expr,RestPreds,_,check_well_definedness), % We could do check_well_definedness only if preference set
    conjunct_predicates(RestPreds,RestPred),
    not_occurs_in_predicate(Blacklist,RestPred),
    !,
    % TO DO: use create_optimized exists; also treat inner existential quantification and merge
    (debug_mode(off) -> true
      ;  print('Encode domain of lambda abstraction: '),translate:print_bexpr(RestPred),nl).

cleanup_post(precondition(TP,TS),subst,_,S,subst,I,multi/remove_triv_precondition) :-
    % remove trivial preconditions
    get_texpr_expr(TP,truth),!,
    get_texpr_expr(TS,S),get_texpr_info(TS,I).

cleanup_post(parallel(Statements),subst,I,Result,subst,I, single/parallel_to_multi_assign) :- 
 % this merges multiple assignments into a single one: advantage: only one waitflag set up
 % should probably not be done in INITIALISATION
    % print(parallel(Statements)),nl,trace,
    extract_assignments(Statements,LHS,RHS,Rest,Nr), % print(extracted(Nr,LHS)),nl,
    Nr>1,!,
    (debug_mode(on) -> 
       print('Parallel to Assignment: '), translate:print_subst(b(parallel(Statements),subst,I)),nl
      ; true),
    (Rest == []  ->   Result  =  assign(LHS,RHS)
     ;  Result = parallel([b(assign(LHS,RHS),subst,[])|Rest])).
    %translate:print_subst(b(Result,subst,I)),nl.

cleanup_post(evb_becomes_such(Ids,Pred),subst,I,any(Primed,Pred,Assignment),subst,I,multi/ev2_becomes_such_any) :-
    prime_identifiers(Ids,Primed),!,
    create_texpr(assign(Ids,Primed),subst,[],Assignment).
cleanup_post(evb2_becomes_such(Ids,Pred),subst,I,becomes_such(Ids,Pred),subst,I,multi/ev2_becomes_such) :- !.
cleanup_post(successor,Type,I,Compset,Type,[was(successor)|I],multi/successor) :- !,
    % translation of succ
    pred_succ_compset(add,Compset).
cleanup_post(predecessor,Type,I,Compset,Type,[was(predecessor)|I],multi/predecessor) :- !,
    % translation of pred
    pred_succ_compset(minus,Compset).
cleanup_post(becomes_such(Ids1,Pred),subst,I,becomes_such(Ids2,Pred),subst,I,single/becomes_such) :- !,
    find_used_primed_ids(Pred,BeforeAfter),
    add_before_after_info(Ids1,BeforeAfter,Ids2).
cleanup_post(select([CHOICE|Rest]),subst,_,S,subst,I,single/remove_select) :-
    CHOICE=b(select_when(TRUTH,Subst),subst,_),
    is_truth(TRUTH),!, 
    (Rest=[] -> true ; print('Removing useless SELECT'),nl),
    get_texpr_expr(Subst,S),get_texpr_info(Subst,I).
cleanup_post(select([CHOICE|_],_ELSE),subst,_,S,subst,I,single/remove_select_else) :-
    CHOICE=b(select_when(TRUTH,Subst),subst,_),
    is_truth(TRUTH),!, print('Removing useless SELECT ELSE'),nl,
    get_texpr_expr(Subst,S),get_texpr_info(Subst,I).
cleanup_post(Expr,Type,I,Expr,Type,[contains_wd_condition|I],single/possibly_undefined) :-
    nonmember(contains_wd_condition,I),
     % print(' - CHECK WD: '), translate:print_bexpr(Expr),nl, %%
    is_possibly_undefined(Expr),!,
     %% print('CONTAINS WD: '), translate:print_bexpr(Expr),nl, %%
     %(translate:translate_bexpression(Expr,'{min(xunits)}') -> trace ; true),
    true.
% if a substitution has a sub-expression that is a substitution with that refers to the original
% value of a variable, we mark this substitution, too.
cleanup_post(Subst,subst,I,Subst,subst,NI,single/refers_to_old_state) :-
    safe_syntaxelement(Subst,Subs,_,_,_),
    % check if a child contains the refers_to_old_state flag
    setof(Reference, (member(Sub,Subs),
                         get_texpr_info(Sub,SubInfo),
                         memberchk(refers_to_old_state(References),SubInfo),
                         member(Reference,References)), ReferedIds),
    !,
    sort(ReferedIds,SortedIds),
    NI = [refers_to_old_state(SortedIds)|I].
% If the substitution of an operation contains a while whose invariant contains references x$0
% to the original value of a variable x, we must insert a LET substitution to preserve the original value.
cleanup_post(operation(Id,Results,Args,Body),Type,I,operation(Id,Results,Args,NewBody),Type,I,single/refers_to_old_state_let) :-
    get_texpr_info(Body,BodyInfo),
    memberchk(refers_to_old_state(References),BodyInfo),!,
    create_equalities_for_let(References,Ids,Equalities),
    conjunct_predicates(Equalities,P),
    insert_let(Body,Ids,P,NewBody).
cleanup_post(let_expression([],[],TExpr),Type,_,Expr,Type,I,multi/remove_let_expression) :- !,
    % remove trivial let expressions without any introduced identifiers
    % this rule makes only sense in combination with the next rule which removes
    % simple let identifiers
    get_texpr_expr(TExpr,Expr),
    get_texpr_info(TExpr,I).
cleanup_post(let_predicate([],[],TExpr),Type,Iin,Expr,Type,Iout,multi/remove_let_predicate) :- !,
    % same as above, just for predicates
    get_texpr_expr(TExpr,Expr),
    get_texpr_info(TExpr,I), debug_println(9,removed_let_predicate),
    % The next is done to prevent removing position information (in case of Event-B invariants)
    ( nonmember(nodeid(_),I), member(nodeid(P),Iin) -> Iout = [nodeid(P)|I]
    ; otherwise                                     -> Iout = I).
cleanup_post(let([],Pred,TExpr),Type,_,Expr,Type,I,multi/remove_let) :- is_truth(Pred),!,
    % remove trivial let expressions without any introduced identifiers
    % this rule makes only sense in combination with the next rule which removes
    % simple let identifiers
    get_texpr_expr(TExpr,Expr),
    get_texpr_info(TExpr,I).
cleanup_post(let_expression(Ids,Exprs,Expr),Type,I,
             let_expression(NIds,NExprs,NExpr),Type,I,multi/remove_let_expression2) :-
    simplify_let(Ids,Exprs,Expr,NIds,NExprs,NExpr),!.
cleanup_post(let_predicate(Ids,Exprs,Expr),Type,I,
             let_predicate(NIds,NExprs,NExpr),Type,I,multi/remove_let_predicate2) :-
    simplify_let(Ids,Exprs,Expr,NIds,NExprs,NExpr),!, 
    true. 
    %print(simplified_let),nl, translate:print_bexpr(b(let_predicate(Ids,Exprs,Expr),pred,I)), print(' INTO : '),nl, translate:print_bexpr(b(let_predicate(NIds,NExprs,NExpr),pred,I)),nl, nl.
% was disabled because simplify_let_subst also replaced in RHS of assignments
% but now we check that a LET/ANY variable cannot be assigned to statically
% so it should be safe now to replace simple equalities:
cleanup_post(let(Ids,Pred,Subst),Type,I,
             let(NIds,NPred,NSubst),Type,I,multi/remove_let_subst2) :-
    simplify_let_subst(Ids,Pred,Subst,NIds,NPred,NSubst),! ,
    true. % translate:print_subst(b(let(NIds,NPred,NSubst),subst,[])),nl.
/* this is not used currently; as the generic code's performance has been improved
   and better propagation is ensured by b_compute_arith_expression which (partially) instantiates its argument (known to be an integer)
cleanup_post(ArithPredicate,pred,Info,ArithPredicate,pred,NewInfo,single/add_arith_pred_info) :-
    comparison(ArithPredicate,LHS,RHS,_,_,_),
    \+ clpfd_arith_integer_expression(LHS),
    \+ clpfd_arith_integer_expression(RHS),
    !,
    % store the information that there is no use in doing b_compute_arith_expression special treatment
    NewInfo = [no_clpfd_arith_integer_expression|Info].
*/

cleanup_post(forall([ID],LHS,RHS),pred,IOld,Res,pred,IOld,single/expand_forall_set_extension) :-
 % expand !x.(x:{a,b,...} => RHS) into conjunction
  % can be useful e.g. for KODKOD when we pick an element from a set of sets
   preferences:get_preference(try_kodkod_on_load,true), % at the moment only enable in Kodkod mode
   % TO DO: enable always; but maybe check that set_extension can be computed (which eval_set_extension will do) to avoid duplicating checks (!y.(y:{v,w} => expensive_pred) + what if v=w
   nonmember(do_not_optimize_away,IOld),
   get_texpr_expr(LHS,member(ID2,Set)),
   same_texpr(ID,ID2),
   is_set_extension(Set,SList),
   get_texpr_id(ID,AID),
   format('Expanding forall ~w ',[AID]),
   findall(C, (member(SEL,SList),replace_id_by_expr(RHS,AID,SEL,C)),Conjuncts),
   conjunct_predicates(Conjuncts,ExpandedForAll),
   translate:print_bexpr(ExpandedForAll),nl,
   get_texpr_expr(ExpandedForAll,Res).

% moved these rule below the simplifciation rules above to avoid re-computing used_id infos 
cleanup_post(forall(Ids,Lhs,Rhs),pred,IOld,forall(Ids,NLhs,NRhs),pred,INew,single/forall_used_identifier) :-
    inner_predicate_level_optimizations(Lhs,NLhs),
    inner_predicate_level_optimizations(Rhs,NRhs),
    conjunct_predicates([NLhs,NRhs],P),
    %print(forall(Ids,IOld)),nl,
    add_used_identifier_info(Ids,P,IOld,INew).
cleanup_post(exists(Ids,P),pred,IOld,exists(Ids,NP),pred,INew,single/exists_used_identifier) :-
    inner_predicate_level_optimizations(P,NP),
    add_used_identifier_info(Ids,NP,IOld,INew).
cleanup_post(Construct,Type,I,NewConstruct,Type,I,single/detect_partitions) :- 
    contains_predicate(Construct,Type,Pred,NewConstruct,NewPred),!, % something like a select or other substitution
    predicate_level_optimizations(Pred,NewPred).
cleanup_post(Construct,Type,I,NewConstruct,Type,I,single/detect_partitions2) :- 
    contains_predicates(Construct,Type,Preds,NewConstruct,NewPreds),!,
    maplist(predicate_level_optimizations,Preds,NewPreds)  % TO DO: do CSE together (in some cases) !
    . %, (Preds=NewPreds -> true ; print('Found partitions: '), translate:print_subst(b(NewConstruct,Type,I)),nl).

%% COMMENT IN NEXT LINE TO CHECK validity of AST per NODE (helps find bugs)
%%cleanup_post(Expr,pred,I,Expr,pred,I,single/checked) :- check_ast(true,b(Expr,pred,I)),fail.
%%

simplify_let_subst(Ids,Pred,Subst,NIds,RestPred,NewSubst) :-
    % remove let identifiers whose definitions are very simple, i.e. identifiers
    Eq = b(equal(TypID,TExpr),pred,_),
    can_be_optimized_away(TypID),
    get_texpr_id(TypID,ToReplace),
    b_interpreter:member_conjunct(Eq,Pred,RestPred),
    is_simple_expression(TExpr),
    nth0(_N,Ids,TI,NIds),get_texpr_id(TI,ToReplace),
    !,
    % Intitially there was an issue as we may also replace in the LHS of assignments
    % see e.g. TestLet = LET cnt BE cnt=1 IN  IF cnt=0 THEN ABORT ELSE cnt :: {0,1} END END; in SubstitutionLaws
    % However, now the static type checker rejects those assignments
    replace_id_by_expr(Subst,ToReplace,TExpr,NSubst),
     % TO DO: it seems like cleanup rules are not applied on NSubst, e.g., function for set_extension rules
     debug_println(9,replaced_let_subst_id(ToReplace)),
    (Subst==NSubst -> NewSubst=NSubst ; clean_up(NSubst,[],NewSubst)).

can_be_optimized_away(b(_,_,I)) :- nonmember(do_not_optimize_away,I).

simplify_let(Ids,Exprs,Expr,NIds,NExprs,CleanNewExpr) :-
    % remove let identifiers whose definitions are very simple, i.e. identifiers
    nth0(N,Exprs,TExpr,NExprs),is_simple_expression(TExpr),
    nth0(N,Ids,TI,NIds),get_texpr_id(TI,ToReplace),
    can_be_optimized_away(TI),
    % TO DO: do not do this to outer variables which the user cares about !!
    !,
    debug_println(9,simple_expression_id(ToReplace)),
    replace_id_by_expr(Expr,ToReplace,TExpr,NExpr),
    clean_up(NExpr,[],CleanNewExpr). % clean up adjust eg used_ids info; necessary for test 568 in prob_safe_mode
simplify_let(Ids,Exprs,Expr,NIds,NExprs,CleanNewExpr) :-
    % push the let expression down the AST. E.g. "LET a=E IN (4*(a+z*a) + z)" would
    % be transformed to "4*(LET a=E IN (a+z*a)) + z"
    % If the final form is like "LET a=E IN a" it will be simplified to E.
    nth0(N,Ids,TId,NIds),
    get_texpr_id(TId,Id), can_be_optimized_away(TId),
    nth0(N,Exprs,LetExpr,NExprs),
    \+ do_not_to_move_let_inside(Expr),
    ( identifier_sub_ast(Expr,Id,SubPosition) ->
        SubPosition = [_|_], % The LET can actually be moved down
        %tools_printing:print_term_summary(identifier_sub_ast(Id,Expr,SubPosition)),nl,
        exchange_ast_position(SubPosition,Expr,OldInner,NewInner,NExpr),
        get_texpr_type(OldInner,Type),
        ( get_texpr_id(OldInner,Id) -> % There is only one reference to Id,
            debug_println(9,one_reference(Id)),
            NewInner = LetExpr         % replace it with the expression
        % We need to check that we are not simply just exchanging lets with each other
        ; cycle_detection(Id,SubPosition,Expr) -> debug_println(9,cycle(Id)),fail
        ; Type=pred ->  % print(created_let_predicate(N,Id,SubPosition,NIds)),nl,
            extract_important_info_from_subexpressions(LetExpr,OldInner,NewLetInfo), % maybe no longer necessary because of cleanups call below
            create_texpr(let_predicate([TId],[LetExpr],OldInner),pred,NewLetInfo,NewInner)
        ; otherwise -> % print(create_let_expression(TId)),nl,
            extract_important_info_from_subexpressions(LetExpr,OldInner,NewLetInfo), % maybe no longer necessary because of cleanups call below
            create_texpr(let_expression([TId],[LetExpr],OldInner),Type,NewLetInfo,NewInner)),
        clean_up(NExpr,[],CleanNewExpr) % maybe we only need WD post rules ?
    ; always_well_defined(LetExpr) -> % Id does not occur in the expression -> just remove the LET
        CleanNewExpr = Expr),!.

do_not_to_move_let_inside(b(E,_,_)) :- 
   do_not_to_move_let_inside_aux(E).
do_not_to_move_let_inside_aux(forall(_,_,_)). % otherwise we may compute the let multiple times
% what about exists ??
do_not_to_move_let_inside_aux(comprehension_set(_,_)). % ditto
do_not_to_move_let_inside_aux(lambda(_,_,_)). % ditto
do_not_to_move_let_inside_aux(general_sum(_,_,_)). % ditto
do_not_to_move_let_inside_aux(general_product(_,_,_)). % ditto
do_not_to_move_let_inside_aux(quantified_union(_,_,_)). % ditto
do_not_to_move_let_inside_aux(quantified_intersection(_,_,_)). % ditto

cycle_detection(Id,SubPosition,Expr) :-
    get_constructors(SubPosition,Expr,Constructors),
    ((member(CC,Constructors), \+ let_constructor(CC)) -> fail
          ; debug_println(9,cycle_let_detection(Id,Constructors))
            %,print(cycle_let(Id,Constructors)),nl,translate:print_bexpr(Expr),nl
            ).
% we have a let constructor which can be modified by simplify_let:
let_constructor(let_expression).
let_constructor(let_predicate).
let_constructor(let_substitution).

% get a SubPosition path (as produced by identifier_sub_ast) and generate all constructors that are used along the Path
get_constructors([],_,[]).
get_constructors([Pos|T],OldTExpr,[ConstructorForPos|CT]) :- 
    remove_bt(OldTExpr,OldExpr,NewExpr,_NewTExpr),
    functor(OldExpr,ConstructorForPos,_),
    syntaxtransformation(OldExpr,Subs,_Names,_NSubs,NewExpr),
    nth0(Pos,Subs, OldSelected,_Rest),
    get_constructors(T,OldSelected,CT).
    
is_simple_expression(TExpr) :-
    get_texpr_expr(TExpr,Expr),
    is_simple_expression2(Expr),!.
is_simple_expression(TExpr) :-
    is_just_type(TExpr).
is_simple_expression2(identifier(_)).
is_simple_expression2(integer(_)).
is_simple_expression2(boolean_true).
is_simple_expression2(boolean_false).

% detect either Event-B identity or id over full type
is_event_b_identity(b(X,_,_)) :- is_event_b_identity_aux(X).
is_event_b_identity_aux(event_b_identity).
is_event_b_identity_aux(identity(T)) :- is_just_type(T).

% check if we have a set extension and return list of terms
is_set_extension(b(S,T,I),L) :- is_set_extension_aux(S,T,I,L).
is_set_extension_aux(set_extension(L),_,_,L).
% TO DO: detect sequence_extension
is_set_extension_aux(value(avl_set(A)),set(T),I,L) :-  % computed by eval_set_extension
   findall(b(value(M),T,I),avl:avl_member(M,A),L).

recursion_detection_enabled(A,_B,I) :-
    animation_mode(b),                % in B mode,
    memberchk(section(properties),I), % the rule should be only applied to properties
    % and where A is an abstract constant.
    get_texpr_info(A,AInfo),memberchk(loc(_,_,abstract_constants),AInfo).
recursion_detection_enabled(A,_B,_I) :-
    animation_minor_mode(eventb),    % in Event-B,
    %TODO: Limit application to axioms
    get_texpr_info(A,AInfo),         % A must be a constant
    memberchk(loc(_,constants),AInfo).
recursion_detection_enabled(_A,_B,_I) :-
    animation_minor_mode(z).    % use always in Z

construct_union_from_list([X],_,_,Res) :- !, Res=X.
construct_union_from_list([X,Y|T],Type,Info,Res) :-
    construct_union_from_list([Y|T],Type,Info,RHS),
    Res = b(union(X,RHS),Type,Info).

% detect if we have an if-then-else function (which is then applied to a dummy argument)
is_if_then_else(b(comprehension_set([TDummyID1,ID2],CONJ),_Type,_),_DUMMYARG,IFPRED,THEN,ELSE) :-
  % we ignore the _DUMMYARG as here we do not check the value of DUMMYARG in the body
  % TO DO: also allow removal of equalities as in TLA case below
   get_texpr_id(ID2,LambdaID), get_texpr_id(TDummyID1,DummyID),
   is_a_conjunct(CONJ,IMP1,IMP2),
   is_an_implication(IMP1,IFPRED,EQ1),
   EQ1 = b(equal(II1,THEN),pred,_), get_texpr_id(II1,LambdaID),
   is_an_implication(IMP2,NOT_IFPRED,EQ2),
   EQ2 = b(equal(II2,ELSE),pred,_), get_texpr_id(II2,LambdaID),
   is_negation_of(IFPRED,NOT_IFPRED),
   \+ occurs_in_expr(DummyID,IFPRED),\+ occurs_in_expr(LambdaID,IFPRED),
   \+ occurs_in_expr(DummyID,THEN),  \+ occurs_in_expr(LambdaID,THEN),
   \+ occurs_in_expr(DummyID,ELSE),  \+ occurs_in_expr(LambdaID,ELSE).
% Recognize B2TLA encodings as well: %((x).(x=0 & PRED|C1)\/%(x).(x=0 & not(PRED)|C2)) (0)
%is_if_then_else(IF,_,_,_,_) :- nl,print(IF),nl,nl,fail.
is_if_then_else(b(union(COMP1,COMP2),_Type,_),DUMMYARG,IFPRED,THEN,ELSE) :-
   %print(union(COMP1,COMP2)),nl,tools_printing:trace_print(COMP1),nl,
   if_then_else_lambda(COMP1,DUMMYARG,IFPRED,THEN),
   if_then_else_lambda(COMP2,DUMMYARG,NOT_IFPRED,ELSE),
   is_negation_of(IFPRED,NOT_IFPRED).
   

if_then_else_lambda(b(comprehension_set([TDummyID,TLAMBDAID],CONJ),_,_),DUMMYARG,IFPRED,RESULT) :-
   conjunction_to_list(CONJ,CL), get_texpr_id(TDummyID,DummyID),
   get_texpr_id(TLAMBDAID,LambdaID),
   (remove_equality(DummyID,DUMMYVAL,CL,ConjList) % look if there is an equality for the DummyID
     -> same_texpr(DUMMYVAL,DUMMYARG), % TO DO: check that we apply the function with this value
        \+ occurs_in_expr(LambdaID,DUMMYVAL), % ensure this is really the same value
        \+ occurs_in_expr(DummyID,DUMMYVAL)
     ; ConjList=CL),
   remove_equality(LambdaID,RESULT,ConjList,RestList),
   \+ occurs_in_expr(LambdaID,RESULT),
   \+ occurs_in_expr(DummyID,RESULT), % otherwise this is not really a dummy identifier
   conjunct_predicates(RestList,IFPRED),
   \+ occurs_in_expr(DummyID,IFPRED).
   
% remove a dummy equality from list
remove_equality(ID,RHS,ConjList,Rest) :-
   DummyEQ = b(equal(DID,RHS),pred,_), % TO DO: also accept simple typing memberships ?
   get_texpr_id(DID,ID),
   select(DummyEQ,ConjList,Rest). %, print(eq(DID,RHS)),nl.

get_texpr_boolean(b(boolean_true,_,_),boolean_true).
get_texpr_boolean(b(boolean_false,_,_),boolean_false).


is_typing_conjunct(b(member(_,B),_,_)) :- is_just_type(B).
is_typing_predicate(Typing) :- conjunction_to_list(Typing,LT), maplist(is_typing_conjunct,LT).

% TO DO: use create_optimized exists;
create_outer_exists(Ids,b(exists(InnerIds,P),pred,Infos),New) :- append(Ids,InnerIds,NewIds),!,
    New = b(exists(NewIds,P),pred,Infos).
create_outer_exists(Ids,P,N) :- create_exists(Ids,P,N).

contains_predicate(convert_bool(Pred),boolean,Pred,
                   convert_bool(NewP),NewP).
contains_predicate(comprehension_set(CompIds,Pred),_,Pred,
                   comprehension_set(CompIds,NewP),NewP).
contains_predicate(general_sum(Ids,Pred,Expression),integer,Pred,
                   general_sum(Ids,NewP,Expression),NewP).
contains_predicate(general_product(Ids,Pred,Expression),integer,Pred,
                   general_product(Ids,NewP,Expression),NewP).
contains_predicate(if_then_else(Pred,Then,Else),_,Pred,
                   if_then_else(NewP,Then,Else),NewP).
contains_predicate(assertion_expression(Pred,ErrMsg,Expr),_,Pred,
                   assertion_expression(NewP,ErrMsg,Expr),NewP).
contains_predicate(precondition(Pred,Body),subst,Pred,
                   precondition(NewP,Body),NewP).
contains_predicate(assertion(Pred,Body),subst,Pred,
                   assertion(NewP,Body),NewP).
contains_predicate(becomes_such(Vars,Pred),subst,Pred,
                   becomes_such(Vars,NewP),NewP).
contains_predicate(any(Parameters,Pred,Body),subst,Pred,
                   any(Parameters,NewP,Body),NewP).
contains_predicate(lazy_let_expr(ID,SharedExpr,MainExpr),pred, SharedExpr,
                    lazy_let_expr(ID,NewSharedExpr,MainExpr), NewSharedExpr) :-
            get_texpr_type(SharedExpr,pred).
contains_predicate(lazy_let_pred(ID,SharedExpr,MainExpr),pred, MainExpr,
                    lazy_let_pred(ID,SharedExpr,NewMainExpr), NewMainExpr) :-
            \+ get_texpr_type(SharedExpr,pred).
%  while(COND,STMT,INV,VARIANT), select, if  --> can have multiple predicates !!
contains_predicates(while(Cond,   Stmt,Invariant,Variant),subst, [Cond,Invariant],
                    while(NewCond,Stmt,NewInv,   Variant), [NewCond,NewInv]).
contains_predicates(if(Whens),subst,Preds,
                    if(NewWhens),NewPreds) :-
            get_predicates_from_list_of_cases(Whens,Preds,NewWhens,NewPreds).
contains_predicates(select(Whens),subst,Preds,
                    select(NewWhens),NewPreds) :-
            get_predicates_from_list_of_cases(Whens,Preds,NewWhens,NewPreds).
contains_predicates(select(Whens,Else),subst,Preds,
                    select(NewWhens,Else),NewPreds) :-
            get_predicates_from_list_of_cases(Whens,Preds,NewWhens,NewPreds).
contains_predicates(lazy_let_pred(ID,SharedExpr,MainExpr),pred, [SharedExpr,MainExpr],
                    lazy_let_pred(ID,NSharedExpr,NMainExpr), [NSharedExpr,NMainExpr]) :-
            get_texpr_type(SharedExpr,pred).

get_predicates_from_list_of_cases([],[],[],[]).
get_predicates_from_list_of_cases([H|T],Preds,[NewH|NewT],NewPreds) :-
   (get_single_predicate(H,Pred,NH,NewPred)
     -> NewH=NH, Preds=[Pred|TP], NewPreds = [NewPred|NTP]
     ;  NewH=H,  Preds=TP,        NewPreds = NTP
    ),
    get_predicates_from_list_of_cases(T,TP,NewT,NTP).

get_single_predicate(b(E,T,I),Preds,b(NewE,T,I),NewPreds) :-
   get_single_predicate_aux(E,Preds,NewE,NewPreds).
get_single_predicate_aux(select_when(Pred,Body),Pred,select_when(NewPred,Body),NewPred).
get_single_predicate_aux(if_elsif(Pred,Body),Pred,if_elsif(NewPred,Body),NewPred).


:- use_module(b_expression_sharing,[cse_optimize_predicate/2]).
% these are "global" optimizations at the predicate level
% they are only called once a predicate has been completely constructed
predicate_level_optimizations(Pred,NewPred) :-
   predicate_level_optimizations(Pred,NewPred,[]).
predicate_level_optimizations(Pred,NewPred,Path) :-
   inner_predicate_level_optimizations(Pred,Pred1),
   ((get_preference(use_common_subexpression_elimination,true),
    \+ do_not_optimise_in_context(Path))
     -> cse_optimize_predicate(Pred1,NewPred)
      ; NewPred=Pred1
   ). %,print_opt_debug_info(Pred,NewPred,Path).
/*
print_opt_debug_info(Pred,NewPred,Path) :-
   (Pred==NewPred -> true
     ; same_texpr(Pred,NewPred) -> true
     ; format('Optimized pred ~w: ',[Path]), translate:print_bexpr(NewPred),nl
        % , (Path=[] -> trace ; true)
    ).
*/

do_not_optimise_in_context([arg(top_level(invariant),Nr)]) :- 
   get_preference(use_po,true),
   format('% NOT applying CSE to Invariant Nr ~w (PROOF_INFO = TRUE)~n',[Nr]).

:- use_module(partition_detection,[detect_all_partitions_in_predicate/2]).
% this predicate is also called for exists, forall, ...:
inner_predicate_level_optimizations(Pred,NewPred) :-
   detect_all_partitions_in_predicate(Pred,NewPred)
   . %,(Pred==NewPred -> true ; print('Optimized pred: '), translate:print_bexpr(NewPred),nl).

% divide a list of identifiers into domain and range identifiers
get_domain_range_ids([D,R],[D],[R]) :- !.
get_domain_range_ids([D1,D2|T],[D1|DT],R) :- get_domain_range_ids([D2|T],DT,R).


% detect whether there is a pattern of a recursive usage of the identifier:  ID(x) or ID[x] or x:ID
find_recursive_usage(TExpr,ID) :- 
    syntaxtraversion(TExpr,Expr,_,_,Subs,TNames), % print(try_id(ID,Expr,Subs,TNames)),nl,
    ((Expr = function(Fun,_), get_texpr_id(Fun,ID)) -> true
     ; (Expr = image(Rel,_), get_texpr_id(Rel,ID)) -> true
     ; (Expr = member(_,Set), get_texpr_id(Set,ID)) -> true
     ; \+ (member(ID1,TNames),get_texpr_id(ID1,ID)), % new local variable with same name
        member(Sub,Subs), find_recursive_usage(Sub,ID)
     ).

:- use_module(bsyntaxtree,[transform_bexpr/3]).
% find comprehension sets and mark them as recursive if they use the recursive ID
mark_recursion(TExpr,RecID,NewTExpr) :-
   (transform_bexpr(b_ast_cleanup:mark_comprehension_set(RecID),TExpr,NewTExpr)
     -> true
     ;  add_internal_error('Call failed: ',transform_bexpr(b_ast_cleanup:mark_comprehension_set(RecID),TExpr,NewTExpr)),
        NewTExpr=TExpr).

mark_comprehension_set(RecID,b(lambda(Ids,P,Expr),Type,Info),
                             b(lambda(Ids,P,Expr),Type,NInfo)) :-
        (find_recursive_usage(P,RecID) -> true ; find_recursive_usage(Expr,RecID)),
        get_texpr_ids(Ids,AtomicIds),
        format('Recursive lambda using ~w detected (name: ~w)~n',[AtomicIds,RecID]),
        error_manager:print_error_span(Info),nl,
        add_texpr_infos(Info,[prob_annotation('SYMBOLIC'),prob_annotation('RECURSIVE')],NInfo).
mark_comprehension_set(RecID,b(comprehension_set(Ids,P),Type,Info),
                       b(comprehension_set(Ids,P),Type,NInfo)) :-
        find_recursive_usage(P,RecID),
        get_texpr_ids(Ids,AtomicIds),
        format('Recursive comprehension set using ~w detected (name: ~w)~n',[AtomicIds,RecID]),
        error_manager:print_error_span(Info),nl,
        NInfo = [prob_annotation('SYMBOLIC'),prob_annotation('RECURSIVE')|Info]. % TO DO: only add if not already there

singleton_set_extension(set_extension([El]),El).
% singleton_set_extension(value(Set),b(value(El),Type,[])) :- custom_explicit_sets:singleton_set(Set,El). % TO DO: transmit type to be able to construct value
%singleton_set_extension(sequence_extension([El]),1 |-> El). % TO DO:

% comparison operators:
comparison(equal(A,B),A,B,equal(SA,SB),SA,SB).
comparison(not_equal(A,B),A,B,not_equal(SA,SB),SA,SB).
comparison(greater(A,B),A,B,greater(SA,SB),SA,SB).
comparison(less(A,B),A,B,less(SA,SB),SA,SB).
comparison(greater_equal(A,B),A,B,greater_equal(SA,SB),SA,SB).
comparison(less_equal(A,B),A,B,less_equal(SA,SB),SA,SB).
% rules to simplify binary comparison arguments
simplify_comparison_terms(b(A,T,_IA),b(B,T,_IB),RA,RB) :- 
    simplify_comparison_terms2(A,B,RA,RB).
    % TO DO: expand into much better simplifier !
simplify_comparison_terms2(minus(A1,A2),minus(B1,B2),ResA,ResB) :-
   (  same_texpr(A1,B1) -> ResA=B2, ResB=A2    % X-A2 < X-B2  <=>  B2 < A2
    ; same_texpr(A2,B2) -> ResA=A1, ResB=B1).  % A1-X < B1-X  <=>  A1 < B1
simplify_comparison_terms2(add(A1,A2),add(B1,B2),ResA,ResB) :-
   ( same_texpr(A1,B1) -> ResA=A2, ResB=B2     % X+A2 < X+B2  <=>  A2 < B2
    ; same_texpr(A2,B2) -> ResA=A1, ResB=B1    % A1+X < B1+X  <=>  A1 < B1
    ; same_texpr(A1,B2) -> ResA=A2, ResB=B1    % X+A2 < B1+X  <=>  A2 < B1
    ; same_texpr(A2,B1) -> ResA=A1, ResB=B2).  % A1+X < X+B2  <=>  A1 < B2
% multiplication: beware of sign, same with division

/* not used at the moment:
clpfd_arith_integer_expression(b(E,integer,_)) :- clpfd_arith_integer_expression_aux(E).
% check if we have an expression that can be dealt with by b_compute_arith_expression
clpfd_arith_integer_expression_aux(unary_minus(X)) :-
   X \= b(integer(_),integer,_). % if it is an explicit integer we can compute it normally
clpfd_arith_integer_expression_aux(add(_,_)).
clpfd_arith_integer_expression_aux(minus(_,_)).
clpfd_arith_integer_expression_aux(multiplication(_,_)).
*/

extract_unions(A,R) :- get_texpr_expr(A,union(A1,A2)),!,
     extract_unions(A1,R1), extract_unions(A2,R2),
     append(R1,R2,R).
extract_unions(A,[A]).

gen_member_predicates(B,SEl,TExpr) :- create_texpr(member(SEl,B),pred,[],TExpr).

% when constructing an expression/predicate: important to ripple wd information up
extract_important_info_from_subexpression(b(_,_,Info),NewInfo) :-
    (memberchk(contains_wd_condition,Info) -> NewInfo = [contains_wd_condition] ; NewInfo=[]).

extract_important_info_from_subexpressions(b(_,_,Info1),b(_,_,Info2),NewInfo) :-
    (memberchk(contains_wd_condition,Info1) -> NewInfo = [contains_wd_condition]
      ; memberchk(contains_wd_condition,Info2) -> NewInfo = [contains_wd_condition]
      ; NewInfo=[]).

% extract all assignments from a list of statements; last arg is the number of assignments extracted
extract_assignments([],[],[],[],0).
extract_assignments([b(assign_single_id(LHS,RHS),subst,_I)|T],ResLHS,ResRHS,Rest,Nr) :- !,
    ResLHS=[LHS|LT],ResRHS=[RHS|RT],
    extract_assignments(T,LT,RT,Rest,N), Nr = N.
extract_assignments([b(assign(LHS,RHS),subst,Info)|T],ResLHS,ResRHS,Rest,Nr) :- 
    \+ member(assigned_after(_),Info), % these assignments are treated in a special way
    \+ member(modifies(_),Info),
    !,
    extract_assignments(T,LT,RT,Rest,N),
    append(LHS,LT,ResLHS), append(RHS,RT,ResRHS), Nr is N+1.
extract_assignments([H|T],LT,RT,[H|Rest],Nr) :- 
    extract_assignments(T,LT,RT,Rest,Nr).
    % assigned_after(Primed),modifies(Var)

% check if we have a simple expression which will not be complicated to calculate
simple_expression(b(E,_,_)) :- simple2(E).
 %  (simple2(E) -> true ; print(not_simple(E)),nl).
simple2(bool_set).
simple2(boolean_false).
simple2(boolean_true).
simple2(empty_sequence).
simple2(empty_set).
simple2(identifier(_)).
simple2(integer_set(_)).
simple2(integer(_)).
simple2(lazy_lookup_expr(_)).
simple2(lazy_let_expr(_,A,B)) :- simple2(B), simple_expr_or_pred(A).
simple2(max_int).
simple2(min_int).
simple2(string_set).
simple2(string(_)).
simple2(value(_)).
simple2(first_of_pair(_)).  simple2(second_of_pair(_)).
simple2(add(A,B)) :- simple_expression(A), simple_expression(B).
simple2(minus(A,B)) :- simple_expression(A), simple_expression(B).
simple2(multiplication(A,B)) :- simple_expression(A), simple_expression(B).
simple2(unary_minus(A)) :- simple_expression(A).
simple2(convert_bool(A)) :- simple_predicate(A).
simple2(sequence_extension(A)) :- maplist(simple_expression,A). 
simple2(set_extension(A)) :- maplist(simple_expression,A). % a bit more expensive than sequence_extension: elements need to be compared

simple_expr_or_pred(b(E,T,_)) :- (T=pred -> simplep2(E) ; simple2(E)).

simple_predicate(b(E,_,_)) :- simplep2(E).
simplep2(equal(A,B)) :- simple_expression(A), simple_expression(B). % could be slightly more expensive if set type
simplep2(not_equal(A,B)) :- simple_expression(A), simple_expression(B). % ditto
simplep2(lazy_let_pred(_,A,B)) :- simple_predicate(B), simple_expr_or_pred(A).
simplep2(less(A,B)) :- simple_expression(A), simple_expression(B).
simplep2(less_equal(A,B)) :- simple_expression(A), simple_expression(B).
simplep2(greater(A,B)) :- simple_expression(A), simple_expression(B).
simplep2(greater_equal(A,B)) :- simple_expression(A), simple_expression(B).

:- use_module(self_check).
:- assert_must_succeed(b_ast_cleanup:nested_couple_to_list(b(couple(b(couple(a,b),x,[]),c),xx,[]),[a,b,c])).
:- assert_must_succeed(b_ast_cleanup:nested_couple_to_list(b(couple(a,c),x,[]),[a,c])).
:- assert_must_succeed(b_ast_cleanup:nested_couple_to_list(b(couple(a,b(couple(b,c),x,[])),xx,[]),[a,b(couple(b,c),x,[])])).
nested_couple_to_list(NC,L) :- nested_couple_to_list_dcg(NC,L,[]).
nested_couple_to_list_dcg(b(couple(A,B),_,_)) --> !,
   nested_couple_to_list_dcg(A), [B].
nested_couple_to_list_dcg(E) --> [E].


create_equalities_for_let(ORefs,Primed,Equalities) :-
    maplist(create_equality_for_let,ORefs,Primed,Equalities).
create_equality_for_let(oref(PrimedId,OrigId,Type),TPrimed,Equality) :-
    create_texpr(identifier(OrigId),Type,[],TOrig),
    create_texpr(identifier(PrimedId),Type,[],TPrimed),
    safe_create_texpr(equal(TPrimed,TOrig),pred,Equality).

% inserts a let statement. If the original statement is a precondition or any, the let is moved
% inside the original statement to prevent strange side-effects. This can be used for other,
% non-value changing substitutions as well.
insert_let(TExpr,Ids,P,NTExpr) :-
    remove_bt(TExpr,Expr,NewExpr,NTExpr),
    move_let_inside(Expr,Old,New,NewExpr),!,
    insert_let(Old,Ids,P,New).
insert_let(TExpr,Ids,P,NTExpr) :-
    create_texpr(let(Ids,P,TExpr),subst,[],NTExpr).
move_let_inside(precondition(Cond,Old),Old,New,precondition(Cond,New)).
move_let_inside(any(Any,Where,Old),Old,New,any(Any,Where,New)).

% find_one_point_rules(+TypedIds,+Blacklist,+Predicates,
%                      -LetIds,-LetExprs,-RemainingIds,-RemainingPredicates, +CheckWellDeff)
% TypedIds: The ids that are quantified in the exists clause
% Blacklist: All ids that must not be used in the found expression
% Predicates: The predicates of the exists (without already used id=E predicates)
% LetIds: The ids that can be introduced as LET
% LetExprs. For each id (in LetIds) the corresponding expression
% RemainingIds: The ids that are not converted into LETs
% RemainingPredicates: The predicates after removing the id=E predicates
% CheckWellDef: check_well_definedness --> only extract equalities from first pred or if always_well_defined
%  (e.g., f = {1|->2} & !e.(2:dom(f) & e=f(2) => e>100) should not generate a WD-error)
find_one_point_rules([],Preds,_,[],[],[],Preds,_).
find_one_point_rules([TId|Irest],Preds,Blacklist,LetIds,Exprs,RestIds,NewPreds,CheckWellDef) :-
    ( select_equality(TId,Preds,Blacklist,Expr,RestPred,_,CheckWellDef) ->
        LetIds = [TId|Lrest], Exprs = [Expr|Erest],
        RestIds = RestIds2, Preds2 = RestPred
    ; otherwise ->
        LetIds = Lrest, Exprs = Erest,
        RestIds = [TId|RestIds2], Preds2 = Preds),
    find_one_point_rules(Irest,Preds2,Blacklist,Lrest,Erest,RestIds2,NewPreds,CheckWellDef).
% select a predicate from Preds of the form id=Expr (or Expr=id) where Expr does not contain
% references to identifiers in Blacklist. Rest is Preds without id=Expr
select_equality(TId,Preds,Blacklist,Expr,Rest,UsedIds,CheckWellDef) :-
    get_texpr_id(TId,Id),
    is_equality(TEqual,TA,TB),
    safe_select(CheckWellDef,TEqual,Preds,Rest),
    ( get_texpr_id(TA,Id),TB=Expr ; get_texpr_id(TB,Id),TA=Expr ),
    find_identifier_uses(Expr,[],UsedIds),
    ord_disjoint(Blacklist,UsedIds).

% the ast_cleanup rules have not run yet on the sub-expressions of the exists: detect equalities
is_equality(TP,TA,TB) :- get_texpr_expr(TP,P), is_equality_aux(P,TA,TB).
is_equality_aux(equal(TA,TB),TA,TB).
is_equality_aux(partition(TA,[TB]),TA,TB).
is_equality_aux(member(TA,b(Set,_,_)),TA,TB) :- singleton_set_extension(Set,TB).
is_equality_aux(negation(b(not_equal(TA,TB),pred,_)),TA,TB).
% others not not_member ?
    
safe_select(check_well_definedness,Element,[H|T],Rest) :- !,
    (Element=H,Rest=T % either first element
      ; % or if later element; then it must be well-defined; otherwise H could fail
      select(Element,T,TRest), Rest=[H|TRest],
      \+ full_check_is_possibly_undefined(Element)  % we cannot use always_well_defined(Element)  in cleanup_pre; it is only computed in cleanup_post at the moment
      %, print(always_wd(Element)),nl
     ).
safe_select(_,Element,List,Rest) :- select(Element,List,Rest).

% split predicate into conjuncts using a certain list of ids and those not
split_predicate(Pred,Ids,UsingIds,NotUsingIds) :-
    conjunction_to_list(Pred,LP),
    get_sorted_ids(Ids,SIds),
    filter(not_occurs_in_predicate(SIds),LP,NP,UP),
    conjunct_predicates(NP,NotUsingIds),
    conjunct_predicates(UP,UsingIds).

not_occurs_in_predicate([],_Pred) :- !.
not_occurs_in_predicate(SortedIDs,Pred) :- SortedIDs = [ID1|_],
    (ID1 = b(_,_,_) -> add_internal_error('Wrapped identifiers: ',not_occurs_in_predicate(SortedIDs,Pred)) ; true),
    find_identifier_uses(Pred,[],UsedIds),
    ord_disjoint(SortedIDs,UsedIds).
get_sorted_ids(Ids,SIds) :- 
    get_texpr_ids(Ids,UnsortedIds),sort(UnsortedIds,SIds).

% basically we redo the work in select_equality: to do: cleanup and merge code
find_lets(Ids,Pred,[BID|IDT],[EQUALITY|T],RestIds,RestPred) :- Ids \= [],
    %print(find_lets(Ids)),nl,
    get_sorted_ids(Ids,SIds),
    b_interpreter:member_conjunct(Eq,Pred,RestPred1),
    id_equality(Eq,ID,RHS),
    select(BID,Ids,RestIds1),get_texpr_id(BID,ID),
    % Check that RHS does not contain any variable from Ids !
    not_occurs_in_predicate(SIds,RHS),
    !,
    %print(found_let(ID)),nl,
    %EQUALITY = b(equal(BID,RHS),pred,[generated_let]),
    safe_create_texpr(equal(BID,RHS),pred,EQUALITY),
    find_lets(RestIds1,RestPred1,IDT,T,RestIds,RestPred).
find_lets(Ids,Pred,[],[],Ids,Pred).
id_equality(Pred,ID,EqExpr) :- 
    get_texpr_expr(Pred,equal(LHS,RHS)),
    (get_texpr_id(LHS,ID), EqExpr=RHS
     ;
     get_texpr_id(RHS,ID), EqExpr=LHS).

pred_succ_compset(Op,comprehension_set([A,B],Pred)) :-
    %get_unique_id('_a_',AId),get_unique_id('_b_',BId),
    BId = '_lambda_result_',
    (Op = add -> AId = '_succ_' ; AId='_pred_'),
    create_texpr(identifier(AId),integer,[],A),
    create_texpr(identifier(BId),integer,[],B),
    create_texpr(integer(1),integer,[],Integer),
    create_texpr(ArithOp,integer,[],TArithOp),
    ArithOp =.. [Op,A,Integer],
    safe_create_texpr(equal(B,TArithOp),pred,Pred).

add_used_identifier_info(Ids,P,IOld,[used_ids(FoundIds)|I1]) :-
    % add used identifiers to information
    get_texpr_ids(Ids,Ignore),
    find_identifier_uses(P,Ignore,FoundIds), 
    %print(computed_used_ids(ignore(Ignore),found(FoundIds))),nl, translate:print_bexpr(P),nl,
    (select(used_ids(_),IOld,I1) -> true ; I1=IOld).

% can be used to check the validity of the used_ids field, e.g., for existential quantifier
check_used_ids_info(Parameters,Predicate,StoredUsedIds,PP) :-
    % get_global_identifiers(Ignored), and add to Parameters ??
    (add_used_identifier_info(Parameters,Predicate,[],[used_ids(UsedIds)])
      -> (StoredUsedIds=UsedIds -> true
         ; ord_subset(UsedIds,StoredUsedIds)
           -> format('Suboptimal used_ids info: ~w (actual ~w) [origin ~w with ~w]~n',[StoredUsedIds,UsedIds,PP,Parameters])
               %, translate:print_bexpr(Predicate),nl
         ;  add_internal_error('Incorrect used_ids info: ', check_used_ids_info(Parameters,Predicate,UsedIds,PP)),
            translate:print_bexpr(Predicate),nl,
            ord_subtract(UsedIds,StoredUsedIds,Delta),
            format('Incorrect used_ids info: ~w (actual ~w)~nNot included: ~w~n',[StoredUsedIds,UsedIds,Delta])
               %, translate:print_bexpr(Predicate),nl
         )
       ; add_internal_error('Could not computed used ids:',
                        add_used_identifier_info(Parameters,Predicate,[],[used_ids(_)]))
    ).

% update used_ids_info for existential and universal quantifier (at top-level only !)
recompute_used_ids_info(b(E,pred,Info1),Res) :- 
   recompute_used_ids_info_aux(E,Info1,Info2),!, Res= b(E,pred,Info2).
recompute_used_ids_info(TE,TE).

recompute_used_ids_info_aux(exists(Parameters,Predicate),Info1,Info2) :-
   add_used_identifier_info(Parameters,Predicate,Info1,Info2).
recompute_used_ids_info_aux(forall(Parameters,Lhs,Rhs),Info1,Info2) :-
   conjunct_predicates([Lhs,Rhs],Predicate),
   add_used_identifier_info(Parameters,Predicate,Info1,Info2).


% generation of unique identifiers
:- dynamic unique_id_counter/1.
unique_id_counter(1).

get_unique_id_inside(Prefix,Pred,ResultId) :-
   (\+ occurs_in_expr(Prefix,Pred) % first try and see whether we need to append a number
       -> ResultId = Prefix
       ;  get_unique_id(Prefix,ResultId)
    ).
get_unique_id_inside(Prefix,Pred,Expr,ResultId) :-
   ((\+ occurs_in_expr(Prefix,Pred),
      \+ occurs_in_expr(Prefix,Expr)) % first try and see whether we need to append a number
       -> ResultId = Prefix
       ;  get_unique_id(Prefix,ResultId)
    ).

get_unique_id(Prefix,Id) :-
    retract(unique_id_counter(Old)),
    New is Old + 1,
    assert(unique_id_counter(New)),
    safe_atom_chars(Prefix,CPrefix,get_unique_id1),
    number_chars(Old,CNumber),
    append(CPrefix,CNumber,CId),
    safe_atom_chars(Id,CId,get_unique_id2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% replace all variables in types by "any"

ground_type(T,_) :- ground(T),!.
ground_type(T,Exceptions) :- var(T),!,
    ( exact_member(T,Exceptions) -> !
    ; otherwise -> T=any).
ground_type(T,Exceptions) :-
    functor(T,_,Arity),
    ground_type_args(Arity,T,Exceptions).
ground_type_args(0,_T,_Exceptions) :- !.
ground_type_args(N,T,Exceptions) :-
    arg(N,T,Arg),ground_type(Arg,Exceptions),
    N2 is N-1,ground_type_args(N2,T,Exceptions).

% put optional before/after usage into the information of the identifiers
% makes only sense in the context of becomes_such substitutions
add_before_after_info([],_,[]).
add_before_after_info([TId|IRest],BeforeAfter,[TId2|ORest]) :-
    def_get_texpr_id(TId,Id),
    ( member(ba(Id,BeforeId),BeforeAfter) ->
        get_texpr_type(TId,Type),
        get_texpr_info(TId,Info),
        create_texpr(identifier(Id),Type,[before_substitution(Id,BeforeId)|Info],TId2)
    ; otherwise ->
        TId = TId2 ),
    add_before_after_info(IRest,BeforeAfter,ORest).

% find all used pairs of before/after variables, e.g. ba(x,'x$0')
% see becomes_such substitutions
find_used_primed_ids(TExpr,Uses) :-
    find_used_primed_ids2(TExpr,[],Uses).
find_used_primed_ids2(TExpr,In,Out) :-
    syntaxtraversion(TExpr,Expr,_,Infos,Subs,_),
    ( Expr = identifier(_), member(before_substitution(Id,FullId),Infos) ->
        ord_add_element(In,ba(Id,FullId),Uses)
    ; otherwise ->
        In = Uses ),
    foldl(find_used_primed_ids2,Subs,Uses,Out).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% find expressions which are possibly not well-defined

is_possibly_undefined(Expr) :-
        syntaxtraversion(_,Expr,_,_,Subs,_),
        member(Sub,Subs),
        get_texpr_info(Sub,Info),
        memberchk(contains_wd_condition,Info),!.
is_possibly_undefined(Expr) :-
        functor(Expr,F,_),has_wd_condition(F,Expr).  %%,print(wd(F,Expr)),nl.

% a version which does the full traversal; can be used before contains_wd_condition has been computed
full_check_is_possibly_undefined(BExpr) :- get_texpr_expr(BExpr,Expr),
        full_check_is_possibly_undefined_aux(Expr).
full_check_is_possibly_undefined_aux(Expr) :-
        functor(Expr,F,_),has_wd_condition(F,Expr),!.
full_check_is_possibly_undefined_aux(Expr) :-
        syntaxtraversion(_,Expr,_,_,Subs,_),
        member(Sub,Subs),
        full_check_is_possibly_undefined(Sub),!.

% division and module must not divide by zero
has_wd_condition(div,Expr) :- arg(2,Expr,DIV),
    \+ definitely_not_zero(DIV).
has_wd_condition(modulo,Expr) :- arg(2,Expr,DIV),
    (\+ definitely_not_zero(DIV) ;
     \+ definitely_not_negative(DIV) ;
      arg(1,Expr,A1), \+ definitely_not_negative(A1)).
% power_of must have a non-negative exponent (?)
has_wd_condition(power_of,Expr) :- arg(2,Expr,EXP),
   \+ definitely_not_negative(EXP).
% min and max need a non-empty set
% also Atelier-B manual requires the set to have an upper limit
has_wd_condition(max,Expr) :- arg(1,Expr,S), \+ definitely_not_empty_and_finite(S).
has_wd_condition(min,Expr) :- arg(1,Expr,S), \+ definitely_not_empty_and_finite(S).
% all the sequence operations must not be applied to non-sequence
% relations and some must not be applied to empty-sequences
has_wd_condition(size,E) :- arg(1,E,S), \+ definitely_sequence(S).
has_wd_condition(first,E) :- arg(1,E,S), \+ definitely_not_empty_sequence(S).
has_wd_condition(last,E) :- arg(1,E,S), \+ definitely_not_empty_sequence(S).
has_wd_condition(front,E) :- arg(1,E,S), \+ definitely_not_empty_sequence(S).
has_wd_condition(tail,E) :- arg(1,E,S), \+ definitely_not_empty_sequence(S).
has_wd_condition(rev,E) :- arg(1,E,S), \+ definitely_sequence(S).
has_wd_condition(concat,_).
has_wd_condition(insert_front,_).
has_wd_condition(insert_tail,_).
has_wd_condition(restrict_front,_).
has_wd_condition(restrict_tail,_).
has_wd_condition(general_concat,_).
% functions must not be applied to values outside their domain
has_wd_condition(function,_).
% the general intersection must not be applied to an empty set of sets
has_wd_condition(general_intersection,_).
has_wd_condition(quantified_intersection,_). % gets translated to general_intersection
% card must not be applied to infinite sets:
has_wd_condition(card,Expr) :- arg(1,Expr,S), \+ definitely_finite(S).

has_wd_condition(freetype_destructor,_).

definitely_not_zero(b(integer(X),integer,_)) :- number(X), X \= 0.
definitely_not_negative(b(integer(X),integer,_)) :- number(X), X >= 0.
% to do add more ?: card(_), ...

definitely_not_empty_and_finite(b(S,_,_)) :- definitely_not_empty2(S).
definitely_not_empty2(set_extension(_)).
definitely_not_empty2(sequence_extension(_)).
definitely_not_empty2(value(S)) :- kernel_objects:not_empty_set(S).

definitely_not_empty_sequence(b(S,_,_)) :- definitely_not_empty_sequence2(S).
definitely_not_empty_sequence2(sequence_extension(_)).

definitely_sequence(b(S,_,_)) :- definitely_sequence2(S).
definitely_sequence2(empty_sequence).
definitely_sequence2(sequence_extension(_)).

definitely_finite(b(S,_,_)) :- definitely_finite2(S).
definitely_finite2(empty_set).
definitely_finite2(empty_sequence).
definitely_finite2(set_extension(_)).
definitely_finite2(sequence_extension(_)).
definitely_finite2(value(S)) :- (S=[] ; S=avl_set(_)).

definitely_maximal_set(b(S,_,_)) :- definitely_maximal2(S).
definitely_maximal2(integer_set('INTEGER')).
definitely_maximal2(bool_set).
definitely_maximal2(string_set).
% TO DO: add more rules


