% (c) 2013-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(b_expression_sharing,
          [ apply_expression_sharing_to_machine/2
          , create_initial_expression_cache/1
          , find_common_subexpressions/4
          , cse_optimize_predicate/2
          , expand_cse_lazy_lets/2 ,expand_cse_lazy_lets_if_cse_enabled/2
          ]).

:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(library(avl)).

:- use_module(tools).
:- use_module(self_check).
:- use_module(error_manager).
:- use_module(debug).

:- use_module(bsyntaxtree).
:- use_module(bmachine_structure).

:- use_module(preferences,[get_preference/2]).

expression_sharing_enabled :-
    get_preference(use_common_subexpression_elimination,true).

create_subcache(Id,cache(Id,Avl)) :- empty_avl(Avl).

create_initial_expression_cache(A) :-
    expression_sharing_enabled,!,
    A=[Cache],create_subcache(top,Cache).
create_initial_expression_cache(disabled).


apply_expression_sharing_to_machine(InMachine,OutMachine) :-
    change_sections(InMachine,InExprs,OutExprs,OutMachine),
    create_stores(0,EmptyStores),
    remove_implicit_negation_l(InExprs,NormedExprs),
    find_common_subexpressions_l(NormedExprs,EmptyStores,true,machine,1,OutExprs,Stores,_Stripped,_UsedIds),
    ground_reference_counters(Stores),!.
    
change_sections(InMachine,InExprs,OutExprs,OutMachine) :-
    change_section2([constraints,properties,invariant,
                     assertions,operation_bodies],
                    InMachine,InExprs,OutExprs,OutMachine).
change_section2([],Machine,[],[],Machine).
change_section2([Sec|Rest],InMachine,InExprs,OutExprs,OutMachine) :-
    select_section_texprs(Sec,LInExprs,LOutExprs,InMachine,InterMachine),
    append(LInExprs,RInExprs,InExprs),
    append(LOutExprs,ROutExprs,OutExprs),
    change_section2(Rest,InterMachine,RInExprs,ROutExprs,OutMachine).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(bsyntaxtree).

% TESTS: x+y >100 & x+y <99 now fails deterministically in REPL with CSE

% Note: we should detect if we are at the invariant level before calling cse_optimize_predicate: do not mess up Rodin rodinpos proof information or syntactically_relevance info

cse_optimize_predicate(Pred,Res) :-
   cse_print_check('Running Common Sub Expressions (CSE) Optimization on ', Pred),
   expand_cse_lazy_lets(Pred,ExpandedPred), % TO DO: see if we can do this more efficiently; do we need it at all ??
   cse_print_check('Expanded Expression: ', ExpandedPred),
   % Note: we can still have unexpanded lazy-lets hanging around in ExpandedPred from outer scopes
   find_maximum_lazy_let_id(ExpandedPred,MaxID), % find maximum ID
   StartNr is MaxID+1, % we will start numbering from there to avoid clashes with existing lazy-lets
   find_common_subexpressions(ExpandedPred,StartNr,true,R,_Refs),
   cse_print_check('Found common sub expressions: ',R),
   empty_avl(E),
   transform_bexpr_with_acc(b_expression_sharing:insert_lazy_let,R,OptimizedPred,E,Acc),
   %portray_avl(Acc),nl,
   cse_check_ast(result,OptimizedPred),
   (empty_avl(Acc) -> true ; 
     add_internal_error('cse_optimize_predicate: AVL Accumulator not empty: ',Acc)),
   cse_print_check('Optimized Expression Result: ',OptimizedPred),
   transform_bexpr(b_expression_sharing:cse_cleanup_infos,OptimizedPred,COPred),
   cse_print_check('Cleaned up Optimized Expression Result: ',COPred),
   !,
   Res= COPred.
cse_optimize_predicate(P,Res) :-
   add_internal_error('CSE FAILED: ',cse_optimize_predicate(P,Res)),
   Res=P.

:- use_module(translate,[print_bexpr/1]).
cse_print_check(Msg,Expr) :- debug:debug_mode(on), %%
   !,
   print(Msg), print_bexpr(Expr), nl,
   cse_check_ast(Msg,Expr).
cse_print_check(_,_).

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_safe_mode,true)).
% check the AST (should be removed in production version of ProB):
cse_check_ast(PP,Pred) :-
   debug_print(4,'Checking AST : '), debug_println(4,PP), % translate:print_bexpr(Pred),nl,
   check_ast(true,Pred).
:- else.
cse_check_ast(_PP,_Pred).
:- endif.


% -----------------------------------------------

% utility to expand lazy let expressions for better CSE if repeatedly applied
% Note: lazy_lookups could remain if only a partial predicate is passed

expand_cse_lazy_lets_if_cse_enabled(E,EE) :-
   (expression_sharing_enabled -> expand_cse_lazy_lets(E,EE) ; E=EE).

expand_cse_lazy_lets(Expr,ExpandedExpr) :-
   transform_bexpr(b_expression_sharing:expand_lazy_let,Expr,ExpandedExpr).

:- public expand_lazy_let/2.
%expand_lazy_let(B,R) :- tools_printing:print_term_summary(B),nl,fail.
expand_lazy_let(b(E,Type,Info),Result) :- expand_lazy_let_aux(E,Type,Info,Result).
expand_lazy_let_aux(lazy_let_expr(TID,SharedExpr,MainExpr),_,_,Result) :-
   get_texpr_id(TID,ID), debug_println(4,expanding(ID)),
   expand_lazy_lookup(ID,SharedExpr,MainExpr,Result).
expand_lazy_let_aux(lazy_let_pred(TID,SharedExpr,MainExpr),_,_,Result) :- 
   get_texpr_id(TID,ID), debug_println(4,expanding(ID)),
   expand_lazy_lookup(ID,SharedExpr,MainExpr,Result).
expand_lazy_let_aux(E,Type,Info,Result) :- E \= value(_), E \= integer(_),
   compute_static_expression(E,Val),
   %print(computed(E,Val)),nl,
   Result=b(value(Val),Type,Info).

expand_lazy_lookup(ID,SharedExpr,MainExpr,Result) :- \+ atom(ID),!,
   add_internal_error('Illegal call: ',expand_lazy_lookup(ID,SharedExpr,MainExpr,Result)),
   Result=MainExpr.
expand_lazy_lookup(ID,SharedExpr,MainExpr,Result) :-
   transform_bexpr(b_expression_sharing:replace_lazy_lookup(ID,SharedExpr),MainExpr,Result).

% TO DO: unify with b_compile and ast_cleanup !
compute_static_texpression(b(E,_,_),Result) :- compute_static_expression(E,Result).
compute_static_int_texpression(b(E,_,_),Result) :- compute_static_expression(E,int(Result)).

compute_static_expression(add(A,B),int(Result)) :- % plus
    compute_static_int_texpression(A,IA), compute_static_int_texpression(B,IB),
    Result is IA+IB.
compute_static_expression(minus(A,B),int(Result)) :- % plus
    compute_static_int_texpression(A,IA), compute_static_int_texpression(B,IB),
    Result is IA-IB.
compute_static_expression(unary_minus(A),int(Result)) :- % plus
    compute_static_int_texpression(A,IA), 
    Result is -IA.
compute_static_expression(multiplication(A,B),int(Result)) :-
    compute_static_int_texpression(A,IA), compute_static_int_texpression(B,IB),
    Result is IA*IB.
compute_static_expression(power_of(A,B),int(Result)) :-
    compute_static_int_texpression(A,IA), compute_static_int_texpression(B,IB), IB >= 0,
    kernel_objects:safe_pown(IA,IB,Result), number(Result).
% TO DO: modulo, division, set_extension
compute_static_expression(integer(I),int(I)).
compute_static_expression(value(I),I).
compute_static_expression(equal(A,B),I) :- compute_static_int_texpression(A,SA), compute_static_int_texpression(B,SB),
    (SA==SB -> I = pred_true ; I = pred_false).
compute_static_expression(not_equal(A,B),I) :- compute_static_int_texpression(A,SA), compute_static_int_texpression(B,SB),
    (SA==SB -> I = pred_false ; I = pred_true).


% -----------------------------------------------

% find maximum id used in lazy lookups; useful to generate new unused identifiers
find_maximum_lazy_let_id(BExpr,MaxID) :-
   reduce_over_bexpr(b_expression_sharing:find_maximum_lazy_let_id_aux,BExpr,-1,MaxID).

:- public find_maximum_lazy_let_id_aux/3.
find_maximum_lazy_let_id_aux(lazy_lookup_expr(ID),MaxSoFar,NewMax) :- !,
   convert_atom_to_nr(ID,IDNr),
   (IDNr @> MaxSoFar -> NewMax=IDNr ; NewMax=MaxSoFar).
find_maximum_lazy_let_id_aux(lazy_lookup_pred(ID),MaxSoFar,NewMax) :-
   convert_atom_to_nr(ID,IDNr),
   (IDNr @> MaxSoFar -> NewMax=IDNr ; NewMax=MaxSoFar).
find_maximum_lazy_let_id_aux(_,V,V).

% -----------------------------------------------

:- public insert_lazy_let/4.
% traverse the syntax tree where the sharing/4 info has been deposited by
% find_common_subexpressions and now insert lazy_lets at the lowest possible pointes
% and replace usage of CSE by lazy_lookups
insert_lazy_let(BE,Res,SubAccs,NewAcc) :- %print(lazy_let(BE)),nl,
   update_info_field(BE,BE2), % we have just treated sub-expressions and we may insert BE into avl tree of shared expressions
   combine_accs(SubAccs,CombinedAcc),
   %portray_avl(CombinedAcc),nl,
   insert_lazy_let1(BE2,Res1,CombinedAcc,NewAcc),
   copy_position_info(BE2,Res1,Res2),
   update_info_field(Res2,Res). % maybe redundant ?

% copy position information; important e.g. for discharged checks for theorems and invariants
copy_position_info(From,b(To,Type,Info1),b(To,Type,Info3)) :-
   get_texpr_pos(From,PositionInfo),!,
   delete(Info1,nodeid(_),Info2),
   Info3 = [nodeid(PositionInfo)|Info2].
copy_position_info(_,E,E).

:- use_module(b_ast_cleanup,[recompute_used_ids_info/2]).
% update info field of a typed expression after a modification of the expression
update_info_field(TE,NewTE) :-
   update_wd_condition(TE,TE2),
   recompute_used_ids_info(TE2,NewTE). % inserting lazy_lookups changes used ids for exists/forall

% check if we need to add contains_wd_condition info field
update_wd_condition(b(E,Type,Info1),Res) :-
   nonmember(contains_wd_condition,Info1), % inlined update_wd_condition_info to just copy E in clause below; not sure this helps memory performance
   sub_expression_contains_wd_condition(E),
   !,
   Res = b(E,Type,[contains_wd_condition|Info1]).
update_wd_condition(E,E).

update_wd_condition_info(E,Info1,Info2) :-
   nonmember(contains_wd_condition,Info1),
   sub_expression_contains_wd_condition(E),
   !,
   Info2 = [contains_wd_condition|Info1].
update_wd_condition_info(_,I,I).

insert_lazy_let1(MainExpression,Res,CombinedAcc,NewAcc) :-
   avl_delete(completed,CombinedAcc,CompletedAVL,Acc1),
   % we have seen all uses of a certain common sub expression: introduce a let now
   % Note: we need to insert the lower-level ones first (i.e., at the outer level)
   !,
   MainExpression = b(_Expr,Type,Info),
   avl_to_list(CompletedAVL,ComplList), 
   %print(complete(ComplList)),nl,
   insert_lazy_lookups(MainExpression,SubRes,Acc1,NewAcc),
   insert_completed_lazy_lets(ComplList,Type,Info,Res,SubRes).
insert_lazy_let1(MainExpression,Res,CombinedAcc,NewAcc) :-
   insert_lazy_lookups(MainExpression,Res,CombinedAcc,NewAcc).

insert_completed_lazy_lets([],_,_,Res,Sub) :- update_info_field(Sub,Res).
insert_completed_lazy_lets([ID-cshared(_SameNr,SharedExpression)|_T],_Type,_Info,_Res,_Sub) :-
   SharedExpression = b(lazy_lookup_expr(ID2),_T2,_I2),
   convert_to_atom(ID,AID),
   print(simple_equality(AID,ID2)),nl, % @ID = @ID2
   % TO DO: replace AID by SharedExpression in T and MainExpression above
   fail.
insert_completed_lazy_lets([ID-cshared(SameNr,SharedExpression)|T],Type,Info,Res,Sub) :-
   %T = [ID2-cshared(SameNr,OtherSharedExpression)|T2],
   append(T1,[ID2-cshared(SameNr,OtherSharedExpression)|T2],T), % find an ID with SameNr of uses in which ID is used
   cse_id_used(ID,OtherSharedExpression), %
   % Then SharedExpression occurs *only* in OtherSharedExpression; we do not need outer LET
   convert_to_atom(ID,AID),
   delete_sharing_info(SharedExpression,SharedExpression2),
   debug_println(4,nested_let(ID,ID2,nr(SameNr),AID)),
   transform_bexpr(b_expression_sharing:replace_lazy_lookup(AID,SharedExpression2),OtherSharedExpression,NewOther),
   !,
   cse_print_check('Nested LET Replaced : ',NewOther),
   append(T1,[ID2-cshared(SameNr,NewOther)|T2],NewCShList), % insert updated ID2 equality at right place again in List
   insert_completed_lazy_lets(NewCShList,Type,Info,Res,Sub).
insert_completed_lazy_lets([ID-cshared(Nr,SharedExpression)|T],Type,Info,Result,Sub) :-
   % all uses of ID lie below in the AST: introduce the LET
   debug_println(4,completed(Type,ID,Nr,SharedExpression)),
   %translate:print_bexpr(Sub),nl,
   delete_sharing_info(SharedExpression,SharedExpression2),
   insert_completed_lazy_lets(T,Type,Info,SubRes,Sub),
   %print('Sub res: '),translate:print_bexpr(SubRes),nl,
   gen_lazy_let_with_check(Type,ID,SharedExpression2,SubRes,LET),
   Result=LET,
   cse_print_check('LET: ',LET).

% Remove sharing/4 info from SharedExpression + avoid reusing it for CSE if algorithm applied second time
delete_sharing_info(b(E,T,Info),b(E,T,NewInfo)) :-
   delete(Info,sharing(_,_,_,_),NewInfo).

:- public cse_cleanup_infos/2.
% Remove info that is no longer necessary once CSE has run to completion:
cse_cleanup_infos(b(E,T,Info),b(NewE,T,NewInfo)) :-
   delete(Info,cse_used_ids(_),Info2),
   ((nonmember(contains_wd_condition,Info2), % TO DO: allow in DISPROVER_MODE
     post_cse_simplifier(E,SE)) -> NewE = SE, NewInfo=Info2 % Note: maybe some lets no longer necessary now
   ; reorder_lazy_let_expr(E,T,Info2,NewE,NewInfo) -> true
   ; NewE=E,NewInfo=Info2).

% reorder/move lazy lets outside if this leads to better optimizations being applicable
reorder_lazy_let_expr(Expr,Type,Info,NewExpr,NewInfo) :- 
    reorder_argument(Expr,FunctorID,Match,ArgToReorder,OtherArgs),
    ArgToReorder = b(lazy_let_expr(ID,ShExpr,BodyA),_TypeA,_InfoA), % TO DO: peel off multiple lazy_lets
    get_texpr_expr(BodyA,Match),
    print(reorder_lazy_let_expr(ID,FunctorID)),nl,
    !,
    reorder_argument(NewBinExpr,FunctorID,_,BodyA,OtherArgs),
    !, % FunctorID ensures only one possible solution for reorder_argument
    NewExpr = lazy_let_expr(ID,ShExpr,b(NewBinExpr,Type,Info)),
     % copy important info from InfoA ??
    update_wd_condition_info(NewExpr,Info,NewInfo).
% TO DO: add more cases + what if both arguments need reordering --> fixpoint computation ?

% reorder argument 1 in case we match a certain pattern
% reorder_argument(Expression, ID, MATCH,  ARG_TO_REORDER,  OtherArgs)
reorder_argument(function(A,B),function_1,set_extension(_),A,B). % there is optimized code: b_apply_function_set_extension
reorder_argument(max(A),max,set_extension(_),A,nil). % we have optimized code for min({a,b,...}) and max
reorder_argument(min(A),min,set_extension(_),A,nil).
reorder_argument(card(A),card,set_extension(_),A,nil).
reorder_argument(image(A,B),image_1,closure(_),A,B).
reorder_argument(image(B,A),image_2,set_extension(_),A,B).

% TO DO: bring this in accordance with do_not_share_this rules:

% detect certain obvious inconsistencies / tautologies introduced by CSE
post_cse_simplifier(equal(A,B),truth) :- same_lazy_lookup(A,B).
post_cse_simplifier(not_equal(A,B),falsity) :- same_lazy_lookup(A,B).
post_cse_simplifier(greater(A,B),falsity) :- same_lazy_lookup(A,B).
post_cse_simplifier(greater_equal(A,B),truth) :- same_lazy_lookup(A,B).
post_cse_simplifier(less(A,B),falsity) :- same_lazy_lookup(A,B).
post_cse_simplifier(less_equal(A,B),truth) :- same_lazy_lookup(A,B).
post_cse_simplifier(subset(A,B),truth) :- same_lazy_lookup(A,B).
post_cse_simplifier(not_subset(A,B),falsity) :- same_lazy_lookup(A,B).
post_cse_simplifier(strict_subset(A,B),falsity) :- same_lazy_lookup(A,B).
post_cse_simplifier(not_strict_subset(A,B),truth) :- same_lazy_lookup(A,B).

same_lazy_lookup(A,B) :- get_lazy_lookup_id(A,ID), get_lazy_lookup_id(B,ID).

get_lazy_lookup_id(b(E,Type,_),Type,ID) :- get_lazy_lookup_id_aux(E,ID).
get_lazy_lookup_id(b(E,_,_),ID) :- get_lazy_lookup_id_aux(E,ID).
get_lazy_lookup_id_aux(lazy_lookup_expr(ID),ID).
get_lazy_lookup_id_aux(lazy_lookup_pred(ID),ID).

gen_lazy_let_with_check(Type, Identifier, SharedExpression,  SubExpressionInsideLet,  Result) :-
    cse_check_ast(before_gen_lazy_let_shared,SharedExpression),
    cse_check_ast(before_gen_lazy_let_body,SubExpressionInsideLet),
    (gen_lazy_let(Type, Identifier, SharedExpression,  SubExpressionInsideLet,  Result)
      -> cse_check_ast(gen_lazy_let,Result)
      ; add_internal_error('Call failed: ',gen_lazy_let_with_check(Type, Identifier, SharedExpression,  SubExpressionInsideLet,  Result)),
        fail).

% TO DO: maybe check that we are not in a bad context to prevent optimized treatment (e.g. {1|->a,2|->b}(x) ), rather than delegating this to reorder lazy_let code above ?
% gen_lazy_let(Type, Identifier, SharedExpression,  SubExpressionInsideLet,  Result)
gen_lazy_let(Type,AID,SE,Sub,LET) :- var(Sub),
    print(illegal_call_var_subexpression(gen_lazy_let(Type,AID,SE,Sub,LET))),nl,fail.
gen_lazy_let(Type,AID,SE,Sub,LET) :- \+ ground(Sub),
    print(nonground_subexpression(gen_lazy_let(Type,AID,SE,Sub,LET))),nl,fail.
gen_lazy_let(pred,AID,SharedExpr,b(forall(Ids,A,B),pred,I),Result) :-
   % check if Ids occur in SharedExpr; otherwise no need to add extra implication
    bsyntaxtree:find_identifier_uses(SharedExpr, [], UsedIds),
    get_texpr_ids(Ids,ParIDs),
    member(ForallID,ParIDs), member(ForallID,UsedIds), % Forall id used in SharedExpr; we cannot wrap the let around the forall
    !,
    debug_println(4,forall_shared_id(ForallID,let_pred(AID))), %trace,
    % we must put the let inside the forall as it references IDs introduced by the forall
    % TO DO: if the expression is too simple it may not be worthwhile to introduce the additional implication !!
    convert_to_atom(AID,ID),
    % Disable CSE of this ID completely; TO DO: improve the forall creation below to intelligently create universal quantifiers
    expand_lazy_lookup(ID,SharedExpr,b(forall(Ids,A,B),pred,I),Result), print(expanded(ID)),nl,
    cse_check_ast(gen_lazy_let_forall,Result),
    translate:print_bexpr(Result),nl.
    /*
    gen_id(AID,SharedExpr,ID),
    collect_important_info(A,B,Impl_Info),
    Body = b(implication(A,B),pred,Impl_Info),
    FORALL = forall(Ids,Truth,b(lazy_let_pred(ID,SharedExpr,Body),pred,LLP_Info)),
    Truth = b(truth,pred,[]),
    Result = b(FORALL,pred,LLP_Info),
    collect_important_info(SharedExpr,Body,LLP_Info) ,translate:print_bexpr(Result),nl,nl. */
gen_lazy_let(pred,CSEID,SharedExpr,Body,Result) :- 
    % check if we can split off Prefix and/or Suffix of Body not using CSEID
    cse_split_conjunct(Body, CSEID, Prefix, MainBody, Suffix),
    !,
    debug_println(4,optimizing_let_with_conjunction_body(CSEID)),
    gen_id(CSEID,SharedExpr,ID),
    cse_conjunct_predicates(MainBody,NewBody),
    %get_texpr_info(NewBody,NewInfo),
    collect_important_info(SharedExpr,NewBody,NewInfo), % we could remove ID from cse_used_ids
    LET = b(lazy_let_pred(ID,SharedExpr,NewBody),pred,NewInfo),
    append(Prefix,[LET|Suffix],ConjunctionList),
    cse_conjunct_predicates(ConjunctionList,Result).
gen_lazy_let(pred,A,SharedExpr,Body,b(R,pred,Info)) :- 
    !,
    gen_id(A,SharedExpr,ID), debug_println(4,generating_lazy_let_pred(A,ID)),
    R = lazy_let_pred(ID,SharedExpr,Body),
    collect_important_info(SharedExpr,Body,Info). % we could remove ID from cse_used_ids
gen_lazy_let(Type,A,B,C,b(lazy_let_expr(ID,B,C),Type,Info)) :-
    gen_id(A,B,ID),
    collect_important_info(B,C,Info). % we could remove ID from cse_used_ids

% split a conjunct into  PREFIX &  MAIN  & SUFFIX   such that PREFIX/SUFFIX do not use CSEID identifier
cse_split_conjunct(Conjunct,CSEID,  PREFIX,MAIN,SUFFIX) :-
    is_a_conjunct(Conjunct,_,_),
    % try removing conjuncts which do not use CSEID by flattening first:
    conjunction_to_list(Conjunct,List),
    find_prefix_not_using_cse_id(List,CSEID,PREFIX,Rest1),
    reverse(Rest1,RevRest1),
    find_prefix_not_using_cse_id(RevRest1,CSEID,RevSuffix,RevMain),
    (PREFIX \= [] ; RevSuffix \= []), % only succeed if there is a non-trivial split
    reverse(RevMain,MAIN),
    reverse(RevSuffix,SUFFIX).
   

collect_important_info(SubA,SubB,Info) :- (var(SubA) ; var(SubB)), !,
    add_internal_error('Illegal call: ',collect_important_info(SubA,SubB,Info)),fail.
collect_important_info(SubA,SubB,Info) :-
    extract_info(SubA,SubB,Info1), % WD info (bsyntaxtree utility)
    collect_uses_cse_id([SubA,SubB],IDS), % CSE usage info
    exclude(is_cse_used_ids,Info1,Info2), % is this necessary ??
    Info = [cse_used_ids(IDS)|Info2].
    

gen_lazy_lookup(pred,A,R) :- !, convert_to_atom(A,ID), R = lazy_lookup_pred(ID).
gen_lazy_lookup(_,A,lazy_lookup_expr(ID)) :- convert_to_atom(A,ID).

% generate atoms: many identifier treating predicates expect atoms
gen_id(A,B,b(identifier(ID),Type,[])) :- get_texpr_type(B,Type),convert_to_atom(A,ID).

convert_to_atom(A,ID) :- var(A),!, add_internal_error('Illegal call: ', convert_to_atom(A,ID)),fail.
convert_to_atom(A,ID) :- atom(A),!, ID=A.
convert_to_atom(Nr,ID) :- number_codes(Nr,C), atom_codes(ID,[64|C]). % 64 = @

convert_atom_to_nr(Atom,Nr) :- var(Atom),!, add_internal_error('Illegal call: ',convert_atom_to_nr(Atom,Nr)),fail.
convert_atom_to_nr(Atom,Nr) :- atom_codes(Atom,[64|C]), number_codes(Nr,C).

% find prefix list of conjuncts which do not use CSEID:
find_prefix_not_using_cse_id([],_CSEID,[],[]).
find_prefix_not_using_cse_id([H|T],CSEID,Prefix,Rest) :-
   (uses_cse_id(CSEID,H)
    -> Prefix=[], Rest= [H|T]
     ; % print('NOT USING '), print(CSEID),nl, translate:print_bexpr(H),nl,
       Prefix = [H|TP],
       find_prefix_not_using_cse_id(T,CSEID,TP,Rest)).

% a version of conjunction which recomputes the cse_used_ids INFO
cse_conjunct_predicates(L,Res) :- cse_conjunct_predicates2(L,b(E,pred,I)),
    (collect_uses_cse_id(L,CSEINFO)
      -> Res = b(E,pred,[cse_used_ids(CSEINFO)|I1]),
         exclude(is_cse_used_ids,I,I1) %delete(I,cse_used_ids(_),I1)
      ;  print('*** Could not collect uses_cse_id INFO :'), print(L),nl,
         Res = b(E,pred,I)).

% sometimes we get identical conjuncts; TO DO: generalize to longer lists,...  (test 351 still leaves redundant conjuncts inside let body)
cse_conjunct_predicates2([A,B],R) :- A==B, !, R=A.
cse_conjunct_predicates2(L,R) :- conjunct_predicates(L,R).

is_cse_used_ids(cse_used_ids(_)).

collect_uses_cse_id([],[]).
collect_uses_cse_id([H|T],Res) :-
    get_cse_used_ids_info(H,CSE),!,
    collect_uses_cse_id(T,CSET),
    ord_union(CSE,CSET,Res).

uses_cse_id(ID,Expr) :- get_cse_used_ids_info(Expr,List),!,
   member(ID,List).
uses_cse_id(_ID,Expr) :- print('Could not extract cse_used_ids info: '), print(Expr),nl,
   true.

% CSE Identifier usage infor either form info field or from direct usage by lazy_lookups
get_cse_used_ids_info(TExpr,IDList) :- get_texpr_expr(TExpr,E),
   directly_uses_cse_id(E,UsedID),!,
   IDList = [UsedID].
get_cse_used_ids_info(TExpr,IDList) :- get_texpr_info(TExpr,Info),
   member(cse_used_ids(IDList),Info),!.
get_cse_used_ids_info(TE,L) :- print('Could not get cse_used_ids info: '), print(TE),
   L=[].

directly_uses_cse_id(E,ID) :- var(E),!, add_internal_error('Illegal call: ',directly_uses_cse_id(E,ID)),fail.
directly_uses_cse_id(lazy_lookup_pred(AtomID),ID) :- convert_atom_to_nr(AtomID,ID).
directly_uses_cse_id(lazy_lookup_expr(AtomID),ID) :- convert_atom_to_nr(AtomID,ID).

insert_lazy_lookups(BExpr,b(LAZYLOOKUP,Type,NewInfo),CombinedAcc,NewAcc) :-
   BExpr = b(_Expr,Type,Info),
   select(sharing(ID,TotalCount,_,_),Info,NewInfo), % this expressions is marked as being shared
   TotalCount > 1, % with at least two occurrences
   !,
   gen_lazy_lookup(Type,ID,LAZYLOOKUP),
   %print(lookup(ID)),nl, translate:print_bexpr(BExpr), portray_avl(CombinedAcc),nl,
   cse_check_ast(avl_store(ID),BExpr),
   avl_store(ID,CombinedAcc,shared(1,TotalCount,BExpr),NewAcc).
insert_lazy_lookups(BExpr,BExpr,Acc,Acc).

% check if a typed expression references a shared Common-Sub-Expression ID
cse_id_used(ID,TypedExpr) :- get_texpr_info(TypedExpr,Info),
   member(cse_used_ids(L),Info),
   member(ID,L).

:- public replace_lazy_lookup/4.
:- use_module(btypechecker, [unify_types_strict/2]).
% TO DO: maybe remove cse_used_ids ??
%replace_lazy_lookup(_,_,T,_) :- print(T),nl,fail.
replace_lazy_lookup(ID,SharedExpr,BExpr,SharedExpr) :- 
   get_lazy_lookup_id(BExpr,Type,TID),
    !,
   TID=ID,
   get_texpr_type(SharedExpr,SharedType),
   (my_unify_types(SharedType,Type) -> true 
    ; add_internal_error('Incompatible type: ',replace_lazy_lookup(ID,'\\='(SharedType,Type),BExpr,SharedExpr))).
replace_lazy_lookup(_,_,E,Res) :- update_info_field(E,Res). % probably not necessary ?!

:- if(environ(prob_safe_mode,true)).
my_unify_types(T1,T2) :- (\+ ground(T1) ; \+ ground(T2)),
   format('*** Non-ground type unification: ~w = ~w~n~n',[T1,T2]),
   fail.
:- endif.
my_unify_types(T1,T2) :- unify_types_strict(T1,T2).

% -----------------

% combine accumulators bottom-up from subexpressions:
% merge the information from number of encountered uses of CSEs in different branches
% mark a CSE as completed if we have encountered all uses now
combine_accs([],E) :- empty_avl(E).
combine_accs([H],R) :- !, R=H.
combine_accs([H|T],R) :- combine_accs(T,RT),
   merge_avls(H,RT,R).

% merge two AVL accumulators
merge_avls(AVL1,AVL2,ResultAVL) :-
   avl_to_list(AVL1,L1),
   merge2(L1,AVL2,ResultAVL).

merge2([],A,A).
merge2([ID-shared(Nr1,TotalCount,BE)|T],Acc,R) :- check_id(ID),
   (avl_fetch(ID,Acc,shared(Nr2,TotalCount,_))
     -> Nr is Nr1+Nr2,
        (Nr=TotalCount -> avl_fetch_completed(Acc,CompletedAVL),
                     avl_store(ID,CompletedAVL,cshared(Nr,BE),CompletedAVL1), % we have reached the lowest point where we can introduce a let for this identifier covering all uses
                     avl_store(completed,Acc,CompletedAVL1,Acc1), 
                     avl_delete(ID,Acc1,_,Acc2)
                  ;  avl_store(ID,Acc,shared(Nr,TotalCount,BE),Acc2))
     ;  avl_store(ID,Acc,shared(Nr1,TotalCount,BE),Acc2) % necessary ???
  ), merge2(T,Acc2,R).

% lookup if we have information about completed CSEs (i.e., those for which all occurences have now been seen); return empty_avl otherwise
avl_fetch_completed(Acc,CompletedAVL) :-
    (avl_fetch(completed,Acc,CompletedAVL) -> true ; empty_avl(CompletedAVL)).

check_id(completed) :- print('UNEXPECTED completed INFO'),nl,fail.
check_id(_).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% common subexpression detection

:- assert_must_succeed(
   (   find_common_subexpressions(b(identifier(i),integer,[]),0,false,R),
       R == b(identifier(i),integer,[sharing(0,1,top,[]),cse_used_ids([])])
   )).
:- assert_must_succeed(
   (   find_common_subexpressions(b(identifier(i),integer,[]),0,true,R),
       R == b(identifier(i),integer,[cse_used_ids([])])
   )).
:- assert_must_succeed(
   (   I = b(identifier(i),integer,[]),
       O = b(identifier(i),integer,[sharing(0,2,top,[]),cse_used_ids([])]),
       find_common_subexpressions(b(add(I,I),integer,[]),0,false,R),
       R == b(add(O,O),integer,[sharing(1,1,top,[]),cse_used_ids([0])])
   )).
:- assert_must_succeed(
   (   I = b(identifier(i),integer,[]),
       O = b(identifier(i),integer,[cse_used_ids([])]),
       find_common_subexpressions(b(add(I,I),integer,[]),0,true,R),
       R == b(add(O,O),integer,[sharing(0,1,top,[]),cse_used_ids([])])
   )).
% two quantifiers with the same variable name must not share
% expressions that contain the variable, other expressions should be
% shared
:- assert_must_succeed(
   (   C = b(identifier(c),integer,[]),
       X = b(identifier(x),integer,[]),
       I = b(identifier(i),set(integer),[]),
       O = b(integer(1),integer,[]),
       %  !x.(x:I => x<c+1)
       F = b(forall([X],
                    b(member(X,I),pred,[]),
                    b(less(X,b(add(C,O),integer,[])),pred,[])),pred,[]),
       %  !x.(x:I => x>c+1)
       Fx = b(forall([X],
                    b(member(X,I),pred,[]),
                    b(greater(X,b(add(C,O),integer,[])),pred,[])),pred,[]),
       C1 = b(identifier(c),integer,[sharing(3,2,top,[]),cse_used_ids([])]),
       X1 = b(identifier(x),integer,[sharing(0,3,7,[]),cse_used_ids([])]),
       I1 = b(identifier(i),set(integer),[sharing(1,2,top,[]),cse_used_ids([])]),
       O1 = b(integer(1),integer,[sharing(4,2,top,[]),cse_used_ids([])]),
       F1 = b(forall([X1],
                     b(member(X1,I1),pred,[sharing(2,1,7,[]),cse_used_ids([0,1])]),
                     b(less(X1,b(add(C1,O1),integer,[sharing(5,2,top,[]),cse_used_ids([3,4])])),
                           pred,[sharing(6,1,7,[]),cse_used_ids([0,3,4,5])])),
                pred,[sharing(7,1,top,[0]),cse_used_ids([0,1,2,3,4,5,6])]),
       % e.g. c+1 should be shared (ref. 5)
       % x and x:I must not be shared, they get a new reference (8 and 9)
       X2 = b(identifier(x),integer,[sharing(8,3,11,[]),cse_used_ids([])]),
       F2 = b(forall([X2],
                     b(member(X2,I1),pred,[sharing(9,1,11,[]),cse_used_ids([1,8])]),
                     b(greater(X2,b(add(C1,O1),integer,[sharing(5,2,top,[]),cse_used_ids([3,4])])),
                                      pred,[sharing(10,1,11,[]),cse_used_ids([3,4,5,8])])),
                pred,[sharing(11,1,top,[8]),cse_used_ids([1,3,4,5,8,9,10])]),
       find_common_subexpressions(b(conjunct(F,Fx),pred,[]),0,false,R),
       R == b(conjunct(F1,F2),pred,[sharing(12,1,top,[]),cse_used_ids([0,1,2,3,4,5,6,7,8,9,10,11])])
   )).

% find_common_subexpressions/4 searches for common subexpressions and adds an
%   information field sharing/4 to each AST node.
% find_common_subexpressions(+TExpr,+Start,+SkipSimple,-NTExpr):
%   TExpr: the original AST
%   Start: at which number the IDs should start (usually 0)
%   NTExpr: the AST, each node annotated with an sharing(ID,Counter,SubId,References) information field
%      two syntax nodes share the same ID when they should yield always the same
%      value. Counter counts the number of occurrences of the same expression in
%      the given AST. SubId is the unique reference of the quantified expression
%      where a used ID is introduced, top if there is no such quantifier.
%      References is a list of ids that are are sub-expressions of the node which
%        occur more than once and which depend on the newly introduced identifiers
%      (Please note: The declaration of a variable in a quantified expression
%       is an occurrence of its own.)
%   SkipSimple: set to true if expression sharing should not be applied to simple
%      expressions like identifier(_), integer(_), etc.
find_common_subexpressions(TExpr,Start,SkipSimple,NTExpr) :-
    find_common_subexpressions(TExpr,Start,SkipSimple,NTExpr,_).
find_common_subexpressions(TExpr,Start,SkipSimple,NTExpr,References) :-
    create_stores(Start,EmptyStores),
    find_common_subexpressions2(TExpr,EmptyStores,SkipSimple,root/1,NTExpr,Stores,_Stripped,_UsedIds,_UsedCSEIds),
    %print(grounding(Stores)),nl,
    ground_reference_counters(Stores), %print(stores(Stores)),nl, print(NTExpr),nl,
    get_current_references(Stores,References). % needed ??

:- if(environ(prob_safe_mode,true)).
add_cse_used_ids_info(Ids,OldInfos,NewInfos) :- (var(Ids) ; \+ ground(OldInfos)),!,
   add_internal_error('Illegal Call: ',add_cse_used_ids_info(Ids,OldInfos,NewInfos)),
   NewInfos = OldInfos.
:- endif.
add_cse_used_ids_info(Ids,OldInfos,NewInfos) :-
    exclude(is_cse_used_ids,OldInfos,OldInfos1),
    NewInfos=[cse_used_ids(Ids)|OldInfos1].

% find_common_subexpressions2(+TExpr,+InStores,+SkipSimple,+OuterFunctor,-NTExpr,-OutStores,-Stripped,-UsedIds,-UsedCSEIds):
find_common_subexpressions2(TExpr,InStores,SkipSimple,OuterFunctor,NTExpr,OutStores,Stripped,UsedIds,UsedCSEIds) :-
    %print(find(TExpr,InStores,outer(OuterFunctor))),nl,
    % decompose the typed expression into it components (Subs), look up the newly introduced
    % identifiers (NewIds) and create a new typed expression (NTExpr) with the modified components
    % (NSubs) and the new Info field (Info)
    fc_parts(TExpr,Expr,Type,Subs,NewIds,NSubs,OldInfos,NewInfos,NTExpr),
    % introduce a map from expression to reference that is only visible to this node
    % and it subnodes if new identifiers (NewIds) are introduced
    % OutStores1 and OutStores contain a pattern such that if OutStores1 is set,
    % OutStores is exactly like OutStores1 but without visibility of the newly introduced
    % identifiers.
    create_substore(NewIds,ID,InStores,SubStores),
    % apply subexpression detection on the subexpressions
    functor(Expr,OuterExprFunctor,_), % memorise functor for sub-expressions (allows deciding if sharing makes sense)
    find_common_subexpressions_l(Subs,SubStores,SkipSimple,OuterExprFunctor,1,
                                 NSubs,SubStores1,StrippedSubs,UsedSubIds,SubUsedCSEIds),
    % create a stripped (i.e. no info field, no types) version of the syntax node for look ups
    create_stripped(Expr,StrippedSubs,Stripped),
    %print(stripped(Stripped)),nl,
    LookupKey = cse_lookup_key(Stripped,Type), % we also store the type along with the stripped AST (avoid sharing, e.g., two prj1/2 functions with different types, see test 255 !)
    % merge the used ids of the subnodes with the used ids in this node,
    % subtract newly introduced identifiers because they are not visible from outside
    update_used_ids(Expr,NewIds,UsedSubIds,UsedIds),
    % if a substore was introduced due to newly introduced identifiers, remove it
    remove_last_substore(NewIds,SubStores1,Stores1,References),
    % try to find the expression in the stores
    ( SkipSimple==true,do_not_share_this(Stripped,UsedIds,OuterFunctor,Type,OldInfos) ->
        OutStores = Stores1,
        add_cse_used_ids_info(SubUsedCSEIds,OldInfos,NewInfos),
        UsedCSEIds = SubUsedCSEIds,
        store_expression(LookupKey,UsedIds,Stores1,_,_,ID) /* just ground ID; used for Sub expressions */
        %, print(not_sharing(ID))
    ;  lookup_expression(LookupKey,UsedIds,Stores1,OutStores,SharingInfo,ID) ->
        /* it is shared */
        NewInfos=[SharingInfo|OldInfos1],
        add_cse_used_ids_info(SubUsedCSEIds,OldInfos,OldInfos1),
        ord_union([ID],SubUsedCSEIds,UsedCSEIds),
        set_references(References,SharingInfo)
        % In this case we do not need to look for sharing in the sub-expressions !!
        %,check_sharing_info(lookup,SharingInfo)
    ; otherwise ->
        % if not found, store the expression: it will be shared if we use it again later
        store_expression(LookupKey,UsedIds,Stores1,OutStores,SharingInfo,ID),
        ord_union([ID],SubUsedCSEIds,UsedCSEIds),
        %print(new_id(ID,SubUsedCSEIds)),nl,
        % TO DO: what if later we find out that it is shared after all -> we should remove sub-expression counts
        add_cse_used_ids_info(SubUsedCSEIds,OldInfos,OldInfos1),
        NewInfos=[SharingInfo|OldInfos1],
        set_references(References,SharingInfo)
        %,check_sharing_info(new,SharingInfo)
    ). %  print(done_find(NTExpr,OutStores,UsedIds,UsedCSEIds)),nl.

%check_sharing_info(_,sharing(Id,_RefCounter,Level,Refs)) :- number(Id), ground(Level), ground(Refs),!.
%check_sharing_info(PP,SI) :- print(illegal_sharing_info(PP,SI)),nl.

fc_parts(TExpr,Expr,Type,Subs,NewIds,NSubs,OldInfos,NewInfos,NTExpr) :-
    create_texpr(Expr,Type,OldInfos,TExpr),
    syntaxtransformation(Expr,Subs,Names,NSubs,NExpr),
    get_texpr_ids(Names,NewIds),
    create_texpr(NExpr,Type,NewInfos,NTExpr).

find_common_subexpressions_l([],Stores,_SkipSimple,_OuterFunctor,_,[],Stores,[],[],[]).
find_common_subexpressions_l([Expr|Erest],InStores,SkipSimple,OuterFunctor,ArgNr,
                             [NExpr|Nrest],OutStores,[Stripped|Srest],UsedIds,UsedCSEIds) :-
    find_common_subexpressions2(Expr,InStores,SkipSimple,outer(OuterFunctor,ArgNr),NExpr,InterStores,Stripped,UsedIds1,UsedCSEIds1),
    A1 is ArgNr+1,
    find_common_subexpressions_l(Erest,InterStores,SkipSimple,OuterFunctor,A1,Nrest,OutStores,Srest,UsedIds2,UsedCSEIds2),
    ord_union(UsedIds1,UsedIds2,UsedIds), % the B identifiers used
    ord_union(UsedCSEIds1,UsedCSEIds2,UsedCSEIds). % the identifiers to CSE used

boolean(boolean_true).
boolean(boolean_false).

do_not_share_this(_Stripped,_,_Functor,_Type,Infos) :- 
  memberchk(contains_wd_condition,Infos), 
  get_preference(use_common_subexpression_wd_only,true),
  !. % also examine DISPROVER MODE ?
do_not_share_this(_Stripped,_,_Functor,Type,_Infos) :- Type=pred,
  get_preference(use_common_subexpression_also_for_predicates,false),
  !.
do_not_share_this(Stripped,UsedIds,OuterFunctor,_Type,_Infos) :-  % print(chk(Stripped,OuterFunctor)),nl,
  do_not_share_this_3(Stripped,UsedIds,OuterFunctor).

do_not_share_this_3(set_extension(X),_,outer(OuterFunctor,ArgNr)) :- !,
   %nl,print(sx(X,OuterFunctor/ArgNr)),nl,
   (X = [V], simple_explicit_value(V) ;
   set_extension_optimized_treatment_available(OuterFunctor,ArgNr)).
do_not_share_this_3(comprehension_set(_,_),_,M) :- nmem(M). % optimized treatment available; important for test 1079

:- use_module(external_functions,[not_declarative/1]).
%do_not_share_this_3(closure(_L),_,outer(OuterFunctor,ArgNr)) :- !,
%   closure1_optimized_treatment_available(OuterFunctor,ArgNr). 
%do_not_share_this_3(sequence_extension(_L),_,outer(OuterFunctor,ArgNr)) :- !,  % we need to add b_compute_card2 for seq ?
/*

do_not_share_this_3(X,_,M) :- nmem(M), functor(X,F),
   (kernel_mappings:symbolic_closure_unary_operator(F) ;
    kernel_mappings:symbolic_closure_binary_operator(F)).

% case for unary_in_boolean_type solutions:
*/
do_not_share_this_3(pow_subset(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(fin_subset(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(pow1_subset(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(fin1_subset(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(struct(_),_,M) :- nmem(M).
do_not_share_this_3(seq(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(iseq(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(perm(_),_,M) :- nmem(M).
do_not_share_this_3(domain(_),_,M) :- nmem(M).
do_not_share_this_3(seq1(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(iseq1(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(identity(X),_,M) :- nmem(M) ; infinite_set(X).
do_not_share_this_3(closure(_),_,M) :- nmem(M).
do_not_share_this_3(external_pred_call(FunName,_Args),_,_) :- not_declarative(FunName). % in principle any element containing this should not be shared (but anyway there are little gurantees about preservation and order of side-effects)
do_not_share_this_3(external_function_call(FunName,_Args),_,_) :- not_declarative(FunName).
% case for binary_not_in_boolean_type solutions:
/*
do_not_share_this_3(partial_bijection(_,_),_,M) :- nmem(M).
do_not_share_this_3(total_relation(_,_),_,M) :- nmem(M).
do_not_share_this_3(surjection_relation(_,_),_,M) :- nmem(M).
do_not_share_this_3(total_surjection_relation(_,_),_,M) :- nmem(M).
do_not_share_this_3(parallel_product(_,_),_,M) :- nmem(M).
% set_subtraction, intersection not covered
*/
do_not_share_this_3(interval(A,B),_,M) :- nmem(M) ; small_interval(A,B).
do_not_share_this_3(total_function(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).
do_not_share_this_3(total_injection(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).
do_not_share_this_3(total_surjection(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).
do_not_share_this_3(total_bijection(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).
do_not_share_this_3(partial_function(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).
do_not_share_this_3(partial_injection(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).  % >+>
do_not_share_this_3(partial_surjection(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).
do_not_share_this_3(relations(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).
do_not_share_this_3(cartesian_product(A,B),_,M) :- nmem(M) ; infinite_set(A) ; infinite_set(B).
%do_not_share_this_3(div(A,B),_,_) :- A=identifier(xx), B=integer(1000000000). % xx / 1000000000 ; purpose: test vesg performance issue
%do_not_share_this_3(range(identifier('Block_Id')),_,_). % for vesg performance test in INVARIANT

do_not_share_this_3(Stripped,UsedIds,_OuterFunctor) :- too_simple_for_sharing2(Stripped,UsedIds).

% check if expression used as second arg of member check:
nmem(outer(M,2)) :- (M=member ; M=not_member).
nmem(outer(M,_)) :- (M=partial_function ; M=total_function ; M=relations).
small_interval(integer(A),integer(B)) :- number(A),number(B), B < A+10000000.
infinite_set(integer_set(_)). % :- (X='INTEGER' ; X = 'NATURAL' ; X = 'NATURAL1').
infinite_set(string_set).
% TO DO : couple
simple_explicit_value(integer(_)).
simple_explicit_value(boolean_true).
simple_explicit_value(boolean_false).

set_extension_optimized_treatment_available(card,1).
set_extension_optimized_treatment_available(min,1). % we have optimized code for min({a,b,...}) and max
set_extension_optimized_treatment_available(max,1).
set_extension_optimized_treatment_available(function,1).
set_extension_optimized_treatment_available(subset,2). % there is a rule in cleanup_post which will rewrite this to member checks {x1,x2,...} <: B  <=>  x1:B & x2:B & ...; we could also try and ensure that this runs first !
set_extension_optimized_treatment_available(image,2). % ??

%closure1_optimized_treatment_available(image,1). % there is some custom code for: image_for_closure1_wf; but it is also called if we do CSE ??!!

:- use_module(external_functions,[performs_io/1]).
too_simple_for_sharing2(equal(A,B),_) :- 
 (A = identifier(_Id), boolean(B) ;
  B = identifier(_Id), boolean(A) ). 
too_simple_for_sharing2(identifier(Id),[Id]).
too_simple_for_sharing2(unary_minus(Int),[]) :- Int = integer(_).
too_simple_for_sharing2(integer(_),[]).
too_simple_for_sharing2(value(_),[]).
too_simple_for_sharing2(boolean_true,[]).
too_simple_for_sharing2(boolean_false,[]).
too_simple_for_sharing2(integer_set(_),[]).
too_simple_for_sharing2(bool_set,[]).
too_simple_for_sharing2(string(_),[]).
too_simple_for_sharing2(string_set,[]).
too_simple_for_sharing2(empty_set,[]).
too_simple_for_sharing2(empty_sequence,[]).
too_simple_for_sharing2(lazy_lookup_pred(_),_).
too_simple_for_sharing2(lazy_lookup_expr(_),_).
too_simple_for_sharing2(lazy_let_expr(_,_,_),_).
too_simple_for_sharing2(lazy_let_pred(_,_,_),_).
too_simple_for_sharing2(couple(_,_),_). % maybe could be useful nonetheless ??

%too_simple_for_sharing2(interval(integer(_),integer(_)),_).
% TO DO: do not share arithmetic expressions which can be constant folded ??
% or maybe perform constant folding at the same time ??

too_simple_for_sharing2(truth,[]).
too_simple_for_sharing2(falsity,[]).

% TO DO: completely avoid sharing those expressions ? but usually they should not occur inside predicates/expressions anyway
too_simple_for_sharing2(external_pred_call(FunName,_Args),_) :-
        external_functions:performs_io(FunName).
too_simple_for_sharing2(external_function_call(FunName,_Args),_) :-
        external_functions:performs_io(FunName).



set_references(References,sharing(_Id,_RefCounter,_Level,References)).

% the stores data type has the following form:
% stores(Counter,References,Top,Substores)
% - Counter is the next ID for new expressions, the same expressions
%   share the same id
% - Complete is a mapping from each expression to its id (is this sensible?)
% - References is a mapping from an ID to a term rc(C,R) where C
%   contains the number of occurrences of the expressions so far and R the
%   final number of occurrences of this expression. Until the end, R is a variable
%   that will be unified with the last value of C
% - Top is a mapping from a stripped expression to an ID. This mapping is for all
%   expressions that do not contain references to quantified variables
% - Substores is a list of terms substore(Ids,SubStore) where Ids is a list of
%   introduced identifiers in a qualified expression and Substore is a mapping from
%   a stripped expression to its ID. The expression contains references to identifiers
%   in Ids. In case of nested quantified expressions, the inner expression has
%   a substore entry before the outer expression in the list.
create_stores(Start,stores(Start,Empty,Empty,[])) :-
    empty_avl(Empty).
create_substore([],_SubId,Stores,Stores) :- !.
create_substore(NewIds,ID,InStores,SubStores) :-
    force_create_substore(NewIds,ID,InStores,SubStores).
force_create_substore(Ids,SubId,
                stores(Counter,IR,Top, InSubs),
                stores(Counter,IR,Top, [substore(OrdIds,Empty,SubId,[])|InSubs])) :-
    list_to_ord_set(Ids,OrdIds),
    empty_avl(Empty).

% remove_last_substore(+Names,+InSubStores,-OutSubStores,-References)
remove_last_substore([],Stores,Stores,[]).
remove_last_substore([_|_],
                     stores(Counter,IR,Top,[Substore|Subs]),
                     stores(Counter,IR,Top,Subs),
                     References) :-
    Substore = substore(_Names,_Store,_SubId,References).
    
% try and get references to common sub expression ids + the corresponding expressions
get_current_references(stores(_Counter,IR,TopStore,_Subs),References) :-
   avl_to_list(IR,Refs),
   convlist(reused_expression(TopStore),Refs,References).
reused_expression(TopStore,ID-rc(Count,Count),cse(ID,Expr,Count)) :- Count>1,
   avl_member(Expr,TopStore,ID). % reverse lookup; not efficient !

lookup_expression(Stripped,UsedIds,InStores,OutStores,sharing(Id,RefCounter,Level,_References),Id) :-
    InStores =  stores(Counter,InRefCounter, TopStore,OldSubStores),
    OutStores = stores(Counter,OutRefCounter,TopStore,NewSubStores),
    lookup_expression2(OldSubStores,TopStore,Stripped,UsedIds,Id,Level,NewSubStores),!,
    avl_fetch(Id,InRefCounter,rc(OC,RefCounter)),
    NC is OC+1,
    avl_store(Id,InRefCounter,rc(NC,RefCounter),OutRefCounter).
lookup_expression2([],TopStore,Stripped,_UsedIds,Id,top,[]) :-
    avl_fetch(Stripped,TopStore,Id).
lookup_expression2([substore(Names,Store,SubId,OldRefs)|Rest],_TopStore,Stripped,_UsedIds,Id,SubId,
                   [substore(Names,Store,SubId,NewRefs)|Rest]) :-
    avl_fetch(Stripped,Store,Id),!,
    ord_add_element(OldRefs,Id,NewRefs).
lookup_expression2([substore(Names,Store,SubId,Refs)|Rest],TopStore,Stripped,UsedIds,Id,Level,
                   [substore(Names,Store,SubId,Refs)|NRest]) :-
    ord_disjoint(UsedIds,Names),!,
    lookup_expression2(Rest,TopStore,Stripped,UsedIds,Id,Level,NRest).

%lookup_next_id(stores(Id,_,_,_),Id).

store_expression(Stripped,UsedIds,InStores,OutStores,sharing(Id,RefCounter,Level,_References),Id) :-
    InStores = stores(Id,InRefCounter,InTopStore,InSubStores),
    Counter is Id+1, % Generate a new CSE Identifier
    OutStores = stores(Counter,OutRefCounter,OutTopStore,OutSubStores),
    store_expression2(InSubStores,InTopStore,Stripped,UsedIds,Id,OutSubStores,OutTopStore,Level),
    avl_store(Id,InRefCounter,rc(1,RefCounter),OutRefCounter).
store_expression2([],InTopStore,Stripped,_UsedIds,Id,[],OutTopStore,top) :-
    avl_store(Stripped,InTopStore,Id,OutTopStore).
store_expression2([substore(Names,InStore,SubId,Refs) |Srest],TopStore,Stripped,UsedIds,Id,
                  [substore(Names,OutStore,SubId,Refs)|Srest],TopStore,SubId) :-
    ord_intersect(Names,UsedIds),!, % the Shared Expression makes use of identifiers/names introduced at that level; it cannot be re-used higher up
    avl_store(Stripped,InStore,Id,OutStore).
store_expression2([Substore|ISrest],InTopStore,Stripped,UsedIds,Id,
                  [Substore|OSrest],OutTopStore,Level) :-
    store_expression2(ISrest,InTopStore,Stripped,UsedIds,Id,OSrest,OutTopStore,Level).

update_used_ids(Expr,NewlyIntroducedIds,UsedSubIds,UsedIds) :-
    update_used_ids2(Expr,UsedSubIds,UsedIds1),
    ord_subtract(UsedIds1,NewlyIntroducedIds,UsedIds). % remove NewlyIntroducedIds; they are not visible outside
update_used_ids2(identifier(Id),UsedSubIds,UsedIds) :- !, % we directly use an identifier
    ord_add_element(UsedSubIds,Id,UsedIds).
update_used_ids2(_Expr,UsedSubIds,UsedIds) :- !, UsedSubIds=UsedIds.

% We do not have to use strip_and_norm_ast/2 from bsyntaxtree though
% the result should be the same. We already have the stripped versions
% of the subexpressions, so we can re-use that information
create_stripped(Expr,StrippedSubs,NormStripped) :-
    syntaxtransformation(Expr,_Subs,_Names,StrippedSubs,Stripped),
    norm_strip(Stripped,NormStripped).

ground_reference_counters(stores(_Counter,RefCounter,_TopStore,_SubStores)) :-
    avl_to_list(RefCounter,List),
    maplist(ground_reference_counter,List).
ground_reference_counter(_Id-rc(C,C)).

:- use_module(library(samsort)).
norm_strip(empty_sequence,R) :- !, R = empty_set.
norm_strip(greater_equal(A,B),R) :- !, R = less_equal(B,A).
norm_strip(greater(A,B),R) :- !, R = less(B,A).
norm_strip(set_extension(SX),R) :- !,
    sort(SX,SSX),
    R = set_extension(SSX).
norm_strip(Old,New) :-
    functor(Old,Functor,2),
    is_commutative(Functor),!,
    norm_commutative(Functor,Old,New).
norm_strip(Old,Old).

norm_commutative(Functor,Old,New) :- is_associative(Functor), !,
    Old =.. [Functor,OA,OB],
    % now collect all args with same functor in left and right:
    get_top_level_functor_arguments(OA,Functor,LeftArgs),
    get_top_level_functor_arguments(OB,Functor,RightArgs),
    append(LeftArgs,RightArgs,Args),
    samsort(Args,SortedArgs),
    New =.. [Functor,SortedArgs].
norm_commutative(Functor,Old,New) :- % the functor is not associative (or cannot appear at top-level in OA,OB)
    Old =.. [Functor,OA,OB],
    New =.. [Functor,NA,NB],
    ( OA @>= OB -> OA=NB,OB=NA
    ; otherwise -> OA=NA,OB=NB).

% we could use this instead of samsort
% but we get error message e.g. for % (x\/y\/v\/x) /= (v\/x\/y) & x<:BOOL  as an expression appears within itself ?!
sort_args(Functor,Args,SortedArgs) :-
 (idempotent(Functor) -> sort(Args,SortedArgs) % we can remove duplicates
      ; samsort(Args,SortedArgs)).


get_top_level_functor_arguments(Top,Functor,Args) :- % must also be associative
    functor(Top,Functor,N), !,
    (N=2 -> Top =.. [Functor|Args] % no normalisation applied
       ; arg(1,Top,Args)). % Note: we do not recurse: we assume normalisation has already been applied fully below
get_top_level_functor_arguments(Top,_,[Top]).

is_commutative(conjunct). % commutative only if there are no WD issues !
is_commutative(disjunct). % commutative only if there are no WD issues !
is_commutative(equivalence).
is_commutative(equal).
is_commutative(not_equal).
is_commutative(add).
is_commutative(multiplication).
is_commutative(union).
is_commutative(intersection).


is_associative(conjunct). % even if there are WD issues !
is_associative(disjunct). % even if there are WD issues !
is_associative(add).
is_associative(multiplication).
is_associative(union).
is_associative(intersection).
is_associative(concat).

idempotent(conjunct).
idempotent(disjunct).
idempotent(union).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% removing implicit negation:
%  predicates of the form a/=b (resp. a/:b, ...) are
%  rewritten to not(a=b), (resp. not(a:b), ...) to enable sharing of the
%  predicate a=b.

remove_implicit_negation(TExpr,NTExpr) :-
    remove_bt(TExpr,Expr,NewExpr,TExpr2),
    syntaxtransformation(Expr,Subs,_Names,NSubs,NewExpr),
    remove_implicit_negation_l(Subs,NSubs),
    remove_implicit_negation2(TExpr2,NTExpr).

remove_implicit_negation_l([],[]).
remove_implicit_negation_l([OExpr|ORest],[NExpr|NRest]) :-
    remove_implicit_negation(OExpr,NExpr),
    remove_implicit_negation_l(ORest,NRest).

remove_implicit_negation2(TExpr,NTExpr) :-
    get_texpr_expr(TExpr,Expr),
    remove_implicit_negation3(Expr,NTExpr),!.
remove_implicit_negation2(TExpr,TExpr).

remove_implicit_negation3(negation(P),R) :-
    get_texpr_expr(P,negation(Q)),!,R=Q. % eliminate double negation
remove_implicit_negation3(not_equal(A,B),R) :- negate_bexpr(equal(A,B),A,B,R).
remove_implicit_negation3(not_member(A,B),R) :- negate_bexpr(member(A,B),A,B,R).
remove_implicit_negation3(not_subset(A,B),R) :- negate_bexpr(subset(A,B),A,B,R).
remove_implicit_negation3(not_subset_strict(A,B),R) :- negate_bexpr(subset_strict(A,B),A,B,R).

negate_bexpr(Expr,SubA,SubB,R) :- extract_info(SubA,SubB,Info),
    create_texpr(Expr,pred,Info,P),
    create_negation(P,R).


% TO DO: optimize away unnecessary nested LETs as below:
% Optimized pred: 
% !(sddb1,sddb2).((sddb1 : INTEGER & sddb2 : INTEGER) & (sddb1 |-> sddb2 : dom(SDDBs_Signal) & (Signals_Cap__Name~)(Routes_Cap__Opposite_Signal_ID(route)) : ran(SDDBs_Signal(sddb1 |-> sddb2))) => (LET (@9)==ran(tses) IN (LET (@10)==dom(@9) IN sddb1 /: @10 & sddb2 /: @10)))
