% (c) 2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(before_after_predicates, [before_after_predicate/2,
                                    before_after_predicate_list_conjunct/2,
                                    before_after_predicate_list_disjunct/2,
                                    tcltk_before_after_predicate_for_operation/2,
                                    before_after_predicate_with_equalities/3,
                                    before_after_predicate_list_conjunct_with_equalities/3,
                                    before_after_predicate_list_disjunct_with_equalities/3,
                                    list_renamings_to_primed_vars/2]).

:- use_module(library(lists)).

:- use_module(bsyntaxtree, [get_texpr_expr/2,
                            get_texpr_id/2,
                            safe_create_texpr/3,
                            get_texpr_type/2,
                            conjunct_predicates/2,
                            conjunction_to_list/2,
                            create_implication/3,
                            create_couple/3,
                            disjunct_predicates/2,
                            find_typed_identifier_uses/2,
                            remove_all_infos/2,
                            rename_bt/3,
                            create_exists/3,
                            create_negation/2]).
:- use_module(error_manager, [add_error_fail/3]).
:- use_module(bmachine_eventb, [prime_variable/2]).
:- use_module(bmachine,[b_get_machine_operation/4,
                        b_get_machine_variables/1]).
%:- use_module(specfile,[animation_minor_mode/1]).


:- use_module(module_information,[module_info/2]).
:- module_info(group,misc).
:- module_info(description,'This module computes before-after-predicates of B and Event-B actions.').

tcltk_before_after_predicate_for_operation(OpName,list(Res)) :-
    b_get_machine_operation(OpName,_Results,_Parameters,OpBody),
    b_get_machine_variables(Variables),
    before_after_predicate_with_equalities(OpBody,Variables,BAPredicate),
    translate:print_bexpr(BAPredicate),nl,
    conjunction_to_list(BAPredicate,LP),
    maplist(translate:translate_bexpression,LP,Res).

:- mode before_after_predicate(+Action,+AllIdentifiers,-Predicate).
% does not add missing y' = y equalities for unchanged variables
before_after_predicate(Action,Predicate) :-
    before_after_predicate_list_conjunct([Action],Predicate).
before_after_predicate_list_conjunct(Actions,BAPredicate) :-
    maplist(before_after_predicate_no_equalities,Actions,AllPredicates),
    conjunct_predicates(AllPredicates,BAPredicate).
before_after_predicate_list_disjunct(Actions,BAPredicate) :-
    maplist(before_after_predicate_no_equalities,Actions,AllPredicates),
    disjunct_predicates(AllPredicates,BAPredicate).
% these add missing y' = y equalities for unchanged variables
before_after_predicate_with_equalities(Action,AllIdentifiers,Predicate) :-
    before_after_predicate_list_conjunct_with_equalities([Action],AllIdentifiers,Predicate).
before_after_predicate_list_conjunct_with_equalities(Actions,AllIdentifiers,BAPredicate) :-
    maplist(before_after_predicate_no_equalities,Actions,AllPredicates),
    maplist(add_missing_equalities(AllIdentifiers),AllPredicates,BAPredicates),
    disjunct_predicates(BAPredicates,BAPredicate).
before_after_predicate_list_disjunct_with_equalities(Actions,AllIdentifiers,BAPredicate) :-
    maplist(before_after_predicate_no_equalities,Actions,AllPredicates),
    maplist(add_missing_equalities(AllIdentifiers),AllPredicates,BAPredicates),
    disjunct_predicates(BAPredicates,BAPredicate).

% does not add missing y' = y equalities for unchanged variables
before_after_predicate_no_equalities(Action,Predicate) :-
    get_texpr_expr(Action,Expr),
    before_after_predicate_no_equalities_aux(Expr,Predicate).
before_after_predicate_no_equalities_aux(assign_single_id(Id,Pred),Predicate) :- !,
    before_after_predicate_no_equalities_aux(assign([Id],[Pred]),Predicate).
before_after_predicate_no_equalities_aux(assign(Ids,Preds),Predicate) :- !,
    maplist(ba_predicate_equal,Ids,Preds,Predicates),
    conjunct_predicates(Predicates,Predicate).
before_after_predicate_no_equalities_aux(becomes_element_of(Ids,Set),Predicate) :- !,
    maplist(ba_predicate_member(Set),Ids,Predicates),
    conjunct_predicates(Predicates,Predicate). % should we generate Set /= {} => Predicate
before_after_predicate_no_equalities_aux(becomes_such(Ids,Pred),Predicate) :- !,
    ba_predicate_becomes_such(Ids,Pred,Predicate).
before_after_predicate_no_equalities_aux(choice(Choices),Predicate) :- !,
    maplist(before_after_predicate_no_equalities,Choices,BAChoices),
    disjunct_predicates(BAChoices,Predicate).
before_after_predicate_no_equalities_aux(precondition(Pre,Body),Predicate) :- !,
    before_after_predicate_no_equalities(Body,BABody),
    conjunct_predicates([Pre,BABody],Predicate). % should probably be an implication
before_after_predicate_no_equalities_aux(select_when(Pre,Body),Predicate) :- !,
    before_after_predicate_no_equalities(Body,BABody),
    conjunct_predicates([Pre,BABody],Predicate).
before_after_predicate_no_equalities_aux(select(Whens),Predicate) :- !,
    maplist(before_after_predicate_no_equalities,Whens,WhensPredicates),
    disjunct_predicates(WhensPredicates,Predicate).
before_after_predicate_no_equalities_aux(select(Whens,Else),Predicate) :- !,
    maplist(before_after_predicate_no_equalities,Whens,WhensPredicates),
    maplist(get_select_conditions,Whens,SelectConditions),
    disjunct_predicates(SelectConditions,ElseTemp),
    create_negation(ElseTemp,ElseLHS),
    before_after_predicate_no_equalities(Else,ElseRHS),
    conjunct_predicates([ElseLHS,ElseRHS],ElsePred),
    disjunct_predicates([ElsePred|WhensPredicates],Predicate).
% TODO: case, sequence, while
before_after_predicate_no_equalities_aux(if(IfList),Predicate) :- !,
    ba_predicate_if(IfList,Predicate).
before_after_predicate_no_equalities_aux(let(P,D,B),Predicate) :- !,
    before_after_predicate_no_equalities_aux(any(P,D,B),Predicate).
before_after_predicate_no_equalities_aux(any(Parameters,PreCond,Body),Predicate) :- !,
    before_after_predicate_no_equalities(Body,BABody),
    conjunct_predicates([PreCond,BABody],Inner),
    create_exists(Parameters,Inner,Predicate).
before_after_predicate_no_equalities_aux(parallel(P),Predicate) :- !,
    maplist(before_after_predicate_no_equalities,P,Preds),
    conjunct_predicates(Preds,Predicate).
before_after_predicate_no_equalities_aux(skip,Predicate) :- !,
    Predicate = b(truth,pred,[]).
before_after_predicate_no_equalities_aux(var(Parameters,Body),Predicate) :- !,
    before_after_predicate_no_equalities(Body,BABody),
    create_exists(Parameters,BABody,Predicate).
before_after_predicate_no_equalities_aux(rlevent(_Name,_Section,_Stat,_Parameters,Guards,_Thm,Actions,_VWit,_PWit,_Unmod,_AbsEvents),Predicate) :- !,
    maplist(before_after_predicate_no_equalities,Actions,BAActions),
    conjunct_predicates([Guards|BAActions],Predicate). % should probably be an implication? see precondition.
before_after_predicate_no_equalities_aux(Expr,Predicate) :- !,
    add_error_fail(before_after_predicate,'before-after-predicate implementation missing: ',[Expr,Predicate]).

get_select_conditions(b(select_when(Condition,_),_,_),Condition).

ba_predicate_equal(b(function(FUN,ARG),Type,_Info),RHS,TPred) :- % FUN(ARG) := RHS
    prime_variable(FUN,FUNPrimeId),
    % TO DO: also treat nested function updates f(g)(x) := y where FUN is not an identifier
    create_couple(ARG,RHS,Update), % ARG|->RHS
    safe_create_texpr(set_extension([Update]),Type,UpdateSET), % {ARG|->RHS}
    construct_override(FUN,UpdateSET,Override), % FUN <+ {ARG |-> RHS}
    safe_create_texpr(equal(FUNPrimeId,Override),pred,TPred). % FUN' = FUN <+ {ARG |-> RHS}
ba_predicate_equal(Id,RHS,TPred) :-
    % prime identifiers on the lhs
    prime_variable(Id,PrimedId),
    safe_create_texpr(equal(PrimedId,RHS),pred,TPred). % Id' = RHS

construct_override(FunID,Update,Override) :-
    get_texpr_type(Update,Type),
    Override = b(overwrite(FunID,Update),Type,[]).

ba_predicate_member(Set,Id,TPred) :-
    % prime identifiers on the lhs
    prime_variable(Id,PrimedId),
    safe_create_texpr(member(PrimedId,Set),pred,TPred).

ba_predicate_becomes_such(Ids,Pred,Predicate) :-
    list_renamings_to_primed_vars(Ids,RenamingList),
    rename_bt(Pred,RenamingList,Predicate).

list_renamings_to_primed_vars([],[]).
list_renamings_to_primed_vars([Var|Vars],[rename(VarId,PrimedId)|T]) :-
    get_texpr_id(Var,VarId),
    atom_concat(VarId,'\'',PrimedId),
    list_renamings_to_primed_vars(Vars,T).

add_missing_equalities(AllVariables,PredWithoutEqualities,PredWithEqualities) :-
    find_typed_identifier_uses(PredWithoutEqualities,UsedIdentifiers),
    exclude(primed_var_is_used(UsedIdentifiers),AllVariables,VariablesNotOccuringPrimed),
    maplist(ba_predicate_equal,VariablesNotOccuringPrimed,VariablesNotOccuringPrimed,EqualityPredicates),
    conjunct_predicates([PredWithoutEqualities|EqualityPredicates],PredWithEqualities).

primed_var_is_used(UsedIdentifiers,Identifier) :-
    prime_variable(Identifier,PrimedIdentifier),
    remove_all_infos(PrimedIdentifier,PrimedIdentifierWithoutInfos),
    member(PrimedIdentifierWithoutInfos,UsedIdentifiers).

ba_predicate_if([IF|IFs],BA) :-
    get_texpr_expr(IF,if_elsif(Test,Body)),
    % if-part
    before_after_predicate_no_equalities(Body,InnerBAPredIf),
    safe_create_texpr(implication(Test,InnerBAPredIf),pred,BAPredIf),
    % if there are further else/elseif parts, create pos
    (ba_predicate_if(IFs,ElsePred)
    -> create_negation(Test,NegTest),
       safe_create_texpr(implication(NegTest,ElsePred),pred,BAPredElse),
       disjunct_predicates([BAPredIf,BAPredElse],BA)
    ;  BA = BAPredIf).
