% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(bsyntaxtree,
          [is_texpr/1, % checks if the given argument is a typed expression

           get_texpr_expr/2, % get the expression part of a typed expression
           get_texpr_type/2, % get the type of a typed expression
           get_texpr_info/2, % get the list of information of a typed expression
           get_texpr_id/2, % get the id of a typed identifier, fails if it's not an identifier
           def_get_texpr_id/2, % same as above, but raises error if arg1 is not an identifier
           create_typed_id/3, % create a typed identifier
           get_texpr_id_names/2, % translate a typed id list into a plain list of id names
           same_id/3, % true if two typed identifiers have the same id
           get_texpr_exprs/2, % list variant of the above
           get_texpr_types/2, % list variant of the above
           get_texpr_infos/2, % list variant of the above
           get_texpr_ids/2,   % list variant of the above
           get_texpr_pos/2,   % get the position of a typed expression
           extract_pos_infos/2, % get the position info sub list of an info list
           get_info_pos/2,    % get the position directly from info field

           same_texpr/2, % check if two typed expressions are equal modulo Info fields
           different_texpr_values/2, % check if two type expressions are WD and definitely denote different values

           create_texpr/4, % creates a typed expression by giving expression, type and infos
           add_texpr_infos/3, % adds some more information to a typed expression
           safe_create_texpr/3, % a version of create_texpr which extracts wd-info from sub-expressionssafe_create_texpr
           safe_create_texpr/4,
           sub_expression_contains_wd_condition/1, % utility to check if sub-expression contains wd condition

           get_rodin_name/2, get_rodin_model_name/2, get_texpr_label/2,
           get_texpr_output_pragma/3,
           always_well_defined/1, always_well_defined_or_disprover_mode/1,
           is_truth/1, is_falsity/1,
           conjunct_predicates/2,
           is_a_conjunct/3, is_a_disjunct/3, is_an_implication/3, is_an_equivalence/3, is_a_negation/2,
           conjunction_to_list/2, member_in_conjunction/2, select_member_in_conjunction/3,
           disjunct_predicates/2,
           disjunction_to_list/2,

           predicate_components/2, % split a predicate into components which use distinct identifiers
           predicate_components_in_scope/3, % ditto with an optional list of local variables
           predicate_components_with_restriction/4,
           predicate_identifiers/2, predicate_identifiers_in_scope/3,

           project_predicate_on_identifiers/5,

           find_identifier_uses/3, find_identifier_uses_l/3,
           find_typed_identifier_uses/3, find_typed_identifier_uses/2,
           find_typed_identifier_uses_l/3,
           get_global_identifiers/1,
           occurs_in_expr/2, some_id_occurs_in_expr/2,
           create_exists/3, create_exists_opt/3, create_exists_opt/4,
           create_forall/3,
           create_negation/2, is_negation_of/2,
           create_implication/3,
           create_couple/3, create_couple/2,

           replace_id_by_expr/4,
           replace_ids_by_exprs/4,
           remove_used_id_from_info/3, % remove an id from used_id info field if it exists

           rename_bt/3,
           rename_bt_l/3,
           remove_bt/4,
           syntaxtransformation/5,

           map_over_bexpr/2, map_over_typed_bexpr/2, map_over_typed_bexpr/3,
           reduce_over_bexpr/4,
           transform_bexpr/3, transform_bexpr_with_acc/5,
           uses_implementable_integers/1,
           min_max_integer_value_used/3, min_max_integer_value_used/5,
           syntaxtraversion/6,
           safe_syntaxelement/5,

           remove_all_infos/2, extract_info/2, extract_info/3,
           bsyntax_pattern/2,

           remove_all_infos_and_ground/2,

           check_if_typed_predicate/1,
           check_if_typed_expression/1,
           check_if_typed_substitution/1,

           strip_and_norm_ast/2,
           is_set_type/2, get_set_type/2, get_texpr_set_type/2,
           is_just_type/1, is_just_type/2,

           create_recursive_compset/6,
           unique_typed_id/3,
           mark_bexpr_as_symbolic/2,

           identifier_sub_ast/3, exchange_ast_position/5,

           has_declared_identifier/2, add_declaration_for_identifier/3,
           check_ast/1, check_ast/2
          ]).

:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module provides operations on the type-checked AST.').

:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(library(avl)).
:- use_module(library(terms)).

:- use_module(self_check).
:- use_module(error_manager).
:- use_module(translate).

:- use_module(typing_tools,[is_infinite_type/1]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% basic access to enriched syntax tree

is_texpr(b(_,_,_)).

get_texpr_expr(b(Name,_,_),R) :- !,R=Name.
get_texpr_expr(E,_) :- add_error_fail(get_texpr_expr,'B Expression not properly wrapped: ',E).

get_texpr_type(b(_,Type,_),R) :- !, R=Type.
get_texpr_type(E,_) :- add_error_fail(get_texpr_type,'B Expression not properly wrapped: ',E).
get_texpr_info(b(_,_,Info),R) :- !, R=Info.
get_texpr_info(E,_) :- add_error_fail(get_texpr_info,'B Expression not properly wrapped: ',E).

get_texpr_id(b(N,_,_),Id) :- !, N=identifier(Id).
get_texpr_id(E,_) :- add_error_fail(get_texpr_id,'B Expression not properly wrapped: ',E).
/* same as above: but must succeed */
def_get_texpr_id(b(identifier(Id),_,_),R) :- !,R=Id.
def_get_texpr_id(ID,_) :- add_error_fail(def_get_texpr_id,'Could not extract identifier: ',ID).

create_typed_id(IDName,Type,b(identifier(IDName),Type,[])).

same_id(X,Y,ID) :- get_texpr_id(X,ID), get_texpr_id(Y,ID).


% translate a list of ids with the b/3 wrapper into a plain id list
get_texpr_id_names([],[]).
get_texpr_id_names([H|TT],[ID|T]) :- def_get_texpr_id(H,ID), get_texpr_id_names(TT,T).

% check if two wrapped expressions are equal (modulo associated Info, e.g. source loc info)
same_texpr(b(E1,Type,_),b(E2,Type,_)) :- % should we do a == check first ?? probably not as it may traverse the entire structure (see email exchange with SICStus 28.1.2015)
  same_functor(E1,E2),
  safe_syntaxelement(E1,Subs1,_Names1,_List1,Constant),
  safe_syntaxelement(E2,Subs2,_Names2,_List2,Constant2),
  Constant==Constant2, % in case we have values with variables inside !
  same_sub_expressions(Subs1,Subs2).

same_sub_expressions([],[]).
same_sub_expressions([H1|T1],[H2|T2]) :- same_texpr(H1,H2), same_sub_expressions(T1,T2).

% check if two wrapped expressions definitely denote a different value (and are well-defined)
% only detects certain cases !
different_texpr_values(b(E1,Type,_),b(E2,Type,_)) :-
 different_value(E1,E2).
different_value(boolean_false,boolean_true) :- !.
different_value(boolean_true,boolean_false) :- !.
different_value(integer(X),integer(Y)) :- !, X\=Y.
different_value(string(X),string(Y)) :- !, X\=Y.


get_texpr_exprs([],[]).
get_texpr_exprs([E|Rest],[N|NRest]) :- get_texpr_expr(E,N),get_texpr_exprs(Rest,NRest).
get_texpr_types([],[]).
get_texpr_types([E|Rest],[T|TRest]) :- get_texpr_type(E,T),get_texpr_types(Rest,TRest).
get_texpr_infos([],[]).
get_texpr_infos([E|Rest],[I|IRest]) :- get_texpr_info(E,I),get_texpr_infos(Rest,IRest).
get_texpr_ids([],[]).
get_texpr_ids([E|Rest],[I|IRest]) :- get_texpr_id(E,I),get_texpr_ids(Rest,IRest).

get_texpr_pos(TExpr,Pos) :-
    get_texpr_info(TExpr,Infos),
    ( member(nodeid(Pos1),Infos) -> !,Pos=Pos1
    ; Pos = none).

% similar to get_texpr_pos, but returns sub-list of infos relating to position
extract_pos_infos(Infos,InfoRes) :-
    ( member(nodeid(Pos1),Infos) -> !,InfoRes=[nodeid(Pos1)]
    ; InfoRes = []).

get_info_pos(Infos,Pos) :- (member(nodeid(Pos1),Infos) -> Pos=Pos1 ; Pos=none).

create_texpr(Expr,Type,Info,b(Expr,Type,Info)).

add_texpr_infos(b(Expr,Type,Old),Infos,b(Expr,Type,New)) :-
    append(Infos,Old,New).


% try and extract the Rodin name of a predicate or expression
get_rodin_name(Expression,Name) :- get_texpr_pos(Expression,rodinpos(Name,_)), Name \= [].
get_rodin_name(Expression,Name) :- get_texpr_pos(Expression,rodinpos(_Model,Name,_)), Name \= []. % new rodinpos

get_rodin_model_name(Expression,Model) :-
   get_texpr_pos(Expression,rodinpos(Model,Name,_)), Name \= []. % new rodinpos

% try and extract the Rodin name or label pragma of a predicate or expression
get_texpr_label(TExpr,Label) :-
    get_texpr_info(TExpr,Infos),
    member(I,Infos), info_label(I,Label).
info_label(nodeid(rodinpos(Name,_)),[Name]) :- Name \= [].
info_label(nodeid(rodinpos(Model,Name,_)),[FullName]) :- Name \= [], % new rodinpos
    ajoin([Model,':',Name], FullName). % TO DO: if only one level; don't do this
info_label(label(LabelList),LabelList).

get_texpr_output_pragma(TExpr,[Pragma],[Content]) :-
    get_texpr_info(TExpr,Infos),
    member(output_pragma(Pragma,Content),Infos).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- assert_must_succeed( (E = "1/2",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "3 mod 2",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "1/0",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, \+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "2*(3 mod card({}))",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, \+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "1/(2+1-3)",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, \+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "max({1,2,3,0})",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "min({1,2,3,0})",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "min({1,2,3,0}-(0..4))",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, \+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "2+max({1,2,3, 1/0})",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none,\+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "2-card({1,2,3})",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "2-card({1,2,3, 1/0})",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none,\+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "bool(2-size([1,2,3, 1/0])=3)",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==boolean, Err==none, \+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "bool(2-size([1,2,3, 1/2, 3 mod 2, 3**2])=size([]))",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==boolean, Err==none, always_well_defined(ET) ) ).

always_well_defined_or_disprover_mode(BE) :-
  (preferences:preference(disprover_mode,true) -> true % we can assume all calls are well-defined
    ; always_well_defined(BE)).

always_well_defined(b(E,_,Infos)) :- !,
  (always_wd(E) -> true  % some special rules
   ; nonmember(contains_wd_condition,Infos)).
always_well_defined(E) :- add_error_fail(always_well_defined,'Illegal call: ',always_well_defined(E)).


:- assert_must_succeed( always_wd(div( b(integer(2),integer,[]), b(integer(2),integer,[]) )) ).
:- assert_must_fail( always_wd(div( b(integer(2),integer,[]), b(integer(0),integer,[]) )) ).
:- assert_must_succeed( always_wd(mod( b(integer(2),integer,[]), b(integer(2),integer,[]) )) ).
:- assert_must_fail( always_wd(mod( b(integer(2),integer,[]), b(integer(0),integer,[]) )) ).
:- assert_must_fail( always_wd(mod( b(integer(2),integer,[]), b(integer(-2),integer,[]) )) ).
:- assert_must_fail( always_wd(mod( b(integer(-1),integer,[]), b(integer(2),integer,[]) )) ).
:- assert_must_succeed( always_wd(power_of( b(integer(2),integer,[]), b(integer(3),integer,[]) )) ).
:- assert_must_succeed( always_wd(power_of( b(integer(2),integer,[]), b(integer(0),integer,[]) )) ).
:- assert_must_fail( always_wd(power_of( b(integer(2),integer,[]), b(integer(-1),integer,[]) )) ).

% catch cases where the construct is currently so instantiated that we can determine that it is well-defined
% can happen e.g. during closure compilation
:- use_module(custom_explicit_sets,[check_in_domain_of_avlset/2]).
always_wd(power_of(X,Y)) :- integer_value(Y,Val), number(Val),Val>=0, always_well_defined(X).
always_wd(div(X,Y)) :-  integer_value(Y,Val), number(Val),Val\=0, always_well_defined(X).
always_wd(mod(X,Y)) :- integer_value(Y,Val), number(Val),Val>0,
   integer_value(X,ValX), number(ValX), ValX>=0. % TO DO: maybe treat this differently in TLA / Event-B minor mode ?
always_wd(function(X,Y)) :- nonvar(X), nonvar(Y),
   X= b(value(AVLSET),_,_), nonvar(AVLSET), AVLSET=avl_set(AVL),
   Y= b(value(Val),_,_),
   ground(Val), check_in_domain_of_avlset(Val,AVL). % ,print(ok),nl.
always_wd(min(X)) :- non_empty_set_value(X).
always_wd(max(X)) :- non_empty_set_value(X).
always_wd(card(X)) :- finite_set_value(X).
always_wd(size(X)) :- finite_seq_value(X).
always_wd(general_intersection(X)) :- non_empty_set_value(X).
always_wd(value(_)). % we have already computed the value and will raise WD error; to be 100 % safe we could restrict this to ground values
% other candidates: size(_), first(_) last(_) tail(_) front(_) restrict_front(_,_) restrict_tail(_,_) rel_iterate(_,_)
always_wd(typeset).

integer_value(b(E,integer,_),Val) :- integer_value2(E,Val).
integer_value2(integer(X),R) :- !, R=X.
integer_value2(value(int(X)),R) :- !, R=X.
%integer_value2(X,_) :- print(not_integer_value2(X)),nl,fail.

non_empty_set_value(b(E,_,_)) :- non_empty_set2(E).
non_empty_set2(value(X)) :- definitely_not_empty(X).
non_empty_set2(set_extension(S)) :- l_always_well_defined(S). % the set_extension could contain wd_errors !!
non_empty_set2(sequence_extension(S)) :- l_always_well_defined(S). % ditto
definitely_not_empty(X) :- var(X),!,fail.
definitely_not_empty(avl_set(_)).
definitely_not_empty([_|_]).

finite_set_value(b(E,_,_)) :- finite_set2(E).
finite_set2(empty_set).
finite_set2(empty_sequence).
finite_set2(value(X)) :- X==[] -> true ; definitely_not_empty(X).
finite_set2(set_extension(S)) :- l_always_well_defined(S). % the set_extension could contain wd_errors !!
finite_set2(sequence_extension(S)) :- l_always_well_defined(S). % ditto

finite_seq_value(b(E,_,_)) :- finite_seq2(E).
finite_seq2(empty_set).
finite_seq2(empty_sequence).
finite_seq2(value(X)) :- X==[]. % it could be expensive to check if non empty value is a sequence ??
finite_seq2(sequence_extension(S)) :- l_always_well_defined(S).

l_always_well_defined([]).
l_always_well_defined([H|T]) :- always_well_defined(H), l_always_well_defined(T).

is_truth(b(truth,pred,_)).
is_falsity(b(falsity,pred,_)).

% conjunction of a list of predicates
% NOTE: bsyntaxtree:conjunct_predicates([P1,P2,P3],R).  -->  generates R = b(conjunct(b(conjunct(P1,P2),pred,[]),P3),pred,[])
conjunct_predicates(V,R) :- var(V),!, add_internal_error('Variable conjunction list: ',conjunct_predicates(V,R)),fail.
conjunct_predicates([],b(truth,pred,[])).
conjunct_predicates([P|Rest],Result) :- conjunct2(Rest,P,Result).
conjunct2([],P,P).
conjunct2([Q|Rest],P,Result) :- conjunct3(P,Q,R), conjunct2(Rest,R,Result).
conjunct3(b(truth,_,_),P,P) :- !.
conjunct3(P,b(truth,_,_),P) :- !.
conjunct3(A,B,b(conjunct(A,B),pred,NewInfo)) :- extract_info(A,B,NewInfo).

% disjunction of a list of predicates
disjunct_predicates([],b(falsity,pred,[])).
disjunct_predicates([P|Rest],Result) :- disjunct2(Rest,P,Result).
disjunct2([],P,P).
disjunct2([Q|Rest],P,Result) :- disjunct3(P,Q,R), disjunct2(Rest,R,Result).
disjunct3(b(falsity,_,_),P,R) :- !, R=P.
disjunct3(b(truth,T,I),_,R) :- !, R=b(truth,T,I).
disjunct3(P,b(falsity,_,_),R) :- !, R=P.
disjunct3(_,b(truth,T,I),R) :- !, R=b(truth,T,I).
disjunct3(A,B,b(disjunct(A,B),pred,NewInfo)) :- extract_info(A,B,NewInfo).

% a version of create_texpr which collects automatically important infos from sub-expressions
safe_create_texpr(Expr,Type,b(Expr,Type,Info)) :- %
   (sub_expression_contains_wd_condition(Expr) -> Info = [contains_wd_condition] ; Info=[]).

safe_create_texpr(Expr,Type,Info,b(Expr,Type,FullInfo)) :- %
   (sub_expression_contains_wd_condition(Expr) -> FullInfo = [contains_wd_condition|Info] ; FullInfo=Info).

sub_expression_contains_wd_condition(Expr) :-
   safe_syntaxelement(Expr,Subs,_Names,_L,_C),
   member(b(_,_,Infos),Subs),
   (var(Infos)
    -> add_internal_error('Expression not sufficiently instantiated: ',sub_expression_contains_wd_condition(Expr)),
       fail
    ; memberchk(contains_wd_condition,Infos)).

% extract important info from one sub-expression:
extract_info(b(_,_,Info1),NewInfo) :-
   extract_info_aux(Info1,[],I1),!, NewInfo = I1.
extract_info(A,R) :- add_internal_error('Could not extract info: ',extract_info(A,R)),
 R=[].
% extract imortant info fields from two (sub-)expressions
extract_info(b(_,_,Info1),b(_,_,Info2),NewInfo) :-
   extract_info_aux(Info1,[],I1),
   extract_info_aux(Info2,I1,I2),!, NewInfo = I2.
%:- use_module(library(ordsets),[ord_intersection/3]).
%   ord_intersection(Info1,Info2,NewInfo),!.
extract_info(A,B,R) :- add_internal_error('Could not extract info: ',extract_info(A,B,R)),
 R=[].

extract_info_aux([],Acc,Acc).
extract_info_aux([H|T],Acc,Res) :-
  ((important_info(H), \+ member(H,Acc)) -> extract_info_aux(T,[H|Acc],Res)
    ; extract_info_aux(T,Acc,Res)).
important_info(contains_wd_condition).
important_info(prob_annotation(_)).

is_a_conjunct(b(conjunct(A,B),pred,_),A,B).

conjunction_to_list(C,L) :- var(C),!,
   add_error_fail(conjunction_to_list,'Internal error: var :',conjunction_to_list(C,L)).
conjunction_to_list(C,List) :-
    conjunction_to_list2(C,List,[]).
conjunction_to_list2(X,I,O) :- X=b(E,_,_),
   (E=conjunct(LHS,RHS) -> conjunction_to_list2(LHS,I,Inter),
                           conjunction_to_list2(RHS,Inter,O)
    ; E = truth -> I=O
    ; I = [X|O]).

member_in_conjunction(M,Conj) :- is_a_conjunct(Conj,LHS,RHS),!,
   (member_in_conjunction(M,LHS) ; member_in_conjunction(M,RHS)).
member_in_conjunction(M,M).

% not terribly efficient way to select and remove a conjunct from a predicate
select_member_in_conjunction(M,Conj,Rest) :- is_a_conjunct(Conj,LHS,RHS),!,
   (select_member_in_conjunction(M,LHS,RL), conjunct_predicates([RL,RHS],Rest)
    ; select_member_in_conjunction(M,RHS,RR), conjunct_predicates([LHS,RR],Rest)).
select_member_in_conjunction(M,M,b(truth,pred,[])).

is_a_disjunct(b(disjunct(A,B),pred,_),A,B).
is_a_negation(b(negation(A),pred,_),A).

disjunction_to_list(C,L) :- var(C),!,
   add_error_fail(disjunction_to_list,'Internal error: var :',disjunction_to_list(C,L)).
disjunction_to_list(C,List) :-
    disjunction_to_list2(C,List,[]).
disjunction_to_list2(C,I,O) :- is_a_disjunct(C,LHS,RHS),!,
   disjunction_to_list2(LHS,I,Inter),
   disjunction_to_list2(RHS,Inter,O).
disjunction_to_list2(X,[X|R],R).

is_an_implication(b(implication(A,B),pred,_),A,B).

is_an_equivalence(b(equivalence(A,B),pred,_),A,B).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% remove all info fields
remove_all_infos(TExpr,TNExpr) :-
    var(TExpr),!,TExpr=TNExpr.
remove_all_infos(TExpr,TNExpr) :-
    get_texpr_expr(TExpr,Expr),
    get_texpr_type(TExpr,Type),
    syntaxtransformation(Expr,Subs,_,NSubs,NExpr),
    create_texpr(NExpr,Type,_,TNExpr),
    maplist(remove_all_infos,Subs,NSubs).

% replace all info fields
remove_all_infos_and_ground(TExpr,TNExpr) :-
    var(TExpr),!,TExpr=TNExpr.
remove_all_infos_and_ground(TExpr,TNExpr) :-
    get_texpr_expr(TExpr,Expr),
    get_texpr_type(TExpr,Type),
    syntaxtransformation(Expr,Subs,_,NSubs,NExpr),
    create_texpr(NExpr,Type,[],TNExpr),
    maplist(remove_all_infos_and_ground,Subs,NSubs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% create_exists_opt(TIds,Preds,NewPred)
%   creating an existential quantification with some optimisations.
% TIds: The typed identifiers that are quantified
% Preds: A list of predicates
% NewPred: The (optimised) quantified expression
% Basically two optimisations are performed:
%  - identifiers that are not used at all are removed from the quantifier
%  - predicates that do not use one of the quantified identifiers are moved
%    outside the quantification, resulting in a predicate of the form "P & #x.Q(x)"

create_exists_opt(TIds,Preds,NewPred) :-
    create_exists_opt(TIds,Preds,NewPred,_).
create_exists_opt([],Preds,NewPred,Modified) :- !, Modified = false,
    conjunct_predicates(Preds,NewPred).
create_exists_opt(TIds,Preds,NewPred,Modified) :-
    get_texpr_ids(TIds,UnsortedIds), sort(UnsortedIds,Ids),
    create_exists_opt2(Preds,Ids,[],Inner,Outer,UsedIds),
    ( Inner = []    -> create_texpr(truth,pred,[],Exists)
    ; % Inner = [_|_] ->
        ceo_filter_used_ids(TIds,UsedIds,UsedTIds), % filter out unused variables
        conjunct_predicates(Inner,I),
        create_exists(UsedTIds,I,Exists)),
    append(Outer,[Exists],All),
    (Outer=[_|_] -> Modified=true % TO DO: check if TIds=UsedTIds
      ; UsedIds == Ids -> Modified= false
      ; Modified=true),
    conjunct_predicates(All,NewPred).
create_exists_opt2([],_Ids,UsedIds,[],[],UsedIds) :- !.
create_exists_opt2([Pred|Prest],Ids,UsedIdsIn,Inner,Outer,UsedIdsOut) :- !,
    find_identifier_uses(Pred, [], LocalUses),
    ord_intersection(Ids,LocalUses,LocalExUses),
    ( LocalExUses = [] ->
        Inner = Irest,
        Outer = [Pred|Orest], % move outside of existential quantifier
        UsedIdsIn = Urest
    ; LocalExUses = [_|_] ->
        Inner = [Pred|Irest],
        Outer = Orest,
        ord_union(UsedIdsIn,LocalExUses,Urest)
    ),
    create_exists_opt2(Prest,Ids,Urest,Irest,Orest,UsedIdsOut).
create_exists_opt2(Pred,Ids,UsedIdsIn,Inner,Outer,UsedIdsOut) :-
   add_error(create_exists_opt,'Expexting predicate list: ',Pred),
   create_exists_opt2([Pred],Ids,UsedIdsIn,Inner,Outer,UsedIdsOut).
ceo_filter_used_ids([],_UsedIds,[]).
ceo_filter_used_ids([TId|Irest],UsedIds,Filtered) :-
    get_texpr_id(TId,Id),
    ( ord_member(Id,UsedIds) -> Filtered = [TId|Frest] % id used: keep it
    ; otherwise              -> Filtered = Frest),
    ceo_filter_used_ids(Irest,UsedIds,Frest).

create_exists([],Pred,NewPred) :- !,Pred=NewPred.
create_exists(Ids,Pred,NewPred) :-
    find_identifier_uses(Pred, [], Used),
    extract_info(Pred,NewImportantInfo),
    create_texpr(exists(Ids,Pred),pred,[used_ids(Used)|NewImportantInfo],NewPred).
create_forall([],Pred,NewPred) :- !,Pred=NewPred.
create_forall(Ids,Pred,NewPred) :-
    find_identifier_uses(Pred, [], Used),
    split_forall_body(Pred,LHS,RHS),
    extract_info(LHS,RHS,NewImportantInfo),
    create_texpr(forall(Ids,LHS,RHS),pred,[used_ids(Used)|NewImportantInfo],NewPred).
split_forall_body(b(implication(LHS,RHS),_,_),LHS,RHS) :- !.
split_forall_body(RHS,b(truth,pred,[]),RHS).

create_implication(b(truth,pred,_),P,Res) :- !, Res=P.
create_implication(b(falsity,pred,_),P,Res) :- !, create_negation(P,Res).
create_implication(Lhs,Rhs,b(implication(Lhs,Rhs),pred,NewInfo)) :-
   extract_info(Lhs,Rhs,NewInfo).


create_negation(b(truth,pred,I),R) :- !, R=b(falsity,pred,I).
create_negation(b(falsity,pred,I),R) :- !, R=b(truth,pred,I).
create_negation(b(negation(Pred),pred,_),R) :- !, R=Pred. % not(not(P)) = P
% we could add some rules about negating member <-> not_member, ...
create_negation(Pred,b(negation(Pred),pred,NewInfo)) :- extract_info(Pred,NewInfo).

% check if something is the negation of something else (quite stupid at the moment)
is_negation_of(P,NP) :-
   create_negation(P,NotP),
   same_texpr(NotP,NP).
is_negation_of(b(P,pred,_),b(NP,pred,_)) :- is_negation_aux(P,NP).

is_negation_aux(equal(A,B),NP) :-
   is_negation_of_equality(NP,A,B).
is_negation_aux(subset(XA,SA),not_subset(XB,SB)) :- same_texpr(XA,XB), same_texpr(SA,SB).
is_negation_aux(not_subset(XA,SA),subset(XB,SB)) :- same_texpr(XA,XB), same_texpr(SA,SB).
is_negation_aux(subset_strict(XA,SA),not_subset_strict(XB,SB)) :- same_texpr(XA,XB), same_texpr(SA,SB).
is_negation_aux(not_subset_strict(XA,SA),subset_strict(XB,SB)) :- same_texpr(XA,XB), same_texpr(SA,SB).
is_negation_aux(member(XA,SA),not_member(XB,SB)) :- same_texpr(XA,XB), same_texpr(SA,SB).
is_negation_aux(not_member(XA,SA),member(XB,SB)) :- same_texpr(XA,XB), same_texpr(SA,SB).
is_negation_aux(less(A,B),greater_equal(AA,BB)) :- same_texpr(A,AA), same_texpr(B,BB).
is_negation_aux(less(A,B),less_equal(BB,AA)) :- same_texpr(A,AA), same_texpr(B,BB).
is_negation_aux(less_equal(B,A),less(AA,BB)) :- same_texpr(A,AA), same_texpr(B,BB).
is_negation_aux(less_equal(A,B),greater(AA,BB)) :- same_texpr(A,AA), same_texpr(B,BB).
is_negation_aux(greater_equal(A,B),NP) :- is_negation_aux(less_equal(B,A),NP).
is_negation_aux(greater(A,B),NP) :- is_negation_aux(less(B,A),NP).
% TO DO: detect more ?? x < y   <=>  not ( x > y-1 )

is_negation_of_equality(not_equal(AA,BB),A,B) :- same_texpr(A,AA), same_texpr(B,BB).
is_negation_of_equality(equal(AA,BB),A,B) :- same_texpr(A,AA), neg_value(BB,B).

neg_value(b(boolean_true,_,_),b(boolean_false,_,_)).
neg_value(b(boolean_false,_,_),b(boolean_true,_,_)).


create_couple(A,B,b(couple(A,B),couple(TA,TB),Infos)) :-
      get_texpr_type(A,TA), get_texpr_type(B,TB),
      extract_info(A,B,Infos).

create_couple([A],Couple) :- !,A=Couple.
create_couple([A,B|Rest],Couple) :-
    create_couple(A,B,Couple1),
    create_couple([Couple1|Rest],Couple).


% --------------------------------------


% find uses of identifiers

occurs_in_expr(ID,TExpr) :- var(ID),!,
   add_internal_error('Variable id: ',occurs_in_expr(ID,TExpr)),fail.
occurs_in_expr(ID,TExpr) :- ID=b(_,_,_),!, add_internal_error('Non-atomic identifier: ',occurs_in_expr(ID,TExpr)),fail.
occurs_in_expr(ID,TExpr) :-
 (find_identifier_uses(TExpr,[],Used) -> member(ID,Used)
   ; add_failed_call_error(find_identifier_uses(TExpr,[],_)),fail).
  % TO DO: optimize so that we only look for ID; we don't have to keep track of other IDs

some_id_occurs_in_expr([H|T],TExpr) :- (var(H) ; \+ atomic(H)),
    add_internal_error('Must be (sorted) atomic identifiers: ',some_id_occurs_in_expr([H|T],TExpr)),
    fail.
some_id_occurs_in_expr(SortedIDs,TExpr) :-
    find_identifier_uses(TExpr,[],Used),
    list_to_ord_set(Used,UsedSorted),
    ord_intersect(UsedSorted,SortedIDs).

:- use_module(tools,[safe_sort/3]).
find_identifier_uses(TExpr,Ignored,Ids) :-
    check_is_texpr(TExpr,find_identifier_uses),
    find_typed_identifier_uses(TExpr,Ignored,TIds),
    get_texpr_ids(TIds,Ids).
find_identifier_uses_l(TExprs,Ignored,Ids) :-
    find_typed_identifier_uses_l(TExprs,Ignored,TIds),
    get_texpr_ids(TIds,Ids).

check_is_texpr(X,Context) :-
   (get_texpr_expr(X,_) -> true ; add_internal_error('Expected TExpr: ', check_is_texpr(X,Context))).

find_typed_identifier_uses(TExpr,Ids) :-
    get_global_identifiers(Ignored), % ignore all global sets and global constants; hence find_typed_identifier_uses/2 usually should only be called for top level expressions
    find_typed_identifier_uses(TExpr,Ignored,Ids).

find_typed_identifier_uses(TExpr,Ignored,Ids) :- var(TExpr),!,
    add_internal_error('Variable typed expression: ', find_typed_identifier_uses(TExpr,Ignored,Ids)),
    Ids = [].
find_typed_identifier_uses(TExpr,Ignored,Ids) :-
    find_typed_identifier_uses_l([TExpr],Ignored,Ids).
find_typed_identifier_uses2(TExpr,Ignored,Ids,Rest) :-
    syntaxtraversion_fixed(TExpr,Expr,_,_,QSubs,Subs,TNames),
    ( uses_an_identifier(Expr,Id) ->
        ( memberchk(Id,Ignored) -> Ids=Rest % TO DO: use ord_member and esnure Ignored always sorted or use avl_set
        ; otherwise             ->
            get_texpr_type(TExpr,Type),
            create_texpr(identifier(Id),Type,[],TId),
            Ids = [TId|Rest])
    ; otherwise ->
        find_typed_identifier_uses2_l(QSubs,Ignored,Ids,Ids2),
        get_texpr_ids(TNames,Names),
        append(Names,Ignored,Ignored2),
        find_typed_identifier_uses2_l(Subs,Ignored2,Ids2,Rest)).
find_typed_identifier_uses2_l([],_) --> [].
find_typed_identifier_uses2_l([Expr|Rest],Ignored) -->
    find_typed_identifier_uses2(Expr,Ignored),
    find_typed_identifier_uses2_l(Rest,Ignored).


uses_an_identifier(identifier(Id),Id).
uses_an_identifier(lazy_lookup_pred(Id),Id).
uses_an_identifier(lazy_lookup_expr(Id),Id).

find_typed_identifier_uses_l(TExprs,Ignored,Ids) :-
    find_typed_identifier_uses2_l(TExprs,Ignored,Unsorted,[]),
    safe_sort(find_typed_identifier_uses,Unsorted,Ids).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% break predicate into components with disjunct identifiers
%  Pred: a predicate
%  Components: a list of terms component/2 where the first argument
%    is a predicate, the second is the set of used identifiers
predicate_components(Pred,Res) :- predicate_components_in_scope(Pred,[],Res).
predicate_components_in_scope(Pred,LocalVars,Res) :-
   predicate_components_with_restriction(Pred,LocalVars,all,Res).

predicate_components_with_restriction(Pred,LocalVars,RestrictionList,Res) :-
    conjunction_to_list(Pred,Preds),
    l_predicate_identifiers(Preds,LocalVars,PredIds),
    ( RestrictionList=all -> R=all
    ; otherwise           -> list_to_ord_set(RestrictionList,R)),
    ((preferences:get_preference(partition_predicates_inline_equalities,true),
      preferences:get_preference(try_kodkod_on_load,false), % otherwise we currently get  mapping ProB value to Kodkod failed errors
      find_and_remove_equalities(PredIds,PredIds2)) -> true ; PredIds2=PredIds),
    %%print(find_components(R)),nl,nl,
    %%(member(pred(P,Ids,X),PredIds2), nl,print(pred(Ids,X)),nl,translate:print_bexpr(P),nl,fail ; true),
    find_components(PredIds2,R,Components),
    !,
    Components=Res.
predicate_components_with_restriction(Pred,_,_,[component(Pred,[])]) :-
    add_internal_error('predicate_components failed: ',Pred).

% get the list of used identifiers for each predicate
l_predicate_identifiers([],_LocalVars,[]).
l_predicate_identifiers([Pred|PRest],LocalVars,[pred(Pred,Ids,_Selected)|IRest]) :-
    predicate_identifiers_in_scope(Pred,LocalVars,Ids),
    l_predicate_identifiers(PRest,LocalVars,IRest).

:- use_module(debug,[debug_println/2]).
% find and apply obvious equalities so that they do not interfere with partitioning into components
% example:  c = 1 &  f: 1..c --> A  &  g: 1..c --> B
% TO DO: preprocess and do one pass to detect potential equalities
find_and_remove_equalities([],[]).
find_and_remove_equalities(List,[pred(P,PIds,Sel)|FT]) :- %print(pred(P,PIds)),nl,
    select(pred(P,PIds,Sel),List,Rest),
    PIds = [Id],
    get_texpr_expr(P,equal(TId,Value)),
    get_texpr_id(TId,Id),
    %%(value_which_can_be_replaced(Value) -> true ; nl,print('Not replaced: '),translate:print_bexpr(Value),nl,fail),
    value_which_can_be_replaced(Value),
    debug_println(9,replace_simple_equality(Id,Value,PIds)),
    maplist(apply_to_pred(Id,Value),Rest,RT),
    !,
    %print(applied(AT)),nl,
    find_and_remove_equalities(RT,FT).
find_and_remove_equalities(R,R).

value_which_can_be_replaced(b(E,T,_)) :- value_which_can_be_replaced2(E,T).
   %(value_which_can_be_replaced2(E,T) -> true ; print(not_val(E)),nl,fail).
value_which_can_be_replaced2(value(_),_).
value_which_can_be_replaced2(integer(_),_).
%value_which_can_be_replaced2(identifier(I),global(G)) :- b_global_constant(G), id_not_used_anywhere(I).
value_which_can_be_replaced2(integer_set(_),_).
value_which_can_be_replaced2(unary_minus(_),_).
% should we add: add, minus, .... ? see also simple2 in b_ast_cleanup
value_which_can_be_replaced2(max_int,_).
value_which_can_be_replaced2(min_int,_).
value_which_can_be_replaced2(string(_),_).
value_which_can_be_replaced2(string_set,_).
value_which_can_be_replaced2(boolean_true,_).
value_which_can_be_replaced2(boolean_false,_).
value_which_can_be_replaced2(bool_set,_).
value_which_can_be_replaced2(empty_set,_).
value_which_can_be_replaced2(empty_sequence,_).
value_which_can_be_replaced2(couple(A,B),_) :- value_which_can_be_replaced(A), value_which_can_be_replaced(B).
value_which_can_be_replaced2(interval(A,B),_) :- value_which_can_be_replaced(A), value_which_can_be_replaced(B).
value_which_can_be_replaced2(set_extension(L),_) :- maplist(value_which_can_be_replaced,L).
value_which_can_be_replaced2(sequence_extension(L),_) :- maplist(value_which_can_be_replaced,L).
%value_which_can_be_replaced2(identifier(_),_).
% identifiers can also be replaced: we check above that the only identifier in the predicate is the equality identifier
% TO DO: enable this but then we need to fix replace_id_by_expr2 to updated the used_ids info ! + we can have scoping issues !?; see test 1358

apply_to_pred(ID,Expr,pred(E1,PIds1,X),pred(E2,PIds2,X)) :-
   (select(ID,PIds1,PIds2) -> replace_id_by_expr(E1,ID,Expr,E2)  %,print(applied(ID)),nl,translate:print_bexpr(E2),nl
   % TO DO: replace_id_by_expr does not seem to update used_ids info !! check_used_ids_info fails for test 1358 if we allo identifiers inside Expr
     ; E1=E2,PIds1=PIds2).


% project a predicate : keep only those Predicates that are directly or
%  indirectly relevant for Ids; FullIds are all identifiers used by ProjectedPredicate
project_predicate_on_identifiers(Pred,Ids,ProjectedPredicate,FullIds, RestList) :-
   (debug:debug_mode(on)
     -> print('project: '),translate:print_bexpr(Pred),nl, print('  on : '), print(Ids),nl
     ;  true),
   conjunction_to_list(Pred,Preds),
    l_predicate_identifiers(Preds,[],PredIds), % TO DO: allow LocalVariables to be passed
   % print(predids(PredIds)),nl,
   extract_all_predicates(Ids,all,Ids,PredIds,ProjectedPredicates, RestList,FullIds),
   conjunct_predicates(ProjectedPredicates,ProjectedPredicate),
   (debug:debug_mode(on)
     -> print('*result: '),translate:print_bexpr(ProjectedPredicate),nl,
        print('  on : '),    print(FullIds),nl
     ; true).

%print_preds([]).
%print_preds([pred(P,_IDs,_)|T]) :- translate:print_bexpr(P), nl, print('   '), print_preds(T).

:- use_module(b_global_sets,[b_get_global_constants/1, b_get_global_sets/1]).
predicate_identifiers(Pred,Ids) :- predicate_identifiers_in_scope(Pred,[],Ids).
predicate_identifiers_in_scope(Pred,LocalVariables,Ids) :-
    get_global_identifiers(IS),
    list_to_ord_set(LocalVariables,LV),
    ord_subtract(IS,LV,Ignore2), % Do not ignore any local variable; it will be used instead of enumerate set element
    find_identifier_uses(Pred,Ignore2,Ids1),  % Ignore enumerated set names
    list_to_ord_set(Ids1,Ids).

% get global set and constant identifiers which you usually want to exclude for find_identifier_uses
get_global_identifiers(IDs) :-
    b_global_sets:b_get_global_constants(EnumeratedSetCsts),
    % b_global_sets:b_get_global_enumerated_sets(GSets), % is there a reason to exclude deferred sets ??; cardinality inference,... are all done before partitioning ?
    b_global_sets:b_get_global_sets(GSets),
    append(GSets,EnumeratedSetCsts,GE),
    list_to_ord_set(GE,IDs).

% find_components(ListOf_pred, Restrict, Out:ListOfComponents)
% role of Restrict: all if we do normal partition or List of VariableIDs on which we restrict our attention to (for Existential quantifier construction)
find_components([],_,[]).
find_components([pred(P,PIds,true)|PRest],Restrict,[component(Pred,Ids)|CRest]) :-
    % find all predicates which are using identifiers occuring in PIds
    % (and additionally those which use common identifiers )
    ( Restrict=all -> InterIDs = PIds
    ; otherwise    -> ord_intersection(PIds,Restrict,InterIDs)),
    ( InterIDs =[] ->
        % we simply copy this predicate; not with the scope of Restricted IDs
        % print(skip(PIds,P)),nl, %
        Pred=P, Ids=PIds, PRest=Rest
    ; otherwise ->
        extract_all_predicates(PIds,Restrict,PIds,PRest,Preds,Rest,Ids),
        conjunct_predicates([P|Preds],Pred)
    ),
    find_components(Rest,Restrict,CRest).
extract_all_predicates([],_,_,Preds,Found,Rest,[]) :- !,  % selecting done at end: keep same order of conjuncts
   select_predicates(Preds,Found,Rest).
extract_all_predicates(Ids,Restrict,OldIds,Preds,Found,Rest,ResultIds) :-
    % search for all predicates that directly use one of the
    % identifiers in "Ids"
    select_all_using_preds(Preds,Ids,FoundIds),
    ord_subtract(FoundIds,OldIds,NewIds),
    ( Restrict=all -> NewIdsToExtract=NewIds
    ; otherwise    -> ord_intersection(NewIds,Restrict,NewIdsToExtract)),
    ord_union(OldIds,FoundIds,OldIds2),
    % now recursively do this which the new identifiers that we have found
    extract_all_predicates(NewIdsToExtract,Restrict,OldIds2,Preds,Found,Rest,RestIds),
    ord_union([Ids,NewIds,RestIds],ResultIds).
select_all_using_preds([],_,[]).
select_all_using_preds([pred(_P,PIds,Selected)|PRest],ComponentIds,FoundIds) :-
    ( (Selected==true ; ord_disjoint(PIds,ComponentIds)) ->
        FoundIds1 = []
    ; otherwise ->         % we select the predicate for the component
        FoundIds1 = PIds, Selected=true
    ),
    select_all_using_preds(PRest,ComponentIds,FoundIds2),
    ord_union(FoundIds1,FoundIds2,FoundIds).

select_predicates(Predicates,FoundPreds,OtherPredIds) :-
    split_list(is_selected_predicate,Predicates,FoundPredIds,OtherPredIds),
    maplist(extract_found_predicate,FoundPredIds,FoundPreds).
is_selected_predicate(pred(_P,_PIds,Selected)) :- ground(Selected).
extract_found_predicate(pred(P,_PIds,_Selected),P).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% replace multiple identifiers by expressions
% this could be used to replace replace_id_by_expr by mapping everything in a
% singleton list. however, that would involve list operations instead of unifications
replace_ids_by_exprs(TExpr,Ids,Exprs,Replaced) :-
    replace_ids_by_exprs2(Ids,Exprs,TExpr,Replaced,_).
replace_ids_by_exprs2(Ids,Exprs,TExpr,Replaced,WDC) :-
    remove_bt(TExpr,Expr,NewExpr,TNewExpr),
    ( Expr = identifier(Id), nth0(N, Ids, Id) ->
        nth0(N,Exprs,Inserted),
        Replaced = Inserted,
        get_texpr_info(Inserted,Infos),
        (memberchk(contains_wd_condition,Infos)
        -> WDC = true ; WDC = false) % WDC = true means we have added a wd-condition where previously there was none
    ; otherwise ->
        syntaxtransformation_fixed(Expr,QSubs,Subs,Names,NQSubs,NSubs,NewExpr),
        l_replace_ids_by_exprs2(QSubs,Ids,Exprs,NQSubs,WDC1),
        remove_hidden_names(Names,Ids,NonHiddenIds),
        ( NonHiddenIds = [] -> % all Ids are now hidden for the inner expressions
            NSubs=Subs, WDC=WDC1
        ; otherwise ->
            l_replace_ids_by_exprs2(Subs,NonHiddenIds,Exprs,NSubs,WDC2),
            and_wdc(WDC1,WDC2,WDC)
        ),
        TNewExpr = b(E1,T1,Info1),
        remove_used_ids_from_info(Ids,Info1,Info2), % TO DO: add Used Ids of TExpr !!
        %(E1 = exists(P,_) -> print(exists(P,Id,Info1,Info2)),nl ; true),
        add_wd_if_needed(WDC,b(E1,T1,Info2),Replaced)
    ).

l_replace_ids_by_exprs2([],_,_,[],false).
l_replace_ids_by_exprs2([H|T],Ids,Exprs,[IH|IT],WDC) :-
    replace_ids_by_exprs2(Ids,Exprs,H,IH,WDC1),
    l_replace_ids_by_exprs2(T,Ids,Exprs,IT,WDC2),
    and_wdc(WDC1,WDC2,WDC).

remove_hidden_names([],Ids,Ids).
remove_hidden_names([Name|Names],Ids,NonHiddenIds) :-
    write(remove_hidden_names([Name|Names],Ids,NonHiddenIds)).

remove_used_ids_from_info([],I,I).
remove_used_ids_from_info([Id|Ids],IIn,IOut) :-
    remove_used_id_from_info(IIn,Id,IT),
    remove_used_ids_from_info(Ids,IT,IOut).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% replace an identifier by an expression
replace_id_by_expr(TExpr,Id,Inserted,Replaced) :- \+ atomic(Id),!,
   add_internal_error('Id not atomic: ',replace_id_by_expr(TExpr,Id,Inserted,Replaced)),
   Replaced = TExpr.
replace_id_by_expr(TExpr,Id,Inserted,Replaced) :-
    %find_all_relevant_quantified_vars(Id,TExpr,QVars),
    replace_id_by_expr2(Id,Inserted,TExpr,Replaced,_).
replace_id_by_expr2(Id,Inserted,TExpr,Replaced,WDC) :-
    remove_bt(TExpr,Expr,NewExpr,TNewExpr),
    ( Expr = identifier(Id) ->
        Replaced = Inserted,
        get_texpr_info(Inserted,Infos),
        (memberchk(contains_wd_condition,Infos)
         -> WDC = true ; WDC = false) % WDC = true means we have added a wd-condition where previously there was none
    ; otherwise ->
        syntaxtransformation_fixed(Expr,QSubs,Subs,Names,NQSubs,NSubs,NewExpr),
        %maplist(replace_id_by_expr2(Id,Inserted),QSubs,NQSubs),
        l_replace_id_by_expr2(QSubs,Id,Inserted,NQSubs,WDC1),
        get_texpr_id(TId,Id),
        ( memberchk(TId,Names) -> % the Id is now hidden for the inner expressions
            NSubs=Subs, WDC=WDC1
        ; otherwise ->
            l_replace_id_by_expr2(Subs,Id,Inserted,NSubs,WDC2),
            and_wdc(WDC1,WDC2,WDC)
        ),
        TNewExpr = b(E1,T1,Info1),
        remove_used_id_from_info(Info1,Id,Info2), % TO DO: add Used Ids of TExpr !!
        %(E1 = exists(P,_) -> print(exists(P,Id,Info1,Info2)),nl ; true),
        add_wd_if_needed(WDC,b(E1,T1,Info2),Replaced)
    ).

l_replace_id_by_expr2([],_,_,[],false).
l_replace_id_by_expr2([H|T],Id,Inserted,[IH|IT],WDC) :-
   replace_id_by_expr2(Id,Inserted,H,IH,WDC1),
   l_replace_id_by_expr2(T,Id,Inserted,IT,WDC2),
   and_wdc(WDC1,WDC2,WDC).

and_wdc(true,_,R) :- !,R=true.
and_wdc(_,true,R) :- !, R=true.
and_wdc(_,_,false).

% add contains_wd_condition if a change occured during replacement of id by expression
add_wd_if_needed(true,b(E,T,Infos),Replaced) :-
   nonmember(contains_wd_condition,Infos),
   !,
   Replaced = b(E,T,[contains_wd_condition|Infos]).
add_wd_if_needed(_,T,T).

/*
find_all_relevant_quantified_vars(Id,TExpr,Vars) :-
    find_all_relevant_quantified_vars1(Id,[],TExpr,FoundQs,[]),
    sort(FoundQs,Vars).

find_all_relevant_quantified_vars1(Id,Quantified,TExpr,In,Out) :-
    syntaxtraversion_fixed(TExpr,Expr,_,_,QSubs,Subs,TNames),
    ( Expr=identifier(Id) ->
        append(Quantified,Out,In)
    ; otherwise ->
        get_texpr_ids(TNames,Ids),
        append(Ids,Quantified,NewQuantified),
        foldl(find_all_relevant_quantified_vars1(Id,Quantified),QSubs,In,I1),
        foldl(find_all_relevant_quantified_vars1(Id,NewQuantified),Subs,I1,Out)
    ).
*/

% syntaxtransformation_fixed/7 is the same as syntaxtransformation/5 with the exception
% that we distinguish between subexpressions that have the newly introduced identifiers (Names)
% in scope and those who don't. The only expressions where the latter case is relevant
% are let_expression and let_predicate.
% TODO: This is a quick'n dirty fix for only some cases.
syntaxtransformation_fixed(OExpr,OExprs,OSubs,Names,NExprs,NSubs,NExpr) :-
    is_let(OExpr,OTIds,OExprs,OExprPred,LetType),
    !,
    Names=OTIds,
    is_let(NExpr,NTIds,NExprs,NExprPred,LetType),
    same_length(OExprs,NExprs),same_length(OTIds,NTIds),
    last(OTIds,OExprPred,OSubs),last(NTIds,NExprPred,NSubs),
    !.
syntaxtransformation_fixed(Expr,[],Subs,Names,[],NSubs,NExpr) :-
    syntaxtransformation(Expr,Subs,Names,NSubs,NExpr).

is_let(let_expression(TIds,Exprs,Expr),TIds,Exprs,Expr,let_expression).
is_let(let_predicate(TIds,Exprs,Expr), TIds,Exprs,Expr,let_predicate).
is_let(lazy_let_pred(ID,SharedExpr,Expr), [ID],[SharedExpr],Expr, lazy_let_pred).
is_let(lazy_let_expr(ID,SharedExpr,Expr), [ID],[SharedExpr],Expr, lazy_let_expr).


:- mode syntaxtraversion_fixed(+TExpr,-Expr,-Type,-Infos,-QSubs,-Subs,-Names).
syntaxtraversion_fixed(b(Expr,Type,Infos),Expr,Type,Infos,QSubs,Subs,Names) :-
   syntaxtraversion_fixed_aux(Expr,QSubs,Subs,Names).

% for performance: a few often used cases (not really sure it buys a lot):
syntaxtraversion_fixed_aux(truth,QSubs,Subs,Names) :- !, QSubs=[],Subs=[],Names=[].
syntaxtraversion_fixed_aux(falsity,QSubs,Subs,Names) :- !, QSubs=[],Subs=[],Names=[].
syntaxtraversion_fixed_aux(identifier(_),QSubs,Subs,Names) :- !, QSubs=[],Subs=[],Names=[].
syntaxtraversion_fixed_aux(string(_),QSubs,Subs,Names) :- !, QSubs=[],Subs=[],Names=[].
syntaxtraversion_fixed_aux(integer(_),QSubs,Subs,Names) :- !, QSubs=[],Subs=[],Names=[].
syntaxtraversion_fixed_aux(value(_),QSubs,Subs,Names) :- !, QSubs=[],Subs=[],Names=[].
syntaxtraversion_fixed_aux(conjunct(A,B),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A,B],Names=[].
syntaxtraversion_fixed_aux(disjunct(A,B),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A,B],Names=[].
syntaxtraversion_fixed_aux(implication(A,B),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A,B],Names=[].
syntaxtraversion_fixed_aux(equal(A,B),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A,B],Names=[].
syntaxtraversion_fixed_aux(not_equal(A,B),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A,B],Names=[].
syntaxtraversion_fixed_aux(function(A,B),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A,B],Names=[].
syntaxtraversion_fixed_aux(image(A,B),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A,B],Names=[].
syntaxtraversion_fixed_aux(couple(A,B),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A,B],Names=[].
syntaxtraversion_fixed_aux(member(A,B),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A,B],Names=[].
syntaxtraversion_fixed_aux(union(A,B),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A,B],Names=[].
syntaxtraversion_fixed_aux(set_subtraction(A,B),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A,B],Names=[].
syntaxtraversion_fixed_aux(intersection(A,B),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A,B],Names=[].
syntaxtraversion_fixed_aux(interval(A,B),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A,B],Names=[].
syntaxtraversion_fixed_aux(total_function(A,B),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A,B],Names=[].
syntaxtraversion_fixed_aux(less(A,B),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A,B],Names=[].
syntaxtraversion_fixed_aux(record_field(A,_),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A],Names=[].
syntaxtraversion_fixed_aux(range(A),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A],Names=[].
syntaxtraversion_fixed_aux(domain(A),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A],Names=[].
syntaxtraversion_fixed_aux(covert_bool(A),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A],Names=[].
syntaxtraversion_fixed_aux(card(A),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A],Names=[].
syntaxtraversion_fixed_aux(min(A),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A],Names=[].
syntaxtraversion_fixed_aux(max(A),QSubs,Subs,Names) :- !, QSubs=[],Subs=[A],Names=[].
%:- use_module(hit_profiler,[add_profile_hit/1]). %hit_profiler:profile_statistics.
syntaxtraversion_fixed_aux(Expr,QSubs,Subs,Names) :-
   %hit_profiler:add_profile_hit(Expr),
    (is_let(Expr,TIds,Exprs,ExprPred,_LetType)
      -> Names=TIds, QSubs=Exprs,
         (last(TIds,ExprPred,Subs) -> true)
      ;  QSubs=[],safe_syntaxelement(Expr,Subs,Names,_,_)).

% bsyntaxtree:syntaxtraversion_fixed(b(value(X),integer,[]),E,T,I,Q,S,N).
% bsyntaxtree:syntaxtraversion_fixed(b(equal(a,b),integer,[]),E,T,I,Q,S,N).
% bsyntaxtree:syntaxtraversion_fixed(b(record_field(a,b),integer,[]),E,T,I,Q,S,N).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% rename identifier

rename_bt(Expr,[],Res) :- !,Res=Expr.
rename_bt(OExpr,Renamings,NExpr) :-
    create_texpr(Old,Type,OInfo,OExpr),
    create_texpr(New,Type,NInfo,NExpr),
    rename_in_infos(OInfo,Renamings,NInfo),
    rename_bt2(Old,Renamings,New).
rename_bt2(identifier(Old),Renamings,identifier(New)) :-
    !, rename_id(Old,Renamings,New).
rename_bt2(lazy_lookup_expr(Old),Renamings,lazy_lookup_expr(New)) :-
    !, rename_id(Old,Renamings,New).
rename_bt2(lazy_lookup_pred(Old),Renamings,lazy_lookup_pred(New)) :-
    !, rename_id(Old,Renamings,New).
rename_bt2(OExpr,Renamings,NExpr) :-
    syntaxtransformation(OExpr,Subs,TNames,NSubs,NExpr),
    get_texpr_exprs(TNames,Names),
    remove_renamings(Names,Renamings,NRenamings),
    rename_bt_l(Subs,NRenamings,NSubs).
rename_bt_l([],_,[]).
rename_bt_l([Old|ORest],Renamings,[New|NRest]) :-
    rename_bt(Old,Renamings,New),
    rename_bt_l(ORest,Renamings,NRest).

remove_renamings([],Renamings,Renamings).
remove_renamings([identifier(Name)|Rest],Old,New) :-
    ( select(rename(Name,_),Old,Inter1) -> true  % Name no longer visible to renaming
          ; Old = Inter1),
    (member(rename(_OldName,Name),Inter1) ->
        gensym:gensym('__FRESH__',FRESHID),
        %print(variable_capture_in_rename(Name,from(OldName),FRESHID)),nl,
        Inter2 = [rename(Name,FRESHID)|Inter1]
      ; Inter2 = Inter1),
    remove_renamings(Rest,Inter2,New).

rename_in_infos(Old,Renamings,New) :-
    ( has_info_to_rename(Old) ->
        maplist(rename_in_infos2(Renamings),Old,New)
    ; otherwise ->
        Old = New).
rename_in_infos2(Renamings,OInfo,NInfo) :-
    ( infos_to_rename(OInfo,OIds,SortedNIds,NInfo) ->
        rename_ids(OIds,Renamings,NIds),
        sort(NIds,SortedNIds)
    ; otherwise ->
        OInfo = NInfo).
rename_ids([],_,[]).
rename_ids([OId|Orest],Renamings,[NId|Nrest]) :-
    rename_id(OId,Renamings,NId),
    rename_ids(Orest,Renamings,Nrest).
rename_id(Old,Renamings,New) :-
    ( memberchk(rename(Old,New),Renamings) -> true
    ; otherwise -> Old=New).
has_info_to_rename(Infos) :-
    member(I,Infos),infos_to_rename(I,_,_,_),!.
infos_to_rename(modifies(O),O,N,modifies(N)).
infos_to_rename(reads(O),O,N,reads(N)).
infos_to_rename(used_ids(O),O,N,used_ids(N)).

%remove_renamings([rename(Old,New)|Rest],Names,Renamings) :-
%    ( member(identifier(Old),Names) ->
%        Renamings = RenamingsRest
%    ; otherwise ->
%        Renamings = [rename(Old,New)|RenamingsRest]),
%    remove_renamings(Rest,Names,RenamingsRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% remove type information for transformations

remove_bt(b(Expr,Type,Infos),Expr,NExpr,b(NExpr,Type,Infos)).

%remove_bt_l([],[],[],[]).
%remove_bt_l([OT|OTRest],[O|ORest],[N|NRest],[NT|NTRest]) :-
%   remove_bt(OT,O,N,NT),
%    remove_bt_l(OTRest,ORest,NRest,NTRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% traversations
% Takes a predicate or expression or substitution and extracts:
%  the expression itself Expr, its type Type, the syntaxnode information Infos
%  + the subexpressions as a list Subs and the local identifiers declared
syntaxtraversion(b(Expr,Type,Infos),Expr,Type,Infos,Subs,Names) :- !,
    safe_syntaxelement(Expr,Subs,Names,_,_).
syntaxtraversion(IExpr,Expr,Type,Infos,Subs,Names) :-
    add_internal_error('Not properly wrapped', syntaxtraversion(IExpr,Expr,Type,Infos,Subs,Names)),
    fail.

:- meta_predicate map_over_bexpr(1,?).
map_over_bexpr(P,BExpr) :-
   syntaxtraversion(BExpr,Expr,_,_,Subs,_TNames),
   (call(P,Expr)
    ;
    member(Sub,Subs), map_over_bexpr(P,Sub)
   ).
% same as above but gets typed expressions:
:- meta_predicate map_over_typed_bexpr(1,?).
map_over_typed_bexpr(P,BExpr) :-
   syntaxtraversion(BExpr,_Expr,_,_,Subs,_TNames),
   (call(P,BExpr)
    ;
    member(Sub,Subs), map_over_typed_bexpr(P,Sub)
   ).
% same as above but returns value:
:- meta_predicate map_over_typed_bexpr(2,?,?).
map_over_typed_bexpr(P,BExpr,Result) :-
   syntaxtraversion(BExpr,_Expr,_,_,Subs,_TNames),
   (call(P,BExpr,Result)
    ;
    member(Sub,Subs), map_over_typed_bexpr(P,Sub,Result)
   ).

:- meta_predicate reduce_over_bexpr(3,?,?,?).
reduce_over_bexpr(P,BExpr,InitialValue,ResultValue) :-
   syntaxtraversion(BExpr,Expr,_,_,Subs,_TNames),
   call(P,Expr,InitialValue,I1), % print(reduce(P,Expr,InitialValue,I1)),nl,
   scanlist(reduce_over_bexpr(P),Subs,I1,ResultValue).

% apply a predicate over a syntax tree (bottom-up)
:- meta_predicate transform_bexpr(2,?,?).
transform_bexpr(Pred,b(Expr,Type,Info),NewBExpr) :-
   syntaxtransformation(Expr,Subs,_Names,NSubs,NewExpr1),
   l_transform_bexpr(Subs,Pred,NSubs),
   (call(Pred,b(NewExpr1,Type,Info),NewBExpr) -> true ; NewBExpr = b(NewExpr1,Type,Info)).

l_transform_bexpr([],_,[]).
l_transform_bexpr([SubH|T],Pred,[TSubH|TT]) :-
   transform_bexpr(Pred,SubH,TSubH),
   l_transform_bexpr(T,Pred,TT).

% apply a predicate over a syntax tree (bottom-up) with Accumulator result
% Accumulator is constructed bottom up; Pred receives all accumulators of sub expressions
:- meta_predicate transform_bexpr_with_acc(4,?,?,?,?).
transform_bexpr_with_acc(Pred,b(Expr,Type,Info),NewBExpr,EmptyAcc,Acc) :-
   syntaxtransformation(Expr,Subs,_Names,NSubs,NewExpr1),
   l_transform_bexpr_with_acc(Subs,Pred,NSubs,EmptyAcc,SubAccs),
   (call(Pred,b(NewExpr1,Type,Info),NewBExpr,SubAccs,Acc) -> true
    ; NewBExpr = b(NewExpr1,Type,Info), Acc = EmptyAcc).

l_transform_bexpr_with_acc([],_,[],_,[]).
l_transform_bexpr_with_acc([SubH|T],Pred,[TSubH|TT],EmptyAcc,[Acc1|RestAcc]) :-
   transform_bexpr_with_acc(Pred,SubH,TSubH,EmptyAcc,Acc1),
   l_transform_bexpr_with_acc(T,Pred,TT,EmptyAcc,RestAcc).

min_max_integer_value_used(BExpr,Min,Max) :-
   min_max_integer_value_used(BExpr,none,none,Min,Max).
min_max_integer_value_used(BExpr,IMin,IMax,Min,Max) :-
   reduce_over_bexpr(min_max_aux,BExpr,minmax(IMin,IMax),minmax(Min,Max)).

min_max_aux(sequence_extension(L),minmax(Min,Max),minmax(NMin,NMax)) :- !,
  length(L,Len), % we use implicitly numbers from 1..Len
  ((number(Min),1>Min) -> NMin=Min ; NMin=1),
  ((number(Max),Len<Max) -> NMax=Max ; NMax=Len).
min_max_aux(integer(N),minmax(Min,Max),minmax(NMin,NMax)) :- !,
  ((number(Min),N>Min) -> NMin=Min ; NMin=N),
  ((number(Max),N<Max) -> NMax=Max ; NMax=N).
min_max_aux(_,V,V).

% check if a B expression uses something like NAT,NAT1,INT, MAXINT or MININT.
uses_implementable_integers(BExpr) :-
   map_over_bexpr(uses_implementable_integers_aux,BExpr).

uses_implementable_integers_aux(maxint).
uses_implementable_integers_aux(minint).
uses_implementable_integers_aux(integer_set(X)) :-
   (X='NAT1' ; X='NAT' ; X='INT').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% some checks
check_if_typed_predicate(b(Pred,X,_)) :- ground(X), X=pred,  % at runtime there can be value(X) with variables inside !
    syntaxelement(Pred,_,_,_,_,pred).
check_if_typed_expression(b(Expr,Type,_)) :-
    syntaxelement(Expr,_,_,_,_,expr),
    Type \== pred, Type \== subst, ground(Type).
check_if_typed_substitution(b(Subst,X,_)) :- ground(X), X=subst,
    syntaxelement(Subst,_,_,_,_,subst).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% transformations

syntaxtransformation(Expr,Subs,Names,NSubs,NExpr) :-
    functor(Expr,F,Arity),
    functor(NExpr,F,Arity),
    safe_syntaxelement(Expr,Subs,Names,Lists,Constant),
    all_same_length(Lists,NLists),
    syntaxelement(NExpr,NSubs,_,NLists,Constant,_).

safe_syntaxelement(Expr,Subs,Names,Lists,Constant) :-
    ( syntaxelement(Expr,SubsX,NamesX,Lists,ConstantX,_) ->
        Subs=SubsX, Names=NamesX, Constant=ConstantX
    ; otherwise ->
        functor(Expr,F,Arity),
        add_error_fail(bsyntaxtree,'Uncovered syntax element: ', F/Arity)
        ).

% syntaxelement(Expr,SubExprs,Identifiers,Lists,Constant,Type):
% Expr: the expression itself
% SubExprs: a list of sub-expressions
% Identifiers: a list of identifiers that are newly introduced (e.g. by a quantifier)
% Lists: A list of lists in the expression, to prevent infinite loops when having variable parts
% Constant: A part of the expression that is not a sub-expression (e.g. the number in integer(...))
% Type: Fundamental type of the element (predicate, expression, etc)

% predicates
syntaxelement(truth,                 [],   [], [], [], pred).
syntaxelement(falsity,               [],   [], [], [], pred).
syntaxelement(conjunct(A,B),         [A,B],[], [], [], pred).
syntaxelement(negation(A),           [A],  [], [], [], pred).
syntaxelement(disjunct(A,B),         [A,B],[], [], [], pred).
syntaxelement(implication(A,B),      [A,B],[], [], [], pred).
syntaxelement(equivalence(A,B),      [A,B],[], [], [], pred).
syntaxelement(equal(A,B),            [A,B],[], [], [], pred).
syntaxelement(not_equal(A,B),        [A,B],[], [], [], pred).
syntaxelement(member(A,B),           [A,B],[], [], [], pred).
syntaxelement(not_member(A,B),       [A,B],[], [], [], pred).
syntaxelement(subset(A,B),           [A,B],[], [], [], pred).
syntaxelement(subset_strict(A,B),    [A,B],[], [], [], pred).
syntaxelement(not_subset(A,B),       [A,B],[], [], [], pred).
syntaxelement(not_subset_strict(A,B),[A,B],[], [], [], pred).
syntaxelement(less_equal(A,B),       [A,B],[], [], [], pred).
syntaxelement(less(A,B),             [A,B],[], [], [], pred).
syntaxelement(greater_equal(A,B),    [A,B],[], [], [], pred).
syntaxelement(greater(A,B),          [A,B],[], [], [], pred).
syntaxelement(forall(Ids,D,P),       [D,P|Ids],Ids,[Ids], [], pred).
syntaxelement(exists(Ids,P),         [P|Ids],  Ids,[Ids], [], pred).
syntaxelement(finite(A),             [A],  [], [], [], pred/only_typecheck).
syntaxelement(partition(S,Es),       [S|Es],[],[Es],[],pred).
syntaxelement(kodkod(PId,Ids),       Ids,[],[Ids],PId, pred).
syntaxelement(external_pred_call(F,Args),Args,[],[Args],F,pred).

% expressions
syntaxelement(value(V),                [],   [], [], V,  expr).
syntaxelement(boolean_true,            [],   [], [], [], expr).
syntaxelement(boolean_false,           [],   [], [], [], expr).
syntaxelement(max_int,                 [],   [], [], [], expr).
syntaxelement(min_int,                 [],   [], [], [], expr).
syntaxelement(empty_set,               [],   [], [], [], expr).
syntaxelement(bool_set,                [],   [], [], [], expr).
syntaxelement(string_set,              [],   [], [], [], expr).
syntaxelement(convert_bool(A),         [A],  [], [], [], expr).
syntaxelement(add(A,B),                [A,B],[], [], [], expr).
syntaxelement(minus(A,B),              [A,B],[], [], [], expr).
syntaxelement(minus_or_set_subtract(A,B),[A,B],[], [], [], expr/only_typecheck).
syntaxelement(unary_minus(A),          [A],  [], [], [], expr).
syntaxelement(multiplication(A,B),     [A,B],[], [], [], expr).
syntaxelement(mult_or_cart(A,B),       [A,B],[], [], [], expr/only_typecheck).
syntaxelement(cartesian_product(A,B),  [A,B],[], [], [], expr).
syntaxelement(div(A,B),                [A,B],[], [], [], expr).
syntaxelement(modulo(A,B),             [A,B],[], [], [], expr).
syntaxelement(power_of(A,B),           [A,B],[], [], [], expr).
syntaxelement(successor,               [],   [], [], [], expr).
syntaxelement(predecessor,             [],   [], [], [], expr).
syntaxelement(max(A),                  [A],  [], [], [], expr).
syntaxelement(min(A),                  [A],  [], [], [], expr).
syntaxelement(card(A),                 [A],  [], [], [], expr).
syntaxelement(couple(A,B),             [A,B],[], [], [], expr).
syntaxelement(pow_subset(A),           [A],  [], [], [], expr).
syntaxelement(pow1_subset(A),          [A],  [], [], [], expr).
syntaxelement(fin_subset(A),           [A],  [], [], [], expr).
syntaxelement(fin1_subset(A),          [A],  [], [], [], expr).
syntaxelement(interval(A,B),           [A,B],[], [], [], expr).
syntaxelement(union(A,B),              [A,B],[], [], [], expr).
syntaxelement(intersection(A,B),       [A,B],[], [], [], expr).
syntaxelement(set_subtraction(A,B),    [A,B],[], [], [], expr).
syntaxelement(general_union(A),        [A],  [], [], [], expr).
syntaxelement(general_intersection(A), [A] , [], [], [], expr).
syntaxelement(relations(A,B),          [A,B],[], [], [], expr).
syntaxelement(identity(A),             [A],  [], [], [], expr).
syntaxelement(event_b_identity,        [],  [], [], [], expr). % for Rodin 1.0, TO DO: Daniel please check
syntaxelement(reverse(A),              [A],  [], [], [], expr).
syntaxelement(first_projection(A,B),   [A,B],[], [], [], expr/only_typecheck).
syntaxelement(first_of_pair(A),        [A],  [], [], [], expr).
syntaxelement(event_b_first_projection(A),[A],  [], [], [], expr/only_typecheck).
syntaxelement(event_b_first_projection_v2,[],  [], [], [], expr/only_typecheck). % for Rodin 1.0, TO DO: Daniel please check
syntaxelement(second_projection(A,B),  [A,B],[], [], [], expr/only_typecheck).
syntaxelement(event_b_second_projection_v2,[],  [], [], [], expr/only_typecheck). % for Rodin 1.0, TO DO: Daniel please check
syntaxelement(second_of_pair(A),       [A],  [], [], [], expr).
syntaxelement(event_b_second_projection(A),[A],  [], [], [], expr/only_typecheck).
syntaxelement(composition(A,B),        [A,B],[], [], [], expr).
syntaxelement(ring(A,B),               [A,B],[], [], [], expr/only_typecheck).
syntaxelement(direct_product(A,B),     [A,B],[], [], [], expr).
syntaxelement(parallel_product(A,B),   [A,B],[], [], [], expr).
syntaxelement(trans_function(A),       [A],  [], [], [], expr).
syntaxelement(trans_relation(A),       [A],  [], [], [], expr).
syntaxelement(iteration(A,B),          [A,B],[], [], [], expr).
syntaxelement(reflexive_closure(A),    [A],  [], [], [], expr).
syntaxelement(closure(A),              [A],  [], [], [], expr).
syntaxelement(domain(A),               [A],  [], [], [], expr).
syntaxelement(range(A),                [A],  [], [], [], expr).
syntaxelement(image(A,B),              [A,B],[], [], [], expr).
syntaxelement(domain_restriction(A,B), [A,B],[], [], [], expr).
syntaxelement(domain_subtraction(A,B), [A,B],[], [], [], expr).
syntaxelement(range_restriction(A,B),  [A,B],[], [], [], expr).
syntaxelement(range_subtraction(A,B), [A,B],[], [], [], expr).
syntaxelement(overwrite(A,B),          [A,B],[], [], [], expr).
syntaxelement(partial_function(A,B),   [A,B],[], [], [], expr).
syntaxelement(total_function(A,B),     [A,B],[], [], [], expr).
syntaxelement(partial_injection(A,B),  [A,B],[], [], [], expr).
syntaxelement(total_injection(A,B),    [A,B],[], [], [], expr).
syntaxelement(partial_surjection(A,B), [A,B],[], [], [], expr).
syntaxelement(total_surjection(A,B),   [A,B],[], [], [], expr).
syntaxelement(total_bijection(A,B),    [A,B],[], [], [], expr).
syntaxelement(partial_bijection(A,B),  [A,B],[], [], [], expr).
syntaxelement(total_relation(A,B),     [A,B],[], [], [], expr).
syntaxelement(surjection_relation(A,B),[A,B],[], [], [], expr).
syntaxelement(total_surjection_relation(A,B),[A,B],[], [], [], expr).
syntaxelement(seq(A),                  [A],  [], [], [], expr).
syntaxelement(seq1(A),                 [A],  [], [], [], expr).
syntaxelement(iseq(A),                 [A],  [], [], [], expr).
syntaxelement(iseq1(A),                [A],  [], [], [], expr).
syntaxelement(perm(A),                 [A],  [], [], [], expr).
syntaxelement(empty_sequence,          [],   [], [], [], expr).
syntaxelement(size(A),                 [A],  [], [], [], expr).
syntaxelement(first(A),                [A],  [], [], [], expr).
syntaxelement(last(A),                 [A],  [], [], [], expr).
syntaxelement(front(A),                [A],  [], [], [], expr).
syntaxelement(tail(A),                 [A],  [], [], [], expr).
syntaxelement(rev(A),                  [A],  [], [], [], expr).
syntaxelement(concat(A,B),             [A,B],[], [], [], expr).
syntaxelement(insert_front(A,B),       [A,B],[], [], [], expr).
syntaxelement(insert_tail(A,B),        [A,B],[], [], [], expr).
syntaxelement(restrict_front(A,B),     [A,B],[], [], [], expr).
syntaxelement(restrict_tail(A,B),      [A,B],[], [], [], expr).
syntaxelement(general_concat(A),       [A],  [], [], [], expr).
syntaxelement(function(A,B),           [A,B],[], [], [], expr).
syntaxelement(external_function_call(F,Args),Args,[],[Args],F,expr).
syntaxelement(identifier(I),           [],   [], [], I,  expr).
syntaxelement(lazy_lookup_expr(I),           [],   [], [], I,  expr).
syntaxelement(lazy_lookup_pred(I),           [],   [], [], I,  pred).
syntaxelement(integer(I),              [],   [], [], I,  expr).
syntaxelement(integer_set(T),          [],   [], [], T,  expr).
syntaxelement(string(S),               [],   [], [], S,  expr).
syntaxelement(set_extension(L),        L,    [], [L], [], expr).
syntaxelement(sequence_extension(L),   L,    [], [L], [], expr).
syntaxelement(comprehension_set(Ids,P),[P|Ids],  Ids,[Ids], [], expr).
syntaxelement(event_b_comprehension_set(Ids,E,P),[E,P|Ids],  Ids,[Ids], [], expr/only_typecheck).
syntaxelement(lambda(Ids,P,E),         [P,E|Ids],Ids,[Ids], [], expr).
syntaxelement(general_sum(Ids,P,E),             [P,E|Ids],Ids,[Ids], [], expr).
syntaxelement(general_product(Ids,P,E),         [P,E|Ids],Ids,[Ids], [], expr).
syntaxelement(quantified_union(Ids,P,E),        [P,E|Ids],Ids,[Ids], [], expr).
syntaxelement(quantified_intersection(Ids,P,E), [P,E|Ids],Ids,[Ids], [], expr).
syntaxelement(struct(Rec), [Rec], [], [], [], expr).
syntaxelement(rec(Fields), FContent, [], [FContent], FNames, expr) :- syntaxfields(Fields,FContent, FNames).
syntaxelement(record_field(R,I), [R], [], [], I, expr).
syntaxelement(assertion_expression(Cond,ErrMsg,Expr), [Cond,Expr], [], [], ErrMsg, expr).
syntaxelement(typeset, [], [], [], [], expr/only_typecheck).

% substitutions
syntaxelement(skip,                    [],   [], [], [], subst).
syntaxelement(precondition(A,B),       [A,B],[], [], [], subst).
syntaxelement(assertion(A,B),          [A,B],[], [], [], subst).
syntaxelement(if_elsif(A,B),           [A,B],[], [], [], subst/elsif).
syntaxelement(while(A,B,C,D),          [A,B,C,D],[], [], [], subst).
% used only internally in the interpreter, contains last value of variant:
syntaxelement(while1(A,B,C,D,E),          [A,B,C,D],[], [], E, subst).
syntaxelement(select_when(A,B),        [A,B],[], [], [], subst/when).
syntaxelement(block(S),[S],[],[],[], subst/only_typecheck).
syntaxelement(assign(Lhs,Rhs),Exprs,[],[Lhs,Rhs], [], subst) :- append(Lhs,Rhs,Exprs).
syntaxelement(assign_single_id(Id,Rhs),[Id,Rhs],[],[], [], subst).
syntaxelement(any(Ids,P,S),[P,S|Ids],Ids,[Ids], [], subst).
syntaxelement(var(Ids,S),  [S|Ids],  Ids,[Ids], [], subst).
syntaxelement(if(Ifs),   Ifs,  [], [Ifs], [], subst).
syntaxelement(parallel(Ss), Ss, [], [Ss], [], subst).
syntaxelement(sequence(Ss), Ss, [], [Ss], [], subst).
syntaxelement(becomes_element_of(Ids,E), [E|Ids], [], [Ids], [], subst).
syntaxelement(becomes_such(Ids,P), [P|Ids], [], [Ids], [], subst).
syntaxelement(evb_becomes_such(Ids,P), [P|Ids], [], [Ids], [], subst/only_typecheck).
syntaxelement(evb2_becomes_such(Ids,P), [P|Ids], [], [Ids], [], subst/only_typecheck).
syntaxelement(let(Ids,P,S), [P,S|Ids], Ids, [Ids], [], subst).
syntaxelement(operation_call(Id,Rs,As), [Id|Exprs], [], [Rs,As], [], subst) :- append(Rs,As,Exprs).
syntaxelement(case(E,Eithers,Else), [E,Else|Eithers], [], [Eithers], [], subst).
syntaxelement(case_or(Es,S), [S|Es], [], [Es], [], subst/caseor).
syntaxelement(choice(Ss), Ss, [], [Ss], [], subst).
syntaxelement(select(Whens), Whens, [], [Whens], [], subst).
syntaxelement(select(Whens,Else), [Else|Whens], [], [Whens], [], subst).
syntaxelement(operation(I,Rs,As,B), [I,B|Ids], Ids, [Rs,As], [], subst) :- append(Rs,As,Ids).
syntaxelement(external_subst_call(F,Args),Args,[],[Args],F,subst).

% elements of a VALUES clause
syntaxelement(values_entry(I,E),[I,E],[],[],[],values_entry).

% syntax for Event-B events
syntaxelement(rlevent(I,Sec,St,Ps,G,Ts,As,VWs,PWs,Ums,Rs), Subs, [], [Ps,Ts,As,VWs,PWs,Ums,Rs], [I,Sec], subst) :-
    append([[St],Ps,[G],Ts,As,VWs,PWs,Ums,Rs],Subs).
syntaxelement(witness(I,P), [I,P], [], [], [], witness).

% extended syntax for Z
syntaxelement(let_predicate(Ids,As,Pred), Exprs, Ids, [Ids,As], [], pred) :- append([Ids,As,[Pred]],Exprs).
syntaxelement(let_expression(Ids,As,Expr), Exprs, Ids, [Ids,As], [], expr) :- append([Ids,As,[Expr]],Exprs).
syntaxelement(lazy_let_expr(TID,A,Expr), Exprs, [TID], [[TID],[A]], [], expr) :-
   append([[TID],[A],[Expr]],Exprs).
syntaxelement(lazy_let_pred(TID,A,Expr), Exprs, [TID], [[TID],[A]], [], pred) :-
   append([[TID],[A],[Expr]],Exprs).
syntaxelement(if_then_else(If,Then,Else),[If,Then,Else], [], [], [], expr).
syntaxelement(compaction(A), [A],  [], [], [], expr).
syntaxelement(mu(A),         [A],  [], [], [], expr).
syntaxelement(bag_items(A),  [A],  [], [], [], expr).

syntaxelement(freetype_set(Id), [], [], [], Id, expr).
syntaxelement(freetype_case(Type,Case,Expr), [Expr], [], [], [Type,Case], pred).
syntaxelement(freetype_constructor(Type,Case,Expr), [Expr], [], [], [Type,Case], expr).
syntaxelement(freetype_destructor(Type,Case,Expr), [Expr], [], [], [Type,Case], expr).

syntaxelement(ordinary,    [], [], [], [], status).
syntaxelement(anticipated(Variant), [Variant], [], [], [], status).
syntaxelement(convergent(Variant),  [Variant], [], [], [], status).

% Just one ID expected
syntaxelement(recursive_let(Id,C),[Id,C],[Id],[],[], expr).


% fields of records
syntaxfields([],[],[]).
syntaxfields([field(N,C)|Rest],[C|CRest],[N|NRest]) :- syntaxfields(Rest,CRest,NRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% helper for declaration of quantified identifiers

% has_declared_identifier/2 returns a list of identifiers which are declared in
% this AST node. The main difference to the Names variable when doing a
% syntaxtraversion/6 or similiar is that the identifiers are described by
% predicates.
has_declared_identifier(TExpr,Ids) :-
    get_texpr_expr(TExpr,Expr),
    ( default_declaration(Expr,_,Ids,_)
    ; non_default_declaration(Expr,Ids)).

add_declaration_for_identifier(TExpr,Decl,NTExpr) :-
    remove_bt(TExpr,Expr,NExpr,NTExpr),
    ( default_declaration(Expr,Predicate,Ids,Constant) ->
        same_functor(Expr,NExpr),
        conjunct_predicates([Decl,Predicate],NPredicate),
        default_declaration(NExpr,NPredicate,Ids,Constant)
    ; non_default_declaration(Expr,_Ids) ->
        add_non_default_declaration(Expr,Decl,NExpr)).

default_declaration(forall(Ids,D,P),D,Ids,P).
default_declaration(exists(Ids,P),P,Ids,[]).
default_declaration(comprehension_set(Ids,P),P,Ids,[]).
default_declaration(lambda(Ids,P,E),P,Ids,E).
default_declaration(general_sum(Ids,P,E),P,Ids,E).
default_declaration(general_product(Ids,P,E),P,Ids,E).
default_declaration(quantified_union(Ids,P,E),P,Ids,E).
default_declaration(quantified_intersection(Ids,P,E),P,Ids,E).
default_declaration(any(Ids,P,S),P,Ids,S).
default_declaration(becomes_such(Ids,P),P,Ids,[]).
default_declaration(rlevent(I,Sec,St,Ps,G,Ts,As,VWs,PWs,Ums,Rs),G,Ps,
                        [I,Sec,St,Ps,Ts,As,VWs,PWs,Ums,Rs]).

non_default_declaration(operation(_I,Rs,As,_TBody),Ids) :-
    append(Rs,As,Ids).

add_non_default_declaration(operation(I,Rs,As,TBody),Decl,operation(I,Rs,As,NTBody)) :-
    ( remove_bt(TBody,precondition(P,S),precondition(NP,S),NTBody) ->
        conjunct_predicates([Decl,P],NP)
    ; otherwise ->
        create_texpr(precondition(Decl,TBody),subst,[],NTBody)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pattern definitions (not yet finished)

bsyntax_pattern(Expr,TExpr) :-
    var(Expr),!,Expr=TExpr.
bsyntax_pattern(-Expr,TExpr) :-
    !,remove_all_infos(Expr,TExpr).
bsyntax_pattern(Expr,TExpr) :-
    functor(Expr,b,3),!,Expr=TExpr.
bsyntax_pattern(Expr:Type/Info,TExpr) :-
    !,bsyntax_pattern2(Expr,Type,Info,TExpr).
bsyntax_pattern(Expr:Type,TExpr) :-
    !,bsyntax_pattern2(Expr,Type,_Info,TExpr).
bsyntax_pattern(Expr/Info,TExpr) :-
    !,bsyntax_pattern2(Expr,_Type,Info,TExpr).
bsyntax_pattern(Expr,TExpr) :-
    !,bsyntax_pattern2(Expr,_Type,_Info,TExpr).

bsyntax_pattern2(Pattern,Type,Info,TExpr) :-
    functor(Pattern,Functor,Arity),
    functor(R,Functor,Arity),
    create_texpr(R,Type,Info,TExpr),
    syntaxelement(Pattern,PSubs,_,PLists,Const,EType),
    syntaxelement(R,RSubs,_,RLists,Const,EType),
    ( EType==pred -> Type=pred
    ; EType==subst -> Type=subst
    ; otherwise -> true),
    all_same_length(PLists,RLists),
    maplist(bsyntax_pattern,PSubs,RSubs).

all_same_length([],[]).
all_same_length([A|Arest],[B|Brest]) :-
    ( var(A),var(B) ->
        add_error_fail(bsyntaxtree,'At least one list should be nonvar',[A,B])
    ; otherwise ->
        same_length(A,B)),
    all_same_length(Arest,Brest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% strip an AST into a more compact form (without b/3 terms)

strip_and_norm_ast(TExpr,SNExpr) :-
    get_texpr_expr(TExpr,Expr),
    syntaxtransformation(Expr,Subs,_,NSubs,SExpr),
    strip_and_norm_ast_l(Subs,NSubs),
    norm_strip(SExpr,SNExpr).
strip_and_norm_ast_l([],[]).
strip_and_norm_ast_l([TExpr|Trest],[NExpr|Nrest]) :-
    strip_and_norm_ast(TExpr,NExpr),
    strip_and_norm_ast_l(Trest,Nrest).
norm_strip(greater_equal(A,B),less_equal(B,A)) :- !.
norm_strip(greater(A,B),less(B,A)) :- !.
norm_strip(Old,New) :-
    functor(Old,Functor,2),
    is_commutative(Functor),!,
    Old =.. [Functor,OA,OB],
    New =.. [Functor,NA,NB],
    ( OA @>= OB -> OA=NB,OB=NA
    ; otherwise -> OA=NA,OB=NB).
norm_strip(Old,Old).


is_commutative(conjunct).
is_commutative(disjunct).
is_commutative(equivalence).
is_commutative(equal).
is_commutative(add).
is_commutative(multiplication).
is_commutative(union).
is_commutative(intersection).

% check if a ProB type is a set type:
is_set_type(set(Type),Type).
is_set_type(seq(Type),couple(integer,Type)).
% should we have a rule for any ?? : in all cases using is_set_type any seems not possible; better that we generate error message in get_set_type or get_texpr_set_type
% is_set_type(any,any).

get_set_type(TypeX,Res) :-
   (is_set_type(TypeX,SetType) -> Res=SetType ; add_error_fail(get_set_type,'Not a set type: ',TypeX)).

get_texpr_set_type(X,Res) :- get_texpr_type(X,TypeX),
   get_set_type(TypeX,Res).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% check if a set contains all elements of a type, i.e., maximal type
is_just_type(Expr) :- is_just_type(Expr,[]).
% is_just_type(+Expr,+RTT):
%  like is_just_type/1 but RTT is a list of identifiers (without type information)
%  of variables or constants that are known to be types, too.
%  E.g. is_just_type(a ** INTEGER) would fail but is_just_type(a ** INTEGER,[a]) would
%  succeed.
is_just_type(b(E,T,I),RTT) :- is_just_type3(E,T,I,RTT).


is_just_type3(identifier(G),set(Type),Infos,RefsToTypes) :-
    ( Type = global(G), memberchk(given_set,Infos)
    ; memberchk(G,RefsToTypes)),!.
is_just_type3(freetype_set(_),_,_,_).
is_just_type3(pow_subset(E),_,_,RTT) :- is_just_type(E,RTT).
is_just_type3(fin_subset(E),T,_,RTT) :- \+ is_infinite_type(T),is_just_type(E,RTT).
is_just_type3(integer_set('INTEGER'),set(integer),_,_).
is_just_type3(bool_set,set(boolean),_,_).
is_just_type3(string_set,set(string),_,_).
is_just_type3(cartesian_product(A,B),_,_,RTT) :-
    is_just_type(A,RTT),is_just_type(B,RTT).
is_just_type3(mult_or_cart(A,B),_,_,RTT) :-    % is_just_type/1 could be called before the cleanup
    is_just_type(A,RTT),is_just_type(B,RTT).   % of an expression has finished
is_just_type3(relations(A,B),_,_,RTT) :-
    is_just_type(A,RTT),is_just_type(B,RTT).
is_just_type3(struct(b(rec(Fields),_,_)),_,_,RTT) :-
    maplist(field_is_just_type(RTT),Fields).
is_just_type3(typeset,_,_,_).
field_is_just_type(RTT,field(_,Set)) :- is_just_type(Set,RTT).


% Generate a custom matching / visitor predicate
% bsyntaxtree:gen_visitor(bexpr_variables_aux,bexpr_variables)
/*
gen_visitor(Pred, BodyPred) :- syntaxelement(Expr,SubsX,_NamesX,_Lists,_ConstantX,_),
    print_clause(Pred,BodyPred,Expr,SubsX),
    fail.
gen_visitor(Pred, BodyPred) :- nl.

print_clause(Pred,BodyPred,Expr,SubsX) :- E=..[Op|SubsX], format('~w(~w) :- ',[Pred,Expr]).
*/

/*
 * create_recursive_compset(+Ids,+Cond,+Type,+Infos,-RecId,-TCompSet)
 *  creates a recursive comprehension set:
 *   Ids      is the list of the introduced typed identifiers
 *   Cond     is the condition (a typed predicate)
 *   Infos    are additional informations for the comprehension set syntax element
 *   RecId    is the identifier (untyped, an atom) that can be used in Cond to refer to the
 *            comprehension set recursively. Usually RecId is used in Cond as a variable
 *   TCompSet is the generated recursive comprehension set, the recursion parameter is introduced
 *            by a recusive(Id,CompSet) syntax element
 */
create_recursive_compset(Ids,Cond,Type,Infos,RecId,TCompSet) :-
    unique_id("recursive.",RecId),
    create_texpr(comprehension_set(Ids,Cond),Type,[prob_annotation('SYMBOLIC')|Infos],TCompSet1),
    create_typed_id(RecId,Type,TRecId),
    create_texpr(recursive_let(TRecId,TCompSet1),Type,[],TCompSet).

unique_typed_id(Prefix,Type,TId) :-
    unique_id(Prefix,Id),
    create_typed_id(Id,Type,TId).


mark_bexpr_as_symbolic(b(E,T,I),b(E2,T,[prob_annotation('SYMBOLIC')|I]))
  :- mark_aux(E,E2).
mark_aux(union(A,B),R) :- !, R=union(SA,SB),
   mark_bexpr_as_symbolic(A,SA), mark_bexpr_as_symbolic(B,SB).
% union is currently the only operator that is kept symbolically
mark_aux(A,A).

% identifier_sub_ast(+TExpr,+Identifier,-SubPosition):
% Find occurences of an identifier "Identifier" in an expression "TExpr"
% Returns a list of positions (beginning with 0) that describes the position of a sub-expression
% that contains all occurrences of Identifier in TExpr.
% E.g. [1,0,1] means "second sub-expression of TExpr, then the first sub-expression of that,
% then the second sub-expression of that".
% When an identifier has multiple occurences, the position of the sub-expression is returned
% where all occurences of it are located.
% If the identifier is not found, the call fails.
% Sub-expressions are meant like the ones syntaxtraversion or syntaxtransformation return.
identifier_sub_ast(TExpr,Identifier,SubPosition) :-
    syntaxtraversion(TExpr,Expr,_Type,_Infos,AllSubs,Names),
    identifier_sub_ast_aux(Expr,AllSubs,Names,Identifier,SubPosition).
identifier_sub_ast_aux(identifier(Identifier),[],[],Identifier,[]) :- !.
identifier_sub_ast_aux(_Expr,AllSubs,Names,Identifier,SubPosition) :-
    get_texpr_id(TId,Identifier),nonmember(TId,Names),
    % For each sub-expression E in the list AllSubs create a term Pos-E
    % where Pos is E's position (starting with 0) in AllSubs
    foldl(annotate_pos,AllSubs,AllAnnotatedSubs,0,_),
    % Introduced identifiers occur alsa as sub-expressions, remove those
    foldl(select_sub,Names,AllAnnotatedSubs,RealAnnSubs),
    convlist(identifier_sub_ast_aux2(Identifier),RealAnnSubs,SubPositions),
    ( SubPositions = [SubPosition] -> true
    ; SubPositions = [_,_|_] -> % More then one occurrences - make this the found node
        SubPosition = []).
identifier_sub_ast_aux2(Identifier,Pos-TExpr,[Pos|SubPosition]) :-
    identifier_sub_ast(TExpr,Identifier,SubPosition).
annotate_pos(TExpr,I-TExpr,I,I2) :- I2 is I+1.
select_sub(Name,AnnotatedSubs,Result) :-
    selectchk(_Pos-Name,AnnotatedSubs,Result).

% exchange_ast_position(+SubPosition,+OldTExpr,-OldInner,+NewInner,-NewTExpr):
% exchanges a sub-expression in the expression "OldTExpr".
% SubPosition is a list of positions like identifier_sub_ast/3 returns it.
% OldInner is the sub-expression found in OldTExpr.
% NewInner is the new sub-expression.
% NewTExpr is the new expression that originates from replacing OldInner by NewInner.
exchange_ast_position([],Old,Old,New,New).
exchange_ast_position([Pos|RestPos],OldTExpr,OldInner,NewInner,NewTExpr) :-
    remove_bt(OldTExpr,OldExpr,NewExpr,NewTExpr),
    syntaxtransformation(OldExpr,Subs,_Names,NSubs,NewExpr),
    nth0(Pos,Subs, OldSelected,Rest),
    nth0(Pos,NSubs,NewSelected,Rest),
    exchange_ast_position(RestPos,OldSelected,OldInner,NewInner,NewSelected).

% -----------------------
% remove an Identifier from used_ids Info field if it exists
remove_used_id_from_info(Info,ID_to_remove,Info2) :-
   (select(used_ids(IDS),Info,Info1) ->
      (select(ID_to_remove,IDS,NewIDS) -> Info2 = [used_ids(NewIDS)|Info1]
        ; print(projecting_away_unknown_id(ID_to_remove,IDS)),nl,
          Info2 = Info)
    ; Info2 = Info).

% -----------------------

% a simple checker to see if an AST is well-formed:
% can be tested e.g. as follows: b_get_invariant_from_machine(I), check_ast(I).

check_ast(TE) :- check_ast(false,TE).
check_ast(AllowVars,TExpr) :- map_over_typed_bexpr(check_ast_texpr(AllowVars),TExpr).
check_ast(_,_).

:- use_module(typing_tools,[valid_ground_type/1]).
%check_ast_texpr(X) :- print(check(X)),nl,fail.
check_ast_texpr(AllowVars,b(E,Type,Infos)) :-
   (check_expr(E,Type,Infos) -> true ; add_error(check_ast_texpr,'Invalid Expr: ', b(E,Type,Infos))),
   (check_type(Type,AllowVars) -> true ; add_error(check_ast_texpr,'Invalid Type: ',b(E,Type,Infos))),
   (check_infos(Infos) -> true ; add_error(check_ast_texpr,'Invalid Infos: ',b(E,Type,Infos))),
   fail. % to force backtracking in map_over_typed_bexpr

check_type(X,AllowVars) :- var(X),!,
  (AllowVars==true -> true ; add_error(check_infos,'Variable type: ',X),fail).
check_type(pred,_) :- !.
check_type(subst,_) :- !.
check_type(op(Paras,Returns),AllowVars) :- !,
    maplist(check_normal_type(AllowVars),Paras),
    maplist(check_normal_type(AllowVars),Returns).
check_type(T,AllowVars) :- check_normal_type(AllowVars,T).

check_normal_type(AllowVars,T) :-
 (AllowVars \== true -> valid_ground_type(T)
   ; ground(T) -> valid_ground_type(T)
   ; true).  % TO DO: call valid_ground_type but pass AllowVars


check_infos(X) :- var(X),!, add_error(check_infos,'Info field list not terminated: ',X),fail.
check_infos([]).
check_infos([H|_]) :- \+ ground(H),!, add_error(check_infos,'Info field not ground: ',H),fail.
check_infos([cse_used_ids(H)|T]) :- member(cse_used_ids(H2),T),!,
   add_error(check_infos,'Multiple cse_used_ids entry: ',[H,H2]),fail.
check_infos([_|T]) :- check_infos(T).

check_expr(Expr,Type,Infos) :- nonmember(contains_wd_condition,Infos),
   sub_expression_contains_wd_condition(Expr),
   TE = b(Expr,Type,Infos),
   (Type = subst -> fail % AST cleanup is not called for substitutions; WD-info not available for substitutions at the moment
                    %translate:translate_substitution(TE,PS)
     ; translate:translate_bexpression(TE,PS)),
   functor(Expr,Functor,_),
   add_warning(check_expr,'Node does not contain wd condition info: ',Functor:PS,TE).
check_expr(member(LHS,RHS),Type,Infos) :- is_just_type(RHS),!,
   TE = b(member(LHS,RHS),Type,Infos),
   translate:translate_bexpression(TE,PS),
   add_warning(check_expr,'AST contains redundant typing predicate: ',PS,TE).
check_expr(identifier(ID),_,_) :- illegal_id(ID),!,
   add_error(check_infos,'Illegal identifier: ', identifier(ID)).
check_expr(lazy_lookup_expr(ID),_,_) :- illegal_id(ID),!,
   add_error(check_infos,'Illegal identifier: ', lazy_lookup_expr(ID)).
check_expr(lazy_lookup_pred(ID),_,_) :- illegal_id(ID),!,
   add_error(check_infos,'Illegal identifier: ', lazy_lookup_pred(ID)).
check_expr(exists(Parameters,Condition),pred,Infos) :- !,
   b_interpreter:get_exists_used_ids(Parameters,Condition,Infos,_Used). % this will do some checks
check_expr(_,_,_).

illegal_id(ID) :- var(ID),!.
illegal_id(op(ID)) :- !, \+ atom(ID).
illegal_id(ID) :- \+ atom(ID).
