% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(btypechecker,
          [btype/7,                % typecheck B syntax node
           btype_l/6,        % typecheck list of B syntax nodes
           btype_ground/6,   % typecheck and check if all new identifier have ground types
           btype_ground_dl/7, % typecheck and check if all new identifier have ground types
           check_ground_types_dl/4, % check if a list of identifiers has only ground types
           unify_types_werrors/5, % unify two types, return list of errors in case of type mismatch
           unify_types_werrors_l/5, % unify two list of types
           unify_types_strict/2, % unify two types, just fail in case of type mismatch

           ext2int/7, % convert a (raw) syntax element into a typed element
           remove_pos/2, % remove position info from a (raw) syntax element
           % TODO[DP, 15.03.2013]: In case of a re-structuring of modules,
           % maybe it is a good idea to move the type environment predicates to a module of its own
           env_empty/1,  % create an empty type environment
           env_lookup_type/3, % lookup a type of an identifier in the environment
           env_lookup_infos/3, % lookup additional infos of an identifier in the environment
           env_store/5, % store the type of an identifier in the environment
           env_store_operator/4, % store a theory operator into the environment
           env_identifiers/2,
           add_identifiers_to_environment/3,
           operation_infos/1, % give a list of (uninstantiated) infos that operations need for type-checking

           couplise_list/2,     % [a,b,c] -> couple(couple(a,b),c)
           prime_identifiers/2,

           fasttype/2, % manually create a syntax tree with type infos

           openenv/2,  % create an "open" type environment, where unknown identifiers will be added on demand
           openenv_identifiers/2, % get the introduced identifiers of an open type environment
           opentype/5, % type-check an expression without having every identifier declared in advance

           reset_typechecker/0,
           machine_string/1, % succeeds for every explicit string seen so far by the type checker
           get_accessed_vars/4,
           store_typecheck_error/4
          ]).

:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'ProBs type checker and type inference module.').

:- use_module(library(lists)).
:- use_module(library(terms)).
:- use_module(library(avl)).
:- use_module(library(ordsets)).

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(kernel_records,[normalise_record_type/2]).

:- use_module(translate).

:- use_module(bsyntaxtree).
:- use_module(b_ast_cleanup).
:- use_module(debug).
:- use_module(self_check).
:- use_module(pragmas).
:- use_module(input_syntax_tree,[raw_replace/3]).
:- use_module(preferences,[get_preference/2, preference/2]).

% operator for type inference rules
:- op(800,xfx,::).

%:- dynamic op_modifies_reads_cache/3. % store modifies and reads information for operations
:- dynamic machine_string/1. /* keep a list of all explicit strings seen in the machine so far */
/* Probably better: do it in a separeate traversal; talk to Daniel about this */
reset_typechecker :- reset_machine_strings. %, retractall(op_modifies_reads_cache(_,_,_)).
reset_machine_strings :- retractall(machine_string(_)).
assert_machine_string(S) :- machine_string(S) -> true ; assert(machine_string(S)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% miniature B type checker

btype_ground(Exprs,Env,NonGroundExceptions,Types,TExprs,Errors) :-
    btype_ground_dl(Exprs,Env,NonGroundExceptions,Types,TExprs,Errors,[]).
btype_ground_dl(Exprs,Env,NonGroundExceptions,Types,TExprs,EIn,EOut) :-
    ~~mnf(same_length(Exprs,Types)), ~~mnf(same_length(Exprs,TExprs)),
    ~~mnf(get_texpr_types(TExprs,Types)),
    ~~mnf(do_typecheck_dl(Exprs,Env,[ground(NonGroundExceptions)],TExprs,EIn,EOut)).

% do_typecheck type-checks the expression and checks if the types
% of introduced variables (like in quantifiers) are ground
do_typecheck(Exprs,Env,Options,TExprs,Errors) :-
    do_typecheck_dl(Exprs,Env,Options,TExprs,Errors,[]).

do_typecheck_dl(Exprs,Env,Options,TExprs,ErrorsIn,ErrorsOut) :-
    typecheck_result_init(ErrorsIn,Delayed,Identifiers,TRStart),
    % type-check the expressions
    btype_l(Exprs,Env,_Type,TExprs,TRStart,TR),
    typecheck_result_get_errors(TR,E1),
    typecheck_result_get_delayed_rules(TR,[]),
    typecheck_result_get_identifiers(TR,[]),
    foldl(trigger_delayed_rule,Delayed,E1,E2),
    % get the newly introduced identifiers and check if their types are
    % ground
    ( memberchk(ground(NonGroundExceptions),Options) ->
        check_ground_types_dl(Identifiers,NonGroundExceptions,E2,ErrorsOut)
    ; otherwise ->
        E2=ErrorsOut).

trigger_delayed_rule(dr(Trigger,Errors),Ein,Eout) :-
    % ground trigger, enforce blocked rule
    Trigger = 0,
    % by now Errors should have given a ground value by the co-routine
    add_all(Errors,Ein,Eout).
add_all([]) --> !.
add_all([E|Rest]) --> [E], add_all(Rest).

typecheck_result_init(Errors,DR,Ids,tr(Errors,DR,Ids)).
typecheck_result_add_error(Error,tr([Error|Errors],DR,Ids),tr(Errors,DR,Ids)).
typecheck_result_add_delayed_rule(Trigger,RuleErrors,tr(Errors,[dr(Trigger,RuleErrors)|Rules],Ids),tr(Errors,Rules,Ids)).
typecheck_result_add_identifier(Id,tr(Errors,Rules,[Id|Ids]),tr(Errors,Rules,Ids)).
typecheck_result_get_errors(tr(Errors,_DR,_Ids),Errors).
typecheck_result_get_delayed_rules(tr(_Errors,Rules,_Ids),Rules).
typecheck_result_get_identifiers(tr(_Errors,_Rules,Ids),Ids).
%typecheck_result_finalize(tr([],[],[])). % not used at the moment

% btype/7 is the central predicate of the type-checker
% type-checking is done in three "stages":
%  btype/5:   extracts the obligatory position information from the input syntax element
%                decouples the type in the argument and the type of the expression by
%                  an explicit unify_types/5 call
%                calls btype1
%  btype1/7:  calls btype_rewrite for optional rewriting rules
%                if no more rewriting rules can be applied, btype2 is called
%  btype2/7:  does the actual type-checking

% btype(+Src,+Ext,+Env,Type,-Result,TRin,TRout):
%  Src:    Where the call comes from
%  Ext:    The AST of the expression to type check as it comes from the parser
%  Env:    The type environment (see the env_... predicates)
%  Type:   The type of the expression (could be as input or output)
%  Result: The type checked AST
%  TRin/TRout: typecheck_result data structure, see typecheck_result_... predicates
btype(Src,Ext,Env,Type,Result,TRin,TRout) :-
    (var(Ext)
      -> add_internal_error('btype Ext is variable: ',Src:Ext)
       ; true),
    if(btype_aux(Ext,Env,Type,Result,TRin,TRout),true,
       add_error_and_fail(btype,'btype failed: ',Src:Ext)),
    (var(TRout) -> add_error(btype,'TRout is variable: ',Src:Ext) ; true).

btype_aux(Ext,Env,Type,Result,TRin,TRout) :-
    % convert the external representation Ext (first argument is the position)
    % into the typed representation
    ext2int(btype,Ext,Expr,Pos,Type,TExpr,Infos,Result),
    % type-check the expression
    btype1(Expr,Pos,Env,TExpr,LType,Infos1,TRin,TR1),
    % an explicit unify decouples the expected and inferred type
    unify_types(Type,LType,Pos,TExpr,TR1,TRout),
    % check if a pragma is defined for this position and add an
    % appropiate info field
    apply_pragmas(Pos,TExpr,Type,Infos1,Infos).


% btype1/7 tries to apply rewriting rules until no more rules
% are applyable, then btype2/7 is called
% btype1(+Expr,+Pos,+Env,-TExpr,-Type,-Infos,+TRin,-TRout):
btype1(Expr,Pos,Env,TExpr,Type,Infos,TRin,TRout) :-
    catch( btype_try_rewrite(Expr,Pos,Env,TExpr,Type,Infos,TRin,TRout),
           E,
           handle_rewrite_exception(E,Pos,TExpr,Infos,TRin,TRout)
         ).
btype_try_rewrite(Expr,Pos,Env,TExpr,Type,Infos,TRin,TRout) :-
    btype_rewrite(Expr,Env,Pos,NewExpr),!,
    btype1(NewExpr,Pos,Env,TExpr,Type,Infos,TRin,TRout).
btype_try_rewrite(Expr,Pos,Env,TExpr,Type,Infos,TRin,TRout) :-
    btype2(Expr,Pos,Env,TExpr,Type,Infos,TRin,TRout).

handle_rewrite_exception(rewrite_exception(Msg,Name),Pos,TExpr,Infos,TRin,TRout) :- !,
    store_typecheck_error(Msg,Pos,TRin,TRout),
    % Type remains unbound, Infos are empty
    Infos = [],
    TExpr = identifier(Name).
handle_rewrite_exception(E,_Pos,_TExpr,_Infos,_TRin,_TRout) :- throw(E).

% btype2/8 encodes the type checking rules
btype2(identifier(I1), Pos, Env, identifier(I), Type, Infos, TRin, TRout) :-
    !, % In case the identifier is a list, join it with '.' (e.g. ['M',x] -> 'M.x')
    ( is_list(I1) -> ajoin_with_sep(I1,'.',I)
    ; otherwise   -> I1=I),
    % get the identifier's type from the type environment
    lookup_type(I,Env,Pos,Type,TRin,TR1),
    % copy also additional information from the environment,
    lookup_infos(I,Env,AllInfos),
    % but not all (e.g. no position information)
    extract_id_information(AllInfos,Infos),
    % an access to this identifier might be illegal, check this
    ( member(error(Error),AllInfos) -> store_typecheck_error(Error, Pos, TR1, TRout)
    ; otherwise -> TR1=TRout).
btype2(primed_identifier(Id,N), Pos, Env, identifier(FullId), Type, Infos, TRin, TRout) :-
    !, % a primed identifier is of the form x$0, we just convert it to
    % a normal identifier called "x$0" and type-check it.
    number_suffix(Id,N,FullId),
    btype(primed_identifier(Id,N),identifier(Pos,FullId),Env,Type,TId,TRin,TRout),
    get_texpr_info(TId,Infos).
btype2(forall(Ids,EImplication), _, Env, forall(TIdentifiers,TDomain,TPred), pred, [], TRin, TRout) :-
    !,add_ext_variables_with_info(Ids,Env,[introduced_by(forall)],TIdentifiers,Subenv,TRin,TR1),
    ext2int(btype2,EImplication,Implication,IPos,_,_,_,_),
    % we expect an implication as predicate, if not, generate an error
    ( Implication=implication(Domain,Pred) ->
        btype(forall_domain(Ids),Domain,Subenv,pred,TDomain,TR1,TR2),
        btype(forall_pred(Ids),Pred,Subenv,pred,TPred,TR2,TRout)
    ; otherwise ->
        store_typecheck_error('expected implication in universal quantification ',IPos,TR1,TR2),
        create_texpr(truth,pred,[],TDomain),
        btype(foralle_impl(Ids),EImplication,Subenv,pred,TPred,TR2,TRout)
    ).
btype2(external_function_call(FunName,FunParams,_FunDef,TypeParams,Decl), Pos, Env,
       external_function_call(FunName,TFunParams), OutputType , [], TRin, TRout) :-
       /* note: position info is automatically added to info field; hence [] ok */
       (FunParams = []  -> InputType = empty_type, FullType = set(OutputType)
         ; FullType = set(set(couple(InputType,OutputType)))),
      !, btype_external_call(FunName,TypeParams,FunParams,Decl,FullType,
                           InputType,Pos,Env,TFunParams,TRin,TRout).
btype2(external_pred_call(FunName,FunParams,_FunDef,TypeParams,Decl), Pos, Env,
       external_pred_call(FunName,TFunParams), pred, [], TRin, TRout) :-
    !, btype_external_call(FunName,TypeParams,FunParams,Decl,set(InputType),InputType,
                           Pos,Env,TFunParams,TRin,TRout).
btype2(external_subst_call(FunName,FunParams,_FunDef,TypeParams,Decl), Pos, Env,
       external_subst_call(FunName,TFunParams), subst, [], TRin, TRout) :-
    !, btype_external_call(FunName,TypeParams,FunParams,Decl,set(InputType),InputType,
                           Pos,Env,TFunParams,TRin,TRout).
btype2(integer(I), _, _, integer(I), integer, [], TR, TR) :- !.
btype2(integer_set(S), _, _, integer_set(S), set(integer), [], TR, TR) :- !.
btype2(string(I), _, _, string(I), string, [], TR, TR) :- !, assert_machine_string(I).
btype2(assign(IdFuns,Exprs), Pos, Env, assign(TIdFuns,TExprs), subst, [], TRin, TRout) :-
    !,
    ( same_length(IdFuns,Exprs) ->
        btype_l(IdFuns,Env,Types,TIdFuns,TRin,TR1),
        btype_l(Exprs, Env,Types,TExprs, TR1, TR2),
        foldl(check_assign_lhs(Env),TIdFuns,TR2,TRout)
    ; otherwise ->
        store_typecheck_error('different number of variables on left and right side of a substitution ',Pos,TRin,TRout)
    ).
btype2(becomes_element_of(Ids,Expr), _, Env, becomes_element_of(TIds,TExpr), subst, [], TRin, TRout) :-
    !,btype_l(Ids,Env,Types,TIds,TRin,TR1),
    foldl(check_assign_lhs(Env),TIds,TR1,TR2),
    couplise_list(Types,Type),
    btype(becomes_element_of(Ids),Expr,Env,set(Type),TExpr,TR2, TRout).
btype2(becomes_such(Ids,Pred), _, Env, becomes_such(TIds,TPred), subst, [], TRin, TRout) :-
    !,btype_l(Ids,Env,_,TIds,TRin,TR1),
    foldl(check_assign_lhs(Env),TIds,TR1,TR2),
    % for each LHS identifier x (in Ids), we add an identifier of the form
    % x$0 with the same type to the environment
    foldl(add_primed_id,TIds,Env,Subenv),
    btype(becomes_such,Pred,Subenv,pred,TPred,TR2,TRout).
btype2(evb_becomes_such(Ids,Pred), _, Env, evb_becomes_such(TIds,TPred), subst, [], TRin, TRout) :-
    !,check_evb_becomes_such(evb_becomes_such,Ids,Pred,Env,TIds,TPred,TRin,TRout).
btype2(evb2_becomes_such(Ids,Pred), _, Env, evb2_becomes_such(TIds,TPred), subst, [], TRin, TRout) :-
    !,check_evb_becomes_such(evb2_becomes_such,Ids,Pred,Env,TIds,TPred,TRin,TRout).
btype2(operation_call(Op,Res,Args), Pos, Env, TOp, subst, Infos, TRin, TRout) :-
    !,type_operation_call(Op,Res,Args,Pos,Env,TOp,Infos,TRin,TRout).
btype2(case(Expr,Eithers,Else), _, Env, case(TExpr,TEithers,TElse), subst, [], TRin, TRout) :-
    !,btype(case,Expr,Env,Type,TExpr,TRin,TR1),
    foldl(btype_caseor(Env,Type),Eithers,TEithers,LExprs,TR1,TR2),
    append(LExprs,Exprs),
    btype(case_else,Else,Env,subst,TElse,TR2,TR3),
    % The next check could be made dependend on a preference (non-literal values allowed in
    % CASE statements), in that case add TR3=TRout as an alternative
    check_case_expressions(Exprs,TR3,TRout).
btype2(struct(Fields), Pos, Env, struct(TRecord), set(RType), [], TRin, TRout) :-
    !, btype(struct,rec(Pos,Fields),Env,_,TRecord,TRin,TR1),
    get_texpr_expr(TRecord,rec(TFields)),
    foldl(extract_field_type,TFields,FieldTypes,TR1,TRout),
    normalise_record_type(record(FieldTypes),RType).
btype2(rec(Fields), _, Env, rec(TFields), RType, [], TRin, TRout) :-
    !, foldl(btype_field_rec(Env),Fields,FTypes,TFields1,TRin,TRout),
    normalise_record_type(record(FTypes),RType),
    % we sort the field arguments here, so we do not have to normalise them every
    % time in the interpreter
    RType = record(NormedFieldTypes),
    sort_record_fields(NormedFieldTypes,TFields1,TFields).
btype2(record_field(Record,I), Pos, Env, record_field(TRecord,Id), Type, [], TRin, TRout) :-
    !,btype(record_field,Record, Env, RType, TRecord, TRin, TR1),
    remove_pos(I,identifier(Id)),
    unify_types(record([field(Id,Type)|_]),RType,Pos,TRecord,TR1,TRout).
btype2(operation(Id,Results,Args,Body), Pos, Env, operation(TId,TResults,TArgs,TBody), Type, [modifies(Modifies),reads(Reads)], TRin, TRout) :-
    !, %print(checking_operation(Id)),nl,
    % type-check the identifier to look up the operation's type
    % in the environment.
    % we call ext2int and btype2/7 directly instead of going over
    % btype/5, because we use op(Name) instead of Name to look up
    % the type. This is done to seperate operation and variable name
    % spaces (Note: that was a stupid and ugly hack)
    ext2int(btype2_operation,Id,identifier(Opid),IPos,Type,Topid,IdInfos,TId),
    btype2(identifier(op(Opid)),IPos,Env,Topid,Type,IdInfos,TRin,TR1),
    % introduce a sub-environment Subenv by adding the argument and
    % result variables
    add_ext_variables_with_info(Results,Env,[introduced_by(operation)],TResults,Subenv1,TR1,TR2),
    add_ext_variables_with_info(Args,Subenv1,[readonly,introduced_by(operation)],TArgs,Subenv,TR2,TR3),
    % get their types and determine the resulting operation's type
    get_texpr_types(TResults,LResultTypes),
    get_texpr_types(TArgs,LArgTypes),
    unify_types(op(LArgTypes,LResultTypes),Type,Pos,TId,TR3,TR4),
    % type the operation's body in the sub-environment
    btype(operation(Id),Body, Subenv, subst, TBody, TR4, TRout),
    % the list of modified machine variables is the list of changed
    % variables by the substitutions in the body minus the result variables
    append(TResults,TArgs,TLocals),
    get_texpr_ids(TLocals,Locals),
    get_accessed_vars(TBody,Locals,Modifies,Reads),
    %print(modified_reads(Opid,Modifies,Reads)),nl,
    %assert(op_modifies_reads_cache(Opid,Modifies,Reads)), % cache information for operation_calls; possibly temporary solution until we understand the code for operation_call [Note: we can have multiple versions ! at different refinement levels !!]
    % make sure that the List of modified variables of this operation
    % is also up-to-date in the type environment
    memberchk(modifies(EnvModifies),IdInfos),
    memberchk(reads(EnvReads),IdInfos),
    % Check if the environment variables and the local variables are unifiable.
    % If not, an operation with the same name must have been defined before.
    % We can ignore this, later an error message will be raised.
    ( EnvModifies=Modifies, EnvReads=Reads -> true; true).
btype2(while(Pred,Body,Inv,Var), _, Env, while(TPred,TBody,TInv,TVar), subst, FullInfo, TRin, TRout) :- !,
    btype(while,Pred,Env,pred,TPred,TRin,TR1),
    btype(while_body,Body,Env,subst,TBody,TR1,TR2),
    % In contrast to the other substitutions, in the invariant and variant
    % part of a while loop it is possible to reference abstract variables,
    % as they are not used to influence the behaviour of the substitution.
    % We create a new environment InvEnv where those access restrictions are
    % removed.
    allow_access_to_abstract_vars(Env,InvEnv),
    add_primed_old_value_variables(InvEnv,PInvEnv),
    btype(while_inv,Inv,PInvEnv,pred,TInv,TR2,TR3),
    btype(while_variant,Var,InvEnv,integer,TVar,TR3,TRout),
    % tag this statement with an appropiate info if a reference to the original value
    % of a variable is made in the invariant.
    % This info is used in the clean_up (see b_ast_cleanup) to insert a LET statement.
    check_for_old_state_references(TInv,Info),
    FullInfo = [modifies(Modifies),reads(AllReads)|Info],
    % TO DO: ideally we should compute those in one go for all substitutions:
    %  here we compute information again which we have already computed for operations
    get_accessed_vars(b(while(TPred,TBody,TInv,TVar),subst,Info),[],Modifies,Reads),
    (member(refers_to_old_state(SortedRefs),Info)
      -> findall(Primed,member(oref(Primed,_,_),SortedRefs),PrimedDollarVars), sort(PrimedDollarVars,SPDV),
         ord_union(SPDV,Reads,AllReads)  %, print(add(SPDV,Reads)),nl
      ;  AllReads = Reads).

btype2(parallel(Substs),Pos,Env,parallel(TSubsts), subst, [], TRin, TRout) :-
    !, btype_l(Substs,Env,_,TSubsts,TRin,TR1),
    check_no_double_assignment(TSubsts,Pos,TR1,TRout).
btype2(minus_or_set_subtract(A,B), Pos, Env, minus_or_set_subtract(TA,TB), Type, [], TRin, TRout) :-
    !,btype_l([A,B],Env,[TypeA,TypeB],[TA,TB],TRin,TR1),
    delayed_type_minus(TypeA,TypeB,Type,Pos,A,B,TR1,TRout).
btype2(mult_or_cart(A,B), Pos, Env, mult_or_cart(TA,TB), Type, [], TRin, TRout) :-
    !,btype_l([A,B],Env,[TypeA,TypeB],[TA,TB],TRin,TR1),
    delayed_type_times(TypeA,TypeB,Type,Pos,A,B,TR1,TRout).
btype2(substitution(Subst,Pred), _Pos, Env, TSubst, pred, [], TRin, TRout) :-
    !, type_substitution_expression(Subst,Pred,Env,TSubst,TRin,TRout).
btype2(extended_expr(Id,Args,Args2), Pos, Env, Result, Type, [was_operator(Id)|Info], TRin, TRout) :-
    !, lookup_eventb_operator(Id,Args,Args2,Pos,Env,expr,Result,Type,Info,TRin,TRout).
btype2(extended_pred(Id,Args,Args2), Pos, Env, Result, Type, [was_operator(Id)|Info], TRin, TRout) :-
    !, lookup_eventb_operator(Id,Args,Args2,Pos,Env,pred,Result,Type,Info,TRin,TRout).
btype2(recursive_let(Id,Set),_Pos,Env,recursive(TId,TSet),set(Type),[],TRin,TRout) :- !,
    add_ext_variables([Id], Env, [TId], SubEnv, TRin, TR1),
    get_texpr_type(TId,set(Type)),
    btype(recursive,Set,SubEnv,set(Type),TSet,TR1,TRout).
btype2(set_extension(RawIntegers),_Pos,_Env,value(V),set(Type),[],TR,TR) :-
    % optimisation: is this a set of integers or pair of integers?
    % -> skip type checking the content and just generate an AVL set directly
    simple_set_extension_type(RawIntegers,Type),
    maplist(read_raw_integers(Type),RawIntegers,Integers),!,
    custom_explicit_sets:convert_to_avl(Integers,V).
btype2(typeset,_Pos,_Env,typeset,set(_),[],TR,TR) :- !.
btype2(typeof(InnerExpr,TypeExpr),_Pos,Env,Expr,Type,Info,TRin,TRout) :-
    !, btype(typeof,TypeExpr,Env,set(Type),_TTypeExpr,TRin,TR1),
    btype(typeof,InnerExpr,Env,Type,TExpr,TR1,TRout),
    get_texpr_expr(TExpr,Expr),get_texpr_info(TExpr,Info).
btype2(Expr, Pos, Env, TExpr, Type, [], TRin, TRout) :-
    % this is the most simple case of a type-check
    safe_functor(btype2_expr,Expr,Functor,Arity),
    safe_functor(btype2_lookup,Lookup,Functor,Arity),
    Lookup :: Type,!,
    safe_functor(btype2_texpr,TExpr,Functor,Arity),
    type_and_unify_args(1,Arity,Pos,Env,Expr,Lookup,TExpr,TRin,TRout).
btype2(Expr, _Pos, _, _, _, _, TR, TR) :-
    safe_functor(btype2_uncovered,Expr,F,Arity),
    add_internal_error('Uncovered operation in type checker: ',F/Arity),
    fail.

% check if a set extension is of type integer or couple(integer,...) by inspecting
% the first element
simple_set_extension_type([First,_|_],Type) :- % At least two elements
    simple_set_extension_type2(First,Type).    % of type integer or couple of integers
simple_set_extension_type2(integer(_,_),integer).
simple_set_extension_type2(couple(Pos,List),Type) :-
    couplise_list_pos(List,Pos,Couple),
    simple_set_extension_type2(Couple,Type).
simple_set_extension_type2(couple(_Pos,A,B),couple(TA,TB)) :-
    simple_set_extension_type2(A,TA),
    simple_set_extension_type2(B,TB).
% convert a raw integer or pairs of integers into a value
read_raw_integers(integer,integer(_Pos,I),int(I)).
read_raw_integers(couple(TA,TB),couple(_Pos,List),(VA,VB)) :-
    read_raw_integer_couple(TA,VA,List,Rest),
    read_raw_integer_couple(TB,VB,Rest,[]).
read_raw_integer_couple(integer,int(I)) --> [integer(_Pos,I)].
read_raw_integer_couple(couple(TA,TB),(VA,VB)) -->
    read_raw_integer_couple(TA,VA),
    read_raw_integer_couple(TB,VB).


check_no_double_assignment(TSubsts,Pos,TRin,TRout) :-
    maplist(modified_var,TSubsts,Mods),
    ( (preferences:preference(allow_simultaneous_assignments,false),
       find_common_variables(Mods,Var)) ->
        ajoin(['Simultaneous assignment to variable ',Var],Msg),
        store_typecheck_error(Msg,Pos,TRin,TRout)
    ; otherwise ->
        TRin = TRout).
modified_var(TSubst,Modified) :-
    get_accessed_vars(TSubst,[],Modified,_Read).
find_common_variables([Vars|Rest],Var) :-
    find_common_variables2(Rest,Vars,Var).
find_common_variables2([Vars|Vrest],Prior,Var) :-
    ord_intersection(Vars,Prior,FoundVars),
    ( memberchk(Var,FoundVars) -> true
    ; otherwise ->
        ord_union(Vars,Prior,NewVars),
        find_common_variables2(Vrest,NewVars,Var)).

type_substitution_expression(Subst,Pred,Env,
                             let_predicate(TypedLhs,TypedRhs,TPred),TRin,TRout) :-
    ( remove_pos(Subst,assign(Lhs,Rhs)) ->
        ( same_length(Lhs,Rhs) ->
            /* Typecheck the right side of the substitution */
            btype_l(Rhs,Env,Types,TypedRhs,TRin,TR1),
            /* The left site of the substitution introduces new variables which have the
               same type as the corrresponding element of the RHS */
            add_ext_variables_with_info(Lhs,Env,[introduced_by(let)],TypedLhs,SubEnv,TR1,TR2),
            get_texpr_types(TypedLhs,Types),
            /* Type the predicates in the sub-environment */
            btype(substitution_expression,Pred,SubEnv,pred,TPred,TR2,TRout)
        ; otherwise /* LHS hat not the same number of elements as RHS */ ->
            ext2int(Subst,_,Pos,_,_,_,_),
            TypedLhs=[], TypedRhs=[], create_texpr(truth,pred,[],TPred),
            store_typecheck_error('Expected same number of elements on both sides of an assignment',
                        Pos,TRin,TRout))
    ; otherwise /* Substitution is no assignment (the parser should not allow this) */ ->
        add_internal_error('Expected assignment in substitution expression, but was: ',Subst),
        fail).

delayed_type_minus(TypeA,TypeB,ResType,Pos,ExprA,ExprB,TRin,TRout) :-
    arg(1,ExprA,PosA), arg(1,ExprB,PosB),
    % In case of A-B=C we know that A, B and C have the same type
    unify_types_l([ResType,ResType],[TypeA,TypeB],[PosA,PosB],TRin,TR1),
    % but not if it is integer or set(_)
    typecheck_result_add_delayed_rule(Trigger,Errors,TR1,TRout),
    delayed_type_minus2(ResType,Trigger,Pos,Errors).
:- block delayed_type_minus2(-,-,?,?).
delayed_type_minus2(Type,_Trigger,Pos,Errors) :-
    ( var(Type) ->
        Errors = [error('Ambiguous expression (arithmetic minus or set minus?)',Pos)]
    ; Type = integer ->
        Errors = []
    ; otherwise ->
        unify_types_werrors_l([set(_)],[Type],[Pos],Errors,[])
    ).

% delayed_type_times(TypeA,TypeB,ResType,Pos,ExprA,ExprB,TRin,TRout):
%   "A*B=R"
%   TypeA:   type of A, TypeB: type of B, ResType: type of R
%   Pos:     position info for error messages
%   ExprA/B: expressions to generate pretty error messages
delayed_type_times(TypeA,TypeB,ResType,Pos,ExprA,ExprB,TRin,TRout) :-
    arg(1,ExprA,PosA), arg(1,ExprB,PosB),
    ( maplist(var,[TypeA,TypeB,ResType]) ->
        typecheck_result_add_delayed_rule(Trigger,Errors,TRin,TRout),
        delayed_type_times2(TypeA,TypeB,ResType,Trigger,Pos,PosA,PosB,Errors)
    ; contains_integer_type([TypeA,TypeB,ResType]) ->
        % One type is an integer
        unify_types_l([integer,integer,integer],[ResType,TypeA,TypeB],[Pos,PosA,PosB],TRin,TRout)
    ; otherwise ->
        % A type is known, no need for a delay
        unify_types_l([set(couple(X,Y)),set(X),set(Y)],[ResType,TypeA,TypeB],[Pos,PosA,PosB],TRin,TRout)
    ).
:- block delayed_type_times2(-,-,-,-,?,?,?,?).
delayed_type_times2(A,B,Res,_Trigger,Pos,PosA,PosB,Errors) :-
    ( maplist(var,[A,B,Res]) ->
        Errors = [error('Ambiguous expression (multiplication or cartesian product?)',Pos)]
    ; contains_integer_type([A,B,Res]) ->
        unify_types_werrors_l([integer,integer,integer],[Res,A,B],[Pos,PosA,PosB],Errors,[])
    ; otherwise ->
        unify_types_werrors_l([set(couple(X,Y)),set(X),set(Y)],[Res,A,B],[Pos,PosA,PosB],Errors,[])
    ).

contains_integer_type([T|_Rest]) :- T==integer,!.
contains_integer_type([_T|Rest]) :- contains_integer_type(Rest).

type_operation_call(Op,Res,Args,Pos,Env,operation_call(TOp,TRes,TArgs),Infos,TRin,TRout) :-
    ext2int(btype2_opcall,Op,identifier(Opid),IPos,_,_,_,_),
    btype2(identifier(op(Opid)),IPos,Env,identifier(OpId),op(ArgTypes1,ResTypes1),OpInfos,TRin,TR1),
    create_texpr(identifier(OpId),subst,OpInfos,TOp),
    btype_l(Args,Env,_ArgTypes,TArgs,TR1,TR2),
    btype_l(Res,Env,_ResTypes,TRes,TR2,TR3),
    foldl(check_assign_lhs(Env),TRes,TR3,TR4),
    ( same_length(Args,ArgTypes1) ->
        ( same_length(Res,ResTypes1) ->
            % number of arguments resp. results ok
            type_operation_call2(OpId,Env,Pos,TRes,TArgs,ArgTypes1,ResTypes1,Infos,TR4,TRout)
        ; otherwise ->
            % wrong number of result identifiers given
            add_operation_call_error(ResTypes1,Res,Pos,Opid,'result identifiers(s)',Infos,TR4,TRout))
    ; otherwise ->
        % wrong number of arguments given
        add_operation_call_error(ArgTypes1,Args,Pos,Opid,'arguments(s)',Infos,TR4,TRout)).
type_operation_call2(OpId,Env,Pos,TRes,TArgs,ArgTypes,ResTypes,Infos,TRin,TRout) :-
    check_arguments(TArgs,ArgTypes,TRin,TR1), % check the arguments
    check_arguments(TRes,ResTypes,TR1,TR2),   % check the result

    % if the called operation is not found in the environment,
    % an error was stored above, and we skip the following section
    ( env_lookup_infos(OpId,Env,EnvInfos) ->
        % we currently return the modifies/1 information of the called operation
        % the only information about this node. This information is needed by
        % get_modified_vars/3.
        get_operation_infos(EnvInfos,Infos),
        % check if we are in a section where we can only call operations
        % that make no changes to the state ("inquiry")
        ( memberchk(dontcall,EnvInfos) ->
            % it is not allowed do call the operation: this is the case if a local
            % operation is used (only operations of included/used/seen machines are allowed to call)
            % TO DO: we are allowed to call LOCAL_OPERATIONS from other OPERATIONS
            store_typecheck_error('a local operation must not be called',Pos,TR2,TRout)
        ; memberchk(inquiry,EnvInfos) ->
            % check whether we call a query-operation, if the target it marked "inquiry only"
            ( memberchk(modifies([]),Infos) -> TR2=TRout
            ; otherwise -> store_typecheck_error('query-operation expected ',Pos,TR2,TRout))
        ; otherwise ->
            % ok, we may call even state-changing operations
            TR2=TRout
        )
    ; otherwise ->
        % operation not found, skip additional checks above and
        % state (maybe incorrectly, but errors are reported anyway),
        % that the operation call does not modify any variables
        Infos = [modifies([]),reads([])],
        TR2=TRout).
check_arguments(TArgs,ArgTypes1,TRin,TRout) :-
    get_texpr_types(TArgs,ArgTypes), maplist(get_texpr_pos,TArgs,ArgPos),
    unify_types_l(ArgTypes1,ArgTypes,ArgPos,TRin,TRout).

% Generator an error for the wrong numbers of arguments resp. result identifiers
add_operation_call_error(ExpTypes,GivenExpr,Pos,Opid,Title,Infos,TRin,TRout) :-
    Infos = [modifies([]),reads([])],     % Info field must be set, even in case of an error
    length(ExpTypes,Expected),length(GivenExpr,Given),
    ajoin(['Expected ',Expected,' ',Title,' for operation ',Opid,
           ', but ',Given,' ',Title,' given'],Msg),
    store_typecheck_error(Msg,Pos,TRin,TRout).


% common typing patterns:
% type a list of expressions, return the list of types and typed expressions
btype_l([],_Env,[],[],TR,TR).
btype_l([Expr|Rest],Env,[Type|TRest],[TExpr|TERest],TRin,TRout) :-
    btype(btype_l,Expr,Env,Type,TExpr,TRin,TR1),
    btype_l(Rest,Env,TRest,TERest,TR1,TRout).

% type a list of expressions whose type is the same
btype_same([],_Env,_Type,[],TR,TR).
btype_same([Expr|ERest],Env,Type,[TExpr|TERest],TRin,TRout) :-
    btype(btype_same,Expr, Env, ElemType, TExpr, TRin, TR1),
    get_texpr_pos(TExpr,Pos),
    unify_types(Type,ElemType,Pos,TExpr,TR1,TR2),
    btype_same(ERest,Env,Type,TERest,TR2,TRout).

% for a list of identifiers, create a subenvironment with
% their primed versions. I.e., for each identifier "x", introduce
% a variable "x$0" into the environment.
% The type and additional information is copied from the original
% identifier, aditionally information about what the original
% identifier was ist stored.
add_primed_id(TId,InEnv,SubEnv) :-
    def_get_texpr_id(TId,Id),
    get_texpr_type(TId,Type),
    get_texpr_info(TId,Infos),
    number_suffix(Id,0,FullId),
    % the additional information is used in the interpreter to store the
    % values before the substitution under the alternative names
    env_store(FullId,Type,[before_substitution(Id,FullId)|Infos],InEnv,SubEnv).

% for an identifier "x" and a number N, create a new identifier "x$N"
number_suffix(Id,N,FullId) :-
    safe_atom_chars(Id,IdChars,number_suffix1),number_chars(N,NChars),
    append(IdChars,['$'|NChars],FullIdChars),
    safe_atom_chars(FullId,FullIdChars,number_suffix2).

% special cases

% TODO(DP,2014-05-13): I have no idea why there are two event-b becomes such assignments
check_evb_becomes_such(Label,Ids,Pred,Env,TIds,TPred,TRin,TRout) :-
    btype_l(Ids,Env,_,TIds,TRin,TR1),
    foldl(check_assign_lhs(Env),TIds,TR1,TR2),
    prime_identifiers(TIds,TPrimed),
    add_identifiers_to_environment(TPrimed,Env,Subenv),
    btype(Label,Pred,Subenv,pred,TPred,TR2,TRout).

% CASE ... EITHER .. OR ... END
btype_caseor(Env,Type,CaseOr,TCaseOr,TExprs,TRin,TRout) :-
    ext2int(btype_caseor,CaseOr,case_or(Exprs,Subst),_,pred,case_or(TExprs,TSubst),[],TCaseOr),
    % a case expresiion can have multiple expressions that may match, all of the same type "Type"
    btype_same(Exprs,Env,Type,TExprs,TRin,TR1),
    btype(btype_caseor,Subst,Env,subst,TSubst,TR1,TRout).
    % TO DO: check that Field1 of case_or(Field1,_) is a literal value
check_case_expressions(Exprs,TRin,TRout) :-
    maplist(is_literal,Exprs),!,  % only literals allowed
    ( contains_duplicate_literal(Exprs,Duplicate) ->
        translate_bexpression(Duplicate,DStr),
        get_texpr_pos(Duplicate,Pos),
        ajoin(['Duplicate case in CASE statement: ',DStr],Msg),
        store_typecheck_error(Msg,Pos,TRin,TRout)
    ; otherwise ->              % ok, all literals, no duplicates
        TRin=TRout
    ).
check_case_expressions(Exprs,TRin,TRout) :-
    % there are elements that are not literals
    exclude(is_literal,Exprs,NonLiterals),
    foldl(non_literal_error,NonLiterals,TRin,TRout).
is_literal(TExpr) :-
    get_texpr_expr(TExpr,Expr),
    get_texpr_info(TExpr,Info),
    is_literal2(Expr,Info).
is_literal2(identifier(_),Info) :-
    memberchk(enumerated_set_element,Info).
is_literal2(integer(_),_Info).
is_literal2(boolean_true,_Info).
is_literal2(boolean_false,_Info).
is_literal2(string(_),_Info).
non_literal_error(TExpr,TRin,TRout) :-
    translate_bexpression(TExpr,Str),
    get_texpr_pos(TExpr,Pos),
    ajoin(['CASE argument \"',Str,'\" is not a literal value.'],Msg),
    store_typecheck_error(Msg,Pos,TRin,TRout).
contains_duplicate_literal(TExprs,Duplicate) :-
    % This can be done, because we know that TExprs are literals, so there are no
    % sub-expressions with positions information or similar problems
    append(_Prefix,[Dup|Rest],TExprs),
    get_texpr_expr(Dup,Expr),
    get_texpr_expr(Duplicate,Expr),
    memberchk(Duplicate,Rest),!.

btype_field_rec(Env,Ext,field(Id,Type),field(Id,TExpr),TRin,TRout) :-
    remove_pos(Ext,rec_entry(Identifier,Expr)),
    remove_pos(Identifier,identifier(Id)),
    btype(field,Expr,Env,Type,TExpr,TRin,TRout).

extract_field_type(field(Id,TExpr),field(Id,Type),TRin,TRout) :-
    get_texpr_type(TExpr,SetType),
    get_texpr_pos(TExpr,Pos),
    unify_types(set(Type),SetType,Pos,TExpr,TRin,TRout).

sort_record_fields([],_,[]).
sort_record_fields([field(Id,_Type)|Trest],Fields,[Field|Frest]) :-
    Field = field(Id,_Value),
    member(Field,Fields),!,
    sort_record_fields(Trest,Fields,Frest).

lookup_eventb_operator(Id,Args1,Args2,Pos,Env,ExPr,Result,Type,Info,TRin,TRout) :-
    ( env_lookup_infos(operator(Id),Env,[Module:Callback]) ->
        append(Args1,Args2,Args), % do not distinguish between expressions and predicates
        btype_l(Args,Env,_Types,TArgs,TRin,TR1),
        call(Module:Callback,Id,TArgs,Pos,Env,ExPr,TExpr,TR1,TRout),
        create_texpr(Result,Type,Info,TExpr)
    ; otherwise ->
        ajoin(['Unknown operator identifier ',Id],Msg),
        store_typecheck_error(Msg,Pos,TRin,TRout),
        failure_syntax_element(ExPr,Id,Result,Type)).
% TODO[DP,2013-01-31]: copied from bmachine_eventb -> refactor!
failure_syntax_element(expr,Id,identifier(Id),_).
failure_syntax_element(pred,_Id,falsity,pred).

get_operation_infos(Info,OpInfos) :-
    operation_infos(OpInfos),
    get_operation_infos2(OpInfos,Info).
get_operation_infos2([],_).
get_operation_infos2([Info|Rest],AllInfos) :-
    memberchk(Info,AllInfos),
    get_operation_infos2(Rest,AllInfos).

% the following information is required for each operation identifier
% it is "filled" by the type-checker (the operation/4 case of btype2)
operation_infos(Infos) :- findall(I,operation_info(I),Infos).
operation_info(modifies(_)).
operation_info(reads(_)).


% unify type arguments of two terms
%   N is the number of the argument which should be processed
%   Arity is the arity of the term, so 1 <= N <= Arity
%   Env is the type environment
%   TypeTerm is a term that corresponds to the syntax element
%     e.g. "1<5" ~ less(1,5) has a type term less(integer,integer)
%     in the arguments of the TypeTerm, special constructs may be used:
%     "[Type]" declares that there is a list of expressions with the same Type
%     "ids" declares that there is a list of identifiers wich build a
%       sub-environment which will be used in the following arguments
%       (strict left-to-right)
%     "ids(T)" is like "ids" but the types of the identifiers are merged
%
%   TExpr is the typed expression of the complete expression
%   GI as usual: Introduced identifier and errors
type_and_unify_args(N,Arity,Pos,Env,Expr,TypeTerm,TExpr,TRin,TRout) :-
    ( N =< Arity ->
        arg(N,Expr,Arg), % Expr is the raw expression of the Nth argument
        arg(N,TypeTerm,ArgType), % ArgType is the corresponding type
        arg(N,TExpr,ArgTExpr), % TExpr is the corresponding typed expression
        type_and_unify_args_aux(ArgType,Arg,Pos,Expr,Env,ArgTExpr,NextEnv,TRin,TR2),
        N2 is N+1,
        type_and_unify_args(N2,Arity,Pos,NextEnv,Expr,TypeTerm,TExpr,TR2,TRout)
    ; otherwise ->
        TRin=TRout
    ).
type_and_unify_args_aux(ArgType,Arg,Pos,Expr,Env,ArgTExpr,NextEnv,TRin,TRout) :-
    % new identifiers are introduced, they are
    % available in the environment of the following arguments
    nonvar(ArgType),id_introduced(ArgType,CoupledType,Duplicates,Infos),!,
    functor(Expr,ExprFunctor,_), % Stored inside the information of newly created variables
    flatten_illegal_couples(Arg,FixedArg),
    ( Duplicates = unique  -> check_for_duplicates(FixedArg,Env,TRin,TR1)
    ; Duplicates = allowed -> TRin=TR1),
    add_introduced_by_info(ExprFunctor,Infos,NewInfos),
    add_ext_variables_with_info(FixedArg,Env,NewInfos,ArgTExpr,NextEnv,TR1,TR2),
    % convert the list of types to a "couplised" type,
    %   (e.g. [a] -> a, or [a,b] -> couple(a,b) )
    % in many cases (if ArgType==ids), the "couplised" type is just ignored
    idlist_to_type(ArgTExpr,CoupledType1),
    unify_types(CoupledType,CoupledType1,Pos,ArgTExpr,TR2,TRout).
type_and_unify_args_aux(ArgType,Arg,_Pos,_Expr,Env,ArgTExpr,NextEnv,TRin,TRout) :-
    % if the argument is of the form [...] we expect a list of
    % expressions with the same type
    nonvar(ArgType),ArgType = [CommonType],!,
    btype_same(Arg,Env,CommonType,ArgTExpr,TRin,TRout),
    NextEnv = Env.
type_and_unify_args_aux(ArgType,Arg,Pos,Expr,Env,ArgTExpr,NextEnv,TRin,TRout) :-
    % standard case: just type-check the argument
    btype(type_and_unify_args(Pos,Expr,Arg,ArgType),Arg,Env,ArgType,ArgTExpr,TRin,TRout),
    NextEnv = Env.

% Note: Nothing seems to use introduced_by(_); except for exists in b_intepreter
add_introduced_by_info(exists,Infos,Res) :- !, Res=[introduced_by(exists)|Infos].
add_introduced_by_info(_,Infos,Infos).

% id_introduced(+Pattern,-CoupledType,-DuplicateIdentifiers,-Infos):
%  Pattern is a pattern in the type checking database (::/2).
%  id_introduced/4 extracts the CoupledType (may be a free variable),
%    (see the declaration of "comprehension_set" for an example use)
%  and returns a flag if an introduction of an already existing variable
%  is allowed. (E.g. LET does not allow to declare an existing identifier).
%  Additional infos that are added to the type environment are supported.
id_introduced(ids,_,allowed,[]).
id_introduced(ids(CoupledType),CoupledType,allowed,[]).
id_introduced(new_readonly_ids,_,unique,[readonly]).
id_introduced(new_readonly_ids(CoupledType),CoupledType,unique,[readonly]).

% check_for_duplicates(IdList,Env,TRin,TRout):
%  IdList is a list of (not typechecked) identifiers
%  If an identifier already exists in the environment Env, an error
%  is added.
check_for_duplicates(IdList,Env,TRin,TRout) :-
    foldl(check_for_duplicates_aux(Env),IdList,TRin,TRout).
check_for_duplicates_aux(Env,identifier(Pos,Id),TRin,TRout) :-
    ( env_lookup_type(Id,Env,_Type) ->
        (env_lookup_position_string(Id,Env,PosStr)
         -> ajoin(['Identifier ',Id,' has already been declared at (line:column) =',PosStr],Msg)
         ; ajoin(['Identifier ',Id,' has already been declared'],Msg)),
        store_typecheck_error(Msg,Pos,TRin,TRout)
    ; otherwise ->
        TRin=TRout
    ).

% remove illegal couple terms introduced by Prolog BParser: TO DO: adapt Prolog BParser for UNION/INTER/...
flatten_illegal_couples([],Res) :- !, Res=[].
flatten_illegal_couples([couple(_Pos,Ids)],Res) :- !, Res = Ids.
flatten_illegal_couples([H|T],Res) :- Res = [H|T].

% environment access
lookup_type(Id,Env,Pos,Type,TRin,TRout) :-
    ( env_lookup_type(Id,Env,Type) ->
        TRin=TRout
    ; otherwise ->
        ( Id = op(Op) ->
            ajoin(['Unknown operation ',Op],Msg)
        ; otherwise ->
            ajoin(['Unknown identifier ',Id],Msg)),
        store_typecheck_error(Msg,Pos,TRin,TRout)).
lookup_infos(Id,Env,Infos) :-
    ( env_lookup_infos(Id,Env,Infos) -> !
    ; otherwise -> Infos=[]).

% check if the left hand side of an assignment is writeable
check_assign_lhs(Env,TExpr,TRin,TRout) :-
    ( id_or_function(TExpr,Id) ->
        get_texpr_type(TExpr,Type),
        lookup_infos(Id,Env,Infos),
        ( member(readonly,Infos) ->
            get_texpr_pos(TExpr,Pos),
            store_typecheck_error('read-only expression on left hand side of assignment ',Pos,TRin,TRout)
        ; nonvar(Type), Type = op(_,_) ->
            get_texpr_pos(TExpr,Pos),
            store_typecheck_error('operation on left hand side of assignment ',Pos,TRin,TRout)
        ; otherwise ->
            TRin=TRout
        )
    ; otherwise ->
        get_texpr_pos(TExpr,Pos),
        store_typecheck_error('unsupported expression on left hand side of assignment ',Pos,TRin,TRout)).


id_or_function(TIdFunc,Id) :-
    get_texpr_expr(TIdFunc,IdFunc),
    id_or_function2(IdFunc,Id).
id_or_function2(identifier(Id),Id).
id_or_function2(function(Func,_Arg),Id) :-
    id_or_function(Func,Id).

% extract information for identifiers from the environment
% currently, use all information but the position
extract_id_information(InfosIn,InfosOut) :-
    exclude(is_information_to_remove,InfosIn,InfosOut).
is_information_to_remove(nodeid(_)).
is_information_to_remove(remove(_)).


% some helper predicates
:- assert_must_succeed(btypechecker:couplise_list([a],a)).
:- assert_must_succeed(btypechecker:couplise_list([a,b],couple(a,b))).
:- assert_must_succeed(btypechecker:couplise_list([a,b,c],couple(couple(a,b),c))).
:- assert_must_succeed(btypechecker:couplise_list([a,b,c,d],couple(couple(couple(a,b),c),d))).
couplise_list([A],R) :- !,R=A.
couplise_list([A,B],R) :- !,R=couple(A,B).
couplise_list([A,B|Rest],Result) :-
    couplise_list([couple(A,B)|Rest],Result).

couplise_list_pos([A],_,A) :- !.
couplise_list_pos([A,B],Pos,couple(Pos,A,B)) :- !.
couplise_list_pos([A,B|Rest],Pos,Result) :-
    couplise_list_pos([couple(Pos,A,B)|Rest],Pos,Result).

idlist_to_type(Exprs,Type) :-
    get_texpr_types(Exprs,Types),
    couplise_list(Types,Type).

% remove choice_or constructs in a list
% there has to be at least one occurrence of choice_or
remove_choice_ors(Options,Substs) :-
    memberchk(choice_or(_,_),Options), % TODO: Is it ok that this could fail?
    maplist(remove_choice_or2,Options,Substs).
remove_choice_or2(choice_or(_,Subst),Subst) :- !.
remove_choice_or2(Subst,Subst).

add_identifiers_to_environment(Ids,In,Out) :-
    foldl(add_identifier_to_environment,Ids,In,Out).
add_identifier_to_environment(TExpr,In,Out) :-
    def_get_texpr_id(TExpr,Id),
    get_texpr_type(TExpr,Type),
    get_texpr_info(TExpr,Info),
    env_store(Id,Type,Info,In,Out).

prime_identifiers(TExprs,Primed) :-
    maplist(prime_identifier,TExprs,Primed).
prime_identifier(TExpr,PExpr) :-
    def_get_texpr_id(TExpr,Id),
    get_texpr_type(TExpr,Type),
    atom_concat(Id,'\'',Primed),
    create_texpr(identifier(Primed),Type,[],PExpr).

% allow access to abstract variables by removing all "error/1"
% information from identifiers in the environment which also
% carry the "abstraction" information
allow_access_to_abstract_vars(OldEnv,NewEnv) :-
    env_transform_infos(OldEnv,_,OldInfos,NewInfos,NewEnv),
    maplist(allow_access_to_abstract_var,OldInfos,NewInfos).
% Iterate over the information list
allow_access_to_abstract_var(OldInfo,NewInfo) :-
    % abstract variable where no access is permitted
    member(abstraction,OldInfo),
    % remove the error flag
    select(error(_),OldInfo,NewInfo),!.
allow_access_to_abstract_var(Info,Info).

% for each local concrete_variable, add a variable x$0 of the same
% type to the scope. (Needed in while loop invariants)
% TODO: This is not completely correct. The variable should only be added
% if there is an abstract variable of the same name.
add_primed_old_value_variables(InvEnv,PInvEnv) :-
    env_transform_infos(InvEnv,Ids,Infos,_,_),
    foldl(add_primed_old_value_variable,Ids,Infos,InvEnv,PInvEnv).
add_primed_old_value_variable(Id,Info,In,Out) :-
    % We add a primed version if it is a concrete variable declared in this machine
    memberchk(loc(local,_,Sec),Info),memberchk(Sec,[concrete_variables,abstract_variables]),!,
    atom_concat(Id,'$0',Primed),
    env_lookup_type(Id,In,Type),
    env_store(Primed,Type,[refers_to_old_state(Id)],In,Out).
add_primed_old_value_variable(_Id,_Info,Env,Env).

% check if a reference to the original value of a variable is made.
% This is only possible in the loop invariant of a while statement
% The reference to an original value is an identifier marked with an refers_to_old_state(Id) tag,
% see add_primed_old_value_variables/2 above.
check_for_old_state_references(TExpr,Info) :-
    check_for_old_state_references2(TExpr,Refs,[]),
    sort(Refs,SortedRefs),
    ( SortedRefs = []    -> Info = []
    ; SortedRefs = [_|_] -> Info = [refers_to_old_state(SortedRefs)]).
check_for_old_state_references2(TExpr) -->
    { syntaxtraversion(TExpr,Expr,Type,Infos,Subs,_) },
    check_for_old_state_references3(Expr,Type,Infos),
    foldl(check_for_old_state_references2,Subs).
check_for_old_state_references3(identifier(Id),Type,Infos) -->
    {memberchk(refers_to_old_state(OrigId),Infos),!}, [oref(Id,OrigId,Type)].
check_for_old_state_references3(_Expr,_Type,_Infos) --> !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% error handling
store_typecheck_error(Msg,Pos,TRin,TRout) :-
    typecheck_result_add_error(error(Msg,Pos),TRin,TRout).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% type unification

unify_types_werrors_l(A,B,Pos,Ein,Eout) :-
    typecheck_result_init(Ein,[],[],TRStart),
    unify_types_l(A,B,Pos,TRStart,TR),
    typecheck_result_get_errors(TR,Eout),
    typecheck_result_get_delayed_rules(TR,[]),
    typecheck_result_get_identifiers(TR,[]).
unify_types_werrors(A,B,Pos,Ein,Eout) :-
    unify_types_werrors_l([A],[B],[Pos],Ein,Eout).

unify_types_l([],[],[],TR,TR).
unify_types_l([A|Arest],[B|Brest],[Pos|Prest],TRin,TRout) :-
    unify_types(A,B,Pos,no_expression,TRin,TR1),
    unify_types_l(Arest,Brest,Prest,TR1,TRout).



unify_types(A,B,_,_,TR,TR) :-
    unify_types_strict(A,B),!.
unify_types(A,B,Pos,Expr,TRin,TRout) :-
    unify_failed(A,B,Pos,Expr,TRin,TRout).
unify_failed(A,B,Pos,Expr,TRin,TRout) :-
    pretty_type(A,PA), pretty_type(B,PB),
    ( nonvar(Expr),Expr\=no_expression,translate_bexpression(Expr,ExprString) ->
        ajoin(['Type mismatch: Expected ', PA, ', but was ', PB,
               ' in \"', ExprString, '\"'], Msg)
    ; otherwise ->
        ajoin(['Type mismatch: Expected ', PA, ', but was ', PB], Msg)),
    store_typecheck_error(Msg,Pos,TRin,TRout).

:- assert_must_succeed(( btypechecker:unify_types_strict(set(couple(boolean,string)), set(couple(boolean,string))) )).
:- assert_must_succeed(( btypechecker:unify_types_strict(seq(integer), set(couple(integer,integer))) )).
:- assert_must_succeed(( btypechecker:unify_types_strict(seq(set(couple(integer,integer))),
                                                         set(couple(integer,seq(integer)))) )).
:- assert_must_succeed(( btypechecker:unify_types_strict(seq(integer), any) )).
:- assert_must_succeed(( btypechecker:unify_types_strict(any, set(couple(integer,integer))) )).
:- assert_must_succeed(( btypechecker:unify_types_strict(seq(any), set(couple(integer,boolean))) )).

unify_types_strict(A,B) :-
    ( var(A) ->
        ( var(B)    -> A=B
        ; otherwise -> unify_types1(B,A))
    ; otherwise ->
        unify_types1(A,B)
    ),!.

unify_types1(A,B) :- var(B),contains_var(B,A),!,fail.
unify_types1(A,B) :- unify_types2(A,B).

unify_types2(any,B) :- !,
   (var(B) -> B=any
            ; true). % TO DO: set inner variables to any ?
unify_types2(_,B) :- B==any, !. % Maybe these two cases should be removed and type errors generated
                                % where the type cannot be inferred
unify_types2(set(A),set(B)) :- !, unify_types_strict(A,B).
unify_types2(seq(A),seq(B)) :- !, unify_types_strict(A,B).
unify_types2(set(A),seq(B)) :- !, unify_types_strict(seq(B),set(A)).
unify_types2(seq(A),set(B)) :- !, unify_types_strict(couple(integer,A),B).
unify_types2(couple(A,B),couple(X,Y)) :- !,unify_types_strict(A,X),unify_types_strict(B,Y).
unify_types2(string,string).
unify_types2(integer,integer).
unify_types2(boolean,boolean).
unify_types2(global(G),global(G)).
unify_types2(pred,pred).
unify_types2(subst,subst).
unify_types2(record(A),record(B)) :- !, unify_fields(B,A).
unify_types2(op(ParamsA,ResultsA),op(ParamsB,ResultsB)) :-
    unify_types_l(ParamsA,ParamsB),
    unify_types_l(ResultsA,ResultsB).
unify_types2(freetype(Name),freetype(Name)).
unify_types2(freetype(Name,TypeParams1),freetype(Name,TypeParams2)) :-
    unify_types_l(TypeParams1,TypeParams2).
unify_types2(witness,witness).
unify_types2(status,status).


unify_types_l(A,B) :- var(A),!,A=B.
unify_types_l([],[]) :- !.
unify_types_l([T1|Rest1],[T2|Rest2]) :-
    unify_types_strict(T1,T2),unify_types_l(Rest1,Rest2).

unify_fields(Sub,Super) :-
    nonvar(Sub),nonvar(Super),!,
    ( Sub = [field(Name,Type)|SubRest] ->
        extract_field(Super,Name,Type,SuperRest),
        unify_fields(SubRest,SuperRest)
    ; Sub = [] ->
        Super = []).
unify_fields(Fields,Fields).

extract_field(L,Name,Type,Rest) :-
    extract_field2(L,Name,Type,Rest).
extract_field2(L,Name,Type,Rest) :- nonvar(L),!,
    L = [field(FName,FType)|LRest],
    ( Name = FName ->
        unify_types_strict(Type,FType),
        Rest=LRest
    ; otherwise ->
        Rest=[field(FName,FType)|NRest],
        extract_field2(LRest,Name,Type,NRest)).
extract_field2(L,Name,Type,LRest) :-
    L = [field(Name,Type)|LRest].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% check if all types of the variables are ground
check_ground_types_dl(TExprs,Exceptions,Ein,Eout) :-
    foldl(check_ground_type_dl(Exceptions),TExprs,Ein,Eout).
check_ground_type_dl(Exceptions,TExpr,Ein,Eout) :-
    get_texpr_type(TExpr,Type),
    ( is_ground_type(Exceptions,Type) ->
        Ein=Eout
    ; otherwise ->
        get_texpr_expr(TExpr,identifier(Id)),
        ((preferences:get_preference(allow_untyped_identifiers,true),
         instantiate_non_ground_type(Exceptions,Type))
          -> translate:pretty_type(Type,TS),
             format('Instantiated type of ~w to ~w~n',[Id,TS]),
             Ein=Eout
          ;  get_texpr_pos(TExpr,Pos),
             ajoin(['Could not infer type of ',Id], Msg),
             Ein=[error(Msg,Pos)|Eout]
        )
    ).
is_ground_type(_Exceptions,Type) :- ground(Type),!.
is_ground_type(Exceptions,Type) :- var(Type),!,exact_member(Type,Exceptions).
is_ground_type(Exceptions,Type) :-
    Type =.. [_|TArgs],
    maplist(is_ground_type(Exceptions),TArgs).

:- use_module(b_global_sets,[generate_fresh_supplementary_global_set/1]).
instantiate_non_ground_type(_Exceptions,Type) :- ground(Type),!.
instantiate_non_ground_type(Exceptions,Type) :- var(Type),!,
   (exact_member(Type,Exceptions) -> true
      ; % if preference set to true we allow those untyped vars and use new deferred sets for them
        %% Type = string  % temporary : TO DO: generate a new deferred set
        generate_fresh_supplementary_global_set(GS),
        Type = global(GS)
    ).
instantiate_non_ground_type(Exceptions,Type) :-
    Type =.. [_|TArgs],
    maplist(instantiate_non_ground_type(Exceptions),TArgs).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% rewrite rules to clean up the syntax tree

% substitute definitions
btype_rewrite(RawExpr,Env,_Pos,Result) :-
    is_definition_call(RawExpr,Env),!,
    rewrite_definition(RawExpr,Env,Result).
% other (not definition-related) rules
btype_rewrite(Expr,_Env,Pos,New) :- btype_rewrite2(Expr,Pos,New).

% a shorter version of the
is_definition_call(identifier(Name),Env) :-
    lookup_definition(Name,Env,_Params,_Body).
is_definition_call(function(DefIdP,_Args),Env) :-
    remove_pos(DefIdP,identifier(Name)),
    % the definition should have parameters, if not, it could be
    % that a function is defined and normal function application is
    % needed
    % Note: There is a rewrite rule for functions that converts a list
    %   of arguments into couples. That rewrite rule must not be called
    %   before this rule.
    lookup_definition(Name,Env,[_|_],_Body).
is_definition_call(definition(_Name,_Args),_Env).

rewrite_definition(identifier(Name),Env,Result) :-
    rewrite_definition2(Name,[],Env,Result).
rewrite_definition(function(DefIdP,Args),Env,Result) :-
    % A function application where the function is an identifier
    % may be a definition with arguments
    remove_pos(DefIdP,identifier(Name)),
    rewrite_definition2(Name,Args,Env,Result).
rewrite_definition(definition(Name,Args),Env,Result) :-
    % the standard way of using a definition
    rewrite_definition2(Name,Args,Env,Result).

rewrite_definition2(Name,Args,Env,Result) :-
    ( lookup_definition(Name,Env,Params,Body) ->
        btype_rewrite_definition(Params,Name,Env,Args,Body,NewBody),
        remove_pos(NewBody,Result) % remove position info
    ; otherwise ->
        ajoin(['Definition ',Name,' not found'],ErrMsg),
        throw(rewrite_exception(ErrMsg,Name))).

btype_rewrite_definition(Params,Name,Env,Args,Body,NewBody) :-
    ( same_length(Params,Args) ->
        maplist(definition_expansion_on_raw_ast_with_pos(Env),Args,NewArgs),
        maplist(create_def_replace,Params,NewArgs,Replaces),
        raw_replace(Body,Replaces,NewBody)  %%  ,print(replaced_def(Name,NewBody)),nl %%
    ; otherwise ->
        length(Params,PN),length(Args,AN),
        ajoin(['Expected ',PN,' instead of ',AN,' arguments for definition ',Name],ErrMsg),
        throw(rewrite_exception(ErrMsg,Name))).
create_def_replace(identifier(_,Old),New,replace(Old,New)).
% lookup DEFINITION with Name in Environment Env
lookup_definition(Name,Env,Params,Body) :-
    env_lookup_type(Name,Env,Type), nonvar(Type),
    Type = definition(_,Params,Body).

% expressions
btype_rewrite2(couple([A,B]),_,couple(A,B)) :- !.
btype_rewrite2(couple(Elements),Pos,couple(X,Y)) :- couplise_list_pos(Elements,Pos,couple(_,X,Y)).
btype_rewrite2(function(Fun,Args),Pos,function(Fun,Arg)) :-
    couplise_list_pos(Args,Pos,Arg).

% substitutions and operations
btype_rewrite2(if(Pred,Subst,Elsifs,Else),Pos,if(IfList)) :-
    (Else=skip(_) -> IfList = [if_elsif(Pos,Pred,Subst)|Elsifs]
     ; append([if_elsif(Pos,Pred,Subst)|Elsifs],[if_elsif(none,truth(none),Else)],IfList)).
btype_rewrite2(case(Expr,Either,EitherSubst,Ors,Else),Pos,
                  case(Expr,[case_or(Pos,Either,EitherSubst)|Ors],Else)).
btype_rewrite2(select(Pred,Then,Whens,Else),Pos,select([select_when(Pos,Pred,Then)|Whens],Else)).
btype_rewrite2(select(Pred,Then,Whens),Pos,select([select_when(Pos,Pred,Then)|Whens])).
btype_rewrite2(choice([]),_,choice([])).
btype_rewrite2(choice(Options),_,choice(Substs)) :-
    remove_choice_ors(Options,Substs).

% integer sets
btype_rewrite2(integer_set,_,integer_set('INTEGER')).
btype_rewrite2(natural_set,_,integer_set('NATURAL')).
btype_rewrite2(natural1_set,_,integer_set('NATURAL1')).
btype_rewrite2(nat_set,_,integer_set('NAT')).
btype_rewrite2(nat1_set,_,integer_set('NAT1')).
btype_rewrite2(int_set,_,integer_set('INT')).

% applying rewriting of definitions on the raw (un-typed input) syntax tree
% this is used for parameters of definitions
definition_expansion_on_raw_ast_with_pos(Env,RawExprPos,OutAst) :-
    compound(RawExprPos),ext2int(RawExprPos,RawExpr,Pos,_,_,_,_),!,
    definition_expansion_on_raw_ast_without_pos(Env,RawExpr,OutAst1),
    OutAst1 =.. [Functor|Args], OutAst =.. [Functor,Pos|Args].
definition_expansion_on_raw_ast_with_pos(_Env,Ast,Ast).
definition_expansion_on_raw_ast_without_pos(Env,RawExpr,OutAst) :-
    is_definition_call(RawExpr,Env),!,
    rewrite_definition(RawExpr,Env,InterAst),
    definition_expansion_on_raw_ast_without_pos(Env,InterAst,OutAst).
definition_expansion_on_raw_ast_without_pos(Env,InAst,OutAst) :-
    compound(InAst),
    InAst =.. [Functor|Args],!,
    same_functor(InAst,OutAst),
    OutAst =.. [Functor|NewArgs],
    maplist(definition_expansion_on_raw_ast_with_pos(Env),Args,NewArgs).
definition_expansion_on_raw_ast_without_pos(_Env,Ast,Ast).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% extract syntax tree positions

remove_pos(Ext,Expr) :-
    ext2int(removepos,Ext,Expr,_,_,_,_,_).

% create a enhanced syntax element from a raw input
% ext2int(+Ext,-Expr,-Pos,+Type,+TExpr,+RestInfos,-Result):
%  Ext:  raw information with position info (from parser)
%  Expr: the underlying expression without position info (e.g. identifier(x))
%  Pos:  the position extracted from the raw expression
%  Type: the type used for the typed expression
%  TExpr: the new expression in the enhanced syntax tree format
%        (but without the information for the root element)
%  RestInfos: Additional list of informations that will be stored inside
%        the element aside from the position information
%  Result: the new completed element (with type and additional information)
%     Example: "1+x"
%          Ext:    add(pos1,integer(pos2,1),identifier(pos3,x)
%          Expr:   add(integer(pos2,1),identifier(pos3,x))
%          Result: b(TExpr,Type,[nodeid(pos1)|RestInfos])
ext2int(Ext,Expr,Pos,Type,TExpr,RestInfos,Res) :-
    ext2int(unknown_src,Ext,Expr,Pos,Type,TExpr,RestInfos,Res).
ext2int(Src,Ext,Expr,_Pos,_Type,_TExpr,_RestInfos,_) :-
    var(Ext),!,add_error(ext2int,'Variable Ext: ',Src:Expr,Expr),fail.
ext2int(_Src,Ext,Expr,Pos,Type,TExpr,RestInfos,b(TExpr,Type,[nodeid(Pos)|Infos])) :-
    % the first argument contains the position information
    safe_functor(ext2int_ext,Ext,Functor,Arity), arg(1,Ext,ExtPos),
    extract_optional_external_info(ExtPos,Pos,Infos,InfosWithPragma),
    % now construct a copy without the first argument
    NArity is Arity-1,
    safe_functor(ext2int_expr,Expr,Functor,NArity),
    copy_ext_term(Arity,Ext,Expr),
    apply_pragmas(Pos,TExpr,Type,RestInfos,InfosWithPragma).
% It is possible to inject information into the AST from outside
% by using info(Pos,Info) or info(Pos,[Info1,...,InfoN]) as position
% Those terms are copied directly into the AST.
% The result is returned in form of a difference list I1-I2.
extract_optional_external_info(info(Pos,ExtInfo),Pos,I1,I2) :- !,
    ( ExtInfo = []    -> I1=I2
    ; ExtInfo = [_|_] -> append(ExtInfo,I2,I1)
    ; otherwise       -> I1=[ExtInfo|I2]).
extract_optional_external_info(Pos,Pos,Infos,Infos).

copy_ext_term(1,_,_) :- !.
copy_ext_term(N1,Ext,Int) :-
    N2 is N1-1, arg(N1,Ext,Arg), arg(N2,Int,Arg),
    copy_ext_term(N2,Ext,Int).

% add variables to the environment
% - the variables are given as raw parser terms
%   and returned in typed format
% - additionally the identifier is stored in the list of newly introduced
% identifiers
add_ext_variables(Identifiers, Env, TExprs, NewEnv, TRin, TRout) :-
    add_ext_variables_with_info(Identifiers, Env, [], TExprs, NewEnv, TRin, TRout).

add_ext_variables_with_info(Identifiers, Env, AddInfo, TExprs, NewEnv, TRin, TRout) :-
    foldl(add_ext_variable_with_info(AddInfo), Identifiers, TExprs, TRin/Env, TRout/NewEnv).
add_ext_variable_with_info(AddInfo, Identifier, TExpr, TRin/Env, TRout/NewEnv) :-
    % convert the raw format to the typed format
    Expr = identifier(Name),
    % Krings: in the following line, AddInfo was set to []
    % thus, additional information was only added to the type environment,
    % not to the identifier itself. introduced_by was only store inside of
    % expressions, not at the identifier declarations. this might break something?
    ext2int(add_ext_variables_with_info,Identifier,Expr,_,Type,Expr,AddInfo,TExpr),!,
    % add the identifier's type and info to the type environment
    get_texpr_info(TExpr,Infos),
    % append not needed anymore, as additional information is stored in
    % the identifier. see comment above.
    %append(AddInfo,Infos,AllInfos),
    env_store(Name,Type,Infos,Env,NewEnv),
    % store the identifier in the list of newly introduced identifiers
    typecheck_result_add_identifier(TExpr,TRin,TRout).
add_ext_variable_with_info(_AddInfo, Expr, TExpr, TRin/Env, TRout/Env) :-
    % We expected an identifier but there was something different
    ext2int(add_ext_variables_with_info,Expr,_,Pos,_,_,_,_),
    ( translate_raw_expr_functor(Expr,TS) -> ajoin(['Expected identifier, but got operator: ',TS],Msg)
    ; otherwise                           -> Msg='Expected identifier'),
    store_typecheck_error(Msg,Pos,TRin,TRout),
    % just to make sure that there is no variable as AST element:
    create_texpr(identifier('not an identifier'),_,[found(Expr)],TExpr).

translate_raw_expr_functor(Expr,Translation) :-
    % note: we cannot call translate_bexpression yet; not yet wrapped and typed !
    nonvar(Expr), functor(Expr,F,_),
    translate_prolog_constructor(F,Translation).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% type rules

% predicates
truth :: pred.
falsity :: pred.
conjunct(pred,pred) :: pred.
negation(pred) :: pred.
disjunct(pred,pred) :: pred.
implication(pred,pred) :: pred.
equivalence(pred,pred) :: pred.
equal(A,A) :: pred.
not_equal(A,A) :: pred.
member(A,set(A)) :: pred.
not_member(A,set(A)) :: pred.
subset(set(A),set(A)) :: pred.
subset_strict(set(A),set(A)) :: pred.
not_subset(set(A),set(A)) :: pred.
not_subset_strict(set(A),set(A)) :: pred.
less_equal(integer,integer) :: pred.
less(integer,integer) :: pred.
greater_equal(integer,integer) :: pred.
greater(integer,integer) :: pred.
finite(set(_)) :: pred.
exists(ids,pred) :: pred.
partition(set(T),[set(T)]) :: pred.

% expressions
boolean_true :: boolean.
boolean_false :: boolean.
max_int :: integer.
min_int :: integer.
empty_set :: set(_).
bool_set :: set(boolean).
string_set :: set(string).
convert_bool(pred) :: boolean.
add(integer,integer) :: integer.
minus(integer,integer) :: integer.
unary_minus(integer) :: integer.
multiplication(integer,integer) :: integer.
cartesian_product(set(A),set(B)) :: set(couple(A,B)).
div(integer,integer) :: integer.
modulo(integer,integer) :: integer.
power_of(integer,integer) :: integer.
successor :: set(couple(integer,integer)).
predecessor :: set(couple(integer,integer)).
max(set(integer)) :: integer.
min(set(integer)) :: integer.
card(set(_)) :: integer.
couple(A,B) :: couple(A,B).
pow_subset(set(A)) :: set(set(A)).
pow1_subset(set(A)) :: set(set(A)).
fin_subset(set(A)) :: set(set(A)).
fin1_subset(set(A)) :: set(set(A)).
interval(integer,integer) :: set(integer).
union(set(A),set(A)) :: set(A).
intersection(set(A),set(A)) :: set(A).
set_subtraction(set(A),set(A)) :: set(A). /* ?? */
general_union(set(set(A))) :: set(A).
general_intersection(set(set(A))) :: set(A).
relations(set(A),set(B)) :: set(set(couple(A,B))).
identity(set(A)) :: set(couple(A,A)).
event_b_identity :: set(couple(A,A)). % for Rodin 1.0
reverse(set(couple(A,B))) :: set(couple(B,A)). % function inverse ~
first_projection(set(A),set(B)) :: set(couple(couple(A,B),A)).
event_b_first_projection(set(couple(A,B))) :: set(couple(couple(A,B),A)). % should no longer appear after Rodin 1.0
event_b_first_projection_v2 :: set(couple(couple(A,_B),A)). % for Rodin 1.0
second_projection(set(A),set(B)) :: set(couple(couple(A,B),B)).
event_b_second_projection(set(couple(A,B))) :: set(couple(couple(A,B),B)). % should no longer appear after Rodin 1.0
event_b_second_projection_v2 :: set(couple(couple(_A,B),B)). % for Rodin 1.0
composition(set(couple(A,B)),set(couple(B,C))) :: set(couple(A,C)).
ring(set(couple(B,C)),set(couple(A,B))) :: set(couple(A,C)).
direct_product(set(couple(E,F)),set(couple(E,G))) :: set(couple(E,couple(F,G))).
parallel_product(set(couple(E,F)),set(couple(G,H))) :: set(couple(couple(E,G),couple(F,H))).
trans_function(set(couple(A,B))) :: set(couple(A,set(B))).
trans_relation(set(couple(A,set(B)))) :: set(couple(A,B)).
iteration(set(couple(A,A)),integer) :: set(couple(A,A)).
reflexive_closure(set(couple(A,A))) :: set(couple(A,A)).
closure(set(couple(A,A))) :: set(couple(A,A)).
domain(set(couple(A,_))) :: set(A).
range(set(couple(_,B))) :: set(B).
image(set(couple(A,B)),set(A)) :: set(B).
domain_restriction(set(A),set(couple(A,B))) :: set(couple(A,B)).
domain_subtraction(set(A),set(couple(A,B))) :: set(couple(A,B)).
range_restriction(set(couple(A,B)),set(B)) :: set(couple(A,B)).
range_subtraction(set(couple(A,B)),set(B)) :: set(couple(A,B)).
overwrite(set(couple(A,B)),set(couple(A,B))) :: set(couple(A,B)).
partial_function(A,B)::T    :- relations(A,B)::T.
total_function(A,B)::T      :- relations(A,B)::T.
partial_injection(A,B)::T   :- relations(A,B)::T.
total_injection(A,B)::T     :- relations(A,B)::T.
partial_surjection(A,B)::T  :- relations(A,B)::T.
total_surjection(A,B)::T    :- relations(A,B)::T.
total_bijection(A,B)::T     :- relations(A,B)::T.
partial_bijection(A,B)::T     :- relations(A,B)::T.
total_relation(A,B)::T      :- relations(A,B)::T.
surjection_relation(A,B)::T :- relations(A,B)::T.
total_surjection_relation(A,B)::T :- relations(A,B)::T.
seq(set(A)) :: set(seq(A)).
seq1(X)::T  :- seq(X)::T.
iseq(X)::T  :- seq(X)::T.
iseq1(X)::T :- seq(X)::T.
perm(set(A)) :: set(seq(A)).
empty_sequence :: seq(_).
size(seq(_)) :: integer.
first(seq(A)) :: A.
last(seq(A)) :: A.
front(seq(A)) :: seq(A).
tail(seq(A)) :: seq(A).
rev(seq(A)) :: seq(A).
concat(seq(A),seq(A)) :: seq(A).
insert_front(A,seq(A)) :: seq(A).
insert_tail(seq(A),A) :: seq(A).
restrict_front(seq(A),integer) :: seq(A).
restrict_tail(seq(A),integer) :: seq(A).
general_concat(seq(seq(A))) :: seq(A).
function(set(couple(A,B)),A) :: B.
set_extension([T]) :: set(T).
sequence_extension([T]) :: seq(T).
comprehension_set(ids(T),pred) :: set(T).
event_b_comprehension_set(ids,T,pred) :: set(T).
lambda(ids(A),pred,R) :: set(couple(A,R)).
general_sum(ids,pred,integer) :: integer.
general_product(ids,pred,integer) :: integer.
quantified_union(ids,pred,set(T)) :: set(T).
quantified_intersection(ids,pred,set(T)) :: set(T).
mu(set(T)) :: T.
if_then_else(pred,T,T) :: T.

% substitutions
skip :: subst.
block(subst) :: subst.
precondition(pred,subst) :: subst.
sequence([subst]) :: subst.
any(new_readonly_ids,pred,subst) :: subst.
if([subst]) :: subst.
if_elsif(pred,subst) :: subst.
assertion(pred,subst) :: subst.
let(new_readonly_ids,pred,subst) :: subst.
var(ids,subst) :: subst.
select([subst],subst) :: subst.
select([subst]) :: subst.
select_when(pred,subst) :: subst.
choice([subst]) :: subst.

% other stuff
witness(_,pred) :: witness.
ordinary :: status.
anticipated(_) :: status.
convergent(_) :: status.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% check for query-only operations

/* get_accessed_vars/4 computes the variables that are read or written to
   by a substitution.
   get_accessed_vars(+Subst,+LocalVars,-Modified,-Reads):
    Subst is the substitution (as a typed syntax tree)
    LocalVars is a list of (untyped) ids that are local variables.
       Local variables are ignored and not included in the list of read or written
       variables.
    Modified is the list of (untyped) identifiers that are written by the substitution and
    Reads is the list of (untyped) identifiers that are read. */
get_accessed_vars(TSubst,LocalVars,Modified,Reads) :-
    get_accessed_vars1(TSubst,LocalVars,In,Out),
    extract_reads_and_modifies(In,Out,Modified,Reads).
extract_reads_and_modifies((UnsortedMods,UnsortedReads),([],[]),Modified,Reads) :-
    sort(UnsortedMods,Modified),
    sort(UnsortedReads,Reads).
get_accessed_vars1(TSubst,LocalVars,In,Out) :-
    get_texpr_expr(TSubst,Subst),
    ( Subst = operation_call(Operation,Results,Arguments) ->
        get_texpr_info(TSubst,Info),
        def_get_texpr_id(Operation,op(OperationName)), %print(info(OperationName,Info)),nl,
        memberchk(modifies(MVars),Info),
        memberchk(reads(RVars),Info),
        %op_modifies_reads_cache(OperationName,MVarsC,RVarsC) %this is no longer necessary
        ((ground(MVars), ground(RVars)) -> true
          ; add_error(btypechecker, 'Cannot obtain modifies/reads info for called operation (probably illegal local call): ',OperationName),
             % TEST 655 used to reach this branch
            MVars=[], RVars=[]
           % this should only happen when local operation called !?!
           % or ensure that operations are processed before we reach operation calls
           % this may be problematic for local operations
        ),
        %print(compute_acc_vars_for_op_call(modifies(MVars),Subst)),nl,trace,
        accessed_vars_for_operation_call(MVars,RVars,LocalVars,Results,Arguments,In,Out)
    ; get_accessed_vars2(Subst,LocalVars,In,Out) ->
        true
    ; otherwise ->
        add_internal_error('Could not extract information about accessed variables for substitution',
                           get_accessed_vars2(Subst,LocalVars,In,Out)), fail).
accessed_vars_for_operation_call(MVars,RVars,LocalVars,Results,Arguments) -->
    add_modified_vars2(MVars),
    add_read_vars2(RVars),
    add_modified_vars(Results,LocalVars),
    add_read_vars_l(Arguments,LocalVars).

get_accessed_vars2(skip,_) --> !.
get_accessed_vars2(block(Subst),LocalVars) --> get_accessed_vars1(Subst,LocalVars).
get_accessed_vars2(assertion(A,Subst),R) -->   add_read_vars(A,R),get_accessed_vars1(Subst,R).
get_accessed_vars2(assign(LHS,E),R) -->        %Note: LHS can be nested function application f(a)(b) :=
      {maplist(get_lhs_assigned_identifier,LHS,Ids)}, add_modified_vars(Ids,R),
      {convlist(get_lhs_read_expression,LHS,LHS_R), append(LHS_R,E,RE)}, add_read_vars_l(RE,R).
get_accessed_vars2(assign_single_id(Id,E),R) --> add_read_vars_l([E],R),add_modified_vars([Id],R).
get_accessed_vars2(precondition(P,Subst),R) -->add_read_vars(P,R),get_accessed_vars1(Subst,R).
get_accessed_vars2(assertion(Subst),R) -->     get_accessed_vars1(Subst,R).
get_accessed_vars2(choice(Substs),R) -->       get_accessed_vars_for_choice(Substs,R).
get_accessed_vars2(if(Substs),R) -->           get_accessed_vars_for_if(Substs,R).
get_accessed_vars2(select(Whens,Else),R) -->   get_accessed_vars_for_select(Whens,Else,R).
get_accessed_vars2(select(Whens),R) -->        get_accessed_vars_for_select(Whens,none,R).
get_accessed_vars2(case(Expr,Substs,Else),R) -->  get_accessed_vars_for_case(Expr,Substs,Else,R).
get_accessed_vars2(case_or(Exprs,Subst),R) --> add_read_vars_l(Exprs,R),get_accessed_vars1(Subst,R).
get_accessed_vars2(any(Ids,G,Subst),R) -->     append_vars(R,Ids,L),add_read_vars(G,L),get_accessed_vars1(Subst,L).
get_accessed_vars2(let(Ids,G,Subst),R) -->     append_vars(R,Ids,L),add_read_vars(G,L),get_accessed_vars1(Subst,L).
get_accessed_vars2(becomes_element_of(Ids,E),R) --> add_modified_vars(Ids,R),add_read_vars(E,R).
get_accessed_vars2(becomes_such(Ids,E),LocalVars0) -->  add_modified_vars(Ids,LocalVars0),
    {append_primed_vars(Ids,LocalVars0,LocalVars1)}, % all Ids are already marked as accessed, any reference to Id$0 is a read of the old value; do not add Id$0 to the accessed variables
    add_read_vars(E,LocalVars1).
get_accessed_vars2(evb_becomes_such(Ids,E),R) --> add_modified_vars(Ids,R),add_read_vars(E,R).
get_accessed_vars2(sequence(Substs),R) -->     get_accessed_vars_l(Substs,R).
get_accessed_vars2(parallel(Substs),R) -->     get_accessed_vars_l(Substs,R).
get_accessed_vars2(while(Cond,Subst,Inv,Variant),R) -->
    get_accessed_vars_for_while(Cond,Subst,Inv,Variant,R).
get_accessed_vars2(var(Ids,Subst),R)         --> append_vars(R,Ids,L),get_accessed_vars1(Subst,L).
% TODO: we need information about which variables are modified by an external function
get_accessed_vars2(external_subst_call(_Name,Exprs),R) --> add_read_vars_l(Exprs,R).

get_accessed_vars_for_while(Cond,Subst,Inv,Variant,Locals) -->
    get_accessed_vars_for_while2(Inv,Locals),
    add_read_vars(Cond,Locals),
    add_read_vars(Variant,Locals),
    get_accessed_vars1(Subst,Locals).
get_accessed_vars_for_while2(Inv,Locals,In,Out) :-
    add_read_vars(Inv,Locals,LI,LO),
    extract_reads_and_modifies(LI,LO,Modified,Reads0),
    maplist(remove_optional_prime,Reads0,Reads),
    add_read_vars2(Reads,In,Inter),
    add_modified_vars2(Modified,Inter,Out).
remove_optional_prime(Id,Orig) :-  atom_concat(Orig,'$0',Id),!.
remove_optional_prime(Id,Id).

get_accessed_vars_for_if(ElseIfs,Locals) -->
    { maplist(split_elseif,ElseIfs,Conds,Subst),
      ( if_has_else_block(Conds) -> HasElse = has_else
      ; otherwise                -> HasElse = no_else) },
    % mark all used variables in the conditions as read
    add_read_vars_l(Conds,Locals),
    get_accessed_vars_for_if_when(HasElse,Subst,Locals).
split_elseif(Expr,Cond,Subst) :- get_texpr_expr(Expr,if_elsif(Cond,Subst)).

get_accessed_vars_for_select(Whens,Else,Locals) -->
    { maplist(split_when,Whens,Conds,Subst1),
      ( Else = none ->
          HasElse=no_else, Subst = Subst1
      ; otherwise ->
          HasElse=has_else, Subst = [Else|Subst1]) },
    % mark all used variables in the conditions as read
    add_read_vars_l(Conds,Locals),
    get_accessed_vars_for_if_when(HasElse,Subst,Locals).
split_when(Expr,Cond,Subst) :- get_texpr_expr(Expr,select_when(Cond,Subst)).

get_accessed_vars_for_if_when(HasElse,Subst,Locals) -->
    { get_accessed_vars_for_if_when2(HasElse,Subst,Locals,Modified,Reads) },
    add_read_vars2(Reads),
    add_modified_vars2(Modified).

get_accessed_vars_for_if_when2(has_else,Subst,Locals,ModifiedBySome,ReadBySome) :-
    maplist(get_accessed_vars_x(Locals),Subst,Modifies,Reads),
    % We want to calculate those elements that are modified by some (in the union)
    % but not by all (in the intersection)
    ord_union(Modifies,ModifiedBySome),
    ord_intersection(Modifies,ModifiedByAll),
    ord_subtract(ModifiedBySome,ModifiedByAll,ModifiedByNotAll),
    % we add those values that are not modified by all to the set of read variables
    % because e.g. "IF P THEN x:=1 END" is equivalent to "IF P THEN x:=1 ELSE x:=x END"
    ord_union([ModifiedByNotAll|Reads],ReadBySome).
get_accessed_vars_for_if_when2(no_else,Subst,Locals,ModifiedBySome,ReadBySome) :-
    % no else block present, we must assume that each written variable
    % can also be treaten like a read variable (see example in the other condition)
    maplist(get_accessed_vars_x(Locals),Subst,Modifies,Reads),
    ord_union(Modifies,ModifiedBySome),
    ord_union([ModifiedBySome|Reads],ReadBySome).

get_accessed_vars_for_choice(Substs,Locals) -->
    get_accessed_vars_for_if_when(has_else,Substs,Locals).

get_accessed_vars_for_case(Expr,Eithers,Else,Locals) -->
    % we can treat a case statement analogously to an IF-THEN-ELSE statement
    % where an ELSE is present
    % additionally, we have to add all expressions to the read set
    add_read_vars(Expr,Locals),
    { maplist(split_case_or,Eithers,ExprsList,Substs), append(ExprsList,Exprs) },
    add_read_vars_l(Exprs,Locals),
    get_accessed_vars_for_if_when(has_else,[Else|Substs],Locals).
split_case_or(CaseOr,Exprs,Subst) :- get_texpr_expr(CaseOr,case_or(Exprs,Subst)).

% just the same as get_accessed_vars, but with different order of parameters
% to allow call with maplist
get_accessed_vars_x(Locals,Subst,Modifies,Reads) :-
    get_accessed_vars(Subst,Locals,Modifies,Reads).

if_has_else_block(Conds) :-
    last(Conds,Cond),is_truth(Cond).

append_vars(Vars,TIds,NewVars,X,X) :-
    get_texpr_ids(TIds,Ids),
    append(Vars,Ids,NewVars).

add_read_vars_l([],_LocalVars) --> !.
add_read_vars_l([Expr|Rest],LocalVars) -->
    add_read_vars(Expr,LocalVars),
    add_read_vars_l(Rest,LocalVars).

% extract modified identifier from LHS of assign:
get_lhs_assigned_identifier(b(function(LHS,_),_,_),Res) :- !, get_lhs_assigned_identifier(LHS,Res).
get_lhs_assigned_identifier(ID,ID).

% extract part of LHS of assign which reads information:
get_lhs_read_expression(b(function(LHS,Read1),_,_),Res) :- !,
   (get_lhs_read_expression(LHS,Read2) -> create_couple(Read1,Read2,Res)
     ; Res = Read1).

add_read_vars(Expression,LocalVars) -->
    {find_identifier_uses(Expression,LocalVars,ReadVars)},
    add_read_vars2(ReadVars).
add_read_vars2(Vars,(Modified,InRead),(Modified,OutRead)) :- append(Vars,OutRead,InRead).

add_modified_vars(Ids,LocalVars) -->
    {filter_result_vars(Ids,LocalVars,ModifiedVars)},
    add_modified_vars2(ModifiedVars).
add_modified_vars2(Vars,(InModified,Read),(OutModified,Read)) :-
    (var(Vars) -> add_internal_error('Modified Vars variable: ',add_modified_vars2(Vars,(InModified,Read),(OutModified,Read))) ; true),
    append(Vars,OutModified,InModified).

filter_result_vars(Ids,LocalVars,ModifiedVars) :-
    % extract_changed_id might fail if the user entered
    % a wrong LHS of an assignment. This should lead to an
    % type error, handled by the assign type checking rule.
    convlist(extract_changed_id,Ids,Names),
    remove_all(Names,LocalVars,ModifiedVars).
extract_changed_id(Id,Name) :-
    get_texpr_id(Id,Name),!.
extract_changed_id(Func,Name) :-
    get_texpr_expr(Func,function(TId,_)), !,
    extract_changed_id(TId,Name). % we can have nested function calls f(x)(y) := z
extract_changed_id(X,Name) :-
    add_internal_error('Could not extract changed id: ',extract_changed_id(X,Name)),fail.

get_accessed_vars_l([],_) --> [].
get_accessed_vars_l([S|Rest],R) --> get_accessed_vars1(S,R),get_accessed_vars_l(Rest,R).

:- assert_must_succeed(( btypechecker:append_primed_vars([b(identifier(x),integer,[])],[],R), R = ['x$0'])).
append_primed_vars(TIds,R,L) :-
    maplist(add_dollar_zero,TIds,Primed),
    append(Primed,R,L).
:- assert_must_succeed(( btypechecker:add_dollar_zero(b(identifier(x),integer,[]),R), R = 'x$0')).
add_dollar_zero(TId,P) :-
    get_texpr_id(TId,Id),
    number_suffix(Id,0,P).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%+
% type environment
env_empty(env(Avl)) :- empty_avl(Avl).

env_lookup_type(Id,Env,Type) :-
    env_lookup2(Env,Id,Type,_Infos).
env_lookup_infos(Id,Env,Infos) :-
    env_lookup2(Env,Id,_Type,Infos).
env_lookup_position_string(Id,Env,PosStr) :-
    env_lookup2(Env,Id,_Type,Infos),
    get_info_pos(Infos,Pos), Pos \= none,
    translate:translate_span(Pos,PosStr).
env_lookup2(env(Avl),Id,Type,Infos) :-
    avl_fetch(Id,Avl,entry(Type,Infos)).
env_lookup2(openenv(Open,Env),Id,Type,Infos) :-
    ( env_lookup2(Env,Id,Type,Infos) ->
        true
    ; otherwise ->
        openenv_lookup(Open,Id,Type,Infos)).

env_store(Id,Type,Infos,In,Out) :-
    env_store2(In,Id,Type,Infos,Out).
env_store2(env(In),Id,Type,Infos,env(Out)) :-
    avl_store(Id,In,entry(Type,Infos),Out).
env_store2(openenv(Open,In),Id,Type,Infos,openenv(Open,Out)) :-
    env_store2(In,Id,Type,Infos,Out).

% store a callback predicate in the environment
env_store_operator(Id,Module:Callback,In,Out) :-
    % the type is ignored, we just use the info field
    env_store(operator(Id),boolean,[Module:Callback],In,Out).

% the open environment acts like the normal type environment with the
% exception that a lookup of an identifier always works, in case the
% identifier was not in the environment, it will be added to it.
openenv(Env,openenv(_,Env)).
openenv_lookup([Head|Tail],Id,Type,Infos) :-
    Infos=[],
    ( Head = entry(Id,_,_) ->
        Head = entry(Id,Type,Infos)
    ; otherwise ->
        openenv_lookup(Tail,Id,Type,Infos)).
openenv_identifiers(openenv(List,_),Ids) :-
    openenv_identifiers2(List,Ids).
openenv_identifiers2(List,[]) :- var(List),!.
openenv_identifiers2([],[]).
openenv_identifiers2([entry(Id,Type,Infos)|Rest],[TId|IdRest]) :-
    create_texpr(identifier(Id),Type,Infos,TId),
    openenv_identifiers2(Rest,IdRest).

env_identifiers(env(Avl),Ids) :-
    avl_domain(Avl,Ids).
env_identifiers(openenv(Open,Env),Ids) :-
    env_identifiers(Env,Ids1),
    open_env_identifiers(Open,Ids2),
    append(Ids2,Ids1,UIds),sort(UIds,Ids).
open_env_identifiers(L,Ids) :- var(L),!,Ids=[].
open_env_identifiers([],[]).
open_env_identifiers([entry(Id,_,_)|Orest],[Id|Irest]) :-
    open_env_identifiers(Orest,Irest).


% gives a list of all known identifiers and their information in a type
% environment. A second list of the same size is used to build up a new
% environment whith the same identifiers and types but with new information
env_transform_infos(env(Avl),Ids,OldInfos,NewInfos,env(NewAvl)) :-
    % transform a standard environment by converting it to a list,
    % iterate over the list and convert the result back
    avl_to_list(Avl,OldEntries),
    maplist(env_transinfo2,OldEntries,Id_OldInfo_NewInfo,NewEntries),
    maplist(split_slash,Id_OldInfo_NewInfo,Ids,Old_New_Info),
    maplist(split_slash,Old_New_Info,OldInfos,NewInfos),
    ord_list_to_avl(NewEntries,NewAvl).
env_transform_infos(openenv(Open,Env),Ids,OldInfos,NewInfos,openenv(NewOpen,NewEnv)) :-
    % an open environment iterates first over the (open) list,
    % calls then the code for the standard environment and concatenates
    % the generated lists.
    env_transinfo_open(Open,Ids1,OldInfos1,NewInfos1,NewOpen),
    env_transform_infos(Env,Ids2,OldInfos2,NewInfos2,NewEnv),
    append(Ids1,Ids2,Ids),
    append(OldInfos1,OldInfos2,OldInfos),
    append(NewInfos1,NewInfos2,NewInfos).
% iterate over standard environments
env_transinfo2(Id-entry(Type,OldInfo),Id/(OldInfo/NewInfo),Id-entry(Type,NewInfo)).
% iterate over the open list (last element is not [] but a variable)
env_transinfo_open(OpenEnv,[],[],NewInfo,NewEnv) :- var(OpenEnv),!,OpenEnv=NewEnv,NewInfo=[].
env_transinfo_open([],[],[],[],[]).
env_transinfo_open([entry(Id,Type,OldInfo)|ORest],[Id|IdRest],[OldInfo|OIRest],[NewInfo|NIRest],[entry(Id,Type,NewInfo)|NRest]) :-
    env_transinfo_open(ORest,IdRest,OIRest,NIRest,NRest).
split_slash(A/B,A,B).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% fast manual generation of syntax trees
%   this code is only for testing, not for use in the regular type checking
%   of user-given expressions

:- if(current_prolog_flag(dialect,sicstus)).
:- if((current_prolog_flag(version_data,sicstus(4,X,_,_,_)),X<3)).
:- use_module(library(terms),[term_variables/2]). % is built-in in SICSTUS 4.3
:- endif.
:- endif.

fasttype(Expr,TExpr) :-
    term_variables(Expr,OldVars),
    ( Expr= +_  -> fasttype2(Expr,TExpr,unknown)
    ; otherwise -> fasttype2(Expr,TExpr,_Type)),
    term_variables(TExpr,NewVars),
    ( fasttype_check_new_vars(NewVars,OldVars) -> true
    ; add_error(btypechecker, 'fasttype/2 introduced new variables'),fail).
fasttype2(Type<<Expr,b(TExpr,Type,[]),Type) :- !,
    safe_functor(fasttype2,Expr,F,Arity),
    safe_functor(fasttype2,TExpr,F,Arity),
    safe_functor(fasttype2,Pattern,F,Arity),
    ( Pattern :: Type ->
        fasttype_args(Arity,Expr,TExpr,Pattern)
    ; otherwise ->
        fasttype_all_unknown(Arity,Pattern)),
    fasttype_args(Arity,Expr,TExpr,Pattern).
fasttype2(+Expr,TExpr,Type) :- !,
    ( Type==unknown ->
        safe_functor(fasttype2,Expr,F,Arity),
        safe_functor(fasttype2,Pattern,F,Arity),
        ( fasttype_lookup(Pattern,NType) ->
            fasttype2(NType<<Expr,TExpr,NType)
        ; otherwise ->
            add_error(btypechecker,'Not able to simply infer type:',Expr),fail)
    ; otherwise ->
        fasttype2(Type<<Expr,TExpr,Type)).
fasttype2([H|T],[TH|TT],_) :- !,fasttype2(H,TH,unknown),fasttype2(T,TT,unknown).
fasttype2(Expr,Expr,_).

fasttype_all_unknown(0,_) :- !.
fasttype_all_unknown(N,Pattern) :-
    arg(N,Pattern,unknown),
    N1 is N-1, fasttype_all_unknown(N1,Pattern).

fasttype_args(0,_,_,_) :- !.
fasttype_args(N,Expr,NExpr,Pattern) :-
    arg(N,Expr,Arg), arg(N,NExpr,NArg),
    arg(N,Pattern,ArgType),
    fasttype2(Arg,NArg,ArgType),
    N1 is N-1,
    fasttype_args(N1,Expr,NExpr,Pattern).

fasttype_lookup(integer(_),integer).
fasttype_lookup(Pattern,Type) :- Pattern :: Type.

fasttype_check_new_vars(Old,New) :-
    sort(Old,SOld), sort(New, SNew),
    fasttype_check_new_vars2(SOld,SNew).
fasttype_check_new_vars2([],_).
fasttype_check_new_vars2([N|NRest],[O|ORest]) :-
    ( N==O      -> fasttype_check_new_vars(NRest,ORest)
    ; otherwise -> fasttype_check_new_vars([N|NRest],ORest)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% type expressions without declaring the identifiers

% opentype type-checks an expression without having to declare every used
% identifier in advance.
%   Exprs        - a list of raw expressions to be type-checked
%   InitEnv      - an initial environment
%   Identifiers  - used identifiers that weren't in the initial environment
%   DeferredSets - new introduced deferred sets for variables whose type couldn't be infered
%   CleanExprs   - the type-checked (and cleaned-up) list of expressions
opentype(Exprs,InitEnv,Identifiers,DeferredSets,CleanExprs) :-
    openenv(InitEnv,OpenEnv),
    do_typecheck(Exprs,OpenEnv,[],TypedExprs,Errors),
    ( Errors = [] ->
        true
    ; Errors = [_|_] ->
        throw(type_errors(Errors))),
    % all identifier I whose type have the form set(_),
    % introduce implicitly a new type global(I)
    foldl(declare_deferred_sets,TypedExprs,DS1,[]),
    % ensure that all bounded identifiers have ground types
    ground_name_types_by_defsets(TypedExprs,DS2),
    % ensure that all free identifiers have ground types
    openenv_identifiers(OpenEnv,AllIdentifiers),
    ground_free_types_by_defsets(AllIdentifiers,DS3),
    % join all deferred sets
    append([DS1,DS2,DS3],DeferredSets),
    % the returned identifiers should not contain the global sets
    exclude(tid_in_idlist(DeferredSets),AllIdentifiers,Identifiers),
    % clean the expression up
    clean_up_l_wo_optimizations(TypedExprs,[],CleanExprs,opentype).

% used by opentype/4
tid_in_idlist(IdList,TId) :-
    def_get_texpr_id(TId,Id),
    memberchk(Id,IdList).

% used by opentype/4
declare_deferred_sets(TExpr) -->
    {syntaxtraversion(TExpr,Expr,Type,_,Subs,_)},
    declare_deferred_sets2(Expr,Type),
    foldl(declare_deferred_sets,Subs).
declare_deferred_sets2(identifier(Id),Type) -->
    {nonvar(Type),Type=set(SType),var(SType),!,SType=global(Id)},[Id].
declare_deferred_sets2(_,_) --> [].

% used by opentype/4, get all types of introduced identifiers
% (e.g. in quantifiers) and ground them by introducing deferred sets
ground_name_types_by_defsets(TExprs,DefSets) :-
    get_all_name_types_l(TExprs,Types),
    ground_by_defsets(Types,'_NSet',DefSets).
% used by opentype/4, get all types of the given identifiers
% and ground them by introducing deferred sets
ground_free_types_by_defsets(Ids,DefSets) :-
    get_texpr_types(Ids,Types),
    ground_by_defsets(Types,'_FSet',DefSets).
% takes a list of types and ground them by introducing
% deferred sets, each set's name starting with Prefix
ground_by_defsets(Types,Prefix,DefSets) :-
    term_variables(Types,VarTypes),
    foldl(ground_by_defsets2(Prefix),VarTypes,DefSets,1,_).
ground_by_defsets2(Prefix,global(DS),DS,N,N2) :-
    number_chars(N,CN),
    safe_atom_chars(Prefix,CS,ground_by_defsets2_1),
    append(CS,CN,Chars),
    safe_atom_chars(DS,Chars,ground_by_defsets2_2),
    N2 is N+1.

% get the types of all introduced identifiers
% (e.g. in quantifiers)
get_all_name_types(TExpr,Types) :-
    syntaxtraversion(TExpr,_,_,_,Subs,Names),
    get_texpr_types(Names,NTypes),
    append(NTypes,RTypes,Types),
    get_all_name_types_l(Subs,RTypes).
get_all_name_types_l(Exprs,Types) :-
    maplist(get_all_name_types,Exprs,LTypes),
    append(LTypes,Types).

%check_nonvar(Src,V) :- var(V) -> add_error(check_nonvar,'Is a variable: ',Src:V) ; true.

% type-checking of external functions/predicates/substitutions
btype_external_call(FunName,rewrite_protected(TypeParams),FunParams,rewrite_protected(Decl),
                    DeclType,InputType,Pos,Env,TFunParams,TRin,TRout) :-
    add_ext_variables(TypeParams,Env,_TTypeParams,DeclSubenv,TRin,TR1),
    btype(extcall_decl,Decl,DeclSubenv,DeclType,_TypedDecl,TR1,TR2),
    % TODO: check against originally (for Atelier-B) defined function
    % first identify the types of the parameters
    btype_l(FunParams,Env,ParamTypes,TFunParams,TR2,TR3),
    % now check if the function parameters match the input type
    (ParamTypes = []
     ->  /* the external call has no parameters */
         (InputType = empty_type -> TR3=TRout
          ; add_error(btype_external_call,'External Definition has no arguments:',FunName),
            fail)
     ; couplise_list(ParamTypes,JoinedParamType),
       unify_types(InputType,JoinedParamType,Pos,no_expression,TR3,TRout)).
