% (c) 2009-2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(debug,
  [(~~)/1,
   global_debug_flag/0,
   debug_print/1, debug_nl/0,
   debug_print/2, debug_nl/1, /* extra argument with urgency of message 0: not urgent */
   debug_println/2, debug_println/1,
   println/2, println/1, % allow easy commenting in of debug_println statements
   log/1,
   debug_stats/1, print_debug_stats/0,
   
   tcltk_turn_debugging_on/0, tcltk_turn_debugging_on/1,
   tcltk_turn_debugging_off/0,
   debug_mode/1,
   
   print_quoted/1, print_quoted_with_max_depth/2,
   time/1,  % call with timing information, time_call
   time_if_debug/1,
   watch/1, watch/2, watch_det/2, det_check/1, det_check/2,
   nl_time/0,
   
   new_pp/2, new_sol/2, reset_pp/1,
   if_det_check/3,
   
   trace_in_debug_mode/0,
   
   (timer_call)/1, (timer_call)/2, timer_statistics/0]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'This module provides predicates to output debugging information when needed.').

:- meta_predicate ~~(0).
~~(G) :- call(G).
 
 :- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_debug_flag,true)).
global_debug_flag.
:- else.
global_debug_flag :-   /* can be used to control debugging using if pre-processor primitives */
    %% true. %%
     fail. %%
:- endif.

use_timer :-
    %%  fail true. %%
     fail. %%

:- if(debug:global_debug_flag).
% A useful definition that ensures that all standard error exceptions causes the debugger to enter trace mode, is as follows: 
  :- multifile user:error_exception/1.
 user:error_exception(error(_,_)).  % then use ?- debug,go. e.g. for Tcl/Tk Version
:- endif.

:- if(debug:use_timer).
:- use_module(extension('timer/timer'),[timer_call/1,timer_call/2,timer_init/0,timer_statistics/0]).
:- timer_init.  
:- print('Timer initialised. Use debug:timer_statistics. to obtain statistics.'),nl.
:- else.
:- op(300,fx,timer_call). 
:- meta_predicate timer_call(0).
:- meta_predicate timer_call(-,0).
timer_call(X) :- call(X).
timer_call(_PP,Call) :- call(Call).
timer_statistics :- print('*** timer_statistics NOT available; set enable_timer in debug.pl to true'),nl.
:- endif.


debug_println(X) :- debug_print(X), debug_nl.
debug_println(P,X) :- debug_print(P,X), debug_nl(P).
println(_,X) :- print_with_max_depth(X,10),nl.
println(X) :- print_with_max_depth(X,10),nl.

debug_print(X) :- debug_print(6,X).
debug_nl :- debug_nl(6).

:- dynamic debug_print/2, debug_nl/1.

debug_print(_,_X).
debug_nl(_).

:- dynamic debug_mode/1.
debug_mode(off).

:- dynamic debug_level/1.

debug_level(5). /* only messages with priority of 5 or higher are printed */

tcltk_turn_debugging_on :- debug_level(X),tcltk_turn_debugging_on(X).

tcltk_turn_debugging_on(Level) :- atomic(Level), number(Level),
       retract(debug_level(_)),
       assert(debug_level(Level)),
       retract(debug_mode(_)), !, assert(debug_mode(on)),
       retractall(debug_nl(_)),
       assert((debug_nl(U) :- (U < Level -> true ; nl))),
       retractall(debug_print(_,_)),
       assert((debug_print(U,X) :- (U < Level -> true ; print_with_max_depth(X,10)))),
       print('Debugging mode: On :'), print(Level),nl.
tcltk_turn_debugging_on(_).

tcltk_turn_debugging_off :-
       retract(debug_mode(on)), !, assert(debug_mode(off)),
       retractall(debug_nl(_)),assert(debug_nl(_)),
       retractall(debug_print(_,_)), assert(debug_print(_,_)),
       print('Debugging mode: Off'),nl.
tcltk_turn_debugging_off.

trace_in_debug_mode :- (debug_mode(off) -> true ; trace).

print_with_max_depth(X,Max) :- write_term(X,
                            [max_depth(Max),portrayed(true)]).
print_quoted_with_max_depth(X,Max) :- write_term(X,
                            [quoted(true),max_depth(Max),portrayed(true)]).
print_quoted(X) :- write_term(X,
                            [quoted(true),max_depth(0),
                             numbervars(true),portrayed(true)]).


debug_stats(Info) :- debug_mode(on),!,print(Info),print_debug_stats.
debug_stats(_).

print_debug_stats :-
      tools:print_memory_used_wo_gc,
      statistics(walltime,[WT,_]), print(' '),print(WT),print(' ms'),nl.

:- meta_predicate time_if_debug(0).
:- meta_predicate time(0).
time_if_debug(Call) :- debug_mode(off),!,Call.
time_if_debug(Call) :- time(Call).
   
:- use_module(tools_printing,[print_term_summary/1]).
time(Call) :-
   nl,print('calling: '),print_term_summary(Call),
   statistics(runtime,[Start,_]),
   statistics(total_runtime,[StartT,_]),
   statistics(walltime,[StartW,_]),
   call(Call),
   statistics(runtime,[End,_]),
   statistics(total_runtime,[EndT,_]),
   statistics(walltime,[EndW,_]),
   print('exit: '), print_term_summary(Call),
   Tot is End-Start,
   print('Runtime: '), print(Tot), print(' ms'),
   TotT is EndT-StartT,
   print(' Total Runtime: '), print(TotT), print(' ms'),
   TotW is EndW-StartW,
   print(' Walltime: '), print(TotW), print(' ms'),nl.

:- volatile sol_found/2.
:- dynamic det_id/1, sol_found/2.
det_id(0).
gen_det_id(X) :- retract(det_id(X)), X1 is X+1,
 assert(det_id(X1)).

:- meta_predicate watch_det(*,0).
:- meta_predicate watch(0).
:- meta_predicate watch(*,0).
watch_det(Limit,Call) :-
   gen_det_id(ID),
   watch(Limit,Call),
   (retract(sol_found(ID,Nr))
     -> nl,print('*** '), print_term_summary(Call),
        print('*** NON-DETERMINATE SOLUTION #'),
        N1 is Nr+1, print(N1),nl,nl,
        assert(sol_found(ID,N1))  %,trace
      ; assert(sol_found(ID,1))
   ).

:- meta_predicate det_check(0).
:- meta_predicate det_check(0,0).
det_check(Call) :- det_check(Call,true).

det_check(Call,ErrCode) :-
   gen_det_id(ID),
   call(Call),  %tools:print_bt_message(det_check_sol(ID,Call)),
   (retract(sol_found(ID,Nr))
     -> nl,print('*** '), print_term_summary(Call),
        print('*** NON-DETERMINATE SOLUTION #'),
        N1 is Nr+1, print(N1),nl,
        assert(sol_found(ID,N1)),
        call(ErrCode), % trace,
        nl
      ; assert(sol_found(ID,1))
   ).
   
watch(Call) :- watch(30,Call).

watch(Limit,Call) :-
   statistics(runtime,[Start,_]),
   call(Call),
   statistics(runtime,[End,_]),
   Tot is End-Start,
   (Tot>Limit -> nl, %nl,print(Call),nl,
                 print('*** '),print_term_summary(Call),
                 print('*** exceeded limit: '), print(Tot), print(' ms'),nl
   ; true).

% print new line with time info
nl_time :- statistics(runtime,[Start,SinceLast]), statistics(walltime,[WStart,WSinceLast]),
           format(' [~w total, ~w ms (wall ~w, ~w)]~n',[Start,SinceLast,WStart,WSinceLast]).

% ------------------------
% utility similar to det_check; but allows to set spy point for tracing and checks redo

:- volatile pp_nr/1, spy_pp/1, pp_goal/2.
:- dynamic pp_nr/1, spy_pp/1, pp_goal/2.
% debug:reset_pp(XX).
pp_nr(0). pp_goal(0,true).
%spy_pp(115).
reset_pp(Spy) :- retract(pp_nr(_)), assert(pp_nr(0)),
                 retractall(pp_goal(_,_)), assert(spy_pp(Spy)).
:- volatile sol_found/1.
:- dynamic sol_found/1.
new_pp(C,Nr) :- retract(pp_nr(X)), Nr is X+1, assert(pp_nr(Nr)),
          assert(pp_goal(Nr,C)),
          (spy_pp(Nr) -> trace ; true).
new_sol(S,Nr) :- new_sol_no_redo(S,Nr).
new_sol(S,Nr) :- %spy_pp(Nr),
   print('### REDO PP: '), print(Nr),nl,
   print('###  SOL: '), print(S),nl,spy_trace(Nr),
   fail.

new_sol_no_redo(S,Nr) :-
   (sol_found(Nr) -> nl,print('### Non-Deterministic Program Point !'),nl,
                 print('### '), print(Nr),nl,
                 print('### GOAL: '), pp_goal(Nr,G), print(G),nl,
                 print('###  SOL: '),  print(S),nl,
                 nl,trace
               ; assert(sol_found(Nr)), % print(sol_found(Nr,S)),nl,nl,
                 spy_trace(Nr)
   ).

spy_trace(Nr) :- (spy_pp(Nr) -> trace ; true).

% an if predicate which checks if the Test part is deterministic
:- meta_predicate if_det_check(0,0,0).
:- meta_predicate if_det_check_pp(0,0,0,*).
if_det_check(Test,Then,Else) :- new_pp(if_det_check_test(Test),PP),
   if_det_check_pp(Test,Then,Else,PP).
if_det_check_pp(Test,Then,_Else,PP) :-
   Test,
   new_sol_no_redo(Test,PP),
   Then.
if_det_check_pp(_Test,_Then,Else,PP) :- \+ sol_found(PP), Else.

% -------------------------------------------
:- dynamic hit_counter_fact/1.
hit_counter_fact(1).

hit_counter(X) :- retract(hit_counter_fact(X)), X1 is X+1, assert(hit_counter_fact(X1)).

% -------------------------------------------


:- dynamic tdepth/1.
tdepth(0).
:- public trace_point/1.
trace_point(E) :- retract(tdepth(T)), T1 is T+1, assert(tdepth(T1)),
    print_tab(T), print('> ENTER '),print(T), print(' : '),print_term_summary(E),nl.
trace_point(E) :- retract(tdepth(T1)), T is T1-1, assert(tdepth(T)),
    print_tab(T), print('> exit  '), print(T), print(' : '), print_term_summary(E),nl,fail.
:- public print_tabs/0.
print_tabs :- tdepth(T), print_tab(T).
print_tab(0) :- !, print(' ').
print_tab(N) :- N>0, print('|-+-'), N1 is N-1, print_tab(N1).

% -------------------------------------------

% use to log terms into a file (alternative to writeln_log)
log(Term) :-  F='~/problog.pl',
          open(F,append,S),
          write_term(S,Term,[quoted(true)]), write(S,'.'),nl(S),
          close(S).


