% (c) 2013-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(disprover, [disprove/5, disprove_with_opts/6]).

:- use_module(module_information).
:- module_info(group,cbc).
:- module_info(description,'Rodin Prover / Disprover').

% provide access to the Disprover algorithm with all hypotheses, selected hypotheses and goal

:- use_module(extension('user_signal/user_signal'), [user_interruptable_call_det/2]).

:- use_module(solver_interface, [solve_predicate/5,solver_result/4,type_check_in_machine_context/2]).
:- use_module(preferences, [temporary_set_preference/2, reset_temporary_preference/1,
                            get_preference/2, set_preference/2]).
:- use_module(btypechecker,[env_empty/1,add_identifiers_to_environment/3]).
:- use_module(bmachine, [b_get_machine_variables/1, b_get_machine_constants/1]).
:- use_module(b_global_sets, [is_b_global_constant/3,b_global_set/1,
                              contains_unfixed_deferred_set/1]).
:- use_module(bsyntaxtree, [conjunct_predicates/2, create_negation/2,
                            find_typed_identifier_uses/3,
                            find_typed_identifier_uses_l/3,
                            get_texpr_types/2]).
:- use_module(error_manager, [get_error/2,reset_errors/0,add_error/2]).
:- use_module(library(lists)).
:- use_module(tools,[start_ms_timer/1, stop_ms_timer/1]).
:- use_module(debug, [debug_println/2,debug_mode/1]).

disprove(Goal,AllHypotheses,SelectedHypotheses,Timeout,OutResult) :-
    % default options:
    Opts = [use_smt_mode/true,use_clpfd_solver/true,use_chr_solver/true],
    % Note: disprover_mode is always set to true
    disprove_with_opts(Goal,AllHypotheses,SelectedHypotheses,Timeout,Opts,OutResult).

disprove_with_opts(Goal,AllHypotheses,SelectedHypotheses,Timeout,Opts,OutResult) :-
    temporary_set_preference(disprover_mode,true),
    temporary_set_preference(time_out,Timeout),
    start_ms_timer(Timer),
    user_interruptable_call_det(disprove2(Goal,AllHypotheses,SelectedHypotheses,Opts,Result),InterruptResult),
    %write(ir(InterruptResult)), nl,
    catch_internal_timeout(InterruptResult,Result,OutResult),
    (debug_mode(on) -> write(disprove_result(OutResult)),nl,stop_ms_timer(Timer),nl ; true),
    reset_temporary_preference(disprover_mode),
    reset_temporary_preference(time_out).

catch_internal_timeout(interrupted,_Result,interrupted) :-
    reset_errors.
catch_internal_timeout(ok,Result,Out) :-
    get_error(warning(X),Y), atom_codes(Y,Codes), !,
    (append("TIME-OUT occurred while ProB was expanding", _, Codes)
     -> Out = time_out ; Out = Result, add_error(warning(X),Y)).
catch_internal_timeout(ok,Result,Result).

solve_disprover_predicate(Predicate,State,Opts,CBCResult) :-
   TimeoutFactor = 1,
   solve_predicate(Predicate,State,TimeoutFactor,Opts,CBCResult).

:- public disprove2/5. % it is called by user_interruptable_call_det above; we should also add a meta_predicate annotation
:- use_module(bsyntaxtree, [predicate_identifiers/2]).
disprove2(Goal,AllHypotheses,SelectedHypotheses,Opts,ResultOut) :-
    type_check_in_machine_context([Goal|AllHypotheses],[TGoal|TAllHypotheses]),
    type_check_in_machine_context(SelectedHypotheses,TSelectedHypotheses),

    % build predicate H1 & H2 .... & not Goal
    create_negation(TGoal,NegatedGoal),

    %(debug_mode(on) -> write(NegatedGoal), nl ; true),

    (debug_mode(on) -> nl,write('Negated Goal:'), nl,
                             translate:nested_print_bexpr_as_classicalb(NegatedGoal),
                             write_hypothesis(TAllHypotheses)
                          ;  true),

    conjunct_predicates([NegatedGoal|TAllHypotheses],Predicate),

    solve_disprover_predicate(Predicate,State,Opts,CBCResult),
    solver_result(CBCResult,Predicate,State,Result),
    debug_println(20,disprover_result_with_all_hypotheses(Result)),
    
    disprove_double_check(Result,NegatedGoal,TAllHypotheses,TSelectedHypotheses,Opts,ResultOut).

% res <— solve(  All Hyp & not G )
% if res = solution then
%    return CounterExample Found,  Goal definitely not provable
% else if unfixed deferred set used in selected Hypotheses or Goal then
%    return Unknown
% else if res = fail (without time-out & without enumeration warning & no unfixed deferred set in All Hypotheses) then
%    return Proof Found, Goal proven
% else 
%    res <— solve (Selected Hyp & not G)    /* maybe use smaller time-out here ?? */
%    if res = fail (without time-out & without enumeration warning) then
%      return Proof Found, Goal proven
%    else if res = solution then
%      return CounterExample for selected Hypotheses found, Goal not provable from selected Hypotheses (but may be provable with all Hyp)
%    else
%      return Unknown
%    fi
%  fi

% we found a solution / counter example. no need to double check here.
% just clear up the resulting state, keeping only variables occuring in the goal.
disprove_double_check(solution(X),NegatedGoal,_Hypotheses,_SelectedHypotheses,_,ResultOut) :-
    predicate_identifiers(NegatedGoal,GoalIdentifiers),
    findall(binding(Id,Value,PP), (member(binding(Id,Value,PP),X), member(Id,GoalIdentifiers)), FilteredState),
    ResultOut = solution(FilteredState).
% check if the unfixed deferred sets occur in selected hyps or goal
disprove_double_check(no_solution_found(unfixed_deferred_sets),NegatedGoal,_Hypotheses,SelectedHypotheses,_,no_solution_found(unfixed_deferred_sets)) :-
    find_typed_identifier_uses(NegatedGoal,[],GoalIdentifiers),
    get_texpr_types(GoalIdentifiers,GoalTypes),
    find_typed_identifier_uses_l(SelectedHypotheses,[],HypIdentifiers),
    get_texpr_types(HypIdentifiers,HypTypes),
    append(GoalTypes,HypTypes,Types),
    % exclude all types that are not unfixed
    include(contains_unfixed_deferred_set,Types,UnfixedTypes),
    % see if an unfixed type is left
    UnfixedTypes = [_|_],
    debug_println(4,goal_or_sel_hyp_contains_unfixed_deferred_sets(UnfixedTypes)),
    !.
% contradiction was found, i.e. there was neither a timeout nor an enumeration warning
disprove_double_check(contradiction_found,_NegatedGoal,_Hypotheses,_SelectedHypotheses,_,contradiction_found).
disprove_double_check(Result,_NegatedGoal,Hypotheses,SelectedHypotheses,_Opts,ResultOut) :-
    same_conjunction_list(Hypotheses,SelectedHypotheses),!,
    (debug_mode(on) -> nl,write('Selected hypotheses identical to all hypotheses.'),nl ; true),
    Result=ResultOut.
% see what solving only selected hypotheses & not G gives us
disprove_double_check(_OldResult,NegatedGoal,_Hypotheses,SelectedHypotheses,Opts,ResultOut) :-
    % reset existing enumeration warnings,
    % that might have been generated in a previous run
    reset_errors,
    (debug_mode(on) 
      -> nl,write('Trying to use only selected hypotheses.\nNegated Goal:'), nl,
         translate:nested_print_bexpr_as_classicalb(NegatedGoal),
         write_hypothesis(SelectedHypotheses)
       ; true),
    conjunct_predicates([NegatedGoal|SelectedHypotheses],Predicate),
    solve_disprover_predicate(Predicate,State,Opts,CBCResult),
    solver_result(CBCResult,Predicate,State,Result),
    disprove_triple_check(Result,NegatedGoal,ResultOut).
%  we still have a contradiction
disprove_triple_check(contradiction_found,_,contradiction_found).
disprove_triple_check(solution(X),NegatedGoal,ResultOut) :-
    predicate_identifiers(NegatedGoal,GoalIdentifiers),
    findall(binding(Id,Value,PP), (member(binding(Id,Value,PP),X), member(Id,GoalIdentifiers)), FilteredState),
    ResultOut = solution_on_selected_hypotheses(FilteredState).
disprove_triple_check(time_out,_,time_out).
disprove_triple_check(no_solution_found(X),_,no_solution_found(X)).

:- use_module(probsrc(bsyntaxtree),[same_texpr/2]).
% check if two lists of conjuncts represent the same conjunction (modulo info fields)
same_conjunction_list(List1,List2) :-
    same_length(List1,List2,_), % avoid possibly expensive comparisons if lists clearly different
    % we could strip info field and the sort; but here we just want to catch the obvious case where the disprover is called with SelectedHyp = AllHyp
    maplist(same_texpr,List1,List2).

write_hypothesis([]).
write_hypothesis([Hyp|Hyps]) :-
    write(' & /* Hypothesis: */ '), nl,
    %(debug_mode(on) -> write(Hyp), nl ; true),
    translate:nested_print_bexpr_as_classicalb(Hyp),
    write_hypothesis(Hyps).
