% (c) 2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(enable_graph, [print_enable_graph_dot/2, path_enable_graph/8
         ,add_vertex/3, add_neighbour/5, edge_labels/2 % should be deleted (exported just for testing)
         ]).

/* Modules and Infos for the code coverage analysis */
:- use_module(module_information).
:- module_info(group,model_checker).
:- module_info(description,'This module provides operations of directed enable graphs.').

% Standard SICSTUS prolog libraries
:- use_module(library(lists)).

% Classical B prolog modules
:- use_module(probsrc(bmachine),[b_get_machine_operation/4,b_top_level_operation/1]).
:- use_module(probsrc(b_interpreter),[b_test_boolean_expression_wf/3]).
:- use_module(probsrc(bsyntaxtree), [get_texpr_expr/2,safe_create_texpr/3]).

% Event-B prolog modules
:- use_module(probsrc(b_interpreter_eventb),[b_event/6, b_event_with_change_set/7]).

% Prolog modules handling the state space
:- use_module(probsrc(store), [empty_state/1,store_updates_and_normalise/3,normalise_store/2]).

% Prolog modules declaring predicates relevant for partial order reduction optimisation
:- use_module(probsrc(ample_sets),[dependent_actions/3,dependent_act/2]).
%:- use_module(probsrc(enabling_analysis),[enable_analysis/6]).

% Other modules (used mostly for debugging)
:- use_module(probsrc(dot_graph_generator),[gen_dot_graph/6]).
:- use_module(probsrc(translate),[translate_bexpression/2, print_bexpr/1, print_bstate/1]).
:- use_module(probsrc(preferences),[get_preference/2]).

% Importing unit tests predicates
:- use_module(probsrc(self_check)).

/* 
 An enable graph is represented as a list of pairs (vertex-edgelist), the edge list is a list of pairs (neighbor-enabling_predicate).
 The module follows the definition of enable graph from 'Automatic Flow Analysis for Event-B'. An enable graph can be used, for example,
 for the analysis of the event relations in a Event-B machine that is necessary for optimisations like partial order reduction and 
 partial guard evaluation.
*/

% path_enable_graph(+V1,+V2,+N,+ST,+VertexSet,+EnableGraph,-Path,-Length)
% is true if 'Path' is a list of vertices constituting a path of the length 'Length'
% from 'V1' to 'V2' in 'EnableGraph' and thereby each vertex of the 'Path' is
% not in 'VertexSet' and 'ST' is a state (predicate expresison) in respect of which the enable guard of the first 
% edge will be evaluated.
% Special cases: 
% 1. If 'ST' is the empty state and 'VertexSet' is an empty list, then path_enable_graph can be used to
% find some path from 'V1' to 'V2' in 'EnableGraph' without taking into account the enabling predicates of the edges.
% 2. If 'N' is equal to 0, then the predicate will search just for a path connecting 'V1' with 'V2'.

path_enable_graph(V1,V2,N,ST,VertexSet,EnableGraph,Path,Length) :- 
    path(V1,V2,N,ST,VertexSet,EnableGraph,Path,Length).


% Unit tests
:- assert_must_succeed((enable_graph: vertices_edges_predicates_to_egraph([a,b,d,e,q],[a-1-d,a-2-b,d-4-q,e-1-a,e-4-b,q-3-c],R), R = [a-[b-2,d-1],b-[],d-[q-4],e-[a-1,b-4],q-[c-3]])).
:- assert_must_succeed((enable_graph: vertices_edges_predicates_to_egraph([],[a-1-d,a-2-b,d-4-q,e-1-a,e-4-b,q-3-c],R), R = [a-[b-2,d-1],d-[q-4],e-[a-1,b-4],q-[c-3]])).
:- assert_must_succeed((enable_graph: vertices_edges_predicates_to_egraph([],[],R), R = [])).
:- assert_must_succeed((enable_graph: add_neighbour([2-[3-v],4-[1-w,4-p],5-[]],2,q,3,R), R = [2-[3-q,3-v],4-[1-w,4-p],5-[]])).
:- assert_must_succeed((enable_graph: add_neighbour([2-[3-v],4-[1-w,4-p],5-[]],1,q,3,R), R = [1-[3-q],2-[3-v],4-[1-w,4-p],5-[]])).
:- assert_must_succeed((enable_graph: add_neighbour([2-[3-v],4-[1-w,4-p],5-[]],6,q,3,R), R = [2-[3-v],4-[1-w,4-p],5-[],6-[3-q]])).
:- assert_must_succeed((enable_graph: add_vertex([a-2,b-3,d-4],aa,R), R = [a-2,aa-[],b-3,d-4])).
:- assert_must_succeed((enable_graph: add_vertex([a-2,b-3,d-4],'A',R), R = ['A'-[],a-2,b-3,d-4])).
:- assert_must_succeed((enable_graph: add_vertex([],'A',R), R = ['A'-[]])).

vertices_edges_predicates_to_egraph(Vertices,Edges,EnableGraph) :-
    add_vertices([],Vertices,EnableGraph1),
    add_edges(EnableGraph1,Edges,EnableGraph).

 add_vertices(EnableGraph,[],EnableGraph).
 add_vertices(EnableGraph,[V|Vs],NewEnableGraph) :-
    add_vertex(EnableGraph,V,EnableGraph1),
    add_vertices(EnableGraph1,Vs,NewEnableGraph).

%% Using difference lists for constructing the enable graph
add_vertex([],V,[V-[]]).
add_vertex([V-NB|Rest],V,[V-NB|Rest]) :- !.
add_vertex([W-NB|Rest],V,[V-[],W-NB|Rest]) :-
    W @> V,!.
add_vertex([W-NB|Rest],V,[W-NB|S]) :-
    add_vertex(Rest,V,S).

add_edges(EnableGraph,[],EnableGraph).
add_edges(EnableGraph,[From-Predicate-To|Edges],NewEnableGraph) :-
    add_neighbour(EnableGraph,From,Predicate,To,EnableGraph1),
    add_edges(EnableGraph1,Edges,NewEnableGraph).

add_neighbour([],From,Predicate,To,[From-[To-Predicate]]).
add_neighbour([V-Neighbours|Rest],From,Predicate,To,[V-NewNeighbours|Rest]) :-
    From==V,!,
    (member(To-Predicate,Neighbours) ->
       NewNeighbours = Neighbours
    ; otherwise -> %trace,
       sort([To-Predicate|Neighbours],NewNeighbours)
    ).
add_neighbour([V-Neighbours|Rest],From,Predicate,To,[From-[To-Predicate],V-Neighbours|Rest]) :-
    V @> From,!.
add_neighbour([V-Neighbours|Rest],From,Predicate,To,[V-Neighbours|S]) :-
    add_neighbour(Rest,From,Predicate,To,S).

edge_labels(EnableGraph,Labels) :-
    edge_labels(EnableGraph,[],Labels).

edge_labels([],Labels,Labels).
edge_labels([_V-NB|Rest],LabelsSoFar,Labels) :-
    maplist(get_vertex_label,NB,Ls),
    append(LabelsSoFar,Ls,LabelsSoFar1),
    edge_labels(Rest,LabelsSoFar1,Labels).

get_vertex_label(_Dst-P,P).

% Predicates for finding a path from Src to Goal in the enable graph
path(EnableGraph,Src,Goal,Module,Predicate,EvaluationDepth,[Src|Path],Length) :-
    path1(EnableGraph,Src,Goal,Module,Predicate,EvaluationDepth,[],Path,0,Length).

path1(_EnableGraph,Goal,Goal,_Module,_Predicate,_EvaluationDepth,_CurPath,[],Length,Length).
path1(EnableGraph,Src,Goal,Module,Predicate,EvaluationDepth,CurPath,Path,CurLength,Length) :-
    Src\=Goal,
    (EvaluationDepth==0 -> EvaluatePredicate=fail; EvaluatePredicate=true),
    predicate_edge(EnableGraph,Src,Next,CurPath,Module,Predicate,EvaluatePredicate,NewPredicate),
    NewLength is CurLength +1,
    decrease_evaluation_depth(EvaluationDepth,NewEvaluationDepth),
    Path = [Next|PathTail],
    path1(EnableGraph,Next,Goal,Module,NewPredicate,NewEvaluationDepth,[Next|CurPath],PathTail,NewLength,Length).

decrease_evaluation_depth(EvaluationDepth,NewEvaluationDepth) :-
    (EvaluationDepth>0 ->
        NewEvaluationDepth is EvaluationDepth -1
    ;   NewEvaluationDepth=EvaluationDepth
    ).

% Checks if there is an edge between V1 and V2 and additionally tests whether the 
% Label (e.g. the enabling predicate) evaluates to true. If one of both is not 
% fullfilled then the call should fail.
predicate_edge(EnableGraph,V1,V2,CurPath,Module,Predicate,EvaluatePredicate,NewPredicate) :-
    edge(EnableGraph,V1,V2,P),
    nonmember(V2,CurPath),
    call(Module:Predicate,P,EvaluatePredicate,V1,V2,NewPredicate).

edge(EnableGraph,V1,V2,Predicate) :-
    member(V1-NB,EnableGraph),
    member(V2-Predicate,NB).

% TODO: The implementation should be removed to another module
% In the context of POR the DoNotVisit set is the set of the currently computed ample set.
% predicate(State,DoNotVisit,++EvaluatePredicate,)
evaluate_weakest_precondition_for_por(State,DoNotVisit,P,EvaluatePredicate,OpName1,OpName2,NewPredicate) :-
    nonmember(OpName2,DoNotVisit),
    empty_state(LocalState),
    ((EvaluatePredicate,State\=[]) ->
        % V2 is enabled after V1, we need to execute V1 in oder to generate the next state
        % TODO: maybe we should set a sort of time-out to decrease the overhead in some cases
        !,
        %% print_bexpr(P),nl,
        %% print_bstate(State),nl,
        b_test_boolean_expression_wf(P,LocalState,State),
        % the success of the predicate should guarantee that no well-definedness-errors occur 
        % when OpName1 is executed
        %% print(b_test_boolean_expression_wf_true),nl,
        %% print(execute_operation(OpName1)),nl,
        execute_operation(OpName1,LocalState,State,NextState) % TODO: handle case for non-deterministic operations
        %% ,print_bstate(NextState),nl
    ; otherwise -> % just checking whether there is an edge between V1 and V2
        NextState=[] % TODO: what should we give as next state
    ),
    NewPredicate =.. [evaluate_weakest_precondition_for_por,NextState,DoNotVisit].

execute_operation(OpName,LocalState,State,NewState) :-
    %% (OpName = 'T3writebus' -> trace; true),
    b_get_machine_operation(OpName,_Res,Par,Body),
    %% print(expr(Body)),nl,
    get_texpr_expr(Body,Expr),
    (Par = [] ->
         execute_expression(OpName,Expr,LocalState,State,Updates)
    ;    b_interpreter: b_execute_top_level_operation_update(OpName,_FullOp,State,Updates,_Path)
         %b_interpreter: b_execute_operation_with_parameters(OpName,State,Par,Updates,Res,_Path,_WF,output_required)
    ),
    %print(executed_operation(OpName)),nl,
    % guarantee that Updates is not a variable here
    %% print(states(Updates,State)),nl,
    store_updates_and_normalise(Updates,State,NewState).

% do not evaluate the precondition
execute_expression(_OpName,precondition(_Condition,Body),LocalState,State,Updates) :- !,
    b_interpreter: b_execute_inner_statement(Body,LocalState,State,Updates,_WF,_IPath,_OR).
execute_expression(OpName,Stmt,_LocalState,State,Updates) :-
    functor(Stmt,rlevent,11),!,
    % Stmt = rlevent(Name,Section,_Status,_Params,Guard,_Theorems,_Actions,_VWitnesses,_PWitnesses,_Unmod,_AbstractEvents)
    assert(user: do_not_evaluate_guard),
    %safe_create_texpr(Stmt,event,Event),
    %b_event(Event,[],[],State,OutputState,_Path),
    %b_event_with_change_set(Event,[],[],State,_ChangeSet,Updates,_Path),
    b_interpreter:b_execute_top_level_operation_update(OpName,_Operation,State,Updates,_PathInfo),
    retractall(user: do_not_evaluate_guard).
execute_expression(Stmt,LocalState,State,Updates) :-
    b_interpreter: b_execute_statement(Stmt,LocalState,State,Updates,_WF,_Path).
% TODO: export b_execute_inner_statement/7 and b_execute_statement/6 in b_interpreter.pl module

:- dynamic asserted_enable_graph/1.

print_enable_graph_dot(EnableGraph,File) :- 
    assert(asserted_enable_graph(EnableGraph)),
    gen_dot_graph(File,enable_graph,eop_node_predicate,enable_trans_predicate,none,none),
    retractall(asserted_enable_graph(_)).

eop_node_predicate(NodeID,SubGraph,NodeDesc,Shape,Style,Color):- 
    enabling_analysis: eop_node_predicate(NodeID,SubGraph,NodeDesc,Shape,Style,Color).

enable_trans_predicate(NodeID,Label,SuccID,Color,Style) :-
    b_top_level_operation(OpName1),
    b_top_level_operation(OpName2),
    asserted_enable_graph(EnableGraph),
    edge(EnableGraph,OpName1,OpName2,Predicate),
    translate_bexpression(Predicate,Res),
    translate_enable_res(Res,NodeID,SuccID,Label,Color,Style).
