% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(enabling_analysis,[reset_enabling_analysis/0,
                             tcltk_cbc_cfg_analysis/1, tcltk_dot_cfg_analysis/1,
                             tcltk_cbc_enabling_analysis/1, tcltk_cbc_enabling_analysis/2,
                             tcltk_cbc_simultaneous_enabling_analysis/1,
                             tcltk_cbc_dependence_analysis/1, print_enable_table/1,
                             tcltk_create_enable_graph/1,tcltk_create_enable_graph_por/3,
		                     tcltk_create_dependence_graph/1,
                             dependent_actions/4, disable_graph/1,
			                 b_get_read_write/3,b_get_read_write/5,
			                 compute_action_relations_graph/2, enable_analysis/6,tcltk_enabling_analysis/2,tcltk_disabling_analysis/1,
			                 feasible_operation/2, feasible_operation/4,
			                 feasible_operation_cache/2,infeasible_operation/1, infeasible_operation_cache/1,
			                 operation_sequence_possible/3, operation_can_be_enabled_by/3]).

% --------------

:- use_module(bmachine,            [b_get_machine_operation/4, b_get_operation_normalized_read_write_info/3, b_top_level_operation/1,
                                    b_get_invariant_from_machine/1, b_get_properties_from_machine/1]).
:- use_module(b_state_model_check, [get_negated_guard/3,get_guard/2]).
:- use_module(sap).
:- use_module(library(ordsets)).
:- use_module(library(lists)).
:- use_module(error_manager).




% -------------------- BINARY STATIC ANALYSIS -------------------------

% certain static analyses used by CBC Test case generation

% check if OpName1 can influence the truth value of the guard of OpName2 by looking at read/write info
activation_dependent(OpName1,OpName2) :-
    b_top_level_operation(OpName1),
    b_get_read_write(OpName1,_GuardReads1,_ActionReads1_,_,Writes1),
    b_top_level_operation(OpName2),
    b_get_read_write(OpName2,GuardReads2,_ActionReads2_,_,_Writes2),
    (ord_intersect(Writes1,GuardReads2) -> true
     ; %print(indep(OpName1,Writes1,OpName2,GuardReads2)),nl,
       fail
     ).


% return Res=impossible if from Invariant we can never execute OpName1 followed by OpName2
operation_sequence_possible(OpName1,OpName2,Res) :-
    \+ activation_dependent(OpName1,OpName2),
    !,
    Res = activation_independent.
operation_sequence_possible(OpName1,OpName2,Res) :-
  % we assume OpName2 is feasible on its own
    get_negated_guard(OpName2,PosGuard2,_NegGuard2),
    Timeout1 = 200,
    % print('Guard: '),translate:print_bexpr(PosGuard2),nl,
    (testcase_path_timeout_catch(invariant,Timeout1,[OpName1],PosGuard2,_Csts,_Ops,_TestS,_TI,Res)
      -> true
      ;  Res = impossible
    ).


% return whether OpName1 can enable a previously disabled OpName2
operation_can_be_enabled_by(OpName1,OpName2,Res) :-
    b_top_level_operation(OpName1),
    b_get_read_write(OpName1,_Reads1,Writes1),
    b_top_level_operation(OpName2),
    format("~nCHECKING if ~w (~w) can enable ~w~n",[OpName1,Writes1,OpName2]),
    get_negated_guard(OpName2,PosGuard2,_NegGuard2),
    filter_predicate(PosGuard2,Writes1,FilteredPosGuard2),
    create_negation(FilteredPosGuard2,FilteredNegGuard2),
    TIMEOUT = 500,
   % print('Before: '), translate:print_bexpr(FilteredNegGuard2),nl,print('After: '), translate:print_bexpr(PosGuard2),nl,
    testcase_path_timeout_catch(pred(FilteredNegGuard2),TIMEOUT,[OpName1],PosGuard2,_,_,_,_,Res),
    print(result(Res)),nl.
    
    

% -------------------- CFG ANALYSIS -------------------------
:-use_module(dot_graph_generator,[gen_dot_graph/6]).
tcltk_dot_cfg_analysis(File) :-
       gen_dot_graph(File,enabling_analysis,eop_node_predicate,cfg_trans_predicate,none,none).

:- use_module(tools,[ajoin/2, cputime/1]).
op_node_predicate(NodeID,SubGraph,NodeDesc,Shape,Style,Color) :- SubGraph=none,
     Shape=record, Style=none, 
     b_top_level_operation(Op), NodeID = Op,
     %NodeDesc = NodeID,
     b_get_read_write(Op,Reads1,_ReadsAct,Reads2,Writes1),
     insert_commas(Reads1,0,R1), insert_commas(Reads2,0,R2), insert_commas(Writes1,0,W1),
     append(W1,['}|'],W2),
     append(R1,['\\n|reads: '|R2],Reads),
     append(Reads,['\\n|writes: '|W2],As),
     ajoin(['|{',Op,'\\n|reads (guard): '|As],NodeDesc),
     Color=blue.
     
insert_commas([],_,[]).
insert_commas([H],_,R) :- !, R=[H].
insert_commas([H|T],N,[H,Sep|IT]) :- 
    (N>5 -> N1=0, Sep=',\\n' ; N1 is N+1, Sep=', '),
    insert_commas(T,N1,IT).

cfg_trans_predicate(NodeID,Label,SuccID,Color,Style) :- 
    cbc_quick_cfg_analysis(NodeID,SuccID,Res),
    translate_res(Res,NodeID,SuccID,Label,Color,Style).

translate_res(cannot_enable,_,_,_,_,_) :- !, fail.
translate_res(syntactic_independent,X,Y,'independent',yellow,bold) :- !,X=Y. % only show if source & dest identical
translate_res(race_dependent,X,Y,'race_independent',lightgray,bold) :- !,X=Y. % only show if source & dest identical
translate_res(possible,_,_,'',black,solid) :- !.
translate_res(timeout_possible,_,_,'',tomato,solid) :- !.
translate_res(X,_,_,X,red,dashed).

% compute a Control Flow Graph very quickly
% ideal when we have program counter style variables
tcltk_cbc_cfg_analysis(list([list(['Origin'|SOps])|Result])) :- 
  findall(Op, b_top_level_operation(Op), Ops), sort(Ops,SOps),
  findall(list([OpName1|EnableList]),
     (b_get_sorted_op(OpName1),
      findall(Possible,cbc_quick_cfg_analysis(OpName1,_OpName2,Possible),EnableList)),
     Result),
  print_enable_table([list(['Origin'|Ops])|Result]).

/*  TO DO: complete into a determinacy analysis ?
cbc_quick_det_analysis(OpName1,OpName2,Res) :-
    b_top_level_operation(OpName1),  % top_level
    ((b_get_sorted_op(OpName2), OpName2 \= OpName1,
   % get_negated_guard(OpName1,PosGuard1,NegGuard1),
    get_negated_guard(OpName2,PosGuard2,NegGuard2),
    sap:testcase_path_timeout(pred(PosGuard2),1200,[OpName1],b(truth,pred,[]),_,_,_,_,Res))
     -> format(user_output,'Operation ~w can be simultaneously enabled with ~w (~w).~n',[OpName1,OpName2,Res])
     ;  format(user_output,'Operation ~w cannot be simultanously enabled with another operation.~n',[OpName1]), Res=det
     ).
*/

cbc_quick_cfg_analysis(OpName1,OpName2,Res) :-
   OpName1='INITIALISATION',
   b_top_level_operation(OpName2),
   b_get_read_write(OpName2,Reads2,Writes2),
   format(user_output,"COMPUTING CFG: INITIALISATION  -->  ~w r:~w / w:~w~n~n",
				 [OpName2,Reads2,Writes2]),
    ( sap:testcase_path_timeout(init,250,[OpName2],b(truth,pred,[]),_Constants,_Ops,_TestS,_TI,R1)
      ->  format(user_output," ~w can be enabled by ~w (~w)!~n",[OpName2,OpName1,R1]),
		  (is_timeout(R1) -> Res=timeout_possible; Res = possible)
	    ; format(user_output," ~w cannot be enabled by ~w!~n",[OpName2,OpName1]),
		  Res = cannot_enable
	).
cbc_quick_cfg_analysis(OpName1,OpName2,Res) :-
    b_top_level_operation(OpName1),  % top_level
    b_get_read_write(OpName1,Reads1,Writes1),
    %get_negated_guard(OpName1,PosGuard1,NegGuard1),
    b_get_sorted_op(OpName2),
    (b_get_read_write(OpName2,Reads2,Writes2),
		format(user_output,"COMPUTING CFG: ~w r:~w / w:~w  -->  ~w r:~w / w:~w~n~n",
				 [OpName1,Reads1,Writes1,OpName2,Reads2,Writes2]),
		(\+ ord_intersect(Writes1,Reads2)
		  -> (ord_intersect(Writes1,Writes2)
		      -> Res = race_dependent, format(user_output," ~w cannot be enabled/disabled by ~w!~n",[OpName2,OpName1])
		       ; Res = syntactic_independent, format(user_output," ~w cannot be enabled/disabled/modified by ~w!~n",[OpName2,OpName1])
		      )
		  ;  get_negated_guard(OpName2,PosGuard2,NegGuard2),
		     garbage_collect,
		     print('Guard: '), translate:print_bexpr(PosGuard2),nl,
			 ((sap:testcase_path_timeout(typing(NegGuard2),60,[OpName1],PosGuard2,_,_,_,_,_R0), % quick check without invariants & properties
			 % TO DO: project constants, variables onto the ones really needed
			   sap:testcase_path_timeout(pred(NegGuard2),250,[OpName1],PosGuard2,_Constants,_Ops,_TestS,_TI,R1))
			   ->  format(user_output," ~w can be enabled by ~w (~w)!~n",[OpName2,OpName1,R1]),
				   (is_timeout(R1) -> Res=timeout_possible; Res = possible)
			  ; format(user_output," ~w cannot be enabled by ~w!~n",[OpName2,OpName1]),
			    Res = cannot_enable)
		)
    ).

%use_module(enabling_analysis),enabling_analysis:cbc_quick_cfg_analysis('SEQUENCER2','SEQUENCER3',Res).
%use_module(sap),use_module(enabling_analysis),enabling_analysis:cbc_quick_cfg_analysis('COMPUTE_SDDBs_points','prop_COMPUTE_0',Res).
%use_module(sap),use_module(enabling_analysis),enabling_analysis:cbc_quick_cfg_analysis('COMPUTE_SDDBs_points','COMPUTE_SDDBs_length',Res).

% -------------------- DEPENDENCE ANALYSIS ----------------------
/*

Independency between two actions in a B/Event-B model can be expressed also by means of LTL-formulas:
    1. Two actions a and b are independent if the following LTL-formula is satisfied by the model:
           "G ((e(b) & [a] => X e(b)) & (e(a) & [b] => X e(a)))" (ind)
    2. If (ind) is violated by the model then a and b are dependent actions.

Note: The race_dependent condition cannot be covered by the above LTL-formula. If two actions are race dependent, but never simultaniously enabled in a state
      from the state space, then (ind) will be satisfied.

*/

tcltk_cbc_dependence_analysis(list([list(['Origin'|SOps])|Result])) :- 
  %findall(Op, b_get_machine_operation(Op,_,_,_), Ops), sort(Ops,SOps),
  findall(Op, b_top_level_operation(Op), Ops), sort(Ops,SOps),
  findall(list([OpName1|EnableList]),
     (b_get_sorted_op(OpName1),
      findall(Enable,cbc_dependence_analysis(OpName1,_OpName2,Enable),EnableList)),
     Result).
  %,print_enable_table([list(['Origin'|Ops])|Result]).

b_get_sorted_op(Op) :- findall(Op, b_top_level_operation(Op), Ops),
    sort(Ops,SOps), member(Op,SOps).

cbc_dependence_analysis(OpName1,OpName2,Res) :-
    b_top_level_operation(OpName1),
    b_get_read_write(OpName1,GReads1,AReads1,Reads1,Writes1),
    get_negated_guard(OpName1,PosGuard1,NegGuard1),
    b_get_sorted_op(OpName2),
    (OpName1=OpName2 -> action_dependent_to_itself(OpName1,GReads1,Writes1,Res) % TO DO: check if it is possible that for two different parameter values of the same event there is dependence
      ; OpName2 @< OpName1 -> Res = '-'  % our checking is symmetric; only check one pair
      ; b_get_read_write(OpName2,GReads2,AReads2,Reads2,Writes2),
		format("CHECKING DEPENDENCE: ~w gr:~w / ar:~w / w:~w  <-->  ~w gr:~w / ar:~w / w:~w~n",
				 [OpName1,GReads1,AReads1,Writes1,OpName2,GReads2,AReads2,Writes2]),
		(syntactical_independence(Reads1,Writes1,Reads2,Writes2)
		  -> Res = syntactic_independent
		  ;  ord_intersect(Writes1,Writes2) -> Res = race_dependent
		     % TO DO: in this case check if there is indeed a race dependence (e.g. in scheduler new and ready_active are actually independent !)
		     % Set up [OpName1,OpName2], [OpName2,OpName1] and see if for ord_intersect(Writes1,Writes2) we can find different values
		  ;  (\+ord_intersect(AReads1,Writes2),\+ord_intersect(AReads2,Writes1) ->
                 get_negated_guard(OpName2,PosGuard2,NegGuard2),
                 conjunct_predicates([PosGuard1,PosGuard2],GuardsConj),
			     (sap:testcase_path_timeout(pred(GuardsConj),300,[OpName2],NegGuard1,_Constants,_Ops,_TestS,_TI,R1)
			       ->  format(" ~w may disable ~w (~w)!~n",[OpName2,OpName1,R1]),
				       (is_timeout(R1) -> Res=timeout_dependent; Res = dependent)
			       ; % TO DO: check for enumeration warnings !
			         sap:testcase_path_timeout(pred(GuardsConj),300,[OpName1],NegGuard2,_,_,_,_,R2)
				       -> format(" ~w may disable ~w (~w)!~n",[OpName1,OpName2,R2]),
				       (is_timeout(R2) -> Res=timeout_dependent; Res = dependent)
			      ; Res = independent)
             ; Res = dependent)
		)
    ).

% --------------------------- Used for the ample sets (Begin) -----------------------------%

:- use_module(bsyntaxtree, [find_identifier_uses/3, find_identifier_uses_l/3, conjunct_predicates/2]).

dependent_actions(OpName1,OpName2,Timeout,Res) :-
    b_get_read_write(OpName1,GReads1,AReads1,Reads1,Writes1),
    get_negated_guard(OpName1,PosGuard1,NegGuard1),
    (OpName1=OpName2 -> action_dependent_to_itself(OpName1,GReads1,Writes1,Res)
      ; OpName2 @< OpName1 -> Res = '-'  % our checking is symmetric; check only one pair
      ; b_get_read_write(OpName2,GReads2,AReads2,Reads2,Writes2),
		(ord_intersect(Writes1,Writes2) 
			-> Res = race_dependent
        ; (\+ ord_intersect(Writes1,Reads2),\+ ord_intersect(Writes2,Reads1))
		    -> Res = syntactic_independent
		  ;  (get_preference(use_cbc_analysis,true) -> % sometimes we don't want to use the CBC
                 (\+ord_intersect(AReads1,Writes2),\+ord_intersect(AReads2,Writes1) ->
                     get_negated_guard(OpName2,PosGuard2,NegGuard2),
                     conjunct_predicates([PosGuard1,PosGuard2],GuardsConj),
				     ((ord_intersect(GReads1,Writes2), catch_and_ignore_well_definedness_error(
                                                      sap:testcase_path_timeout(pred(GuardsConj),Timeout,[OpName2],NegGuard1,_,_,_,_,_R1),_Result1))
				   	                                  -> Res = dependent
				     ;(ord_intersect(Writes1,GReads2),catch_and_ignore_well_definedness_error(
                                                      sap:testcase_path_timeout(pred(GuardsConj),Timeout,[OpName1],NegGuard2,_,_,_,_,_R2),_Result2))
					                                  -> Res = dependent
				     ; Res = independent
				     )
		     ; otherwise ->
				  Res=dependent
		     )
         ; otherwise ->
                  Res=dependent
         )
		)
    ).

:- use_module(b_interpreter,[b_get_machine_operation_for_animation/6]).
action_dependent_to_itself(OpName,Reads,Writes,Res) :-
    b_get_machine_operation_for_animation(OpName,_Res,Params,_,_,true),
    length(Params,Len),
    (Len>0,\+ord_intersect(Reads,Writes) ->
	Res = self_independent
    ; Len = 0 ->
	Res = '='
    ; otherwise ->
	Res = self_dependent
    ).

/*
syntactic_independent_actions(OpName1,OpName2) :-
    b_get_read_write(OpName1,GReads1,_Reads1_,Writes1),
    b_get_read_write(OpName2,GReads2,_Reads1_,Writes2),
    syntactical_independence(GReads1,Writes1,GReads2,Writes2).
*/

syntactical_independence(Reads1,Writes1,Reads2,Writes2) :-
    \+ ord_intersect(Writes1,Reads2), % op1 does not modify guard/effect of op2
    \+ ord_intersect(Writes2,Reads1), % op2 does not modify guard/effect of op1
    \+ ord_intersect(Writes1,Writes2). % no race condition

% --------------------------- Used for the ample sets (End) -----------------------------%


% -------------- ENABLING ANALYSIS --------------

/*

The meaning of the following comments here is to give an idea how to validate the results from the Enabling analysis.
We use here the Linear Time Logic (proposed by Amir Pnueli) in order to give LTL-formulas to each possible table entry result that can prove that
the relation results between the operations of machine M produced by the Enabling analysis are correct. For instance, 
if the Enabling analysis says that it is impossible operation B to be preceded by operation A in the state space of M (i.e. the result in 
cell (A,B) of the table is 'impossible') then this property can be expressed by the LTL-formula "G ([a] => not X e(b))" and checked by
the LTL model checker of ProB.

* impossible:
    The impossibility an action 'a' to enable action 'b' can be expressed by means of LTL-formulas as follows:
            "G ([a] => not X e(b))" (dis)

    If a disables b then (dis) should be satisfied by the model. The result of the enabling analysis 
    returns 'impossible' in cell (a,b) of the Enabling analysis table.

* guaranteed:
    Action 'a' guaranteed enables action 'b' if there is an execution in the state space where after 'a' follows 'b' and 
    'a' and 'b' are never enabled simultaneously in the same state:
            "G ( ([a] => X e(b)) & (not (e(a) & e(b))) )" (en)

* keep:
    Action 'a' keeps action 'b' enabled:
                "(G ((e(b) & [a]) => X e(b))) or (G ((not e(b) & [a]) => X not e(b)))" (keep)
*/

:- dynamic cbc_enable_analysis_cache/4.

% check which operations can be enabled after executing another operation
% to do: move maybe to another module; provide proper CSV export using Sebastian's modules
cbc_enable_analysis(OpName1,OpName2,Enable,ExtraTimeout) :- 
   init_or_op(OpName1),
   if(cbc_enable_analysis_cache(OpName1,OpName2,Result,ExtraTimeout),
      Result = Enable, % already computed
      (% now compute all enablings form OpName1
       cbc_enable_analysis_calc(OpName1,Op2,R,ExtraTimeout),
       assert(cbc_enable_analysis_cache(OpName1,Op2,R,ExtraTimeout)),
       fail
       ;
       % now look up:
       cbc_enable_analysis_cache(OpName1,OpName2,Enable,ExtraTimeout)
      )).


cbc_enable_analysis_calc(OpName1,OpName2,Enable,ExtraTimeout) :- 
   OpName1='INITIALISATION',print('CHECKING ENABLING AFTER INITIALISATION'),nl,
   b_top_level_operation(OpName2),
    print('INITIALISATION'), print(' ---> '), print(OpName2), print('  :: '),
    Timeout is 200+ExtraTimeout,
    ( sap:testcase_path_timeout(init,Timeout,[OpName2],b(truth,pred,[]),_Constants,_Ops,_TestS,_TI,R)
      -> print(R), print(' : '),
         get_negated_guard(OpName2,_,NegGuard),
         (sap:testcase_path_timeout(init,Timeout,[],NegGuard,_Constants2,_Ops2,_TestS2,_TI2,R2)
           -> print(R2), print(' : '),Enable=possible
            ; Enable=guaranteed)
      ;  Enable=impossible),
      print(Enable),nl.
cbc_enable_analysis_calc(OpName1,OpName2,Enable,ExtraTimeout) :-
    Timeout1 is 300+ExtraTimeout,
    Timeout2 is 200+ExtraTimeout,
    b_top_level_operation(OpName1),
    b_get_read_write(OpName1,Reads1,_AReads1_,_Reads1_,Writes1),
    %get_negated_guard(OpName1,PosGuard1,_NegGuard1),
    format("CHECKING ENABLING AFTER: ~w r:~w / w:~w~n",[OpName1,Reads1,Writes1]),
    b_top_level_operation(OpName2),
    b_get_read_write(OpName2,Reads2,_AReads2_,_Reads2_,Writes2),
    get_negated_guard(OpName2,PosGuard2,NegGuard2),
    filter_predicate(PosGuard2,Writes1,FilteredPosGuard2),
    create_negation(FilteredPosGuard2,FilteredNegGuard2),
    print(OpName1), print(' ---> '), print(OpName2), print('  :: '),
    (\+ ord_intersect(Writes1,Reads2) -> WR=independent, print('(wr_indep) ')
      ; WR=dependent, print('(wr_dep) ')),
    ((WR=independent,\+ ord_intersect(Reads1,Reads2))
     -> (\+ ord_intersect(Writes1,Writes2) -> Enable = independent
         ; Enable = activation_indepdendent  /* independent for enabling, but effect of OpName2 can be modified */
        )
        ;otherwise -> 
% sap:testcase_path_timeout(invariant,Timeout1,[OpName1,OpName2],b(truth,pred,[]),_Csts,_Ops,_TestS,_TI,R)
      (sap:testcase_path_timeout(invariant,Timeout1,[OpName1],PosGuard2,_Csts,_Ops,_TestS,_TI,R) % advantage over version with [OpName1,OpName2] : one less state to setup and enumerate; but inner guards may not be checked
      -> print(enabled_after(R)), print(' : '),
         %print('Pos Guard: '),translate:print_bexpr(PosGuard2),nl,
         %print('Neg Guard: '),translate:print_bexpr(NegGuard2),nl,
         (sap:testcase_path_timeout(invariant,Timeout1,[OpName1],NegGuard2,_Csts2,_Ops2,_TestS2,_TI2,R2)
           -> print(disabled_after(R2)), print(' : '),
            % TO DO: test if NegGuard2 holds initially it is possible to execute [OpName1,OpName2]; if not: OpName1 cannot enable OpName2, only keep it -> Enable=keep_possible
            % then we could check if OpName1 can disable OpName2: PosGuard2,[OpName1],NegGuard2
              (WR=dependent 
              -> % b_get_invariant_from_machine(INV), % assume invariant holds ?
              ((OpName1\=OpName2, % otherwise OpName2 must be enabled if [OpName1] can be executed
                %sap:testcase_path_timeout(pred(NegGuard2),Timeout2,[OpName1,OpName2],b(truth,pred,[]),_,_,_,_,R3)
                sap:testcase_path_timeout(invariant,Timeout2,[OpName1],PosGuard2,_,_,_,_,R3))
               -> print(enable(R3)), print(' : '),
                 %nl, translate:print_bexpr(PosGuard2),nl,print(OpName1),nl,translate:print_bexpr(NegGuard2),nl,
                 %print('FILTERED: '), translate:print_bexpr(FilteredNegGuard2),nl,
                  (sap:testcase_path_timeout(pred(PosGuard2),Timeout2,[OpName1],FilteredNegGuard2,_,_,_,_,R4) 
                  -> print(disable(R4)), print(' : '),
                     (contains_timeout([R,R2,R3,R4]) -> Enable=timeout_possible ; Enable=possible)
                  ;  (contains_timeout([R,R2,R3]) -> Enable=timeout_possible_enable ; Enable=possible_enable) /* Opname cannot disable OpName2; only enable it */
                  )
               ; /* OpName1 cannot enable OpName2; only preserve it */
                  (sap:testcase_path_timeout(pred(PosGuard2),Timeout2,[OpName1],FilteredNegGuard2,_,_,_,_,R4) 
                  -> print(disable(R4)), print(' : '),
                     (contains_timeout([R,R2,R4]) -> Enable=timeout_possible_disable; Enable=possible_disable)
                  ;  (contains_timeout([R,R2]) -> Enable=timeout_keep; Enable=keep) /* Opname cannot enable or disable */
                  )
              )
              ; % possible after, can also be disabled after but op1 cannot modify guard variables of op2
                (contains_timeout([R,R2])
                 -> Enable=timeout_keep
                 ; Enable=keep))  % was possible_act_indep
            ; (is_timeout(R) -> Enable=timeout_guaranteed; Enable=guaranteed))
      ;  Enable=impossible)
     ),
     print('Enable='),print(Enable),nl.

% check if contains timeout or similar result
contains_timeout(List) :- (member(timeout,List) -> true ; member(clpfd_overflow,List) -> true).
is_timeout(R) :- (R==timeout -> true ; R==clpfd_overflow -> true).

:- use_module(bsyntaxtree,[conjunction_to_list/2,some_id_occurs_in_expr/2,create_negation/2]).
% remove all predicates which are not modified
filter_predicate(Pred,ModifiedVars,FilteredPred) :-
    conjunction_to_list(Pred,List),
    include(some_id_occurs_in_expr(ModifiedVars),List,FilteredList),
    conjunct_predicates(FilteredList,FilteredPred).

predicate_modified_by_write_vars(Pred,SortedWriteVars) :-
    some_id_occurs_in_expr(SortedWriteVars,Pred).



tcltk_cbc_enabling_analysis(L) :- tcltk_cbc_enabling_analysis(0,L).
tcltk_cbc_enabling_analysis(ExtraTimeout,list([list(['Origin'|Ops])|Result])) :- 
  findall(Op, b_top_level_operation(Op), Ops),
  %statistics(walltime,[WT1,_]),
  findall(list([OpName1|EnableList]),
     (init_or_op(OpName1),
      findall(Enable,cbc_enable_analysis(OpName1,_OpName2,Enable,ExtraTimeout),EnableList)),
     Result).
  %statistics(walltime,[WT2,_]), Time is WT2-WT1,
  %format('Runtime for enabling analysis: ~w ms~n',[Time])
  %,print_enable_table([list(['Origin'|Ops])|Result]).


% ----------------------------------------------

% --------------------------- Enabling anaysis used for the computation of ample sets (Begin) -----------------------------%

:- use_module(ample_sets,[dependent_act/2,clear_dynamic_predicates_for_POR/0,compute_dependendency_relation_of_all_events_in_the_model/3]).

enable_analysis(OpName1,OpName2,_FindInvViolations,_UseEnableGraph,Timeout,Enable) :-
   OpName1='INITIALISATION',%print('CHECKING ENABLING AFTER INITIALISATION'),nl,
   b_top_level_operation(OpName2),
    %print('INITIALISATION'), print(' ---> '), print(OpName2), print('  :: '),
    ( sap:testcase_path_timeout(init,Timeout,[OpName2],b(truth,pred,[]),_Constants,_Ops,_TestS,_TI,_R)
      -> %print(R), print(' : '),
         get_negated_guard(OpName2,_,NegGuard),
         (sap:testcase_path_timeout(init,Timeout,[],NegGuard,_Constants2,_Ops2,_TestS2,_TI2,_R2)
           -> %print(R2), print(' : '),
              Enable=possible
            ; Enable=guaranteed)
      ;  Enable=impossible).
      %,print(Enable),nl.
enable_analysis(OpName1,OpName2,FindInvViolations,UseEnableGraph,Timeout,Enable) :- %trace,
    b_top_level_operation(OpName1),
    b_top_level_operation(OpName2),
    %print(OpName1), print(' ---> '), print(OpName2), print('  :: '),
    (OpName1 == OpName2 -> 
        Enable=impossible % on top level an operation cannot enable itself
    ; otherwise ->
        b_get_read_write(OpName1,_GReads1,_AReads1,_Reads1,Writes1),
        b_get_read_write(OpName2,GReads2,_AReads2,_Reads2,_Writes2),
        get_negated_guard(OpName2,PosGuard2,NegGuard2),
        ( get_preference(use_cbc_analysis,false)
            -> (ord_intersect(Writes1,GReads2)
                -> (UseEnableGraph=true -> compute_weakest_precondition(OpName1,PosGuard2,FindInvViolations,Enable); Enable=possible_enable)
                ;  Enable=possible_keep
                )
        ; otherwise ->
            get_guard(OpName1,Guard1),
            get_conj_inv_predicate([Guard1,NegGuard2],FindInvViolations,StartPred),
            ((OpName1\=OpName2,ord_intersect(Writes1,GReads2),catch_and_ignore_well_definedness_error(
                                    sap:testcase_path_timeout(pred(StartPred),Timeout,[OpName1],PosGuard2,_,_,_,_,R),_)) ->
                (UseEnableGraph=true -> compute_weakest_precondition(OpName1,PosGuard2,FindInvViolations,Enable); (is_timeout(R)-> Enable=timeout; Enable=possible_enable))
            ; otherwise ->
                Enable=possible_keep_or_disable
            )
        )
		%format(user_output,"Read Write Vars: ~w w:~w  -->  ~w gr:~w ~n",[OpName1,Writes1,OpName2,GReads2])
    ).%,print(Enable),nl.

% conjoin a predicate with invariant and properties if specified
get_conj_inv_predicate(Preds,FindInvViolations,StartPred) :-
    (FindInvViolations==1 ->
        b_get_invariant_from_machine(Inv),
        b_get_properties_from_machine(Prop),
        conjunct_predicates([Prop,Inv|Preds],StartPred) % TO DO: project away
    ;   (FindInvViolations==0
          -> true
           ; add_internal_error('Illegal flag: ',get_conj_inv_predicate(Preds,FindInvViolations,StartPred))),
        conjunct_predicates(Preds,StartPred)
    ).

reads_writes_intersection(Writes, Reads, ComputedResult, PositiveResult, Res) :-
    (ord_intersect(Writes,Reads) ->
        Res=PositiveResult
    ; otherwise ->
        Res=ComputedResult
    ).

tcltk_enabling_analysis(list([list(['Origin'|Ops])|Result]),POR) :-
  get_preference(timeout_cbc_analysis,Timeout),
  compute_dependendency_relation_of_all_events_in_the_model(1,por(ample_sets,0,0,0),EnableGraph),
  suitable_for_por(EnableGraph,POR),
  findall(Op, b_top_level_operation(Op), Ops),
  findall(list([OpName1|EnableList]),
     (init_or_op(OpName1),
      findall(Enable,enable_analysis(OpName1,_OpName2,1,1,Timeout,Enable),EnableList)),
     Result)
  . %,print_enable_table([list(['Origin'|Ops])|Result]).

suitable_for_por(EnGraph,Result) :-
    (not_all_reachable(EnGraph)
        -> Result=true
        ;  Result=false
    ).

not_all_reachable(EnGraph) :-
    ample_sets: vertices(EnGraph,Vertices),
    member(V,Vertices),
    \+ (ample_sets: reachable(V,EnGraph,Reachable),Reachable=Vertices).

% --------------------------- Used for the ample sets (End) -----------------------------%

:- use_module(weakest_preconditions, [weakest_precondition/3]).
:- use_module(bsyntaxtree, [get_texpr_expr/2, conjunct_predicates/2]).
:- use_module(translate, [translate_bexpression/2, translate_substitution/2]).
:- use_module(library(timeout)).
:- use_module(debug).

:- use_module(library(wgraphs)).

:- use_module(probsrc(preferences),[get_preference/2]).
:- dynamic disable_graph/1.

compute_action_relations_graph(FindInvViolations,DisableGraph) :-
   (disable_graph(DisableGraph) -> true
   ; otherwise -> enter_new_error_scope(ScopeID),
                  call_cleanup(compute_disable_graph(FindInvViolations,DisableGraph),
                               my_exit_error_scope(ScopeID))
                      
   ).
my_exit_error_scope(ScopeID) :- 
  (specific_event_occurred_at_level(ScopeID,Event) ->
   format('### ERROR/ENUMERATION WARING OCCURED : ~w !~n',[Event])),fail.
my_exit_error_scope(ScopeID) :-  %error_manager:print_error_scopes,
  clear_all_errors_in_error_scope(ScopeID),
  exit_error_scope(ScopeID,_).

compute_disable_graph(FindInvViolations,DisableGraph) :-
        findall(Action, init_or_op(Action),AllActions),
	    print('********** DETERMINE ENABLE/DISABLE GRAPH *************'),nl,
        get_preference(timeout_cbc_analysis,Timeout),
   	    cputime(T1),
   	    determine_action_relations(AllActions,FindInvViolations,Timeout,DisableGraph,determine_neighbors),
   	    debug_println(9,disgraph(DisableGraph)),
   	    cputime(T2),
  	    print('ENABLE/DISABLE ANALYSIS Checking Time: '), D is T2-T1, print(D),print(' ms.'),nl,
  	    !,
   	    assert(disable_graph(DisableGraph)).


determine_action_relations(AllActions,FindInvViolations,Timeout,DisGraph,Predicate) :-
   vertices_edges_to_wgraph([],[],Graph),
   determine_action_relations_graph(AllActions,AllActions,FindInvViolations,Timeout,Graph,DisGraph,Predicate).

determine_action_relations_graph([],_AllActions,_FindInvViolations,_Timeout,Graph,Graph,_Predicate).
determine_action_relations_graph([Action|Actions],AllActions,FindInvViolations,Timeout,Graph,DisGraph,Predicate) :-
   call(Predicate,Action,AllActions,FindInvViolations,Timeout,Graph,ResGraph),
   determine_action_relations_graph(Actions,AllActions,FindInvViolations,Timeout,ResGraph,DisGraph,Predicate).

determine_neighbors(_Act,[],_FindInvViolations,_Timeout,Graph,ResGraph) :- !, ResGraph=Graph.
determine_neighbors(Act1,[Act2|Actions],FindInvViolations,Timeout,Graph,ResGraph) :-
   ((get_preference(use_pge,true),get_preference(use_pge_dis,true),possibly_disables(Act1,Act2,FindInvViolations,Timeout)) ->
   	      debug_println(9,possibly_disables(Act1, Act2)),
          add_vertices(Graph,[Act1,Act2],Graph1),
       	  add_edges(Graph1,[Act2-(Act1-3)],Graph2)
   ; (get_preference(por,sleep_sets),dependent_actions(Act1,Act2,Timeout,Result),(Result==independent;Result==syntactic_independent)) ->
          debug_println(9,independent_actions(Act1, Act2)),
          add_vertices(Graph,[Act1,Act2],Graph1),
       	  add_edges(Graph1,[Act2-(Act1-2)],Graph2)
   ; (get_preference(use_pge,true),get_preference(use_pge_keep,true),keeps_enabled_syntactic(Act1,Act2)) ->
	      debug_println(9,keeps_enabled_syntactic(Act1, Act2)),
          add_vertices(Graph,[Act1,Act2],Graph1),
          add_edges(Graph1,[Act2-(Act1-1)],Graph2)
   ; otherwise ->
       	  Graph2=Graph
   ),
   determine_neighbors(Act1,Actions,FindInvViolations,Timeout,Graph2,ResGraph).

possibly_disables('INITIALISATION', OpName2,_FindInvViolations,Timeout) :-
    b_top_level_operation(OpName2),
    get_guard(OpName2,PosGuard2),
    (/*catch_and_ignore_well_definedness_error(*/sap:testcase_path_timeout(init,Timeout,[],PosGuard2,_,_,_,_,_R)/*,_Result)*/ ->
	fail
    ; otherwise ->
        % OpName1 always disables OpName2
        debug_println(9,possibly_disables('INITIALISATION', OpName2))
    ).
possibly_disables(OpName1, OpName2, FindInvViolations, Timeout) :-
    b_top_level_operation(OpName1),
    b_top_level_operation(OpName2),
    get_guard(OpName1,PosGuard1),
    get_guard(OpName2,PosGuard2),
    get_conj_inv_predicate([PosGuard1],FindInvViolations,StartPred),
    (/*catch_and_ignore_well_definedness_error(*/sap:testcase_path_timeout(pred(StartPred),Timeout,[OpName1],PosGuard2,_,_,_,_,_R)/*,_Result)*/
        -> fail
    ; otherwise % OpName2 is impossible to be executed after OpName1
        -> debug_println(9,possibly_disables(OpName1, OpName2))
    ).

% be careful here: keeps_enabled_syntactic/2 is not the same as keeps_enabled/2!
% keeps_enabled_syntactic/2 is a legal keep relationship with regard to the current constellation
% of both operations in the state space.
% If s -OpName2-> s' & s -OpName1-> t & curState = t & wr(OpName1) /\ rd(OpName2) = {} then t -OpName2-> t'
keeps_enabled_syntactic(OpName1, OpName2) :-
   b_get_read_write(OpName1,_Reads1,Writes1),
   b_get_read_write(OpName2,GReads2,_AReads2,_Reads2,_Writes2),
   \+ord_intersect(Writes1,GReads2).

/*
check_predicate_timeout(Pred,Timeout,State,Result) :-
    time_out(solve_predicate(Pred,S,R),
             Timeout,
             T),
    ( T==success -> Result=R,State=S
    ; otherwise  -> Result=timeout,State=empty).    
*/

tcltk_disabling_analysis(list([list(['Origin'|Ops])|Result])) :-
  get_preference(timeout_cbc_analysis,Timeout),
  findall(Op, b_top_level_operation(Op), Ops),
  findall(list([OpName1|DisableList]),
     (init_or_op(OpName1),
      findall(Disables,(b_top_level_operation(OpName2),(possibly_disables(OpName1,OpName2,1,Timeout) -> Disables=yes; Disables=no)),DisableList)),
     Result),
  print_enable_table([list(['Origin'|Ops])|Result]).

:- use_module(probsrc(error_manager),[real_error_occurred/0, logged_error/4, reset_errors/0, add_warning/3]).

:- meta_predicate catch_and_ignore_well_definedness_error(0).
catch_and_ignore_well_definedness_error(Call,Result) :-
  catch_enumeration_warning(Call, true),
  (real_error_occurred ->
	  ((logged_error(Error,_S,_TContext,_Span),Error==well_definedness_error) ->
		  reset_errors,
		  print('Well definedness error occurred, but it will be ignored here.'),nl,
		  Result=error
    ; otherwise ->
		  print('No well definedness error occurred. Error will not be ignored here.'),nl,
		  Result=unexpected_error_occurred
	  )
  ; otherwise ->
	  Result=success
  ).


%:- meta_predicate catch_enumeration_warning(0).
catch_enumeration_warning(Call,Handler) :-
    % throw/1 predicate raises instantiation_error
    on_exception(enumeration_warning(enumerating(_),_Type,_,_,critical),Call,call(Handler)).

testcase_path_timeout_catch(Pred,TIMEOUT,Seq,P2,Csts,Ops,TestS,TI,Res) :-
  catch_enumeration_warning(sap:testcase_path_timeout(Pred,TIMEOUT,Seq,P2,Csts,Ops,TestS,TI,Res),
         Res=virtual_time_out).

%enumeration_warning(enumerating(n),'INTEGER',1:sup,1:3,critical)

compute_weakest_precondition(OpName,Pred,FindInvViolations,Res) :-
    b_get_machine_operation(OpName,_Res,Par,Body),
    (FindInvViolations==1 ->
      b_get_invariant_from_machine(Inv),
      conjunct_predicates([Inv,Pred],Pred1)
    ;otherwise ->
      Pred1 = Pred
    ),
    % TODO: add case for Event-B 
    % compute wp when the source operation is deterministic
    get_preference(enable_graph_depth,Depth),
    ((Par\=[]; Depth==0) ->
        Res = pred(b(truth,pred,[]))
        %add_warning(weakest_precondition,'Weakest precondition cannot be computed: ',(body(Body),pred(Pred))),
        %Res = invariant
    ; otherwise ->
        weakest_precondition(Body,Pred1,WP),
        print_weakest_precondition(Body,Pred,WP),
        remove_nodeid_infos(WP,WP1),
        Res=pred(WP1)
    ).

compute_wp(b(Body),Pred,WP) :-
    weakest_precondition(Body,Pred,WP).

/* For debugging only */
print_weakest_precondition(Body,Guard,WP) :- 
   translate_bexpression(Guard,Guard_PP),
   translate_bexpression(WP,WP_PP),
   translate_substitution(Body,RealBody),
   debug_print(9,'['),debug_print(9,RealBody),debug_print(9,']'),
   debug_print(9,'('),debug_print(9,Guard_PP),debug_print(9,')'),
   debug_print(9,' = '),debug_print(9,WP_PP),debug_println(9,'').

init_or_op('INITIALISATION').
init_or_op(OpName1) :- b_top_level_operation(OpName1). %b_get_machine_operation(OpName1,_,_,_).

b_get_read_write(O,R,W) :- b_get_operation_normalized_read_write_info(O,R,W),!.
b_get_read_write(_O,[all],[all]). % in Event-B no read information stored at the moment; all calls to b_get_operation_normalized_read_write_info fail

/* We differ between read variables in the guard and read variables in the event statement */
b_get_read_write(Operation,ReadGuard,ReadAction,Read,Write) :-
	b_get_read_write(Operation,Read,Write),
    b_get_machine_operation(Operation,_RR,_P,Body),
    bsyntaxtree: get_texpr_expr(Body,RLEVENT),
    (RLEVENT = rlevent(_Name,_Section,_Stat,_Parameters,Grd,_Thm,Act,_VWit,_PWit,_Unmod,_AbsEvents) ->
		G=Grd,A=Act
	; otherwise ->
	    get_guard(Operation,G),
        get_action(Operation,A)
	),
	% RGuardVars can comprise also local variables by operations with parameters
	find_identifier_uses(G,[],RGuardVars),
	list_to_ord_set(RGuardVars,ReadGuard),
    (is_list(A) % in case of Event-B events the action block is a list of assignments
        -> find_identifier_uses_l(A,[],RActionVars)
        ;  find_identifier_uses(A,[],RActionVars)
    ),
	list_to_ord_set(RActionVars,RActionVars1),
    ord_intersection(RActionVars1,Read,ReadAction).
	%,debug_println(9,b_get_read_write(O,RGuard,R,W)).

get_action(OpName,Action) :-
    b_get_machine_operation_for_animation(OpName,_Results,_Parameters,TBody,_OType,true),
    (get_texpr_expr(TBody,precondition(_Precondition,Action)) -> true; Action=TBody).

% print table in CSV format
print_enable_table(R) :- print_enable_table_list(R).
print_enable_table_list([]) :- nl,nl.
print_enable_table_list([list(L)|T]) :-
   print_csv_list(L), print_enable_table_list(T).
print_csv_list([]) :- !,nl.
print_csv_list([H]) :- !,print(H),nl.
print_csv_list([H|T]) :- !,print(H), print(','), print_csv_list(T).
print_csv_list(X) :- print(illegal_list(X)),nl, add_internal_error('Illegal list: ', print_csv_list(X)).

% ---------- create enable graph ------------------

tcltk_create_enable_graph(File) :-
       gen_dot_graph(File,enabling_analysis,eop_node_predicate,cbc_enable_trans_predicate,none,none).

tcltk_create_enable_graph_por(File,Inv,UseEnableGraph) :-
    compute_dependendency_relation_of_all_events_in_the_model(Inv,por(ample_sets,UseEnableGraph,1,0),_EnableGraph),
    EnablePred =.. [cbc_enable_trans_predicate_por,Inv,UseEnableGraph],
    gen_dot_graph(File,enabling_analysis,eop_node_predicate,EnablePred,none,none).

eop_node_predicate(NodeID,SubGraph,NodeDesc,Shape,Style,black) :-
     op_node_predicate(NodeID,SubGraph,NodeDesc,Shape,Style,_Color).
eop_node_predicate('INITIALISATION',SubGraph,'INITIALISATION',Shape,Style,Color) :- SubGraph=none,
     Shape=hexagon, Style=none, Color=olivedrab2.

cbc_enable_trans_predicate_por(Inv,UseEnableGraph,NodeID,Label,SuccID,Color,Style) :-
    get_preference(timeout_cbc_analysis,Timeout),
    enable_analysis(NodeID,SuccID,Inv,UseEnableGraph,Timeout,Res),
    (Res=possible_enable; Res=possible; Res=guaranteed; Res=timeout; Res=pred(_Expr)),
    translate_enable_res(Res,NodeID,SuccID,Label,Color,Style).

cbc_enable_trans_predicate(NodeID,Label,SuccID,Color,Style) :- 
    cbc_enable_analysis(NodeID,SuccID,Res,500),
    translate_enable_res(Res,NodeID,SuccID,Label,Color,Style).

translate_enable_res(impossible,_,_,'impossible',red,_) :- !, fail.
translate_enable_res(guaranteed,_,_,'guaranteed',olivedrab2,solid) :- !.
translate_enable_res(independent,X,Y,'independent',yellow,bold) :- !,X=Y. % only show if source & dest identical
translate_enable_res(activation_indepdendent,X,Y,'activation_indepdendent',lightgray,bold) :- !,X=Y. % only show if source & dest identical
translate_enable_res(possible,_,_,'possible',black,solid) :- !.
translate_enable_res(keep,_,_,'keep',lightgray,solid) :- !.
translate_enable_res(possible_enable,_,_,'(enable)',black,solid) :- !.
translate_enable_res(possible_disable,_,_,'(disable)',lightblue,solid) :- !.
translate_enable_res(timeout_keep,_,_,'keep?',tomato,dashed) :- !.
translate_enable_res(timeout_possible,_,_,'possible?',tomato,dashed) :- !.
translate_enable_res(timeout_possible_disable,_,_,'(disable?)',tomato,dashed) :- !.
translate_enable_res(timeout_possible_enable,_,_,'(enable?)',tomato,dashed) :- !.
translate_enable_res(pred(Expr),_,_,String,black,dashed) :- !, translate_bexpression(Expr,String),debug_println(9,translate_enable_res(Expr,String)).
translate_enable_res(X,_,_,X,red,solid).

remove_nodeid_infos(Expr,Res) :-
  var(Expr),!,
  Res=Expr.
remove_nodeid_infos(Expr,Res) :-
  is_list(Expr),!,
  maplist(remove_nodeid_infos,Expr,Res).
remove_nodeid_infos(Expr,Res) :-
  Expr =.. [F|Args],
  remove_nodeid_infos1(F,Args,Res).

remove_nodeid_infos1(nodeid,_Args,Res) :-
  % ignore additional node infos
  !, Res =.. [nodeid,none].
remove_nodeid_infos1(F,Args,Res) :-
  maplist(remove_nodeid_infos,Args,ArgsRes),
  Res =.. [F|ArgsRes].

% -------------

:- use_module(flow).

dot_edge([]).
dot_edge([(A1,A2,Enable)|T]) :- 
	flow:eventnumber(A1,A1N), 
	flow:eventnumber(A2,A2N),  
	format_edge(A1N,A2N,Enable), 
	dot_edge(T).

format_edge(FromNumber,ToNumber,Enable) :-
   print(FromNumber), print('->'), print(ToNumber), print(' '),
   (  Enable==dependent         -> print('[color="forestgreen" label=dependent]')
	; Enable==race_dependent    -> print('[color="seablue" label=race_dependent]')
	; Enable==guaranteed        -> print('[color="forestgreen" label=guaranteed]') 
	; Enable==possible          -> print('[color="blue" label=possible]')
	; Enable==possible_enable   -> print('[color="blue" label=possible_enable]')
	; Enable==timeout           -> print('[color="brown" label=timeout]')
	; Enable==clpfd_overflow    -> print('[color="red" label=clpfd_overflow]')
	; otherwise                 -> print('[color="red" label=error]')
	),nl.


tcltk_create_dependence_graph(File) :- 
	findall(Action, (b_get_machine_operation(Action,_,_,_)),AllActions),
	ample_sets: determine_dependent_enable_action_relations(AllActions,300,0,_EnGraph),
	flow: reset_count,
	flow: enumerate_events(AllActions),
	get_dependency_edges(AllActions,Edges),
	to_dot_edges(Edges,DotEdges),
	print(writing_to_dot_file(File)),nl,
	tell(File),
	flow: dot_header('DepGraph'),
	flow: dot_node(AllActions),
	dot_edge(DotEdges),
	flow: dot_footer,
	told,
	clear_dynamic_predicates_for_POR.

to_dot_edges([],[]).
to_dot_edges([X-Y|T],[(X,Y,dependent)|TRes]) :- 
	delete(T,Y-X,TR),
	to_dot_edges(TR,TRes).

get_dependency_edges(AllActions,Edges) :- 
	get_dependency_edges(AllActions,AllActions,EdgeTuples),
	append(EdgeTuples,Edges).

get_dependency_edges([],_AllActions,[]).
get_dependency_edges([Act|Actions],AllActions,[Edges|EdgesRest]) :- 
	get_action_edges(Act,AllActions,Edges),
	get_dependency_edges(Actions,AllActions,EdgesRest).

get_action_edges(_Act,[],[]).
get_action_edges(Act,[Act1|Actions],[Act-Act1|Rest]) :- 
	dependent_act(Act,Act1),!,
	get_action_edges(Act,Actions,Rest).
get_action_edges(Act,[_Act1|Actions],Rest) :- 
	get_action_edges(Act,Actions,Rest).

% ---------------------------------------------------

% Simultaneous Enabling Analysis
% Find events which cannot be enabled simultaneously
% Usage: 
%  a) if one event is known to be enabled we do not need to check the impossible ones
%  b) if one event is known to be disabled we do not need check events which guarantee it

% use_module(enabling_analysis), tcltk_cbc_simultaneous_enabling_analysis(R).

tcltk_cbc_simultaneous_enabling_analysis(list([list(['Origin'|SOps])|Result])) :- 
  findall(Op, b_top_level_operation(Op), Ops),
  sort(Ops,SOps), % sort operations to ensure nice triangular display
  statistics(walltime,[WT1,_]),
  findall(list([OpName1|EnableList]),
     (member(OpName1,SOps),
      findall(SimulRes,tcltk_simult(OpName1,SOps,SimulRes),EnableList)),
     Result),
  statistics(walltime,[WT2,_]), Time is WT2-WT1,
  format('Runtime for simultaneous enabling analysis: ~w ms~n',[Time])
  . %,print_enable_table([list(['Origin'|Ops])|Result]).

tcltk_simult(Op1,SOps,Res) :- member(Op2,SOps),tcltk_simult2(Op1,Op2,Res).
tcltk_simult2(Op1,Op2,Res) :-
  Op1 = Op2, % Op2 @=< Op1, % the simultaneous impossible analysis is symmetric:  TO DO: do not compute/display info twice
  !,
   Res = '-'. 
tcltk_simult2(Op1,Op2,TRes) :-
   simult_enable_analysis(Op1,Op2,1,fixed_time_out(150),TRes).

simult_enable_analysis(OpName1,OpName2,UseInvariant,TimeoutFactor,Res) :-
    b_top_level_operation(OpName1),
    b_top_level_operation(OpName2),
    %OpName2 @>OpName1,
	get_negated_guard(OpName1,PosGuard1,_NegGuard1),
    get_negated_guard(OpName2,PosGuard2,NegGuard2),
	get_conj_inv_predicate([PosGuard1,PosGuard2],UseInvariant,Pred),
	%print(checking_simultaneous(OpName1,OpName2)),nl,
	(solve_predicate_with_chr(Pred,_State,TimeoutFactor,RR)
	  -> %print(state(_State)),nl, print(result(RR)),nl,
	     (RR=no
	      -> Res = impossible % we do not compute PosGuard1,NegGuard2 in this case: otherwise this would mean that Operation 1 can never fire !
	      ;  get_conj_inv_predicate([PosGuard1,NegGuard2],UseInvariant,Pred2),
	         solve_predicate_with_chr(Pred2,_State2,TimeoutFactor,RR2),
	         combine_simult_result(RR,RR2,Res))
	   ; print('********* ERROR ***********'),nl,fail).

combine_simult_result(yes,yes,R) :- !,R=possible.
combine_simult_result(yes,no,R) :- !, R=guaranteed.
combine_simult_result(A,B,Res) :- functor(A,FA,_), functor(B,FB,_),
   ajoin([FA,'_',FB],Res).

% ---------------------------------------------------

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(specification_initialised,reset_enabling_analysis,
                           'Initialise module enabling analysis.').

:- dynamic feasible_operation_cache/2. 
reset_enabling_analysis :-
   retractall(disable_graph(_)),
   retractall(feasible_operation_cache(_,_)),
   retractall(cbc_enable_analysis_cache(_,_,_,_)).

% in principle you should only call this for Op with operation_name_not_yet_covered(Op)
feasible_operation(Op,Res) :- feasible_operation_cache(Op,CR),!,Res=CR.
feasible_operation(Op,Res) :- feasible_operation(Op,1,fixed_time_out(500),CR),
   assert(feasible_operation_cache(Op,CR)),
   Res=CR.

:- use_module(state_space,[operation_not_yet_covered/1]).
infeasible_operation(Op) :- feasible_operation_cache(Op,CR),!, CR=impossible.
infeasible_operation(Op) :- operation_not_yet_covered(Op),
   feasible_operation(Op,impossible).

infeasible_operation_cache(Op) :- feasible_operation_cache(Op,impossible).

:- use_module(solver_interface, [solve_predicate/5]).
solve_predicate_with_chr(Pred,State,TimeoutFactor,RR) :-
    solve_predicate(Pred,State,TimeoutFactor,['CHR','CLPFD','SMT'],RR).

:- use_module(debug, [debug_mode/1]).
% check if an operation is feasible
% UseInvariant = 1 : assume invariant
feasible_operation(OpName1,UseInvariant,TimeoutFactor,Res) :-
    b_top_level_operation(OpName1),
	get_negated_guard(OpName1,PosGuard1,_NegGuard1),
	get_conj_inv_predicate([PosGuard1],UseInvariant,Pred),
	(debug_mode(on) -> print(checking_feasibility(OpName1)),nl, translate:print_bexpr(Pred),nl ; true),
	% we could also call: sap:testcase_path_timeout(invariant,250,[OpName1],b(truth,pred,[]),_Constants,_Ops,_TestS,_TI,R1)
	(solve_predicate_with_chr(Pred,State,TimeoutFactor,RR)
	  -> %print(state(_State)),nl, print(result(RR)),nl,
	     (RR=no
	      -> Res = impossible % we do not compute PosGuard1,NegGuard2 in this case: otherwise this would mean that Operation 1 can never fire !
	      ;  RR=yes -> Res = possible,
	         (debug_mode(on) -> print('SOLUTION: '),translate:print_bstate(State),nl ; true)
	      ;  functor(RR,Res,_) )
	   ; print('********* ERROR ***********'),nl,fail).


