% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(error_manager,[add_warning/2, add_warning/3, add_warning/4,
           reset_errors/0, add_error/2, add_error/3, add_error/4,
           add_error_and_fail/3, add_error_and_fail/2,
           add_error_fail/3,
           add_failed_call_error/1, add_internal_error/2, add_internal_error/4,
           add_message/2, add_message/3,
           add_all_perrors/1, add_all_perrors/2, add_all_perrors/3,
           get_all_errors/1, get_all_errors_and_reset/1,
           get_error/2,
           get_error_span_linecol/4,
           position_is_not_in_main_file/1,
           real_error_occurred/0, reset_real_error_occurred/0, assert_real_error_occurred/0,
           test_error_occurred/1, count_errors_occurred/2,

           set_error_context/1, clear_error_context/0, get_error_context/1,

           extract_line_col/5, extract_all_line_col/5,
           extract_line_col_for_main_file/5,
           extract_subsidiary_tail_file_name/2,
           extract_file_number_and_name/3,
           print_error_span/1,

           trace_upon_error/0, turn_trace_upon_error_on/0,turn_trace_upon_error_off/0,
           turn_show_error_source_on/0, turn_show_error_source_off/0,

            enter_new_error_scope/1,  enter_new_clean_error_scope/1,
            exit_error_scope/2, error_occurred_at_level/1, specific_event_occurred_at_level/2,
            throw_errors_in_scope/2,
            add_new_event_in_error_scope/1, add_new_event_in_error_scope/2,
            event_occurred_in_error_scope/1,
            clear_events_in_error_scope/2, clear_events_in_current_error_scope/1,
            clear_all_errors_in_error_scope/1,
            call_in_fresh_error_scope_for_one_solution/1,
            enumeration_warning_occured_in_error_scope/0,
            on_enumeration_warning/2, on_enumeration_warning_with_continue/3,
            throw_enumeration_warnings_in_scope/2,
            catch_enumeration_warning_exceptions/2, catch_enumeration_warning_exceptions/3,
            call_with_enumeration_warning/1,
            critical_enumeration_warning_occured_in_error_scope/4, critical_enumeration_warning_occured_in_error_scope/0,
            clear_enumeration_warnings/0,
            time_out_with_enum_warning_one_solution/3, % use this for finding one solution
            time_out_with_enum_warning_one_solution_no_new_error_scope/3,
            is_time_out_result/1,
            time_out_with_enum_warning_for_findall/3, % use this if you cycle through all solutions
            time_out_with_enum_warning_for_findall_in_current_error_scope/4,
            logged_error/4
            ]).

:- use_module(library(lists)).

:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'The error_manager controls error messages. Error messages can be added, retrieved, counted,etc. The module also takes care of extracting source spans from error messages.').

%:- use_module(translate).
:- use_module(debug, [debug_mode/1]).

% INTERNAL ERRORS

:- meta_predicate add_failed_call_error(0).
:- meta_predicate add_internal_error(-,0).

add_failed_call_error(Call) :- !,
   add_internal_error('Internal Error; call failed: ',Call).

add_internal_error(Module,Msg,Term,Span) :-
   add_error(internal_error(Module),Msg,Term,Span).
add_internal_error(Msg,Module:Call) :- !,
   add_error(internal_error(Module),Msg,Call).
add_internal_error(Msg,Call) :-
   add_internal_error(Msg,unknown_module:Call).


% ERROR SCOPING

:- dynamic cur_block_level/1.
cur_block_level(0).
:- volatile error_occurred_at_level/1.
:- dynamic error_occurred_at_level/1.
:- volatile throw_error/2.
:- dynamic throw_error/2.

enter_new_error_scope(Level) :-
   enter_new_clean_error_scope(Level),
   L1 is Level - 1,
   copy_throw_errors(L1,Level).
copy_throw_errors(L1,L) :- throw_error(L1,Error), %print(copy(throw(L1,L,Error))),nl,
   assert(throw_error(L,Error)),fail.
copy_throw_errors(_,_).
enter_new_clean_error_scope(Level) :-
  retract(cur_block_level(L)), L1 is L+1, assert(cur_block_level(L1)),
  retractall(throw_error(L1,_)),
  % print(enter(L1)),nl,trace,
  Level=L1.
% mark that everything unifying with Error should be thrown rather than collected
throw_errors_in_scope(Level,Error) :-
   (\+ throw_error(Level,Error)  % TO DO: check that we are an instance of existing error !
    -> assert(throw_error(Level,Error))
     ; true).

clear_all_errors_in_error_scope(Level) :- %print(clearing(Level)),nl,
   clear_events_in_error_scope(Level,_),
   error_occurred_at_level(L1), L1>Level, %print(clearing(L1)),nl,
   clear_events_in_error_scope(Level,_),fail.
clear_all_errors_in_error_scope(_).

exit_error_scope(Level,ErrorOcc) :- retract(cur_block_level(CurL)),
   % print(exit(Level,CurL)),nl,
   (CurL = Level -> NewL is CurL-1
    ; Level < CurL -> NewL is Level-1,  format('*** exiting multiple error blocks (~w -> ~w)***~n',[CurL,NewL])
    ; NewL is CurL -> print('*** block already exited ***'),nl
   ), assert(cur_block_level(NewL)),
   copy_error_scope_events(CurL,NewL),
   retractall(throw_error(CurL,_)),
   (retract(error_occurred_at_level(CurL))
    -> ErrorOcc=true,
       (error_occurred_at_level(NewL) -> true ; assert(error_occurred_at_level(NewL)))
    ; ErrorOcc=false).

assert_real_error_in_current_block :- cur_block_level(L),L>0,
  \+ error_occurred_at_level(L), assert(error_occurred_at_level(L)),!.
assert_real_error_in_current_block.


reset_error_scopes :- retractall(cur_block_level(_)), assert(cur_block_level(0)),
   retractall(error_occurred_at_level(_)),
   retractall(specific_event_occurred_at_level(_,_)).

% for debugging:
:- public print_error_scopes/0.
print_error_scopes :- cur_block_level(L), 
   format(' Cur Block Level = ~w~n',[L]),
   fail.
print_error_scopes :-
   findall(L,error_occurred_at_level(L),LL),
   format(' Errors occured at levels = ~w~n',[LL]),
   specific_event_occurred_at_level(Level,Err),
   format(' ~w  -->  ~w~n',[Level,Err]),
   fail.
print_error_scopes.

:- meta_predicate call_in_fresh_error_scope_for_one_solution(0).
% call a predicate in fresh error scope (assuming one solution is searched for only)
% and then discard all errors
call_in_fresh_error_scope_for_one_solution(Call) :-
     enter_new_error_scope(Level),
     call_cleanup(Call,
                   (clear_all_errors_in_error_scope(_),
                    exit_error_scope(Level,_))).

:- volatile specific_event_occurred_at_level/2.
:- dynamic specific_event_occurred_at_level/2.

% can be used to keep track of whether specific events occurred within an error_scope
:- meta_predicate add_new_event_in_error_scope(-,1).
add_new_event_in_error_scope(Event) :-
    add_new_event_in_error_scope(Event,true).
add_new_event_in_error_scope(Event,PrintMessageCall) :- cur_block_level(L),
   (specific_event_occurred_at_level(L,Event) -> fail
     ; throw_error(L,Event) -> (debug:debug_mode(off) -> true ; print_message(throwing(Event))), % trace, % <---
                               call(PrintMessageCall,throwing),
                               throw(Event)
     ; assert(specific_event_occurred_at_level(L,Event)),
       call(PrintMessageCall,not_throwing)
   ).

% call if you just want to add the error information; without throwing
assert_event_in_error_scope_without_throw(Event) :- cur_block_level(L),
   (specific_event_occurred_at_level(L,Event) -> true
    ; assert(specific_event_occurred_at_level(L,Event))).

event_occurred_in_error_scope(Event) :- cur_block_level(L),specific_event_occurred_at_level(L,Event).

copy_error_scope_events(FromLevel,ToLevel) :- retract(specific_event_occurred_at_level(FromLevel,Event)),
   \+ specific_event_occurred_at_level(ToLevel,Event),
   % print(copying_event(FromLevel,ToLevel,Event)),nl,
   assert(specific_event_occurred_at_level(ToLevel,Event)),fail.
copy_error_scope_events(_,_).

clear_events_in_current_error_scope(Event) :- cur_block_level(L),clear_events_in_error_scope(L,Event).

%clear_events_in_error_scope(L,Event) :-retract(specific_event_occurred_at_level(L,Event)), print(retract(L,Event)),nl,fail.
clear_events_in_error_scope(L,Event) :-
   retractall(specific_event_occurred_at_level(L,Event)),
   (specific_event_occurred_at_level(L,_) -> true
    ; retractall(error_occurred_at_level(L))).


throw_enumeration_warnings_in_scope(Level,Critical) :-
   throw_errors_in_scope(Level,enumeration_warning(_,_,_,_,Critical)).

:- meta_predicate on_enumeration_warning(0,0).
:- meta_predicate on_enumeration_warning_with_continue(0,0,0).
on_enumeration_warning(Call, BackupCall) :-
        on_enumeration_warning_with_continue(Call, BackupCall, true).
% sets up a new error scope and forces enumeration_warnings to be thrown
% if an enumeration_warning happens the BackupCall is executed
on_enumeration_warning_with_continue(Call, BackupCall, TrueContinuation) :-
    enter_new_error_scope(Level),
	throw_enumeration_warnings_in_scope(Level,_),
	on_exception(enumeration_warning(_,_,_,_,_),
		    if(Call,
		         (exit_error_scope(Level,_),call(TrueContinuation)),
		         (exit_error_scope(Level,_),fail)
		       ),
		    (exit_error_scope(Level,_),BackupCall)).
:- meta_predicate catch_enumeration_warning_exceptions(0,0).
:- meta_predicate catch_enumeration_warning_exceptions(0,0,-).
catch_enumeration_warning_exceptions(Call, BackupCall) :-
    catch_enumeration_warning_exceptions(Call,BackupCall,false).

% AddAsEventError = true means that caught exceptions are added as event_errors in scope
% (this can be important if we later want to decide whether an enumeration warning occured)
catch_enumeration_warning_exceptions(Call, BackupCall,AddAsEventError) :-
    enter_new_error_scope(Level),
	% difference: we do not force to throw_errors_in_scope
	on_exception(enumeration_warning(A,B,C,D,E),
		         call_cleanup(Call, exit_error_scope(Level,_)),
		         ((AddAsEventError=false -> true
		            ; assert_event_in_error_scope_without_throw(enumeration_warning(A,B,C,D,E))),
		          BackupCall)).

:- meta_predicate call_with_enumeration_warning(0).
% sets up a new error scope and forces enumeration_warnings to be thrown
% the exceptions are not caught !
call_with_enumeration_warning(Call) :-
    enter_new_error_scope(Level),
	throw_enumeration_warnings_in_scope(Level,_),
	on_exception(E,
		    if(Call,
		         exit_error_scope(Level,_),
		         (exit_error_scope(Level,_),fail)
		       ),
		    (exit_error_scope(Level,_),throw(E))).

:- use_module(tools,[safe_time_out/3]).
:- meta_predicate time_out_with_enum_warning_one_solution(0,-,-).
% a version of time_out which treats enumeration_warnings like time-outs
% and also checks enumeration_warnings in case of failure
% assumes we only look for one solution (otherwise error scope cleanup has to be changed !! )
time_out_with_enum_warning_one_solution(Call,TO,TimeOutRes) :-
    enter_new_error_scope(Level),
    %% print(time_out_with_enum_warning_one_solution(TO)),nl, %%
    call_cleanup(
        on_exception(enumeration_warning(A,B,C,D,E),
                 time_out_check_enum_warning_for_failure(Call,TO,TimeOutRes,no_clear), % no need to clear: we leave error_scope below
                 %% (print(enum_warning),nl,TimeOutRes=time_out)). %%
                 TimeOutRes=virtual_time_out(thrown_enumeration_warning(A,B,C,D,E))),
        exit_error_scope(Level,_)).

enumeration_warning_occured_in_error_scope :-
   event_occurred_in_error_scope(enumeration_warning(_,_,_,_,_)).

% just like above; but does not set up new error_scope (for performance) and clears enumeration warnings in case of failure and success
:- meta_predicate time_out_with_enum_warning_one_solution_no_new_error_scope(0,-,-).
time_out_with_enum_warning_one_solution_no_new_error_scope(Call,TO,TimeOutRes) :-
    (enumeration_warning_occured_in_error_scope
     -> trace,add_internal_error('Error block already has enumeration warnings',time_out_with_enum_warning_one_solution_no_new_error_scope(Call,TO,TimeOutRes))
      ; true),
    on_exception(enumeration_warning(A,B,C,D,E),
                 time_out_check_enum_warning_for_failure(Call,TO,TimeOutRes,clear_errors), % clear errors so we can continue using same error scope
                 %% (print(enum_warning),nl,TimeOutRes=time_out)). %%
                 TimeOutRes=virtual_time_out(thrown_enumeration_warning(A,B,C,D,E))).


% check if a result is a time_out:
is_time_out_result(time_out).
is_time_out_result(virtual_time_out(_)).

% a version which is to be used not for a single call but for backtracking
% generate virtual time_out at the end if enumeration warning occured
:- meta_predicate time_out_with_enum_warning_for_findall(0,-,-).
:- meta_predicate time_out_with_enum_warning_for_findall2(-,0,-,-).

time_out_with_enum_warning_for_findall(Call,TO,TimeOutRes) :-
    enter_new_error_scope(Level),
    time_out_with_enum_warning_for_findall2(Level,Call,TO,TimeOutRes).

time_out_with_enum_warning_for_findall2(_Level,Call,TO,TimeOutRes) :- %print(enter(_Level)),nl,
    on_exception(enumeration_warning(_,_,_,_,_),
                 time_out_check_enum_warning_for_failure(Call,TO,TimeOutRes,no_clear),
                 TimeOutRes=virtual_time_out(exception)). % virtual time_out
time_out_with_enum_warning_for_findall2(Level,_Call,_TO,_TimeOutRes) :- % print(exit(Level)),nl,
    clear_events_in_error_scope(Level,enumeration_warning(_,_,_,_,_)),
    exit_error_scope(Level,_),fail.

% a version of time_out_with_enum_warning_for_findall where you manually have to enter_new_error_scope before
% and exit_error_scope afterwards; useful if you repeatedly call this predicate and want to set up the block only once
:- meta_predicate time_out_with_enum_warning_for_findall_in_current_error_scope(-,0,-,-).
time_out_with_enum_warning_for_findall_in_current_error_scope(_Level,Call,TO,TimeOutRes) :- %print(enter(_Level)),nl,
    on_exception(enumeration_warning(_,_,_,_,_),
                 time_out_check_enum_warning_for_failure(Call,TO,TimeOutRes,no_clear),
                 TimeOutRes=virtual_time_out(exception)). % virtual time_out
time_out_with_enum_warning_for_findall_in_current_error_scope(Level,_Call,_TO,_TimeOutRes) :-
    % does not exit level
    clear_events_in_error_scope(Level,enumeration_warning(_,_,_,_,_)),fail.

% a version of time_out which upon failure checks if enumeration warnings were generated
time_out_check_enum_warning_for_failure(Call,TO,TimeOutRes,ClearErrors) :-
    safe_time_out(Call,TO,TimeOutRes),  % a safe version which checks that time-out value TO is not too large
    (TimeOutRes = time_out -> !  %, print(real_timeout),nl
       ; (ClearErrors=clear_errors -> clear_enumeration_warnings ; true)
         %,print(sol),nl
    ).
time_out_check_enum_warning_for_failure(_Call,_TO,TimeOutRes,ClearErrors) :- % print(failure),nl,
    event_occurred_in_error_scope(enumeration_warning(A,B,C,D,critical)), % print(virtual),nl,
    TimeOutRes = virtual_time_out(failure_enumeration_warning(A,B,C,D,critical)), % we have not investigated all possibilities; generate virtual time_out
    (ClearErrors=clear_errors -> clear_enumeration_warnings ; true).

critical_enumeration_warning_occured_in_error_scope :-
   critical_enumeration_warning_occured_in_error_scope(_,_,_,_).
critical_enumeration_warning_occured_in_error_scope(A,B,C,D) :-
    event_occurred_in_error_scope(enumeration_warning(A,B,C,D,critical)).

clear_enumeration_warnings :-
    clear_events_in_current_error_scope(enumeration_warning(_,_,_,_,_)).

% -----------------------

:- volatile logged_error/4, real_error_occurred/0, logged_error_spans/2.
:- dynamic logged_error/4.
:- dynamic real_error_occurred/0.
:- dynamic logged_error_spans/2.

reset_real_error_occurred :- retractall(real_error_occurred).
reset_errors :- retractall(real_error_occurred),
                retractall(logged_error(_,_,_,_)),retractall(logged_error_spans(_,_)),
                reset_error_scopes,
                clear_error_context.

reset_error_span_if_required :-
	(logged_error(_,_,_,_) -> true
		; retractall(logged_error_spans(_,_))).



add_error(Source,ErrMsg,ErrTerm,Span) :-
   add_error2(Source,ErrMsg,ErrTerm,Span),
   assert(logged_error_spans(Source,Span)),
   print_error_span(Span).

print_error_span(unknown) :- !.
print_error_span(Span) :-
   extract_line_col(Span,StartLine,StartCol,_EndLine,_EndCol), !,
   (extract_file_name(Span,Filename) ->
      ajoin(['Line: ',StartLine,' Column: ',StartCol, ' in file ',Filename],ErrMsg)
      ;  ajoin(['Line: ',StartLine,' Column: ',StartCol],ErrMsg)
   ),print_error(ErrMsg).
print_error_span(Span) :- print_error(Span). % span not recognized !

add_error_fail(Source,ErrMsg, ErrTerm) :- add_error_and_fail(Source,ErrMsg, ErrTerm).
add_error_and_fail(Source,ErrMsg, ErrTerm) :- add_error(Source,ErrMsg, ErrTerm),fail.
add_error_and_fail(Source,ErrMsg) :- add_error(Source,ErrMsg),fail.

add_error(Source,ErrMsg, ErrTerm) :-
    add_error2(Source,ErrMsg, ErrTerm,unknown).

%:- use_module(translate,[translate_error_term/2]).

trace_upon_error :- preferences:get_preference(trace_upon_error,true).
turn_trace_upon_error_on :- preferences:set_preference(trace_upon_error,true).
turn_trace_upon_error_off :-  preferences:set_preference(trace_upon_error,false).
:- dynamic show_error_source/0.
show_error_source. % True if error source and span to be printed
turn_show_error_source_on :- (show_error_source -> true ; assert(show_error_source)).
turn_show_error_source_off :-  retractall(show_error_source).

trace_if_user_wants_it :- (preferences:get_preference(trace_upon_error,true) -> trace ; true).


%:- use_module(translate, [translate_error_term/2, translate_context/2]).
add_error2(Source,ErrMsg,ErrTerm,Span) :-
    (show_error_source -> print_error('An error occurred !') ; true),
    assert_real_error_occurred,
    (show_error_source -> print_error(source(Source)) ; true),
    print_error(ErrMsg),
    safe_translate_error_term(ErrTerm,RString),
    (RString = '' -> true ; print_error(RString)),
    string_concatenate(ErrMsg,' ',ST),
    string_concatenate(ST,RString,S),
    safe_translate_context(Context),
    ((Context = '' ; \+show_error_source) -> true ; print_error(Context)),
    reset_error_span_if_required,
    assertz_logged_error(Source,S,Context,Span).


safe_translate_error_term(ErrTerm,RString) :-
    preferences:temporary_set_preference(expand_avl_upto,5,CHNG),
    translate:translate_error_term(ErrTerm,RString),
    preferences:reset_temporary_preference(expand_avl_upto,CHNG).

safe_translate_context(Result) :-
% avoid infinite loop when error occurs inside translate_context
    get_error_context(Context),
    set_error_context(translate_context),
    preferences:temporary_set_preference(expand_avl_upto,5,CHNG),
    (translate:translate_context(Context,TContext)
      -> true ; TContext=Context),
    preferences:reset_temporary_preference(expand_avl_upto,CHNG),
    set_error_context(Context),
    Result=TContext.


add_error(Source,ErrMsg) :- print_error('An error occurred !'),
    assert_real_error_occurred,
    add_error_or_warning(Source,ErrMsg,unknown).

add_error_or_warning(Source,ErrMsg,Span) :-
    (preferences:get_preference(strict_raise_warnings,true)
    -> assert_real_error_occurred ; true),
    (show_error_source -> print_error(source(Source)) ; true),
    print_error(ErrMsg),
    safe_translate_context(Context),
    ((Context = '' ; \+show_error_source) -> true ; print_error(Context)),
    reset_error_span_if_required,
    assertz_logged_error(Source,ErrMsg,Context,Span),
    print_error_span(Span).

assert_real_error_occurred :- (real_error_occurred -> true ; assert(real_error_occurred)),
    assert_real_error_in_current_block.

assertz_logged_error(Source,S,TContext,Span) :-
    assertz(logged_error(Source,S,TContext,Span)),trace_if_user_wants_it.

add_warning(Source,ErrMsg, ErrTerm) :-
   add_warning(Source,ErrMsg, ErrTerm,unknown).
add_warning(Source,ErrMsg, ErrTerm,Span) :-
    safe_translate_error_term(ErrTerm,RString),
    string_concatenate(ErrMsg,RString,S),
    add_error_or_warning(warning(Source),S,Span).
add_warning(Source,Message) :- add_error_or_warning(warning(Source),Message,unknown).

add_message(Source,ErrMsg, ErrTerm) :- /* same as add_error but does not add err msg */
    safe_translate_error_term(ErrTerm,RString),
    string_concatenate(ErrMsg,RString,S),
    add_message(Source,S).
add_message(Source,ErrMsg) :-
   (show_error_source -> print_error(source(Source)) ; true),
   print_error(ErrMsg).
   %assertz(logged_error(Source,ErrMsg,_,_)).

% adds a list of errors with position information
% [error(Msg,Pos)]
add_all_perrors(Errors) :- add_all_perrors(Errors,[],bmachine).
add_all_perrors(Errors,Files) :- add_all_perrors(Errors,Files,bmachine).
add_all_perrors(Errors,Files,Context) :- add_all_perrors(Errors,Files,Context,0,100).
add_all_perrors([],_,_,_,_).
add_all_perrors([H|Rest],_Filenames,ContextOfError,NrOfErrors,MaxErrors) :-
    NrOfErrors >= MaxErrors,!,
    length([H|Rest],Len),
    add_error(ContextOfError,'Reached maximium number of errors; number of errors not reported: ',Len).
add_all_perrors([error(Msg,Pos)|Rest],Filenames,ContextOfError,NrOfErrors,MaxErrors) :-
	( Pos = pos(_,Filenumber,Srow,Scol,Erow,Ecol) ->
	    Ecol1 is Ecol+1,
	    ( get_filename(Filenumber,Filenames,Filename) ->
		add_error(ContextOfError,Msg,Filename,src_span(Filenumber,Filename,Srow,Scol,Erow,Ecol1,[],[]))
	    ; otherwise ->
		add_error(ContextOfError,Msg,'',src_span(Srow,Scol,Erow,Ecol1,[],[])))
	; Pos = pos(Srow,Scol,Filename) ->
	    add_error(ContextOfError,Msg,'',src_position_with_filename(Srow,Scol,1,Filename))
	; Pos = pos(Srow,Scol) -> /* deprecated, should no longer appear */
	    print(deprecated_pos(Srow,Scol)),nl,
	    add_error(ContextOfError,Msg,'',src_position(Srow,Scol,none,1)) /* none=offset */
	; otherwise ->
	    add_error(ContextOfError,Msg,'',Pos)),
	N1 is NrOfErrors+1,
	add_all_perrors(Rest,Filenames,ContextOfError,N1,MaxErrors).

get_filename(Filenumber,Filenames,Filename) :-
	number(Filenumber),!,
	nth1(Filenumber,Filenames,Filename).
get_filename(N,F,R) :- add_error(get_filename,'Internal error: ',get_filename(N,F,R)),
   F=[R]. % if there is only one file then return that file

get_error(Source,ErrMsg) :-
   retract(logged_error(Source,ErrMsg,_,_)).

get_error(Msg) :-
    retract(logged_error(Source,ErrMsg,Context,Span)),
    (extract_line_col(Span,StartLine,StartCol,_EndLine,_EndCol)
      -> (string_concatenate(StartCol,';;',Sp0),
          string_concatenate(', Column: ',Sp0,Sp1),
          string_concatenate(StartLine,Sp1,Sp2),
          string_concatenate('; ### Line: ',Sp2,Sp3),
          (extract_file_name(Span,Filename)
            -> string_concatenate(Filename,Sp3,Sp4),
               string_concatenate('; ### File: ',Sp4,EndStr)
             ; EndStr = Sp3
          )
         )
      ;  (EndStr = ';;')
     ),
    ((Context='' ; Context =[])
      ->  string_concatenate(ErrMsg,EndStr,S0)
       ;  safe_translate_error_term(Context,CString),
          string_concatenate(CString,EndStr,CString2),
          string_concatenate(':  Context = ',CString2,CString3),
          string_concatenate(ErrMsg,CString3,S0)
     ),
    (debug:debug_mode(off) -> Msg = S0
      ;  string_concatenate(Source,': ',S1),
         string_concatenate(S1,S0,Msg)
     ).


get_all_errors(All) :-
    findall(Err,get_error(Err),All),
    All \= [], clear_error_context.

get_all_errors_and_reset(All) :- findall(Err,get_error(Err),All),reset_errors,All \= [].

test_error_occurred(Source) :- logged_error(Source,_,_,_),real_error_occurred,!.

%:- use_module(library(lists),[length/2]). length is builtin

count_errors_occurred(Source,NrOfErrors) :-
  findall(1,logged_error(Source,_,_,_),Ls),
  length(Ls,NrOfErrors).


:- volatile current_context/1.
:- dynamic current_context/1.

/* add context information; will be added to any error messages */
/* will be removed upon backtracking */
%push_error_context(_Ctx) :- !.
set_error_context(Ctx) :- % print(ctx(Ctx)),nl,
  retractall(current_context(_)),
  asserta(current_context(Ctx)).
%push_error_context(_Ctx) :-
%  (retract(current_context(_Ctx)) -> true ; print_error(could_not_retract_context(_Ctx))).

clear_error_context :- %print(clearing_ctx),nl,
   retractall(current_context(_)).

%pop_error_context :-
%  (retract(current_context(_Ctx)) -> true ; print_error(could_not_pop_context(_Ctx))).

get_error_context(C) :- findall(Ct,current_context(Ct),C).


get_error_span(Source,Span) :-
   retract(logged_error_spans(Source,Span)).


get_error_span_linecol(Line,Col,EndLine,EndCol) :-
    get_error_span(_Source,Span),  % looks at logged_error_spans
    extract_line_col(Span,Line,Col,EndLine,EndCol).


extract_line_col([],_,_,_,_) :- !,fail.
extract_line_col([H|T],Srow,Scol,Erow,Ecol) :- !,
   (member(nodeid(N),[H|T])
     -> extract_line_col(N,Srow,Scol,Erow,Ecol) % we have a B Expression info field
   % TO DO: add rodinpos
    ; (member(N,[H|T]),extract_line_col(N,Srow,Scol,Erow,Ecol)) -> true
   ).
extract_line_col(b(_,_,Infos),Srow,Scol,Erow,Ecol) :- !,
  extract_line_col(Infos,Srow,Scol,Erow,Ecol).
extract_line_col(span_predicate(P,_,_),Srow,Scol,Erow,Ecol) :- !,
  extract_line_col(P,Srow,Scol,Erow,Ecol).
extract_line_col(pos(_,_Filenumber,Srow,Scol,Erow,Ecol),Srow,S1,Erow,E1) :-  number(Srow),
  % _Filenumber is an index into list provided by b_get_all_used_filenames
  number(Scol), number(Erow), number(Ecol),!,
  (Scol>0 -> S1 is Scol-1 ; S1 = Scol),  % the BParser starts column numbering at 1
  E1=Ecol. %Somehow the Ecol info is ok for Tcl
extract_line_col(unknown,_,_,_,_) :- !,fail.
extract_line_col(src_span(_Filenumber,_Filename,StartLine,SC,EndLine,EC,Offset,LenBytes),StartLine,StartCol,EndLine,EndCol)
 :- !, extract_line_col(src_span(StartLine,SC,EndLine,EC,Offset,LenBytes),StartLine,StartCol,EndLine,EndCol).
extract_line_col(src_span(StartLine,SC,EndLine,EC,_Off,[]),StartLine,StartCol,EndLine,EndCol)
 :-  !, StartCol is SC-1, EndCol is EC-1.
extract_line_col(src_span(SL,SC,EL,EC,_Off,_LenBytes),StartLine,StartCol,EndLine,EndCol) :- !,
   StartLine=SL,StartCol is SC-1,EndLine = EL, EndCol is EC-1.
 %:- !, StartCol is SC-1, EndCol is StartCol+LenBytes, EndLine = StartLine. %% fix because new library generates erroneous EndLine/EndCol info
% EndCol is EC-1.
extract_line_col(span_info(_Info,Span),SL,SC,EL,EC) :- !, extract_line_col(Span,SL,SC,EL,EC).
extract_line_col(src_span_operator(SP1,SP2),SL,SC,EL,EC) :- !,
 (extract_line_col(SP1,SL,SC,EL,EC) ; extract_line_col(SP2,SL,SC,EL,EC)).
extract_line_col(multi_span(SL,SC,EL,EC,_Off,_Len,TAIL),StartLine,StartCol,EndLine,EndCol) :- !,
  ((StartLine = SL, StartCol is SC-1,
     EndLine = EL, EndCol is EC-1
     )
   ; extract_line_col(TAIL,StartLine,StartCol,EndLine,EndCol)).
extract_line_col(src_position(StartLine,SC,_Off,LenBytes),StartLine,StartCol,EndLine,EndCol) :- !,
   StartCol is SC-1, EndCol is StartCol+LenBytes, EndLine is StartLine.
extract_line_col(src_position_with_filename(StartLine,SC,LenBytes,_Filename),StartLine,StartCol,EndLine,EndCol) :- !,
   StartCol is SC-1, EndCol is StartCol+LenBytes, EndLine is StartLine.
extract_line_col(lineCol(StartLine,StartCol),StartLine,StartCol,EndLine,EndCol) :- !,
   EndLine=StartLine, EndCol=StartCol.
extract_line_col(Struct,StartLine,StartCol,EndLine,EndCol) :-
    \+ atomic(Struct), Struct =.. [_|Args],
      /* recursively look for error spans inside */
    member(Arg,Args), extract_line_col(Arg,StartLine,StartCol,EndLine,EndCol),!.
%extract_line_col(S,_,_,_,_) :- print('COULD NOT EXTRACT LINE/COL INFO: '), print(S),
%    functor(S,F,N), print(' '),print(F/N),nl,nl,fail.


/* For TCL TK: extract all line_col information in one go and in 4 lists */
extract_all_line_col(Span, StartLineList,StartColList,EndLineList,EndColList) :-
    findall(lc(SL,SC,EL,EC), extract_line_col(Span,SL,SC,EL,EC),All),
    %print(all(All,Span)),nl,
    split_lc(All,StartLineList,StartColList,EndLineList,EndColList).

split_lc([], [],[],[],[]).
split_lc([lc(SL,SC,EL,EC)|T],[SL|T1],[SC|T2],[EL|T3],[EC|T4]) :-
   split_lc(T,T1,T2,T3,T4).

% TO DO: merge with line_col and in Tcl/Tk use info to highlight proper file !

extract_file_name(Pos,Filename) :- extract_file_number_and_name(Pos,_,Filename).
extract_subsidiary_tail_file_name(Pos,Filename) :-
   extract_file_number_and_name(Pos,Nr,FullFilename),
   not_main_file(Nr,FullFilename),
   tools:get_tail_filename(FullFilename,Filename).

% a version of extract_line_col which fails if the error is not in the main file
extract_line_col_for_main_file(Span,Line,Col,EndLine,EndCol) :-
   extract_line_col(Span,Line,Col,EndLine,EndCol),
   \+((extract_file_number_and_name(Span,Nr,FullFilename),
       not_main_file(Nr,FullFilename))).


extract_file_number_and_name(pos(_,Filenumber,_,_,_,_), Filenumber,Filename) :-
   bmachine:b_get_all_used_filenames(FN), get_filename(Filenumber,FN,Filename),!.
extract_file_number_and_name(pos(_,Filenumber,_,_,_,_), Filenumber,unknown(Filenumber)) :- !.
extract_file_number_and_name(src_span(_,Filename,_,_,_,_,_,_), unknown, Filename) :- !.
extract_file_number_and_name(src_position_with_filename(_,_,_,Filename), unknown, Filename) :- !.
extract_file_number_and_name(Struct,Filenumber,Filename) :-
    \+ atomic(Struct), Struct =.. [_|Args],
      /* recursively look for error spans inside */
    member(Arg,Args), extract_file_number_and_name(Arg,Filenumber,Filename),!.

%:- use_module(bmachine,[b_get_all_used_filenames/1,b_get_main_filenumber/1]).
position_is_not_in_main_file(pos(_,Filenumber,_,_,_,_)) :- !, % B position info
   not_main_filenumber(Filenumber).
position_is_not_in_main_file(src_span(_,Filename,_,_,_,_,_,_)) :- !,
   specfile:currently_opened_file(Main), Main \= Filename.
position_is_not_in_main_file(src_position_with_filename(_,_,_,Filename)) :- !,
   specfile:currently_opened_file(Main), Main \= Filename.
position_is_not_in_main_file(Struct) :- \+ atomic(Struct), Struct =.. [_|Args],
      /* recursively look for error spans inside */
    member(Arg,Args), position_is_not_in_main_file(Arg),!.


not_main_file(Filenumber,_) :- number(Filenumber),!, % B Position info
    not_main_filenumber(Filenumber).
not_main_file(_,Filename) :- % probably CSP position info; or unknown
    specfile:currently_opened_file(Main), Main \= Filename.

not_main_filenumber(Filenumber) :-
  bmachine:b_get_main_filenumber(MainFN),
  MainFN \= Filenumber.
