% (c) 2009-2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(evalstores,[
                      evalstore_reset/0
                     ,evalstore_create_type_env_by_state/2
                     ,evalstore_create_store_by_state/2
                     ,evalstore_eval_formula/4
                     ,evalstore_bindings/2
                     ]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,animator).
:- module_info(description,'This module provides methods to construct stores dynamically and evaluate expressions in them.').

:- use_module(library(lists)).
:- use_module(library(samsort),[samsort/3]).
:- use_module(tools,[safe_time_out/3]). %:- use_module(library(timeout),[time_out/3]).

:- use_module(state_space,     [visited_expression/2]).
:- use_module(specfile,        [state_corresponds_to_set_up_constants/2
                               ,state_corresponds_to_initialised_b_machine/1]).
:- use_module(bsyntaxtree,     [create_texpr/4
                               ,get_texpr_type/2]).
:- use_module(b_global_sets,   [is_b_global_constant/3
                               ,b_global_set/1]).
:- use_module(bmachine,        [full_b_machine/1
                               ,b_get_machine_constants/1
                               ,b_get_machine_variables/1]).
:- use_module(bmachine_construction,[type_open_formula/7]).
:- use_module(btypechecker,    [env_empty/1
                               ,add_identifiers_to_environment/3
                               ,env_lookup_type/3
                               ,env_identifiers/2]).
:- use_module(store,           [empty_state/1
                               ,normalise_store/2
                               ,merge_updates/3]).
:- use_module(kernel_waitflags,[init_wait_flags/1
                               ,ground_wait_flags/1]).
:- use_module(b_interpreter,   [set_up_typed_localstate/6
                               ,b_test_boolean_expression/4
                               ,b_compute_expression/5]).
:- use_module(b_enumerate,     [b_tighter_enumerate_values/2]).
:- use_module(error_manager,   [get_all_errors_and_reset/1
                               ,reset_errors/0]).

:- use_module(extension('user_signal/user_signal'),[protect_from_user_interrupt_det/1]).
:- use_module(extension('probhash/probhash'),[hash_term/2]).

% Each node represents a variable binding like a state in the state space
% The main difference to a state in the state space is that we usually have an evolution
% of nodes by adding successively new variables.

% A node is encoded with terms that describe combinations / references to other states
%  (see expand_node/2 for details).
% 

:- dynamic stored_evalstores/3.          % stored_evalstores(EvalstoreId,Store,TypeEnvId)
:- dynamic evalstore_hashes/2.           % evalstore_hashes(Hash,EvalstoreId)
:- dynamic next_evalstore_id/1.          % next_evalstore_id(EvalstoreId)

:- dynamic stored_type_environment/2.   % stored_type_environment(TypeEnvId,TypeEnv)
:- dynamic type_environment_hashes/2.   % type_environment_hashes(Hash,TypeEnvId)
:- dynamic next_type_environment_id/1.  % next_type_environment_id(TypeEnvId)
%:- dynamic standard_type_env/2.         % standard_type_env(Description,TypeEnvId)

% remove all existing nodes
evalstore_reset :-
    protect_from_user_interrupt_det(reset_evalstores2).
reset_evalstores2 :-
    retractall(stored_evalstores(_,_,_)),
    retractall(evalstore_hashes(_,_)),
    retractall(next_evalstore_id(_)),
    assert(next_evalstore_id(1)),
    retractall(stored_type_environment(_,_)),
    retractall(next_type_environment_id(_)),
    retractall(type_environment_hashes(_,_)),
    assert(next_type_environment_id(1)).


store_type_environment(Env,Id) :-
    hash_type_env(Env,Hash),
    ( type_environment_hashes(Hash,ExistingId) ->
        Id = ExistingId
    ; otherwise ->
        assert_type_env(Env,Hash,Id)).
assert_type_env(Env,Hash,Id) :-
    protect_from_user_interrupt_det(assert_type_env2(Env,Hash,Id1)),
    Id1=Id.
assert_type_env2(Env,Hash,Id) :-
    retract(next_type_environment_id(Id)),
    Next is Id+1,
    assert(next_type_environment_id(Next)),
    assert(stored_type_environment(Id,Env)),
    assert(type_environment_hashes(Hash,Id)).
    
create_new_node(NodeExpr,TypeEnvId,EvalstoreId) :-
    expand_node(NodeExpr,Store),
    hash_node(Store,TypeEnvId,Hash),
    ( evalstore_hashes(Hash,Id) ->
        EvalstoreId = Id
    ; otherwise ->
        assert_new_node(NodeExpr,Hash,TypeEnvId,EvalstoreId)).

hash_node(Store,TypeEnvId,Hash) :-
    hash_term(Store/TypeEnvId,Hash).

assert_new_node(Node,Hash,TypeEnvId,Id) :-
    % the assert should be protected from interrupts to prevent an inconsistent database
    protect_from_user_interrupt_det(assert_new_node2(Node,Hash,TypeEnvId,Id2)),
    Id=Id2.
assert_new_node2(Node,Hash,TypeEnvId,Id) :-
    retract(next_evalstore_id(Id)),
    Next is Id+1,
    assert(next_evalstore_id(Next)),
    assert(evalstore_hashes(Hash,Id)),
    assert(stored_evalstores(Id,Node,TypeEnvId)).

% expand_node(NodeExpr,Store):
%  get the full store that the node represents
expand_node(Node,Store) :-
    expand_node2(Node,UStore),
    samsort(cmp_bind,UStore,Store).
expand_node2(empty,Store) :-
    empty_state(Store).
expand_node2(state(StateId),Store) :- % refers to a state in the state space
    ( StateId=root -> expand_node2(empty,Store)
    ; otherwise ->
        visited_expression(StateId,State),
        state_corresponds_to_set_up_constants(State,Store)).
expand_node2(store(Store),Store).       % <- directly defines a store
expand_node2(update(Orig,New),Store) :- % <- updates the values of a store by another one,
    expand_node2(Orig,OrigStore),       %    the new store overrides old values
    expand_node2(New,NewStore),
    merge_updates(OrigStore,NewStore,Store).

cmp_bind(bind(A,_),bind(B,_)) :- A @< B.

% Generate a hash of the normalised type environment
hash_type_env(Env,Hash) :-
    env_identifiers(Env,Ids), % Ids is sorted
    maplist(lookup_for_type(Env),Ids,Types),
    hash_term(Types,Hash).
lookup_for_type(Env,Id,type(Id,Type)) :-
    env_lookup_type(Id,Env,Type).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Type Environments
extract_type_environment(State,Env) :-
    % by starting with a list of sorted identifiers, we try to make
    % sure that two environments with the same identifiers and types
    % are treated as equal.
    extract_type_environment2(State,UIds),
    create_normalised_type_environment(UIds,Env).
create_normalised_type_environment(UIds,Env) :-
    sort(UIds,Ids),
    env_empty(Empty),
    add_identifiers_to_environment(Ids,Empty,Env).

extract_type_environment2(root,Ids) :-
    % the root node has already some identifiers in the type environment:
    % deferred sets, enumerated sets and their elements
    !,global_identifiers(Ids,[]).
extract_type_environment2(State,Ids) :-
    state_corresponds_to_initialised_b_machine(State),!,
    global_identifiers(Ids,I1),
    constant_identifiers(I1,I2),
    variable_identifiers(I2,[]).
extract_type_environment2(_State,Ids) :-
    global_identifiers(Ids,I1),
    constant_identifiers(I1,[]).

global_identifiers -->
    % the enumerated set elements:
    findall(TId,
            (is_b_global_constant(G,_,N),create_texpr(identifier(N),global(G),[],TId))),
    % the deferred/enumerated sets:
    findall(TId,
            (b_global_set(G),create_texpr(identifier(G),set(global(G)),[],TId))).
constant_identifiers(In,Out) :-
    b_get_machine_constants(Csts),
    append(Csts,Out,In).
variable_identifiers(In,Out) :-    
    b_get_machine_variables(Vars),
    append(Vars,Out,In).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% evaluation
eval_in_evalstore(TypedFormula,StoreExpr,NewIdentifiers,Timeout,Result) :-
    reset_errors,
    expand_node(StoreExpr,Store),
    get_texpr_type(TypedFormula,Type),
    empty_state(Empty),
    set_up_typed_localstate(NewIdentifiers,_NewValues,TypedVals,Empty,LocalStore,positive),
    init_wait_flags(WF),
    b_tighter_enumerate_values(TypedVals,WF),
    time_out_eval(Timeout,Type,TypedFormula,LocalStore,Store,WF,Result1,TOResult),
    ( TOResult \== success ->
        Result = timeout
    ; error_occurred(Result1,Msg) ->
        Result = error(Msg)
    ; Result1 = ok(V,ResultStore) ->
        update_store(StoreExpr,ResultStore,NewStore),
        Result = ok(V,NewStore)
    ).
update_store(OldStore,AdditionalStore,NewStore) :-
    empty_state(AdditionalStore),!,OldStore=NewStore.
update_store(OldStore,AdditionalStore,update(OldStore,store(AdditionalStore))).

time_out_eval(none,Type,TypedFormula,LocalStore,Store,WF,Result,TOResult) :-
    !,evaluate_formula(Type,TypedFormula,LocalStore,Store,WF,Result1),
    Result=Result1,TOResult=success.
time_out_eval(Timeout,Type,TypedFormula,LocalStore,Store,WF,Result,TOResult) :-
    safe_time_out( evaluate_formula(Type,TypedFormula,LocalStore,Store,WF,Result),
              Timeout,
              TOResult),!.

evaluate_formula(pred,F,LS,S,WF,ok(pred(V),ResultStore)) :- !,
    ( b_test_boolean_expression(F,LS,S,WF), ground_wait_flags(WF) ->
        normalise_store(LS,ResultStore), V=true
    ; otherwise ->
        empty_state(ResultStore), V=false).
evaluate_formula(subst,_F,_LS,_S_,_WF,error('Substitutions are not yet supported')) :- !.
evaluate_formula(_Type,F,LS,S,WF,ok(value(Value),ResultStore)) :-
    b_compute_expression(F,LS,S,Value,WF),!,
    empty_state(ResultStore).
    
:- use_module(library(codesio),[write_to_codes/2]).
error_occurred(error(ErrorMsg),ErrorMsg) :- !.
error_occurred(_,ErrorMsg) :-
    get_all_errors_and_reset([Error|_]),
    write_to_codes(Error,ErrorCodes),
    limit_error(ErrorCodes,ErrorCodes2),
    atom_codes(ErrorMsg,ErrorCodes2).

% copied from bvisual2:
limit_error(ErrorCodes,ErrorCodes) :-
    length(ErrorCodes,L),L < 200,!.
limit_error(In,Out) :-
    prefix_length(In,Start,200),
    append(Start,"...",Out).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% type checking
evalstore_typecheck(Raw,EnvId,Result) :-
    stored_type_environment(EnvId,Env),
    full_b_machine(Machine),
    type_open_formula(Raw,[env(Env)],Machine,_Type,NewIdentifiers,TExpr,Errors),
    ( Errors = [] ->
        evalstore_typecheck_update_typeenv(NewIdentifiers,Env,EnvId,NewEnvId),
        Result=ok(TExpr,NewIdentifiers,NewEnvId)
    ; otherwise ->
        Result=errors(Errors)).
evalstore_typecheck_update_typeenv([],_Env,EnvId,NewEnvId) :-
    !, EnvId = NewEnvId.
evalstore_typecheck_update_typeenv(NewIdentifiers,Env,_EnvId,NewEnvId) :-
    add_identifiers_to_environment(NewIdentifiers,Env,NewEnv),
    store_type_environment(NewEnv,NewEnvId).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% entry points
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% type environments

evalstore_create_type_env_by_state(StateId,EnvId) :-
    ( StateId=root -> State=root
    ; otherwise    -> visited_expression(StateId,State)),
    extract_type_environment(State,Env),
    store_type_environment(Env,EnvId).

evalstore_create_store_by_state(StateId,EvalstoreId) :-
    evalstore_create_type_env_by_state(StateId,EnvId),
    create_new_node(state(StateId),EnvId,EvalstoreId).

evalstore_eval_formula(EvalstoreId,Raw,Timeout,Result) :-
    stored_evalstores(EvalstoreId,StoreExpr,EnvId),
    evalstore_typecheck(Raw,EnvId,TypeResult),
    evalstore_eval_formula2(TypeResult,StoreExpr,Timeout,Result).
evalstore_eval_formula2(errors(Errors),_StoreExpr,_Timeout,errors(Errors)).
evalstore_eval_formula2(ok(TExpr,NewIdentifiers,NewEnvId),StoreExpr,Timeout,Result) :-
    eval_in_evalstore(TExpr,StoreExpr,NewIdentifiers,Timeout,EvalResult),
    evalstore_eval_formula3(EvalResult,NewIdentifiers,NewEnvId,Result).
evalstore_eval_formula3(error(Msg),_Ids,_EnvId,errors([Msg])).
evalstore_eval_formula3(timeout,_Ids,_EnvId,timeout).
evalstore_eval_formula3(ok(Value,NewStore),Ids,EnvId,ok(Value,Ids,StoreId)) :-
    create_new_node(NewStore,EnvId,StoreId).

evalstore_bindings(EvalstoreId,Binding) :-
    stored_evalstores(EvalstoreId,StoreExpr,_),
    expand_node(StoreExpr,Binding).

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(specification_initialised,evalstore_reset,
                           'Initialise module evalstore').
