% (c) 2012-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(external_functions,[call_external_function/5, call_external_predicate/6,
                              call_external_substitution/6,
                              reset_external_functions/0,
                              external_function_argument_list/2,
                              external_pred_always_true/1,
                              check_if_predicates_were_evaluated/1,
                              reset_side_effect_occurred/0,
                              side_effect_occurred/1,
                              performs_io/1, not_declarative/1,
                              reset_argv/0, set_argv_from_list/1, set_argv_from_atom/1,
                              
                              observe_parameters/2
                              ]).

% Note: there is also the predicate external_procedure_used/1 in bmachine_construction

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(library(avl)).
:- use_module(library(lists)).
:- use_module(translate).

% TO DO: add external Prolog files
% Should we do auto-conversion of B to Prolog types ,... ?

:- use_module(kernel_waitflags).
:- use_module(debug).

:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This module provides a way to call external functions and predicates.').

:- dynamic side_effect_occurred/1.
% TRUE if a side effect has occured
% possible values (user_output, file)
reset_side_effect_occurred :- retractall(side_effect_occurred(_)).
assert_side_effect_occurred(Location) :-
  (side_effect_occurred(Location) -> true ; assert(side_effect_occurred(Location))).

:- discontiguous expects_waitflag/1.
:- discontiguous expects_spaninfo/1.
:- discontiguous performs_io/1.
:- discontiguous is_not_declarative/1. % does not perform io but is still not declarative
:- discontiguous does_not_modify_state/1.
:- discontiguous external_function_argument_list/2.
:- discontiguous expects_unevaluated_args/1.
:- discontiguous external_pred_always_true/1. % external predicates which are always true

not_declarative(FunName) :- (performs_io(FunName) -> true ; is_not_declarative(FunName)).


:- assert_must_succeed(( external_functions:call_external_function('COS',[int(0)],int(100),unknown,_WF) )).
:- assert_must_succeed(( external_functions:call_external_function('SIN',[int(0)],int(0),unknown,_WF) )).
% first a few hard-wired external functions for efficiency:
call_external_function('COS',[X],Value,_,_WF) :- !, 'COS'(X,Value). % on_exception ?
call_external_function('SIN',[X],Value,_,_WF) :- !, 'SIN'(X,Value).
call_external_function('TAN',[X],Value,_,_WF) :- !, 'TAN'(X,Value).
% now the generic case:
call_external_function(F,Args,Value,SpanInfo,WF) :-
   (expects_waitflag(F) -> E1=[WF] ; E1=[]),
   (expects_spaninfo(F) -> E2=[SpanInfo|E1] ; E2=E1),
   append(Args,[Value|E2],AV),
    % Check if there is better way to do this
   Call =.. [F|AV],
   %print('EXTERNAL FUNCTION CALL: '), print(F), print(' '), print(Call),nl,
   safe_call(Call,SpanInfo)
   %, print(result(Value)),nl.
   .

call_external_predicate(F,Args,EvaluatedArgs,PREDRES,SpanInfo,WF) :-
   (expects_waitflag(F) -> E1=[WF] ; E1=[]),
   (expects_spaninfo(F) -> E2=[SpanInfo|E1] ; E2=E1),
   (expects_unevaluated_args(F) -> E3=[Args|E2] ; E3=E2),
   append(EvaluatedArgs,[PREDRES|E3],AV),
   Call =.. [F|AV],
   % print('EXTERNAL PREDICATE CALL: '), print(Call),nl, %%
   safe_call(Call,SpanInfo).

call_external_substitution(FunName,Args,InState,OutState,Info,WF) :-
 % TO DO: signal if substitution can be backtracked
   get_wait_flag_for_subst(FunName,WF,Flag),
   (does_not_modify_state(FunName) -> OutState=[] ; OutState = OutState2),
   call_external_substitution2(Flag,FunName,Args,InState,OutState2,Info,WF).

get_wait_flag_for_subst(FunName,WF,Flag) :- perform_directly(FunName),!,
   get_wait_flag1(WF,Flag). % ensure that precondition, ... have been evaluated
get_wait_flag_for_subst(FunName,WF,Flag) :- performs_io(FunName),!,
   get_enumeration_finished_wait_flag(WF,Flag). % ensure that precondition, ... have been evaluated
%   get_last_wait_flag(external(FunName),WF,Flag).
get_wait_flag_for_subst(FunName,WF,Flag) :- get_binary_choice_wait_flag(external(FunName),WF,Flag).

:- block call_external_substitution2(-,?,?,?,?,?,?).
call_external_substitution2(_,F,Args,InState,OutState,_Info,WF) :-
   (expects_waitflag(F) -> E1=[WF] ; E1=[]),
   (expects_spaninfo(F) -> E2=[SpanInfo|E1] ; E2=E1),
   append(Args,[InState,OutState|E2],AV),
   Call =.. [F|AV],
   % print('EXTERNAL SUBSTITUTION CALL: '), print(Call),nl, %%
   safe_call(Call,SpanInfo).

safe_call(X,SpanInfo) :-
         on_exception(Exception, call(X),
                (add_exception_error(Exception,X,SpanInfo),fail)).

%safe_is(Context,X,Y) :- safe_is(Context,X,Y,unknown).
safe_is(Context,X,Y,Span) :-   % print(safe_is(Context,X,Y)),nl,
         on_exception(Exception, X is Y,
                (add_exception_error(Exception,Context,Span),fail)). %, print(res(X)),nl.

add_exception_error(error(existence_error(procedure,external_functions:FUN),_),X,SpanInfo) :- !,
   add_error(external_functions,'### EXTERNAL function does not exist: ',FUN:X,SpanInfo).
add_exception_error(error(undefined,ERR),X,SpanInfo) :-
   is_error(ERR),!,
   add_error(external_functions,'### Arithmetic operation not defined: ',X,SpanInfo).
add_exception_error(error(evaluation_error(undefined),ERR),X,SpanInfo) :-
   is_error(ERR),!,
   add_error(external_functions,'### Arithmetic operation not defined: ',X,SpanInfo).
add_exception_error(error(evaluation_error(zero_divisor),ERR),X,SpanInfo) :-
   is_error(ERR),!,
   add_error(external_functions,'### Division by 0 in arithmetic operation: ',X,SpanInfo).
add_exception_error(error(existence_error(source_sink,File),_),X,SpanInfo) :- !,
   add_error(external_functions,'### File does not exist (calling external function): ',File:X,SpanInfo).
add_exception_error(error(consistency_error(Msg,_,format_arguments),_),_X,SpanInfo) :- !,
   add_error(external_functions,'### Illegal format string (see SICStus Prolog reference manual for syntax): ',Msg,SpanInfo).
add_exception_error(enumeration_warning(A,B,C,D,E),X,SpanInfo) :-
   Exception = enumeration_warning(A,B,C,D,E),!,
   add_warning(external_functions,'### Enumeration warning occured while calling external function: ',X:Exception,SpanInfo),
   % or should we just print a message on the console ?
   throw(Exception).
add_exception_error(user_interrupt_signal,_X,_SpanInfo) :- !,
    print('User Interrupt during external function evaluation'),nl,
    throw(user_interrupt_signal).
add_exception_error(Exception,X,SpanInfo) :-
   add_error(external_functions,'### Exception occurred while calling external function: ',X:Exception,SpanInfo).

is_error(evaluation_error('is'(_,_),2,_,_)).
is_error(domain_error('is'(_,_),2,_,_)).

% STANDARD EXTERNAL FUNCTIONS


% -------------------------------

% MATHEMATICAL FUNCTIONS

:- public 'COS'/2.
:- block 'COS'(-,?). % redundant; we could add CLPFD constraints for R
'COS'(int(X),int(R)) :- cos2(100,X,R,unknown).
expects_spaninfo('COSx').
:- public 'COSx'/4.
:- block 'COSx'(-,?,?,?),'COSx'(?,-,?,?).
'COSx'(int(PrecisionMultiplier),int(X),int(R),Span) :- cos2(PrecisionMultiplier,X,R,Span).
:- block cos2(-,?,?,?),cos2(?,-,?,?).
cos2(Multiplier,X,R,Span) :- safe_is('COSx',R,round(Multiplier*cos(X/Multiplier)),Span).

:- public 'SIN'/2.
:- block 'SIN'(-,?). % redundant; we could add CLPFD constraints for R
'SIN'(int(X),int(R)) :- sin2(100,X,R,unknown).
expects_spaninfo('SINx').
:- public 'SINx'/4.
:- block 'SINx'(-,?,?,?),'SINx'(?,-,?,?).
'SINx'(int(PrecisionMultiplier),int(X),int(R),Span) :- %print('SINx'(PrecisionMultiplier,X,R)),nl,
  sin2(PrecisionMultiplier,X,R,Span).
:- block sin2(-,?,?,?),sin2(?,-,?,?).
sin2(Multiplier,X,R,Span) :- safe_is('SINx',R,round(Multiplier*sin(X/Multiplier)),Span).

:- public 'TAN'/2.
:- block 'TAN'(-,?). % redundant; we could add CLPFD constraints for R
'TAN'(int(X),int(R)) :- tan2(100,X,R,unknown).
expects_spaninfo('TANx').
:- public 'TANx'/4.
:- block 'TANx'(-,?,?,?),'TANx'(?,-,?,?).
'TANx'(int(PrecisionMultiplier),int(X),int(R),Span) :- tan2(PrecisionMultiplier,X,R,Span).
:- block tan2(?,-,?,?),tan2(-,?,?,?).
tan2(Multiplier,X,R,Span) :- safe_is('TANx',R,round(Multiplier*tan(X/Multiplier)),Span).


expects_spaninfo('PROLOG_FUN').
:- public 'PROLOG_FUN'/5.
'PROLOG_FUN'(string(Function),int(PrecisionMultiplier),int(Argument),int(Result),Span) :-
  prolog_fun2(Function,PrecisionMultiplier,Argument,Result,Span).
% Function can be:
% sin, cos, tan, cot, sinh, cosh, tanh, asin, acos, atan, acot, asinh, acosh, atanh, acoth, sqrt, log, exp
:- block prolog_fun2(?,-,?,?,?),prolog_fun2(-,?,?,?,?), prolog_fun2(?,?,-,?,?).
prolog_fun2(Function,Multiplier,X,R,Span) :-
   E =.. [Function,X/Multiplier],
   safe_is(Function,R,round(Multiplier*E),Span).

expects_spaninfo('LOGx').
:- public 'LOGx'/5.
:- block 'LOGx'(-,?,?,?,?),'LOGx'(?,-,?,?,?), 'LOGx'(?,?,-,?,?).
'LOGx'(int(PrecisionMultiplier),int(Base),int(X),int(R),Span) :-
  log2(PrecisionMultiplier,Base,X,R,Span).
:- block log2(?,-,?,?,?),log2(-,?,?,?,?),log2(?,?,-,?,?).
log2(Multiplier,Base,X,R,Span) :-
  safe_is('LOGx',R,round(Multiplier*log(Base/Multiplier,X/Multiplier)),Span).

:- public 'GCD'/3.
:- block 'GCD'(-,?,?),'GCD'(?,-,?).
'GCD'(int(X),int(Y),int(R)) :- gcd2(X,Y,R).
:- block gcd2(?,-,?),gcd2(-,?,?).
% we could assert abs(X) #>= R, abs(Y) #>= R
gcd2(X,Y,R) :- R is gcd(X,Y).

:- public 'ABS'/2.
:- use_module(clpfd_interface,[clpfd_eq_expr/2]).
:- block 'ABS'(-,-). % do we need this ??
'ABS'(int(X),int(R)) :-
   clpfd_eq_expr(R,abs(X)).


expects_spaninfo('MSB').
:- public 'MSB'/3.
:- block 'MSB'(-,?,?).
'MSB'(int(X),int(R),Span) :- msb2(X,R,Span). % most significant bit
:- block msb2(-,?,?).
msb2(X,R,Span) :- safe_is('MSB',R,msb(X),Span). % TO DO: could be made reversible by propagating constraint if R known

% -------------------------------

% THE CHOOSE operator (for TLA, recursive functions over sets,...)
% or choice operator from Abrial Book, page 60ff

:- use_module(custom_explicit_sets).
expects_waitflag('CHOOSE').
expects_spaninfo('CHOOSE').
:- public 'CHOOSE'/4.
:- block 'CHOOSE'(-,?,?,?).
'CHOOSE'([],_,Span,WF) :-
   add_wd_error_span('CHOOSE applied to empty set: ',[],Span,WF).
'CHOOSE'(CS,H,_,_) :- is_interval_closure_or_integerset(CS,Low,Up),!, choose_interval(Low,Up,H).
'CHOOSE'([H|T],C,_,_) :- when(ground([H|T]),(sort([H|T],SL),SL=[C|_])).
'CHOOSE'(avl_set(A),H,_,_) :- avl_min(A,H).
'CHOOSE'(global_set(GS),H,_,_) :- H = fd(1,GS). % assume GS not empty, integer sets covered above
'CHOOSE'(freetype(ID),H,Span,WF) :-
    (kernel_freetypes:enumerate_freetype(basic,FV,freetype(ID)) -> equal_object(FV,H)
     ; add_internal_error('Could not generate freetype element: ','CHOOSE'(freetype(ID),H,Span,WF))).
'CHOOSE'(closure(P,T,B),H,Span,WF) :-
   custom_explicit_sets:expand_custom_set(closure(P,T,B),ER),
   'CHOOSE'(ER,H,Span,WF).

:- block choose_interval(-,?,?).
choose_interval(Low,_Up,H) :- number(Low),!,H=int(Low).
choose_interval(_Low,Up,H) :- choose_interval2(Up,H). % lower-bound is minus infinity
:- block choose_interval2(-,?).
choose_interval2(Up,H) :- number(Up),!,H=int(Up). % for those intervals we choose Up
choose_interval2(_,int(0)). % we have INTEGER or something equivalent; simply choose 0


% Prolog term order on encoding
:- public 'LESS'/2.
:- block 'LESS'(-,?), 'LESS'(?,-).
'LESS'(string(A),string(B)) :- !, less2(A,B).
'LESS'(int(A),int(B)) :- !, less2(A,B). % use CLPFD less ?
'LESS'(fd(A,G),fd(B,G)) :- !, less2(A,B). % use CLPFD less
'LESS'(pred_false,B) :- !, B=pred_true.
'LESS'(pred_true,_) :- !, fail.
% TO DO: expand finite closures, ...
'LESS'(X,Y) :- when((ground(X),ground(Y)), X @< Y).

:- block less2(-,?), less2(?,-).
less2(A,B) :- A @< B.

% -------------------------------

% a first prototype for implementing recursive functions
% still a bit clumsy to define recursive functions (using STRING parameter for names)
% a restriction is that we can only have one version per machine, i.e.
%  one can only define recursive function in the CONSTANTS section and only
%  in a deterministic fashion not depending on enumerated constants,...
% also: we still need to ensure that the recurisve closures are not expanded
%  this should probably go hand in hand with the symbolic pragma

reset_external_functions :-
    retractall(recursive_function(_,_)),
    retractall(evaluated_predicate(_)),
    retractall(last_walltime(_)),
    reset_profiler, % maybe we should do this only if profiler was used ?
    close_all_files.

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(clear_specification,reset_external_functions,
                           'Reset B External Functions.').

:- volatile recursive_function/2.
:- dynamic recursive_function/2.

expects_waitflag('REC'). % DEPRECATED
expects_spaninfo('REC').
:- public 'REC'/5.
:- block 'REC'(?,-,?,?,?).
'REC'(string(FunName),ArgValue,Value,Span,WF) :-
  % print('REC'(FunName,ArgValue)),nl,
   (recursive_function(FunName,Fun)
    -> bsets_clp:apply_to(Fun,ArgValue,Value,Span,WF)
    ;  add_error(external_functions,'### RECURSIVE function does not exist: ',FunName,Span),
       fail).

expects_spaninfo('RECx'). % DEPRECATED
:- public 'RECx'/3.
:- block 'RECx'(-,?,?).
'RECx'(string(FunName),Result,Span) :-
       print('DEPRECATED RECx'(FunName)),nl, %%
   (recursive_function(FunName,FunBody)
    -> kernel_objects:equal_object(Result,FunBody,'RECx')
    ;  add_error(external_functions,'### RECURSIVE function does not exist: ',FunName,Span),
       fail).

:- use_module(custom_explicit_sets,[mark_closure_as_recursive/2]).
%expects_waitflag('REC_LET').
expects_spaninfo('REC_LET').  % DEPRECATED
external_function_argument_list('REC_LET',[normal,symbolic]).
:- public 'REC_LET'/4.
'REC_LET'(string(FunName),FunBody,Result,Span) :-
   when(ground(FunBody),
       ( mark_closure_as_recursive(FunBody,SFunBody),  % what we actually need is that FunBody is evaluated already with a symbolic annotation !
       % TO DO: also marks as possibly being non-wd (infinite loop; avoid over-eager computation by b_interpreter_check)
         store_recursive_function(FunName,SFunBody,Span),
         kernel_objects:equal_object(Result,FunBody,'REC_LET') % pass the argument through
         )).

store_recursive_function(FunName,FunBody,Span) :-
  print('STORING RECURSIVE FUNCTION : '), print(FunName),nl,  % print(FunBody),nl, %
  (recursive_function(FunName,OtherBody)
    -> (OtherBody=FunBody -> true
         ; add_error(external_functions,'### Multiple definitions of recursive function not yet allowed: ',FunName:FunBody:OtherBody,Span)
        )
    ;  assert(recursive_function(FunName,FunBody))
  ).

% -------------------------------

expects_waitflag('FORCE').
:- public 'FORCE'/3.
:- block 'FORCE'(-,?,?).
'FORCE'(avl_set(A),Result,_) :- !, equal_object(avl_set(A),Result).
'FORCE'([],Result,_) :- !, equal_object([],Result).
'FORCE'(closure(P,T,B),Result,_) :- !,
   print('FORCING : '),translate:print_bvalue(closure(P,T,B)),nl,
   custom_explicit_sets:expand_closure_to_avl_or_list(P,T,B,ExpandedValue,no_check),
   print('RESULT : '),translate:print_bvalue(ExpandedValue),nl,
   print('  ASSIGNING TO : '), translate:print_bvalue(Result),nl,
   equal_object(ExpandedValue,Result).
'FORCE'(A,B,_) :- equal_object_optimized(A,B,'FORCE').


% force enumeration of certain basic datatypes; at the moment this is quite limited : TO DO: extend
% maybe we also want to add priorities or several variables in one go (several variables can be sent as a pair though)
expects_waitflag('ENUMERATE').
:- public 'ENUMERATE'/3.
:- block 'ENUMERATE'(-,?,?).
'ENUMERATE'(Obj,pred_true,WF) :- get_wait_flag1(WF,WF1),
    format('*** ENUMERATE: ~w~n',Obj),
    enumerate_aux(Obj,WF1,WF).

:- block enumerate_aux(?,-,?).
enumerate_aux(fd(X,GS),_,WF) :-  !, kernel_objects:enumerate_basic_type_wf(fd(X,GS),global(GS),WF).
enumerate_aux(int(X),_,WF) :-  !, kernel_objects:enumerate_basic_type_wf(int(X),integer,WF).
enumerate_aux((A,B),WF1,WF) :- !, enumerate_aux2(A,WF1,WF), enumerate_aux2(B,WF1,WF).
enumerate_aux(Obj,_,_WF) :- add_error(external_functions,'Cannot enumerate value yet: ',Obj).

:- block enumerate_aux2(-,?,?).
enumerate_aux2(Obj,WF1,WF) :- enumerate_aux(Obj,WF1,WF).

% a predicate to ensure that certain variables are enumerated together:
% as soon as first value is known, the second value gets enumerated
expects_waitflag('LINKED_ENUMERATION').
:- public 'LINKED_ENUMERATION'/4.
:- block 'LINKED_ENUMERATION'(-,?,?,?).
'LINKED_ENUMERATION'(Obj1,Obj2,pred_true,WF) :- link_enum_aux(Obj1,Obj2,WF).

link_enum_aux(fd(X,_),Obj2,WF) :- !, link_enum_block(X,Obj2,WF).
link_enum_aux(int(X),Obj2,WF) :- !, link_enum_block(X,Obj2,WF).
link_enum_aux(string(X),Obj2,WF) :- !, link_enum_block(X,Obj2,WF).
link_enum_aux(Obj1,Obj2,WF) :- when(ground(Obj1),link_enum_block(1,Obj2,WF)).

:- block link_enum_block(-,?,?).
link_enum_block(_X,Obj2,WF) :- %print(linked_enumeration(_X,Obj2)), tools_printing:print_arg(Obj2),nl,
   get_wait_flag1(WF,WF1), % give chance to run deterministic propagations
   enumerate_aux(Obj2,WF1,WF).

% -------------------------------

:- assert_must_succeed(( external_functions:call_external_function('STRING_APPEND',[string(a),string(b)],string(ab),unknown,_WF) )).
% STRING MANIPULATION FUNCTIONS
:- use_module(kernel_waitflags).
% function to append two strings, reversible
expects_waitflag('STRING_APPEND').
:- public 'STRING_APPEND'/4.
:- block 'STRING_APPEND'(-,?,-,?), 'STRING_APPEND'(?,-,-,?).
'STRING_APPEND'(string(A),string(B),string(C),WF) :- %print(app(A,B,C)),nl,
  app2(A,B,C,WF).
:- block app2(-,?,-,?), app2(?,-,-,?).
app2(A,B,C,_WF) :- %print(app2(A,B,C)),nl,
   nonvar(A),nonvar(B),!,atom_concat(A,B,CC), /* overcome bug in SICStus; delay unification */
   CC=C. %, print(app2(A,B,C)),nl.
app2(A,B,C,WF) :- atom_codes(C,CC), app3(A,B,CC,WF).
app3(A,B,CC,WF) :-
   (   nonvar(B) -> atom_codes(B,BB), append(AA,BB,CC), atom_codes(A,AA)
     ; nonvar(A) -> atom_codes(A,AA), append(AA,BB,CC), atom_codes(B,BB)
     ; length(CC,CLen), %print(enum(CC,CLen)),nl,
       get_wait_flag(CLen,'STRING_APPEND',WF,LWF),
       app4(A,B,CC,WF,LWF)
    ).
:- block app4(-,?,-,?,-), app4(?,-,-,?,-).
app4(A,B,CC,WF,_) :- (nonvar(A) ; nonvar(B)), !, % no need to enumerate
   app3(A,B,CC,WF).
app4(A,B,CC,_WF,_) :- %print(enumerating(CC)),nl,
   append(AA,BB,CC), % will be non-deterministic
   atom_codes(A,AA), atom_codes(B,BB).

:- public 'STRING_CHARS'/2.
:- block 'STRING_CHARS'(-,-).
% to do: make it reversible
'STRING_CHARS'(Str,SeqRes) :- nonvar(Str), Str=string(A), ground(A),!,
   string_chars2(A,SeqRes).
'STRING_CHARS'(Str,SeqRes) :- when((ground(Str);ground(SeqRes)),string_chars1(Str,SeqRes)).

string_chars1(Str,SeqRes) :- nonvar(Str), Str=string(A), ground(A),!,
   string_chars2(A,SeqRes).
string_chars1(Str,SeqRes) :-
   custom_explicit_sets:expand_custom_set_to_list(SeqRes,ExpandedAtomList,Done,string_chars1),
   string_chars3(Str,SeqRes,ExpandedAtomList,Done).

string_chars2(A,SeqRes) :- atom_codes(A,AA), generate_char_seq(AA,1,CharSeq),
   equal_object_optimized(CharSeq,SeqRes,string_chars2).

:- block string_chars3(-,?,?,-).
string_chars3(Str,SeqRes,_ExpandedAtomList,_Done) :- nonvar(Str), Str=string(A), ground(A),!,
   string_chars2(A,SeqRes).
string_chars3(Str,_SeqRes,ExpandedAtomList,Done) :- nonvar(Done),!,
   sort(ExpandedAtomList,SL),
   maplist(drop_index,SL,IL),
   convert_b_to_prolog_atoms(IL,PL,Done2),
   when(nonvar(Done2),
      (tools:ajoin(PL,Atom),
       equal_object(Str,string(Atom)))).
string_chars3(Str,SeqRes,ExpandedAtomList,Done) :- % Str is only partially instantiated
   when((ground(Str);nonvar(Done)),string_chars3(Str,SeqRes,ExpandedAtomList,Done)).
generate_char_seq([],_,[]).
generate_char_seq([Code|T],Nr,[(int(Nr),string(CS))|TSeq]) :-
   atom_codes(CS,[Code]),
   N1 is Nr+1, generate_char_seq(T,N1,TSeq).

:- public 'STRING_CODES'/2.
:- block 'STRING_CODES'(-,-).
'STRING_CODES'(string(A),SeqRes) :- string_codes2(A,SeqRes).
:- block string_codes2(-,-).
string_codes2(A,SeqRes) :- nonvar(A),!,string_codes4(A,SeqRes).
string_codes2(A,SeqRes) :- SeqRes==[],!, empty_string_atom(A).
string_codes2(A,SeqRes) :- expand_custom_set_to_list(SeqRes,SeqList,_,string_codes2),
   when((nonvar(A);ground(SeqList)), string_codes3(A,SeqList)).

string_codes3(A,SeqRes) :- nonvar(A),!,string_codes4(A,SeqRes).
string_codes3(A,SeqRes) :- sort(SeqRes,SSeqRes),extract_codes(SSeqRes,1,Codes),
   atom_codes(A,Codes).
string_codes4(A,SeqRes) :- atom_codes(A,AA), generate_code_seq(AA,1,CodeSeq),
   equal_object_optimized(CodeSeq,SeqRes,string_codes4).

generate_code_seq([],_,[]).
generate_code_seq([Code|T],Nr,[(int(Nr),int(Code))|TSeq]) :-
   N1 is Nr+1, generate_code_seq(T,N1,TSeq).

extract_codes([],_,[]).
extract_codes([(int(Nr),int(Code))|T],N,[Code|CT]) :-
  (Nr==N -> true ; add_error(extract_codes,'Unexpected index: ',(Nr,N))),
  N1 is N+1, extract_codes(T,N1,CT).


empty_string(string('')).
empty_string_atom('').

:- assert_must_succeed(( external_functions:call_external_function('STRING_LENGTH',[string(a)],int(1),unknown,_WF) )).
% function to get length of a string
:- public 'STRING_LENGTH'/2.
'STRING_LENGTH'(SA,int(L)) :- greater_than_equal(int(L),int(0)),
   string_len2(SA,L).
:- block string_len2(-,-).
string_len2(SA,L)  :- L==0,!, empty_string(SA).
string_len2(string(A),L) :- string_len3(A,L).
:- block string_len3(-,-).
string_len3(A,L)  :- L==0,!, empty_string_atom(A).
% in case A is not known and L=1 we could enumerate chars ??
string_len3(A,L) :- string_len4(A,L).
:- block string_len4(-,?).
string_len4(A,L) :- atom_length(A,LL), LL=L. % delay unification due to bug in SICStus atom_length
% bug in SICStus: dif(X,1), atom_length(a,X) succeeds in 4.2.0 and 4.2.1

% function to split a string into a list of strings which were delimited by a separator
% WARNING: if the seperator is of length more than one, then first match-strategy will be used
:- public 'STRING_SPLIT'/3.
:- block 'STRING_SPLIT'(?,-,?), 'STRING_SPLIT'(-,?,-).
'STRING_SPLIT'(string(A),string(B),R) :- string_split2(A,B,R).
:- block string_split2(?,-,?),string_split2(-,?,-).
string_split2(Atom,Seperator,SplitAtomList) :- (var(Atom) ; var(Seperator)),!,
  custom_explicit_sets:expand_custom_set_to_list(SplitAtomList,ExpandedSplitAtomList,Done,string_split2),
  string_split3(Atom,Seperator,SplitAtomList,Done,ExpandedSplitAtomList).
string_split2(Atom,Separator,SplitAtomList) :- % normal forward mode: atom and separator known:
  string_split_forward(Atom,Separator,SplitAtomList).

string_split_forward(Atom,Separator,SplitAtomList) :-
  split_atom_string(Atom,Separator,List), % safe_call ?
  convert_prolog_to_b_list(List,SplitAtomList).

:- block string_split3(?,-,?,?,?),string_split3(-,?,?,-,?). % we need to know the seperator: TO DO : improve this
string_split3(Atom,Seperator,SplitAtomList,Done,_ExpandedSplitAtomList) :- var(Done),!,
   string_split_forward(Atom,Seperator,SplitAtomList).
string_split3(Atom,Seperator,_SplitAtomList,_Done,ExpandedSplitAtomList) :-
   ExpandedSplitAtomList \= [],  % split("",sep) --> [""] not the empty list
   sort(ExpandedSplitAtomList,SL),
   maplist(drop_index,SL,IL),
   convert_b_to_prolog_atoms(IL,PL,Done),
   atom_codes(Seperator,SepCodes),
   append(SepCodes,_,SepCodes2),
   when(nonvar(Done),
      (maplist(not_suffix_atom(SepCodes2),PL), % check that seperator occurs in no split atom: e.g. STRING_SPLIT(r,"_") = ["a","_","c"] should fail
       insert_sep(PL,Seperator,PL2),
       tools:ajoin(PL2,Atom))).
insert_sep([],_,[]).
insert_sep([H],_,R) :- !, R=[H].
insert_sep([H|T],Sep,[H,Sep|IT]) :- insert_sep(T,Sep,IT).

not_suffix_atom(SepCodes,Atom) :- \+ suffix_atom(SepCodes,Atom).
suffix_atom(SepCodes,Atom) :-
   atom_codes(Atom,AL),
   append(_,SepCodes,AL).

:- public 'STRING_JOIN'/3.
:- block 'STRING_JOIN'(?,-,?), 'STRING_JOIN'(-,?,?).
'STRING_JOIN'(SplitAtoms,string(Sep),Res) :- string_join2(SplitAtoms,Sep,Res).

% this is not reversible  ["a","b"],"_"  ->  "a_b"  but ["a_b"],"_" -> "a_b"
:- block string_join2(?,-,?),string_join2(-,?,?).
string_join2(SplitAtomList,Seperator,Result) :- 
  custom_explicit_sets:expand_custom_set_to_list(SplitAtomList,ExpandedSplitAtomList,Done,string_split2),
  string_join3(Result,Seperator,SplitAtomList,Done,ExpandedSplitAtomList).

:- block string_join3(?,-,?,?,?),string_join3(?,?,?,-,?). % we need to know the seperator: TO DO : improve this
string_join3(Result,Seperator,_SplitAtomList,_Done,ExpandedSplitAtomList) :-
   ExpandedSplitAtomList \= [],  % split("",sep) --> [""] not the empty list
   sort(ExpandedSplitAtomList,SL),
   maplist(drop_index,SL,IL),
   convert_b_to_prolog_atoms(IL,PL,Done2),
   when(nonvar(Done2),
     (insert_sep(PL,Seperator,PL2),
      tools:ajoin(PL2,Atom),
      equal_object(Result,string(Atom)))).


:- assert_must_succeed((external_functions:split_atom_string('ef,g',',',R), R==[ef,g])).
:- assert_must_succeed((external_functions:split_atom_string('ab,cd,ef,g',',',R), R==['ab','cd',ef,g])).
:- assert_must_succeed((external_functions:split_atom_string('ab','a',R), R==['','b'])).
:- assert_must_succeed((external_functions:split_atom_string('"STR1"','',R), R==['"STR1"'])).

split_atom_string(Atom,Sep,SplitList) :-
  atom_chars(Sep,SepAscii),
  (SepAscii=[] -> SplitList = [Atom]
   ; atom_chars(Atom,ListAscii),split3(ListAscii,SepAscii,[],SplitList)).

split3([],_,MatchSoFar,R) :- !,reverse(MatchSoFar,RM), atom_chars(Atom,RM),R=[Atom].
split3(List,Sep,MatchSoFar,Res) :-
   append(Sep,Tail,List),!, % we have a match with a separator
   reverse(MatchSoFar,RM), atom_chars(Atom,RM),
   Res=[Atom|R2], split3(Tail,Sep,[],R2).
split3([H|T],Sep,MatchSoFar,Res) :- % no match
   split3(T,Sep,[H|MatchSoFar],Res).


convert_prolog_to_b_list(PL,BL) :- convert_prolog_to_b_list_aux(PL,1,CPL),
  equal_object(CPL,BL,convert_prolog_to_b_list).

convert_prolog_to_b_list_aux([],_,[]).
convert_prolog_to_b_list_aux([H|T],Index,[(int(Index),CH)|CT]) :-
  convert_prolog_to_b_term(H,CH),
  I1 is Index+1, convert_prolog_to_b_list_aux(T,I1,CT).

convert_prolog_to_b_term(N,R) :- number(N),!, R=int(N).
convert_prolog_to_b_term(A,R) :- atomic(A),!, R=string(A).
convert_prolog_to_b_term(A,R) :-
   add_internal_error('Illegal Prolog term: ',convert_prolog_to_b_term(A,R)), R=A.


drop_index((int(_),R),R).

% a version that delays converting and sets Done to done when all B Atoms have been grounded
convert_b_to_prolog_atoms([],[],done).
convert_b_to_prolog_atoms([BAtom|T],[PrologAtom|PT],Done) :-
   convert_b_to_prolog_atoms_aux(BAtom,T,PrologAtom,PT,Done).

:- block convert_b_to_prolog_atoms_aux(-,?,?,?,?).
convert_b_to_prolog_atoms_aux(pred_true,T,PrologAtom,PT,Done) :- convert_b_to_prolog_atoms_aux2('TRUE',T,PrologAtom,PT,Done).
convert_b_to_prolog_atoms_aux(pred_false,T,PrologAtom,PT,Done) :- convert_b_to_prolog_atoms_aux2('FALSE',T,PrologAtom,PT,Done).
convert_b_to_prolog_atoms_aux(string(S),T,PrologAtom,PT,Done) :- 
   convert_b_to_prolog_atoms_aux2(S,T,PrologAtom,PT,Done).
convert_b_to_prolog_atoms_aux(int(I),T,PrologAtom,PT,Done) :- 
   convert_b_to_prolog_atoms_aux2(I,T,PrologAtom,PT,Done).

:- block convert_b_to_prolog_atoms_aux2(-,?,?,?,?).
convert_b_to_prolog_atoms_aux2(Atom,T,Atom,PT,Done) :- convert_b_to_prolog_atoms(T,PT,Done).


:- public 'STRING_TO_ENUM'/2.
:- block 'STRING_TO_ENUM'(-,-).
'STRING_TO_ENUM'(string(S),FD) :- string_to_enum2(S,FD).

:- block string_to_enum2(-,-).
string_to_enum2(S,FD) :- nonvar(S),!,
   (b_global_sets:is_b_global_constant(GS,Nr,S) ->
       FD = fd(Nr,GS)
     ; add_error('Could not convert string to enumerated set element: ',S),fail).
string_to_enum2(S,fd(Nr,GS)) :-
   string_to_enum3(S,Nr,GS).

:- block string_to_enum3(-,-,?),string_to_enum3(-,?,-).
string_to_enum3(S,Nr,GS) :- nonvar(S),!,
   (b_global_sets:is_b_global_constant(SGS,SNr,S) -> (SGS,SNr) = (GS,Nr)
     ; add_error('Could not convert string to enumerated set element: ',S),fail).
string_to_enum3(S,Nr,GS) :-
   (b_global_sets:is_b_global_constant(GS,Nr,GNS) -> GNS=S
     ; add_error('Could not convert deferred set element to string: ',GS:Nr),fail).

:- public 'STRING_TO_INT'/2.
:- block 'STRING_TO_INT'(-,-).
'STRING_TO_INT'(string(S),I) :- string_to_int2(S,I).

:- block string_to_int2(-,?).
string_to_int2(S,Res) :-
    atom_codes(S,C),
    on_exception(error(syntax_error(_),_), number_codes(Num,C),
         add_error_and_fail(external_functions,'### Could not convert string to integer: ',S)),
    Res=int(Num).

:- public 'INT_TO_STRING'/2.
:- block 'INT_TO_STRING'(-,?).
'INT_TO_STRING'(int(I),S) :- int_to_string2(I,S).

:- block int_to_string2(-,?).
int_to_string2(Num,Res) :- number_codes(Num,C), atom_codes(S,C), Res=string(S).

% -------------------------------

% STRING FILE SYSTEM FUNCTIONS


:- use_module(kernel_objects).
:- use_module(library(file_systems)).
:- public 'FILES'/2.
:- block 'FILES'(-,?).
'FILES'(string(A),L) :- files2(A,L).
:- block files2(-,?).
files2(Dir,List) :- b_absolute_file_name(Dir,ADir),
   findall(string(F),file_member_of_directory(ADir,F,_FLong),L),
   equal_object_optimized(L,List,files2).
:- public 'FULL_FILES'/2.
:- block 'FULL_FILES'(-,?).
'FULL_FILES'(string(A),L) :- full_files2(A,L).
:- block full_files2(-,?).
full_files2(Dir,List) :-
   b_absolute_file_name(Dir,ADir),
   findall(string(FLong),file_member_of_directory(ADir,_F,FLong),L),
   equal_object_optimized(L,List,full_files2).


:- public 'DIRECTORIES'/2.
:- block 'DIRECTORIES'(-,?).
'DIRECTORIES'(string(A),L) :- dir2(A,L).
:- block dir2(-,?).
dir2(Dir,List) :- b_absolute_file_name(Dir,ADir),
   findall(string(F),directory_member_of_directory(ADir,F,_FLong),L),
   equal_object_optimized(L,List,dir2).
:- public 'FULL_DIRECTORIES'/2.
:- block 'FULL_DIRECTORIES'(-,?).
'FULL_DIRECTORIES'(string(A),L) :- full_dir2(A,L).
:- block full_dir2(-,?).
full_dir2(Dir,List) :- b_absolute_file_name(Dir,ADir),
   findall(string(FLong),directory_member_of_directory(ADir,_,FLong),L),
   equal_object_optimized(L,List,full_dir2).

:- public 'FILE_EXISTS'/2.
:- block 'FILE_EXISTS'(-,?).
'FILE_EXISTS'(string(A),PREDRES) :- file_exists2(A,PREDRES).
:- block file_exists2(-,?).
file_exists2(A,PREDRES) :- b_absolute_file_name(A,AA),
   (file_exists(AA) -> PREDRES=pred_true ; PREDRES=pred_false).
:- public 'DIRECTORY_EXISTS'/2.
:- block 'DIRECTORY_EXISTS'(-,?).
'DIRECTORY_EXISTS'(string(A),PREDRES) :- dir_exists2(A,PREDRES).
:- block dir_exists2(-,?).
dir_exists2(A,PREDRES) :-  b_absolute_file_name(A,AA),
   (directory_exists(AA) -> PREDRES=pred_true ; PREDRES=pred_false).


b_absolute_file_name(File,AbsFileName) :-
   bmachine:b_get_main_filename(MainFileName),
   absolute_file_name(File,AbsFileName,[relative_to(MainFileName)]).

:- public 'FILE_PROPERTY'/3.
:- block 'FILE_PROPERTY'(-,?,?), 'FILE_PROPERTY'(?,-,?).
'FILE_PROPERTY'(string(A),string(P),BoolRes) :-
   file_property2(A,P,BoolRes).
:- block file_property2(-,?,?), file_property2(?,-,?).
file_property2(File,Prop,BoolRes) :-
    b_absolute_file_name(File,AFile),
    file_property(AFile,Prop) -> BoolRes=pred_true ; BoolRes=pred_false.
% possible properties: readable, writable, executable

:- public 'FILE_PROPERTY_VALUE'/3.
:- block 'FILE_PROPERTY_VALUE'(-,?,?), 'FILE_PROPERTY_VALUE'(?,-,?).
'FILE_PROPERTY_VALUE'(string(A),string(P),int(R)) :-
   file_property_val2(A,P,R).
:- block file_property_val2(-,?,?), file_property_val2(?,-,?).
file_property_val2(File,Prop,Res) :-
    b_absolute_file_name(File,AFile),
    file_property(AFile,Prop,Val) -> Res=Val
         ; add_error(external_functions,'Illegal FILE_PROPERTY',Prop),Res=int(-1).
% possible properties:
%  size_in_bytes, create_timestamp, modify_timestamp, access_timestamp
% + on unix the follwing is also an integer: owner_user_id, owner_group_id

:- public 'DIRECTORY_PROPERTY'/3.
:- block 'DIRECTORY_PROPERTY'(-,?,?), 'DIRECTORY_PROPERTY'(?,-,?).
'DIRECTORY_PROPERTY'(string(A),string(P),BoolRes) :-
   dir_property2(A,P,BoolRes).
:- block dir_property2(-,?,?), dir_property2(?,-,?).
dir_property2(File,Prop,BoolRes) :-
    b_absolute_file_name(File,AFile),
    directory_property(AFile,Prop) -> BoolRes=pred_true ; BoolRes=pred_false.

:- public 'DIRECTORY_PROPERTY_VALUE'/3.
:- block 'DIRECTORY_PROPERTY_VALUE'(-,?,?), 'DIRECTORY_PROPERTY_VALUE'(?,-,?).
'DIRECTORY_PROPERTY_VALUE'(string(A),string(P),int(R)) :-
   dir_property_val2(A,P,R).
:- block dir_property_val2(-,?,?), dir_property_val2(?,-,?).
dir_property_val2(File,Prop,Res) :-
    b_absolute_file_name(File,AFile),
    directory_property(AFile,Prop,Val) -> Res=Val
           ; add_error(external_functions,'Illegal DIRECTORY_PROPERTY',Prop), Res=int(-1).
% --------------------------------

% I/O


% printf as predicate; can also be used as a function to BOOL
expects_spaninfo('fprintf').
external_pred_always_true('fprintf').
:- public fprintf/5.
:- block fprintf(-,?,?,?,?), fprintf(?,-,?,?,?), fprintf(?,?,-,?,?).
fprintf(string(File),string(FormatString),Value,PredRes,Span) :- !, PredRes = pred_true,
   when((ground(FormatString),ground(Value)),
   fprintf2(File,FormatString,Value,Span)).
fprintf(File,F,V,P,S) :- add_internal_error('Illegal call: ',fprintf(File,F,V,P,S)), P=pred_true.

expects_spaninfo('printf').
external_pred_always_true('printf').
:- assert_must_succeed(external_functions:printf(string('test ok = ~w~n'),pred_true,pred_true,unknown)).
:- block printf(-,?,?,?), printf(?,-,?,?).
printf(string(FormatString),Value,PredRes,Span) :- !, PredRes = pred_true,
   when((ground(FormatString),ground(Value)),
         fprintf2(user_output,FormatString,Value,Span)).
printf(F,V,P,S) :- add_internal_error('Illegal call: ',printf(F,V,P,S)), P=pred_true.

expects_spaninfo('printf_nonvar'). % print as soon as value becomes nonvar (printf waits until the value is ground)
external_pred_always_true('printf_nonvar').
:- block printf_nonvar(-,?,?,?), printf_nonvar(?,-,?,?).
printf_nonvar(string(FormatString),Value,PredRes,Span) :- !, PredRes = pred_true,
   when((ground(FormatString),nonvar(Value)),
         fprintf2(user_output,FormatString,Value,Span)).
printf_nonvar(F,V,P,S) :- add_internal_error('Illegal call: ',printf(F,V,P,S)), P=pred_true.


expects_spaninfo('dprintf'). % a version of printf which delays setting itself to TRUE until printing is performed
external_pred_always_true('dprintf').
:- public dprintf/4.
:- block dprintf(-,?,?,?), dprintf(?,-,?,?).
dprintf(string(FormatString),Value,PredRes,Span) :- !,
   when((ground(FormatString),ground(Value)),
         (fprintf2(user_output,FormatString,Value,Span),PredRes = pred_true)).
dprintf(F,V,P,S) :- add_internal_error('Illegal call: ',dprintf(F,V,P,S)), P=pred_true.

fprintf2(File,FormatString,Value,Span) :-
   translate:translate_bvalue(Value,TV),
   file_format(File,FormatString,[TV],Span),
   print_stats,
   fprint_log_info(File,Span).
%fprintf2(File,FormatString,Value,Span) :- print('BACKTRACKING ') , file_format(File,FormatString,['?'],Span),nl,fail.
% warning: format can call an argument as goal with ~@; this is dangerous
% and we should prohibit this
% Luckily PRINTF("@: <>~@<>~n","print(danger)") where print(danger)
% is arbitrary code is not a security exploit:
% our PRINTF translates the argument to an atom, i.e., one could
% only call Prolog predicates without arguments, anything else leads to an exception
% (even PRINTF("@: <>~@<>~n","nl") leads to an exception)

:- dynamic nr_of_printfs/1.
nr_of_printfs(0).
print_stats :-
  external_fun_pref(printf_stats,PFS),   % use SET_PREF("printf_stats","TRUE")
  PFS='TRUE',
  retract(nr_of_printfs(N)),
  !,
  N1 is N+1,
  assert(nr_of_printfs(N1)),
  statistics(walltime,[X,_]),
  format('% Call ~w at walltime ~w ms~n',[N1,X]).
print_stats.

:- dynamic external_fun_pref/2.
% valid keys: log_info
valid_external_fun_pref(log_info).
valid_external_fun_pref(printf_stats).

fprint_log_info(File,Span) :- external_fun_pref(log_info,X),
   'GET_INFO'(string(X),string(Res)),
   file_format(File,'%%% ~w~n',[Res],Span).
fprint_log_info(_,_).

% example usage SET_PREF("log_info","time")
:- public 'SET_PREF'/3.
:- block 'SET_PREF'(-,?,?), 'SET_PREF'(?,-,?).
'SET_PREF'(string(S),string(Val),R) :-
   preferences:eclipse_preference(S,Pref),!, R=pred_true,
   format('Setting ProB preference ~w := ~w~n',[S,Val]),
   preferences:set_preference(Pref,Val).
'SET_PREF'(string(S),string(Val),R) :- valid_external_fun_pref(S),!,
   R=pred_true,
   retractall(external_fun_pref(S,_)),
   format('Setting external function preference ~w := ~w~n',[S,Val]),
   assert(external_fun_pref(S,Val)).
'SET_PREF'(P,V,R) :-
   add_internal_error('Illegal call: ','SET_PREF'(P,V)), R=pred_false.

% printf followed by fail; useful for print fail loops in the REPL
expects_spaninfo('printfail').
:- assert_must_fail(external_functions:printfail(string('test ok = ~w~n'),pred_true,pred_true,unknown)).
:- block printfail(-,?,?,?), printfail(?,-,?,?).
printfail(string(FormatString),Value,PredRes,Span) :- !, PredRes = pred_true,
   when((ground(FormatString),ground(Value)),
         (fprintf2(user_output,FormatString,Value,Span),fail)).
printfail(F,V,P,S) :-
   add_internal_error('Illegal call: ',printfail(F,V,P,S)), P=pred_true.

% a version of printf with automatic indenting and backtracking feedback
expects_spaninfo('iprintf').
external_pred_always_true('iprintf').
:- assert_must_succeed(external_functions:iprintf(string('test ok = ~w~n'),pred_true,pred_true,unknown)).
:- block iprintf(-,?,?,?), iprintf(?,-,?,?).
iprintf(string(FormatString),Value,PredRes,Span) :- !, PredRes = pred_true,
   when((ground(FormatString),ground(Value)),
         ifprintf2(user_output,FormatString,Value,no_trace,Span)).
iprintf(F,V,P,S) :- add_internal_error('Illegal call: ',iprintf(F,V,P,S)), P=pred_true.
ifprintf2(File,FormatString,Value,Trace,Span) :-
   translate:translate_bvalue(Value,TV),
   indent_format(File,FormatString,[TV],Trace),
   fprint_log_info(File,Span).
expects_spaninfo('itprintf').
external_pred_always_true('itprintf').
:- public itprintf/4.
:- assert_must_succeed(external_functions:iprintf(string('test ok = ~w~n'),pred_true,pred_true,unknown)).
:- block itprintf(-,?,?,?), itprintf(?,-,?,?).
itprintf(string(FormatString),Value,PredRes,Span) :- PredRes = pred_true,
   when((ground(FormatString),ground(Value)),
         ifprintf2(user_output,FormatString,Value,trace,Span)).


:- public observe/5.
expects_spaninfo('observe').
expects_unevaluated_args('observe').
expects_waitflag('observe').
external_pred_always_true('observe').
% observe pairs of values
observe(Value,PredRes,UnEvalArgs,Span,WF) :-
  observe2(Value,PredRes,UnEvalArgs,no_trace,Span,WF).

:- public tobserve/5.
expects_spaninfo('tobserve'). % observe but with tracing enabled
expects_unevaluated_args('tobserve').
expects_waitflag('tobserve').
external_pred_always_true('tobserve').
tobserve(Value,PredRes,UnEvalArgs,Span,WF) :-
   observe2(Value,PredRes,UnEvalArgs,trace,Span,WF).

observe2(Value,PredRes,[Arg],Trace,Span,WF) :- PredRes=pred_true,
   %kernel_waitflags:portray_waitflags(WF),
   kernel_waitflags:get_wait_flag0(WF,WF0),
   kernel_waitflags:get_wait_flag1(WF,WF1),
   when(nonvar(WF0),(indent_format(user_output,'WAITFLAG 0 set~n',[],no_trace),kernel_waitflags:portray_waitflags(WF))),
   when(nonvar(WF1),(indent_format(user_output,'WAITFLAG 1 set~n',[],no_trace),kernel_waitflags:portray_waitflags(WF))),
   observe_aux(Value,Arg,1,_,Trace,Span),
   iprintf(string("observe full value complete: ~w~n~n"),Value,_,Span).

observe_aux((V1,V2),b(couple(A1,A2),_,_),Nr,R,Trace,Span) :- !,
    observe_aux(V1,A1,Nr,N1,Trace,Span),
    observe_aux(V2,A2,N1,R,Trace,Span).
observe_aux(V,Arg,Nr,R,Trace,Span) :- R is Nr+1, 
   translate:translate_bexpression(Arg,ArgString), %print(obs(Nr,ArgString)),nl,
    atom_codes(ArgString,ArgCodes),
    observe_value(V,ArgCodes,Nr,Trace,Span).

observe_value(Value,ArgCodes,Nr,Trace,Span) :-  %print(obs(Nr,ArgString)),nl,
   number_codes(Nr,NrCodes), append(NrCodes,") = ~w~n",Tail1), append(" (",Tail1,Tail2),
   append(ArgCodes,Tail2,FormatString),
   observe_set(Value,ArgCodes,1,Trace,Span),
   when((ground(FormatString),ground(Value)),ifprintf2(user_output,FormatString,Value,Trace,Span)).

:- block observe_set(-,?,?,?,?).
observe_set([],_ArgCodes,_,_Trace,_Span) :- !.
observe_set([Value|T],ArgCodes,Pos,Trace,Span) :- !, %print(obs(Value,Pos)),nl,
   append(ArgCodes," -> ",ArgN),
   observe_value(Value,ArgN,Pos,Trace,Span),
   %tools:put_codes(ArgN,user_output),print(Pos),print(' = '),tools_printing:print_arg(Value),nl,
   P1 is Pos+1, observe_set(T,ArgCodes,P1,Trace,Span).
observe_set(_,_,_,_,_).

:- use_module(bsyntaxtree,[get_texpr_id/2]).
% utility function for other modules to use observe on parameters
observe_parameters(Parameters,LocalState) :-
   skel(LocalState,SetupDone),
   reset_ident,
   when(nonvar(SetupDone),observe_parameters_aux(Parameters,1,LocalState)).
:- block skel(-,?).
skel([],pred_true).
skel([_|T],Done) :- skel(T,Done).
:- block observe_parameters_aux(-,?,?), observe_parameters_aux(?,?,-).
observe_parameters_aux([],_,_).
observe_parameters_aux([TID|T],Nr,LS) :-
   get_texpr_id(TID,ID), atom_codes(ID,PCodes),
   store:lookup_value(ID,LS,Value),
   print(observing(ID,Value,LS)),nl,
   observe_value(Value,PCodes,Nr,no_trace,unknown),
   N1 is Nr+1,
   observe_parameters_aux(T,N1,LS).

:- public tprintf/4.
expects_spaninfo('tprintf').
external_pred_always_true('tprintf').
% a debugging version that switches into trace mode
:- block tprintf(-,?,?,?), tprintf(?,-,?,?).
tprintf(string(FormatString),Value,PredRes,Span) :- PredRes = pred_true,
   when((ground(FormatString),ground(Value)),
         (fprintf2(user_output,FormatString,Value,Span),trace)).

:- public 'prolog_printf'/5.
expects_spaninfo('prolog_printf').
expects_waitflag('prolog_printf').
external_pred_always_true('prolog_printf').
prolog_printf(string(FormatString),Value,PredRes,Span,WF) :- PredRes = pred_true,
   file_format(user_output,FormatString,[initial(Value,WF)],Span),
   when(nonvar(Value),file_format(user_output,FormatString,[nonvar(Value,WF)],Span)),
   when((ground(FormatString),ground(Value)),
         fprintf2(user_output,FormatString,Value,Span)).

% vprintf: expression which returns argument as value
expects_spaninfo('vprintf').
:- public vprintf/4.
vprintf(Format,Value,ValueRes,Span) :-
    printf(Format,Value,pred_true,Span),
    equal_object(Value,ValueRes,vprintf). % we could generate a version which only instantiates when printing done
expects_spaninfo('fvprintf').
:- public fvprintf/5.
fvprintf(File,Format,Value,ValueRes,Span) :-
    fprintf(File,Format,Value,pred_true,Span),
    equal_object(Value,ValueRes,fvprintf). % we could generate a version which only instantiates when printing done


% PRINT a single value as substitution
performs_io('PRINT').
does_not_modify_state('PRINT').
:- public 'PRINT'/3.
:- block 'PRINT'(-,?,?), 'PRINT'(?,-,?).
'PRINT'(Val,_Env,OutEnv) :- %print('PRINT : '), print(Env),nl,
   assert_side_effect_occurred(user_output),
   print_value(Val),nl, OutEnv=[].

% PRINT a single value as substitution; show backtracking
perform_directly('PRINT_BT').
does_not_modify_state('PRINT_BT').
:- public 'PRINT_BT'/3.
:- block 'PRINT_BT'(-,?,?), 'PRINT_BT'(?,-,?).
'PRINT_BT'(Val,_Env,OutEnv) :- %print('PRINT : '), print(Env),nl,
   assert_side_effect_occurred(user_output),
   print_value(Val),nl, OutEnv=[].
'PRINT_BT'(Val,_Env,_OutEnv) :-
   assert_side_effect_occurred(user_output),
   print('BACKTRACKING: '),print_value(Val),nl, fail.

print_value(Val) :- translate:translate_bvalue(Val,TV), print(TV).


performs_io('DEBUG_PRINT_STATE').
does_not_modify_state('DEBUG_PRINT_STATE').
:- public 'DEBUG_PRINT_STATE'/3.
:- block 'DEBUG_PRINT_STATE'(-,?,?).
'DEBUG_PRINT_STATE'(Value,Env,OutEnv) :-
    assert_side_effect_occurred(user_output),
    print_value(Value),translate:print_bstate(Env),nl, OutEnv=[].

 % PRINTF as substitution
expects_spaninfo('PRINTF').
performs_io('PRINTF').
does_not_modify_state('PRINTF').
:- public 'PRINTF'/5.
:- block 'PRINTF'(-,?,?,?,?), 'PRINTF'(?,-,?,?,?), 'PRINTF'(?,?,-,?,?).
'PRINTF'(string(FormatString),Value,_Env,OutEnvModifications,Span) :-
  % print('PRINTF'(FormatString,Value,_Env)),nl,
   when((ground(FormatString),ground(Value)),
         (fprintf2(user_output,FormatString,Value,Span),
          OutEnvModifications=[] /* substitution finished */
         )).
expects_spaninfo('FPRINTF').
performs_io('FPRINTF').
does_not_modify_state('FPRINTF').
:- public 'FPRINTF'/6.
:- block 'FPRINTF'(-,?,?,?,?,?), 'FPRINTF'(?,-,?,?,?,?), 'FPRINTF'(?,?,-,?,?,?).
'FPRINTF'(string(File),string(FormatString),Value,_Env,OutEnvModifications,Span) :-
   when((ground(FormatString),ground(Value)),
         (fprintf2(File,FormatString,Value,Span),
          OutEnvModifications=[] /* substitution finished */
         )).

file_format(user_output,FormatString,Args,Span) :- !,
    assert_side_effect_occurred(user_output),
    safe_call(myformat(user_output,FormatString,Args),Span).
file_format(File,FormatString,Args,Span) :-
    open(File,append,Stream),
    assert_side_effect_occurred(file),
    safe_call(myformat(Stream,FormatString,Args),Span),
    close(Stream). % TO DO: call_cleanup

:- public myformat/3.
myformat(Stream,FormatString,Args) :-
   on_exception(error(consistency_error(_,_,format_arguments),_),
               format(Stream,FormatString,Args),
               % backup: the user probably has not provided ~w annotations:
               (format(Stream,FormatString,[]), format_args(Stream,Args))),
   flush_output(Stream).

format_args(_Stream,[]).
format_args(Stream,[H|T]) :- format(Stream,'~w~n',H), format_args(Stream,T).

% a format with indentation and backtracking info:
:- dynamic indentation_level/1.
indentation_level(0).
inc_indent :- retract(indentation_level(X)), X1 is X+1, assert(indentation_level(X1)).
dec_indent :- retract(indentation_level(X)), X1 is X-1, assert(indentation_level(X1)).
reset_ident :- retractall(indentation_level(_)), assert(indentation_level(0)).
indent(Stream) :- indentation_level(L), %(L>28 -> trace ; true), 
    indent_aux(L,Stream).
indent_aux(X,Stream) :- X < 1, !, write(Stream,' ').
indent_aux(X,Stream) :- write(Stream,'+-'), X1 is X-1, indent_aux(X1,Stream).

indent_format(Stream,Format,Args,Trace) :- inc_indent,
  indent(Stream), format(Stream,Format,Args), flush_output(Stream),
  (Trace=trace -> trace ; true).
indent_format(Stream,Format,Args,Trace) :-
  indent(Stream), write(Stream,'BACKTRACKING '),
                  format(Stream,Format,Args),  flush_output(Stream),
  dec_indent,
  (Trace=trace -> trace ; true),
  fail.

:- public 'TRACE'/2.
:- block 'TRACE'(-,?).
'TRACE'(pred_true,R) :- !,trace,R=pred_true.
'TRACE'(pred_false,pred_true).

% --------------------------------------------

% ARGV, ARGC

:- dynamic argv_string/2, argc_number/1.
argc_number(0).
reset_argv :-
   retractall(argv_string(_,_)),
   retractall(argc_number(_)),
   assert(argc_number(0)).

% call to set the commandline arguments, e.g., from probcli
set_argv_from_atom(X) :-
   retractall(argv_string(_,_)),
   retractall(argc_number(_)),
   split_atom_string(X,' ',List),
   set_argv_from_list(List).
set_argv_from_list(List) :-
   retractall(argv_string(_,_)),
   retractall(argc_number(_)),
   set_argv_from_list_aux(List).
set_argv_from_list_aux(List) :-
   length(List,Len),
   assert(argc_number(Len)),
   debug:debug_println(9,argc(Len)),
   nth1(N,List,String), % we number the arguments in B style
   debug:debug_println(9,argv(N,String)),
   assert(argv_string(N,String)),
   fail.
set_argv_from_list_aux(_).

:- public 'ARGV'/2.
:- block 'ARGV'(-,?).
'ARGV'(int(N),S) :- argv2(N,S).
:- block argv2(-,?).
argv2(Nr,Res) :- % print(get_argv(Nr,Res)),nl,
  (argv_string(Nr,S) -> Res=string(S)
   ; add_error(external_functions,'Command-line argument does not exist: ',Nr), Res=string('')).

:- public 'ARGC'/1.
'ARGC'(int(X)) :- argc_number(X).

% --------------------------------------------

% RANDOM

:- assert_must_succeed(( external_functions:call_external_function('RANDOM',[int(0),int(1)],int(0),unknown,_WF) )).
:- use_module(library(random)).
% Generate a random number >= Low and < Up
% warning: this is not a mathematical function
:- public 'RANDOM'/3.
is_not_declarative('RANDOM').
:- block 'RANDOM'(-,?,?), 'RANDOM'(?,-,?).
'RANDOM'(int(Low),int(Up),int(R)) :- random2(Low,Up,R).
:- block random2(-,?,?), random2(?,-,?).
random2(L,U,R) :- random(L, U, R).

% pick a random element from a set
:- public 'random_element'/3.
is_not_declarative('random_element').
expects_waitflag('random_element').
expects_spaninfo('random_element').
:- block 'random_element'(-,?,?,?).
random_element([],_,Span,WF) :- !,
   add_wd_error_span('random_element applied to empty set: ',[],Span,WF).
random_element([H|T],El,_,_) :- !,
   block_length(T,1,Len),
   random_select_from_list([H|T],Len,El).
random_element(CS,Res,_,_) :- is_interval_closure_or_integerset(CS,Low,Up),!,
   U1 is Up+1,
   random(Low,U1,H), Res = int(H).
random_element(CS,X,Span,WF) :-
   custom_explicit_sets:expand_custom_set(CS,ER),
   random_element(ER,X,Span,WF).

:- block block_length(-,?,?).
block_length([],Acc,Acc).
block_length([_|T],Acc,R) :- A1 is Acc+1, block_length(T,A1,R).

:- block random_select_from_list(?,-,?).
random_select_from_list(List,Len,Res) :-
   random(0,Len,RandNr),
   nth0(RandNr,List,El),
   kernel_objects:equal_object(El,Res).

:- use_module(library(system)).

:- assert_must_succeed(( external_functions:call_external_function('TIME',[string(year)],int(Y),unknown,_WF), number(Y), Y>2011 )).
:- assert_must_succeed(( external_functions:call_external_function('TIME',[string(month)],int(M),unknown,_WF), number(M), M>0, M<13 )).
:- public 'TIME'/2.
is_not_declarative('TIME'). % as time can change with next call
'TIME'(string(S),int(N)) :- time2(S,N).
:- block time2(-,?).
time2(now,N) :- !, now(N). % provide UNIX timesstamp
time2(year,N) :- !,datime(datime(N,_,_,_,_,_)).
time2(month,N) :- !,datime(datime(_,N,_,_,_,_)).
time2(day,N) :- !,datime(datime(_,_,N,_,_,_)).
time2(hour,N) :- !,datime(datime(_,_,_,N,_,_)).
time2(min,N) :- !,datime(datime(_,_,_,_,N,_)).
time2(sec,N) :- !,datime(datime(_,_,_,_,_,N)).
time2(OTHER,_) :- add_error(time2,'Illegal time field (must be now,year,month,day,hour,min,sec): ',OTHER),fail.

:- assert_must_succeed(( external_functions:call_external_function('GET_INFO',[string(time)],string(M),unknown,_WF), atom(M) )).
:- public 'GET_INFO'/2.
'GET_INFO'(string(S),Res) :- get_info2(S,Res).
:- use_module(library(codesio)).
:- block get_info2(-,?).
get_info2(time,Res) :- !, datime(datime(Year,Month,Day,Hour,Min,Sec)),
    (Min<10
      -> format_to_codes('~w/~w/~w - ~wh0~w ~ws',[Day,Month,Year,Hour,Min,Sec],Codes)
       ; format_to_codes('~w/~w/~w - ~wh~w ~ws',[Day,Month,Year,Hour,Min,Sec],Codes)
    ),
    atom_codes(INFO, Codes), Res=string(INFO).
get_info2(OTHER,_) :- add_error(time2,'Illegal GET_INFO field (must be time): ',OTHER),fail.

:- public 'TIMESTAMP_INFO'/3.
is_not_declarative('TIMESTAMP_INFO'). % as time can change with next call
:- block 'TIMESTAMP_INFO'(-,-,-).
'TIMESTAMP_INFO'(string(S),int(Ticks),int(Res)) :- time_info2(S,Ticks,Res).
:- block time_info2(-,?,?), time_info2(?,-,?).
time_info2(year,Ticks,N) :- !,datime(Ticks,datime(N,_,_,_,_,_)).
time_info2(month,Ticks,N) :- !,datime(Ticks,datime(_,N,_,_,_,_)).
time_info2(day,Ticks,N) :- !,datime(Ticks,datime(_,_,N,_,_,_)).
time_info2(hour,Ticks,N) :- !,datime(Ticks,datime(_,_,_,N,_,_)).
time_info2(min,Ticks,N) :- !,datime(Ticks,datime(_,_,_,_,N,_)).
time_info2(sec,Ticks,N) :- !,datime(Ticks,datime(_,_,_,_,_,N)).
time_info2(OTHER,_,_) :- add_error(time_info2,'Illegal time field (must be now,year,month,day,hour,min,sec): ',OTHER),fail.

:- public 'TIMESTAMP'/7.
:- block 'TIMESTAMP'(-,-,-,-,-,-,-).
'TIMESTAMP'(int(Y),int(M),int(D),int(H),int(Min),int(S),int(Res)) :-
   timestamp2(Y,M,D,H,Min,S,Res).

:- block timestamp2(-,?,?,?,?,?,-), timestamp2(?,-,?,?,?,?,-), timestamp2(?,?,-,?,?,?,-),
         timestamp2(?,?,?,-,?,?,-), timestamp2(?,?,?,?,?,-,-).
timestamp2(Y,M,D,H,Min,S,ResTimeStamp) :- datime(ResTimeStamp, datime(Y,M,D,H,Min,S)).


:- public 'WALLTIME'/1.
is_not_declarative('WALLTIME'). % as time can change with next call
'WALLTIME'(int(X)) :- statistics(walltime,[X,_]).

:- volatile last_walltime/1.
:- dynamic last_walltime/1.
:- public 'DELTA_WALLTIME'/1.
is_not_declarative('DELTA_WALLTIME'). % as time can change with next call
'DELTA_WALLTIME'(int(D)) :- statistics(walltime,[X,_]),
  (retract(last_walltime(L)) -> D is X-L ; D = 0),
  assert(last_walltime(X)).

:- public 'RUNTIME'/1.
is_not_declarative('RUNTIME'). % as time can change with next call
'RUNTIME'(int(X)) :- statistics(runtime,[X,_]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% check if a given expression has already a ground value when
% the predicate is called

% the predicate is true when the value is ground, false otherwise
:- public 'IS_DETERMINED'/2.
'IS_DETERMINED'(Value,PredRes) :- ground(Value),!,PredRes=pred_true.
'IS_DETERMINED'(_Value,pred_false).

% other variant: the predicate is always true, the information if the variable is written
% to a file to check it later.
expects_spaninfo('IS_DETERMINED_INTO_FILE').
:- public 'IS_DETERMINED_INTO_FILE'/5.
'IS_DETERMINED_INTO_FILE'(FileValue,VariableValue,Value,PredRes,SpanInfo) :-
    check_value_is_string(FileValue,'IS_DETERMINED_INTO_FILE','file name',SpanInfo,File),
    check_value_is_string(VariableValue,'IS_DETERMINED_INTO_FILE','variable name',SpanInfo,Varname),
    ( ground(File) ->
        open(File,write,S),
        call_cleanup( check_determined2(Value,S,Varname), close(S))
    ; otherwise ->
        add_error(external_functions,'### EXTERNAL Unspecified file in check_determined','',SpanInfo)),
    PredRes = pred_true.
check_determined2(Value,S,Varname) :-
    check_determined3(Value,Varname,Term),
    writeq(S,Term),write(S,'.\n').
check_determined3(Value,Name,Term) :- ground(Value),!,Term=determined_value(Name,Value).
check_determined3(_Value,Name,not_determined_value(Name)).

check_value_is_string(string(S),_PredName,_Location,_SpanInfo,String) :-
    !, String=S.
check_value_is_string(Value,PredName,Location,SpanInfo,_String) :-
    atom_concat('### EXTERNAL ',PredName,Msg1),
    atom_concat(Msg1,': Type error, expected string as ',Msg2),
    atom_concat(Msg2,Location,Msg),
    add_error(external_functions,Msg,Value,SpanInfo),
    fail.


:- public 'SEE'/3.
is_not_declarative('SEE').
:- dynamic open_stream/2.
'SEE'(string(File),_S,[]) :- b_absolute_file_name(File,AFile),
   open_stream(AFile,_),!,
   print('File already open: '), print(File),nl.
'SEE'(string(File),_S,[]) :-
   b_absolute_file_name(File,AFile),
   open_absolute_file_for_reading(AFile,_Stream).

open_file_for_reading_if_necessary(File,Stream) :-
   b_absolute_file_name(File,AFile),
   (open_stream(AFile,S) -> Stream=S
     ; open_absolute_file_for_reading(AFile,Stream)).
open_absolute_file_for_reading(AFile,Stream) :-
   open(AFile,read,Stream),
   print(opened_file(AFile,Stream)),nl,
   assert_side_effect_occurred(file),
   assert(open_stream(AFile,Stream)).

correct_file_open_mode(read).
correct_file_open_mode(write).
correct_file_open_mode(append).

:- public 'GET_CODE'/2.
performs_io('GET_CODE').
'GET_CODE'(string(File),int(Code)) :-
   b_absolute_file_name(File,AFile),
   open_stream(AFile,S),
   assert_side_effect_occurred(file),
   get_code(S,Code). %,print(got_code(S,Code)),nl.

:- public 'GET_CODE_STDIN'/1.
'GET_CODE_STDIN'(int(Code)) :- get_code(user_input,Code).

:- public 'SEEN'/3.
is_not_declarative('SEEN').
'SEEN'(string(File),_S,[]) :- b_absolute_file_name(File,AFile),
   retract(open_stream(AFile,S)),
   assert_side_effect_occurred(file),
   close(S).

close_all_files :- retract(open_stream(_File,S)),
   close(S),fail.
close_all_files.


:- public 'READ_FILE_AS_STRINGS'/3.
performs_io('READ_FILE_AS_STRINGS').
expects_spaninfo('READ_FILE_AS_STRINGS').
:- block 'READ_FILE_AS_STRINGS'(-,?,?).
'READ_FILE_AS_STRINGS'(string(File),Res,SpanInfo) :-
    b_absolute_file_name(File,AFile),
    safe_call(open(AFile,read,Stream),SpanInfo),!,
    format('% Opened file: ~w~n',[File]),
    % no side-effect: as everything read in one go
    call_cleanup(read_lines(Stream,Res),
                 close(Stream)).
'READ_FILE_AS_STRINGS'(string(File),_Res,_SpanInfo) :-
   format('% Opening file failed: ~w~n',[File]),fail.
read_lines(File,Res) :- read_lines_loop(1,File,Lines),
   kernel_objects:equal_object(Res,Lines).
read_lines_loop(Nr,Stream,Lines) :- read_line(Stream,Line),
   debug:debug_println(9,read_line(Nr,Line)),
   (Line = end_of_file -> Lines = []
    ; atom_codes(Atom,Line), Lines = [(int(Nr),string(Atom))|T],
      N1 is Nr+1,read_lines_loop(N1,Stream,T)).

:- public 'READ_LINE'/3.
expects_spaninfo('READ_LINE').
performs_io('READ_LINE').
'READ_LINE'(string(File),Res,_SpanInfo) :-
   open_file_for_reading_if_necessary(File,Stream),
   read_line(Stream,Line),
   debug:debug_println(9,read_line(Line)),
   assert_side_effect_occurred(file),
   (Line = end_of_file
    -> add_error(external_functions,'Attempting to read past EOF: ',File),
       Res = ""
     ; atom_codes(Atom,Line), Res = string(Atom)).

:- public 'EOF'/2.
'EOF'(string(File),Res) :- eof2(File,Res).
:- block eof2(-,?).
eof2(File,Res) :-
   open_file_for_reading_if_necessary(File,Stream),
   ((peek_code(Stream,C), C = -1) -> Res = pred_true ; Res = pred_false).  % for Linux compatibility
%   (at_end_of_stream(Stream) -> Res = pred_true ; Res=pred_false).

:- public 'EOF_STDIN'/2.
'EOF_STDIN'(_,Res) :- % dummy argument to enable valid external predicate type declarations
   ((peek_code(user_input,C), C = -1) -> Res = pred_true ; Res = pred_false).


:- public 'CURRENT_FILE_POSITION'/3.
'CURRENT_FILE_POSITION'(string(File),string(Field),Res) :-
   open_file_for_reading_if_necessary(File,Stream),
   stream_position(Stream,Pos),
   % POS must be one of byte_count [only for binary streams], line_count, character_count, line_position
   stream_position_data(Field,Pos,R), Res=int(R).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- public 'ADD_ERROR'/5.
expects_spaninfo('ADD_ERROR').
'ADD_ERROR'(string(Msg),Expr,_S,[],SpanInfo) :-
   add_error(b_user_program,Msg,Expr,SpanInfo).


:- public 'ASSERT_TRUE'/5.
expects_spaninfo('ASSERT_TRUE').
expects_waitflag('ASSERT_TRUE').

'ASSERT_TRUE'(PredVal,string(Msg),PredRes,SpanInfo,WF) :-
    assert_true(PredVal,Msg,PredRes,SpanInfo,WF).

:- block assert_true(-,?,?,?,?).
assert_true(pred_true,_Msg,pred_true,_SpanInfo,_WF).
assert_true(pred_false,Msg,Res,SpanInfo,WF) :- print(Msg),nl,
    add_error_ef('ASSERT_TRUE',Msg,'',SpanInfo,WF,Res).


add_error_ef(Source,Msg,Term,Span,WF,Res) :-
    get_enumeration_finished_wait_flag(WF,AWF),
    add_error_ef_aux(AWF,Source,Msg,Term,Span,WF,Res).

:- block add_error_ef_aux(-,?,?,?,?,?,?).
add_error_ef_aux(_AWF,Source,Msg,Term,Span,_WF,Res) :-
    add_error(Source,Msg,Term,Span),
    Res = pred_false.

:- public 'ASSERT_EXPR'/6.
expects_spaninfo('ASSERT_EXPR').
expects_waitflag('ASSERT_EXPR').

'ASSERT_EXPR'(PredVal,string(Msg),ExprVal,Res,SpanInfo,WF) :-
    assert_expr(PredVal,Msg,ExprVal,Res,SpanInfo,WF).

:- block assert_expr(-,?,?,?,?,?).
assert_expr(pred_true,_Msg,ExprVal,Res,_SpanInfo,_WF) :-
    Res = ExprVal.
assert_expr(pred_false,Msg,_ExprVal,Res,SpanInfo,WF) :-
    add_error_ef('ASSERT_EXPR',Msg,'',SpanInfo,WF,Res).

% --------------------------
% Reflection Functions

:- public 'STATE_AS_STRING'/3.
expects_waitflag('STATE_AS_STRING').
'STATE_AS_STRING'(int(X),StateStr,WF) :-
    state_space:get_state_space_stats(NrNodes,_NrTransitions,_ProcessedNodes),
    get_wait_flag(NrNodes,'STATE_AS_STRING',WF,LWF),
    get_state2(X,StateStr,LWF).
:- block get_state2(-,-,-).
get_state2(X,StateStr,_) :-
    (number(X) -> state_space:visited_expression(X,S)
     ; state_space:visited_expression(ID,S), ID \= root, X=ID), % avoid CLPFD errors if constraints set up for X
    translate:translate_bstate(S,Str),
    StateStr = string(Str).

:- public 'STATE_SUCC'/4.
expects_waitflag('STATE_SUCC').
'STATE_SUCC'(int(X),int(Y),Res,WF) :-
    state_space:get_state_space_stats(NrNodes,_NrTransitions,_ProcessedNodes),
    get_wait_flag(NrNodes,'STATE_SUCC',WF,LWF),
    get_succ2(X,Y,Res,LWF).
:- block get_succ2(-,-,?,-).
get_succ2(X,Y,Res,_) :-
    (number(X) -> 
      if(state_space:transition(X,_,Y),Res=pred_true,Res=pred_false)
     ; if((state_space:transition(ID,_,Y), ID \= root, X=ID),Res=pred_true,Res=pred_false) % avoid CLPFD errors on X
    ).


:- public 'STATE_SAT'/4.
expects_waitflag('STATE_SAT').
% check whether formula as string holds in a state
:- block 'STATE_SAT'(?,-,?,?).
'STATE_SAT'(int(X),string(Formula),Res,WF) :- %print(parsing(Formula)),nl, trace,
    parse_pred(Formula,TFormula),
    state_space:get_state_space_stats(NrNodes,_NrTransitions,_ProcessedNodes),
    get_wait_flag(NrNodes,'STATE_SAT',WF,LWF),
    state_sat2(X,TFormula,Res,LWF,WF).
:- block state_sat2(?,-,?,?,?),state_sat2(-,?,?,-,?).
state_sat2(X,TFormula,PredRes,_,WF) :- 
    (number(X) ->  test_formula_in_state(X,TFormula,PredRes,WF)
     ; test_formula_in_state(ID,TFormula,PredRes,WF), ID \= root, X=ID % avoid CLPFD errors on X
    ).

:- use_module(b_global_sets,[add_prob_deferred_set_elements_to_store/3]).
:- use_module(specfile,[expand_const_and_vars_to_full_store/2,state_corresponds_to_initialised_b_machine/2]).
test_formula_in_state(ID,Pred,PredRes,WF) :-
    state_space:visited_expression(ID,State),
    state_corresponds_to_initialised_b_machine(State,FullState),
    add_prob_deferred_set_elements_to_store(FullState,FullState1,visible),
    debug_println(9,check(Pred,FullState1,PredRes)),
    b_interpreter:b_try_check_boolean_expression_wf(Pred,[],FullState1,WF,PredRes).


:- public 'EVAL'/4.
expects_waitflag('EVAL'). % external function with unknown return type
'EVAL'(int(X),string(Expr),Value,WF) :-
     'STATE_EVAL'(int(X),string(Expr),Value,pred_true,WF).
:- public 'STATE_EVAL'/5.
expects_waitflag('STATE_EVAL'). % external predicate
% evaluate expression and check whether value corresponds to provided value
% (needed for type inference of the external predicate)
:- block 'STATE_EVAL'(?,-,?,?,?).
'STATE_EVAL'(int(X),string(Expr),Value,PredRes,WF) :-
    % string must be provided ; we assume it will be ground
    parse_expr(Expr,TExpr),
    state_space:get_state_space_stats(NrNodes,_NrTransitions,_ProcessedNodes),
    get_wait_flag(NrNodes,'STATE_SAT',WF,LWF),
    state_eval2(X,TExpr,Value,LWF,PredRes).
:- block state_eval2(?,-,?,?,?),state_eval2(-,?,?,-,?).
state_eval2(X,TExpr,Value,_,PredRes) :-
    (number(X) ->  eval_formula_in_state(X,TExpr,Value,PredRes)
     ; eval_formula_in_state(ID,TExpr,Value,PredRes), ID \= root, X=ID % avoid CLPFD errors on X
    ).
eval_formula_in_state(ID,TExpr,Value,PredRes) :-
    state_space:visited_expression(ID,State),
    state_corresponds_to_initialised_b_machine(State,FullState),
    add_prob_deferred_set_elements_to_store(FullState,FullState1,visible),
    debug_println(9,eval(TExpr,FullState1,Value)),
    kernel_equality:equality_objects(Value,TExprValue,PredRes), % TO DO: Stronger Checks that types compatible
    b_interpreter:b_compute_expression(TExpr,[],FullState1,TExprValue).
    
:- use_module(library(terms),[term_hash/2]).
:- dynamic parse_pred_cache/3.
parse_pred(PredStr,TPred) :- term_hash(PredStr,Hash), parse_pred(PredStr,Hash,TPred).
parse_pred(PredStr,Hash,TPred) :- parse_pred_cache(Hash,PredStr,TC),!, TPred=TC.
parse_pred(PredStr,Hash,TPred) :-
    bmachine:b_parse_machine_predicate(PredStr,[prob_ids(visible),variables],TC),
    assert(parse_pred_cache(Hash,PredStr,TC)),
    TPred=TC.

:- dynamic parse_expr_cache/3.
parse_expr(ExprStr,TExpr) :- term_hash(ExprStr,Hash), parse_expr(ExprStr,Hash,TExpr).
parse_expr(ExprStr,Hash,TExpr) :- parse_expr_cache(Hash,ExprStr,TC),!, TExpr=TC.
parse_expr(ExprStr,Hash,TExpr) :-
    atom_codes(ExprStr,Codes),  %print(parsing(Expr,Codes)),nl,
    bmachine:b_parse_machine_expression_from_codes_with_prob_ids(Codes,TC),
    assert(parse_expr_cache(Hash,ExprStr,TC)),
    TExpr=TC.

add_int(X,int(X)).    
:- public 'HISTORY'/1.
'HISTORY'(Value) :-
   state_space:history(RH), reverse(RH,[root|Rest]),
   generate_code_seq(Rest,1,Hist),
   kernel_objects:equal_object(Value,Hist).

:- public 'INITIAL_STATE'/3.
expects_waitflag('INITIAL_STATE'). % external function with unknown return type
'INITIAL_STATE'(int(InitialState),PredRes,WF) :- 
    get_wait_flag(2,'INITIAL_STATE',WF,LWF),
    is_init2(InitialState,PredRes,LWF).
:- block is_init2(-,-,?), is_init2(-,?,-).
is_init2(InitialState,PredRes,_) :-
   PredRes == pred_true,!,
   is_initial_id(InitialState).
is_init2(InitialState,PredRes,_) :- number(InitialState),!,
   (is_initial_id(InitialState) -> PredRes=pred_true ; PredRes=pred_false).
is_init2(InitialState,PredRes,_LWF) :-
   PredRes == pred_false,!,
   when(nonvar(InitialState), \+is_initial_id(InitialState)).
is_init2(InitialState,PredRes,LWF) :-
   add_internal_error('Uncovered case: ', is_init2(InitialState,PredRes,LWF)),fail.

is_initial_id(InitialState) :-
  state_space:transition(root,_,State),
  state_space:packed_visited_expression(State,P),
  (P = concrete_constants(_)
    -> state_space:transition(State,_,InitialState)
    ;  InitialState=State).


:- public 'MACHINE_INFO'/4.
expects_waitflag('MACHINE_INFO').
:- block 'MACHINE_INFO'(?,-,?,?).
'MACHINE_INFO'(M,string(I),Value,WF) :- 
    get_wait_flag(2,'INITIAL_STATE',WF,LWF), % TO DO: use number of machines
    print(minfo(I,M,Value)),nl,
    machine_info2(I,M,Value,LWF).

:- block machine_info2(-,?,?,?), machine_info2(?,-,?,-).
machine_info2('TYPE',M,Value,_LWF) :-
   (ground(M) -> M = string(MachineName) ; true),
   b_machine_hierarchy:machine_type(MachineName,Type),
   (M,Value) = (string(MachineName),string(Type)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% store the information that a predicate has actually been used
% We can check this later with check_if_predicates_were_evaluated/1
% below. The CLI offers a way to use the predicate.

:- use_module(tools, [assert_once/1]).
:- dynamic evaluated_predicate/1.
expects_spaninfo('IS_EVALUATED').
:- public 'IS_EVALUATED'/3.
'IS_EVALUATED'(Value,PredRes,SpanInfo) :-
    check_value_is_string(Value,'IS_EVALUATED',name,SpanInfo,Name),
    assert_once( evaluated_predicate(Name) ),
    PredRes = pred_true.

check_if_predicates_were_evaluated(List) :-
    maplist(check_if_predicate_was_evaluated,List).
check_if_predicate_was_evaluated(evaluated(P)) :-
    format('Checking if "~w" has been evaluated.~n',[P]),
    ( evaluated_predicate(P) -> true
    ; otherwise -> add_error(external_functions,'Predicate has not been evaluated', P)).
check_if_predicate_was_evaluated(not_evaluated(P)) :-
    format('Checking if "~w" has not been evaluated.~n',[P]),
    ( evaluated_predicate(P) -> add_error(external_functions,'Predicate has been evaluated', P)
    ; otherwise -> true).

% --------------------------

:- use_module(hit_profiler,[add_hit/2, add_hit_if_not_covered/2, profile_statistics/0, reset_profiler/0,
                        get_profile_statistics/2, profile_functor/1]). %profiler:profile_statistics.

:- block 'GET_PROFILE_STATISTICS'(-,?,?).
is_not_declarative('GET_PROFILE_STATISTICS').
'GET_PROFILE_STATISTICS'(string(PP),_DummyType,Res) :- % DummyType is just there for typing purposes
  if(profile_functor(PP),
      (get_profile_statistics(PP,Hits),
       maplist(translate_hit_to_b,Hits,List)),
      List = []),
  equal_object_optimized(List,Res, 'GET_PROFILE_STATISTICS').
  %equal_object(List,Res,'GET_PROFILE_STATISTICS'), print(res(Res)),nl.

translate_hit_to_b(hit(Arg,Nr),(Arg,int(Nr))).

'GET_PROFILE_POINTS'(Res) :- findall(string(S),profile_functor(S),List),
  equal_object_optimized(List,Res,'PROFILE_POINT').

expects_spaninfo('PROFILE').
:- block 'PROFILE'(-,?,?,?,?), 'PROFILE'(?,-,?,?,?). %, 'PROFILE'(?,-,?,?,?).
'PROFILE'(string(PP),Count,Value,ValueRes,SpanInfo) :-
    add_profile_exf_hit(Value,PP,SpanInfo,Count),
    equal_object(Value,ValueRes,'PROFILE').

:- block add_profile_exf_hit(-,?,?,?), add_profile_exf_hit(?,-,?,?).
add_profile_exf_hit(pred_true,PP,SpanInfo,Count) :- !, add_value_hit(Count,pred_true,PP,SpanInfo).
add_profile_exf_hit(pred_false,PP,SpanInfo,Count) :- !, add_value_hit(Count,pred_false,PP,SpanInfo).
add_profile_exf_hit(int(X),PP,SpanInfo,Count) :- !, add_int_hit(X,PP,SpanInfo,Count).
add_profile_exf_hit(string(X),PP,SpanInfo,Count) :- !, add_string_hit(X,PP,SpanInfo,Count).
add_profile_exf_hit(Value,PP,SpanInfo,Count) :- when(ground(Value), add_complex_hit(Value,PP,SpanInfo,Count)).

add_complex_hit(Value,PP,SpanInfo,Count) :- print(PP), print(' : '), translate:print_bvalue(Value),nl,
   add_value_hit(Count,Value,PP,SpanInfo).

:- block add_int_hit(-,?,?,?).
add_int_hit(X,PP,SpanInfo,Count) :- add_value_hit(Count,int(X),PP,SpanInfo).
:- block add_string_hit(-,?,?,?).
add_string_hit(X,PP,SpanInfo,Count) :- add_value_hit(Count,string(X),PP,SpanInfo).

add_value_hit(pred_true,Val,PP,_) :- add_hit(PP,Val).
add_value_hit(pred_false,Val,PP,_) :- add_hit_if_not_covered(PP,Val).
