% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 
:- module(gtree, 
[
    visit_tree/5,
    visit_tree/4,
    visit_tree/3,
    visit_tree/2,
    walk_tree/5,
    walk_tree/4,
    walk_tree/3,
    walk_tree/2
]
).


%%
% Hierarchical Visitor Pattern for "trees"
%
% For this implementation a tree is a (root) node and a functor, which
% maps a node to it's child nodes. For example a Prolog term can be
% understand as a tree by the functor:
%   children_fkt(Term, Subterms):- Term=..[_|Subterms].
%
% Important: Do not use trees with cycles. This implementation will not
% detect them and will not terminate.
%
% The indented calling is:
%   visit_tree(+VisitorFkt, +ChildrenFkt, +Node, +EnvIn, -EnvOut)
%
% If no functor is specified, the former mentioned one is used.
% EnvIn and EnvOut are Variables for intermediate values or states,
% which you might need to maintain between calls to VisitorFkt. You 
% might omit them, if you do not need them.
%
% During the visiting process the VisitorFkt functor is called for each
% node starting with the "root". Calls are made in following order:
%   - VisitorFkt(Node, EnvIn, EnvOut), if Node is a leaf, 
%     i.e. ChildrenFkt(Node, []).
%   - VisitorFkt(enter, Node, Children, EnvIn, EnvOut), before branching,
%   - VisitorFkt(leave, Node, Children, EnvIn, EnvOut), after branching 
%     to Node's child nodes.
%  !- Only if VisitorFkt(enter, ...) succeeds, child nodes will be 
%     visited and VisitorFkt(leave, ...) will be called.
%

:- use_module(library(lists), [scanlist/4, maplist/2, group/4]).

:- meta_predicate read_option(2,-,-).
:- meta_predicate visit_tree(3,2,-,-,-).

visit_tree(VisitorFkt, ChildrenFkt, Node, EnvIn, EnvOut):-
    call(ChildrenFkt, Node, Children),
    ( Children == [] ->
        call(VisitorFkt, Node, EnvIn, EnvOut)
        ;
        ( call(VisitorFkt, enter, Node, Children, EnvIn, Env1) ->
            scanlist(visit_tree(VisitorFkt, ChildrenFkt), Children, Env1, Env2),
            call(VisitorFkt, leave, Node, Children, Env2, EnvOut)
            ;
            EnvOut = EnvIn
        )
    ),
    !.

%%
% Shortcuts
%

visit_tree(VisitorFkt, Node, EnvIn, EnvOut):-
    visit_tree(VisitorFkt, x_default_children_fkt, Node, EnvIn, EnvOut).

visit_tree(VisitorFkt, ChildrenFkt, Node):-
    visit_tree(VisitorFkt, ChildrenFkt, Node, _In, _Out).

visit_tree(VisitorFkt, Node):-
    visit_tree(VisitorFkt, Node, _In, _Out).


%%
% Default ChildrenFkt mentioned before
%
:- public x_default_children_fkt/2.
x_default_children_fkt(Node, Children):-
    Node =.. [_|Children].



%%
% Walking through for "trees"
%
% The definition of Tree is same as before.
% Remarks on visit_tree about calling apply also to walk_tree.
% 
:- meta_predicate walk_tree(5,2,-,-,-).
:- meta_predicate walk_tree_impl(5,2,-,-,-).
:- use_module(tools,[safe_on_exception/3]).
walk_tree(WalkerFkt, ChildrenFkt, Node, EnvIn, EnvOut):-
	%% TODO: print message api
    print('Walk through start...'),
    nl,
    print('Enter "help" for descriptions'),
    nl, 
    
    safe_on_exception(
        x_walk_tree_abort_exception,
        (walk_tree_impl(WalkerFkt, ChildrenFkt, Node, EnvIn, EnvOut)),
        (true)
    ),
    
    %% TODO: print message api
    print('Walk through end'),
    nl.


%%
% Shortcuts
%

walk_tree(WalkerFkt, Node, EnvIn, EnvOut):-
    walk_tree(WalkerFkt, x_default_children_fkt, Node, EnvIn, EnvOut).
    
walk_tree(WalkerFkt, ChildrenFkt, Node):-
    walk_tree(WalkerFkt, ChildrenFkt, Node, _, _).
    
walk_tree(WalkerFkt, Node):-
    walk_tree(WalkerFkt, Node, _, _).


%%
% the real thing
%

walk_tree_impl(WalkerFkt, ChildrenFkt, Node, EnvIn, EnvOut):-
    call(ChildrenFkt, Node, Children),
    ( call(WalkerFkt, enterMsg, Node, Children, EnvIn, Env1) ->
    	%% TODO: print message api
        read_option(allowed_options, Cmd, _),
        ( memberchk(Cmd, [enter, y, n, r, s]) ->
            ( Cmd == s ->
                Env2 = Env1
                ;
			    safe_on_exception(
			        x_walk_tree_out_exception(Env2),
					(scanlist(walk_tree_impl(WalkerFkt, ChildrenFkt), Children, Env1, Env2)),
					(true)
				)
            ),
            walk_tree_leave_impl(WalkerFkt, ChildrenFkt, Node, Children, Env2, EnvOut)
            ;
            Cmd = x ->
                x_walk_tree_call_walker_fkt(WalkerFkt, Node, Children, EnvIn, _),
                walk_tree_impl(WalkerFkt, ChildrenFkt, Node, EnvIn, EnvOut)
            ;
            x_walk_tree_actions(Cmd, WalkerFkt, EnvIn),
            walk_tree_impl(WalkerFkt, ChildrenFkt, Node, EnvIn, EnvOut)
        )
        ;
        EnvIn = EnvOut
    ),
    !.


walk_tree_leave_impl(WalkerFkt, ChildrenFkt, Node, Children, EnvIn, EnvOut):-
    call(WalkerFkt, leaveMsg, Node, Children, EnvIn, Env3),
    %% TODO: print message api
    read_option(allowed_options, Cmd, _),
    ( memberchk(Cmd, [enter, y, n, s]) ->
        EnvOut = Env3
        ;
        Cmd = r ->
            walk_tree_impl(WalkerFkt, ChildrenFkt, Node, EnvIn, EnvOut)
        ;
        Cmd = x ->
            x_walk_tree_call_walker_fkt(WalkerFkt, Node, Children, EnvIn, _),
            walk_tree_leave_impl(WalkerFkt, ChildrenFkt, Node, Children, EnvIn, EnvOut)
        ;
        x_walk_tree_actions(Cmd, WalkerFkt, EnvIn),
        walk_tree_leave_impl(WalkerFkt, ChildrenFkt, Node, Children, EnvIn, EnvOut)
    ).


%%
% Helpers
%

x_walk_tree_call_walker_fkt(WalkerFkt, Node, Children, EnvIn, EnvOut):-
	%% TODO: print message api
    print('Please enter a term for the Walker functor:'),
    nl,
    read_term(Term, [consume_layout(true)]),
    ( call(WalkerFkt, Term, Node, Children, EnvIn, EnvOut) ->
        true;
        print('Call fails'),
        nl
    ).

x_walk_tree_actions(xh, WalkerFkt, _):-
	x_print_help(WalkerFkt).
	
x_walk_tree_actions(o,_,EnvIn):-
	throw(x_walk_tree_out_exception(EnvIn)).
	
x_walk_tree_actions(a,_,_):-
	throw(x_walk_tree_abort_exception).


%%
% Keys for options, their description and actions
%    
:- public allowed_options/2.

allowed_options(enter,  'default').
allowed_options(y,      'Yes, like enter').
allowed_options(n,      'Next, like enter').
allowed_options(x,      'Call the Walker functor').
allowed_options(xh,     'List terms accepted by Walker functor').
allowed_options(s,      'Skip, do not enter child nodes').
allowed_options(o,      'Out, return to parent node').
allowed_options(a,      'Abort, same like the abort predicate!!').
allowed_options(r,      'Repeat, reenter a node at leaving!!').
allowed_options(help,   'View this help').


%%
% Helpers
%

%%
% Read an option (a word) from standard input.
% Everything after the first word is keep in Args as string (code-list).
% OptionsFkt functor controls which option is allowed. See 
% allowed_options/2 above.
%

read_option(OptionsFkt, Opt, Args):-
    repeat,
        read_line(L),
        x_strip_white_space_prefix(L, L0),
        ( x_split_option(L0, Opt, Args) ->
            ( 
                Opt == help -> 
                    x_print_help(OptionsFkt),
                    fail
                ;
                call(OptionsFkt, Opt, _) -> 
                    true
                ;
                format('Unknown Option: "~p". Choose help for help.~N', [Opt]),
                fail
            ),
            IsFailed = false
            ;
            IsFailed = true
        ),
    !,
    ( IsFailed == true -> fail; true).
    

%:- use_module(library(lists), [group/4]).

x_split_option([], enter, []).
    
x_split_option(L, Opt, Args):-
    group(x_non_white_space, L, OptStr, ArgsStr),
    name(Opt, OptStr),
    ( ArgsStr == [_|Args],!; Args = ArgsStr).


x_strip_white_space_prefix(String, StrippedString):-
    group(x_white_space, String, _, StrippedString).


x_non_white_space(X):- 
    \+char_code(' ', X).

x_white_space(X):-
    char_code(' ', X).
    
x_print_help(Fkt):-
	safe_on_exception(
		Error,
		(
		    findall([Opt, HelpText], call(Fkt, Opt, HelpText), L),
    		print('Possible options: '),
    		nl,
		    maplist(format('~2|~p~24|~p~N'), L)
	    ),
	    (
	    	Error = error(existence_error(procedure, _:Fkt0/_), _),
	    	functor(Fkt, Fkt0, _),
	    	print('No options or description available.'),
	    	nl
    	)
	).







%%
% For test and debug purpose only
%


:- public x_print_term/1.
x_print_term(T):-
    visit_tree(x_print_visitor, T, x, _).


x_print_visitor(enter, Node, Children, In, s(In)):-
    x_print_indent(In),
    functor(Node, Fkt,_),
    length(Children, Arity),
    print(Fkt/Arity),
    nl.
    
x_print_visitor(leave, _Node, _Children, s(Out), Out).

x_print_visitor(Node, In, In):-
    x_print_visitor(enter, Node, [], In, _).

x_print_indent(s(X)):- !,
    print('  '),
    x_print_indent(X).
x_print_indent(_).

/*
x_walk_term(T, Out):-
    walk_tree(x_print_walker, x_default_children_fkt, T, x, Out).
    
x_print_walker(enterMsg, Node, Children, In, s(In)):-
    x_print_indent(In),
    functor(Node, Fkt,_),
    length(Children, Arity),
    format('>> ~p~N', [Fkt/Arity]).

x_print_walker(leaveMsg, Node, _Children, s(Out), Out):-
    x_print_indent(Out),
    functor(Node, Fkt,_),
    format('<< ~p~N', [Fkt]).

x_print_walker(details, Node, _,_,_):-
    print(Node),
    nl.
    
x_print_walker(details, 'Show Details').
*/