% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(hashing,[my_term_hash/2, env_hash/2, efficient_index/1]).

:- use_module(library(terms),[term_hash/2,term_hash/3]).


:- use_module(module_information,[module_info/2]).

:- module_info(group,animator).
:- module_info(description,'This module computes hash values (mainly for states) for efficient indexing.').

% --------------------------------------------

:- use_module(self_check).

:- assert_must_succeed(( hashing:my_term_hash([1,2,3],X), hashing:efficient_index(X) )).
:- assert_must_succeed(( hashing:my_term_hash([1,2,3],X), hashing:my_term_hash([1,2,4],Y), X\=Y )).

:- if(tools:platform_is_64_bit).
:- format("64-bit platform~n",[]).
my_term_hash(T,Res) :- super_hash(T,Res).
:- else.
:- format("32-bit platform~n",[]).
%my_term_hash(T,Res) :- env_hash(T,Res).
my_term_hash(T,Res) :- term_hash(T,
  [range(smallint),  /* will use full 32 bit on 64 bit systems, 28 bit on 32 bit system */
   algorithm(sdbm), % other options jenkins, hsieh, default
   depth(infinite),if_var(ignore)],Res).
% collisions seem with more recent versions of SICStus to no longer be a problem !?
% sdbm seems to generate less collisions
:- endif.

% check if a hash value is an efficient for first argument indexing
efficient_index(X) :- number(X),
  prolog_flag(max_tagged_integer,Max), X=<Max,
  prolog_flag(min_tagged_integer,Min), X>=Min.

% On 64-bit machines we can use the following for efficient indexing:
:- public super_hash/2. % not used at the moment
super_hash(Term,Hash) :-
  term_hash(Term,[range(smallint),algorithm(sdbm),depth(infinite),if_var(ignore)],H1),
  term_hash(Term,[range(smallint32),algorithm(jenkins),depth(infinite),if_var(ignore)],H2),
  Hash is H1 + (H2 << 32). % should be in range 0.. 2**60
  % we could still try and use negative numbers


%old_my_term_hash(T,Res) :- term_hash(T,H1),  /* The builtin hash function provides many collisions */
%  (env_hash(T,H2) -> combine_hash(H1,H2,Res)  % Res is H1+H2*4
%    ; Res=H1 %(print(env_hash_failed(T,H2)),nl, Res = H1)
%   ).

:- assert_must_succeed(( hashing:env_hash(concrete_constants([bind(x,[int(1),int(2),int(3)])]),X), hashing:efficient_index(X) )).
:- assert_must_succeed(( hashing:env_hash(const_and_vars(1,[bind(x,[(int(1),int(2),pred_true,pred_false,[])]),
                         bind(y,fd(1,'Name')),bind(z,string('Name')),bind(v,global_set('Name')), 
                         bind(w,rec([field(balance,int(1)),field(name,fd(2,'Name'))])), bind(var,_VAR), bind(ab,abort(t))]),X), hashing:efficient_index(X) )).
:- assert_must_succeed(( hashing:env_hash([bind(x,[int(1),int(2),int(3)])],X), hashing:env_hash([bind(x,[int(1),int(2),int(4)])],Y), X\=Y )).

% not really used anymore !:
env_hash(csp_and_b(_CSP,B),R) :- env_hash(B,R).
env_hash(concrete_constants(X),R) :- env_hash(X,R).
env_hash(const_and_vars(CID,X),R) :- env_hash(X,RX), combine_hash(CID,RX,R).
env_hash([],1).
env_hash([bind(_,Expr)|T],Res) :- expr_hash(Expr,EH),!, env_hash(T,ET),
  combine_hash(EH,ET,Res).
/* TO DO: extend for CSP ! */

expr_hash(X,R) :- var(X),!,R=1.
expr_hash(pred_true /* bool_true */,1) :- !.
expr_hash(pred_false /* bool_false */,0) :- !.
expr_hash(int(X),R) :- !, (ground(X) -> R is X+67 ; R=2).  % what about negative numbers?
expr_hash(fd(X,_),R) :- !, (ground(X) -> R=X ; R=3).
expr_hash([],29) :- !.
expr_hash([H|T],R) :- !, list_hash([H|T],R).
expr_hash((A,B),R) :- !,expr_hash(A,EA), expr_hash(B,EB), combine_hash(EA,EB,RAB), R is RAB+7.
expr_hash(closure(_,_,_),2) :- !.
%expr_hash(closure_x(_,_,_,Exp),R) :- !, var(Exp) -> R=2 ; list_hash(Exp,R).
expr_hash(avl_set(X),R) :- !, (ground(X) -> term_hash(X,R) ; R=1).
expr_hash(global_set(_),5) :- !.
expr_hash(freetype(_),7) :- !.
expr_hash(rec(Fields),R) :- !,list_hash(Fields,R).
expr_hash(field(_Name,Value),R) :- !,expr_hash(Value,R).
expr_hash(string(X),R) :- !,(ground(X) -> term_hash(X,R) ; R=1).
expr_hash(abort(X),R) :- !,(ground(X) -> term_hash(X,R) ; R=1).
expr_hash(term(X),R) :- !,(ground(X) -> term_hash(X,R) ; R=1).
expr_hash(X,1) :- print(unknown_expr(X)),nl.


list_hash(X,R) :- var(X),!,R=0.
list_hash([],1).
list_hash([Expr|T],Res) :- expr_hash(Expr,EH),!, list_hash(T,ET),
 % Res is ((ET * 37) + EH + 1) mod 2097152. %1048576.  %978217. 
  Res is (((ET * 1013 %1031 %16777619
               ) /\  4194303) \ EH).  %1073741824) \ EH. %1048576.  %978217. 
  /* should probably be dealt with slightly differently from env_hash; otherwise
      x={a,b},y={}  and x={},y={a,b} will have the same hash value */

combine_hash(EH,ET,Res) :-
   Res is (((ET * 1019 %1031 %16777619
            ) /\  4194303) \ EH).  %mod 1073741824) \ EH. %1048576. %19872353. %978217.
