% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(kernel_z, [compaction/3,bag_items/3]).
% The module for Z specific kernel operations

:- use_module(library(avl)).

:- use_module(debug).
:- use_module(self_check).
:- use_module(tools).
:- use_module(error_manager).
:- use_module(kernel_objects).
:- use_module(kernel_waitflags,[add_wd_error/3,assert_must_abort_wf/2]).
:- use_module(custom_explicit_sets,[try_expand_custom_set/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,proz).
:- module_info(description,'This module provides operations that are special to the Z notation (sequence compaction, items of a bag).').


/* The compaction of a set integer -> X is the sequence where the "gaps are filled up".
   E.g. compaction of {(1,a),(4,b),(8,c)} is {(1,a),(2,b),(3,c)}. */
   
:- assert_must_succeed(exhaustive_kernel_check_wfdet(kernel_z:compaction([(int(1),int(11)), (int(4),int(44))],[(int(1),int(11)), (int(2),int(44))],WF),WF)).
:- assert_must_succeed(kernel_z:compaction([(int(1),int(11)), (int(4),int(44))],[(int(2),int(44)), (int(1),int(11))],_)).
:- assert_must_succeed((kernel_z:compaction(S,R,_),
  S=[(int(5),int(55)),(int(1),int(11)), (int(4),int(44))], kernel_objects:equal_object(R,[(int(3),int(55)),(int(1),int(11)), (int(2),int(44))]))).
:- assert_must_abort_wf(kernel_z:compaction([(int(1),int(11)), (int(4),int(44)), (int(1),int(22))],_,WF),WF).

:- block compaction(-,?,?).
compaction(Sequence,Value,WF) :- 
   try_expand_custom_set(Sequence,ESeq),
   try_expand_custom_set(Value,EValue),
   compaction2(ESeq,EValue,WF).

compaction2(Relation,Sequence,WF) :-
    empty_avl(Empty),
    collect_sorted_values(Relation,Empty,SortedValues,WF),
    create_sequence(SortedValues,Sequence1),
    equal_object(Sequence1,Sequence).

:- block collect_sorted_values(-,?,?,?).
collect_sorted_values([],Avl,Avl,_WF).
collect_sorted_values([(int(Pos),Element)|Rest],In,Out,WF) :-
    add_value(Pos,Element,Rest,In,Out,WF).

:- block add_value(-,?,?,?,?,?).
add_value(Pos,Element,Rest,In,Out,WF) :-
    ( avl_fetch(Pos,In,Old) ->
        not_equal_object_wf(Old,Element,WF),
        add_wd_error('Input for squash is not a function',
                        [(int(Pos),Old),(int(Pos),Element)],WF),
        Out=failure
    ; otherwise ->
        avl_store(Pos,In,Element,Inter),
        collect_sorted_values(Rest,Inter,Out,WF)).

:- block create_sequence(-,?).
create_sequence(failure,_Sequence) :- !.
create_sequence(SortedValues,Sequence) :-
    create_sequence2(SortedValues,1,Sequence1),
    equal_object(Sequence1,Sequence).
create_sequence2(SortedValues,Pos,Sequence) :-
    ( avl_del_min(SortedValues, _Min, Element, Rest) ->
        Sequence = [(int(Pos),Element)|RestSequence],
        NextPos is Pos+1,
        create_sequence2(Rest,NextPos,RestSequence)
    ; otherwise ->
        Sequence = []).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(kernel_z:bag_items([(int(1),int(11)), (int(2),int(22)), (int(3),int(11))],[(int(11),int(2)), (int(22),int(1))],WF),WF)).
:- assert_must_abort_wf(bag_items([(int(1),int(11)), (int(2),int(22)), (int(4),int(11))],_,WF),WF).
:- assert_must_abort_wf(bag_items([(int(1),int(11)), (int(2),int(22)), (int(1),int(11))],_,WF),WF).

% This predicate computes the effect of the items function in Z.
% It is written for basic bag support without performance in mind.
:- block bag_items(-,?,?).
bag_items(Sequence,Bag,WF) :-
    try_expand_custom_set(Sequence,ESequence),
    empty_avl(Empty),
    bag_items2(ESequence,[],Bag,Empty,WF).

:- block bag_items2(-,?,?,?,?).
bag_items2([],InBag,OutBag,Cnt,WF) :-
    ( check_count(Cnt) ->
        equal_object(InBag,OutBag)
    ; otherwise ->
        add_wd_error('Input for bag_times is not a sequence',[],WF)).
bag_items2([(int(Pos),Elem)|Rest],InBag,OutBag,InCnt,WF) :-
    bag_items3(Pos,Elem,InBag,InterBag,InCnt,OutCnt,WF),
    bag_items2(Rest,InterBag,OutBag,OutCnt,WF).
:- block bag_items3(-,?,?,?,?,?,?).
bag_items3(Pos,Elem,InBag,OutBag,InCnt,OutCnt,WF) :-
    ( avl_fetch(Pos,InCnt) ->
        add_wd_error('Input for bag_times is not a function',[],WF)
    ; otherwise ->
        avl_store(Pos,InCnt,true,OutCnt),
        OldEntry = (Elem,int(OldCount)),
        ( exact_element_of(OldEntry, InBag) ->
            remove_element_wf( OldEntry, InBag, InterBag,WF),!,
            NewCount is OldCount+1
        ; otherwise ->
            InterBag = InBag,
            NewCount = 1),
        add_element_wf( (Elem,int(NewCount)), InterBag, OutBag,WF)).

check_count(Cnt) :-
    avl_domain(Cnt,Domain),
    check_count2(Domain,1).
check_count2([],_).
check_count2([H|T],H) :- I is H+1,check_count2(T,I).
