:- module(kodkod_test, [test_kodkod/1,compare_kodkod_performance/2]).

:- use_module(library(lists)).

:- use_module(probsrc(tools),[start_ms_timer/1,stop_ms_timer/2,average/2,ajoin/2]).
:- use_module(probsrc(module_information),[module_info/2]).
:- use_module(probsrc(error_manager),[add_error/2]).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(bmachine)).
:- use_module(probsrc(b_interpreter)).
:- use_module(probsrc(store)).
:- use_module(probsrc(b_enumerate)).
:- use_module(probsrc(kernel_waitflags)).
:- use_module(probsrc(error_manager)).
:- use_module(probsrc(junit_tests)).
:- use_module(probsrc(translate)).
:- use_module(probsrc(succeed_max)).
:- use_module(probsrc(preferences),[get_preference/2]).
:- use_module(probsrc(self_check),[time_out_call/2]).

:- use_module(kodkod).
:- use_module(kodkod_process).

:- module_info(group,kodkod).
:- module_info(description,'This is a test engine for the Kodkod translation functionality.').

compare_kodkod_performance(Outfile,Iterations) :-
    kodkod_performance(Iterations,Results),
    open(Outfile,write,S),
    maplist(print_result(S),Results),
    close(S).

print_result(S,translation_failed(Pred)) :-
    print_intro(S,Pred),
    write(S,'Status: translation failed\n').
print_result(S,comparision(Pred,_Ks,_Ps,Kodkod,ProB)) :-
    print_intro(S,Pred),
    write(S,'Status: translation succeeded\n'),
    print_result2(S,Kodkod,KN),
    print_result2(S,ProB,PN),
    ( KN=PN     -> true
    ; otherwise -> write(S,'Error: different number of results\n')).
print_intro(S,Pred) :-
    write(S,'___ Comparing for predicate ___\n'),
    conjunction_to_list(Pred,Preds),
    maplist(print_intro2(S),Preds).
print_intro2(S,Pred) :-
    translate_bexpression(Pred,TPred),
    format(S,'Pred: ~w~n',[TPred]).
print_result2(S,kodkod_timeout,_) :-
    print_time(S,'Kodkod total',timeout),
    print_time(S,'Kodkod preparation',timeout),
    print_time(S,'Kodkod computation',timeout),
    print_time(S,'Kodkod Java',timeout).
print_result2(S,kodkod_solved(PrepTime,CompTime,JavaTime,N),N) :-
    memberchk(walltime/PT,PrepTime),
    memberchk(walltime/CT,CompTime),
    TT is PT+CT,
    print_time(S,'Kodkod total',TT),
    print_time(S,'Kodkod preparation',PT),
    print_time(S,'Kodkod computation',CT),
    print_time(S,'Kodkod Java',JavaTime).
print_result2(S,prob_timeout,_) :-
    print_time(S,'ProB total',timeout).
print_result2(S,prob_solved(CompTime,N),N) :-
    memberchk(walltime/TT,CompTime),
    print_time(S,'ProB total',TT).
print_time(S,Name,Time) :-
    ( Time = timeout ->
        format(S,'~w time: ~w~n', [Name,timeout])
    ; otherwise ->
        format(S,'~w time: ~0f~n',[Name,Time])).
    


kodkod_performance(Iterations,Results) :-
    b_get_machine_constants(Constants),
    ( Constants = [_|_] ->
        kodkod_performance2(Constants,Iterations,Results)
    ; otherwise ->
        add_error(kodkod_test,'No constants to check measure performance of Kodkod translation'),
        fail).

kodkod_performance2(Constants,Iterations,Results) :-
    b_get_properties_from_machine(Predicate),
    predicate_components(Predicate,Components),
    maplist(kodkod_performance3(Constants,Iterations),Components,Results).

kodkod_performance3(Constants,Iterations,component(Predicate,Ids),Result) :-
    include(memberid(Ids),Constants,TIds),
    kodkod_performance4(Predicate,TIds,Iterations,Result).

kodkod_performance4(Predicate,Constants,Iterations,comparision(Predicate,Ks,Ps,Ksum,Psum)) :-
    kodkod_translation(Predicate,Constants,_KPred),
    !,
    kodkod_performance5(Predicate,Constants,Iterations,Ks,Ps),
    summarize_kodkod(Ks,Ksum),
    summarize_prob(Ps,Psum).
kodkod_performance4(Predicate,_Constants,_Iterations,translation_failed(Predicate)).

kodkod_performance5(Predicate,Constants,Iterations,[K|Krest],[P|Prest]) :-
    Iterations > 0,!,
    call_kodkod(Predicate,Constants,K),
    call_prob(Predicate,Constants,P),
    I2 is Iterations - 1,
    kodkod_performance5(Predicate,Constants,I2,Krest,Prest).
kodkod_performance5(_Predicate,_Constants,_Iterations,[],[]).

summarize_kodkod(L,kodkod_timeout) :-
    memberchk(timeout,L),!.
summarize_kodkod(L,kodkod_solved(PrepTime,CompTime,JavaTime,NrSolutions)) :-
    maplist(get_kodkod_prep_time,L,PTimes),summarize_times(PTimes,PrepTime),
    maplist(get_kodkod_comp_time,L,CTimes),summarize_times(CTimes,CompTime),
    maplist(get_kodkod_java_time,L,JTimes),average(JTimes,JavaTime),
    maplist(get_kodkod_nsol,L,Sols),check_same(Sols,NrSolutions).
summarize_prob(L,prob_timeout) :-
    memberchk(timeout,L),!.
summarize_prob(L,prob_solved(CompTime,NrSolutions)) :-
    maplist(get_prob_comp_time,L,CTimes),summarize_times(CTimes,CompTime),
    maplist(get_prob_nsol,L,Sols),check_same(Sols,NrSolutions).

summarize_times([T|Rest],Times) :-
    findall(S, member(S/_,T), Secs),
    maplist(sum_time([T|Rest]), Secs, Times).
sum_time(Lists,Sec,Sec/Time) :-
    member(L,Lists),member(Sec/timeout,L),!,Time=timeout.
sum_time(Lists,Sec,Sec/Time) :-
    findall(T, (member(L,Lists),memberchk(Sec/T,L)), Times),
    average(Times,Time).

check_same(L,V) :-
    ( is_same(L,V) -> true
    ; otherwise -> add_error(kodkod_test,'Inconsistent number of solutions'),fail).

is_same([E],V) :- !,E=V.
is_same([V|L],V) :-
    is_same(L,V).
    

get_kodkod_prep_time(kodkod_solved(PrepTime,_CompTime,_JavaTime,_N),PrepTime).
get_kodkod_comp_time(kodkod_solved(_PrepTime,CompTime,_JavaTime,_N),CompTime).
get_kodkod_java_time(kodkod_solved(_PrepTime,_CompTime,JavaTime,_N),JavaTime).
get_kodkod_nsol(kodkod_solved(_PrepTime,_CompTime,_JavaTime,N),N).

get_prob_comp_time(prob_solved(CompTime,_N),CompTime).
get_prob_nsol(prob_solved(_CompTime,N),N).


:- meta_predicate measured_call(0,?).
measured_call(Call,TimeR) :-
    start_ms_timer(Timer),
    time_out_call(Call,Timeout=true),
    stop_ms_timer(Timer,Time),
    ( Timeout = false -> TimeR = Time
    ; Timeout = true  -> TimeR = timeout).

call_kodkod(Predicate,Constants,Result) :-
    measured_call(kodkod_translation(Predicate,Constants,KPred),PrepTime),
    !,
    clear_java_comp_time,
    measured_call(find_solutions(Constants,KPred,States),CompTime),
    ( CompTime == timeout ->
        Result = timeout
    ; otherwise ->
        Result = kodkod_solved(PrepTime,CompTime,JavaTime,NrSolutions),
        get_java_comp_time(JavaTime),
        length(States,NrSolutions)).
call_prob(Predicate,Constants,Result) :-
    measured_call(find_solutions(Constants,Predicate,States),CompTime),
    ( CompTime == timeout ->
        Result = timeout
    ; otherwise ->
        Result = prob_solved(CompTime,NrSolutions),
        length(States,NrSolutions)).

kodkod_translation(Predicate,Constants,KPred) :-
    kodkod:try_predicate_analysis_with_global_sets(Predicate,Constants,TaggedConstants,TaggedPredicate),
    kodkod:apply_transformation2(TaggedConstants,TaggedPredicate,KPred).


memberid(Ids,TId) :-
    get_texpr_id(TId,Id),memberchk(Id,Ids).

test_kodkod(MaxResiduePreds) :-
    catch( (test_kodkod2(MaxResiduePreds),store_testcase(pass)),
           junit(Msg),
           store_testcase(error([Msg]))).
test_kodkod2(MaxResiduePreds) :-
    b_get_machine_constants(Constants),
    b_get_properties_from_machine(Predicate),
    check( Constants=[_|_], 'Test failed, no constants present in machine.'),
    check( find_prob_solutions(Predicate,Constants,ProBStates),
           'finding ProB solutions failed'),
    check( find_kodkod_solutions(Predicate,Constants,KodkodStates,Residue),
           'finding Kodkod solutions failed'),
    ajoin(['too many remaining predicates: ', Residue, ' > ' , MaxResiduePreds],TooManyErrMessage),
    check( Residue =< MaxResiduePreds,TooManyErrMessage),
    check( ProBStates = KodkodStates,
           'different solutions found').

find_prob_solutions(Predicate,Constants,States) :-
    find_solutions(Constants,Predicate,States).

find_kodkod_solutions(Predicate,Constants,States,Residue) :-
    %predicate_size(Predicate,Before),
    replace_by_kodkod(Constants,Predicate,KPredicate),
    check( kodkod_present(KPredicate),
           'no kodkod translation applied'),
    predicate_size(KPredicate,After), Residue is After-1,
    find_solutions(Constants,KPredicate,States).

predicate_size(Predicate,Size) :-
    conjunction_to_list(Predicate,Preds),
    length(Preds,Size).

kodkod_present(Predicate) :-
    conjunction_to_list(Predicate,Preds),
    create_texpr(kodkod(_,_),pred,_,KPred),
    memberchk(KPred,Preds).

find_solutions(Constants,Predicate,States) :-
    get_preference(maxNrOfInitialisations,Max),
    findall(State,
            succeed_max_call(find_solution(Constants,Predicate,State),Max),
            UnsortedStates),
    sort(UnsortedStates,States).

find_solution(Constants,Predicate,NormalizedState) :-
    empty_state(Empty),
    set_up_typed_localstate(Constants,_Values,TypedValues,Empty,State,positive),
    normalise_store(State,NormalizedState),
    init_wait_flags(WF),
    b_enumerate_values(TypedValues,WF),
    b_test_boolean_expression(Predicate,Empty,State,WF),
    ground_wait_flags(WF).

check(Check,Msg) :-
    ( call(Check)   -> true
    ; in_junit_mode -> throw(junit(Msg))
    ; otherwise     -> add_error(kodkod_test,Msg), fail
    ).

store_testcase(Verdict) :-
    in_junit_mode,!,
    b_machine_name(Name),
    create_junit_result(Name,kodkod,comparision_with_prob,0,Verdict,Junit),
    print_junit([Junit],kodkod).
store_testcase(_Verdict).

in_junit_mode :- junit_mode(_),!.
