% (c) 2009 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(kodkod_typing,[kodkod_insert_casts/4]).

:- use_module(library(lists)).
:- use_module(library(avl)).

:- use_module(probsrc(module_information),[module_info/2]).
:- use_module(probsrc(tools),[foldl/4]).

:- use_module(kodkod_tools).

:- module_info(group,kodkod).
:- module_info(description,'This is a small internal type checker for Kodkod expressions. It\'s needed for inserting casts between different integer representations.').

kodkod_insert_casts(Kin,TypeMap,IdMap,Kout) :-
    create_refmap(TypeMap,IdMap,Env),
    type_kodkod(Kin,Env,formula,Kout).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% create Kodkod typechecking environment
create_refmap(TypeMap,IdMap,ktypenv(Env,StartEnv)) :-
    empty_avl(StartEnv),
    create_refmap_types(TypeMap,StartEnv,TypeEnv),
    create_refmap_ids(IdMap,TypeEnv,Env).
lookup_reltypes(Id,ktypenv(Env,_Vars),Types) :-
    avl_fetch(Id,Env,Types).
lookup_vartypes(Id,ktypenv(_Env,Vars),Types) :-
    avl_fetch(Id,Vars,Types).
add_var_to_type_env(Id,Types,ktypenv(E,EnvIn),ktypenv(E,EnvOut)) :-
    avl_store(Id,EnvIn,Types,EnvOut).

create_refmap_types(TypeMap,In,Out) :-
    avl_range(TypeMap,Relations1),
    maplist(remove_relation_wrapper,Relations1,Relations),
    foldl(add_to_refmap,Relations,In,Out).
remove_relation_wrapper(relation(_,_,Rel),Rel).

create_refmap_ids(IdMap,In,Out) :-
    avl_range(IdMap,Relations),
    foldl(add_to_refmap,Relations,In,Out).    
add_to_refmap(R,In,Inter) :-
    get_rel_types(R,Id,Types),
    avl_store(Id,In,Types,Inter).
get_rel_types(reldef(Id,Types,_,_,_),Id,Types).
get_rel_types(intsetrel(Id,_,_),Id,[R]) :-
    intset_relation_kodkod_name(R).
get_rel_types(pow2rel(Id,_,_),Id,[R]) :-
    pow2integer_relation_kodkod_name(R).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% typechecking kodkod expressions with inserting type-casts

type_kodkod(KExpr1, Env, OuterType, OuterExpr) :-
    ( KExpr1 = KExpr:ExpectedType ->
        % The expression has an explicit type associated,
        % first make sure that this expected type matches the type seen
        % by the caller (insert cast if necessary)
        unify_kodkod_types(ExpectedType,OuterType,Expr,OuterExpr)
    ; otherwise ->
        % If there is no explicit type given, the type seen by the
        % outside is used as the expected type. No cast is needed
        KExpr=KExpr1, ExpectedType=OuterType, OuterExpr=Expr
    ),
    % Now type-check the expression itself
    type_kodkod2(KExpr, Env, InnerType, InnerExpr),
    % and try to match the types (insert cat if necessary)
    unify_kodkod_types(ExpectedType,InnerType,InnerExpr,Expr).

:- block unify_kodkod_types(-,-,?,?).
unify_kodkod_types(rel(A),rel(B),Expr,NewExpr) :- !,
    ( A=B -> Expr=NewExpr
        % both relations can be unified, good
    ; A=[IS],intset_relation_kodkod_name(IS),
        B=[P2I],pow2integer_relation_kodkod_name(P2I) ->
        % expected was a set of integers but given was a set representing
        % one integer by its bits
        % TODO: this conversion is not optimal, better try to avoid rel([p2i])
        NewExpr = int2intset(expr2int(Expr))).
unify_kodkod_types(int_expression,int_expression,Expr,Expr) :- !.
unify_kodkod_types(formula,formula,Expr,Expr) :- !.
unify_kodkod_types(rel([R]),int_expression,Expr,int2intset(Expr)) :-
    intset_relation_kodkod_name(R),!.
unify_kodkod_types(rel([R]),int_expression,Expr,int2pow2(Expr)) :-
    pow2integer_relation_kodkod_name(R),!.
unify_kodkod_types(int_expression,rel([_]),Expr,expr2int(Expr)).
%    intset_relation_kodkod_name(R).

type_kodkod2(relref(Id), Env, rel(Types), relref(Id)) :- !,
    lookup_reltypes(Id,Env,Types).
type_kodkod2(varref(Id), Env, VType, K) :- !,
    lookup_vartypes(Id,Env,Types),
    ( Types = [p2i] ->
        VType = int_expression, K=expr2int(varref(Id))
    ; otherwise ->
        VType = rel(Types), K=varref(Id)).
type_kodkod2(int(I), _Env, int_expression, int(I)) :- !.
type_kodkod2(prj(Pos,Arg), Env, rel(P), prj(Pos,NArg)) :-
    !,type_kodkod(Arg,Env,rel(S),NArg),
    nth0s(Pos,S,P).
type_kodkod2(forall(Decls,Formula), Env, formula, forall(NDecls,NFormula)) :-
    !,type_quantifier(Decls,Formula,Env,_,NDecls,NFormula).
type_kodkod2(exists(Decls,Formula), Env, formula, exists(NDecls,NFormula)) :-
    !,type_quantifier(Decls,Formula,Env,_,NDecls,NFormula).
type_kodkod2(comp(Decls,Expr), Env, rel(Types), comp(NDecls,NExpr)) :-
    !,type_quantifier(Decls,Expr,Env,rel(Types),NDecls,NExpr).
type_kodkod2(Expr, Env, Type, NExpr) :-
    functor(Expr,F,Arity),
    functor(Lookup,F,Arity),
    ktype(Lookup,Type),!,
    Expr =.. [_|Args], Lookup =.. [_|Types],
    maplist(type_kodkod3(Env),Args,Types,NArgs),
    NExpr =.. [F|NArgs].
type_kodkod2(Expr, _Env, _Type, _NExpr) :-
    throw( unsupported_kodkod_expression(Expr) ).

% just to swap arguments
type_kodkod3(Env,Expr,Type,NExpr) :-
    type_kodkod(Expr,Env,Type,NExpr).

type_quantifier(Decls,Subexpr,Env,rel(Types),NDecls,NSubexpr) :-
    type_kodkod_declarations(Decls,Env,Env,NDecls,Subenv,Types),
    type_kodkod(Subexpr,Subenv,formula,NSubexpr).

type_kodkod_declarations([],_,Env,[],Env,[]).
type_kodkod_declarations([Decl|Drest],Env,EnvIn,[NDecl|Nrest],EnvOut,Types) :-
    type_kodkod_declaration(Decl,Env,EnvIn,NDecl,EnvInter,Types1),
    safe_append(Types1,Types2,Types),
    type_kodkod_declarations(Drest,Env,EnvInter,Nrest,EnvOut,Types2).
type_kodkod_declaration(decl(Id,Arity,Mult,Expr),Env,EnvIn,
                        decl(Id,Arity,Mult,NExpr),EnvOut,Types) :-
    add_var_to_type_env(Id,Types,Env,Subenv),
    type_kodkod(Expr,Subenv,rel(Types),NExpr),
    add_var_to_type_env(Id,Types,EnvIn,EnvOut).


ktype(true,formula).
ktype(false,formula).
ktype(not(formula),formula).
ktype(and(formula,formula),formula).
ktype(or(formula,formula),formula).
ktype(implies(formula,formula),formula).
ktype(iff(formula,formula),formula).

ktype(one(rel(_)), formula).
ktype(some(rel(_)), formula).
ktype(no(rel(_)), formula).
ktype(lone(rel(_)), formula).
ktype(set(rel(_)), formula).

ktype(in(rel(T),rel(T)), formula).
ktype(equals(T,T),formula).
ktype(pfunc(rel(_),rel(_),rel(_)), formula).
ktype(func(rel(_),rel(_),rel(_)), formula).
ktype(one(rel(_)), formula).
ktype(some(rel(_)), formula).

ktype(empty,rel(_)).
ktype(iden,rel(_)).
ktype(univ,rel(_)).
ktype(intersection(rel(A),rel(A)), rel(A)).
ktype(union(rel(A),rel(A)), rel(A)).
ktype(difference(rel(A),rel(A)), rel(A)).
ktype(overwrite(rel(A),rel(A)), rel(A)).
ktype(closure(rel(A)), rel(A)).
ktype(reflexive_closure(rel(A)), rel(A)).
ktype(product(rel(A),rel(B)), rel(C)) :- safe_append(A,B,C).
ktype(transpose(rel(R)), rel(T)) :- safe_reverse(R,T).
ktype(join(rel(A),rel([Bhead|Btail])), rel(C)) :-
    safe_append(Astart,[Bhead],A),
    safe_append(Astart,Btail,C).
ktype(if(formula,rel(R),rel(R)), rel(R)).

ktype(gt(int_expression,int_expression), formula).
ktype(gte(int_expression,int_expression), formula).
ktype(lt(int_expression,int_expression), formula).
ktype(lte(int_expression,int_expression), formula).

ktype(card(rel(_)), int_expression).
ktype(add(int_expression,int_expression), int_expression).
ktype(sub(int_expression,int_expression), int_expression).
ktype(mul(int_expression,int_expression), int_expression).
ktype(div(int_expression,int_expression), int_expression).
ktype(mod(int_expression,int_expression), int_expression).
ktype(sum(rel([_])), int_expression).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

nth0s(Positions,List,Elements) :- maplist(nth0s2(List),Positions,Elements).
nth0s2(List,Position,Element) :- nth0(Position,List,Element).

safe_append(A,B,C) :-
    complete_list(A,AComplete),
    complete_list(B,BComplete),
    complete_list(C,CComplete),
    safe_append2(A,B,C,AComplete,BComplete,CComplete).
:- block safe_append2(?,?,?,-,-,?), safe_append2(?,?,?,-,?,-).
safe_append2(A,B,C,1,1,1) :- append(A,B,C),!.
:- block complete_list(-,?).
complete_list([],1).
complete_list([_|T],Trigger) :- complete_list(T,Trigger).

safe_reverse(A,B) :-
    safe_reverse2(A,[],B),
    safe_reverse2(B,[],A).
:- block safe_reverse2(-,?,?).
safe_reverse2([],Acc,Acc).
safe_reverse2([H|T],Acc,Rev) :-
    safe_reverse2(T,[H|Acc],Rev).
