% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

% LTL and CTL model checking

:- module(ltl,[ltl_model_check/4,ltl_model_check2/4,
               ltl_check_assertions/2, parse_ltlfile/2,
               ltl_model_check_with_ce/4, parse_and_pp_ltl_formula/2,
               ltl_formula_structure/3,
               pp_ltl_formula/2,
               counterexample_node/1, counterexample_op/1,
               preprocess_tp/2,
               preprocess_formula/2,
               extract_formulas/4,
               reset_ltl/0
              ]).

:- use_module(library(lists)).
:- use_module(library(avl)).
:- use_module(library(codesio)).
:- use_module(library(ordsets)).

:- use_module(probsrc(translate)).
:- use_module(probsrc(debug)).
:- use_module(probsrc(bmachine),[b_get_machine_operation/4,b_type_expression/5,
                        b_definition_prefixed/4, b_get_typed_definition/3, b_top_level_operation/1]).
:- use_module(probsrc(error_manager)).
:- use_module(probsrc(tools)).
:- use_module(probsrc(specfile),[animation_mode/1, csp_mode/0, csp_with_bz_mode/0,
                        state_corresponds_to_initialised_b_machine/2]).
:- use_module(probsrc(self_check)).

:- use_module(probsrc(state_space),[set_max_nr_of_new_impl_trans_nodes/1,impl_trans_term/3,
                           current_state_id/1,visited_expression/2,
                           transition/3,transition/4,
                           state_space_add/2,retract_visited_expression/2,  % used for test cases
                           impl_trans_term_all/2,state_space_reset/0,
                           execute_id_trace_from_current/3]).
:- use_module(probcspsrc(haskell_csp),[filter_formulas_from_pragmas/3]).
:- use_module(probsrc(kernel_objects),[equal_object/2]).
:- use_module(probsrc(b_interpreter),[b_compute_expression/4,b_test_boolean_expression_wf/3]).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(parsercall),[parse_exprs_preds/4]).
:- use_module(probsrc(store),[empty_state/1,lookup_value/3,add_var_to_localstate/4]).

:- use_module(probsrc(ample_sets), [stutter_action/1, visible_action/1,
                           compute_dependendency_relation_of_all_events_in_the_model/3,
                           reset_runtime_dynamic_predicates_POR/0]).
:- use_module(probsrc(b_state_model_check), [get_guard/2]).
:- use_module(probsrc(enabling_analysis),[b_get_read_write/3,b_get_read_write/5]).
:- use_module(probsrc(plugins)).

:- use_module(probltlsrc(ltl_tools)).

:- use_module('../promela/promela_tools',[parse_promela_strings/4,eval_promela/4]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,ltl).
:- module_info(description,'This module provides LTL model checking on models.').

/* LTL relevant PL modules */
% import the low-level C-library wrapper
:- use_module(extension('ltlc/ltlc')).

:- use_module(probltlsrc(ltl_fairness),[initialise_fairness_assumptions/4,ltl_model_check_core_with_fairness/5]).
:- use_module(probltlsrc(ltl_verification)).

trans(A,B,Op) :-
    ( (preferences: preference(use_por_for_ltl,true), animation_mode(b))
        -> user: compute_ample_set2(A),
           transition(A,Op,B) % from state_space.pl
        ;  impl_trans_term(A,Op,B) /* will automatically compute as required */
    ).

:- use_module(probsrc(eclipse_interface),[test_boolean_expression_in_node/2]).
% check an atomic proposition in a given node (i.e. state)
check_ap(bpred(Property),Node) :-
    !, test_boolean_expression_in_node(Node,Property).
check_ap(promela_pred(Predicate),Node) :-
    visited_expression(Node,State),!,
    eval_promela(Predicate,global,State,Result),!,
    ( Result == 0    -> fail
    ; number(Result) -> true
    ; otherwise      -> add_error(ltl,'Promela eval returned a non-boolean value: ',Result),fail).
check_ap(enabled(Op),Node) :- !,
    check_enabled(Op,Node).
check_ap(det(OpList),Node) :- !,
    check_if_maximum_one_enabled(OpList,Node).
check_ap(dlk(OpList),Node) :- !,
    check_if_all_disabled(OpList,Node).
check_ap(ctrl(OpList),Node) :- !,
    is_exactly_one_operation_enabled(OpList,Node).
check_ap(deadlock,Node) :- !, \+ trans(Node,_,_).
check_ap(sink,Node) :- !,
    \+ ( trans(Node,Next,_),Next \= Node ).
check_ap(stateid(SpecNode),Node) :- !, SpecNode = Node.

check_if_maximum_one_enabled(OpList,Node) :-
    check_if_maximum_one_enabled(OpList,0,Node).

check_if_maximum_one_enabled([],_NrEnabled,_Node).
check_if_maximum_one_enabled([Op|Ops],NrEnabled,Node) :-
    (check_enabled(Op,Node) -> 
        NrEnabled == 0, NrEnabled1 = 1
    ;   NrEnabled1 = 0
    ), check_if_maximum_one_enabled(Ops,NrEnabled1,Node).

is_exactly_one_operation_enabled(OpList,Node) :-
    is_exactly_one_operation_enabled(OpList,0,Node).

is_exactly_one_operation_enabled([],NrEnabled,_Node) :-
    NrEnabled=1.
is_exactly_one_operation_enabled([Op|Ops],NrEnabled,Node) :-
    (check_enabled(Op,Node) -> 
        NrEnabled == 0, NrEnabled1 = 1
    ;   NrEnabled1 = NrEnabled
    ), is_exactly_one_operation_enabled(Ops,NrEnabled1,Node).        

check_if_all_disabled([],_Node).
check_if_all_disabled([Op|Ops],Node) :-
    \+check_enabled(Op,Node),
    check_if_all_disabled(Ops,Node).

check_enabled(Spec,SrcNode) :-
        check_transition(Spec,Transition,SrcNode,DestNode),
        trans(SrcNode,DestNode,Transition),!.

% check_transition(Spec,Transition,SrcNode,DstNode)
%   Checks if a transition fulfils the specification, it fails if not.
%   Please note, that this predicate is called _before_ the actual transition
%   is known.
%  Spec: specification that encodes how the transition should look like
%        See below for the supported transitions
%  Transition: The transition itself as it is stored in the second arg of state_space:transition/3
%  SrcNode: The Id of the source state (first arg of transition/3)
%  DstNode: The Id of the destination state (third arg of transition/3)

:- assert_must_succeed( check_transition(bop(dummyop,2,0,[],[]),dummyop(int(1),int(2)),tsrc,tdst) ).
:- assert_must_fail( check_transition(bop(dummyop,2,0,[],[]),otherop(int(1),int(2)),tsrc,tdst) ).
:- assert_must_fail( check_transition(bop(dummyop,2,0,[],[]),dummyop(int(1)),tsrc,tdst) ).
:- assert_must_succeed(call_cleanup(( empty_state(E),
                                      state_space_add(tsrc,E),
                                      create_texpr(integer(2),integer,[],I2),
                                      check_transition(bop(dummyop,2,0,[2/I2],[]),
                                                       dummyop(int(1),int(2)),tsrc,tdst)),
                                    retract_visited_expression(tsrc,_))).
:- assert_must_fail(call_cleanup(( empty_state(E),
                                   state_space_add(tsrc,E),
                                   create_texpr(integer(2),integer,[],I2),
                                   check_transition(bop(dummyop,2,0,[1/I2],[]),
                                                    dummyop(int(1),int(2)),tsrc,tdst)),
                                 retract_visited_expression(tsrc,_))).
:- assert_must_succeed( check_transition(bop(dummyop,1,1,[],[]),'-->'(dummyop(int(1)),[int(2)]),tsrc,tdst) ).
:- assert_must_fail( check_transition(bop(dummyop,1,1,[],[]),dummyop(int(1)),tsrc,tdst) ).
:- assert_must_succeed(call_cleanup(( empty_state(E),
                                      state_space_add(tsrc,E),
                                      create_texpr(integer(2),integer,[],I2),
                                      check_transition(bop(dummyop,1,1,[],[1/I2]),
                                                       '-->'(dummyop(int(1)),[int(2)]),tsrc,tdst)),
                                    retract_visited_expression(tsrc,_))).
:- assert_must_fail(call_cleanup(( empty_state(E),
                                   state_space_add(tsrc,E),
                                   create_texpr(integer(2),integer,[],I2),
                                   check_transition(bop(dummyop,1,1,[],[1/I2]),
                                                    '-->'(dummyop(int(1)),[int(1)]),tsrc,tdst)),
                                 retract_visited_expression(tsrc,_))).
                       
% bop(Name,NumberArgs,NumberResults,ArgPatterns,ResPatterns):
%   The transition is a B operation with name Name, NumberArgs parameters and NumberResults
%   result values.
%   ArgPatterns is a list of Terms of the form Index/Expr where Index is the
%     number of the parameter that should have the value of the expression Expr.
%   ResPatterns is a list analogous to ArgPatterns on the result
check_transition(bop(Name,NumberArgs,NumberResults,ArgPatterns,ResPatterns),
                 Transition,SrcNode,_DstNode) :-
    !, functor(Op,Name,NumberArgs),
    ( NumberResults == 0 -> Transition = Op
    ; otherwise          ->
        Transition = '-->'(Op,ReturnValues),
        length(ReturnValues,NumberResults),
        check_bop_res_patterns(ResPatterns,SrcNode,State,ReturnValues)),
    check_bop_arg_patterns(ArgPatterns,SrcNode,State,Op).
check_transition(btrans(TransPred),Transition,SrcNode,DstNode) :-
    !, check_eventb_transition(TransPred,Transition,SrcNode,DstNode).
check_transition(csp(Channel,Arity,Args),Transition,_SrcNode,_DstNode) :-
    !, check_csp_event(Channel,Arity,Args,Transition).
check_transition(xtl(Op,Arity,PatternArgs),Transition,_SrcNode,_DstNode) :-
    !, check_xtl_transition(Op,Arity,PatternArgs,Transition).
check_transition(plugin_transition_check(Id,Predicate),Transition,_SrcNode,_DstNode) :-
    !, plugin_evaluate_transition_predicate(Id,Predicate,Transition).
check_transition(Spec,_Trans,_SrcNode,_DstNode) :-
    functor(Spec,Name,Arity), add_error(ltl, 'Unknown transition specification: ', Name/Arity), fail.

:- block check_xtl_transition(?,?,?,-).
check_xtl_transition(Op,Arity,PatternArgs,Transition) :-
    ( Arity=0 ->
        functor(Transition,Op,_) /* in this case: pattern match any operation with right number of args */
    ;   %functor(Transition,Op,Arity),
        Transition =.. [Op|Args],
        check_xtl_arguments(Args,PatternArgs)
    ).

check_bop_arg_patterns([],_,_,_).
check_bop_arg_patterns([P|Prest],Node,State,Op) :-
    check_bop_arg_patterns2([P|Prest],Node,State,Op).
:- block check_bop_arg_patterns2(?,?,?,-).
check_bop_arg_patterns2([],_,_,_).
check_bop_arg_patterns2([Index/Expr|Prest],Node,State,Op) :-
    check_bop_arg_pattern(Index,Expr,Node,State,Op),
    check_bop_arg_patterns2(Prest,Node,State,Op).
check_bop_arg_pattern(Index,Expr,Node,State,Op) :-
    arg(Index,Op,OpValue),  % exception if Index is expression like 1+1 <----
    check_b_operation_value(OpValue,Expr,Node,State).

check_bop_res_patterns([],_,_,_).
check_bop_res_patterns([P|Prest],Node,State,OpResult) :-
    check_bop_res_patterns2([P|Prest],Node,State,OpResult).
:- block check_bop_res_patterns2(?,?,?,-).
check_bop_res_patterns2([],_,_,_).
check_bop_res_patterns2([Index/Expr|Prest],Node,State,OpResult) :-
    check_bop_res_pattern(Index,Expr,Node,State,OpResult),
    check_bop_res_patterns2(Prest,Node,State,OpResult).
check_bop_res_pattern(Index,Expr,Node,State,OpResult) :-
    nth1(Index,OpResult,OpValue),
    check_b_operation_value(OpValue,Expr,Node,State).

check_b_operation_value(OpValue,Expr,Node,State) :-
    ( var(State) ->
        visited_expression(Node,RawState),
        state_corresponds_to_initialised_b_machine(RawState,State)
    ; otherwise -> true),
    empty_state(LS),
    b_compute_expression(Expr,LS,State,PatternValue),
    equal_object(PatternValue,OpValue),!.
    

:- block check_eventb_transition(?,-,?,?).
check_eventb_transition(event(Name),Transition,_SrcNode,_DstNode) :- !,
    functor(Transition,Name,_).
check_eventb_transition(event(Name,Predicate,PoststateAccess),Transition,SrcNode,DstNode) :- !,
    functor(Transition,Name,_),
    visited_expression(SrcNode,CurState),
    state_corresponds_to_initialised_b_machine(CurState,CurBState),
    parameter_values(Transition,LocalState1),
    add_post_state_values(PoststateAccess,DstNode,LocalState1,LocalState),
    b_test_boolean_expression_wf(Predicate,LocalState,CurBState).

add_post_state_values([],_DstNode,State,State) :- !.
add_post_state_values(Access,Node,InState,OutState) :-
    visited_expression(Node,DstState1),
    state_corresponds_to_initialised_b_machine(DstState1,DstState),
    add_post_state_values2(Access,DstState,InState,OutState).
add_post_state_values2([],_DstState,State,State).
add_post_state_values2([poststate(Orig,Primed)|Prest],DstState,InState,OutState) :-
    lookup_value(Orig,DstState,Value),
    add_var_to_localstate(Primed,Value,InState,InterState),
    add_post_state_values2(Prest,DstState,InterState,OutState).

parameter_values(Transition,LocalState) :-
    Transition =.. [Name|Arguments],
    b_get_machine_operation(Name,_,Parameters,_),
    empty_state(Empty),
    add_parameter_values2(Parameters,Arguments,Empty,LocalState).
add_parameter_values2([],[],State,State).
add_parameter_values2([TParameter|Prest],[Argument|Arest],InState,OutState) :-
    def_get_texpr_id(TParameter,Parameter),
    add_var_to_localstate(Parameter,Argument,InState,InterState),
    add_parameter_values2(Prest,Arest,InterState,OutState).

/*
:- assert_must_succeed(( ltl:split_operation('ch.2.3','.',ch,A), A=[2,3] )).
:- assert_must_succeed(( ltl:split_operation('ch.2.3','.',ch,[2,3]) )).
:- assert_must_succeed(( ltl:split_operation('ch','.',ch,[]) )).
:- assert_must_succeed(( ltl:split_operation('ch.','.',ch,[A]), A='' )).
split_operation(OpAtom, Sep,  OpName, OpArgs) :-
   atom_codes(OpAtom,ListAscii), atom_codes(Sep,[SepACode]),
   split_op(ListAscii,SepACode,HeadA, ArgsA),
   atom_codes(Head,HeadA), OpName=Head,
   l_name(Args,ArgsA), Args=OpArgs.
l_name([],[]).
l_name([H|T],[AH|AT]) :- name(H,AH),  % using name to generate numbers if possible
                               l_name(T,AT).

split_op([],_,[],[]). % no separator found
split_op([H|T],Sep,OpName,Args) :-
   (H=Sep -> Args=[FirstArg|OtherArgs],OpName=[],split_args(T,Sep,FirstArg,OtherArgs)
    ; OpName = [H|CT], split_op(T,Sep,CT,Args)).

split_args([],_Sep,[],[]).
split_args([H|T],Sep,FirstArg,OtherArgs) :-
   (H=Sep -> FirstArg=[], OtherArgs=[NextArg|TA],split_args(T,Sep,NextArg,TA)
    ; FirstArg = [H|CT], split_args(T,Sep,CT,OtherArgs)).
*/


check_xtl_arguments([],[]).
check_xtl_arguments([],[H|T]) :-
   print('### Extra arguments in LTL/CTL pattern: '),print([H|T]),nl,fail.
check_xtl_arguments([H|T],[]) :- print(ltl_check_xtl_arguments_ignoring([H|T])),nl.
check_xtl_arguments([H|T],[PatH|PatT]) :-
    check_xtl_arg(PatH,H,T,RemArgs),
    check_xtl_arguments(RemArgs,PatT).

check_xtl_arg('_',_,TArgs,RemArgs) :- !,RemArgs=TArgs.
check_xtl_arg('?',_,TArgs,RemArgs) :- !,RemArgs=TArgs.
check_xtl_arg(X,Y,TArgs,RemArgs) :- atomic(Y),!,X=Y,RemArgs=TArgs.
check_xtl_arg(X,Y,TArgs,RemArgs) :- nonvar(Y),!,Y=..[Functor|YArgs], Functor=X, 
  append(YArgs,TArgs,RemArgs).
check_xtl_arg(X,Y,TArgs,RemArgs) :- var(Y),
   print(variable_for_ltl_pattern(X)),nl,
   X=Y,RemArgs=TArgs.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% get LTL formulas from 'DEFINITIONS' for B machines and 
% from pragmas ({-# ... #-}) for CSP-M specifications
:- dynamic ltl_formula_cache/2.

get_ltl_formulas(Names,Strings,Formulas) :-
    (animation_mode(b)
        -> get_ltl_formula_strings(Names,Strings)
    ; animation_mode(cspm)
        -> filter_formulas_from_pragmas(ltl,Names,Strings)
    ; animation_mode(csp_and_b)
        -> get_ltl_formula_strings(BNames,BStrings),
           filter_formulas_from_pragmas(ltl,CSPNames,CSPStrings),
           append(BNames,CSPNames,Names),
           append(BStrings,CSPStrings,Strings)
    ),
    (ltl_formula_cache(Strings,Formulas) -> true
    ; otherwise ->
        retractall(ltl_formula_cache(_,_)),
        temporal_parser_l(Strings,ltl,Formulas),
        assert(ltl_formula_cache(Strings,Formulas))).

get_ltl_formula_strings(Names,Formulas) :-
    findall(ltl_assertion(Name,F),get_ltl_formula_string(Name,F),Assertions),
    findall(N,member(ltl_assertion(N,_),Assertions),Names),
    findall(F,member(ltl_assertion(_,F),Assertions),Formulas).    
get_ltl_formula_string(Tail,S) :-
    b_definition_prefixed(_,'ASSERT_LTL',Tail,DefName),
    b_get_typed_definition(DefName,[],F),
    ( get_texpr_expr(F,string(S)) -> true
    ; get_texpr_pos(F,Pos) ->
        add_all_perrors([error('Expected String for LTL formula',Pos)]),
        fail).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% check all LTL assertions in a machine

ltl_check_assertions(Limit,Outcome) :-
    get_ltl_formulas(Names,Strings,Formulas),
    extract_expectations(Names,Exp),
    ltl_check_assertions2(Names,Strings,Formulas,Exp,Limit,Outcomes),
    merge_outcomes(Outcomes,Outcome).

ltl_check_assertions2([],[],[],[],_,[]).
ltl_check_assertions2([Name|Nrest],[String|Srest],[Formula|Frest],[Exp|Erest],Limit,[Outcome|Orest]) :-
    print('Checking LTL formula '),print(Name),print(': '),print(String),nl,
    flush_output(user_output),
    catch( ltl_model_check2(Formula,Limit,init,Status),
           E,
           Status = exception(E)),
    ( Exp=pass, Status=ok -> Outcome=pass, print('no counter-example found, test passed\n')
    ; Exp=fail, Status=no -> Outcome=pass, print('counter-example found, test passed\n')
    ; Status=exception(E) ->
        add_error(ltl,'Caught exception: ',E),Outcome=error
    ; Exp=pass, Status=no ->
        Outcome=fail, add_error(ltl, 'Expected correct formula, but found counter-example')
    ; Exp=fail, Status=ok ->
        Outcome=fail, add_error(ltl, 'Expected counter-example, but model checker found none')
    ; otherwise ->
        add_error(ltl,'Unexpected result: ',(Exp,Status)),Outcome=fail),
    ltl_check_assertions2(Nrest,Srest,Frest,Erest,Limit,Orest).

extract_expectations([],[]).
extract_expectations([N|Nrest],[E|Erest]) :-
    name(N,Codes),
    ( append(_,"FAIL",Codes) -> E=fail; E=pass),
    extract_expectations(Nrest,Erest).

merge_outcomes(List,Result) :-
    merge_outcomes2(List,no_tests,Result).
merge_outcomes2([],R,R).
merge_outcomes2([O|Rest],P,R) :-
    merge_outcomes3(P,O,I),
    merge_outcomes2(Rest,I,R).
merge_outcomes3(no_tests,X,X) :- !.
merge_outcomes3(_,error,error) :- !.
merge_outcomes3(P,fail,fail) :- P\=error,!.
merge_outcomes3(P,pass,P).
 
%*******************************************************************************

% ltl_model_check(FormulaAsAtom,Max,Mode,Status):
%  FormulaAsAtom: The LTL formula as an atom
%  Max: The maximum number of states that should be newly calculated
%  Mode: Where to start the search in state space
%     starthere: start the evaluation of the formula in the current state
%     init: start in the (possible several) initialisation states of the model
%     checkhere: start the search as in init, but check the formula F in the
%        current state by constructing a new formula (current -> F)
%  Status:
%     no: the model does not fulfil the formula, a counter-example is shown by
%         modifying the history
%     ok: the model does fulfil the formula
%     incomplete: the search was stopped (usually because the limit Max was reached)
%         before finding a counter-example or guarantee the absent of a counter-example
ltl_model_check(FormulaAsAtom,Max,Mode,Status) :-
    print_message(parsing_ltl_formula),
    retractall(counterexample_node(_)),
    retractall(counterexample_op(_)),
    ( temporal_parser(FormulaAsAtom,ltl,Ltl) ->
        true
    ; otherwise ->
        add_error(ltl,'LTL Parser failed: ',FormulaAsAtom),fail),
    ltl_model_check2(Ltl,Max,Mode,Status).

ltl_model_check2(Ltl,Max,Mode,Status) :-
    print_message(initialising_ltlc),
    print(starting_model_checking),nl,
    ( Ltl=syntax_error ->
        Status=syntax_error
    ; otherwise ->
        ltl_model_check3(Ltl,Max,Mode,Result),
        convert_result(Result,Mode,Status)).

ltl_model_check3(Ltl,Max,Mode,Result) :- 
    ( (preferences: preference(use_por_for_ltl,true),animation_mode(b))
        -> prepare_ltl_model_checker_for_por(Ltl)
        ;  true
    ),
    call_c_initialisation,
    set_max_nr_of_new_impl_trans_nodes(Max),
    % initialising all given fairness constraints
    initialise_fairness_assumptions(Ltl,WeakFairnessAssumption,StrongFairnessAssumption,LtlFormula),
    debug_println(9,fairness_constraints(weak(WeakFairnessAssumption),strong(StrongFairnessAssumption))),
    select_ltl_mode(Mode,LtlFormula,Startnodes,Ltl2),
    preprocess_formula(Ltl2,Ltl3),
    debug_println(9,preprocess_formula(LtlFormula,Ltl3)),
    ((WeakFairnessAssumption = [], StrongFairnessAssumption = []) 
      -> ltl_model_check_core(Startnodes,Ltl3,Result) % no fairness assumptions have been made
       ; ltl_model_check_core_with_fairness(Startnodes,Ltl3,WeakFairnessAssumption,StrongFairnessAssumption,Result)
    ),
    call_ltl_verification(Result,Ltl3),
    call_ltl_fairness_verification(Result,WeakFairnessAssumption,StrongFairnessAssumption),
    debug_println(9,ltl_result(Result)).

ltl_model_check_core([],_Ltl,nostart) :- !.
ltl_model_check_core(Startnodes,Ltl,Result) :-
    c_ltl_modelcheck(Ltl, Startnodes, Result, ltl:cltl_callback).

call_c_initialisation :- ltlc_init,!.
call_c_initialisation :- 
    add_error(ltl,'Initialisation of LTL-C module failed'),
    fail.

select_ltl_mode(starthere,Ltl,[Startnode],not(Ltl)) :-
    !,current_state_id(Startnode).
select_ltl_mode(init,Ltl,Startnodes,not(Ltl)) :-
    !,init_states(Startnodes).
select_ltl_mode(checkhere,Ltl,Startnodes,not(globally(implies(ap(current),Ltl)))) :-
    !,init_states(Startnodes).
select_ltl_mode(Mode,Ltl,[],Ltl) :-
    add_error(ltl,'Internal error: Unexpected mode: ', Mode),
    fail.

convert_result(nostart,_,nostart) :- !.
convert_result(none,_,ok) :- !.
convert_result(incomplete,_,incomplete) :- !.
convert_result(model(CE,_Entry),Mode,no) :- !,
    translate_ctrace(CE,Op,N,Dest),
    ( (Mode=init; Mode=checkhere) ->
        ce_get_first_state(CE,UsedInitState),
        set_trace_to_init(UsedInitState)
    ; otherwise ->
        true),
    execute_id_trace_from_current(Dest,Op,N).
convert_result(Result,_,_) :-
    add_error(ltl,'Internal error: Unexpected result: ', Result),
    fail.

ce_get_first_state([atom(State,_,_)|_],State).

:- dynamic counterexample_node/1.
:- dynamic counterexample_op/1.

reset_ltl :-
   retractall(counterexample_node(_)),
   retractall(counterexample_op(_)),
   retractall(ltl_formula_cache(_,_,_)).
  
:- use_module(probsrc(eventhandling),[register_event_listener/3]).
:- register_event_listener(reset_specification,reset_ltl,
                           'Reset LTL Model Checker.').

translate_ctrace(Atoms,OpIds,StateIds,Dest) :-
    translate_ctrace2(Atoms,beginning,OpIds,StateIds,Dest).
translate_ctrace2([atom(StartNode,_,OpId)|RestAtoms],_PreviousState,Ops,Nodes,Dest) :-
    assert(counterexample_node(StartNode)),
    assert(counterexample_op(OpId)),
    ( OpId = none -> Ops=[],Nodes=[],Dest=StartNode
    ; otherwise ->
        transition(StartNode,_Op,OpId,NextNode),
        Ops=[OpId|OpRest], Nodes=[NextNode|NextRest], 
        translate_ctrace2(RestAtoms,NextNode,OpRest,NextRest,Dest)).
translate_ctrace2([],Dest,[],[],Dest).

% currently not called: nor is unparsed_ap called 
translate_preds(In,Out) :-
    unparsed_ap_l(In,Out,Translation,[]),
    parse_preds_exprs(Translation).

parse_preds_exprs([]) :- !. % nothing to parse
parse_preds_exprs(Translation) :-
    split_preds(Translation,PredStrings,Preds,ExprStrings,Exprs),
    animation_mode(AnimMode),
    parse_preds_exprs2(AnimMode,PredStrings,Preds,ExprStrings,Exprs).

parse_preds_exprs2(promela,PredStrings,Preds,ExprStrings,Exprs) :- !,
    annotate_ap_type(Preds,promela_pred,PromPreds),
    parse_promela_strings(ExprStrings,PredStrings,Exprs,PromPreds).

parse_preds_exprs2(Mode,PredStrings,Preds,ExprStrings,Exprs) :-
    is_registered_plugin(Mode),!,
    plugin_parse(Mode,ExprStrings,PredStrings,[],Exprs,Preds,[]).

% default: parse the predicates and expressions as B 
parse_preds_exprs2(_,PredStrings,Preds,ExprStringsWithTypes,Exprs) :-
    annotate_ap_type(Preds,bpred,BPreds),
    split_exprs_types(ExprStringsWithTypes,ExprStrings,Types),
    parse_exprs_preds(ExprStrings,UExprs,PredStrings,UPreds),
    type_exprs(UExprs,Types,Exprs,EErrors),
    type_preds(UPreds,BPreds,PErrors),
    append(EErrors,PErrors,Errors),
    add_all_perrors(Errors,[],type_error).

split_exprs_types([],[],[]).
split_exprs_types([Expr/Type|ETRest],[Expr|ERest],[Type|TRest]) :-
    split_exprs_types(ETRest,ERest,TRest).

type_exprs([],[],[],[]).
type_exprs([Expr|Erest],[Type|TypeRest],[Typed|Trest],Errors) :-
    b_type_expression(Expr,[variables],Type,Typed,LErrors),
    append(LErrors,ErrorRest,Errors),
    type_exprs(Erest,TypeRest,Trest,ErrorRest).
type_preds([],[],[]).
type_preds([Pred|Prest],[Typed|Trest],Errors) :-
    b_type_expression(Pred,[variables],pred,Typed,LErrors),
    append(LErrors,ErrorRest,Errors),
    type_preds(Prest,Trest,ErrorRest).

annotate_ap_type([],_,[]).
annotate_ap_type([AP|APrest],Type,[R|Rrest]) :-
    functor(AP,Type,1),arg(1,AP,R),
    annotate_ap_type(APrest,Type,Rrest).



split_preds([],[],[],[],[]).
split_preds([pred(A,B)|Rest],[A|ARest],[B|BRest],X,Y) :- !,split_preds(Rest,ARest,BRest,X,Y).
split_preds([expr(A,T,B)|Rest],X,Y,[A/T|ARest],[B|BRest]) :- split_preds(Rest,X,Y,ARest,BRest).

% replacing unparsed/1 by atomic propositions
unparsed_ap(ap(P),ap(P)) --> !.
unparsed_ap(action(unparsed_b(OpName/Arity/Types,Exprs)),action(b(OpName/Arity,Es))) --> !,
    add_unparsed_exprs(Exprs,Types,Es).
unparsed_ap(action(b(OpName/Arity/_Types)),action(b(OpName/Arity))) --> !.
unparsed_ap(action(Op),action(Op)) --> !.
unparsed_ap(unparsed(Unparsed),ap(P)) --> !,[pred(Unparsed,P)].
unparsed_ap(F,N) --> {F =..[Functor|Args]},unparsed_ap_l(Args,NArgs),{N=..[Functor|NArgs]}.
unparsed_ap_l([],[]) --> [].
unparsed_ap_l([F|Rest],[N|NRest]) --> unparsed_ap(F,N), unparsed_ap_l(Rest,NRest).

add_unparsed_exprs([],[],[]) --> [].
add_unparsed_exprs([expr(P,U)|UR],[T|TR],[expr(P,E)|ER]) -->
    [expr(U,T,E)], add_unparsed_exprs(UR,TR,ER).

%*******************************************************************************
% set current trace

executetrace(Trace) :-
    translate_trace(Trace,OpIds,Ids,Dest),
    debug_println(9,exec(Dest,OpIds,Ids)),
    execute_id_trace_from_current(Dest,OpIds,Ids),!.
executetrace(Trace) :- add_internal_error('Could not execute counter-example: ',executetrace(Trace)).

translate_trace([Start,Id|RestNodes],[OpId|OpRest],[Id|IdRest],Dest) :-
    !,state_space:transition(Start,_Op,OpId,Id),
    translate_trace([Id|RestNodes],OpRest,IdRest,Dest).
translate_trace([Dest],[],[],Dest).


%*******************************************************************************
% find initialised states
% TODO: does NOT work with XTL (CSP?) and constants

init_states(Init) :-
    animation_mode(Mode),
    ( is_registered_plugin(Mode) ->
        findall(I,find_plugin_inits(root,Mode,_,I),Init)
    ; mode_cst_init(Mode) ->
        findall(I,find_init(root,I,_),Init)
    ; mode_one_step(Mode) ->
        next_states(root,Init)
    ; otherwise ->
        fail).

mode_cst_init(b).
mode_cst_init(z).
mode_cst_init(csp_and_b).

mode_one_step(csp).
mode_one_step(cspm).
mode_one_step(xtl).
mode_one_step(promela).

next_states(Start,States) :-
    impl_trans_term_all(Start,Ops),
    findall(S, member(op(_Id,_,S),Ops), States).

find_init(Start,Init,Trace) :- Start==Init,!,Trace=[]. % usually called with Start=Init=root
find_init(Start,Init,[State|Rest]) :-
    impl_trans_term(Start,O,State),
    find_init2(O,State,Init,Rest).
find_init2(O,Init,Init,[]) :-
    has_functor_and_maybe_tau(O,'$initialise_machine').
find_init2(O,State,Init,Path) :-
    has_functor_and_maybe_tau(O,'$setup_constants'),
    find_init(State,Init,Path).
find_init2(start_cspm_MAIN,State,Init,Path) :-
    find_init(State,Init,Path).
    

% has_functor_and_maybe_tau(Term,Functor)
% checks if Term has the form "Functor(...)" or "tau(Functor(...))"
% this is used for CSP||B specification where the initialisation is wrapped with
% in a tau operator
has_functor_and_maybe_tau(tau(Term),Functor) :-
    has_functor_and_maybe_tau(Term,Functor),!.
has_functor_and_maybe_tau(Term,Functor) :-
    functor(Term,Functor,_).


find_plugin_inits(Start,Id,[Start|Rest],StateId) :-
    visited_expression(Start,State),
    ( plugin_is_initialised_state(Id,State) ->
        StateId = Start,
        Rest = []
    ; otherwise ->
        impl_trans_term(Start,_,Next),
        find_plugin_inits(Next,Id,Rest,StateId)).

set_trace_to_init(Initstate) :-
    animation_mode(Mode),
    find_trace_to_init(Mode,Initstate,Trace),!,
    state_space_reset,
    executetrace(Trace).
set_trace_to_init(Initstate) :- add_internal_error('Could not replay counter-example: ',set_trace_to_init(Initstate)).

find_trace_to_init(Mode,Target,[root,Target]) :-
    mode_one_step(Mode).
find_trace_to_init(Mode,Target,[root|Trace]) :-
    mode_cst_init(Mode),
    find_init(root,Target,Trace).
find_trace_to_init(Mode,Target,Trace) :-
    is_registered_plugin(Mode),
    find_plugin_inits(root,Mode,Trace,Target).
                  

%%% Predicates for determining which actions are visible and which are stutter in regard to the given LTL formula.
%%% An operation Op in a B machine should be identified as visible for a given LTL formula f, if Op writes variables of the predicates
%%% in f, or if Op is used as operation inside [-] or e(-) in f. 
prepare_ltl_model_checker_for_por(Ltl) :- print(prepare_ltl_model_checker_for_por(Ltl)),nl,
    %reset_runtime_dynamic_predicates_POR,now done via event-handling
    % state space should be reseted every time when a new LTL_X formula is checked,
    % because of the stutter condition
    prob2_interface:reset_animator,
    prob2_interface:initialise_animator,
    get_stutter_visible_actions(Ltl,StutterActions,VisibleActions),
    debug_println(9,visible_actions(VisibleActions)),
    debug_println(9,stutter_actions(StutterActions)),
    assert_all_visible_stutter_actions(StutterActions,VisibleActions),
    % if the model is not reloaded _EnableGraph will not be computed every time a new LTL formula is checked
    compute_dependendency_relation_of_all_events_in_the_model(0,por(ample_sets,0,0,0),_EnableGraph).

% get_stutter_visible_actions(+Ltl,-StutterActions,-VisibleActions)
% Ltl: the parsed Ltl formula presented as Prolog term.
% StutterActions: the set of stutter actions in regard to Ltl
% VisibleActions: the set of visible actions in regard to Ltl
get_stutter_visible_actions(Ltl,StutterActions,VisibleActions) :-
    findall(Op,b_top_level_operation(Op),Ops),
    list_to_ord_set(Ops,Operations),
    look_up_for_visible_actions(Ltl,Operations,VisActions),
    list_to_ord_set(VisActions,VisibleActions),
    ord_subtract(Operations,VisibleActions,StutterActions).

look_up_for_visible_actions(ap(bpred(Pred)),Ops,Acts) :- !,
    predicate_identifiers(Pred,Ids),
    list_to_ord_set(Ids,Vars),
    get_actions_modifying_variables(Vars,Ops,Acts).
look_up_for_visible_actions(action(bop(Name,_NArgs,_NRes,_Args,_Outputs)),Ops,Acts) :- !,
    b_get_read_write(Name,Reads,_Writes),
    list_to_ord_set(Reads,Vars),
    get_actions_modifying_variables(Vars,Ops,Acts).
% All actions that can enable or disable an action checked for enabledness should be
% asserted as visible. This means that every action that writes the variables of action
% "a" which is checked for enabledness (e(a)) in the LTL-formula.
look_up_for_visible_actions(enabled(bop(Name,_NArgs,_NRes,_Args,_Outputs)),Ops,Acts) :- !,
    b_get_read_write(Name,GuardReads,_ActionReads,_Reads,_Writes),
    list_to_ord_set(GuardReads,Vars),
    get_actions_modifying_variables(Vars,Ops,Acts).
look_up_for_visible_actions(Term,_Ops,[]) :-
    (atomic(Term);var(Term)),!. % can Term be a variable???
% the recursive clause
look_up_for_visible_actions(Term,Ops,Actions) :-
    compound(Term),!,
    functor(Term,_Functor,N),
    look_up_for_visible_actions_args(N,Term,Ops,[],Actions).

look_up_for_visible_actions_args(0,_,_,Actions,Actions) :- !.
look_up_for_visible_actions_args(N,Term,Ops,Acts_so_far,Actions) :-
    N>0,!,
    arg(N,Term,Arg),
    look_up_for_visible_actions(Arg,Ops,Acts),
    append(Acts_so_far,Acts,Acts_so_far1),
    N1 is N-1,
    look_up_for_visible_actions_args(N1,Term,Ops,Acts_so_far1,Actions).

get_actions_modifying_variables(Vars,Ops,Actions) :-
    maplist(variable_modified_by_action(Vars),Ops,Acts),
    append(Acts,Actions).

variable_modified_by_action(Vars,Op,Res) :-
    b_get_read_write(Op,_Read,Write),
    (ord_intersect(Vars,Write)
        -> Res=[Op]
        ;  Res=[]
    ).

/*
check_action_membership(Name,Ops) :-
    (member(Name,Ops)
        -> true
        ;  add_error_fail(ltl_por,'Action not found in the currently loaded machine: ',Name/Ops)).
*/

% asserting all stutter and visible actions before starting the LTL model checker
assert_all_visible_stutter_actions(StutterActions,VisibleActions) :-
    maplist(assert_stutter_action,StutterActions),
    maplist(assert_visible_action,VisibleActions).

assert_stutter_action(Act) :-
    assert(stutter_action(Act)).

assert_visible_action(Act) :-
    assert(visible_action(Act)).
%*******************************************************************************
% do some preprocessing on a formula:
%  - insert current state id
%  - rewrite CSP declarations in some cases

preprocess_formula(In,Out) :-
    ( preprocess_formula2(In,Out) -> true
    ; otherwise ->
        add_error(ltl,'preprocessing formula failed: ',In),fail).

% base cases
preprocess_formula2(ap(A),ap(B)) :- !,
    (A=bpred(Pred) -> 
        predicate_identifiers(Pred,Ids),
        debug_println(predicate_identifiers(A,Ids))
    ; true),
    preprocess_ap(A,B).
preprocess_formula2(action(A),action(B)) :- !,preprocess_tp(A,B).
preprocess_formula2(ena(SpecA,FA),action(SpecB,FB)) :-
    !,preprocess_tp(SpecA,SpecB),preprocess_formula2(FA,FB).
% idempotency laws
preprocess_formula2(globally(globally(F)),globally(PF)) :-
    !,preprocess_formula2(F,PF).
preprocess_formula2(finally(finally(F)),finally(PF)) :-
    !,preprocess_formula2(F,PF).
preprocess_formula2(until(F,until(F,G)),until(PF,PG)) :-
    !,preprocess_formula2(F,PF),
      preprocess_formula2(G,PG).
preprocess_formula2(until(until(F,G),G),until(PF,PG)) :-
    !,preprocess_formula2(F,PF),
      preprocess_formula2(G,PG).
% absorption laws
preprocess_formula2(finally(globally(finally(F))),globally(finally(PF))) :-
    !,preprocess_formula2(F,PF).
preprocess_formula2(globally(finally(globally(F))),finally(globally(PF))) :-
    !,preprocess_formula2(F,PF).
% otherwise
preprocess_formula2(Term,NewTerm) :-
    functor(Term,Functor,Arity),
    functor(NewTerm,Functor,Arity),
    preprocess_formula_args(Arity,Term,NewTerm).

preprocess_formula_args(0,_,_) :- !.
preprocess_formula_args(N,Term,NewTerm) :-
    arg(N,Term,Arg),
    arg(N,NewTerm,NewArg),
    preprocess_formula2(Arg,NewArg),
    N2 is N-1,
    preprocess_formula_args(N2,Term,NewTerm).

preprocess_ap(current,stateid(StateId)) :- !,current_state_id(StateId).
preprocess_ap(enabled(A),enabled(B)) :- !,preprocess_tp(A,B).
preprocess_ap(AP,AP).

preprocess_tp(csp(Channel,Arity,CspArgs),
              bop(Channel,NumberArgs,NumberResults,ArgPatterns,ResPatterns)) :-
    csp_with_bz_mode,
    b_get_machine_operation(Channel,Results,Parameters,_Body),
    !,
    % check if more arguments are specified that allowed
    length(Parameters,NumberArgs),
    length(Results,NumberResults),
    MaxNumberAllowed is NumberArgs + NumberResults,
    ( Arity > MaxNumberAllowed ->
        add_error(ltl,'Too many arguments for LTL/CTL channel: ', Channel),fail
    ; Arity =< NumberResults ->
        ParamArgs = CspArgs,ResArgs = []
    ; otherwise ->
        MinArgs is min(NumberArgs,Arity),
        append_length(ParamArgs,ResArgs,CspArgs,MinArgs)),
    csp_arguments(ParamArgs,1,Parameters,ArgPatterns,Exprs1),
    csp_arguments(ResArgs,1,Results,ResPatterns,Exprs2),
    append(Exprs1,Exprs2,Exprs), 
    parse_preds_exprs(Exprs).
preprocess_tp(TP,TP).

csp_arguments([],_Index,_Parameters,[],[]).
csp_arguments([P|ParamArgs],Index,[Param|ParamRest],ArgPatterns,Exprs) :-
    ( is_csp_placeholder(P) ->
        ArgPatterns = ArgPatternsRest,
        Exprs = ExprsRest
    ; otherwise ->
        ArgPatterns = [Index/ArgPattern|ArgPatternsRest],
        Exprs = [Expr|ExprsRest],
        csp_argument(P,Param,ArgPattern,Expr)),
    I2 is Index+1,
    csp_arguments(ParamArgs,I2,ParamRest,ArgPatternsRest,ExprsRest).
csp_argument(Argument,Parameter,ParsedExpr,expr(ToParse,Type,ParsedExpr)) :-
    get_texpr_type(Parameter,Type),
    atom_codes(Argument,ToParse).

is_csp_placeholder('_').
is_csp_placeholder('?').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% callback from LTLc:

% cltl_callback(+Node, -Aps, -Out, +Queries, +Specs):
%  Callback predicate that evaluates for a state "Node" all atomic
%  propositions and returns a list of outgoing transitions.
%  It evaluates if each of the transitions met the propositions on transistions.
%  "Queries" is a list of atomic propositions (ap), "Aps" is a lists where
%  each entry is 0 resp 1 if the corresponding atomic proposition is true resp. false.
%  "Specs" is a list of propositions on transitions (tp).
%  "Out" is a list of outgoing transitions, for details, see below check_tps/4.
:- public cltl_callback/5.
cltl_callback(Node, Aps, Out, Queries, Specs) :- 
    check_aps(Queries,Node,Aps),      % generate a list of 0s and 1s, depending on whether the AP is met or not
    impl_trans_term_all1(Node,Trans), % find all outgoing transitions
    check_tps(Trans,Specs,Node,Out),  % evaluate the TP on each transition and generate a list of transitions, each
                                      % annotated with 0s and 1s, depending on whether the TP is met or not
    debug_println(9,cltl_callback(Node, Aps, Out, Queries, Specs)).

impl_trans_term_all1(Node,Trans) :-
    % In case that sink/deadlock is checked or the e(-) predicate is used in the LTL formula,
    % the state will be completely explored by execution of the check_aps/3 predicate (see implementation of trans/3).
    ( (preferences: preference(use_por_for_ltl,true), animation_mode(b))
        -> user: compute_ample_set(Node,Trans), % predicate defined in tcltk_interface.pl
           debug_println(9,compute_ample_set(Node,Trans))
        ;  impl_trans_term_all(Node,Trans)
    ).

% check list of atomic propositions (ap)
check_aps([],_,[]).
check_aps([Query|QRest],Node,[Result|RRest]) :-
    (check_ap(Query,Node) -> !,Result=1; Result=0),
    check_aps(QRest,Node,RRest).

% check list of transition properties:
% check_tps(+Transitions, +Specifications, +State, -Result):
%  "Transitions" is a list of terms of the form op(Id,Operation,NextState)
%          each term represents a transition from "State" to "NextState",
%          with the id "Id" and operation "Operation"
%  "Specifications" is a list of propositions on transistions (tp)
%  "Result" is a list where each element corresponds to a transition in "Transitions"
%          each entry has the form "trans(Next,Id,ResVector)" where "Next" and "Id"
%          are the destination state and the id of the transition and
%          "ResVector" a list where each element corresponds to a proposition (tp)
%          in "Specifications". It is 0 resp. 1 if the proposition is false resp. true.
check_tps([],_,_,[]).
check_tps([op(Id,Op,Next)|OpRest],Specs,Node,[trans(Next,Id,Tp)|TpRest]) :-
    check_tps2(Specs,Node,Next,Op,Tp),
    check_tps(OpRest,Specs,Node,TpRest).
check_tps2([],_,_,_,[]).
check_tps2([Spec|SRest],SrcNode,DstNode,Op,[Result|RRest]) :-
    (check_transition(Spec,Op,SrcNode,DstNode) -> !,Result=1; Result=0),
    check_tps2(SRest,SrcNode,DstNode,Op,RRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

ltl_formula_structure(Formula,Atomics,Structure) :-
    c_ltl_aptp(Formula,ApTp),
    pp_formulas(ApTp,Atomics),
    ltl_structure(Formula,ApTp,Structure).

ltl_structure(fairnessimplication(_FAssumptions,F),Subformulas,Structure) :- !,
    ltl_structure(F,Subformulas,Structure).
ltl_structure(ap(AP),Subformulas,ap(N)) :-
    !,nth0(N,Subformulas,ap(AP)),!.
ltl_structure(action(TP),Subformulas,tp(N)) :-
    !,nth0(N,Subformulas,action(TP)),!.
ltl_structure(F,ApTp,N) :-
    F =.. [Functor|Args],
    same_length(Args,NewArgs),
    N =.. [Functor|NewArgs],
    ltl_structure_l(Args,ApTp,NewArgs).
ltl_structure_l([],_,[]).
ltl_structure_l([F|Frest],ApTp,[N|Nrest]) :-
    ltl_structure(F,ApTp,N),
    ltl_structure_l(Frest,ApTp,Nrest).

ltl_model_check_with_ce(Formula,Max,Mode,Result) :-
    ltl_model_check3(Formula,Max,Mode,R1),
    ce_result(R1,Formula,Result).

ce_result(nostart,_,nostart) :- !.
ce_result(none,_,ok) :- !.
ce_result(incomplete,_,incomplete) :- !.
ce_result(model(CE,LoopEntry),Formula,counterexample(Counterexample,LoopEntry,PathToCE)) :- !,
    ce_result_initpath(CE,PathToCE),
    c_ltl_aptp(Formula,ApTp),
    length(ApTp,N),
    convert_counterexample(CE,N,Counterexample).
ce_result(Result,_,_) :-
    add_error(ltl,'Internal error: Unexpected result: ', Result),
    fail.

ce_result_initpath(CE,Path) :-
    ce_get_first_state(CE,InitState),
    find_init(root,InitState,StatePath),
    find_operation_ids(StatePath,root,Path).
find_operation_ids([],_,[]).
find_operation_ids([Dst|Srest],Src,[(Id,Op,Dst)|Irest]) :-
    transition(Src,Op,Id,Dst),
    find_operation_ids(Srest,Dst,Irest).

/* just restrict the evaluation bits to the atomic formulas */
convert_counterexample([],_N,[]).
convert_counterexample([atom(State,Evals,NextOp) |Arest],N,
                       [atom(State,NEvals,OpTuple)|Erest]) :-
    prefix_length(Evals,NEvals,N),
    ( NextOp = none          -> OpTuple = none
    ; otherwise              -> transition(State,Action,NextOp,Dst),OpTuple=(NextOp,Action,Dst)),
    convert_counterexample(Arest,N,Erest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% verify a counter-example

callback_tp(TP,StateId,TransId) :-
    ( transition(StateId,Op,TransId,DestId) ->
        check_transition(TP,Op,StateId,DestId)
    ; otherwise ->
        add_error(ltl,'Internal error: callback_tp called for unknown transition: ',
                  state_trans(StateId,TransId)),
        fail
    ).
      
call_ltl_verification(model(Atoms,Status),Formula) :-
    !,call_ltl_verification2(Status,Atoms,Formula).
call_ltl_verification(_,_Formula).
call_ltl_verification2(Status,Atoms,Formula) :-
    maplist(convert_atom_to_statetransition,Atoms,StateTransitions),
    ( evaluate_ltl_formula(Formula,StateTransitions,Status,
                           check_ap,callback_tp,Result) ->
        ( Result = true -> true
        ; otherwise ->
            add_error(ltl,'Evaluation of LTL yield unexpected result: ',Result))
    ; otherwise ->
        add_error(ltl,'Evaluation of LTL formula failed: ',Formula)).

convert_atom_to_statetransition(atom(NodeId,_,TransId),strans(NodeId,TransId)).


% fairness
call_ltl_fairness_verification(model(Atoms,Status),WeakDNF,StrongDNF) :-
    !,maplist(convert_atom_to_statetransition,Atoms,ST),
    call_ltl_fairness_verification1(WeakDNF,weak,ST,Status),
    call_ltl_fairness_verification1(StrongDNF,strong,ST,Status).
call_ltl_fairness_verification(_,_WeakDNF,_StrongDNF).

call_ltl_fairness_verification1([],_Type,_ST,_Status) :- !.
call_ltl_fairness_verification1(DNF,Type,ST,Status) :-
    ( call_ltl_fairness_verification2(DNF,Type,ST,Status,Result) ->
        true
    ; otherwise ->
        add_error(ltl,'Evalutation of LTL fairness failed: ',Type/DNF)),
    ( Result = fair -> true
    ; otherwise ->
        print(Type/DNF),nl,
        add_error(ltl,'Evaluation of LTL yield that the result violates fairness: ',Type/Result)).

ev_fairness_id(Spec,StateId,TransId,Op) :-
    transition(StateId,Spec,TransId,DstId),
    check_transition(Spec,Op,StateId,DstId).

call_ltl_fairness_verification2([],_Type,_ST,_Status,unfair([])).
call_ltl_fairness_verification2([Clause|CRest],Type,ST,Status,Result) :-
    call_ltl_fairness_verification3(Clause,Type,ST,Status,CResult),
    ( CResult = fair -> Result=fair
    ; CResult = unfair(Cause) ->
        call_ltl_fairness_verification2(CRest,Type,ST,Status,RResult),
        ( RResult = fair           -> Result=fair
        ; RResult = unfair(Causes) -> Result=unfair([Cause|Causes]))).
call_ltl_fairness_verification3([],_Type,_ST,_Status,fair).
call_ltl_fairness_verification3([Spec|SRest],Type,ST,Status,Result) :-
    evaluate_ltl_fairness(Type,ST,Status,ev_fairness_id(Spec),LResult),
    ( LResult = fair ->
        call_ltl_fairness_verification3(SRest,Type,ST,Status,Result)
    ; LResult = unfair(Id) ->
        Result = unfair(Id)).
