% (c) 2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(ltl_fairness,[initialise_fairness_assumptions/4,ltl_model_check_core_with_fairness/5]).

:- use_module(probsrc(module_information)).
:- module_info(group,ltl).
:- module_info(description,'This module provides LTL fairness checking on models.').

% SICSTUS libraries
:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(library(codesio)). % for read_from_codes/2

% ProB libraries
:- use_module(probsrc(self_check)).
:- use_module(probsrc(error_manager)).
:- use_module(probsrc(debug)).
:- use_module(probsrc(state_space),[transition/4]).

% import the low-level C-library wrapper
:- use_module(extension('ltlc/ltlc')).

% We allow that more than one action can be given in a single LTL formula 
% for checking the found SCC component for fairness in regard to those actions.
% There are two types of fairness which we want to support for the LTL Model Checker: strong and weak fairness. 

% We use the operators SF(-) and WF(-) for picking actions for the fairness assumption. If we want to check 
% an LTL formula "f" by setting strong fairness constraints for the action a, we write then "SF(a) => f".
% It is also possible to express fairness constraints for more than one actions by using the boolean operators
% & and or.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ltl_model_check_core_with_fairness([],_Ltl,_WeakFairnessArgs,_StrongFairnessArgs,nostart) :- !.
ltl_model_check_core_with_fairness(Startnodes,Ltl,WeakFairnessArgs,StrongFairnessArgs,Result) :-
    % it is impossible that both WeakFairnessArgs are empty lists, otherwise the predicate ltl_model_check_core/3 will be called
    ( WeakFairnessArgs = []
      -> c_ltl_modelcheck(Ltl, Startnodes, Result, ltl:cltl_callback, none, 
                      ltl_fairness:get_transition_ids, [],StrongFairnessArgs)
    ; StrongFairnessArgs = []
      -> c_ltl_modelcheck(Ltl, Startnodes, Result, ltl:cltl_callback, none, 
                      ltl_fairness:get_transition_ids, WeakFairnessArgs,[])
    ; otherwise
      -> c_ltl_modelcheck(Ltl, Startnodes, Result, ltl:cltl_callback, none, 
                      ltl_fairness:get_transition_ids, WeakFairnessArgs,StrongFairnessArgs)).


%%%%%%%%%%%%%%%%%%%% Initialising the fairness assumptions (begin) %%%%%%%%%%%%%%%%%%%%%
% temporary implemented prediates for testing the fairness check implementation.

% if you want to check if a trace is strongly fair w.r.t. multiple actions (e.g. Req2 and Req1)
% then conjugate these with the & operator and place every action between two braces

% in case fairness should be ignored, just give an empty atom ''.

initialise_fairness_assumptions(Ltl,WeakFairnessArgs,StrongFairnessArgs,LtlFormula) :-
  (Ltl = fairnessimplication(FairnessFormula,LtlFormula) ->
      get_strong_weak_fairness_arguments(FairnessFormula,WeakFairnessArgs,StrongFairnessArgs),
      debug_println(9,get_strong_weak_fairness_arguments(FairnessFormula,WeakFairnessArgs,StrongFairnessArgs))
  ; otherwise ->
      WeakFairnessArgs = [], StrongFairnessArgs = [], LtlFormula = Ltl).

get_strong_weak_fairness_arguments(FairnessFormula,WeakFairnessArgs,StrongFairnessArgs) :-
  (FairnessFormula = and(strongassumptions(SAssumptions),weakassumptions(WAssumptions))
      -> get_fairness_assumption_actions(SAssumptions,StrongFairnessArgs),
         get_fairness_assumption_actions(WAssumptions,WeakFairnessArgs)
  ; FairnessFormula = and(weakassumptions(WAssumptions),strongassumptions(SAssumptions))
      -> get_fairness_assumption_actions(SAssumptions,StrongFairnessArgs),
         get_fairness_assumption_actions(WAssumptions,WeakFairnessArgs)
  ; FairnessFormula = andfair1(SAssumptions,WAssumptions)
      -> get_fairness_assumption_actions(SAssumptions,StrongFairnessArgs),
         get_fairness_assumption_actions(WAssumptions,WeakFairnessArgs)
  ; FairnessFormula = andfair2(WAssumptions,SAssumptions)
      -> get_fairness_assumption_actions(SAssumptions,StrongFairnessArgs),
         get_fairness_assumption_actions(WAssumptions,WeakFairnessArgs)
  ; FairnessFormula = weakassumptions(F)
      -> get_fairness_assumption_actions(F,WeakFairnessArgs), StrongFairnessArgs =[]
  ; FairnessFormula = strongassumptions(F)
      -> WeakFairnessArgs = [], get_fairness_assumption_actions(F,StrongFairnessArgs)
  ; otherwise
      -> add_internal_error('Unrecognised fairness assumption: ', FairnessFormula)
  ).

get_fairness_assumption_actions(F,Res) :-
   normalise_to_dnf(F,DNF),
   get_or_args(DNF,R),
   list_to_ord_set(R,Res).

:- assert_must_succeed((normalise_to_dnf(and(or(ap(a),ap(b)),ap(c)),R), R == or(and(ap(a),ap(c)),and(ap(b),ap(c))))).
:- assert_must_succeed((normalise_to_dnf(and(ap(a),ap(b)),R), R == and(ap(a),ap(b)))).
:- assert_must_succeed((normalise_to_dnf(or(ap(a),ap(b)),R), R == or(ap(a),ap(b)))).
:- assert_must_succeed((normalise_to_dnf(ap(a),R), R == ap(a))).
:- assert_must_succeed((normalise_to_dnf(and(or(ap(a),ap(b)),or(ap(c),ap(d))),R), 
                   R == or(or(and(ap(a),ap(c)),and(ap(a),ap(d))),or(and(ap(b),ap(c)),and(ap(b),ap(d)))) )).

normalise_to_dnf(F,DNF) :-
   ex2disj(F,DNF).

ex2disj(or(A,B),or(AR,BR)) :- !,
   ex2disj(A,AR),
   ex2disj(B,BR).
ex2disj(and(A,B),R) :- !,
   ex2disj(A,AR),
   ex2disj(B,BR),
   join(AR,BR,R).
ex2disj(ap(X),ap(X)) :- !.

join(or(A,B),C,or(AC,BC)) :- !,
   join(A,C,AC),
   join(B,C,BC).
join(C,or(A,B),or(AC,BC)) :- 
   C\=or(_,_),!,
   join(C,A,AC),
   join(C,B,BC).
join(A,B,and(A,B)) :-
   A\=or(_,_),B\=or(_,_),!.

:- use_module(probltlsrc(ltl), [preprocess_tp/2]).
% left argument of get_and_args/2 is a boolean formula in DNF without any not(-) operators
get_or_args(ap(strong_fair(A)),[[R]]) :- !, ltl: preprocess_tp(A,R).
get_or_args(ap(weak_fair(A)),[[R]]) :- !, ltl: preprocess_tp(A,R).
%get_or_args(ap(A),[[A]]) :- atom(A),!.
get_or_args(and(A,B),Args) :- !,
   get_and_args(and(A,B),Res),
   list_to_ord_set(Res,Res1),
   Args = [Res1].
get_or_args(or(A,B),Res) :- !,
   get_or_args(A,R1),
   get_or_args(B,R2),
   append(R1,R2,Res).

get_and_args(ap(strong_fair(A)),[R]) :- !, ltl: preprocess_tp(A,R).
get_and_args(ap(weak_fair(A)),[R]) :- !, ltl: preprocess_tp(A,R).
%get_and_args(ap(A),[A]) :- atom(A),!.
get_and_args(and(A,B),Res) :- !,
   get_and_args(A,R1),
   get_and_args(B,R2),
   append(R1,R2,Res).
%%%%%%%%%%%%%%%%%%%% Initialising the fairness assumptions (end) %%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%% Callback predicates (begin) %%%%%%%%%%%%%%%%%%%%%
% Callback from LTLc (fairness):

% check_fairness(FairnessAssumption,[anode(AtomId,Nodeid,[trans(NextAtomId,NextNodeId,[TransitionId])])])
% FairnessAssumtion is the Fairness Condition of the LTL formula.
% For instance, if we have the LTL formula "SF ([a] & [b]) => Phi" then FairnessAssumption consists of the actions [a,b]

% Tested only for B and CSP models

get_transition_ids(Arg,StateIds,LTransIds) :-
  get_transition_ids1(Arg,StateIds,LTransIds),
  debug_println(9,get_transition_ids(fainress_argument(Arg),state_ids(StateIds),transition_ids(LTransIds))).

get_transition_ids1(Arg,StateIds,LTransIds) :-
  findall(TransId,(member(NodeId,StateIds),(transition(NodeId,Op,TransId,_Dst),ltl:check_transition(Arg,Op,_,_))),LTransIds).

%%%%%%%%%%%%%%%%%%%% Callback predicates (end) %%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%% Predicates for checking whether the given SCC is a possible counter example for a checked Ltl formula %%%%%%%%%%%%%%%%

check_scc_ce(LtlFormula,SCC) :-
  Max = 10000, Mode = init,
  ( ltl: temporal_parser(LtlFormula,ltl,Ltl) ->
      true
  ; otherwise ->
      add_error_fail(ltl,'LTL Parser failed: ',LtlFormula)),
  ltl: ltl_model_check3(Ltl,Max,Mode,Status),
  clean_scc(Status,Result),
  atom_codes(SCC,Codes),
  read_from_codes(Codes,SCCTerm),
  (match_ce_result(SCCTerm,Result) ->
      true
  ; otherwise ->
      print(result_was(Result)),nl,
      add_error(ltl_scc_check,'No one of the SCCs could be reproduced as counter example: ',SCC)
  ).

% predicates for removing all unnecessary information like atomids and ap/tp evaluations
clean_scc(model(Trace,Entry),Result) :-
  maplist(get_nodeid,Trace,CleanedTrace),
  Result = model(CleanedTrace,Entry).

get_nodeid(atom(StateId,_Evaluation,_AtomId),StateId).
get_nodeid(El,_) :- add_internal_error('Unrecognised trace element: ', El).

match_ce_result([],_) :- !,fail.
match_ce_result([H|T],Result) :-
  (H = Result -> true; match_ce_result(T,Result)).
