% (c) 2009-2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html

:- module(ltl_tools,[temporal_parser/3,temporal_parser_l/3,
                     parse_ltlfile/2,check_csp_event/4,
                     parse_and_pp_ltl_formula/2, pp_ltl_formula/2, pp_formulas/2,
                     typecheck_temporal_formula/3,extract_formulas/4]).

/* SICSTUS Libraries */
:- use_module(library(lists)).
:- use_module(library(codesio)).

/* ProB Libraries */
:- use_module(probsrc(error_manager)).
:- use_module(probsrc(debug)).
:- use_module(probsrc(self_check)).
:- use_module(probsrc(parsercall),[call_ltl_parser/3]).
:- use_module(probsrc(bmachine),[b_type_expression/5, b_get_machine_operation/4]).
:- use_module(probsrc(plugins),[plugin_parse/7, active_plugin/1]).
:- use_module(probsrc(specfile),[animation_mode/1,csp_mode/0]).
:- use_module(probsrc(tools),[split_atom/3, exact_member/2, string_concatenate/3, ajoin/2]).
:- use_module(probcspsrc(haskell_csp),[channel/2, channel_type_list/2,parse_single_csp_expression/3]).
:- use_module(probsrc(bsyntaxtree),[get_texpr_type/2]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,ltl).
:- module_info(description,'This module provides utility predicates for the LTL and CTL model checkers.').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% parse an LTL or CTL formula
% Mode: ltl or ctl
temporal_parser(Formula,Mode,Result) :-
    temporal_parser_l([Formula],Mode,[Result]).

temporal_parser_l(Formulas,Mode,Result) :-
    % call the Java parser for LTL
    call_ltl_parser(Formulas,Mode,Res1),
    % check if everything was fine
    check_syntax_result_l(Res1,Parsed,ParseFlag,Mode), ParseFlag=ok,
    % typecheck atomic propositions and such things
    typecheck_temporal_formula_l(Parsed,Result,TypeFlag), TypeFlag=ok.

check_syntax_result_l([],[],_,_).
check_syntax_result_l([R|Rrest],[F|Frest],Res,Mode) :-
    check_syntax_result(R,F,Res,Mode),
    check_syntax_result_l(Rrest,Frest,Res,Mode).
check_syntax_result(ltl(F),F,_,_Mode).
check_syntax_result(syntax_error(E),_,error,Mode) :-
    add_error(Mode,'Syntax error: ',E).

typecheck_temporal_formula_l([],[],_).
typecheck_temporal_formula_l([F|Frest],[R|Rrest],Res) :-
    typecheck_temporal_formula(F,R,Res),
    typecheck_temporal_formula_l(Frest,Rrest,Res).
typecheck_temporal_formula(ap(AP),ap(Typed),Res) :-
    !,typecheck_ap(AP,Typed,Res).
typecheck_temporal_formula(action(TP),action(Typed),Res) :-
    !,typecheck_tp(TP,Typed,Res).
typecheck_temporal_formula(ena(TP,F),ena(Typed,R),Res) :-  % used in CTL only
    !,typecheck_tp(TP,Typed,Res),
    typecheck_temporal_formula(F,R,Res).
typecheck_temporal_formula(F,R,Res) :-
    F =.. [Functor|InArgs],
    same_length(InArgs,OutArgs),
    R =.. [Functor|OutArgs],
    typecheck_temporal_formula_l(InArgs,OutArgs,Res).

typecheck_ap(BasicAP,AP,_Res) :- is_basic_ap(BasicAP),!,AP=BasicAP.
typecheck_ap(enabled(TP),enabled(R),Res) :-
    !,typecheck_tp(TP,R,Res).
typecheck_ap(available(TP),available(R),Res) :-
    !,typecheck_tp(TP,R,Res).
typecheck_ap(weak_fair(TP),weak_fair(R),Res) :-
    !,typecheck_tp(TP,R,Res).
typecheck_ap(strong_fair(TP),strong_fair(R),Res) :-
    !,typecheck_tp(TP,R,Res).
typecheck_ap(dlk(TPList),dlk(R),Res) :-
    !,typecheck_tp_l(TPList,R,Res).
typecheck_ap(ctrl(TPList),ctrl(R),Res) :-
    !,typecheck_tp_l(TPList,R,Res).
typecheck_ap(det(TPList),det(R),Res) :-
    !,typecheck_tp_l(TPList,R,Res).
typecheck_ap(bpred(Raw),bpred(Typed),Res) :-
    b_type_expression(Raw,[variables],pred,Typed,Errors),
    add_all_perrors(Errors,[],typecheck_atomic_property),
    ( Errors == [] -> true
    ; otherwise    -> Res=error).
typecheck_ap(unparsed_pred(Atom),promela_pred(Predicate),ResultFlag) :-
    animation_mode(promela),!,
    atom_codes(Atom,Codes),
    ( parse_promela_strings([],[Codes],[],[Predicate]) ->
        true
    ; otherwise ->
        add_error(ltl, 'Failed to parse Promela predicate: ', Atom),
        ResultFlag = error).
typecheck_ap(unparsed_pred(Atom),plugin_predicate_check(Id,Predicate),ResultFlag) :-
    active_plugin(Id),!,
    atom_codes(Atom,Codes),
    ( plugin_parse(Id,[],[Codes],[],[],[Predicate],[]) -> true
    ; otherwise                                        -> ResultFlag=error).

is_basic_ap(deadlock).
is_basic_ap(sink).
is_basic_ap(current).
is_basic_ap(stateid(_)).

typecheck_tp(unparsed_trans(Atom),plugin_transition_check(Id,Predicate),ResultFlag) :-
    active_plugin(Id),!,
    atom_codes(Atom,Codes),
    ( plugin_parse(Id,[],[],[Codes],[],[],[Predicate]) -> true
    ; otherwise                                        -> ResultFlag=error).
typecheck_tp(unparsed_trans(Op),xtl(Opname,Arity,Args),ResultFlag) :-
    animation_mode(xtl), !,
    ( split_atom(Op,['.'],[Opname|Args]) -> 
        length(Args,Arity)
    ; otherwise ->
        add_error(ltl,'Failed to parse operation specification'),
        ResultFlag = error).
typecheck_tp(unparsed_trans(Opname),csp(Channel,Arity,ResArgs),ResultFlag) :-
    csp_mode,!,
    ( Opname = tick
        -> Channel=Opname, Arity=0, ResArgs=[] % should we set Arity ??
    ; Opname = tau
        -> Channel=Opname, Arity=0, ResArgs=[] % should we set Arity ??
    ; (split_atom(Opname,['.','!'],[Channel|Args]), channel(Channel,_))
        ->
         ((channel_type_list(Channel, TypeList),length(TypeList,Arity),length(Args,Arity))
             -> (Arity = 0
                 -> ResArgs = []
                 ; (\+memberchk('_',Args) % we want to call the CSP-M parser only once when there is no '_' symbol in Args
                     -> mergeArgsToDotTuple([Channel|Args],DotTupleEvent),
                        parse_single_csp_expression(ltl,DotTupleEvent,ParsedEvent),
                        splitToSingleArgs(ParsedEvent,[Channel|ResArgs],Arity)
                     ;  l_parse_single_csp_expression_undescore(ltl,Args,ResArgs)
                     )
                )
             ;  add_error(ltl,'Number of arguments not correct for: ', Opname),
                ResultFlag=error, ResArgs=[], Arity = 0
         )
    ; otherwise
        -> add_error(ltl,'Channel does not exist: ',Opname),
           ResultFlag = error, ResArgs=[], Arity = 0
    ).
typecheck_tp(bop(oppattern(_Pos,Name,Args)),Result,ResultFlag) :-
    ( b_get_machine_operation(Name,Results,Parameters,_) ->
        Result = bop(Name,NumberArgs,NumberResults,ArgPatterns,ResPatterns),
        length(Parameters,NumberArgs), length(Results,NumberResults),
        % Pattern matching for results is not yet supported
        ResPatterns = [],
        ( Args=[] ->
            ArgPatterns = [],
            ResPatterns = []
        ; length(Args,NumberArgs) ->
            create_b_op_patterns(Args,Parameters,ArgPatterns,ResultFlag)
        ; otherwise ->
            add_error(ltl,'Number of arguments not correct for: ', Name),ResultFlag=error)
    ; csp_mode, Args=[] ->
        debug_println(9,unparsed_trans(Name)),
        typecheck_tp(unparsed_trans(Name),Result,ResultFlag)
    ; otherwise ->
        add_error(ltl,'Could not find operation: ', Name),ResultFlag=error).
typecheck_tp(btrans(event(Name)),btrans(event(Name)),_ResultFlag) :- !.
typecheck_tp(btrans(event(Name,RawPredicate)),btrans(event(Name,TPredicate,Poststate)),ResultFlag) :-
    ( b_get_machine_operation(Name,_,_Parameters,_) ->
        ( bmachine:type_with_errors(RawPredicate,[prepost,operation(Name)],pred,TPredicate) ->
            find_poststate_access(TPredicate,Poststate)
        ; otherwise -> ResultFlag=error)
    ; otherwise ->
        add_error(ltl,'Could not find operation: ', Name),ResultFlag=error).

typecheck_tp_l([],[],_Res).
typecheck_tp_l([TP|TPs],[TTP|TTPs],Res) :-
    typecheck_tp(TP,TTP,ResultFlag),
    (ResultFlag == error -> Res = error
    ; otherwise -> Res=Res1),
    typecheck_tp_l(TPs,TTPs,Res1).

:- use_module(probsrc(bsyntaxtree),[syntaxtraversion/6]).
find_poststate_access(TExpr,Poststate) :-
    find_poststate_access2(TExpr,P,[]),
    sort(P,Poststate).
find_poststate_access2(TExpr) -->
    {syntaxtraversion(TExpr,Expr,_,Infos,Subs,_)},
    find_poststate_access3(Expr,Infos),
    find_poststate_access_l(Subs).
find_poststate_access3(identifier(Id),Infos) -->
    {memberchk(poststate,Infos),!,atom_concat(Orig,'\'',Id)},
    [poststate(Orig,Id)].
find_poststate_access3(_,_) --> [].
find_poststate_access_l([]) --> !.
find_poststate_access_l([TExpr|Rest]) -->
    find_poststate_access2(TExpr),
    find_poststate_access_l(Rest).
        
/* Parsing CSP expressions where '_' will be not parsed */
parse_single_csp_expression_undescore(Context,Expr,Res) :-
    (Expr=='_' -> Res = Expr; parse_single_csp_expression(Context,Expr,Res)).

l_parse_single_csp_expression_undescore(Context,LExpr,LRes) :-
    maplist(parse_single_csp_expression_undescore(Context),LExpr,LRes).
/*******************************************************/

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% parse a file with several LTL formulas:

parse_ltlfile(Filename, Formulas) :-
    ( parse_ltlfile2(Filename,Formulas) -> true
    ; otherwise ->
        add_error(ltl, 'Error while parsing the LTL formula file')).
parse_ltlfile2(Filename, Formulas) :-
    read_file(Filename,Codes),
    parse_lines(Lines,Codes,[]),
    parse_sections(Raws,Lines,[]),
    parse_file_formulas(Raws,Formulas).
parse_file_formulas(In,Out) :-
    extract_formulas(In,Strings,Parsed,Out),
    temporal_parser_l(Strings,ltl,Parsed).

extract_formulas([],[],[],[]).
extract_formulas([formula(Name,Raw)|Rin],[Raw|Rraw],
                 [Parsed|Rparsed],[formula(Name,Parsed)|Rout]) :-
    extract_formulas(Rin,Rraw,Rparsed,Rout).

/*
parse_file2([]) --> [].
parse_file2([F|Rest]) -->
    parse_section(F),
    parse_file2(Rest). */

parse_sections([formula(Name,F)|Rest]) -->
    munch_comments,
    [header(Name)],
    parse_section_formulas("",FCodes),!,
    { atom_codes(F,FCodes) },
    parse_sections(Rest).
parse_sections([]) --> [].

munch_comments --> [comment(_)],!,munch_comments.
munch_comments --> [].

parse_section_formulas(Prev,F) -->
    munch_comments,
    [formula(Snippet)],!,
    {(Prev==[] -> S2=Snippet; append("\n",Snippet,S2) ),append(Prev,S2,F2)},
    parse_section_formulas(F2,F).
parse_section_formulas(F,F) --> munch_comments,[].

parse_lines([]) --> [].
parse_lines([L|Rest]) -->
    parse_line(L),
    parse_lines(Rest).

parse_line(comment(C)) -->
    pl_spaces,
    "#",!,
    pl_until_newline(C).
parse_line(comment("")) -->
    pl_spaces,
    pl_newline,!.
parse_line(header(Title)) -->
    pl_title(Title),
    pl_spaces, pl_newline, !.
parse_line(formula([C|F])) -->
    % no empty formulas
    [C],
    pl_until_newline(F).

pl_title(Title) -->
    "[", pl_spaces, pl_title2(T), {atom_codes(Title,T)}.
pl_title2([]) --> pl_spaces,"]",!.
pl_title2([]) --> pl_newline,!,{fail}. % maybe an exception?
pl_title2([C|Rest]) --> [C],pl_title2(Rest).

pl_spaces --> " ",!,pl_spaces.
pl_spaces --> "\t",!,pl_spaces.
pl_spaces --> [].

pl_until_newline([]) --> pl_newline,!.
pl_until_newline([C|Rest]) --> [C],!,pl_until_newline(Rest).
pl_until_newline([]) --> [].

pl_newline --> "\n\r",!.
pl_newline --> "\n",!.
pl_newline --> "\r".

read_file(Filename,Codes) :-
    open(Filename,read,S),
    read_codes(S,Codes),
    close(S).
read_codes(S,Codes) :-
    get_code(S,Code),
    ( Code < 0 ->
        Codes = []
    ; otherwise ->
        Codes = [Code|Rest],
        read_codes(S,Rest)).

create_b_op_patterns(Args,Parameters,ArgPatterns,Res) :-
    create_b_op_patterns2(Args,Parameters,ArgPatterns,1,Res).
create_b_op_patterns2([],[],[],_,_).
create_b_op_patterns2([Arg|ARest],[Param|ParamRest],Patterns,Index,Res) :-
    create_b_op_pattern(Arg,Param,Pattern,Res),
    ( var(Pattern) -> Patterns = PRest
    ; otherwise -> Patterns = [Index/Pattern|PRest]),
    Index2 is Index+1,
    create_b_op_patterns2(ARest,ParamRest,PRest,Index2,Res).
create_b_op_pattern(undef(_Pos),_,_,_).
create_b_op_pattern(def(_Pos,Raw),TParam,Typed,Res) :-
    get_texpr_type(TParam,Type),
    b_type_expression(Raw,[variables],Type,Typed,Errors),
    add_all_perrors(Errors,[],typecheck_operation_argument),
    ( Errors == [] -> true
    ; otherwise    -> Res=error).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/* Pretty priner predicates */
:- public pp_op/2.
pp_op(Id,PP) :-
    transition(_Src,Action,Id,_Dst),!,
    translate:translate_event(Action,PP).
    
pp_formulas([],[]).
pp_formulas([ApTp|Rest],[PP|PPrest]) :-
    ( pp_formula(ApTp,PP) -> true
    ; otherwise -> PP = '???'),
    pp_formulas(Rest,PPrest).
pp_formula(ap(AP),ap(PP)) :-
    pp_ap(AP,PP).
pp_formula(action(AP),tp(PP)) :-
    pp_tp(AP,PP).

pp_ap(bpred(B),PP) :- !,
    translate:translate_bexpression(B,PP1),
    ajoin(['{',PP1,'}'],PP).
pp_ap(enabled(TP),PP) :- !,
    pp_tp2(TP,TPP),
    ajoin(['e(',TPP,')'],PP).
pp_ap(weak_fair(TP),PP) :- !,
    pp_tp2(TP,TPP),
    ajoin(['WF(',TPP,')'],PP).
pp_ap(strong_fair(TP),PP) :- !,
    pp_tp2(TP,TPP),
    ajoin(['SF(',TPP,')'],PP).
pp_ap(dlk(L),PP) :- !,
    maplist(pp_tp2,L,S),
    tools: ajoin_with_sep(S,',',SText),
    ajoin(['deadlock(',SText,')'],PP).
pp_ap(ctrl(L),PP) :- !,
    maplist(pp_tp2,L,S),
    tools: ajoin_with_sep(S,',',SText),
    ajoin(['controller(',SText,')'],PP).
pp_ap(det(L),PP) :- !,
    maplist(pp_tp2,L,S),
    tools: ajoin_with_sep(S,',',SText),
    ajoin(['deterministic(',SText,')'],PP).
pp_ap(Expr,PP) :-
    write_to_codes(Expr,Codes),
    atom_codes(PP,Codes).

pp_tp(T,PP) :-
    pp_tp2(T,PP1),
    ajoin(['[',PP1,']'],PP).
pp_tp2(btrans(event(Name)),Name) :- !.
pp_tp2(btrans(event(Name,Predicate,_)),PP) :- !,
    translate:translate_bexpression(Predicate,PP1),
    ajoin([Name,' | ',PP1],PP).
pp_tp2(csp(tau,_Arity,[]),PP) :-  !, PP = tau.
pp_tp2(csp(tick,_Arity,[]),PP) :- !, PP = tick.
pp_tp2(Action,PP) :-
    translate_action(Action,PP).
pp_tp2(_F,'???').

translate_action(bop(Name,_NumberArgs,_NumberResults,ArgPatterns,_ResPatterns),Transition) :- !,
    maplist(get_arg,ArgPatterns,ArgPatterns1),
    maplist(translate:translate_bexpression,ArgPatterns1,PPArgPatterns),
    Op =.. [Name|PPArgPatterns],
    Transition = Op.
/*    ( NumberResults == 0 ->
        Transition = Op
    ; otherwise ->
        maplist(translate:translate_bexpression,ResPatterns,PPResPatterns),
        Transition = '-->'(Op,PPResPatterns)
    ).
*/
translate_action(csp(Channel,_Arity,Args),PP) :-
    (specfile:csp_with_bz_mode ->
        % in this mode we need to remove 'CSP:'
        translate: translate_event2(io(Args,Channel,_SPan),[_|S],[])
    ; otherwise ->
        translate: translate_event2(io(Args,Channel,_SPan),S,[])
    ), ajoin(S,PP).

get_arg(N/Arg,Arg) :- nonvar(N),!.
get_arg(Arg,Arg).

parse_and_pp_ltl_formula(FormulaAsAtom,Text) :-
   temporal_parser(FormulaAsAtom,ltl,Ltl),
   ( pp_ltl_formula(Ltl,Text)
      -> true
   ; add_error_fail(ltl, 'Failed to pretty print LTL formula: ', Ltl)
   ).

pp_ltl_formula(Formula,Text) :-
    pp_ltl(Formula,Atoms,[]),
    ajoin(Atoms,Text).

/* Tests for pp_ltl_formula/2 */
    
:- assert_must_succeed((ltl_tools: pp_ltl_formula(implies(globally(finally(until(true,false))),true),Text), 
          Text = '(G (F ((true U false))) => true)')).

:- assert_must_succeed((ltl_tools: pp_ltl_formula(historically(not(release(yesterday(sink),next(deadlock)))),Text), 
          Text = 'H (not((Y (sink) R X (deadlock))))')).

:- assert_must_succeed((
     ltl_tools: pp_ltl_formula(implies(
                                 ap(enabled(btrans(event('A')))),
                                 or(weakuntil(
                                              ap(bpred(equal(identifier(op('x')),integer(1)))),
                                              ap(bpred(equal(identifier(op('x')),integer(2)))) ),
                                    finally(action(btrans(event('A')))))),Text),
           Text = '(e(A) => (({x = 1} W {x = 2}) or F ([A])))')).

:- assert_must_succeed((
     ltl_tools: pp_ltl_formula(implies(
                                 current,
                                 and(once(ap(bpred(equal(identifier(op('x')),integer(1))))),
                                     action(btrans(event('A',equal(identifier(op('x')),integer(2)),_))))),Text),
           Text = '(current => (O ({x = 1}) & [A | x = 2]))')).

:- assert_must_succeed((
     ltl_tools: pp_ltl_formula(fairnessimplication(
                                 and(strongassumptions(and(ap(strong_fair(btrans(event('A')))),
                                                           ap(strong_fair(btrans(event('B')))))),
                                     weakassumptions(or(ap(weak_fair(btrans(event('A')))),
                                                        ap(weak_fair(btrans(event('B'))))))),
                                    true),Text),
           Text = '(((SF(A) & SF(B))) & (WF(A) or WF(B)) => true)')).

:- assert_must_succeed((
     ltl_tools: pp_ltl_formula(fairnessimplication(
                                 and(weakassumptions(and(ap(weak_fair(btrans(event('A')))),
                                                         ap(weak_fair(btrans(event('B')))))),
                                     strongassumptions(or(ap(strong_fair(btrans(event('A')))),
                                                          ap(strong_fair(btrans(event('B'))))))),
                                    true),Text),
           Text = '(((WF(A) & WF(B))) & (SF(A) or SF(B)) => true)')).

/* Simple Pretty Printer for LTL/CTL Formulas */

pp_ltl(true,['true'|S],S).
pp_ltl(false,['false'|S],S).
pp_ltl(deadlock,['deadlock'|S],S).
pp_ltl(sink,['sink'|S],S).
pp_ltl(current,['current'|S],S).
pp_ltl(globally(F),['G ('| S],T) :- !, pp_ltl(F,S,[')'|T]).
pp_ltl(finally(F),['F ('| S],T) :- !, pp_ltl(F,S,[')'|T]).
pp_ltl(next(F),['X ('|S],T) :- !, pp_ltl(F,S,[')'|T]).
pp_ltl(yesterday(F),['Y ('|S],T) :- !, pp_ltl(F,S,[')'|T]).
pp_ltl(since(F,G),['('|S],T) :- !,
     pp_ltl(F,S,[' S '| S1]),
     pp_ltl(G,S1,[')'|T]).
pp_ltl(once(F),['O ('|S],T) :- !, pp_ltl(F,S,[')'|T]).
pp_ltl(historically(F),['H ('|S],T) :- !, pp_ltl(F,S,[')'|T]).
pp_ltl(until(F,G),['('|S],T) :- !,
     pp_ltl(F,S,[' U '| S1]),
     pp_ltl(G,S1,[')'|T]).
pp_ltl(weakuntil(F,G),['('|S],T) :- !,
     pp_ltl(F,S,[' W '| S1]),
     pp_ltl(G,S1,[')'|T]).
pp_ltl(release(F,G),['('|S],T) :- !,
     pp_ltl(F,S,[' R '| S1]),
     pp_ltl(G,S1,[')'|T]).
pp_ltl(implies(F,G),['('|S],T) :- !,
     pp_ltl(F,S,[' => '| S1]),
     pp_ltl(G,S1,[')'|T]).
pp_ltl(fairnessimplication(FairAssump,Formula),['('|S],T) :- !,
     pp_ltl_fairness(FairAssump,S,[' => '| S1]),
     pp_ltl(Formula,S1,[')'|T]).
pp_ltl(or(F,G),['('|S],T) :- !,
     pp_ltl(F,S,[' or '| S1]),
     pp_ltl(G,S1,[')'|T]).
pp_ltl(and(F,G),['('|S],T) :- !,
     pp_ltl(F,S,[' & '| S1]),
     pp_ltl(G,S1,[')'|T]).
pp_ltl(not(F),['not('| S],T) :- !, pp_ltl(F,S,[')'|T]).
pp_ltl(ap(AP),[Pred|S],S) :- !, pp_ap(AP,Pred).
pp_ltl(action(A),[Action|S],S) :- !, pp_tp(A,Action).

/* Pretty Print CTL Part */
pp_ltl(ena(F),['EX ('|S],T) :- !, pp_ltl(F,S,[')'|T]).
pp_ltl(en(F), ['EX ('|S],T) :- !, pp_ltl(F,S,[')'|T]).
pp_ltl(an(F), ['AX ('|S],T) :- !, pp_ltl(F,S,[')'|T]).
pp_ltl(eg(F), ['EG ('|S],T) :- !, pp_ltl(F,S,[')'|T]).
pp_ltl(ef(F), ['EF ('|S],T) :- !, pp_ltl(F,S,[')'|T]).
pp_ltl(ag(F), ['AG ('|S],T) :- !, pp_ltl(F,S,[')'|T]).
pp_ltl(af(F), ['AF ('|S],T) :- !, pp_ltl(F,S,[')'|T]).
pp_ltl(eu(F,G), ['( E '|S],T) :- !,
    pp_ltl(F,S,[' U '|S1]),
    pp_ltl(G,S1,[')'|T]).

/* Pretty Print Fairness Assumptions */
pp_ltl_fairness(strongassumptions(F),['('|S],T) :- !,
    pp_ltl_fairness1(F,S,[')'|T]).
pp_ltl_fairness(weakassumptions(F),['('|S],T) :- !,
    pp_ltl_fairness1(F,S,[')'|T]).
pp_ltl_fairness(and(F,G),S,T) :- !,
    pp_ltl_fairness(F,S,[' & '|S1]),
    pp_ltl_fairness(G,S1,T).
pp_ltl_fairness(andfair1(F,G),S,T) :- !,
    pp_ltl_fairness1(F,S,[' & '|S1]),
    pp_ltl_fairness1(G,S1,T).
pp_ltl_fairness(andfair2(F,G),S,T) :- !,
    pp_ltl_fairness1(F,S,[' & '|S1]),
    pp_ltl_fairness1(G,S1,T).

pp_ltl_fairness1(and(F,G),['('|S],T) :- !,
    pp_ltl_fairness1(F,S,[' & '|S1]),
    pp_ltl_fairness1(G,S1,[')'|T]).
pp_ltl_fairness1(or(F,G),S,T) :- !,
    pp_ltl_fairness1(F,S,[' or '|S1]),
    pp_ltl_fairness1(G,S1,T).
pp_ltl_fairness1(ap(AP),[Pred|S],S) :- !,
    pp_ap(AP,Pred).

%%%%%%%% Other predicates %%%%%%%%%%

:- assert_must_succeed((ltl_tools: mergeArgsToDotTuple([1,2,'<bob,alice>','{bob}'],Res), Res = '1.2.<bob,alice>.{bob}')).
:- assert_must_succeed((ltl_tools: mergeArgsToDotTuple(['(bob,alice)'],Res), Res = '(bob,alice)')).
:- assert_must_succeed(ltl_tools: mergeArgsToDotTuple([],'')).

mergeArgsToDotTuple(List,Res) :- mergeArgsToDotTuple(List,'',Res).

mergeArgsToDotTuple([],Str,Str).
mergeArgsToDotTuple([H|T],Str,Res) :-
        (T\=[] -> string_concatenate(H,'.',HRes) ; HRes=H),
         string_concatenate(Str,HRes,StrRes),mergeArgsToDotTuple(T,StrRes,Res).

splitToSingleArgs(dotTuple(List),List,Arity) :-
    N is Arity +1,
    length(List,N).
splitToSingleArgs(X,[X],0).
splitToSingleArgs(X,_L,N) :- 
    add_internal_error('Internal Error: Arity does not match for channel: ',X/N).

:- assert_must_succeed((check_csp_event(tick,0,[],R), R=tick(_))).
:- assert_must_succeed((check_csp_event(tau,0,[],R), R=tau(_))).
:- assert_must_fail(check_csp_arguments([setValue([alice,bob]),tupleExp([s,t,u])], [setValue([alice,bob]),in(na_tuple([s,t]))],some_event)).
:- assert_must_succeed(check_csp_arguments([setValue([alice,bob]),tupleExp([s,t,u])], [setValue([alice,bob]),in(na_tuple([s,t,u]))],some_event)).
:- assert_must_succeed(check_csp_arguments([setValue([bob,alice]),tupleExp([s,t,u])], [setValue([alice,bob]),in(na_tuple([s,t,u]))],some_event)).
:- assert_must_succeed(check_csp_arguments([a,'_','_'],[in(a),in(int(10)),in(int(20))],true)).

%% Predicate check csp event 
check_csp_event(tick,0,[],R) :- !, R=tick(_).
check_csp_event(tau,0,[],R) :- !, R=tau(_).
check_csp_event(Event,Arity,PatArgs,io(Vals,Event,_SrcSpan)) :- !,
    (channel(Event,_) ->
        ((channel_type_list(Event,List),length(List,Arity)) -> 
            check_csp_arguments(PatArgs,Vals,Event)
        ; otherwise -> add_error(ltl,'Internal Error: Missing/extra arguments in CSP LTL/CTL pattern: ',Event:PatArgs),fail
        )
    ; otherwise -> add_internal_error('Internal Error: Channel not defined: ',Event/Arity)).

:- block check_csp_arguments(?,-,?).
check_csp_arguments([],[],_).
check_csp_arguments([],[H|T],Event) :-
  print('* Missing arguments in CSP LTL/CTL pattern (use .?): '),print(Event),
  print(' : '),print([H|T]),nl.
check_csp_arguments([H|T],[],Event) :- 
  add_error(ltl,'Extra arguments in CSP LTL/CTL pattern: ',Event:[H|T]),fail.
check_csp_arguments([H|T],[PatH|PatT],Event) :-
  check_csp_arg(H,PatH),!,
  check_csp_arguments(T,PatT,Event).

check_csp_arg(X,Y) :- 
    var(Y),!,
    add_internal_error('Internal Error: Event argument was a variable, expected to be: ',X).
check_csp_arg('_',_Y). % underscore should match any argument
check_csp_arg(X,in(Y)) :- 
    check_csp_arg(X,Y).
check_csp_arg(X,out(Y)) :- 
    check_csp_arg(X,Y).

check_csp_arg(list(L),list(L1)) :- !,
  l_check_csp_arg(L,L1,list).
check_csp_arg(setValue(L),setValue(L1)) :- !,
  sort(L,R),
  sort(L1,R1),
  l_check_csp_arg(R,R1,setValue). % elements order is not important within set expressions
check_csp_arg(tupleExp(L),na_tuple(L1)) :- !,
  l_check_csp_arg(L,L1,na_tuple).
check_csp_arg(X,Y) :- eq_csp_val(Y,X).

eq_csp_val(int(X),int(Y)) :- !, Y=X.
eq_csp_val(X,X).

l_check_csp_arg([],[],_):- !.
l_check_csp_arg([],[H|T],Functor) :- 
    add_error_fail(ltl,'Missing elements in CSP LTL pattern: ',Functor:[H|T]).
l_check_csp_arg([H|T],[H1|T1],Functor) :- 
    check_csp_arg(H,H1),!,
    l_check_csp_arg(T,T1,Functor).
