% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(model_checker,[is_deadlocked/1, open_search/11,
  last_step_breadth_first_in_mixed_mode/0, depth_breadth_first_mode/1,
  set_depth_breadth_first_mode/1, reset_depth_breadth_first_mode/0,
  dbf_modes/3,
  reset_model_checker/0,
  expired_static_analysis_time/1,
  proof_info_model_checking_incomplete/5, model_checking_is_incomplete/6]).

:- use_module(module_information).
:- module_info(group,model_checker).
:- module_info(description,'The core model checker (does not treat LTL).').

:- use_module(b_interpreter,[state_violates_assertions/2]).
:- use_module(error_manager,[real_error_occurred/0]).

:- use_module(specfile,[animation_mode/1,csp_mode/1]).
:- use_module(tools,[print_message/1]).
:- use_module(state_space,[visited_expression/2,transition/3, operation_not_yet_covered/1]).
:- use_module(debug,[debug_println/2]).

:- use_module(ample_sets,[compute_dependendency_relation_of_all_events_in_the_model/3]).

:- dynamic depth_breadth_first_mode/1.
depth_breadth_first_mode(mixed).

dbf_modes(0,mixed,'Mixed DF/BF').
dbf_modes(1,breadth_first,'Breadth First').
dbf_modes(2,depth_first,'Depth First').

reset_depth_breadth_first_mode :- set_depth_breadth_first_mode(mixed).
set_depth_breadth_first_mode(MODE) :-
  retractall(depth_breadth_first_mode(_)),
  debug_println(9,depth_breadth_first_mode(MODE)),
  assert(depth_breadth_first_mode(MODE)).
  
reset_model_checker :- retractall(last_step_breadth_first_in_mixed_mode).
:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(reset_specification,reset_model_checker,
                           'Reset Model Checker BF/DF Info.').

/* open_search: search for errors/goals among open nodes */
/* LimitNr: max number of nodes to check, NodesAnalysed: number of nodes checked */
open_search(LimitNr,NodesAnalysed,LimitTime,Res,FindDeadlocks,FindInvViolations,
                           FindGoal,FindAssViolations,StopAtFullCoverage,PartialOrderReduction,PartialGuardsEvaluation) :-
    animation_mode(MODE),
    disable_optimisations_if_not_useful(MODE,FindDeadlocks,FindInvViolations,PartialOrderReduction,PartialGuardsEvaluation,
        PartialOrderReduction1),
    open_search_check_for_optimisation_options(MODE,FindInvViolations,PartialOrderReduction1,PartialGuardsEvaluation,Graphs),
    depth_breadth_first_mode(DFMODE),
    preferences:temporary_set_preference(do_invariant_checking,true,CHNG),
    statistics(walltime,[ReferenceTime,_]),
    get_model_check_stats(States,_Transitions,_ProcessedTotal,Percentage),
    open_search10(0,LimitNr,refstatus(ReferenceTime,Percentage,States,ReferenceTime),
                           NodesAnalysed,LimitTime,Res,FindDeadlocks,FindInvViolations,
                           FindGoal,FindAssViolations,
                           StopAtFullCoverage,PartialOrderReduction1,Graphs,MODE,DFMODE),
    preferences:reset_temporary_preference(do_invariant_checking,CHNG),
    clear_context_state.


:- use_module(library(random)).
open_search10(Nr,LimitNr,RefStatus,NodesAnalysed,LimitTime,Res,
              FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,
              SFC,PartialOrderReduction,Graphs,MODE,DF) :-
  %print(open_search(Nr,LimitNr,RefStatus)),nl,
  Nr < LimitNr,!,
  (\+ number(LimitTime)
   -> open_search_do_one_node(Nr,LimitNr,RefStatus,NodesAnalysed,LimitTime,Res,
					  FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,
					  SFC,PartialOrderReduction,Graphs,MODE,DF)
   ;  statistics(walltime,[NewTime,_]),
	  ( (Nr mod 1000 =:= 0,Nr\=0)
		 ->  RefStatus = refstatus(RefTime,RefPerc,PrevStates,StartTime),
		     Delta is NewTime-RefTime,
		     (Delta>20000  % more than 20 seconds since last message
		      % ; random(0,20000,X), X=<Delta) 
		     -> get_model_check_stats(States,Transitions,ProcessedTotal,Perc),
		        format('~n* ~w states checked (~1f% of total ~w), ~w transitions,',[ProcessedTotal,Perc,States,Transitions]),
			    statistics(memory_used,M), tools:print_mb(M),
			    DeltaStates is (States-PrevStates),
			    NodesPerSec is DeltaStates*1000 / (NewTime-RefTime),
			    format(', current ~2F states/sec~n',[NodesPerSec]),
			    RefStatus1 = refstatus(NewTime,Perc,States,StartTime),
			    (Perc>RefPerc -> P5 is (Perc/100) **5,
			                     %DeltaPerc is P5 - ((RefPerc/100) ** 5),
			                     FullDeltaTime is NewTime - StartTime,
			                     FullDeltaTimeMinutes is FullDeltaTime / 60000.0,
			                     TimeRemaining is (FullDeltaTime/P5 - NewTime) / 60000,
			                     EstimatedStates is integer(States / P5),
			                     format('* walltime ~1f minutes, estimating ~1f minutes remaining with ~w states in total~n',[FullDeltaTimeMinutes,TimeRemaining,EstimatedStates])
			                  ; true)
			  ; RefStatus1 = RefStatus)
		 ;   RefStatus1 = RefStatus),
	  ((NewTime>LimitTime, Nr>0)
		-> Res = [timeout,Nr], NodesAnalysed=Nr
		;  open_search_do_one_node(Nr,LimitNr,RefStatus1,NodesAnalysed,LimitTime,Res,
					  FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,
					  SFC,PartialOrderReduction,Graphs,MODE,DF))
	).
open_search10(N,_LimitNr,_RefStatus,N,_,no ,_,_,_,_,_,_,_,_,_).

get_model_check_stats(States,Transitions,ProcessedTotal,Percentage) :-
     get_state_space_stats(States,Transitions,ProcessedTotal),
	 Percentage is (100*ProcessedTotal / States).

:- use_module(preferences,[preference/2]).
:- use_module(specfile,[prepare_state_for_specfile_trans/2]).
open_search_do_one_node(Nr,LimitNr,RefStatus,NodesAnalysed,LimitTime,Res,
                           FindDeadlocks,FindInvViolations,
                           FindGoal,FindAssViolations,
                           StopAtFullCoverage,PartialOrderReduction,Graphs,MODE,DFMODE) :-
  get_next_node_to_check(FindInvViolations,ID,MODE,DFMODE),!,
  set_context_state(ID), % print(treating_node(ID)),nl,trace,
  ((preferences:preference(ignore_hash_collisions,true),
    preferences:preference(forget_state_space,true))
    -> retract_visited_expression(ID,CurState) %, retractall(id_to_marker(ID,_)) not necessary
    ;  visited_expression(ID,CurState)),
  (FindInvViolations=1 % compute invariant status
    -> specfile: prepare_state_for_specfile_trans(CurState,PCurState),
       user:compute_invariant_if_necessary(ID,PCurState,MODE)
     ; CurState=PCurState), 
    %print(processed_invariant(ID)), statistics(memory_used,M), tools:print_mb(M),nl,
  !,
  compute_transitions(PartialOrderReduction,ID,PCurState,MODE,FindDeadlocks,FindInvViolations,Graphs),
  %print_message(mc(ID)),
    %print(computed_all_transitions(ID)), statistics(memory_used,M2), tools:print_mb(M2),nl,
   % !,
  user:do_one_gui_event,
  (open_search_found_error(ID,PCurState,Res,FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,StopAtFullCoverage,MODE)
   ->  print(found_error(Res,ID)),nl, clear_context_state,
       NodesAnalysed=Nr,
       user:tcltk_execute_trace_to_node(ID)
    ;  N1 is Nr + 1, 
	   open_search10(N1,LimitNr,RefStatus,NodesAnalysed,LimitTime,Res,
		FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,StopAtFullCoverage,
		PartialOrderReduction,Graphs,MODE,DFMODE)
  ).
open_search_do_one_node(N,_LimitNr,_RefStatus,N,_,all,_,_,_,_,_,_,_,_,_) :- 
    print_message('All open nodes visited').

compute_transitions(sleep_sets,ID,PCurState,_MODE,FindDeadlocks,_FindInvViolations,[DisableGraph]) :-
    user: compute_enabled_transitions_only(ID,PCurState,DisableGraph,FindDeadlocks,on).
compute_transitions(por(TYPE,UseEnableGraph,Depth,UsePGE),ID,PCurState,_MODE,_FindDeadlocks,FindInvViolations,[DisableGraph]) :-
    user: compute_ample_set(ID,PCurState,FindInvViolations,DisableGraph,por(TYPE,UseEnableGraph,Depth,UsePGE)).
%% compute_transitions(ample_sets,ID,PCurState,_MODE,_FindDeadlocks,FindInvViolations,[DisableGraph]) :-
%%     user: compute_ample_set(ID,PCurState,FindInvViolations,DisableGraph,without_pge).
compute_transitions(off_pge,ID,PCurState,_MODE,FindDeadlocks,_FindInvViolations,[DisableGraph]) :-
	user: compute_enabled_transitions_only(ID,PCurState,DisableGraph,FindDeadlocks,off).
compute_transitions(off,ID,PCurState,_MODE,_FindDeadlocks,_FindInvViolations,_) :-
    user: compute_all_transitions(ID,PCurState).
compute_transitions(0,ID,PCurState,_MODE,_FindDeadlocks,_FindInvViolations,_) :- % 0 is the same as off; be compatible with other model check Tcl/Tk options where 0 means off
    user: compute_all_transitions(ID,PCurState).



:- use_module(preferences,[get_preference/2]).
:- use_module(specfile,[b_or_z_mode/1]).
%%get_next_node_to_check(1,ID,_) :- state_space:not_invariant_checked(ID). % if there is an existing node whose invariant we have not yet checked: check it now
get_next_node_to_check(_,ID,AnimMODE,DFMODE) :- 
  %(nonvar(ID) -> print(nonvar(ID)),nl ; true),
  get_open_node_to_check(DFMODE,ID),
  (b_or_z_mode(AnimMODE) -> assert(not_invariant_checked(ID)) ; true).
  
:- dynamic last_step_breadth_first_in_mixed_mode/0.
get_open_node_to_check(depth_first,NodeID) :- !,
   pop_id_from_front(NodeID).
get_open_node_to_check(mixed,NodeID) :-  %%depth_breadth_first_mode(DFMODE),
   random(1,3,1),
                  /* random(1,4,1) ==> 33 percent chance of going DF */
                  /* random(1,3,1) ==> 50 percent chance of going DF */
                  /* (random(1,3,1);random(1,3,1)) ==> 75 % chance of going DF */
   retractall(last_step_breadth_first_in_mixed_mode),
   !,
   % We are going depth_first
   pop_id_from_front(NodeID). %, print_message(a(NodeID)).
get_open_node_to_check(DFMODE,NodeID) :-
   ((DFMODE=breadth_first ; last_step_breadth_first_in_mixed_mode)
    -> true ; assert(last_step_breadth_first_in_mixed_mode)),
   %print(going_bf(NodeID)),nl,
   pop_id_from_end(NodeID).

:- use_module(state_space).
open_search_found_error(ID,CurState,Error,
                       _FindDeadlocks,1,_FindGoal,_FindAssViolations,_StopAtFullCoverage,MODE) :-
    find_invariant_error(MODE,ID,CurState,Error).
open_search_found_error(ID,_CurState,deadlock,1,_,_,_,_,_) :-
   is_deadlocked(ID). %, print(deadlock(ID)),nl.
open_search_found_error(ID,CurState,assertion_violation,_,_,_,1,_,_MODE) :-
   state_violates_assertions(ID,CurState).
open_search_found_error(ID,_CurState,ERR,_,_,_,_,_,_MODE) :-
   state_space:state_error(ID,_,Error), Error \== invariant_violated,
   ERR = state_error(Error).
open_search_found_error(_ID,_CurState,general_error_occurred,_,_,_,_,_,_MODE) :-
   error_manager:real_error_occurred.  /* do we need to reset_errors_occurred before model check ?? */
open_search_found_error(ID,_CurState,goal_found,_,_,1,_,_,MODE) :-
   (user:test_goal_in_node(ID) ; is_csp_goal_state(ID,MODE)).
open_search_found_error(_ID,_CurState,full_coverage,_,_,_,_,1,_MODE) :-
   \+ operation_not_yet_covered(_).

find_invariant_error(b,ID,_CurState,invariant_violation) :- !, % here we know that invariant information has been computed
    (invariant_violated(ID) -> true).
find_invariant_error(cspm,ID,_CurState,csp_error) :- !, % we know that only csp_can be possible
    is_csp_error_state(ID).
find_invariant_error(MODE,ID,CurState,invariant_violation) :- 
    user:check_invariantKO(MODE,ID,CurState).
find_invariant_error(csp_and_b,ID,_CurState,csp_error) :- !, % xtl mode cannot apply
    is_csp_error_state(ID).
find_invariant_error(MODE,ID,CurState,xtl_error) :- 
    is_xtl_error_state(ID,CurState,MODE).
   
is_deadlocked(ID) :-
  preferences:preference(forget_state_space,true),!, /*  transitions not added */
  (retract(state_space:transition(ID,_,_,_)) -> fail ; true).
is_deadlocked(ID) :- % ID must be ground upon call
 %%  visited_expression_id(ID),
  deadlocked_state(ID),
  \+ max_reached_or_timeout_for_node(ID).
/* \+((transition(ID,_,NID),NID \= ID)).   <-- to detect more liberal form of DEADLOCK */

is_csp_error_state(ID) :- 
  transition(ID,io(_,error,_),_).  /* <--- THE CSP has raised the error channel */

:- use_module(xtl_interface,[xtl_property/2, promela_property/2]).
is_xtl_error_state(_ID,CurState,AM) :-  
  ( AM=xtl     -> xtl_property(CurState,unsafe)   /* <--- XTL has found unsafe configuration */
  ; AM=promela -> promela_property(CurState,unsafe)
%  ; AM=smv     -> smv_property(CurState,unsafe)
  ).

is_csp_goal_state(ID,MODE) :- csp_mode(MODE),
  visited_expression_id(ID),
  transition(ID,io(_,goal,_),_).


% ---------------------------------------

:- use_module(tools, [ajoin/2]).
:- use_module(symmetry_marker,[hash_model_checking_imprecise/2]).
%model_checking_is_unsound(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg,Term)
model_checking_is_incomplete(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg,'') :-
    proof_info_model_checking_incomplete(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg).
model_checking_is_incomplete(_FindInv,_FindDeadlocks,_FindGoal,_FindAss,Msg,Term) :-
   (hash_model_checking_imprecise(ID,BasicType)
    ->  Msg = '*** Hashing may be imprecise for, e.g., for variable ',
        translate:pretty_type(BasicType,S),
        ajoin([ID,' : ',S],Term)).
model_checking_is_incomplete(1,_FindDeadlocks,_FindGoal,_FindAss,Msg,Term) :-
    (time_out_for_invariant(ID)
      -> Msg = '*** Timeouts occured during invariant computation, e.g., state with id = ', Term=ID).
model_checking_is_incomplete(_FindInv,_FindDeadlocks,_FindGoal,_FindAss,Msg,Term) :-
    (max_reached_for_node(ID)
      -> Msg = '*** Not all transitions were computed, e.g., state with id = ', Term=ID).


:- use_module(bmachine, [b_machine_has_proven_invariants/0, b_machine_has_unproven_assertions/0]).
proof_info_model_checking_incomplete(1,_FindDeadlocks,_FindGoal,0,Msg) :-
    b_machine_has_proven_invariants,
    preferences:get_preference(use_po,true),
    b_machine_has_unproven_assertions,
    Msg = '*** Using proof information (-p PROOF_INFO TRUE) is potentially incomplete when not checking theorems (-noass) !'.

:- use_module(enabling_analysis, [compute_action_relations_graph/2,disable_graph/1]).
:- dynamic expired_static_analysis_time/1.


% disable POR if not applicable + merge POR/PGE info into one atom for easier lookup in compute_transitions
disable_optimisations_if_not_useful(MODE,FindDeadlocks,FindInvViolations,PartialOrderReduction,PartialGuardsEvaluation,
                                    UpdatedPartialOrderReduction) :-
   MODE == b, (FindDeadlocks == 1 ; FindInvViolations == 1),!,
   % in this case turn partial order reduction and partial guard evaluation off
   (  (PartialOrderReduction==off,PartialGuardsEvaluation==1) -> UpdatedPartialOrderReduction = off_pge
    ; (PartialOrderReduction==ample_sets) -> 
       (PartialGuardsEvaluation==1 -> UsePGE=1; UsePGE=0),
       get_preference(enable_graph,UseEnableGraph),
       get_preference(enable_graph_depth,Depth),
       UpdatedPartialOrderReduction = por(ample_sets,UseEnableGraph,Depth,UsePGE)
       %UpdatedPartialOrderReduction = ample_sets_pge
     ; UpdatedPartialOrderReduction = PartialOrderReduction).
disable_optimisations_if_not_useful(_,_,_, _,_, off).

/* Predicate for checking if there are any optimizations activated */
open_search_check_for_optimisation_options(MODE,FindInvViolations,PartialOrderReduction,PartialGuardsEvaluation,Graphs) :-
   % Combining both optimisation options. Here we need to run both relation analyses in order to apply both optimisations during the model checking
   retractall(expired_static_analysis_time(_)),
   statistics(runtime,[T1,_]),
   (MODE == b ->
    % clearing dynamic facts for optimisations, running the mc tests for this kind of optimisations may
    % interfer succeeding tests testing the same optimisations as the facts will not be cleared (the fix is temporary)
     user: clear_dynamic_predicates_for_PGE,
     ample_sets: clear_dynamic_predicates_for_POR,
	   ((PartialOrderReduction = por(ample_sets,UseEnableGraph,Depth,PGE))  ->
		  (PartialGuardsEvaluation == 1 ->
        debug_println(9,combined_optimisation_POR_PGE),
			  % computing dependency and enabling event realtions for the Partial Order Reduction
			  compute_dependendency_relation_of_all_events_in_the_model(FindInvViolations,por(ample_sets,UseEnableGraph,Depth,PGE),_EnableGraph),
			  % computing disabling and independency event realtions for the Partial Guard Evaluation
			  compute_action_relations_graph(FindInvViolations,DisableGraph),
			  % in case that the disable graph is empty, then we do not need to check if the PGE optimisation is switched on
			  Graphs = [DisableGraph]
			  % the following two optimizations should be disabled
			  % by combining POR with PGE, because POR take advantage
			  % of independent events
      ; otherwise ->
        debug_println(9,optimising_by_por),
		    compute_dependendency_relation_of_all_events_in_the_model(FindInvViolations,por(ample_sets,UseEnableGraph,Depth,PGE),_EnableGraph),
		    Graphs = [[]]
		  )
     ;(PartialOrderReduction == sleep_sets;PartialGuardsEvaluation==1) ->
        debug_println(9,optimising_by_pge),
    		compute_action_relations_graph(FindInvViolations,DisableGraph),
    		Graphs=[DisableGraph]
 	   ; otherwise ->
        Graphs=[[]]
	   )
   ; otherwise -> % no optimisation information should be produced for other formalisms
	    Graphs=[[]]
   ),
   statistics(runtime,[T2,_]),
   D is T2-T1,
   assert(expired_static_analysis_time(D)).
