% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(solver_interface, [solve_predicate/3, solve_predicate/4, solve_predicate/5, solver_result/4,
                             set_up_typed_localstate_for_pred/4, type_check_in_machine_context/2,
                             predicate_uses_unfixed_deferred_set/1, predicate_uses_unfixed_deferred_set/2,
                             unfixed_typed_id_in_list/3,
                             call_with_smt_mode_enabled/1]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,cbc).
:- module_info(description,'This module provides an interface to ProB\'s constraint solving kernel.').

:- use_module(bmachine, [b_get_machine_variables/1, b_get_machine_constants/1, b_type_expressions/5, full_b_machine/1]).
:- use_module(btypechecker,[env_empty/1,opentype/5, add_identifiers_to_environment/3, env_store_operator/4]).
:- use_module(b_global_sets, [is_b_global_constant/3,
                              b_global_set/1, unfixed_deferred_set/1,
                              contains_unfixed_deferred_set/1]).
:- use_module(b_interpreter, [set_up_typed_localstate/6]).
:- use_module(self_check, [time_out_with_factor_call/3]).
:- use_module(bsyntaxtree, [find_typed_identifier_uses/3, find_typed_identifier_uses/2,
                 map_over_typed_bexpr/3, get_texpr_id/2, get_texpr_type/2, get_texpr_types/2]).
:- use_module(error_manager).
:- use_module(tools,[assert_once/1,foldl/5]).
:- use_module(translate, [translate_bvalue/2]).
:- use_module(bmachine_structure,[get_section/3]).
:- use_module(debug,[debug_println/2]).


:- use_module(library(lists), [include/3,keys_and_values/3]).

:- use_module(preferences).


:- meta_predicate call_with_smt_mode_enabled(0).
:- meta_predicate call_with_preference(0).
:- meta_predicate call_with_chr(0).
call_with_chr(Call) :- call_with_preference(Call,use_chr_solver,true).

call_with_preference(Call,Pref,Val) :-
    temporary_set_preference(Pref,Val,CHNG1),
    call_cleanup(Call,
                 reset_temporary_preference(Pref,CHNG1)).

% :- meta_predicate call_with_smt_mode_enabled(0). % directive above for Spider
call_with_smt_mode_enabled(Call) :-
    temporary_set_preference(use_smt_mode,true,SMT),
    temporary_set_preference(use_clpfd_solver,true,CLPFD), % we also enable CLPFD
    % TO DO: also enable CHR 
    call_cleanup( (Call,!),
                  (reset_temporary_preference(use_smt_mode,SMT),reset_temporary_preference(use_clpfd_solver,CLPFD)
                  )).

% solve predicate finds *one* solution for predicate in State
solve_predicate(Pred,State,Res) :- solve_predicate_direct(Pred,State,1,Res).
solve_predicate(Pred,State,Timeout,Res) :-
   solve_predicate_with_options([use_smt_mode/true,use_clpfd_solver/true],Pred,State,Timeout,Res).
solve_predicate(Pred,State,Timeout,Options,Res) :-
   solve_predicate_with_options(Options,Pred,State,Timeout,Res).

solve_predicate_with_options([],Pred,State,Timeout,Res) :-
   solve_predicate_direct(Pred,State,Timeout,Res).
solve_predicate_with_options([Option|T],Pred,State,Timeout,Res) :-
    solve_with_option(Option,T,Pred,State,Timeout,Res).

solve_with_option('CHR',T,Pred,State,Timeout,Res) :- solve_with_option(use_chr_solver/true,T,Pred,State,Timeout,Res).
solve_with_option('SMT',T,Pred,State,Timeout,Res) :- solve_with_option(use_smt_mode/true,T,Pred,State,Timeout,Res).
solve_with_option('CLPFD',T,Pred,State,Timeout,Res) :- solve_with_option(use_clpfd_solver/true,T,Pred,State,Timeout,Res).
solve_with_option(Pref/Val,T,Pred,State,Timeout,Res) :-
   call_with_preference(solve_predicate_with_options(T,Pred,State,Timeout,Res),Pref,Val).

solve_predicate_direct(Pred,State,Timeout,Res) :-
    %temporary_set_preference(randomise_enumeration_order,true),
    %temporary_set_preference(find_abort_values,true),  % I am not sure we should do this; we want to find counter-examples quickly, we are less worried with finding WD-problems ?!

    %temporary_set_preference(minint,-2147483648), % disabled to pass tests 1004,1015,..
    %temporary_set_preference(maxint,2147483647),
    (real_error_occurred -> reset_real_error_occurred, PrevErr=true ; PrevErr=false),
    solve_predicate_aux(Pred,State,Timeout,InnerRes),
    % reset_temporary_preference(randomise_enumeration_order),
    (InnerRes = overflow -> Res = overflow ; real_error_occurred -> Res = error; Res = InnerRes),
    (PrevErr -> assert_real_error_occurred ; true). % re-establish real-error occured

:- use_module(clpfd_interface,[clpfd_overflow_error_message/0]).
solve_predicate_aux(Pred,State,Timeout,Res) :-
    enter_new_error_scope(ScopeID), %print(enter_scope(ScopeID)),nl,
    call_cleanup(solve_predicate_aux1(Pred,State,Timeout,Res),
                 exit_error_scope(ScopeID,_)).

solve_predicate_aux1(Pred,State,Timeout,Res) :-
    solve_predicate_aux2(Pred,State,Timeout,Res),
    !. % we only look for one solution
solve_predicate_aux1(_Pred,State,_Timeout,Res) :-
    critical_enumeration_warning_occured_in_error_scope(A,B,C,D), % we did not find a solution: check for enum warnings
    !,
    clear_enumeration_warnings,
    State = no_state,
    Res = virtual_time_out(enumeration_warning(A,B,C,D,critical)).
solve_predicate_aux1(_Pred,State,_Timeout,Res) :-
    State = no_state,
    Res = no.

:- use_module(clpfd_interface,[catch_clpfd_overflow_call2/2]).
solve_predicate_aux2(Pred,State,Timeout,Res) :-
    catch_clpfd_overflow_call2(
               solve_predicate_aux3(Pred,State,Timeout,Res),
               ( %print(overflow),nl,trace,
                 %clpfd_interface:clpfd_overflow_warning_message,  % message is already printed by catch_clpfd_overflow_call2
                 % we could try and solve without CLPFD ?!!
                 Res=overflow,State=no_state)).

/*
filter_global_identifiers([],[]).
filter_global_identifiers([Identifier|IDs],Out) :-
    filter_global_identifiers(IDs,TempOut),
    (is_global(Identifier)
     -> Out = TempOut
     ;  Out = [Identifier|TempOut]
    ).

is_global(Identifier) :-
    get_texpr_id(Identifier,Name), get_texpr_type(Identifier,Type),
    Type = global(Set), is_b_global_constant(Set,_Nr,Name).
is_global(Identifier) :-
    get_texpr_id(Identifier,Name), get_texpr_type(Identifier,Type),
    Type = set(global(Name)), b_global_set(Name).
*/

:- use_module(b_interpreter_components).
solve_predicate_aux3(Pred,State,TimeoutFactor,Res) :-
    % find all identifiers used in the predicate and create a state
    % was done by the caller, leading to code duplication
    find_typed_identifier_uses(Pred,FilteredIdentifiers),
    !,
    % global constants should not appear in the state
    % otherwise they treated are  as variables by the solver
    %  -> we now call find_typed_identifier_uses; we used to call filter_global_identifiers(Identifiers,FilteredIdentifiers),
    % state may be predefined / partially set up
    (var(State)
       -> set_up_typed_localstate_for_pred(FilteredIdentifiers,Pred,TypedVals,State)
          % TO DO: add static_symmetry detection 
        ; true),
    b_interpreter_components:reset_component_info(false),
    time_out_with_factor_call(
        b_interpreter_components:b_trace_test_components(Pred,State,TypedVals),
        TimeoutFactor, TO='time_out'),
    (TO==time_out -> Res = TO ; Res = yes).

:- use_module(static_ordering,[sort_ids_by_usage/4]).
set_up_typed_localstate_for_pred(CV,Pred,TypedVals,State) :-
   (preferences:preference(use_static_ordering,true)
     -> static_ordering:sort_ids_by_usage(CV,Pred,SortedVarlist,no_warnings)
        %,reverse(Constants1,Constants)
        %,print(sorted_by_usage(Constants)),nl
     ;  SortedVarlist =  CV),
   set_up_typed_localstate(SortedVarlist,_FreshVars,TypedVals,[],State,positive).

% translate solver_result for use by Disprover/Prover or similar:
solver_result(time_out,_P,_State,time_out).
solver_result(error,_P,_State,error).
solver_result(no,P,_State,Res) :-
    % there are two cases in which we might have a false negative:
    % we enumerated a critical variable
    %\+ event_occurred_in_error_scope(enumeration_warning(_,_,_,_,critical)), % we now check this in solve_predicate_aux1
    % or there are unfixed / deferred global sets involved.
    (predicate_uses_unfixed_deferred_set(P,CType)
        % no unfixed global sets involved -> real contradiction found
        -> debug_println(10,solver_predicate_contains_unfixed_deferred_set(CType)),
           Res=no_solution_found(unfixed_deferred_sets)
        ;  Res=contradiction_found
    ).
solver_result(virtual_time_out(Reason),_P,_State,no_solution_found(Reason)) :-
    true. % this is now checked in solve_predicate_aux1: event_occurred_in_error_scope(enumeration_warning(A,B,C,D,critical)).
solver_result(overflow,_P,_State,no_solution_found(clpfd_overflow)).
solver_result(yes,_P,State,solution(Solution)) :-
    findall(binding(Id,Value,PPValue),
            (member(bind(Id,Value),State),
                (translate_bvalue(Value,PPValue) -> true ; PPValue='**pretty-print failed**')),
            Solution).

% check if a typed expression references an unfixed deferred set in some inner node
predicate_uses_unfixed_deferred_set(TExpr) :- predicate_uses_unfixed_deferred_set(TExpr,_).

predicate_uses_unfixed_deferred_set(TExpr,CType) :-
   unfixed_deferred_set(_), % check first if some unfixed deferred set exists
   map_over_typed_bexpr(solver_interface:uses_unfixed_ds_aux,TExpr,CType).
uses_unfixed_ds_aux(TExpr,Type) :- get_texpr_type(TExpr,Type),
   contains_unfixed_deferred_set(Type).

% find a typed id whose type references an unfixed deferred set
unfixed_typed_id_in_list(TID,CType,Constants) :-
         unfixed_deferred_set(_), % check first if some unfixed deferred set exists
         member(TID,Constants), get_texpr_type(TID,CType),
         contains_unfixed_deferred_set(CType).


type_check_in_machine_context(Predicates, TPredicates) :-
  % create an empty type environment
  % and store the known types of variables / constants
  % otherwise, we might be unable to type variables from
  % machines somewhere in the refinement hierarchy
  % (their typing predicate might not be in the hypothesis)
  env_empty(E),
  b_get_machine_variables(V), b_get_machine_constants(C),
  % add theory operators
  full_b_machine(Machine),
  get_section(operators,Machine,Operators),
  keys_and_values(Operators,Ids,Ops),
  foldl(env_store_operator,Ids,Ops,E,EWithTheories),
  % constants and variables
  findall(b(identifier(Name),global(GlobalSet),[enumerated_set_element]),is_b_global_constant(GlobalSet,_Nr,Name),GC),
  findall(b(identifier(Set),set(global(Set)),[given_set]),b_global_set(Set),GS),
  add_identifiers_to_environment(V,EWithTheories,EV),
  add_identifiers_to_environment(C,EV,EVC),
  add_identifiers_to_environment(GC,EVC,EVCGC),
  add_identifiers_to_environment(GS,EVCGC,EOut),
  opentype(Predicates, EOut, _Identifiers, DeferredSets, TPredicates),
  % DeferredSets should be empty? I.e. should this be allowed to create new types?
  DeferredSets = [].
  % a possible refactoring for the above - however, it does not type with open identifiers
  % reverted it for now
  % b_type_expressions(Predicates,[constants],_Types,TPredicates,Errors),
  % add_all_perrors(Errors,[],type_expression_error),
  % Errors=[].

