% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(state_space_reduction,
               [
                % a generic state space reduction predicate:
                reduce_state_space/2,
                
                % these are the facts generated by reduce_state_space:
                get_reduced_node/4, reduced_trans/5,
                equivalence_class/2, 
                
                % makes use of reduce_state_space to compute the signature-merge:
                compute_signature_merge/0,
                write_signature_merge_to_dotfile/1, write_signature_merge_to_dotfile/2,
                tcltk_write_signature_merge_to_dotfile/2,
                write_covered_values_for_expression_to_csvfile/2,
                initialised_b_state_exists/0,
                compute_covered_values_for_expression/4,
                tcltk_compute_nr_covered_values_for_all_variables/1,
                
                % makes use of reduce_state_space to compute the transition diagram for an expression:
                compute_transition_diagram/1,
                write_transition_diagram_for_expr_to_dotfile/2,
                generate_node_color/4,
                generate_node_labels/3,
                generate_transition_label/3,
                generate_transition_color_and_style/6,
                reset_ignored_events/0,
                set_ignored_events/1
                ]).


:- use_module(state_space).
:- use_module(error_manager).

:- use_module(extension('counter/counter')).

:- dynamic reduced_node/5, equivalence_class/2, equivalence_class_contains_current_node/1.
:- dynamic reduced_trans/5.
reset :- retractall(reduced_node(_,_,_,_,_)), retractall(equivalence_class(_,_)),
         retractall(equivalence_class_contains_current_node(_)),
         retractall(reduced_trans(_,_,_,_,_)),
         counter_init,
         new_counter(transition_id), new_counter(equiv_class_id).

:- meta_predicate reduce_state_space(3,5).


% ---------------------------

% An implementation of signature merge algorithm:

:- dynamic ignore_event/1.
reset_ignored_events :- %print(reseting_ignore_event),nl,
    retractall(ignore_event(_)).
set_ignored_events([]).
set_ignored_events([H|T]) :- assert(ignore_event(H)),
   format('Ignoring: ~w~n',[H]),
   set_ignored_events(T).
signature_and_inv(ID,S,AbsState) :-
   (invariant_violated(ID) -> AbsState = [invariant_violated|Sig] ; AbsState = Sig),
   signature(ID,S,Sig).
signature(ID,_,AbsState) :-
     findall(Name,(transition(ID,Action,_,_),specfile:get_operation_name(Action,Name),
                   \+ ignore_event(Name)), L),
     (L=[] -> 
       (not_all_transitions_added(ID) -> AbsState = ['NOT YET EXPLORED']
         ; transition(ID,_,_,_) -> AbsState = ['NO SELECTED EVENTS']
         ; AbsState = ['DEADLOCK'])
      ; sort(L,AbsState)
     ).

transition_name(_,Action,_,_,Name) :- specfile:get_operation_name(Action,Name),
      \+ ignore_event(Name).

compute_signature_merge :-
 % we could provide options: not consider invariant, ignore certain transitions, ...
    reduce_state_space(signature_and_inv,transition_name).
 
% use_module(state_space_reduction), write_signature_merge_to_dotfile('~/Desktop/sm.dot').

tcltk_write_signature_merge_to_dotfile(IgnoredEventNrs,File) :- 
       sap:tcl_convert_event_numbers_to_names(list(IgnoredEventNrs),list(IgnoredEvents)),
       %print(tcl_convert_event_numbers_to_names(list(IgnoredEventNrs),list(IgnoredEvents))),nl,
       write_signature_merge_to_dotfile(IgnoredEvents,File).

:-use_module(dot_graph_generator,[gen_dot_graph/6]).
write_signature_merge_to_dotfile(IgnoredEvents,File) :- 
       reset_ignored_events,
       set_ignored_events(IgnoredEvents),
       debug:time(state_space_reduction:compute_signature_merge),
       gen_dot_graph(File,state_space_reduction,sm_node_pred(insert_commas),sm_trans_predicate(true),none,none),
       reset_ignored_events.
write_signature_merge_to_dotfile(File) :-
       write_signature_merge_to_dotfile([],File).

:- use_module(library(lists),[maplist/3]).
:- use_module(tools,[ajoin/2, string_escape/2, wrap_and_truncate_atom/4]).
% node predicate for gen_dot_graph:
sm_node_pred(GenLabels,NodeID,SubGraph,NodeDesc,Shape,Style,Color) :- SubGraph=none,
     Shape=record,
     get_reduced_node(AbsState,C1,Witness,NodeID),
     (equivalence_class_contains_current_node(NodeID) -> Style = bold % bold, diagonals, rounded
      % we could optionally turn this off in case the UI wants to do this itself and update when the animator changes state
       ; Style = none), % or rounded
     %insert_commas(AbsState,0,CL),
     (AbsState = [invariant_violated|Signature]
       -> Labels = ['INVARIANT VIOLATED\\n|'|EscCL]
       ;  Labels = EscCL, Signature = AbsState
     ),
     call(GenLabels,Signature,CL),
     maplist(string_escape,CL,EscCL),
     %maplist : wrap_and_truncate_atom(EscCL,30,121,EscCLW),
     append(['|{' | Labels],['\\n|# states: ',C1,'}|'],Atoms),
     ajoin(Atoms,NodeDesc),
     generate_node_color(Witness,AbsState,C1,Color).


generate_node_color(Witness,AbsState,Count,Color) :-
    (not_all_transitions_added(Witness) -> Color=orange
       ; AbsState = [invariant_violated|_] -> Color=red
       ; AbsState = ['DEADLOCK'] -> Color=red
       ; Count>1 -> Color=blue
       ; Color=green).

generate_node_labels(GenLabels,AbsState,Labels) :-
    (AbsState = [invariant_violated|Signature]
      -> Labels = ['INVARIANT VIOLATED'|CL]
      ;  Labels = CL, Signature = AbsState
    ),
    call(GenLabels,Signature,CL).

generate_transition_label(AbsAction,Count,Label) :-
    (Count>1 
      -> ajoin([AbsAction,' : ',Count],Label) 
      ; Label=AbsAction
    ).

generate_transition_color_and_style(ShowSelfLoops,NodeID,AbsAction,SuccID,Color,Style) :-
    ((reduced_trans(NodeID,AbsAction,_,Other,_),
       Other \= SuccID)
        -> Color = lightgray, Style = dashed % there exists another transition leading to another equivalence class
           % to do: we could use another colour if this transition is not always possible ??
        ;  self_loop_check(NodeID,SuccID,ShowSelfLoops),
           ((abs_signature(NodeID,AlwaysEnabled), nonmember(AbsAction/SuccID,AlwaysEnabled))
             -> Color = blue, Style = dashed % not always enabled; but if it is it reached the same destination equivalence class
             ;  Color = black, Style = solid
           )
    ).

self_loop_check(NodeID,SuccID,ShowSelfLoops) :-
  (NodeID==SuccID -> ShowSelfLoops=true % if we have a single self-loop then only show it if preference set
             ; true).


% this should probably be outsourced to another module:
insert_commas(L,R) :- insert_commas(L,0,R).
insert_commas([],_,[]).
insert_commas([H],_,R) :- !, R=[H].
insert_commas([H|T],N,[H,Sep|IT]) :- 
    (N>5 -> N1=0, Sep=',\\n' ; N1 is N+1, Sep=', '),
    insert_commas(T,N1,IT).

% transition predicate for gen_dot_graph:
sm_trans_predicate(ShowSelfLoops,NodeID,Label,SuccID,Color,Style) :- 
     reduced_trans(NodeID,AbsAction,Count,SuccID,_TID),
     % TO DO: maybe merge all AbsAction with the same Origin & Destination into single transition
     generate_transition_label(AbsAction,Count,Label),
     generate_transition_color_and_style(ShowSelfLoops,NodeID,AbsAction,SuccID,Color,Style).

% ---------------------------

% Implementation of the Transition Diagram for Expression
% evaluate an expression for every state; merge those states that have the same value
% show in the diagram how the value can be changed by various operations
:- use_module(specfile,[ state_corresponds_to_initialised_b_machine/2]).
state_value_for_expr(Expr,_ID,root,AbsState) :- !, AbsState=expr(Expr).
state_value_for_expr(Expr,_ID,State,AbsState) :-
    state_corresponds_to_initialised_b_machine(State,BState),
    b_interpreter:b_compute_expression(Expr,[],BState,AbsState),!.
state_value_for_expr(_Expr,_ID,_State,undefined).

:- use_module(specfile,[ state_corresponds_to_initialised_b_machine/1]).
% check if at least one initialised state exists; otherwise transition diagram makes no sense
initialised_b_state_exists :-
   visited_expression(_,S),
   state_corresponds_to_initialised_b_machine(S).

state_value_for_variable(Variable,_ID,State,Value) :- get_variable_value(State,Variable,Value).
% comment in lines below if you want to use it for transition diagram; however then we need to adapt tcltk_compute_nr_covered_values_for_all_variables
%get_variable_value(root,Variable,Value) :- !, Value = Variable.
get_variable_value(const_and_vars(_ID,S),Variable,Value) :- !, member(bind(Variable,Value),S).
%get_variable_value(concrete_constants(_),_Variable,Value) :- !, Value=undefined.
get_variable_value(csp_and_b(_CSP,BState),Variable,Value) :- !,
    get_variable_value(BState,Variable,Value).
get_variable_value(S,Variable,Value) :- !, member(bind(Variable,Value),S).

% as above but also check invariant status
state_value_for_expr_and_inv(Expr,ID,State,ResAbsState) :-
   state_value_for_expr(Expr,ID,State,AbsState),
   (invariant_violated(ID) -> ResAbsState = [invariant_violated|AbsState]
   % TO DO: also look if invariant_not_yet_checked ?, ... not_all_transitions_added ?
     ; ResAbsState = AbsState).

compute_transition_diagram(Expr) :-
    reduce_state_space(state_value_for_expr_and_inv(Expr),transition_name).

% use_module(state_space_reduction), state_space_reduction:write_transition_diagram_for_expr_to_dotfile('card(waiting)','~/Desktop/sm.dot').
write_transition_diagram_for_expr_to_dotfile(VorE,File) :-
   parse_expression(VorE,TypedExpr),
   compute_transition_diagram(TypedExpr),
   gen_dot_graph(File,state_space_reduction,
               sm_node_pred(gen_label),sm_trans_predicate(false),none,none).
    % the original version only kept those transitions that actually change an expression
    %   we could easily do this here by adapting sm_trans_predicate

:- use_module(bmachine,      [b_parse_machine_expression_from_codes/2]).
parse_expression('$VARIABLES',TypedExpr) :- !,
    bmachine:b_get_machine_variables(DeclaredVars),
    gen_couple_list(DeclaredVars,L),
    ajoin(L,Atom),
    print(Atom),nl,
    parse_expression(Atom,TypedExpr).
parse_expression(VorE,TypedExpr) :-
   atom_codes(VorE,VorECodes),
   % to do: maybe support b_parse_machine_expression_from_codes_with_prob_ids 
   bmachine:b_parse_machine_expression_from_codes(VorECodes,TypedExpr).


:- use_module(bsyntaxtree,[get_texpr_id/2]).
gen_couple_list([],['"NO VARIABLES"']).
gen_couple_list([H|T],['(',HID|CT]) :- get_texpr_id(H,HID), gen_couple_list2(T,CT).

gen_couple_list2([],[')']).
gen_couple_list2([H|T],[',',HID|CT]) :- get_texpr_id(H,HID), gen_couple_list2(T,CT).

gen_label(expr(Expr),[Str]) :- !,
   translate:translate_bexpression_with_limit(Expr,100,Str).
gen_label(Value,[Str]) :-
   (translate:translate_bvalue_with_limit(Value,50,Str)
    -> true; Str = 'fail').

simple_list(List,List).

:- use_module(bsyntaxtree,   [get_texpr_type/2]).
:- use_module(kernel_objects,[max_cardinality/2]).
:- use_module(translate,    [get_bexpression_column_template/4]).
% compute which values are present in the state space for a give Expression
compute_covered_values_for_expression(VorE,MaxTypeCard,TotalNrOfValuesFound,
                     list([list(['Nr'|AllColHeaders])|LResult])) :-
    parse_expression(VorE,TypedExpr),
    translate:translate_bexpression_with_limit(TypedExpr,100,Str),
    get_bexpression_column_template(TypedExpr,Values,ColHeaders,Columns),
    append(ColHeaders,['Occurences', 'Witness ID'],AllColHeaders),
    get_texpr_type(TypedExpr,ExprType),
    (max_cardinality(ExprType,MaxTypeCard) -> true ; MaxTypeCard='??'),
    state_space:get_state_space_stats(NrNodes,_,_),
    format('Computing all values in initialised states (of ~w states in total) for expression: ~w [Max.Card=~w]~n',[NrNodes,Str,MaxTypeCard]),
    reduce_states_for_typed_expr(TypedExpr,false),
    format('Collecting values~n',[]),
    findall(list(StrVals),covered_values(StrVals,Values,Columns),Result),
    add_line_numbers(Result,LResult,TotalNrOfValuesFound).

tcltk_compute_nr_covered_values_for_all_variables(list([list(['Variable', 'Number of Values'])|VL])) :-
    state_space:get_state_space_stats(NrNodes,_,_),
    format('Computing all values in ~w states for all variables~n',[NrNodes]),
    findall(list([ID,Count]),number_of_values_for_variable(ID,Count),VL),
    format('Finished computing all values for all variables~n',[]).

number_of_values_for_variable(ID,Count) :-
    bmachine:b_is_variable(ID),
    reduce_states(state_value_for_variable(ID),false),
    get_counter(equiv_class_id,Count),
    format('~w : ~w~n',[ID,Count]).

reduce_states_for_typed_expr(TypedExpr,AssertEquivClass) :-
    get_texpr_id(TypedExpr,ID),
    bmachine:b_is_variable(ID),!,
    % use optimized, faster state_value_for_variable predicate 
    reduce_states(state_value_for_variable(ID),AssertEquivClass).
reduce_states_for_typed_expr(TypedExpr,AssertEquivClass) :-
    reduce_states(state_value_for_expr(TypedExpr),AssertEquivClass).

add_line_numbers(T,LT,Total) :- add_line_numbers_aux(T,0,LT,Total).
add_line_numbers_aux([],Nr,[],Nr) :- format('Found ~w different values.~n',[Nr]).
add_line_numbers_aux([list(H)|T],Nr,[list([N1|H])|LT],Total) :-
    N1 is Nr+1,
    add_line_numbers_aux(T,N1,LT,Total).

covered_values(AllColumns,Values,Columns) :-
    get_reduced_node(AbsState,Count,Witness,_EquivClassID),
    Witness \= root,
    Witness \= concrete_constants(_),
    AbsState \= undefined,
    % TO DO: maybe add further info: card for sets, dom, ran for relations ?
    translate_abs_value(AbsState,Values,Columns,StrVals),
    append(StrVals,[Count,Witness],AllColumns).

translate_abs_value(AbsState,Values,Columns,StrVals) :-
    AbsState=Values, !,
    maplist(mytranslate,Columns,StrVals).
translate_abs_value(undefined,_Values,Columns,StrVals) :- !,
    maplist(myundefined,Columns,StrVals).
translate_abs_value(AbsState,_Values,_Columns,[StrVal]) :-
    mytranslate(AbsState,StrVal).

myundefined(_,undefined).
mytranslate(Value,StrVal) :-
    translate:translate_bvalue_with_limit(Value,100,StrVal).

    
write_covered_values_for_expression_to_csvfile(VorE,File) :-
    print('Parsing and type checking'),nl,
    parse_expression(VorE,TypedExpr),
    translate:translate_bexpression_with_limit(TypedExpr,100,Str),
    format('Computing all values for expression: ~w~n',[Str]),
    reduce_states(state_value_for_expr(TypedExpr),false),
    format('Writing values to CSV file: ~w~n',[File]),
    open(File,write,Stream),
    format(Stream,'"~w",~w,~w,~n',[Str,'Occurences','WitnessStateID']),
    call_cleanup(write_covered_values_for_expression_to_csvfile_aux(Stream), close(Stream)),
    format('Finished writing values to CSV file: ~w~n',[File]).
write_covered_values_for_expression_to_csvfile(_VorECodes,File) :-
    add_error(state_space_reduction,'Computing Value Coverage Failed',File).

write_covered_values_for_expression_to_csvfile_aux(S) :-
    get_reduced_node(AbsState,Count,Witness,_EquivClassID),
    (Witness=root -> StrVal = '?'
     ; translate:translate_bvalue_with_limit(AbsState,100,StrVal)),
    format(S,'"~w",~w,~w~n',[StrVal,Count,Witness]),
    fail.
write_covered_values_for_expression_to_csvfile_aux(_).


% ---------------------------
% the generic state space reduction Algorithm:
% it takes two predicates:
%  AbstracAndFilterStates: it should succeed for those states that we want to process
%                          and then generate an abstract representation of it
%  AbstractAndFilterTransitions: it should succeed for all transtitions to be processed
%                       and also generate an abstract representation
%  Note: all states with the same abstract representation will be merged;
%        the same is true for transitions


reduce_state_space(AbstracAndFilterStates,
                   AbstractAndFilterTransitions) :-
    retractall(abs_signature(_,_)),
    % first compute equivalence class for every state
    reduce_states(AbstracAndFilterStates,true),
    % now inspect every transition for every node (that has been retained)
    equivalence_class(ID,EquivClassID), %print(id_class(ID,EquivClassID)),nl,
    findall(AbsAction/DestEquivClassID, 
             get_and_add_abstract_transition(ID,AbstractAndFilterTransitions,EquivClassID,
                                             AbsAction,DestEquivClassID),
             UnsortedSignature),
    sort(UnsortedSignature,Signature), %print(sig(Signature)),nl,
    update_signature(EquivClassID,Signature),
    fail.
reduce_state_space(_,_).

:- use_module(library(ordsets),[ord_intersection/3]).
% store those abstract transitions which are always enabled
:- dynamic abs_signature/2.
update_signature(EquivClassID,Signature) :-
   (abs_signature(EquivClassID,OldSignature)
    -> (Signature=OldSignature -> true
        ; OldSignature = [] -> true % already minimal
        ; retract(abs_signature(EquivClassID,OldSignature)),
          ord_intersection(OldSignature,Signature,NewSignature),
          %print(updated_signature(NewSignature)),nl,
          assert(abs_signature(EquivClassID,NewSignature))
        )
    ; assert(abs_signature(EquivClassID,Signature))).
% examine every outgoing transition, compute the abstract action and destination equivalence class
% and add a new transition in the reduced state space if necessary
get_and_add_abstract_transition(ID,AbstractAndFilterTransitions,EquivClassID,AbsAction,DestEquivClassID) :-
    transition(ID,Action,TransId,Destination),
    equivalence_class(Destination,DestEquivClassID),
    call(AbstractAndFilterTransitions,ID,Action,TransId,Destination,AbsAction),
    %print(trans(EquivClassID,DestEquivClassID,AbsAction)),nl,
    add_new_abstract_transition(EquivClassID,DestEquivClassID,ID,Action,TransId,Destination,AbsAction).

reduce_states(AbstracAndFilterStates,AssertEquivClass) :- 
    reset,
    % compute equivalence class for every state
    visited_expression(ID,State),
    call(AbstracAndFilterStates,ID,State,AbsState),
    % should succeed if state to be shown
    %print(state(ID,AbsState)),nl,
    add_new_abstract_node(ID,State,AbsState,AssertEquivClass,_EquivClassID),
    fail.
reduce_states(_,_).

get_reduced_node(AbsState,Count,Witness,EquivClassID) :-
    reduced_node(_Hash,AbsState,Count,Witness,EquivClassID).
add_new_abstract_node(ID,_State,AbsState,AssertEquivClass,EquivClassID) :-
   % TO DO: use hashing to improve lookup ! hashing:my_term_hash
   hashing:my_term_hash(AbsState,Hash),
   (retract(reduced_node(Hash,AbsState,Count,Witness,EquivClassID))
     ->   C1 is Count+1,
          assert(reduced_node(Hash,AbsState,C1,Witness,EquivClassID))
     ;    inc_counter(equiv_class_id,EquivClassID),
          (print_progress(EquivClassID,ID)
            -> format('% Number of equivalence classes: ~w. Treated ~w states.~n',[EquivClassID,ID]) ; true),
          assert(reduced_node(Hash,AbsState,1,ID,EquivClassID))
    ),
    (AssertEquivClass==false
     -> true
     ; assert(equivalence_class(ID,EquivClassID)),
       (current_expression(ID,_) -> assert(equivalence_class_contains_current_node(EquivClassID)) ; true)
     ).
print_progress(EquivClassID,_ID) :- EquivClassID mod 10 =:= 0 ;EquivClassID = 5.

% this version does not count (for efficiency reasons):
add_new_abstract_transition(EquivClassID,DestEquivClassID,_ID,_Action,_TransId,_Destination,AbsAction) :-
   (reduced_trans(EquivClassID,AbsAction,_Count,DestEquivClassID,_TransitionID)
     ->   true
     ;    inc_counter(transition_id,TransitionID),
          assert(reduced_trans(EquivClassID,AbsAction,1,DestEquivClassID,TransitionID))
    ).
% TO DO: make more efficient version by using C counters
/*
add_new_abstract_transition(EquivClassID,DestEquivClassID,_ID,_Action,_TransId,_Destination,AbsAction) :-
   (retract(reduced_trans(EquivClassID,AbsAction,Count,DestEquivClassID,TransitionID))
     ->   C1 is Count+1,
          assert(reduced_trans(EquivClassID,AbsAction,C1,DestEquivClassID,TransitionID))
     ;    inc_counter(1,TransitionID),
          assert(reduced_trans(EquivClassID,AbsAction,1,DestEquivClassID,TransitionID))
    ).
   */
