% (c) 2014-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(unsat_core_generalization,[unsat_core_generalization/7]).

:- use_module(probsrc(module_information)).
:- module_info(group,symbolic_model_checker).
:- module_info(description,'Counter example generalization based on unsat cores.').

:- use_module(library(heaps), [add_to_heap/4]).

:- use_module(probsrc(unsat_cores), [unsat_core_with_fixed_conjuncts/3]).
:- use_module(probsrc(translate), [translate_bexpression/2]).

:- use_module(probsrc(bsyntaxtree), [conjunct_predicates/2,
                                     create_negation/2]).
:- use_module(probsrc(translate), [translate_bexpression/2]).
:- use_module(probsrc(error_manager), [add_error/2]).

:- use_module(symbolic_model_checker(predicate_handling)).
:- use_module(symbolic_model_checker(solver_handling)).

% inspired by "efficient implementation of property directed reachability"
% see ic3 reference implementation
% s state to lift, primary inputs i, T transition relation, t successor
% => s & i & T & not t' is unsat
% extract core to find generalization
% currently, non-determinism in the transitions might render the predicate sat
% hence, no core can be extracted in that case
% the inputs should be used to make it deterministic!
unsat_core_generalization(HIn,N,Frames,State,Inputs,[SuccState|SuccStates],HOut) :-
    in_solver_on_level(N,Frames,IS),
    unsat_core_generalization_aux(IS,State,Inputs,SuccState,Core), !,
    add_to_heap(HIn,N,obligation(Core,Inputs,[SuccState|SuccStates]),HOut).

unsat_core_generalization_aux(IS,State,Inputs,SuccState,Core) :-
    prime_predicate(SuccState,TPrime),
    create_negation(TPrime,NotTPrime),
    conjunct_predicates([NotTPrime,Inputs|IS],Conjunction),

    % verify that the query is actually unsat
    conjunct_predicates([State,Conjunction],VerifyQuery),
    solve(VerifyQuery,Result),
    Result = contradiction_found(_), !,

    %debug output
    translate_bexpression(Conjunction,PPConj),
    translate_bexpression(State,PPState),
    translate_bexpression(Inputs,PPInputs),
    format('unsat core generalization: generalizing~n~w~nby computing unsat core of it with fixed conjuncts~n~w~nand inputs~n~w~n.',[PPState,PPConj,PPInputs]),

    % the unsat core is a lifting to a partial assignment
    unsat_core_with_fixed_conjuncts(State,Conjunction,Core),
    translate_bexpression(Core,PPCore),
    format('unsat core is~n~w~n',[PPCore]).
unsat_core_generalization_aux(_IS,State,_Inputs,_SuccState,Core) :-
    add_error(unsat_core_generalization, 'unsat core generalization failed. query is not unsat!'),
    Core = State.
