% (c) 2009-2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(tools, [exact_member/2,
                  exact_member_lookup/4, exact_member_replace/5,
                  remove/3, remove_all/3, insert/3,
                  remove_variables/3,
                  list_intersection/3, list_difference/3,
                  disjoint_list_union/3, lists_are_disjoint/2,
                  string_concatenate/3, is_upper_case_name/1,
                  write_to_file/2,
                  put_codes/2,
                  print_error/1,
                  print_message/1, print_message_with_max_depth/2,
                  print_short_msg/1,
                  print_bt_message/1,  bt_trace/1,
                  print_wtime/1,
                  prints/1,
                  print_dynamic_pred/3, print_dynamic_fact/1,
                  host_platform/1, platform_is_64_bit/0, max_tagged_integer/1,
                  translate_term_into_atom/2, translate_term_into_atom_with_max_depth/2,
                  get_parent_directory/2, get_tail_filename/2, get_modulename_filename/2, 
                  get_filename_extension/2,
                  get_options/4,get_options/5,
                  arg_is_number/2, arg_is_number_or_wildcard/2,
                  arg_is_integer/2,
                  split_filename/3,
                  safe_absolute_file_name/2, safe_absolute_file_name/3,

                  filter/4,
                  flatten/2,
                  count_occurences/2,
                  split_atom/3,
                  string_escape/2, print_escaped/1, % can also be used for dot
                  read_string_from_file/2,
                  write_lines_to_file/2,

                  predicate_functor/3,

                  is_list_simple/1,
                  ajoin/2,
                  ajoin_with_limit/3,
                  ajoin_with_sep/3,
                  substitute/4, call_residue/2,

                  catch_call/1, safe_on_exception/3,
                  reraise_important_exception/1,
                  catch_matching/3,

                  safe_numbervars/3,
                  convert_list_into_pairs/2, convert_pairs_into_list/3,

                  safe_univ/2,
                  safe_sort/3,
                  safe_functor/4,

                  safe_atom_codes/2, safe_atom_chars/3,
                  atom_codes_with_limit/2, atom_codes_with_limit/3,
                  truncate_atom/3, wrap_and_truncate_atom/4,
                  safe_number_codes/2,
                  ensure_atom/2,

                  print_size_of_table/1,
                  print_runtime/0, start_ms_timer/1, stop_ms_timer/2, stop_ms_timer/1,
                  cputime/1, walltime/1,

                  retract_with_statistics/2,
                  get_memory_used/1, print_memory_used_difference/2, print_memory_used/1,
                  print_memory_used_wo_gc/0,

                  split_list/4,foldl/4,foldl/5,foldl/6,
                  average/2,

                  map_over_bvalue/3,

                  (space_call)/1,
                  assert_once/1,
                  
                  safe_time_out/3,

                  unique_id/2,

                  cputime/1,

                  get_PROBPATH/1
                 ]).

:- use_module(module_information).

:- module_info(group,infrastructure).
:- module_info(description,'This module contains many general helper predicates.').

:- use_module(pathes,[]). % we just import it to set-up pathes, we don't need any predicates
              
:- use_module(library(lists)).
:- use_module(library(system)).
:- use_module(library(file_systems)).
:- use_module(library(codesio)).

:- meta_predicate catch_call(0).
catch_call(Call) :-
  on_exception(Exception,call(Call),
       (add_error(catch_call,'Call raised an exception: ',(Call:Exception)),
        /* read(_), */
        reraise_important_exception(Exception),
        fail)).
 
:- meta_predicate safe_on_exception(*,0,0).
% use if you want to catch any exception; ensures time_out not treated and passed on
safe_on_exception(E,Call,ExcCode) :- 
  on_exception(E,call(Call), (print(exception(E)),nl,
                        reraise_important_exception(E),ExcCode)).
  
% reraise important exceptions
reraise_important_exception(time_out) :- !, raise_exception(time_out).
reraise_important_exception(_).

% like catch/3, but it does not fail if an exception occures that is not
% unifiable with the second argument. Instead it re-throws the original
% exception.
:- meta_predicate catch_matching(0,*,0).
catch_matching(Call,Exception,Handler) :-
    catch(Call, E, (E=Exception -> Handler ; throw(E))).





:- use_module(self_check). % put after search paths have been set

% --------------------------------------

cputime(T) :-
    statistics(runtime,[T,_]).

walltime(WT) :-
    statistics(walltime,[WT,_]).
% --------------------------------------

platform_is_64_bit :- 
  prolog_flag(max_tagged_integer,X),
  X >= 1152921504606846975.

% equals 268435455 on 32-bit systems, i.e., 2**28 - 1
% equals 1152921504606846975 on 64-bit systems 
max_tagged_integer(X) :- prolog_flag(max_tagged_integer,X).


host_platform(Res) :-
  prolog_flag(host_type,HostType),
  name(HostType,AsciiList),
  map_host_platform(AsciiList,Res),!.

map_host_platform(HT,darwin) :- append("powerpc-darwin",_,HT).
map_host_platform(HT,darwin) :- append("i386-darwin",_,HT).
map_host_platform(HT,darwin) :- append("x86_64-darwin",_,HT).
map_host_platform(HT,windows) :- append("x86-win",_,HT).
map_host_platform(HT,linux) :- append("x86-linux",_,HT).
map_host_platform(_,unknown).

% on Windows XP: 'x86-win32-nt-4'
% on MacPro: 'x86_64-darwin-10.6.0'
% on Linux:  'x86-linux-glibc2.7'

write_lines_to_file(FILE,Lines) :- 
   lines_to_code_list(Lines,CodeList),
   write_to_file(FILE,CodeList).

lines_to_code_list([],[]).
lines_to_code_list([Line|Lines],CodeList) :- 
   (number(Line) -> number_codes(Line,L); atom_codes(Line,L)),
   append(L,[10],L1),
   append(L1,RL,CodeList),
   lines_to_code_list(Lines,RL).

write_to_file(FILE,CodesList) :-
   open(FILE,write,S),
   put_codes(CodesList,S),
   close(S).

put_codes([],_).
put_codes([H|T],S) :- put_code(S,H), put_codes(T,S).

print_error(Error) :-
    current_output(X),
    set_output(user_error),
    (var(Error)  -> print_message(error,'_')
    ;  write('! '),write_term(Error,[max_depth(20),numbervars(true)]),nl),
    %% flush_output(user_error), %%
    set_output(X).

:- use_module(error_manager,[add_error/3, add_internal_error/2]).
print_message(Msg) :- print_message_with_max_depth(Msg,20).
print_message_with_max_depth(Msg,MaxDepth) :-
   safe_on_exception(E,print_message2(Msg,MaxDepth),
                       add_error(tools,'Exception in print_message: ',E)). % added because sometimes in Windows/Vista we get an exception here
print_message2(Msg,MaxDepth) :-
    current_output(X),
    set_output(user),
    (var(Msg) -> print_message(informational,'_')
    ; (write('% '),write_term(Msg,[max_depth(MaxDepth)]),nl)),
    set_output(X).
print_short_msg(Msg) :-
    current_output(X),
    set_output(user),
    write(Msg),
    set_output(X).

print_bt_message(Msg) :- print_message(Msg).
print_bt_message(Msg) :- print_message(backtrack(Msg)),
   %(Msg = found_enumeration_of_constants(_,_) -> trace ; true),
   fail.

% trace upon backtrack:
bt_trace(_) :- true.
bt_trace(PP) :- print(' * BACKTRACK: '), print(PP),nl, trace,fail.

print_wtime(PP) :- statistics(walltime,[WT,_]), 
    current_output(X),
    set_output(user),
    print(PP), print(' : '), print(WT), print(' ms'),nl,
    set_output(X).

% a print that will automatically stop after 25 prints and give the user the option to inspect the printed messages
:- dynamic prints_count/1.
prints_count(25).
prints(L) :- print_bt_message(L),
  (retract(prints_count(X)) -> true ; X=25),
  (X<1
    -> print('*** Stopped printing >'),
       read(RT),
       (number(RT) -> X1=RT ; X1=25)
    ;  X1 is X-1
  ), assert(prints_count(X1)).

:- assert_pre(tools:exact_member(_Var,Vs),
                       (list_skeleton(Vs))).
:- assert_post(tools:exact_member(_Var,_Vs), true).
:- assert_must_succeed(tools:exact_member(V,[V])).
:- assert_must_succeed(tools:exact_member(V,[_X,_Z,V])).
:- assert_must_fail(tools:exact_member(_W,[_X,_Z,_V])).
:- assert_must_fail(tools:exact_member(_W,[])).

exact_member(X,[Y|T]) :-
   ((X==Y) -> true ; exact_member(X,T)).
   

:- assert_pre(tools:exact_member_lookup(_Var,_ValRes,Vs,Vals),
                       (list_skeleton(Vs),list_skeleton(Vals))).
:- assert_post(tools:exact_member_lookup(_Var,_ValRes,_Vs,_Vals), true).
:- assert_must_succeed(tools:exact_member_lookup(V,2,[V],[2])).
:- assert_must_succeed(tools:exact_member_lookup(V,2,[_X,_Z,V],[1,3,2])).
:- assert_must_fail(tools:exact_member_lookup(V,3,[_X,_Z,V],[1,3,2])).
:- assert_must_fail(tools:exact_member_lookup(_W,3,[_X,_Z,_V],[1,3,2])).
:- assert_must_fail(tools:exact_member_lookup(_W,3,[],[1,3,2])).
                       
exact_member_lookup(Var,ValRes,[V|TV],[Val|TVal]) :-
   ((Var==V) -> (ValRes=Val) ; exact_member_lookup(Var,ValRes,TV,TVal)).
   

:- assert_pre(tools:exact_member_replace(_Var,_ValRes,Vs,Vals,_),
                       (list_skeleton(Vs),list_skeleton(Vals))).
:- assert_post(tools:exact_member_replace(_Var,_ValRes,_Vs,_Vals,NewVals),
                       list_skeleton(NewVals)).
:- assert_must_succeed(tools:exact_member_replace(V,44,[_X,_Z,V],[1,3,2],[1,3,44])).
:- assert_must_fail(tools:exact_member_replace(_V,44,[_X,_Z,_VV],[1,3,2],[1,3,44])).

exact_member_replace(Var,NewVal,[V|TV],[Val|TVal],[NV|TN]) :-
   ((Var==V) -> (NV=NewVal,TN=TVal)
            ;   (NV=Val,exact_member_replace(Var,NewVal,TV,TVal,TN))).
   


remove([X|T],X,T).
remove([Y|T],X,[Y|DT]) :-  \+(X=Y),   remove(T,X,DT).

:- assert_must_succeed(tools:flatten([],[])).
:- assert_must_succeed(tools:flatten([[1,2,3]],[1,2,3])).
:- assert_must_succeed(tools:flatten([[1],[[99]],[2,3],[4,5]],[1,99,2,3,4,5])).
:- assert_must_succeed(tools:flatten([[]],[])).

flatten(List,FlatList) :- flatten1(List,[],FlatList).
flatten1([],L,L) :- !.
flatten1([H|T],Tail,List) :- !, flatten1(H,FlatList,List), flatten1(T,Tail,FlatList).
flatten1(NonList,Tail,[NonList|Tail]).


:- assert_must_succeed((tools:count_occurences([a,b,a,a,b],R),R == [a-3,b-2])).
:- use_module(library(avl)).
% count number of occurences inside a list
count_occurences(L,R) :- empty_avl(E), count_occ_aux(L,E,R).
count_occ_aux([Term|T],A,Res) :- (avl_fetch(Term,A,Count) -> C is Count+1 ; C = 1),
   avl_store(Term,A,C,A2),
   count_occ_aux(T,A2,Res).
count_occ_aux([],A,L) :- avl_to_list(A,L).


:- assert_must_succeed((tools:filter(var,[1,2,_X,3,_Y],R,Out),R == [_X,_Y],Out == [1,2,3])).
:- assert_must_succeed((tools:filter(nonvar,[1,2,_X,3,_Y],R,Out),R == [1,2,3],Out == [_X,_Y])).

:- meta_predicate filter(1,*,*,*).
filter(_Pred,[],[],[]).
filter(Pred,[H|T],True,False) :-
    (call(Pred,H) -> True = [H|TT], filter(Pred,T,TT,False)
      ; False = [H|FF], filter(Pred,T,True,FF)).

:- assert_must_succeed((tools: filter(tools:is_a_comment('/*','*/'),['This comment ','/* comment */','will be ignored!'],R,Out),
                        R==['/* comment */'],Out==['This comment ','will be ignored!'])).
% checking whether an atom is a comment /* ... */
is_a_comment(Begin,End,Comment) :-
    maplist(atom_codes,[Begin,End,Comment],[BL,EL,CL]),
    prefix(CL,BL),suffix(CL,EL).

:- assert_must_succeed((tools:remove_variables([X,Y,Z],[Y],R),R==[X,Z])).
remove_variables(List,Vars,Remaining) :-
    exclude(exact_member_rev(Vars),List,Remaining).
% a version of exact member with reversed parameters - usefull with
% higher-order functions
exact_member_rev(List,Member) :- exact_member(Member,List).

:- assert_must_succeed(tools:insert([],a,[a])).
:- assert_must_succeed(tools:insert([a,b,c,d],a,[a,b,c,d])).
:- assert_must_succeed(tools:insert([a,b,c,d],b,[a,b,c,d])).
:- assert_must_succeed(tools:insert([a,b,c,d],d,[a,b,c,d])).
:- assert_must_succeed(tools:insert([a,b,c,d],x,[a,b,c,d,x])).
insert([],X,[X]).
insert([H|T],X,R) :- (H=X -> R=[H|T] ; R=[H|R2],insert(T,X,R2)).


:- assert_must_succeed(tools:list_intersection([a,b,c,d],[d,f,b],[b,d])).

list_intersection([],_L,[]).
list_intersection([H|T],L,Res) :-
   (remove(L,H,NL) -> (Res=[H|RR]) ; (Res=RR,NL=L)),
   list_intersection(T,NL,RR).

insert_new([],X,[X]).
insert_new([H|T],X,R) :- (H=X -> fail ; R=[H|R2],insert_new(T,X,R2)).

:- assert_must_succeed(tools:disjoint_list_union([a,b,c,d],[e,f,g],[e,f,g,a,b,c,d])).
:- assert_must_fail(tools:disjoint_list_union([a,b,c,d],[d,f,b],_)).
:- assert_must_fail(tools:disjoint_list_union([a,b,c,d],[d,f,b],[d,f,b,a,c])).
disjoint_list_union([],L,L).
disjoint_list_union([H|T],L,Res) :- insert_new(L,H,L2), disjoint_list_union(T,L2,Res).

:- assert_must_succeed(tools:lists_are_disjoint([a,b,c,d],[e,f,g])).
:- assert_must_succeed(tools:lists_are_disjoint([a,b,c,d],[])).
:- assert_must_succeed(tools:lists_are_disjoint([],[e,f,g])).
:- assert_must_fail(tools:lists_are_disjoint([a,b,c,d],[e,f,g,d])).
lists_are_disjoint([],_).
lists_are_disjoint([H|T],List2) :- \+ member(H,List2), lists_are_disjoint(T,List2).

:- assert_must_succeed(tools:list_difference([a,b,c,d],[b,f,d],[a,c])).

list_difference([],_L,[]).
list_difference([H|T],L,Res) :-
   (remove(L,H,NL) -> (Res=RR) ; (Res=[H|RR],NL=L)),
   list_difference(T,NL,RR).

:- assert_must_succeed(( tools:string_concatenate('5','.10',R), R=='5.10' )).

:- mode string_concatenate(i,i,o).
string_concatenate(X,Y,XY) :- atom(X),atom(Y),!, atom_concat(X,Y,XY).
string_concatenate(X,Y,XY) :- 
    safe_name(X,Xs),safe_name(Y,Ys),append(Xs,Ys,XYs),atom_codes(XY,XYs).
   

   
safe_name(X,N) :- atomic(X),!, name(X,N).
safe_name(X,N) :- var(X),!,atom_codes(var,N).  /* ensure that "5.10" is converted to '5.10' and not 5.1 */
safe_name(lambda_res(X),[114,101,115,95|N]) :- !, name(X,N).
safe_name(X,N) :- functor(X,F,_),name(F,N), print_error(non_atomic_in_safe_name(X)).
   

predicate_functor(X,F,N) :- var(X),!, print_error(var_in_predicate_functor),F='$VAR',N=0.
predicate_functor(_Module:Pred,F,N) :- !,predicate_functor(Pred,F,N).
predicate_functor(P,F,N) :- functor(P,F,N).
   
safe_numbervars(Term,Start,End) :-
  on_exception(E,numbervars(Term,Start,End), 
               (print('Exception during numbervars: '),print(E),nl,nl,reraise_important_exception(E))).


 
translate_term_into_atom(CTerm,Atom) :- 
           copy_term(CTerm,Term), safe_numbervars(Term,0,_),
           write_term_to_codes(Term,Temp,[quoted(true),numbervars(true)]),
                       name(Atom,Temp).
translate_term_into_atom_with_max_depth(Term,Atom) :- atomic(Term),!,Atom=Term.                       
translate_term_into_atom_with_max_depth(Term,Atom) :- 
                %copy_term(CTerm,Term),safe_numbervars(Term,0,_),
                %write_term_to_codes(Term,Temp,[quoted(true),numbervars(true),max_depth(20)]),
                write_term_to_codes(Term,Temp,[quoted(true),numbervars(true),max_depth(20)]),
                atom_codes(Atom,Temp).
   
:- assert_must_succeed(tools:is_upper_case_name('GOODS')).
:- assert_must_succeed(tools:is_upper_case_name('ZZAA')).
:- assert_must_fail(tools:is_upper_case_name('capacity')).
:- assert_must_fail(tools:is_upper_case_name('PARAs')).

is_upper_case_name(Name) :-
  safe_name(Name,AsciiList),
  upper_case_list(AsciiList).
  
upper_case_list([]).
upper_case_list([H|T]) :- H >="A", H=<"Z", upper_case_list(T).   
   
print_dynamic_pred(InModule,Pred,N) :- nl,
   write(':- dynamic '), print(Pred), write('/'),print(N), write('.'),nl,
   functor(Call,Pred,N),
   call(InModule:Call), write_term(Call,[quoted(true)]),write('.'),nl,fail.
print_dynamic_pred(InModule,Pred,N) :- 
   functor(Call,Pred,N),
   (call(InModule:Call) -> true ; (write_with_numbervars(Call), write(' :- fail.'),nl)),
   nl.

write_with_numbervars(Term) :- copy_term(Term,T), numbervars(T,0,_),
   write_term(T,[quoted(true),numbervars(true)]).
print_dynamic_fact(Fact) :-
   functor(Fact,Pred,N),
   print(':- dynamic '), print(Pred), write('/'),print(N), write('.'),nl,
   write_term(Fact,[quoted(true)]),write('.'),nl.



:- assert_must_succeed(tools:get_parent_directory('/aaaa/bbb/cc/d.app','/aaaa/bbb/cc/')).
:- assert_must_succeed(tools:get_parent_directory('/aaaa/bbb/cc/','/aaaa/bbb/cc/')).
    
get_parent_directory(Path,NewPath) :- 
   atom_chars(Path,PathAscii),
   strip_last(PathAscii,[],[],New),
   atom_chars(NewPath,New).
   
:- use_module(library(lists)).

strip_last([],ResSoFar,_,Res) :- reverse(ResSoFar,Res).
strip_last(['/'|Tail],ResSoFar,StripSoFar,Res) :- !,
  append(['/'|StripSoFar],ResSoFar,NewRes),
  strip_last(Tail,NewRes,[],Res).
strip_last([A|Tail],ResSoFar,StripSoFar,Res) :-
  strip_last(Tail,ResSoFar,[A|StripSoFar],Res).

:- assert_must_succeed((tools:split_atom('ef,g',[','],R), R==[ef,g])).
:- assert_must_succeed((tools:split_atom('ef, g',[',',' '],R), R==[ef,g])).
:- assert_must_succeed((tools:split_atom('ab,cd,ef,g',[','],R), R==['ab','cd',ef,g])).
:- assert_must_succeed((tools:split_atom('ab,cd,ef;g',[',',';'],R), R==['ab','cd',ef,g])).

split_atom(Atom,SepList,SplitList) :-
  atom_chars(Atom,ListAscii), 
  split2(ListAscii,SepList,SplitList).

split2([],_,R) :- !,R=[].
split2(List,Sep,Res) :- get_next_word(List,Sep,Word,Tail),!,
  (Word=[] -> split2(Tail,Sep,Res)
   ; Res=[Atom|TA], atom_chars(Atom,Word), split2(Tail,Sep,TA)).

get_next_word([],_Sep,[],[]).
get_next_word([H|T],Sep,Word,Tail) :-
  member(H,Sep) -> Word=[],Tail=T
    ; Word=[H|TR], get_next_word(T,Sep,TR,Tail).

:- assert_must_succeed(tools:split_last('/aaaa/bbb/cc/d.app','/','/aaaa/bbb/cc','d.app')).
:- assert_must_succeed(tools:split_last('/aaaa/bbb/cc/d.app','.','/aaaa/bbb/cc/d','app')).
split_last(Atom, Sep,  Head, Tail) :-
   atom_chars(Atom,ListAscii), atom_chars(Sep,[SepACode]),
   split_last2(ListAscii,SepACode,[],[],HeadA, TailA),
   atom_chars(Head,HeadA), atom_chars(Tail,TailA).

split_last2([],Sep,CurSplit,[Sep|Head],ResH,ResT) :-
   reverse(CurSplit,ResT),
   reverse(Head,ResH).
split_last2([Sep|Tail],Sep,CurSplit,Head,ResH,ResT) :- !,
   append([Sep|CurSplit],Head,NewHead),
   split_last2(Tail,Sep,[],NewHead,ResH,ResT).
split_last2([H|Tail],Sep,CurSplit,Head,ResH,ResT) :-
   split_last2(Tail,Sep,[H|CurSplit],Head,ResH,ResT).

 
 :- assert_must_succeed(tools:split_filename('/aaaa/bbb/cc/d.app','/aaaa/bbb/cc/d','app')).
:- assert_must_succeed((Z='/aaaa/bbb/cc/d',tools:split_filename(Z,R,X),X=='',R==Z)).

split_filename(Filename,Base,Ext) :-
  (split_last(Filename,'.',Base,Ext) -> true ; Base=Filename,Ext='').

  
:- assert_must_succeed(tools:get_tail_filename('/aaaa/bbb/cc/d.app','d.app')).
:- assert_must_succeed(tools:get_tail_filename('d.app','d.app')).
:- assert_must_succeed(tools:get_tail_filename('/aaaa/bbb/cc/','')).
get_tail_filename(Path,Tail) :- (split_last(Path, '/',  _, T) -> Tail=T ; Tail=Path).

:- assert_must_succeed(tools:get_modulename_filename('/aaaa/bbb/cc/d.app','d')).
:- assert_must_succeed(tools:get_modulename_filename('d.app','d')).
get_modulename_filename(Path,Module) :- 
    get_tail_filename(Path,Tail),
    (split_last(Tail, '.',  M, _) -> Module=M ; Module=other).


:- assert_must_succeed(get_filename_extension('/aaaa/bbb/cc/d.app','app')).
get_filename_extension(Path,Ext) :- split_filename(Path,_,Ext).

% also works if numbers passed (can happen by accident in test_runner ...)
safe_absolute_file_name(F,AF,Options) :-
   ensure_atom(F,A),
   absolute_file_name(A,AF,Options).

safe_absolute_file_name(F,AF) :-
   ensure_atom(F,A),
   absolute_file_name(A,AF).



%*******************************************************************************
% remove_all(A,B,Result): Result is the list of elements of A which do not occure in B
remove_all([],_,[]).
remove_all([H|T],Remove,Result) :-
    (member(H,Remove) -> !,Result = Rest ; Result = [H|Rest]),
    remove_all(T,Remove,Rest).

%*******************************************************************************
% get_options/4 for parsing command line arguments

:- meta_predicate get_options(+,4,-,-).
:- meta_predicate get_options(+,4,-,-,0).

get_options(List,Pred,Options,Rest) :-
   get_options(List,Pred,Options,Rest,halt).  
get_options([],_,[],[],_).
get_options([X|T],Module:Recognised,Options,Args,HALTCMD) :-
    Call =.. [Recognised,X,Opt,Values,Action],
    ( call(Module:Call) ->
        ( append(Values, Rest, T) -> true
        ; otherwise ->
            nl, print('Missing/Invalid argument(s) for option: '), print(X),nl,HALTCMD),
        ( call(Action) -> true
        ; otherwise ->
            nl,print('Invalid argument(s) for option '),print(X),
            print(': '),print(Values),nl,HALTCMD),
        RT = Rest,  
        Options = [Opt|OT], Args = AT  
    ; otherwise ->
        Options = OT, Args = [X|AT],
        RT = T
    ),
    get_options(RT,Module:Recognised,OT,AT,HALTCMD).

arg_is_number(Arg,Nr) :- name(Arg,Str),name(Nr,Str),number(Nr).
arg_is_integer(Arg,Nr) :-
    ( append("-",SPos,Arg) -> % negative number
        arg_is_number(SPos,Pos),
        Nr is -Pos
    ; otherwise ->
        arg_is_number(Arg,Nr)).


arg_is_number_or_wildcard('_',R) :- !, R=_.
arg_is_number_or_wildcard(Arg,N) :- arg_is_number(Arg,N).

/* for Ciao 
:- use_module(library(dec10_io)).
:- use_module(library(write)).

prefix([],_).
prefix([H|T],[H|Y]) :- prefix(T,Y).
*/




/* from File: sp4_compatibility_mappings.pl */
/* Created: 08/05/2007 by Michael Leuschel */

:- meta_predicate call_residue(0,*).

call_residue(X,Residue) :- call_residue_vars(X,V),filter_residue_vars(V,Residue).


filter_residue_vars([],[]).
filter_residue_vars([H|T],Res) :-
  frozen(H,FH),
  (FH=true -> Res=RT ; Res = [FH|RT]),
  filter_residue_vars(T,RT).
  

:- assert_must_succeed( tools:convert_list_into_pairs([a],a)).
:- assert_must_succeed( (tools:convert_list_into_pairs([a,b,c],R), R = ((a,b),c) )).

convert_list_into_pairs([X|T],Res) :- conv2(T,X,Res).
convert_list_into_pairs([],[]).
conv2([],X,X).
conv2([X|T],Acc,Res) :- conv2(T,(Acc,X),Res).

:- assert_must_succeed(( tools:convert_pairs_into_list([x],a,R), R==[a] )).
%:- assert_must_succeed(( b_interpreter:convert_pairs_into_list([x,y,z],((a,b),c),R), R == [a,b,c] )).
:- assert_must_succeed(( tools:convert_pairs_into_list([x,y,z],(a,(b,c)),R), R == [a,b,c] )).
:- assert_must_succeed(( tools:convert_pairs_into_list([x,y,z],A,B),nonvar(A),nonvar(B),A=((_,_),_),B=[_,_,_] )).

% the first argument just indicates the length of the list
convert_pairs_into_list([_],X,R) :- !,R=[X].
convert_pairs_into_list([_|Guide],(A,B),[A|Rest]) :-
    convert_pairs_into_list(Guide,B,Rest).
    
    
    

/* ex: substitute(1, [1,2,3,4], 5, X). */
:- assert_must_succeed(( tools:substitute(1, [1,2,3,4], 5, X), X==[5,2,3,4])).
substitute(X,L,Y,Res) :- sub(L,X,Y,Res).
sub([],_,_,[]).
sub([H|T],X,Y,[SH|ST]) :-
  (H=X -> SH=Y ; SH=H),
  sub(T,X,Y,ST).


/* Checks if the argument is a list, but unlike is_list/1 it just
   checks the head and does not iterate through the list */
is_list_simple([]).
is_list_simple([_|_]).

:- assert_must_succeed((tools: ajoin_with_sep([link,a,xa],'.',Text), Text == 'link.a.xa')).
:- assert_must_succeed((tools: ajoin_with_sep([link],'.',Text), Text == 'link')).

ajoin_with_sep(List,Sep,Text) :- ajoin_with_sep_aux(List,'',Sep,Text).

ajoin_with_sep_aux([],Txt,_Sep,Text) :- Text=Txt.
ajoin_with_sep_aux([H|T],Txt,Sep,Text) :- 
   ( Txt == '' -> toAtom(H,TTxt)
   ; ajoin([Txt,Sep,H],TTxt)), ajoin_with_sep_aux(T,TTxt,Sep,Text).

/* Concats a list of atoms, but checks if numbers or compound terms are
   involved, which are converted to simple atoms */
ajoin([],Text) :- Text=''.
ajoin([A|T],Result) :- toAtom(A,Atom),
    ajoin_aux(T,Atom,Result).
ajoin_aux([],Acc,Result) :- Result=Acc.
ajoin_aux([A|T],Acc,Result) :- toAtom(A,Atom),
    safe_atom_concat(Acc,Atom,NewAcc,ExcOcc),
    (ExcOcc==true -> Result = NewAcc ; ajoin_aux(T,NewAcc,Result)).
toAtom(Number,Atom) :- number(Number),!,number_chars(Number,C),atom_chars(Atom,C).
toAtom(Atom,Atom) :- atomic(Atom),!.
toAtom(Term,Atom) :- write_to_codes(Term,Codes),safe_atom_codes(Atom,Codes).

% will concatenate until the Limit is reached or exceeded; it may produce atoms longer than Limit 
% (if first atom already longer than limit + it adds ...
:- assert_must_succeed((tools: ajoin_with_limit(['A','B','C','D'],100,Text), Text == 'ABCD')).
:- assert_must_succeed((tools: ajoin_with_limit(['A','B','C','D'],2,Text), Text == 'AB...')).

ajoin_with_limit([],_,Text) :- 
        !,Text=''.
ajoin_with_limit([A|T],Limit,Result) :- toAtom(A,Atom), atom_length(Atom,Len),
    ajoin_with_limit_aux(T,Atom,Len,Limit,Result).
ajoin_with_limit_aux([],Acc,_,_,Result) :- !,Result=Acc.
ajoin_with_limit_aux([A|T],Acc,Len,Limit,Result) :- toAtom(A,Atom),
    atom_length(Atom,AL), L1 is Len+AL,
    (L1 > Limit -> safe_atom_concat(Acc,'...',Result,_)
      ; safe_atom_concat(Acc,Atom,NewAcc,ExcOcc),
        (ExcOcc==true -> Result = NewAcc ; ajoin_with_limit_aux(T,NewAcc,L1,Limit,Result))).

:- use_module(error_manager,[add_error_and_fail/3]).
safe_univ(Term,List) :- nonvar(Term),!,Term=..List.
safe_univ(Term,List) :- var(List), !,add_error_and_fail(tools,'Arguments to safe_univ (=..) both var:', safe_univ(Term,List)).
safe_univ(Term,List) :- %Term is a variable
                        prolog_flag(max_arity,MA), cut_off_list(List,MA,CL),
                        !, % avoid pending choice points
                        Term =.. CL.


safe_atom_chars(A,B,Loc) :-
    on_exception(error(E1,E2),atom_chars(A,B),
       (add_internal_error('atom_chars error: ',Loc:E1),
        throw(error(E1,E2)))).

safe_atom_concat(A,B,C,ExceptionOcc) :-
    on_exception(error(representation_error(max_atom_length),_),atom_concat(A,B,C),
       (print(exception(max_atom_length)),nl,A=C,ExceptionOcc=true)).

safe_atom_codes(V,C) :- var(V),var(C),!,
   add_internal_error('Variables in call: ',safe_atom_codes(V,C)), C='$VARIABLE$'.
safe_atom_codes(A,C) :-
    on_exception(error(representation_error(max_atom_length),_),atom_codes(A,C),
       (print(exception(max_atom_length)),nl,atom_codes_with_limit(A,1000,C))).

safe_number_codes(V,C) :- var(V),var(C),!,
   add_internal_error('Variables in call: ',safe_number_codes(V,C)), C='$VARIABLE$'.
safe_number_codes(A,C) :-
    on_exception(error(syntax_error(N),_),number_codes(A,C),
       (print(syntax_error_in_number_codes(N)),nl,fail)). % in this case safe_number_codes fails ; we cannot convert the codes into a number

:- assert_must_succeed(ensure_atom(19,'19')).
% ensure that numbers get converted to atoms:
ensure_atom(Var,A) :- var(Var),!, A='_'.
ensure_atom(N,Res) :- number(N),!,number_codes(N,C), atom_codes(A,C), Res=A.
ensure_atom(A,A).

atom_codes_with_limit(A,C) :-
    on_exception(error(representation_error(max_atom_length),_),
                  atom_codes(A,C), 
       (print(exception(max_atom_length)),nl,atom_codes_with_limit(A,1000,C))).
      

atom_codes_with_limit(A,Limit,Codes) :- var(A),!,
    truncate_codes(Codes,Limit,TCodes,_),
    atom_codes(A,TCodes).
atom_codes_with_limit(A,_,Codes) :- atom_codes(A,Codes).

truncate_codes([],_,[],false).
truncate_codes([H|T],Count,Res,Trunc) :-
   Count<1 -> Res = [46,46,46],Trunc=true /* '...' */ 
           ; Res = [H|TT], C1 is Count-1, truncate_codes(T,C1,TT,Trunc).

:- assert_must_succeed((truncate_atom(abcd,100,Text), Text == 'abcd')).
:- assert_must_succeed((truncate_atom(abcd,2,Text), Text == 'ab...')).
:- assert_must_succeed((truncate_atom(abcd,0,Text), Text == '...')).

truncate_atom(Atom,Limit,NewAtom) :- 
   atom_codes(Atom,Codes),
   truncate_codes(Codes,Limit,TCodes,Trunc),
   (Trunc=true -> atom_codes(NewAtom,TCodes) ; NewAtom=Atom).
   
wrap_and_truncate_atom(Atom,LineLength,Limit,NewAtom) :- \+ atom(Atom),!,
   add_internal_error('Argument to wrap_and_truncate_atom not atom: ',wrap_and_truncate_atom(Atom,LineLength,Limit,NewAtom)),
   NewAtom=Atom.
wrap_and_truncate_atom(Atom,LineLength,Limit,NewAtom) :-
   atom_codes(Atom,Codes),
   wrap_and_truncate_codes(Codes,LineLength,LineLength,Limit,NewCodes,Chng),
   (Chng=true -> atom_codes(NewAtom,NewCodes) ; NewAtom=Atom).
 
wrap_and_truncate_codes([],_,_,_,[],false).
wrap_and_truncate_codes([H|T],CharsOnLineCount,MaxLineLength,TotCount,Res,Chng) :-
   (TotCount<1 -> Res = [46,46,46],Chng=true /* '...' */ 
    ; (CharsOnLineCount<1 -> Res=[92,110,H|TT],  /* add newline \n */
                             TR=T, L1 is MaxLineLength, Chng=true 
        ; H=92,T=[H2|T2] -> Res = [H,H2|TT], /* do not split escaped char */
                            TR=T2, L1 is CharsOnLineCount-1, Chng2=Chng
        ; Res = [H|TT], TR=T, L1 is CharsOnLineCount-1, Chng2=Chng),
       TC1 is TotCount-1, wrap_and_truncate_codes(TR,L1,MaxLineLength,TC1,TT,Chng2)
   ).
   

cut_off_list([],_,[]).
cut_off_list([H|T],MA,R) :- (MA<2 -> R=['...'] ; R=[H|TR],MA1 is MA-1, cut_off_list(T,MA1,TR)).


% print_size_of_table counts the number of succeeded calls of the
% given predicate
get_calls_for_table(Module:P/N,Call) :- !,
   functor(Call,P,N),
   call(Module:Call).
get_calls_for_table(P/N,Call) :-
   functor(Call,P,N),
   call(Call).

:- volatile count/1.
:- dynamic count/1.
print_size_of_table(Pred) :- retractall(count(_)),
  assert(count(0)),get_calls_for_table(Pred,_),inc_size_of_table,fail.
print_size_of_table(Pred) :- 
     print('% size of table for '), print(Pred), print(': '), count(X), print(X),nl.
inc_size_of_table :- retract(count(X)),X1 is X+1, assert(count(X1)).


safe_sort(Orig,A,B) :- var(A),!, add_error(tools,'Illegal call: ',mysort(Orig,A,B)),A=B.
safe_sort(_,A,B) :- sort(A,B).

safe_functor(Src,F,A,Term) :- var(F),var(A),var(Term),!,
   add_internal_error('Illegal functor call: ',safe_functor(Src,F,A,Term)),
   fail.
safe_functor(_,F,A,T) :- functor(F,A,T).


print_runtime :- statistics(runtime,[Tot,SinceLast]), print(' Total runtime: '), print(Tot),
  print(' ms, since last: '), print(SinceLast), print(' ms'),nl.


start_ms_timer(timer(R,T,W)) :- statistics(runtime,[R,_]),
   statistics(total_runtime,[T,_]),
   statistics(walltime,[W,_]).
stop_ms_timer(T) :- stop_ms_timer(T,[runtime/RT,total_runtime/RTT,walltime/WT]),
   format('% Runtime: ~w ms (total: ~w ms, walltime: ~w ms)~n',[RT,RTT,WT]).
stop_ms_timer(timer(R,T,W),[runtime/RT,total_runtime/RTT,walltime/WT]) :-!,
   statistics(runtime,[RE,_]),
   statistics(total_runtime,[TE,_]),
   statistics(walltime,[WE,_]),
   RT is RE-R, RTT is TE-T, WT is WE-W.
stop_ms_timer(X,Y) :-
  add_internal_error('Illegal call: ', stop_ms_timer(X,Y)),
  Y = [].

% ---------------------------


retract_with_statistics(Module,ListOfFacts) :-
  nl,print('Retracting Facts'),nl,
  get_memory_used(M),
  print(' Memory usage: '),print_memory_used(M),nl,
  retract_with_statistics(Module,ListOfFacts,M).

retract_with_statistics(_Module,[],_) :- nl,nl.
retract_with_statistics(Module,[Fact|T],Mem) :- 
  print(Fact), print(' : '),
  retractall(Module:Fact),
  get_memory_used(NewMem),
  print_memory_used(NewMem),
  print_memory_used_difference(Mem,NewMem),nl,
  retract_with_statistics(Module,T,NewMem).
  
get_memory_used([M,PU]) :- garbage_collect,garbage_collect_atoms, get_memory_used_wo_gc([M,PU]).
get_memory_used_wo_gc([M,PU]) :- statistics(program,[PU,_]),statistics(memory_used,M).
print_memory_used_wo_gc :- get_memory_used_wo_gc(M), print_memory_used(M).
print_memory_used([M,PU]) :- print_mb(M), print(' ('), print_mb(PU), print(' program) ').  
print_memory_used_difference([M1,_PU1],[NewM2,_PU2]) :- Diff is (M1)-(NewM2),
   (Diff >= 0 -> print(' freed: '), print_bytes(Diff)
    ; print(' allocated: '), D2 is -(Diff), print_bytes(D2)),nl.
   %print(' / '),Diff2 is (PU1)-(PU2), print_mb(Diff2).
print_bytes(X) :- (X<50000 -> print_kb(X) ; print_mb(X)).
print_mb(X) :- XMB is X / 1048576,
  format(' ~3f MB',[XMB]).
print_kb(X) :- XKB is X / 1024,
  format(' ~3f KB',[XKB]).
  
:- op(1150,fx,space_call).
:- meta_predicate space_call(0).
space_call(Call) :- get_memory_used(M1),
    call(Call),
    get_memory_used(M2),
    print_memory_used(M2),
    print_memory_used_difference(M1,M2),nl.

% ---------------------------

read_string_from_file(Filename,String) :-
    open(Filename,read,S),
    read_string(S,String),
    close(S).

read_string(S,String) :-
    get_code(S,C),!,
    (C= -1
    -> String = []
    ;  String=[C|Rest],read_string(S,Rest)).

% ------------------------------


%%
% Escape is needed for &, \/, /\, ", ', etc.
% used mainly for dout output
% It seems dotty do not understand escapes, neither C style nor HTML style.
% Dot does render '&' correctly, no matter of escape.
% 

print_escaped(Atom) :- string_escape(Atom,E), print(E).

string_escape(Atom, EscapedAtom):-
    x_string_escape(Atom, EscapedAtom).

:- use_module(library(lists), [
    maplist/3, 
    scanlist/4]).

x_string_escape(Number, EscapedAtom):-
    number(Number),!, number_chars(Number,C), atom_chars(EscapedAtom,C). % to detect when label=value
x_string_escape(Atom, EscapedAtom):-
    atom(Atom),
    atom_chars(Atom, Chars),
    maplist(x_string_escape_map, Chars, EscapedChars),
    scanlist(x_atom_concat_rev, EscapedChars, '', EscapedAtom),!.
x_string_escape(Term, EscapedTerm):-
    Term =.. [Fkt|Args],
    x_string_escape(Fkt, EscapedFkt),
    maplist(x_string_escape, Args, EscapedArgs),!,
    EscapedTerm =.. [EscapedFkt|EscapedArgs].
x_string_escape(X,X).


x_atom_concat_rev(A,B,BA):-
    atom_concat(B,A,BA).


x_string_escape_map('\n', '\\n'). 
x_string_escape_map('\\', '\\\\').       
x_string_escape_map('"', '\\"').       
x_string_escape_map('\'', '\\\'').
x_string_escape_map('{', '\\{'). % important if used inside dot records
x_string_escape_map('}', '\\}'). % important if used inside dot records
x_string_escape_map('|', '\\|'). % important if used inside dot records
x_string_escape_map('>', '\\>'). % important if used inside dot records
%x_string_escape_map('&', '&amp;').  % michael: habe diese 3 Zeilen auskommentiert
%x_string_escape_map('<', '&lt;').
%x_string_escape_map('>', '&gt;').
% this is the SICSTus encoding for Unicode either \octal_number\  or \xHexNumber\
% See Section 4.1.7.6 Escape Sequences in SICStus Manual (page 60)
x_string_escape_map('\344\', '&auml;'). % a mit Umlaut
x_string_escape_map('\366\', '&ouml;'). % o mit Umlaut
x_string_escape_map('\374\', '&uuml;'). % u mit Umlaut
x_string_escape_map('\304\', '&Auml;'). % A mit Umlaut
x_string_escape_map('\326\', '&Ouml;'). % O mit Umlaut
x_string_escape_map('\334\', '&Uuml;'). % U mit Umlaut

x_string_escape_map('\353\', '&euml;'). % e mit Umlaut
x_string_escape_map('\313\', '&Euml;'). % E mit Umlaut

x_string_escape_map('\350\', '&egrave;'). % e mit Accent
x_string_escape_map('\351\', '&eacute;').
x_string_escape_map('\352\', '&ecirc;').
x_string_escape_map('\310\', '&Egrave;'). % E mit Accent
x_string_escape_map('\311\', '&Eacute;').
x_string_escape_map('\312\', '&Ecirc;').

x_string_escape_map('\340\', '&agrave;'). % a mit Accent
x_string_escape_map('\341\', '&aacute;').
x_string_escape_map('\342\', '&acirc;').
x_string_escape_map('\300\', '&Agrave;'). % A mit Accent
x_string_escape_map('\301\', '&Aacute;').
x_string_escape_map('\302\', '&Acirc;'). 

x_string_escape_map('\354\', '&igrave;'). % i mit Accent
x_string_escape_map('\355\', '&iacute;').
x_string_escape_map('\356\', '&icirc;').
x_string_escape_map('\314\', '&Igrave;'). % I mit Accent
x_string_escape_map('\315\', '&Iacute;').
x_string_escape_map('\316\', '&Icirc;').

x_string_escape_map('\362\', '&ograve;'). % o mit Accent
x_string_escape_map('\363\', '&oacute;').
x_string_escape_map('\364\', '&ocirc;').
x_string_escape_map('\322\', '&Ograve;'). % O mit Accent
x_string_escape_map('\323\', '&Oacute;').
x_string_escape_map('\324\', '&Ocirc;').

x_string_escape_map('\347\', '&ccedil;'). % cedille
x_string_escape_map('\307\', '&Ccedil;'). % Cedille

x_string_escape_map('\337\', '&szlig;'). % scharfes S (sz)
x_string_escape_map('\361\', '&ntilde;'). % n with tilde

x_string_escape_map('\1661\', '&alpha;'). % Greek
x_string_escape_map('\1662\', '&beta;').
x_string_escape_map('\1663\', '&gamma;').
x_string_escape_map('\1664\', '&delta;').
x_string_escape_map('\1665\', '&epsilon;').
x_string_escape_map('\1666\', '&zeta;'). 
x_string_escape_map('\1667\', '&eta;').
x_string_escape_map('\1670\', '&theta;').
x_string_escape_map('\1671\', '&iota;').
x_string_escape_map('\1672\', '&kappa;').
x_string_escape_map('\1673\', '&lambda;').
x_string_escape_map('\1674\', '&mu;').
x_string_escape_map('\1675\', '&nu;').
x_string_escape_map('\1676\', '&xi;').
x_string_escape_map('\1677\', '&omicron;').
x_string_escape_map('\1700\', '&pi;').
x_string_escape_map('\1701\', '&rho;').
x_string_escape_map('\1702\', '&sigmaf;').
x_string_escape_map('\1703\', '&sigma;').
x_string_escape_map('\1704\', '&tau;').
x_string_escape_map('\1705\', '&upsilon;').
x_string_escape_map('\1706\', '&phi;').
x_string_escape_map('\1707\', '&chi;').
x_string_escape_map('\1710\', '&psi;').
x_string_escape_map('\1711\', '&omega;'). 

% to do: add missing upper-case Greek letters:
x_string_escape_map('\1624\', '&Delta;').
x_string_escape_map('\1630\', '&Theta;').
x_string_escape_map('\1633\', '&Lambda;').
x_string_escape_map('\1636\', '&Xi;').
x_string_escape_map('\1645\', '&Upsilon;').
x_string_escape_map('\1647\', '&Chi;').
x_string_escape_map('\1650\', '&Psi;').
x_string_escape_map('\1651\', '&Omega;').

% translate unknown unicode chars to lozenge
x_string_escape_map(Unicode, '&loz;') :- atom(Unicode), atom_codes(Unicode,[Code]), Code>127.

x_string_escape_map(X, X).
 
%
% split_list(Pred,List,ListA,ListB):
%  List contains exactly the same elements as ListA and ListB
%  An element E is member of ListA iff Pred(E) is true and E is element of List.
%  An element E is member of ListB iff Pred(E) is false and E is element of List.
:- meta_predicate split_list(1,?,?,?).
:- meta_predicate split_list2(?,1,?,?).
split_list(Pred,List,A,B) :-
    split_list2(List,Pred,A,B).
split_list2([],_Pred,[],[]).
split_list2([Elem|Rest],Pred,A,B) :-
    ( call(Pred,Elem) -> A=[Elem|AR], B=BR
    ; otherwise       -> A=AR, B=[Elem|BR]),
    split_list2(Rest,Pred,AR,BR).



:- meta_predicate foldl(3,?,?,?).
just_for_unit_test_add(A,B,C) :- C is A+(10*B).
:- assert_must_succeed(( foldl(just_for_unit_test_add,[],0,R), R==0 )).
:- assert_must_succeed(( foldl(just_for_unit_test_add,[3,6,7,2,6],0,R), R==36726 )).

foldl(Module:Pred,List,Start,Result) :-
    foldl2(List,Module:Pred,Start,Result).
foldl2([],_Pred,Value,Value).
foldl2([Elem|Rest],Module:Pred,OldValue,NewValue) :-
    call(Module:Pred,Elem,OldValue,Value),
    foldl2(Rest,Module:Pred,Value,NewValue).


:- meta_predicate foldl(4,?,?,?,?).
just_for_unit_test_add2(A,B,C,D) :- D is 100*C+10*A+B.
:- assert_must_succeed(( foldl(just_for_unit_test_add2,[],[],0,R), R==0 )).
:- assert_must_succeed(( foldl(just_for_unit_test_add2,[3,6,7,2,6],[5,2,9,0,8],0,R), R==3562792068 )).

foldl(Module:Pred,List,List1,Start,Result) :-
    foldl2(List,Module:Pred,List1,Start,Result).
foldl2([],_Pred,[],Value,Value).
foldl2([Elem|Rest],Module:Pred,[H1|R1],OldValue,NewValue) :-
    call(Module:Pred,Elem,H1,OldValue,Value),
    foldl2(Rest,Module:Pred,R1,Value,NewValue).

:- meta_predicate foldl(5,?,?,?,?,?).
just_for_unit_test_add3(A,B,C,D,E) :- E is D*1000+100*A+10*B+C.
:- assert_must_succeed(( foldl(just_for_unit_test_add3,[],[],[],0,R), R==0 )).
:- assert_must_succeed(( foldl(just_for_unit_test_add3,[3,6,7],[5,2,9],[4,1,0],0,R),
                         R==354621790 )).

foldl(Module:Pred,A,B,C,Start,Result) :-
    foldl2(A,Module:Pred,B,C,Start,Result).
foldl2([],_Pred,[],[],Value,Value).
foldl2([Elem|Rest],Module:Pred,[H1|R1],[H2|R2],OldValue,NewValue) :-
    call(Module:Pred,Elem,H1,H2,OldValue,Value),
    foldl2(Rest,Module:Pred,R1,R2,Value,NewValue).

:- assert_must_succeed(( average([2,4,8,10],Avg), D is abs(Avg-6),D<0.001 )).
:- assert_must_succeed(( average([2],Avg), D is abs(Avg-2), D<0.001 )).
average(List,Avg) :-
    length(List,N),sumlist(List,Sum),Avg is Sum/N.

%:- use_module(custom_explicit_sets, [expand_custom_set_to_list_now/2]).
:- meta_predicate map_over_bvalue(2,?,?).
% dispatch. TODO: closure/3 and freetype
map_over_bvalue(_,[],[]) :- !.
map_over_bvalue(Pred,[H|T],[Res|Tail]) :- !,
    map_over_bvalue(Pred,H,Res),
    map_over_bvalue(Pred,T,Tail).
map_over_bvalue(Pred,(A,B),(ResA,ResB)) :- !,
    map_over_bvalue(Pred,A,ResA),
    map_over_bvalue(Pred,B,ResB).
map_over_bvalue(Pred,rec(L),rec(Res)) :- !,
    map_over_bvalue(Pred,L,Res).
map_over_bvalue(Pred,avl_set(A),Res) :- !,
    custom_explicit_sets:expand_custom_set_to_list_now(avl_set(A),L),
    map_over_bvalue(Pred,L,Res).
map_over_bvalue(Pred,global_set(A), Res) :- !,
    custom_explicit_sets:expand_custom_set_to_list_now(global_set(A), L),
    map_over_bvalue(Pred,L,Res).
map_over_bvalue(Pred,field(Name,Value),field(Name,TValue)) :- !,
    map_over_bvalue(Pred,Value,TValue).
% apply
map_over_bvalue(Pred,BValue,OutValue) :- !,
    call(Pred,BValue,OutValue).
        

% assert_once works like assert, but checks if the fact has already been stored before and
% will not store it a second time
:- meta_predicate assert_once(0).
assert_once(Module:Predicate) :-
    ( call(Module:Predicate) -> true
    ; otherwise -> assert(Module:Predicate)).
    
:- use_module(library(timeout),[time_out/3]).
:- meta_predicate safe_time_out(0,*,*).
safe_time_out(Call,TO,Res) :- 
   (TO > 2147483646 -> print('### Warning: Time-Out value too high (>2147483646): '), print(TO),nl,
                       print('### Calling goal without time-out'),nl,
                       call(Call), Res=success
    ; TO < 1 -> print('### Warning: Time-Out value too small: '), print(TO),nl,
                time_out(Call,1,Res)
    ; time_out(Call,TO,Res)).

:- dynamic id_counter/1.
unique_id(Prefix,Id) :-
    ( id_counter(V) -> retractall(id_counter(_))
    ; otherwise     -> V is 0),
    N is V+1,
    assert( id_counter(N) ),
    number_codes(N,NCodes),
    append(Prefix,NCodes,ICodes),
    atom_codes(Id,ICodes).

% PROBPATH is the runtime search path for defintion files not found relative to the original machine.
% By default this is the stdlib directory relative to the prob base dir.
% The user can provide a list of : separated directories in the PROBPATH
% environment variables which are prepended to the default path.
get_PROBPATH(PROBPATH) :-
    environ('PROBPATH', CustomPATH),
    atom_length(CustomPATH, L),
    L > 0, !,
    get_stdlib_path(STDLIB), ajoin([CustomPATH, ':', STDLIB], PROBPATH).

get_PROBPATH(PROBPATH) :- get_stdlib_path(PROBPATH).

get_stdlib_path(STDLIB) :- get_base_dir(Base), atom_concat(Base, 'stdlib', STDLIB).

get_base_dir(Base) :-
    prolog_flag('system_type', runtime), !,
    environ('SP_APP_DIR', Base).
get_base_dir(Base) :-
    prolog_flag('system_type', development), !,
    environ('SP_STARTUP_DIR', Base).
get_base_dir('.') :-
    prolog_flag('system_type', TP),
    print('### Warning: Unexpected sytem_type'), print(TP),nl.
